#include "stdafx.h"
#include "DateExtractor.h"
#include <string>


class TCmpExtractedDate {
private:
  bool IsAsc;
public:
	TCmpExtractedDate(const bool& AscSort=true) : IsAsc(AscSort) { }
	bool operator () (const TExtractedDate& P1, const TExtractedDate& P2) const 
	{
		if (IsAsc) { return P1.Freq < P2.Freq; } else { return P2.Freq < P1.Freq; }
	}
};

// //////
// TExtractedDate
TExtractedDate::TExtractedDate(PJsonVal DateJson)
{
	Ambiguous = DateJson->GetObjBool("Amb", true);
	Freq = DateJson->IsObjKey("Freq") ? DateJson->GetObjNum("Freq", 1) : 1;
	DetectedDate = DateJson->IsObjKey("DetectedDate") ? DateJson->GetObjStr("DetectedDate", "") : "";
	Date = DateJson->GetObjStr("Date");
	BIsDateRange = DateJson->IsObjKey("DateEnd");
	if (BIsDateRange)
		DateEnd = DateJson->GetObjStr("DateEnd");
}

PJsonVal TExtractedDate::GetJson(const bool& IncludeDetectedDateStr) 
{
	PJsonVal DateJson = TJsonVal::NewObj("Amb", Ambiguous);
	if (Freq > 1)
		DateJson->AddToObj("Freq", Freq);
	DateJson->AddToObj("Date", Date);
	// date range props
	if (BIsDateRange)
		DateJson->AddToObj("DateEnd", DateEnd);
	if (IncludeDetectedDateStr) 
		DateJson->AddToObj("DetectedDate", DetectedDate);
	return DateJson;
}


// //////
// TExtractedDateColl
TExtractedDateColl::TExtractedDateColl(const TStr& EncodedDateColl)
{
	PJsonVal ArrayJson = TJsonVal::GetValFromStr(EncodedDateColl);
	if (!ArrayJson->IsArr())
		return;
	int Len = ArrayJson->GetArrVals();
	for (int N=0; N < Len; N++) {
		PJsonVal DateJson = ArrayJson->GetArrVal(N);
		ExtractedDateV.Add(TExtractedDate(DateJson));
	}
}

void TExtractedDateColl::AddExtractedDate(const TExtractedDate& Date)
{
	for (int N=0; N < ExtractedDateV.Len(); N++) {
		if (Date.Date == ExtractedDateV[N].Date)  {
			//  this date is already present. just increase the frequency
			ExtractedDateV[N].IncreaseFreq(Date.Freq);
			return;
		}
	}
	// first time we see the date. add it to the vector
	ExtractedDateV.Add(Date);
}

void TExtractedDateColl::AddColl(const TExtractedDateColl& Coll)
{
	int Dates = Coll.Len();
	for (int N=0; N < Dates; N++) {
		AddExtractedDate(Coll.GetExtractedDate(N));
	}
}

// some TExtractedDate instances can have a missing year (e.g. -MM-DD).
// ImputeMissingYear checks if we can find other instances of TEXtractedDate that have YYYY-MM-DD set (with same MM-DD part)
// if found, then Freq is moved from -MM-DD to YYYY-MM-DD and the -MM-DD instance is removed
// if multiple YYYY-MM-DD are found with the same -MM-DD, the one with the largest Freq is used as the target for merging
// imputing works for TExtractedDate that represent single date as well as date range
void TExtractedDateColl::ImputeMissingYears(PNotify Notify)
{
	// merge dates
	int i=0; 
	while (i < ExtractedDateV.Len()) {
		if (ExtractedDateV[i].IsDateRange()) { 
			i++; continue; 
		} 
		// we don't try to merge the dates with known year
		if (ExtractedDateV[i].Date.GetCh(0) != '-') { 
			i++; continue; 
		}
		int BestMatchInd = -1;
		// find candidates with valid year
		for (int j=0; j < ExtractedDateV.Len(); j++) {
			if (ExtractedDateV[j].IsDateRange()) continue;
			if (i==j) continue;
			// if the MM-DD part is the same then remember it as BestMatchInd
			if (ExtractedDateV[j].Date.GetCh(0) != '-' &&							 // has to have valid year
				ExtractedDateV[j].Date.SearchStr(ExtractedDateV[i].Date) != -1) {    // has to contain the same MM-DD part
				if (BestMatchInd == -1 || ExtractedDateV[j].Freq > ExtractedDateV[BestMatchInd].Freq)
					BestMatchInd = j;
			}
		}
		if (BestMatchInd != -1) {
			if (!Notify.Empty())
				Notify->OnNotifyFmt(ntInfo, "Imputing %s (%d articles) into %s (%d articles)", ExtractedDateV[i].Date.CStr(), ExtractedDateV[i].Freq, ExtractedDateV[BestMatchInd].Date.CStr(), ExtractedDateV[BestMatchInd].Freq);
			ExtractedDateV[BestMatchInd].Freq += ExtractedDateV[i].Freq;
			ExtractedDateV.Del(i);
		}
		else
			i++;
	}
	// merge date ranges
	i=0; 
	while (i < ExtractedDateV.Len()) {
		if (! ExtractedDateV[i].IsDateRange()) { 
			i++; continue; 
		} 
		// we don't try to merge the dates with known year
		if (ExtractedDateV[i].Date.GetCh(0) != '-' && ExtractedDateV[i].DateEnd.GetCh(0) != '-') { 
			i++; continue; 
		} 
		int BestMatchInd = -1;
		// find candidates with valid year
		for (int j=0; j < ExtractedDateV.Len(); j++) {
			if (! ExtractedDateV[j].IsDateRange()) continue;
			if (i==j) continue;
			// if the MM-DD part is the same then remember it as BestMatchInd
			if (ExtractedDateV[j].Date.GetCh(0) != '-' &&	ExtractedDateV[j].DateEnd.GetCh(0) != '-' &&  // has to have valid year
				ExtractedDateV[j].Date.SearchStr(ExtractedDateV[i].Date) != -1 && 
				ExtractedDateV[j].DateEnd.SearchStr(ExtractedDateV[i].DateEnd) != -1) {    // has to contain the same MM-DD part
				if (BestMatchInd == -1 || ExtractedDateV[j].Freq > ExtractedDateV[BestMatchInd].Freq)
					BestMatchInd = j;
			}
		}
		if (BestMatchInd != -1) {
			if (!Notify.Empty())
				Notify->OnStatusFmt("Imputing (%s, %s) (%d articles) into (%s, %s) (%d articles)", ExtractedDateV[i].Date.CStr(), ExtractedDateV[i].DateEnd.CStr(), ExtractedDateV[i].Freq, ExtractedDateV[BestMatchInd].Date.CStr(), ExtractedDateV[BestMatchInd].DateEnd.CStr(), ExtractedDateV[BestMatchInd].Freq);
			ExtractedDateV[BestMatchInd].Freq += ExtractedDateV[i].Freq;
			ExtractedDateV.Del(i);
		}
		else
			i++;
	}
}


void TExtractedDateColl::SortByFreq(const bool& AscSort)
{
	ExtractedDateV.SortCmp(TCmpExtractedDate(AscSort));
}

TStr TExtractedDateColl::EncodeExtractedDates(const int& MaxCount)
{
	if (ExtractedDateV.Len() == 0)
		return "";
	return TJsonVal::GetStrFromVal(EncodeExtractedDatesAsJson(MaxCount));
}

PJsonVal TExtractedDateColl::EncodeExtractedDatesAsJson(const int& MaxCount)
{
	if (ExtractedDateV.Len() == 0)
		return TJsonVal::NewArr();
	PJsonVal DateArr = TJsonVal::NewArr();
	int Count = (MaxCount == -1) ? ExtractedDateV.Len() : min(ExtractedDateV.Len(), MaxCount);
	for (int N=0; N < Count; N++) {
		TExtractedDate Date = ExtractedDateV[N];
		PJsonVal DateJson = Date.GetJson();
		DateArr->AddToArr(DateJson);
	}
	return DateArr;
}

// //////
// TDateExtractor
TDateExtractor::TDateExtractor(void)
{
	AddRegExDateRange("eng", "\\b(?P<month1>jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[a-z]*\\.?\\s+(?P<day1>\\d{1,2})\\.?(?:\\s+to\\s+|\\s*-\\s*|\\s+and\\s+)(?P<day2>\\d{1,2})\\.?(?:[\\s*,]+(?P<year1>\\d{2}|\\d{4}))?\\b");
	// date range jan 1. (to|through|and) feb 4 2010
	AddRegExDateRange("eng", "\\b(?P<month1>jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[a-z]*\\.?\\s+(?P<day1>\\d{1,2})\\.?(?:\\s+(to|through|and)\\s+|\\s*-\\s*)(?P<month2>jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[a-z]*\\.?\\s+(?P<day2>\\d{1,2})\\.?(?:[\\s*,]+(?P<year1>\\d{2}|\\d{4}))?\\b");
		
	//
	// recognize individual date occurences

	// recognize: 2013-3-14 | 2011/3/14 | 
	// don't match: 2013-3/14
	// find all numeric occurences of dates: dddd[-/ ]dd[.-/ ]dd[-/ ]. 
	AddRegExDate("eng", "\\b(?P<year>\\d{4})(?:\\s*[-\\/]\\s*)(?P<month>\\d{1,2})(?:\\s*[-\\/]\\s*)(?P<day>\\d{1,2})\\b");
	// find all numeric occurences of dates: dd[.-/ ]dd[.-/ ](dddd)?. 
	// recognize: 3/14/2011 | 3/14/11 | 03/14/11 | 14/03/2013 | 14/03/11 | 03/14/11 
	// ignore: 3/14 | 3.14 |	
	AddRegExDate("eng", "\\b(?P<day>\\d{1,2})(?:\\s*[-\\/\\.]\\s*)(?P<month>\\d{1,2})(?:\\s*[-\\/\\.]\\s*(?P<year>\\d{2}|\\d{4}))\\b");
	// recognize: 14. 1. | 14.3. | 13.2.
	// ignore: 13.3 | 13/3/2013 | 14. 1. 2013 (covered before)
	AddRegExDate("eng", "\\b(?P<day>\\d{1,2})\\.\\s?(?P<month>\\d{1,2})\\.");
	// find all numeric occurences of dates: day month(in words) (dddd)?
	// match: 22nd September ,2006 | 22nd September 2006 | 22 September 2006 | 2nd February 2007 |  18 October 
	AddRegExDate("eng", "\\b(?P<day>\\d{1,2})(?:\\s*(\\.|st|nd|rd|th)?\\s*)(?P<month>jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[a-z]*\\.?\\s*,?\\s*(?P<year>\\d{2}|\\d{4})?\\b");
	// find all occurences of dates: month(in words) day (dddd)?
	AddRegExDate("eng", "\\b(?P<month>jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[a-z]*\\.?\\s*(the|-)?\\s*(?P<day>\\d{1,2})(?:\\.|st|nd|rd|th)?(?:[\\s*,]+(?P<year>\\d{2}|\\d{4}))?\\b");
	// find all occurences of dates: month(in words) dddd
	// here we don't want to match Jan. 12 (which would mean Jan 2012) - year has to be 4 digits
	AddRegExDate("eng", "(?P<day>)\\b(?P<month>jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[a-z]*\\.?\\s*(?P<year>\\d{4})\\b");
}


TDateExtractor::~TDateExtractor(void)
{
}

void TDateExtractor::SetRegEx(const TStr& TextLang, const TStrV& DateRangeV, const TStrV DateV)
{
	DateH.Clr();
	DateRangeH.Clr();
	for (int N=0; N < DateRangeV.Len(); N++) 
		AddRegExDateRange(TextLang, DateRangeV[N]);
	for (int N=0; N < DateV.Len(); N++) 
		AddRegExDate(TextLang, DateV[N]);
}

void TDateExtractor::AddRegExDate(const TStr& TextLang, const TStr& Date)
{
	if (!DateH.IsKey(TextLang))
		DateH.AddDat(TextLang, TStrV());
	DateH.GetDat(TextLang).Add(Date);
}

void TDateExtractor::AddRegExDateRange(const TStr& TextLang, const TStr& DateRange)
{
	if (!DateRangeH.IsKey(TextLang))
		DateRangeH.AddDat(TextLang, TStrV());
	DateRangeH.GetDat(TextLang).Add(DateRange);
}

void TDateExtractor::ExtractDates(const TStr& Text, const TStr& TextLang, TExtractedDateColl& ExtractedDateColl)
{
	TVec<TIntPr> SeenMatchesV;
	const StringPiece SPText(Text.CStr());
	RE2::Options ReOptions;
	ReOptions.set_case_sensitive(false);
	ReOptions.set_utf8(false);

	map<string, int> Groups;
	StringPiece Matches[10];

	// extract date ranges
	TStrV DateRangeV = DateRangeH.IsKey(TextLang) ? DateRangeH.GetDat(TextLang) : TStrV();
	for (int N = 0; N < DateRangeV.Len(); N++)
		AddDateRanges(SPText, DateRangeV[N].CStr(), ReOptions, TextLang, ExtractedDateColl, SeenMatchesV);

	TStrV DateV = DateH.IsKey(TextLang) ? DateH.GetDat(TextLang) : TStrV();
	for (int N = 0; N < DateV.Len(); N++)
		AddIndividualDates(SPText, DateV[N].CStr(), ReOptions, TextLang, ExtractedDateColl, SeenMatchesV);
		
}


void TDateExtractor::AddIndividualDates(const StringPiece SPText, const char* RegEx, const RE2::Options& ReOptions, const TStr& TextLang, TExtractedDateColl& ExtractedDateColl, TVec<TIntPr>& SeenMatchesV)
{
	int PrintOffset = 20;
	
	StringPiece Matches[10];

	RE2 Re(RegEx, ReOptions);
	map<string, int> Groups = Re.NamedCapturingGroups();
	
	int off = 0;
	while(Re.Match(SPText, off, SPText.size(), RE2::UNANCHORED, Matches, 10)) {
		off = Matches[0].end() - SPText.begin();
		// make sure we haven't match this date already with some previous regex
		// regex expressions are ordered from most to least specific so we don't want to match multiple times cases such as 15 August 2013
		int MatchStart = (int) (Matches[0].begin() - SPText.begin());
		int MatchEnd = (int) (Matches[0].end() - SPText.begin());
		bool alreadySeen = false;
		for (int SeenN = 0; SeenN < SeenMatchesV.Len(); SeenN++) {
			if (SeenMatchesV[SeenN].Val1 <= MatchStart && MatchStart <= SeenMatchesV[SeenN].Val2) {
				NotifyInfo("skipping match '%s' since already seen", Matches[0].ToString().c_str());
				alreadySeen = true;
				break;
			}
		}
		if (alreadySeen)
			continue;
		TStr Day = Groups.count("day") > 0 ? Matches[Groups["day"]].ToString().c_str() : "";
		TStr Month = Groups.count("year") > 0 ? Matches[Groups["month"]].ToString().c_str() : "";
		TStr Year = Groups.count("year") > 0 ? Matches[Groups["year"]].ToString().c_str() : "";
		bool Ambiguous = false;
	
		if (Day != "" && !Day.IsInt() && !Month.IsInt()) 
			continue;
		if (!Day.IsInt() && Month.IsInt()) { TStr Tmp = Day; Day = Month; Month = Tmp; }
		// check if we have valid values for month and day and if we need to switch month and day due to their values
		if (Month.IsInt() && Day.IsInt()) {
			int M = Month.GetInt(-1);
			int D = Day.GetInt(-1);
			Ambiguous = D <= 12 && M <= 12;
			if (M > 12 && D <= 12) { 
				int T = D; D = M; M = T; 
				TStr Tmp = Day; Day =Month; Month = Tmp;
			}
		}
		if (!Month.IsInt())
			Month = GetMonthAsInt(TextLang, Month);
	
		// we require at least two parts of the date to be valid
		if (Day == "" && Year == "")
			continue;
		// we treat as valid values only "" and range (-2000, 3000)
		if (Year != "" && (!Year.IsInt() || Year.GetInt(0) < -2000 || Year.GetInt(0) > 3000)) {
			NotifyInfo("skipping invalid date match '%s'", Matches[0].ToString().c_str());
			continue;
		}
		if (Day != "" && (!Day.IsInt() || Day.GetInt(0) > 31 || Day.GetInt(0) < 1)) {
			NotifyInfo("skipping invalid date match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for day found - exit
		}
		if (!Month.IsInt() || Month.GetInt(0) < 1 || Month.GetInt(0) > 12) {
			NotifyInfo("skipping invalid date match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for month found - exit
		}

		int PrintOffset = 20;
		int StartPrint = max(0, MatchStart-PrintOffset);
		int EndPrint = min(SPText.length() - StartPrint, MatchEnd-MatchStart + 2*PrintOffset);
		TStr PrintOrigContext = SPText.substr(StartPrint, EndPrint).ToString().c_str();
		PrintOrigContext.ChangeStrAll("\n", " ");
		TStr OriginalDate = Matches[0].ToString().c_str();
		if (Year.IsInt() && Year.Len() == 2) 
			Year = (Year.GetInt(0) < 50 ? "20" : "19") + Year;
		if (Day != "" && Day.Len() < 2)
			Day = "0" + Day;
		TStr NormalizedDate = TStr::Fmt("%s-%02d-%s", Year.CStr(), Month.GetInt(0), Day.CStr());
		NotifyInfo("found date: %s in text ...%s...", NormalizedDate.CStr(), PrintOrigContext.CStr());
	
		ExtractedDateColl.AddExtractedDate(TExtractedDate(NormalizedDate, OriginalDate, Ambiguous));
		SeenMatchesV.Add(TIntPr(MatchStart, MatchEnd));
	}
}


void TDateExtractor::AddDateRanges(const StringPiece SPText, const char* RegEx, const RE2::Options& ReOptions, const TStr& TextLang, TExtractedDateColl& ExtractedDateColl, TVec<TIntPr>& SeenMatchesV)
{
	int PrintOffset = 20;
	
	StringPiece Matches[10];

	RE2 Re(RegEx, ReOptions);
	map<string, int> Groups = Re.NamedCapturingGroups();
	int off = 0;
	while(Re.Match(SPText, off, SPText.size(), RE2::UNANCHORED, Matches, 10)) {
		off = Matches[0].end() - SPText.begin();
		int MatchStart = (int) (Matches[0].begin() - SPText.begin());
		int MatchEnd = (int) (Matches[0].end() - SPText.begin());
		bool alreadySeen = false;
		for (int SeenN = 0; SeenN < SeenMatchesV.Len(); SeenN++) {
			if (SeenMatchesV[SeenN].Val1 <= MatchStart && MatchStart <= SeenMatchesV[SeenN].Val2) {
				NotifyInfo("skipping match '%s' since already seen", Matches[0].ToString().c_str());
				alreadySeen = true;
				break;
			}
		}
		if (alreadySeen)
			continue;
		TStr Day1 = Groups.count("day1") > 0 ? Matches[Groups["day1"]].ToString().c_str() : "";
		TStr Day2 = Groups.count("day2") > 0 ? Matches[Groups["day2"]].ToString().c_str() : "";
		TStr Month1 = Matches[Groups["month1"]].ToString().c_str();
		TStr Month2 = Groups.count("month2") > 0 ? Matches[Groups["month2"]].ToString().c_str() : Month1;
		TStr Year1 = Groups.count("year1") > 0 ? Matches[Groups["year1"]].ToString().c_str() : "";
		TStr Year2 = Groups.count("year2") > 0 ? Matches[Groups["year2"]].ToString().c_str() : "";
		if (Year1 == "")
			Year1 = Year2;
		if (Year2 == "")
			Year2 = Year1;
	
		// we require at least two parts of the date to be valid
		if ( (Day1 == "" && Year1 == "") || (Day2 == "" && Year2 == "") )
			continue;
		if (Day1 != "" && !Day1.IsInt() && !Month1.IsInt()) 
			continue;
		if (Day2 != "" && !Day2.IsInt() && !Month2.IsInt()) 
			continue;
		if (!Month1.IsInt())
			Month1 = GetMonthAsInt(TextLang, Month1);
		if (!Month2.IsInt())
			Month2 = GetMonthAsInt(TextLang, Month2);
	
		// we treat as valid values only "" and range (-2000, 3000)
		if (Year1 != "" && (!Year1.IsInt() || Year1.GetInt(0) < -2000 || Year1.GetInt(0) > 3000)) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;
		}
		if (Year2 != "" && (!Year2.IsInt() || Year2.GetInt(0) < -2000 || Year2.GetInt(0) > 3000)) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;
		}
		if (Day1 != "" && (!Day1.IsInt() || Day1.GetInt(0) > 31 || Day1.GetInt(0) < 1) ) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for day found - exit
		}
		if (Day2 != "" && (!Day2.IsInt() || Day2.GetInt(0) > 31 || Day2.GetInt(0) < 1) ) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for day found - exit
		}
		if (!Month1.IsInt() || Month1.GetInt(0) < 1 || Month1.GetInt(0) > 12) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for month found - exit
		}
		if (!Month2.IsInt() || Month2.GetInt(0) < 1 || Month2.GetInt(0) > 12) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for month found - exit
		}

		int StartPrint = max(0, MatchStart-PrintOffset);
		int EndPrint = min(SPText.length() - StartPrint, MatchEnd-MatchStart + 2*PrintOffset);
		TStr PrintOrigContext = SPText.substr(StartPrint, EndPrint).ToString().c_str();
		PrintOrigContext.ChangeStrAll("\n", " ");
		TStr OriginalDate = Matches[0].ToString().c_str();
		if (Year1.IsInt() && Year1.Len() == 2) 
			Year1 = (Year1.GetInt(0) < 50 ? "20" : "19") + Year1;
		if (Year2.IsInt() && Year2.Len() == 2) 
			Year2 = (Year2.GetInt(0) < 50 ? "20" : "19") + Year2;
		if (Day1 != "" && Day1.Len() < 2)
			Day1 = "0" + Day1;
		if (Day2 != "" && Day2.Len() < 2)
			Day2 = "0" + Day2;
		TStr NormalizedDateStart = TStr::Fmt("%s-%02d-%0s", Year1.CStr(), Month1.GetInt(0), Day1.CStr());
		TStr NormalizedDateEnd = TStr::Fmt("%s-%02d-%0s", Year2.CStr(), Month2.GetInt(0), Day2.CStr());
		NotifyInfo("found date range: (%s, %s) in text ...%s...", NormalizedDateStart.CStr(), NormalizedDateEnd.CStr(), PrintOrigContext.CStr());
		ExtractedDateColl.AddExtractedDate(TExtractedDate(NormalizedDateStart, NormalizedDateEnd, OriginalDate, false));
		SeenMatchesV.Add(TIntPr(MatchStart, MatchEnd));
	}

}

TStr TDateExtractor::GetMonthAsInt(const TStr& TextLang, const TStr& Text)
{
	if (Text == "")
		return "";
	const TStr Month = Text.GetLc();
	if (TextLang == "eng") {
		if (Month == "jan" || Month == "january") return "1";
		else if (Month == "feb" || Month == "february" ) return "2";
		else if (Month == "mar" || Month == "march") return "3";
		else if (Month == "apr" || Month == "april") return "4";
		else if (Month == "may") return "5";
		else if (Month == "jun" || Month == "june") return "6";
		else if (Month == "jul" || Month == "july") return "7";
		else if (Month == "aug" || Month == "august") return "8";
		else if (Month == "sep" || Month == "sept" || Month == "september") return "9";
		else if (Month == "oct" || Month == "october") return "10";
		else if (Month == "nov" || Month == "november") return "11";
		else if (Month == "dec" || Month == "december") return "12";
		else return ""; 	// invalid value for month found - exit
	}
	else if (TextLang == "spa") {
		return "";
	}
	else if (TextLang == "deu") {
		return "";
	}
	else if (TextLang == "zho") {
		return "";
	}
	else if (TextLang == "slv") {
		return "";
	}
	else return "";
}

void TDateExtractor::Test()
{
	TStr EngText =  "Jan. 11-15 | Feb 4. - 12. | "
					"Jan. 11- feb 15 | May 4 - Jun 12. | "
					"3/14 | 3/14/2011 | 3/14/11 | 03/14/11 | 14/03/2013 | 14/03/11 | 03/14/11 | "
					"14. 1. | 14.3. | 13.2. | "
					"2013-3-14 | 2011/3/14 | "
					"22nd September ,2006 | 22nd September 2006 | 22 September 2006 | 2nd February 2007 |  18 October | 18 October, 2009 | 14 March 2011 | 18th November, 2009 (OR 18 October, 2009) | 4th October, 2005 (OR 4 October, 2005) | "
					"March 14, 2011 |  March 14th | March 14 | September 22 2006 (without mma) | August 15, 1990. | September 22, 2006 | September 22nd, 2006 | September 22, 2006 | March 14th, 2011 | April-1 | "
					"Jan. 2012 | Feb 2013 | "
				"March the Fourteenth | A the Fourteenth of March, 2011	March the Fourteenth, 2011 | B	14th March 2011 | "
				"C	 | D	14/3/2011	3/14/2011 | E	14/3/11	3/14/11 | "
				" |  30.3.2005 | 19/2/2002 | 30-2-1999 | 26 May 1979  | 30 October 2003 | 18 October, 2003 | 29 January, 2008 | My son was born on 18 October, 2003. April-1";
	TExtractedDateColl Dates;
	ExtractDates(EngText, "eng", Dates);
}

void TDateExtractor::RETest()
{
	int i;
	std::string s;
	bool ok = RE2::FullMatch("ruby:1234", "(\\w+):(\\d+)", &s, &i);
	ok = (s == "ruby");
	ok = (i == 1234);

	RE2 re("(?P<N1>is) (?P<N2>\\d+)|(?P<N3>car)|(object.*movie)");
    re2::StringPiece sp = "car is 12 the animation creation with the object as cars, its a very fine movie";
    int off = 0;
    re2::StringPiece matches[5];
    while(re.Match(sp, off, sp.size(), RE2::UNANCHORED, matches, 5)) {
		map<string, int> map = re.NamedCapturingGroups();
		int ind = map["N3"];
		re2::StringPiece carMatch = matches[ind];
		re2::StringPiece wholeMatch = matches[0];
        printf("match at %d,%d: %s\n", (int)(wholeMatch.begin() - sp.begin()), (int)(wholeMatch.end() - sp.begin()), wholeMatch.as_string().c_str());
        off = wholeMatch.end() - sp.begin();
    }

	//RE2::Arg Month, Day;
	std::string Month, Day, X;
	//RE2::Arg Month, Day, X;
	ok = RE2::PartialMatch("test March 14 20", "(Jan|(March|April)) (\\d+)(th)?", &Month, &Day, &X);

	// Fails: "ruby" cannot be parsed as an integer.
	ok = !RE2::FullMatch("ruby", "(.+)", &i);

	// Success; does not extract the number.
	ok = RE2::FullMatch("ruby:1234", "(\\w+):(\\d+)", &s);

	// Success; skips NULL argument.
	ok = RE2::FullMatch("ruby:1234", "(\\w+):(\\d+)", (void*)NULL, &i);

	// Fails: integer overflow keeps value from being stored in i.
	ok = !RE2::FullMatch("ruby:123456789123", "(\\w+):(\\d+)", &s, &i);

	//std::string s ("Bojan Jan., 10th has a submarine as a subsequence");
	//std::smatch m;
	//std::regex monthsLong("(\\b)(january|february|march|april|may|june|july|august|september|october|november|december)(\\b)",  std::regex_constants::ECMAScript | std::regex_constants::icase );
	//std::regex monthsShort("(\\b)(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[.]?(\\b)",  std::regex_constants::ECMAScript | std::regex_constants::icase );
	//std::regex months("(\\b)((january|february|march|april|may|june|july|august|september|october|november|december)|(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[.]?)(\\b)",  std::regex_constants::ECMAScript | std::regex_constants::icase );

	//std::cout << "Target sequence: " << s << std::endl;
	//std::cout << "The following matches and submatches were found:" << std::endl;

	//while (std::regex_search (s, m, months)) {
	//	int len = m.length();
	//	std::string x = m[1];
	//	/*for (auto x:m) 
	//		std::cout << x << " ";*/
	//	std::cout << std::endl;
	//	s = m.suffix().str();
	//}
}

// notifications
void TDateExtractor::NotifyInfo(const char *FmtStr, ...) 
{
	va_list valist; va_start(valist, FmtStr);	
	Notify(ntInfo, FmtStr, valist); va_end(valist); 
}
void TDateExtractor::NotifyErr(const char *FmtStr, ...) 
{
	va_list valist; va_start(valist, FmtStr);
	Notify(ntErr, FmtStr, valist); va_end(valist); 
}

void TDateExtractor::Notify(const TNotifyType& Type, const char *FmtStr, va_list argptr) 
{
	const int RetVal=vsnprintf(NotifyBuf, 10*1024-2, FmtStr, argptr);
	if (RetVal < 0) return;
	if (! NotifyInst.Empty())
		NotifyInst->OnNotify(Type, NotifyBuf); 
}