
function TopConceptList(options) {
    this.options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        showScores: true,           // show the numbers on the rects
        maxScale: undefined,        // max value for concept. if set then values will be scaled appropriate to it, if not then max value will take the 100% and others appropriatelly less
        decimals: 0,                // num of decimals to show for scores
        barHeight: 20,             // height of the bars
        gap: 2,                      // gap between the bars
        showXAxis: true,
        rectClsNm: "fill-blue",
        scoreClsNm: "fill-white"
    }, options);

    this.conceptData = undefined;
    this.labelNm = "label_eng";
    this.colors = d3.scale.category20();
        
    this.setData = function (conceptData, lang) {
        this.conceptData = conceptData;
        if (lang)
            this.setLang(lang);
        this.redraw();
    }
    
    this.setLang = function(lang) {
        this.labelNm = "label_" + lang;
    }
    
    this.redraw = function () {
        var self = this;
        var opt = this.options;
        
        // clear old data
        d3.select(opt.containerNm + " svg").remove();
        
        if (!this.conceptData || this.conceptData.length == 0)
            return;

        // make labels shorter if necessary
        makeNiceConceptLabels(this.conceptData, this.labelNm);
        
        var maxLabelWidth = d3.max($.map(this.conceptData, function (item, i) {
            return measureTextWidth(item.shortLabel, opt.containerNm + " svg", "name");
        }));
        var graphWidth = $(opt.containerNm).width() - 10,
            xAxisOffset = opt.showXAxis ? 30 : 0,
            conceptsHeight = (opt.barHeight + 2 * opt.gap) * this.conceptData.length,
            conceptLabelOffset = maxLabelWidth + 10;

        var chart = d3.select(opt.containerNm)
            .append('svg')
            .classed('chart', true)
            .attr('width', graphWidth + 10)
            .attr('height', conceptsHeight + xAxisOffset);

        // what should be the maximum value of the concept
        var maxXVal = opt.maxScale == undefined ? d3.max(this.conceptData, function (item) { return item.freq; }) : opt.maxScale;
        var x = d3.scale.linear()
            .domain([0, maxXVal])
            .range([0, graphWidth - conceptLabelOffset]);

        var y = d3.scale.ordinal()
            .domain(d3.range(0, this.conceptData.length))
            .rangeBands([0, conceptsHeight]);

        // show x axis with scale
        if (opt.showXAxis) {
            var xAxis = d3.svg.axis()
                .scale(x)
                .orient("top")
                .tickPadding(3);
            chart.append("g")
                .attr("class", "x axis")
                .attr("transform", "translate(" + conceptLabelOffset + ",27)")
                .call(xAxis);
            chart.append("text")
                .classed("axis-label", true)
                .text("Concept relevance")
                .attr("text-anchor", "end")
                .attr("x", function (val) { return conceptLabelOffset - 10; })
                .attr("y", "18px");
        }

        // rect
        chart.selectAll("rect.rect")
            .data(self.conceptData)
            .enter()
                .append("rect")
                .attr("class", opt.rectClsNm)
                .attr("x", conceptLabelOffset)
                .attr("y", function (val, ind) { return xAxisOffset + y(ind) + opt.gap; })
                .attr("height", opt.barHeight)
                .attr("width", function (val) { return 0; })
                .transition()
                .duration(1000)
                .attr("width", function (val) {
                    return d3.max([x(val.freq), 0]);
                });

        // scores
        if (opt.showScores) {
            chart.selectAll("text.score")
                .data(self.conceptData)
                .enter()
                    .append("text")
                    .attr("class", opt.scoreClsNm)
                    .attr("x", function (val) { return conceptLabelOffset; })
                    .attr("y", function (val, ind) { return xAxisOffset + y(ind) + (y.rangeBand() / 2); })
                    .attr("dx", -5)
                    .attr("dy", ".36em")
                    .attr("text-anchor", "end")
                    .text(function(val) { return val.freq.toFixed(opt.decimals); })
                    .transition()
                    .duration(1000)
                    .attr("x", function (val) {
                        return x(val.freq) + conceptLabelOffset;
                    });
        }

        // labels
        chart.selectAll("text.name")
            .data(self.conceptData)
            .enter()
                .append("text")
                .attr('class', function (val, ind) { return 'name concept ' + val.type;  } )
                .attr("x", conceptLabelOffset - 10)
                .attr("y", function (val, ind) { return xAxisOffset + y.rangeBand() / 2 + y(ind); })
                .attr("dy", ".36em")
                .attr("text-anchor", "end")
                .text(function (val) { return val.shortLabel; });
    }
}

function DiffConceptList(options) {
    this.options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        maxScale: undefined,        // max value for concept. if set then values will be scaled appropriate to it, if not then max value will take the 100% and others appropriatelly less
        decimals: 0,                // num of decimals to show for scores
        barHeight: 8,             // height of the bars
        bigGap: 4,                  // gap between the concepts
        smallGap: 1,                 // gap between the group0 and group1 of the same concept
        showXAxis: true,
        rectClsNm0: "fill-blue2",
        rectClsNm1: "fill-green",
        scoreClsNm: "fill-white"
    }, options);

    this.conceptData = undefined;
    this.labelNm = "label_eng";

    this.setData = function (conceptData, lang) {
        this.conceptData = conceptData;
        if (lang)
            this.setLang(lang);
        this.redraw();
    }

    this.setLang = function (lang) {
        this.labelNm = "label_" + lang;
    }

    this.redraw = function () {
        var self = this;
        var opt = this.options;

        // clear old data
        d3.select(opt.containerNm + " svg").remove();

        if (!this.conceptData || this.conceptData.length == 0)
            return;

        // make labels shorter if necessary
        makeNiceConceptLabels(this.conceptData, this.labelNm);

        var maxLabelWidth = d3.max($.map(this.conceptData, function (item, i) {
            return measureTextWidth(item.shortLabel, opt.containerNm + " svg", "name");
        }));
        var graphWidth = $(opt.containerNm).width() - 10,
            conceptsHeight = (2*opt.barHeight + 2 * opt.bigGap + opt.smallGap) * this.conceptData.length,
            conceptLabelOffset = maxLabelWidth + 10,
            xAxisOffset = opt.showXAxis ? 30 : 0;

        var chart = d3.select(opt.containerNm)
            .append('svg')
            .attr('class', 'chart')
            .attr('width', graphWidth)
            .attr('height', conceptsHeight);

        // what should be the maximum value of the concept
        var maxXVal = opt.maxScale == undefined ? d3.max(this.conceptData, function (item) { return d3.max([item.groupFreq0, item.groupFreq1]); }) : opt.maxScale;
        var x = d3.scale.linear()
            .domain([0, maxXVal])
            .range([0, graphWidth - conceptLabelOffset]);

        var y = d3.scale.ordinal()
            .domain(d3.range(0, this.conceptData.length))
            .rangeBands([0, conceptsHeight]);

        // show x axis with scale
        if (opt.showXAxis) {
            var xAxis = d3.svg.axis()
                .scale(x)
                .orient("top")
                .tickPadding(3);
            chart.append("g")
                .attr("class", "x axis")
                .attr("transform", "translate(" + conceptLabelOffset + ",27)")
                .call(xAxis);
            chart.append("text")
                .classed("axis-label", true)
                .text("Concept relevance")
                .attr("text-anchor", "end")
                .attr("x", function (val) { return conceptLabelOffset - 10; })
                .attr("y", "18px");
        }

        // rect 0
        chart.selectAll("rect.rect0")
            .data(self.conceptData)
            .enter()
                .append("rect")
                .attr("class", opt.rectClsNm0)
                .attr("x", conceptLabelOffset)
                .attr("y", function (val, ind) { return xAxisOffset + y(ind) + opt.bigGap; })
                .attr("height", opt.barHeight)
                .attr("width", function (val) { return 0; })
                .on("mouseenter", function (d) {
                    $(opt.containerNm + " .tooltip-inner").html("Concept importance for left group: <b>{0}</b>".format(d.groupFreq0.toFixed(1)));
                    $(opt.containerNm + " .tooltip").addClass("in")
                        .css("left", (parseFloat(d3.select(this).attr("x")) + parseFloat((d3.select(this).attr("width") - $(".tooltip").width()) / 2)) + "px")
                        .css("top", (parseFloat(d3.select(this).attr("y")) - 53) + "px");
                })
                .on("mouseleave", function () {
                    $(opt.containerNm + " .tooltip").removeClass("in").css("left", "-1000px");
                })
                .transition()
                .duration(1000)
                .attr("width", function (val) { return d3.max([x(val.groupFreq0), 0]); });
    
        // rect 1
        chart.selectAll("rect.rect1")
            .data(self.conceptData)
            .enter()
                .append("rect")
                .attr("class", opt.rectClsNm1)
                .attr("x", conceptLabelOffset)
                .attr("y", function (val, ind) { return xAxisOffset + y(ind) + opt.bigGap + opt.barHeight + opt.smallGap; })
                .attr("height", opt.barHeight)
                .attr("width", function (val) { return 0; })
                .on("mouseenter", function (d) {
                    $(opt.containerNm + " .tooltip-inner").html("Concept importance for right group: <b>{0}</b>".format(d.groupFreq1.toFixed(1)));
                    $(opt.containerNm + " .tooltip").addClass("in")
                        .css("left", (parseFloat(d3.select(this).attr("x")) + parseFloat((d3.select(this).attr("width") - $(".tooltip").width()) / 2)) + "px")
                        .css("top", (parseFloat(d3.select(this).attr("y")) - 53) + "px");
                })
                .on("mouseleave", function () {
                    $(opt.containerNm + " .tooltip").removeClass("in").css("left", "-1000px");
                })
                .transition()
                .duration(1000)
                .attr("width", function (val) { return d3.max([x(val.groupFreq1), 0]); });

        // concept label
        chart.selectAll("text.name")
            .data(self.conceptData.slice(0, this.conceptData.length))
            .enter().append("text")
            .attr('class', function (val, ind) { return 'name concept ' + val.type; })
            .attr("x", conceptLabelOffset - 10)
            .attr("y", function (val, ind) { return xAxisOffset + y(ind) + y.rangeBand() / 2; })
            .attr("dy", ".36em")
            .attr("text-anchor", "end")
            .text(function (val) { return val.shortLabel; });
    }
}