
function DateLineChart(_options) {
    // settings
    var options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        logYScale: false,
        interpolation: "linear",
        minYVal: undefined,           // define min val if you want to have it constant. if undefined it will be set by data's min val
        maxYVal: undefined,
        tooltipTextCallback: undefined,
        defaultZoomEnabled: false
    }, _options);
    this.colors = d3.scale.category10();

    var zoom = undefined;

    this.data = [];
    // tooltip
    $(options.containerNm).append("<div class='lineChart-tooltip hidden' />");
    $(options.containerNm).css("position", "relative");
    var tooltipDiv = $(options.containerNm + " .lineChart-tooltip");
    tooltipDiv.css("position", "absolute");
    
    // parse article date into d3s date format
    var dateFormat = d3.time.format("%Y-%m-%d"),
        parseDateTime = dateFormat.parse,
        bisectDate = d3.bisector(function (d) { return d.date; }).left;
    
    // methods

    this.clearData = function() {
        this.data = [];
    }

    // dataObj should be of form { dataArr: [{date: "2013-12-22", val: 123}, ...], legendLabel: "...", color: "..."}
    this.addData = function (dataObj, redraw) {
        dataObj.dataArr.sort(function (a, b) {
            if (a.date < b.date) return -1;
            if (a.date > b.date) return 1;
            return 0;
        });
        this.data.push(dataObj);
        if (redraw == undefined || redraw == true)
            this.redraw();
    };

    this.redraw = function() {
        function onZoom() {
            //var svg = d3.select("#article-trends-container svg");
            svg.select("g.x.axis").call(xAxis);
            svg.select("g.y.axis").call(yAxis);
            svg.selectAll("path.line").attr("d", line);
            svg.selectAll("circle.dot")
                .attr("cx", function (d) { return xScale(parseDateTime(d.date)); })
                .attr("cy", function (d) { return yScale(d.val); });
        }
        
        function drawLine(lineData, index) {
            var color = lineData.color ? lineData.color : self.colors(lineData.legend);
            // draw area
            chartBody.append("path")
                .datum(lineData.dataArr)
                .attr("class", "line")
                .attr("d", line)
                .attr("stroke", color);

            // draw points for articles
            chartBody.selectAll(".dot" + index)
                .data(lineData.dataArr)
                .enter().append("circle")
                .attr("class", "dot")
                .attr("r", 3)
                .attr("cx", function (d) { return xScale(parseDateTime(d.date)); })
                .attr("cy", function (d) { return yScale(d.val); })
                .style("stroke", color)
                .style("fill", color);
        }
        
        function mousemove() {
            // if we don't have a callback for setting the tooltip text then don't do anything
            if (options.tooltipTextCallback == undefined)
                return;
            var mousePos = d3.mouse(this);
            var x0 = xScale.invert(mousePos[0]);
            if (x0.getHours() > 12) x0.setHours(x0.getHours() + 12);
            var xVal = dateFormat(x0);
            var points = $.map(self.data, function(lineData, lineIndex) {
                var i = bisectDate(lineData.dataArr, xVal, 1),
                    closestI = x0 - parseDateTime(lineData.dataArr[i-1].date) > parseDateTime(lineData.dataArr[i].date) - x0 ? i : i-1;
                var y = yScale(lineData.dataArr[closestI].val);
                var x = xScale(parseDateTime(lineData.dataArr[closestI].date));
                var dist2 = Math.pow(x - mousePos[0], 2) + Math.pow(y - mousePos[1], 2);
                return { lineIndex: lineIndex, pointIndex: closestI, yVal: y, xVal: x, dist: Math.sqrt(dist2) };
            });
            var minDist = d3.min(points, function (point) { return point.dist; });
            if (minDist < 200) {
                var closestPoint = $.grep(points, function (val) { return val.dist == minDist; });
                var x = closestPoint[0].xVal, y = closestPoint[0].yVal;
                tooltipDiv.html(options.tooltipTextCallback(self.data[closestPoint[0].lineIndex], self.data[closestPoint[0].lineIndex].dataArr[closestPoint[0].pointIndex], closestPoint[0].pointIndex));
                tooltipDiv
                    .css("left", x + margin.left - tooltipDiv.width() / 2)
                    .css("top", y + 5 + margin.top)
                    .removeClass("hidden");
            }
            else
                tooltipDiv.addClass("hidden");
        }

        var self = this;
        d3.select(options.containerNm + " svg").remove();
        
        var totWidth = $(options.containerNm).width();
        var totHeight = $(options.containerNm).height();
        var margin = $.extend({ top: 10, right: 40, bottom: 40, left: 40 }, options.margin);
        var width = totWidth - margin.left - margin.right,
            height = totHeight - margin.top - margin.bottom;
        
        var xScale = d3.time.scale().range([0, width]);
        var yScale = options.logYScale ? d3.scale.log().range([height, 0]) : d3.scale.linear().range([height, 0]);

        var xAxis = d3.svg.axis().scale(xScale).orient("bottom");
        var yAxis = d3.svg.axis().scale(yScale).orient("left");
       
        var line = d3.svg.area()
            .interpolate(options.interpolation)
            .x(function (d) { return xScale(parseDateTime(d.date)); })
            .y(function(d, i) { return yScale(d.val); });

        zoom = d3.behavior.zoom();
        
        var svg = d3.select(options.containerNm)
            .append("svg")
            .attr("width", totWidth)
            .attr("height", totHeight)
            .on("mouseout", function () { tooltipDiv.addClass("hidden"); })
            .append("g")
                .attr("transform", "translate(" + margin.left + "," + margin.top + ")")
                .call(zoom);
        
        // create zoomOnOffHandler after call(zoom) - in case default zooming is false we need to disconnect event handler
        var zoomOnOffHandler = new ZoomOnOffHandler(options.containerNm, options.containerNm + " .onoffswitch-checkbox", zoom, onZoom, options.defaultZoomEnabled);

        svg.append("rect")
            .attr("class", "pane" + (options.defaultZoomEnabled ? " resize" : ""))
            .attr("fill", "transparent")
            .attr("width", width)
            .attr("height", height)
            .on("mousemove", mousemove);

        svg.append("clipPath")
            .attr("id", "clip")
            .append("rect")
                .attr("x", 0)
                .attr("y", 0)
                .attr("width", width)
                .attr("height", height);

        svg.append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0," + height + ")")
            .call(xAxis);

        svg.append("g")
            .attr("class", "y axis")
            .call(yAxis);

        var allData = d3.merge(self.data.map(function (val) { return val.dataArr; }));
        xScale.domain(d3.extent(allData.map(function (d) { return parseDateTime(d.date); })));
        xScale.nice(d3.time.day);   // start and end the x scale on a day
        zoom.x(xScale);

        var chartBody = svg.append("g")
            .attr("clip-path", "url(#clip)");

        var maxYVal = options.maxYVal != undefined ? options.maxYVal : d3.max(allData, function (item) { return item.val; });
        var minYVal = options.minYVal != undefined ? options.minYVal : d3.min(allData, function (item) { return item.val; });
        var diff = (maxYVal - minYVal) * 0.02;
        yScale.domain([minYVal - diff, maxYVal + diff]);
        
        $.each(self.data, function (i, data) {
            drawLine(data, i);
        });

        onZoom();
    };
    
}

// //////////////////
// bar chart - used for displaying one series of data as a bar chart (see events timeline example)
function DateBarChart(_options) {
    // settings
    var options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        maxYVal: undefined,
        tooltipTextCallback: undefined,
        colors: d3.scale.category10(),
        minValidYear: 2000,
        maxValidYear: 2015,
        defaultZoomEnabled: false,
        onClickCallback: undefined
    }, _options);

    var zoom = undefined;
    
    // tooltip
    $(options.containerNm).append('<div class="dateBarChart-tooltip graph-tooltip hidden" />');
    $(options.containerNm).css("position", "relative");
    var tooltipDiv = $(options.containerNm + " .dateBarChart-tooltip");
    tooltipDiv.css("position", "absolute");

    // parse article date into d3s date format
    var dateFormat = d3.time.format("%Y-%m-%d"),
        parseDateTime = dateFormat.parse,
        bisectDate = d3.bisector(function (d) { return d.date; }).left;

    this.clearData = function () {
        this.setData(null);
    }

    // dataObj should be of form { dataArr: [{date: "2013-12-22", val: 123}, ...], legendLabel: "...", color: "..."}
    this.setData = function (rawData) {
        function onZoom() {
            svg.select("g.x.axis").call(xAxis);
            svg.select("g.y.axis").call(yAxis);
            svg.select("path.area").attr("d", area);
            svg.selectAll(".dot")
              .data(data)
              .attr("class", "dot")
              .attr("r", function (d) { return d.value > 0 ? dotScale(d.value) : 0; })
              .attr("cx", function (d) { var mid = new Date(d.date); mid.setHours(12); return xScale(mid); })
              .attr("cy", function (d) { return yScale(0) + 35; });
        }

        d3.select(options.containerNm + " svg").remove();

        if (rawData == null)
            return;
        
        var data = [];
        var seenDates = {};
        rawData.forEach(function (d) {
            if (d.date.indexOf("-") == 0 || d.date.lastIndexOf("-") == d.date.length - 1)
                return;
            var date = parseDateTime(d.date);
            if (date.getFullYear() < options.minValidYear || date.getFullYear() > options.maxValidYear)
                return;
            seenDates[d.date] = date;
            data.push({ value: d.count, date: date });
        });
        for (dateStr in seenDates) {
            var dateNext = new Date(seenDates[dateStr].valueOf() + (24 * 1000 * 60 * 60));
            var dateNextStr = dateFormat(dateNext);
            if (seenDates[dateNextStr] == undefined)
                data.push({ value: 0, date: dateNext });
        }
        // sort the dates - otherwise the graph is a mess
        data.sort(function (a, b) {
            if (a.date > b.date) return 1;
            return -1;
        });
        
        var dotScale = d3.scale.log()
            .range([2, 10])
            .domain([1, d3.max(data.map(function (d) { return d.value; }))]);
        
        var totWidth = $(options.containerNm).width() - 5;
        var totHeight = $(options.containerNm).height() - 5;
        var margin = $.extend({ top: 20, right: 10, bottom: 60, left: 50 }, options.margin);
        var width = totWidth - margin.left - margin.right,
            height = totHeight - margin.top - margin.bottom;

        var xScale = d3.time.scale().range([0, width]);
        var yScale = options.logYScale ? d3.scale.log().range([height, 0]) : d3.scale.linear().range([height, 0]);

        var xAxis = d3.svg.axis().scale(xScale).orient("bottom").tickPadding(6);
        var yAxis = d3.svg.axis().scale(yScale).orient("left").tickSize(-width).tickPadding(6);

        var area = d3.svg.area()
            .interpolate("step-after")
            .x(function (d) { return xScale(d.date); })
            .y0(yScale(0))
            .y1(function (d) { return yScale(d.value); });

        zoom = d3.behavior.zoom();

        var svg = d3.select(options.containerNm)
            .append("svg")
            .attr("width", totWidth)
            .attr("height", totHeight)
            .on("mouseout", function () { tooltipDiv.addClass("hidden"); })
            .append("g")
                .attr("transform", "translate(" + margin.left + "," + margin.top + ")")
                .call(zoom);

        // create zoomOnOffHandler after call(zoom) - in case default zooming is false we need to disconnect event handler
        var zoomOnOffHandler = new ZoomOnOffHandler(options.containerNm, options.containerNm + " .onoffswitch-checkbox", zoom, onZoom, options.defaultZoomEnabled);

        svg.append("rect")
            .attr("class", "pane" + (options.defaultZoomEnabled ? " resize" : ""))
            .attr("fill", "transparent")
            .attr("width", width)
            .attr("height", height)

        svg.append("clipPath")
            .attr("id", "clip")
            .append("rect")
                .attr("x", 0)
                .attr("y", 0)
                .attr("width", width)
                .attr("height", totHeight);

        svg.append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0," + height + ")")
            .call(xAxis);

        svg.append("g")
            .attr("class", "y axis")
            .call(yAxis);

        xScale.domain(d3.extent(data.map(function (d) { return d.date; })));
        xScale.nice(d3.time.day);   // start and end the x scale on a day
        zoom.x(xScale);

        var chartBody = svg.append("g")
            .attr("clip-path", "url(#clip)")
            .on("mousemove", function() {
                // if we don't have a callback for setting the tooltip text then don't do anything
                if (options.tooltipTextCallback == undefined || data == undefined || data.length == 0)
                    return;
                var mousePos = d3.mouse(this);
                var x0 = xScale.invert(mousePos[0]);
                var i = bisectDate(data, x0, 1);
                var closestIndex = (i < data.length && dateFormat(data[i].date) == dateFormat(x0)) ? i : i - 1;
                var val = (dateFormat(x0) == dateFormat(data[closestIndex].date)) ? data[closestIndex].value : 0;
                tooltipDiv.html(options.tooltipTextCallback(x0, val));
                var xOffset = mousePos[0] > ($(options.containerNm).width() / 2) ? -tooltipDiv.width() - 10 : 15;
                var yOffset = mousePos[1] > ($(options.containerNm).height() / 2) ? -tooltipDiv.height() - 25 : 5;
                tooltipDiv
                    .css("left", margin.left + mousePos[0] + xOffset)
                    .css("top", margin.top + mousePos[1] + yOffset)
                    .removeClass("hidden");
            })
            .on("mouseout", function() {
                tooltipDiv.addClass("hidden");
            })
            // call callback on clicked rectangle
            .on("click", function () {
                if (!options.onClickCallback)
                    return;
                var mousePos = d3.mouse(this);
                var x0 = xScale.invert(mousePos[0]);
                options.onClickCallback(this, dateFormat(x0));
            })

        var maxYVal = options.maxYVal != undefined ? options.maxYVal : (d3.max(data.map(function (d) { return d.value; })));
        var minYVal = options.minYVal != undefined ? options.minYVal : (d3.min(data.map(function (d) { return d.value; })));
        var diff = (maxYVal - minYVal) * 0.05;
        yScale.domain([minYVal, maxYVal + diff]);

        chartBody.append("path")
            .datum(data)
            .attr("class", "area");
        
        chartBody.selectAll(".dot")
            .data(data)
            .enter().append("circle")
            .attr("clip-path", "url(#clip)")
            .attr("class", "dot");
        
        onZoom();
    };
}

// //////////////////
// bar charts - used for displaying multiple series of data as a bar chart (see events timeline example)
// expected input: [{data:[], lineColor: string, areaColor: string, legend: string}, {data:[], lineColor: string, areaColor: string, legend: string}]
function DateBarCharts(_options) {
    // settings
    var options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        maxYVal: undefined,
        tooltipTextCallback: undefined,
        minValidYear: 2000,
        maxValidYear: 2015,
        colors: d3.scale.category10(),
        defaultZoomEnabled: false
    }, _options);

    var zoom = undefined;
    var areas = [];
    var dataArrs = [];
    
    // tooltip
    $(options.containerNm).append('<div class="dateBarChart-tooltip graph-tooltip hidden" />');
    $(options.containerNm).css("position", "relative");
    var tooltipDiv = $(options.containerNm + " .dateBarChart-tooltip");
    tooltipDiv.css("position", "absolute");

    // parse article date into d3s date format
    var dateFormat = d3.time.format("%Y-%m-%d"),
        parseDateTime = dateFormat.parse,
        bisectDate = d3.bisector(function (d) { return d.date; }).left;

    this.clearData = function () {
        this.setData(null);
    }

    // dataObj should be of form { dataArr: [{date: "2013-12-22", val: 123}, ...], legendLabel: "...", color: "..."}
    this.setData = function (inputData) {
        function onZoom() {
            svg.select("g.x.axis").call(xAxis);
            svg.select("g.y.axis").call(yAxis);
            $.each(areas, function(i, area) {
                svg.select("path.area" + i).attr("d", area);
                svg.selectAll(".dot" + i)
                    .attr("cx", function (d) {
                        var mid = new Date(d.date);
                        mid.setHours(12);
                        return xScale(mid);
                    })
                    .attr("cy", function (d) { return yScale(0) + 40 + i*30; });
            });
        }
        
        var self = this;
        d3.select(options.containerNm + " svg").remove();
        areas = [];
        dataArrs = [];
        
        if (inputData == null)
            return;
        
        var totWidth = $(options.containerNm).width() - 5;
        var totHeight = $(options.containerNm).height() - 5;
        var margin = $.extend({ top: 20, right: 10, bottom: 60, left: 50 }, options.margin);
        var width = totWidth - margin.left - margin.right,
            height = totHeight - margin.top - margin.bottom - (inputData.length * 30) - 20;

        var xScale = d3.time.scale().range([0, width]);
        var yScale = options.logYScale ? d3.scale.log().range([height, 0]) : d3.scale.linear().range([height, 0]);

        var xAxis = d3.svg.axis().scale(xScale).orient("bottom").tickPadding(6);
        var yAxis = d3.svg.axis().scale(yScale).orient("left").tickSize(-width).tickPadding(6);

        [].concat([1, 2], [3, 2])

        var dotScale = d3.scale.log()
            .range([2, 10]);
        
        zoom = d3.behavior.zoom();

        var svg = d3.select(options.containerNm)
            .append("svg")
            .attr("width", totWidth)
            .attr("height", totHeight)
            .on("mouseout", function () { tooltipDiv.addClass("hidden"); })
            .append("g")
                .attr("transform", "translate(" + margin.left + "," + margin.top + ")")
                .call(zoom);

        // create zoomOnOffHandler after call(zoom) - in case default zooming is false we need to disconnect event handler
        var zoomOnOffHandler = new ZoomOnOffHandler(options.containerNm, options.containerNm + " .onoffswitch-checkbox", zoom, onZoom, options.defaultZoomEnabled);

        svg.append("rect")
            .attr("class", "pane" + (options.defaultZoomEnabled ? " resize" : ""))
            .attr("fill", "transparent")
            .attr("width", width)
            .attr("height", height);

        svg.append("clipPath")
            .attr("id", "clip")
            .append("rect")
                .attr("x", 0)
                .attr("y", 0)
                .attr("width", width)
                .attr("height", totHeight);

        svg.append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0," + height + ")")
            .call(xAxis);

        svg.append("g")
            .attr("class", "y axis")
            .call(yAxis);

        var chartBody = svg.append("g")
            .attr("clip-path", "url(#clip)")
            .on("mousemove", function () {
                // if we don't have a callback for setting the tooltip text then don't do anything
                if (options.tooltipTextCallback == undefined || dataArrs.length == 0)
                    return;
                var mousePos = d3.mouse(this);
                var x0 = xScale.invert(mousePos[0]);
                var dataArr = [];
                $.each(dataArrs, function (i, data) {
                    var i = bisectDate(data, x0, 1);
                    var closestIndex = (i < data.length && dateFormat(data[i].date) == dateFormat(x0)) ? i : i - 1;
                    dataArr.push((dateFormat(x0) == dateFormat(data[closestIndex].date)) ? data[closestIndex].value : 0);
                })
                tooltipDiv.html(options.tooltipTextCallback(x0, dataArr));
                var xOffset = mousePos[0] > ($(options.containerNm).width() / 2) ? -tooltipDiv.width() - 10 : 15;
                var yOffset = mousePos[1] > ($(options.containerNm).height() / 2) ? -tooltipDiv.height() - 25 : 5;
                tooltipDiv
                    .css("left", margin.left + mousePos[0] + xOffset)
                    .css("top", margin.top + mousePos[1] + yOffset)
                    .removeClass("hidden");
            })
            .on("mouseout", function () {
                tooltipDiv.addClass("hidden");
            })

        var xMin = undefined, xMax = undefined, yMax = options.maxYVal;
        $.each(inputData, function(i, dataObj) {
            var data = [];
            var seenDates = {};
            dataObj.data.forEach(function(d) {
                if (d.date.indexOf("-") == 0 || d.date.lastIndexOf("-") == d.date.length - 1)
                    return;
                var date = parseDateTime(d.date);
                if (date.getFullYear() < options.minValidYear || date.getFullYear() > options.maxValidYear)
                    return;
                seenDates[d.date] = date;
                data.push({ value: d.count, date: date });
            });
            for (dateStr in seenDates) {
                var dateNext = new Date(seenDates[dateStr].valueOf() + (24 * 1000 * 60 * 60));
                var dateNextStr = dateFormat(dateNext);
                if (seenDates[dateNextStr] == undefined)
                    data.push({ value: 0, date: dateNext });
            }
            // sort the dates - otherwise the graph is a mess
            data.sort(function(a, b) {
                if (a.date > b.date) return 1;
                return -1;
            });

            var area = d3.svg.area()
                .interpolate("step-after")
                .x(function(d) { return xScale(d.date); })
                .y0(yScale(0))
                .y1(function(d) { return yScale(d.value); });

            chartBody.append("path")
                .datum(data)
                .style("stroke", dataObj.lineColor ? dataObj.lineColor : options.colors(i))
                .style("fill", dataObj.areaColor ? dataObj.areaColor : options.colors(i))
                .attr("class", "area area" + i);

            chartBody.selectAll(".dot" + i)
                .data(data)
                .enter()
                .append("circle")
                .attr("r", function (d) { return d.value > 0 ? dotScale(d.value) : 0; })
                .style("fill", dataObj.lineColor ? dataObj.lineColor : options.colors(i))
                .attr("clip-path", "url(#clip)")
                .attr("class", "dot dot" + i);

            dataArrs.push(data);
            areas.push(area);

            var minmaxX = d3.extent(data.map(function(d) { return d.date; }));
            var minmaxY = d3.extent(data.map(function(d) { return d.value; }));
            xMin = (xMin == undefined) ? minmaxX[0] : d3.min([xMin, minmaxX[0]]);
            xMax = (xMax == undefined) ? minmaxX[1] : d3.max([xMax, minmaxX[1]]);
            yMax = (yMax == undefined) ? minmaxY[1] : d3.max([yMax, minmaxY[1]]);
        });
        
        dotScale.domain([1, yMax]);
        
        // legend
        var legend = svg.selectAll('g.legend')
            .data(inputData)
            .enter()
            .append('g')
            .attr('class', 'legend');

        legend.append('circle')
            .attr('cx', 80)
            .attr('cy', function (d, i) { return margin.top + height + 20 + inputData.length * 30 + (i * 30); })
            .attr('r', 7)
            //.attr('height', 10)
            .style('fill', function (d, i) { return d.lineColor; });

        legend.append('text')
            .attr('x', 100)
            .attr('y', function (d, i) { return margin.top + height + 20 + inputData.length * 30 + (i * 30) + 5; })
            .text(function (d, i) { return d.legendText });

        xScale.domain([xMin, xMax]);
        xScale.nice(d3.time.day);   // start and end the x scale on a day
        zoom.x(xScale);
        
        var yMin = yMin != undefined ? yMin : 0;
        var diff = (yMax - yMin) * 0.05;
        yScale.domain([yMin, yMax + diff]);
        
        onZoom();
    };
}

// //////////////////
function LineChart(_options) {
    // settings
    var options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        logXScale: false,
        logYScale: false,
        interpolation: "linear",
        minYVal: undefined,           // define min val if you want to have it constant. if undefined it will be set by data's min val
        maxYVal: undefined,
        tooltipTextCallback: undefined,
        defaultZoomEnabled: false
    }, _options);
    this.colors = d3.scale.category10();

    this.data = [];
    // tooltip
    $(options.containerNm).append("<div class='lineChart-tooltip hidden' />");
    $(options.containerNm).css("position", "relative");
    var tooltipDiv = $(options.containerNm + " .lineChart-tooltip");
    tooltipDiv.css("position", "absolute");

    // parse article date into d3s date format
    var keyBisector = d3.bisector(function (d) { return d.key ; }).left;

    // methods

    this.clearData = function () {
        this.data = [];
    }

    // dataObj should be of form { dataArr: [{date: "2013-12-22", val: 123}, ...], legendLabel: "...", color: "..."}
    this.addData = function (dataObj, redraw) {
        dataObj.dataArr.sort(function (a, b) {
            if (a.key < b.key) return -1;
            if (a.key > b.key) return 1;
            return 0;
        });
        this.data.push(dataObj);
        if (redraw == undefined || redraw == true)
            this.redraw();
    };

    this.redraw = function () {
        function onZoom() {
            //var svg = d3.select("#article-trends-container svg");
            svg.select("g.x.axis").call(xAxis);
            svg.select("g.y.axis").call(yAxis);
            svg.selectAll("path.line").attr("d", line);
            svg.selectAll("circle.dot")
                .attr("cx", function (d) { return xScale(d.key); })
                .attr("cy", function (d) { return yScale(d.val); });
        }

        function drawLine(lineData, index) {
            var color = lineData.color ? lineData.color : self.colors(lineData.legend);
            // draw area
            chartBody.append("path")
                .datum(lineData.dataArr)
                .attr("class", "line")
                .attr("d", line)
                .attr("stroke", color);

            // draw points for articles
            chartBody.selectAll(".dot" + index)
                .data(lineData.dataArr)
                .enter().append("circle")
                .attr("class", "dot")
                .attr("r", 3)
                .attr("cx", function (d) { return xScale(d.key); })
                .attr("cy", function (d) { return yScale(d.val); })
                .style("stroke", color)
                .style("fill", color);
        }

        function mousemove() {
            // if we don't have a callback for setting the tooltip text then don't do anything
            if (options.tooltipTextCallback == undefined)
                return;
            var mousePos = d3.mouse(this);
            var x0 = xScale.invert(mousePos[0]);
            //if (x0.getHours() > 12) x0.setHours(x0.getHours() + 12);
            var points = $.map(self.data, function (lineData, lineIndex) {
                var i = keyBisector(lineData.dataArr, x0, 1);
                if (lineData.dataArr[i - 1] == undefined)
                    console.log(lineData);
                var closestI = x0 - lineData.dataArr[i - 1].key > lineData.dataArr[i].key - x0 ? i : i - 1;
                var y = yScale(lineData.dataArr[closestI].val);
                var x = xScale(lineData.dataArr[closestI].key);
                var dist2 = Math.pow(x - mousePos[0], 2) + Math.pow(y - mousePos[1], 2);
                return { lineIndex: lineIndex, pointIndex: closestI, yVal: y, xVal: x, dist: Math.sqrt(dist2) };
            });
            var minDist = d3.min(points, function (point) { return point.dist; });
            if (minDist < 200) {
                var closestPoint = $.grep(points, function (val) { return val.dist == minDist; });
                var x = closestPoint[0].xVal, y = closestPoint[0].yVal;
                tooltipDiv.html(options.tooltipTextCallback(self.data[closestPoint[0].lineIndex], self.data[closestPoint[0].lineIndex].dataArr[closestPoint[0].pointIndex], closestPoint[0].pointIndex));
                tooltipDiv
                    .css("left", x + margin.left - tooltipDiv.width() / 2)
                    .css("top", y + 5 + margin.top)
                    .removeClass("hidden");
            }
            else
                tooltipDiv.addClass("hidden");
        }

        var self = this;
        d3.select(options.containerNm + " svg").remove();

        var totWidth = $(options.containerNm).width();
        var totHeight = $(options.containerNm).height();
        var margin = $.extend({ top: 10, right: 40, bottom: 40, left: 40 }, options.margin);
        var width = totWidth - margin.left - margin.right,
            height = totHeight - margin.top - margin.bottom;

        var xScale = options.logXScale ? d3.scale.log().range([0, width]) : d3.scale.linear().range([0, width]);
        var yScale = options.logYScale ? d3.scale.log().range([height, 0]) : d3.scale.linear().range([height, 0]);

        var xAxis = d3.svg.axis().scale(xScale).orient("bottom");
        var yAxis = d3.svg.axis().scale(yScale).orient("left");

        var line = d3.svg.area()
            .interpolate(options.interpolation)
            .x(function (d) { return xScale(d.key); })
            .y(function (d, i) { return yScale(d.val); });

        zoom = d3.behavior.zoom();

        var svg = d3.select(options.containerNm)
            .append("svg")
            .attr("width", totWidth)
            .attr("height", totHeight)
            .on("mouseout", function () { tooltipDiv.addClass("hidden"); })
            .append("g")
                .attr("transform", "translate(" + margin.left + "," + margin.top + ")")
                .call(zoom);

        // create zoomOnOffHandler after call(zoom) - in case default zooming is false we need to disconnect event handler
        var zoomOnOffHandler = new ZoomOnOffHandler(options.containerNm, options.containerNm + " .onoffswitch-checkbox", zoom, onZoom, options.defaultZoomEnabled);

        svg.append("rect")
            .attr("class", "pane" + (options.defaultZoomEnabled ? " resize" : ""))
            .attr("fill", "transparent")
            .attr("width", width)
            .attr("height", height)
            .on("mousemove", mousemove);

        svg.append("clipPath")
            .attr("id", "clip")
            .append("rect")
                .attr("x", 0)
                .attr("y", 0)
                .attr("width", width)
                .attr("height", height);

        svg.append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0," + height + ")")
            .call(xAxis);

        svg.append("g")
            .attr("class", "y axis")
            .call(yAxis);

        var allData = d3.merge(self.data.map(function (val) { return val.dataArr; }));
        xScale.domain(d3.extent(allData.map(function (d) { return d.key; })));
        zoom.x(xScale);

        var chartBody = svg.append("g")
            .attr("clip-path", "url(#clip)");

        var maxYVal = options.maxYVal != undefined ? options.maxYVal : d3.max(allData, function (item) { return item.val; });
        var minYVal = options.minYVal != undefined ? options.minYVal : d3.min(allData, function (item) { return item.val; });
        var diff = (maxYVal - minYVal) * 0.02;
        yScale.domain([options.logYScale ? 0.01 : (minYVal - diff), maxYVal + diff]);
        
        $.each(self.data, function (i, data) {
            drawLine(data, i);
        });

        onZoom();
    };
}


// ////////////////////
// BOX PLOT
// //////////////////
function BoxPlot(_options) {
    // settings
    var options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        minYVal: undefined,           // define min val if you want to have it constant. if undefined it will be set by data's min val
        maxYVal: undefined,
        defaultZoomEnabled: false
    }, _options);
    this.colors = d3.scale.category10();
    this.data = undefined;

    // parse article date into d3s date format
    var dateFormat = d3.time.format("%Y-%m-%d"),
        parseDateTime = dateFormat.parse;

    // methods

    this.clearData = function () {
        this.data = [];
    }

    // dataObj should be of form { dataArr: [{date: "2013-12-22", val: 123}, ...], legendLabel: "...", color: "..."}
    this.setData = function (data, redraw) {
        this.data = data;
        data.dataArr.sort(function (a, b) {
            if (a.date < b.date) return -1;
            if (a.date > b.date) return 1;
            return 0;
        });
        if (redraw == undefined || redraw == true)
            this.redraw();
    };

    this.redraw = function () {
        function onZoom() {
            //var svg = d3.select("#article-trends-container svg");
            svg.select("g.x.axis").call(xAxis);
            svg.select("g.y.axis").call(yAxis);
            
            d3.selectAll(options.containerNm + " .box-plot").remove();
            $.each(self.data.dataArr, function (i, dataPoint) {
                drawBoxPlot(dataPoint);
            });
        }

        function drawBoxPlot(dataPoint) {
            var middleX = xScale(parseDateTime(dataPoint.date));
            var dx = xScale(parseDateTime(dataPoint.date).setHours(6)) - middleX;

            var g = chartBody.append("g")
                .classed("box-plot", true)
                .attr("transform", "translate(" + (middleX - dx) + ",0)");
            g.append("line")
                .classed("vert-line", true)
                .attr("x1", dx)
                .attr("y1", yScale(dataPoint.d1))
                .attr("x2", dx)
                .attr("y2", yScale(dataPoint.d9));
            g.append("line")
                .classed("hor-line", true)
                .attr("x1", 0)
                .attr("y1", yScale(dataPoint.d1))
                .attr("x2", 2*dx)
                .attr("y2", yScale(dataPoint.d1));
            g.append("line")
                .classed("hor-line", true)
                .attr("x1", 0)
                .attr("y1", yScale(dataPoint.d9))
                .attr("x2", 2 * dx)
                .attr("y2", yScale(dataPoint.d9));
            g.append("rect")
                .classed("rect", true)
                .attr("x", 0)
                .attr("y", yScale(dataPoint.q3))
                .attr("width", 2 * dx)
                .attr("height", yScale(dataPoint.q1) - yScale(dataPoint.q3));
            g.append("line")
                .classed("middle-line", true)
                .attr("x1", 0)
                .attr("y1", yScale(dataPoint.q2))
                .attr("x2", 2 * dx)
                .attr("y2", yScale(dataPoint.q2));
        }

        var self = this;
        d3.select(options.containerNm + " svg").remove();

        var totWidth = $(options.containerNm).width();
        var totHeight = $(options.containerNm).height();
        var margin = $.extend({ top: 10, right: 40, bottom: 40, left: 40 }, options.margin);
        var width = totWidth - margin.left - margin.right,
            height = totHeight - margin.top - margin.bottom;

        var xScale = d3.time.scale().range([0, width]);
        var yScale = d3.scale.linear().range([height, 0]);

        var xAxis = d3.svg.axis().scale(xScale).orient("bottom");
        var yAxis = d3.svg.axis().scale(yScale).orient("left");

        zoom = d3.behavior.zoom();

        var svg = d3.select(options.containerNm)
            .append("svg")
            .attr("width", totWidth)
            .attr("height", totHeight)
            .append("g")
                .attr("transform", "translate(" + margin.left + "," + margin.top + ")")
                .call(zoom);

        // create zoomOnOffHandler after call(zoom) - in case default zooming is false we need to disconnect event handler
        var zoomOnOffHandler = new ZoomOnOffHandler(options.containerNm, options.containerNm + " .onoffswitch-checkbox", zoom, onZoom, options.defaultZoomEnabled);

        svg.append("rect")
            .attr("class", "pane" + (options.defaultZoomEnabled ? " resize" : ""))
            .attr("fill", "transparent")
            .attr("width", width)
            .attr("height", height);

        svg.append("clipPath")
            .attr("id", "clip")
            .append("rect")
                .attr("x", 0)
                .attr("y", 0)
                .attr("width", width)
                .attr("height", height);

        svg.append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0," + height + ")")
            .call(xAxis);

        svg.append("g")
            .attr("class", "y axis")
            .call(yAxis);

        var domain = d3.extent(this.data.dataArr.map(function (d) { return parseDateTime(d.date); }));
        domain[0].setHours(-12); domain[1].setHours(12);
        xScale.domain(domain);
        zoom.x(xScale);

        var chartBody = svg.append("g")
            .attr("clip-path", "url(#clip)");

        var maxYVal = options.maxYVal != undefined ? options.maxYVal : d3.max(this.data.dataArr, function (item) { return item.d9; });
        var minYVal = options.minYVal != undefined ? options.minYVal : d3.min(this.data.dataArr, function (item) { return item.d1; });
        var diff = (maxYVal - minYVal) * 0.02;
        yScale.domain([minYVal - diff, maxYVal + diff]);

        onZoom();
    };
}
