// Breadcrumb dimensions: width, height, spacing, width of tip/tail.
var b = {
  w: 75, h: 30, s: 3, t: 10
};

function Sunburst(options) {
    this.options = $.extend({
        containerNm: undefined,     // name of the dom element hosting the svg
        sequenceNm: undefined,      // place for the current sequence
        legendNm: undefined,
        onClickCallback: undefined
    }, options);
    
    this.data = undefined;
    // Total size of all segments; we set this later, after loading the data.
    this.totalSize = 0;
    this.colors = d3.scale.category10();
    this.vis = undefined;

    this._nodesToColor = [];

    this.setData = function(data) {
        this.data = this._buildHierarchy(data);
        this.redraw();
    };

    this.redraw = function () {
        this.colors("ShortestPaths");
        var self = this;
        var opt = this.options;

        // clear old data
        d3.select(opt.containerNm + " svg").remove();
        d3.select(opt.sequenceNm + " svg").remove();
        d3.select(opt.legendNm + " svg").remove();
        
        var size = Math.min($(opt.containerNm).width() - 120, window.innerHeight - 100);
        var width = height = size; // make height the same the width
        var radius = Math.min(width, height) / 2;

        var explanation = $(opt.containerNm + " .explanation");
        explanation.css("left", ((width - explanation.width()) / 2) + "px")
            .css("top", ((height - explanation.height()) / 2) + "px");
            
        var vis = d3.select(this.options.containerNm).append("svg:svg")
            .attr("width", width)
            .attr("height", height)
            .append("svg:g")
            .attr("id", "container")
            .attr("transform", "translate(" + width / 2 + "," + height / 2 + ")");
        this.vis = vis;

        var partition = d3.layout.partition()
            .size([2 * Math.PI, radius * radius])
            .value(function(d) { return d.size; });

        var arc = d3.svg.arc()
            .startAngle(function(d) { return d.x; })
            .endAngle(function(d) { return d.x + d.dx; })
            .innerRadius(function(d) { return Math.sqrt(d.y); })
            .outerRadius(function(d) { return Math.sqrt(d.y + d.dy); });

        this._initializeBreadcrumbTrail(width);
        
        // Bounding circle underneath the sunburst, to make it easier to detect
        // when the mouse leaves the parent g.
        vis.append("svg:circle")
            .attr("r", radius)
            .style("opacity", 0);

        // For efficiency, filter nodes to keep only those large enough to see.
        var nodes = partition.nodes(self.data);
        this._nodesToColor = [];
        for (var n = 0; n < self.data.children.length; n++) {
            self.data.children[n].color = self.colors(n);
            this._nodesToColor.push(self.data.children[n]);
        }
        while (this._nodesToColor.length > 0) {
            var node = this._nodesToColor.shift();
            this._setChildrenColors(node);
        }
        
        // individual sections of the circle
        var path = vis.data([self.data]).selectAll(opt.containerNm + " path")
            .data(nodes)
            .enter()
                .append("svg:path")
                .attr("display", function (d) {
                    return (d.depth && d.name != "--end--") ? null : "none";
                })
                .attr("d", arc)
                .attr("fill-rule", "evenodd")
                .style("fill", function (d, i) {
                    return d.color ? d.color : "white";
                })
                .style("opacity", 1)
                .on("mouseover", self._mouseover)
                .on("click", opt.onClickCallback);


        // Add the mouseleave handler to the bounding circle.
        d3.select("#container").on("mouseleave", self._mouseleave);

        // Get total size of the tree = value of root node from partition.
        this.totalSize = path.node().__data__.value;

        this._drawLegend();
    };

    this._setChildrenColors = function(item) {
        function getIndex(prnt, itm) {
            for (var n = 0; n < prnt.children.length; n++) {
                if (prnt.children[n] == itm)
                    return n;
            }
        };
        if (!item.children)
            return;
        var ind = getIndex(item.parent, item);
        var next = (ind + 1) % item.parent.children.length;
        var itemColor = item.color;
        //var nextColor = item.parent.children[next].color;
        var nextColor = undefined;
        var color = d3.scale.linear()
            .range([itemColor, nextColor])
            .interpolate(d3.interpolateRgb);
        for (var i = 0; i < item.children.length; i++) {
            item.children[i].color = color(i / item.children.length);
            this._nodesToColor.push(item.children[i]);
        }
    };

    this._initializeBreadcrumbTrail = function() {
        var width = $(this.options.sequenceNm).width() - 10;
        // Add the svg area.
        var trail = d3.select(this.options.sequenceNm).append("svg:svg")
            .attr("width", width)
            .attr("height", 50)
            .attr("id", "trail");
        // Add the label at the end, for the percentage.
        trail.append("svg:text")
            .attr("id", "endlabel")
            .style("fill", "#000");
    };

    // Fade all but the current sequence, and show it in the breadcrumb trail.
    this._mouseover = function(d) {
        var self = d.self;
        var percentage = (100 * d.value / self.totalSize).toPrecision(3);
        var percentageString = percentage + "%";
        if (percentage < 0.1) {
            percentageString = "< 0.1%";
        }

        d3.select(self.options.containerNm + " .percentage")
            .text(percentageString);

        d3.select(self.options.containerNm + " .explanation")
            .style("visibility", "");

        var sequenceArray = self._getAncestors(d);
        self._updateBreadcrumbs(sequenceArray, percentageString);

        // Fade all the segments.
        d3.selectAll(self.options.containerNm + " path")
            .style("opacity", 0.3);

        // Then highlight only those that are an ancestor of the current segment.
        self.vis.selectAll(self.options.containerNm + " path")
            .filter(function(node) {
                return (sequenceArray.indexOf(node) >= 0);
            })
            .style("opacity", 1);
    };

    // Restore everything to full opacity when moving off the visualization.
    this._mouseleave = function (d) {
        var self = d.self;
        // Hide the breadcrumb trail
        d3.select("#trail")
            .style("visibility", "hidden");

        // Deactivate all segments during transition.
        d3.selectAll(self.options.containerNm + " path").on("mouseover", null);

        // Transition each segment to full opacity and then reactivate it.
        d3.selectAll(self.options.containerNm + " path")
            .transition()
            .duration(1000)
            .style("opacity", 1)
            .each("end", function () {
                d3.select(this).on("mouseover", self._mouseover);
            });

        d3.select(self.options.containerNm + " .explanation")
            .transition()
            .duration(1000)
            .style("visibility", "hidden");
    };
    
    // Generate a string that describes the points of a breadcrumb polygon.
    this._breadcrumbPoints = function (d, i) {
        var width = measureTextWidth(d.name, "body", "#sequence text");
        var off = 15*2;
        var points = [];
        points.push("0,0");
        points.push(width + off + ",0");
        points.push(width + off + b.t + "," + (b.h / 2));
        points.push(width + off + "," + b.h);
        points.push("0," + b.h);
        if (i > 0) { // Leftmost breadcrumb; don't include 6th vertex.
            points.push(b.t + "," + (b.h / 2));
        }
        return points.join(" ");
    };

    // Update the breadcrumb trail to show the current sequence and percentage.
    this._updateBreadcrumbs = function (nodeArray, percentageString) {
        var self = this;
        
        // Data join; key function combines name and depth (= position in sequence).
        var g = d3.select("#trail")
            .selectAll("g")
            .data(nodeArray, function (d) { return d.name + d.depth; });

        // Add breadcrumb and label for entering nodes.
        var entering = g.enter().append("svg:g");

        entering.append("svg:polygon")
            .attr("points", this._breadcrumbPoints)
            .style("fill", function(d, i) { return d.color; });

        entering.append("svg:text")
            .attr("x", "15")
            .attr("y", b.h / 2)
            .attr("dy", "0.35em")
            .text(function (d) { return d.name; });

        // Set position for entering and updating nodes.
        g.attr("transform", function (d, i) {
            var off = 0;
            for (var n = 0; n < i; n++)
                off += measureTextWidth(nodeArray[n].name, "body", "#sequence text");
            return "translate(" + (off + (i * (30 + b.s))) + ", 0)";
        });

        // Remove exiting nodes.
        g.exit().remove();

        // Now move and update the percentage at the end.
        d3.select("#trail").select("#endlabel")
            .attr("x", function(d) {
                var off = 0;
                for (var n = 0; n < nodeArray.length; n++)
                    off += measureTextWidth(nodeArray[n].name, "body", "#sequence text");
                var xoff = (off + (nodeArray.length * (30 + b.s))) + 15;
                return xoff;
            })
            .attr("y", b.h / 2)
            .attr("dy", "0.35em")
            .text(percentageString);

        // Make the breadcrumb trail visible, if it's hidden.
        d3.select("#trail")
            .style("visibility", "");
    };

    this._drawLegend = function () {
        var self = this;
        // Dimensions of legend item: width, height, spacing, radius of rounded rect.
        var li = { w: 75, h: 30, s: 3, r: 3 };

        d3.select(self.options.legendNm + " svg").remove();
        var legend = d3.select(self.options.legendNm).append("svg:svg")
            .attr("width", li.w)
            .attr("height", self.data.children.length * (li.h + li.s));

        var g = legend.selectAll("g")
            .data(self.data.children)
            .enter().append("svg:g")
            .attr("transform", function (d, i) {
                return "translate(0," + i * (li.h + li.s) + ")";
            });

        g.append("svg:rect")
            .attr("rx", li.r)
            .attr("ry", li.r)
            .attr("width", li.w)
            .attr("height", li.h)
            .style("fill", function (d) { return d.color; });

        g.append("svg:text")
            .attr("x", li.w / 2)
            .attr("y", li.h / 2)
            .attr("dy", "0.35em")
            .attr("text-anchor", "middle")
            .text(function (d) { return d.name; });
    };

    // Take a 2-column CSV and transform it into a hierarchical structure suitable
    // for a partition layout. The first column is a sequence of step names, from
    // root to leaf, separated by hyphens. The second column is a count of how 
    // often that sequence occurred.
    this._buildHierarchy = function (data) {
        function getChild(obj, childName) {
            for (var n = 0; n < obj.children.length; n++) {
                if (obj.children[n].name == childName)
                    return obj.children[n];
            }
            return undefined;
        }
        var root = { "name": "root", "children": [], "self": this };
        for (var c = 0; c < data.length; c++) {
            var cat = data[c];
            var count = cat.count;
            var name = cat.title.trimLStr("http://");
            var parts = name.split("/");
            var obj = root;
            for (var p = 1; p < parts.length; p++) {
                var key = parts[p].replace(/_/g, ' ');
                var child = getChild(obj, key);
                if (child == undefined) {
                    child = { name: key, children: [], self: this, uri: cat.title };
                    obj.children.push(child);
                }
                obj = child;
                if (p == parts.length - 1)
                    obj.children.push({ name: "--end--", size: count, self: this });
            }
        }
        return root;
    };

    // Given a node in a partition layout, return an array of all of its ancestor
    // nodes, highest first, but excluding the root.
    this._getAncestors = function (node) {
        var path = [];
        var current = node;
        while (current.parent) {
            path.unshift(current);
            current = current.parent;
        }
        return path;
    };
}





