﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Runtime.Serialization;
using UIControls;
using SimpleJson;
using SlavaGu.ConsoleAppLauncher;
using System.Net;
using GenLib;
using System.Threading;
using System.Windows.Threading;
using System.Net.Mail;

namespace NewsMinerManagement
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
	{
		private string _settingsFName = "NewsMinerManagement-settings.json";
		private ProfileList _profiles = null;
		private int _currentProfileIndex = -1;
		private int _lastLogToKeep = 10000;
		private string _startupDir = null;
		private object _logLock = new object();
		private DispatcherTimer _updateTimer;
		private EmailSender _emailSender = new EmailSender("gleban-test@gmail.com");
		private DispatcherTimer _validateActivity;
		
		private bool _inactivityNotificationSent = false;
		private int _checkInactivityEveryMin = 1;

		public MainWindow()
		{
			InitializeComponent();

			this.Closing += MainWindow_Closing;
			this.Loaded += MainWindow_Loaded;
			_startupDir = Directory.GetCurrentDirectory();
		}

		void MainWindow_Loaded(object sender, RoutedEventArgs e)
		{
			LoadProfiles();
			_updateTimer = new DispatcherTimer();
			_updateTimer.Interval = TimeSpan.FromMilliseconds(1000);
			_updateTimer.Tick += (o, ea) => {
				lock (_logLock) {
					DateTime date = DateTime.Now - TimeSpan.FromMinutes(1);
					ArticlesLog.Text = string.Join("\n", _articlesKeyToOutput.Values.ToList()) + string.Format("\nAvg. added articles: {0:N0}/min", _counts.GetMinuteCount(_addedArtCount, date));
					Clustering1Log.Text = string.Join("\n", _clustering1KeyToOutput.Values.ToList()) + string.Format("\nAvg. clustered articles: {0:N0}/min", _counts.GetMinuteCount(_addedClust1Count, date));
					Clustering2Log.Text = string.Join("\n", _clustering2KeyToOutput.Values.ToList()) + string.Format("\nAvg. clustered articles: {0:N0}/min", _counts.GetMinuteCount(_addedClust2Count, date));
					CategoriesLog.Text = string.Format("Categorized stories on average: {0:N0}/min", _counts.GetMinuteCount(_categoriesCount, date));
					BuildLog.Text = _buildLastLog + string.Format("\nAvg processed files per minute: {0:N0}/min", _counts.GetMinuteCount(_buildCount, date));
				}
			};
			_updateTimer.Start();

			_validateActivity = new DispatcherTimer();
			_validateActivity.Interval = TimeSpan.FromMinutes(_checkInactivityEveryMin);
			_validateActivity.Tick += (o, ea) => {
				if (NotifyOnNoActivityCB.IsChecked.Value == false)
					return;
				lock (_logLock) {
					try {
						// if we've not event started adding the articles (even though the management app is running) then don't send notifications at all
						if (_counts.ValuesForMinuteKey(_addedArtCount) == 0)
							return;
						// if we've intentionally close the app (backup or any other reason) then don't sent the notification
						if (_articlesDesiredStop == true)
							return;
						int sum = 0;
						for (int n = 0; n < 30; n++) {
							DateTime date = DateTime.Now - TimeSpan.FromMinutes(n);
							sum += _counts.GetMinuteCountInt(_addedArtCount, date);
						}
						if (sum == 0 && !string.IsNullOrEmpty(GetNotifyRecipients()) && _inactivityNotificationSent == false) {						
							string content = "In the last 30 minutes there haven't been any articles added to the articles application.\n\n" + GetHourlyActivity();
							SendEmailNotification("Detected inactivity for > 30 minutes", content);
							
							_inactivityNotificationSent = true;
						}
						// once the service is up again reset the flag so that we can send notification again if it falls down again
						else if (sum > 0)
							_inactivityNotificationSent = false;
					}
					catch (Exception ex) {
						LogEvent(ex.Message);
					}
				}
			};
			_validateActivity.Start();
		}
		
		void MainWindow_Closing(object sender, System.ComponentModel.CancelEventArgs e)
		{
			if (_articlesApp != null || _clustering1App != null || _clustering2App != null || _buildApp != null || _categoriesApp != null) {
				var result = MessageBox.Show("Some instances seem to still be running. Are you sure you want to close the application?", "Warning", MessageBoxButton.YesNo, MessageBoxImage.Warning);
				if (result == MessageBoxResult.No) {
					e.Cancel = true;
					return;
				}
			}
			_updateTimer.Stop();
			_updateTimer = null;

			if (_backupTimer != null)
				_backupTimer.Stop();
			_backupTimer = null;

			if (_checkIfClosedtimer != null)
				_checkIfClosedtimer.Stop();
			_checkIfClosedtimer = null;

			if (_validateActivity != null)
				_validateActivity.Stop();
			_validateActivity = null;

			RememberCurrentProfileSettings();
			SaveProfiles();
		}

		#region profile management	
		private void LoadProfiles()
		{
			if (!File.Exists(Path.Combine(_startupDir, _settingsFName))) {
				_profiles = new ProfileList();
				_profiles.Add(new Profile());
			}
			else {
				string text = File.ReadAllText(Path.Combine(_startupDir, _settingsFName));
				_profiles = SimpleJson.SimpleJson.DeserializeObject<ProfileList>(text);
			}
			ProfilesCombo.Items.Clear();
			foreach (var profile in _profiles)
				ProfilesCombo.Items.Add(profile.Name);
			ProfilesCombo.SelectedIndex = 0;
		}

		private void SaveProfiles()
		{
			if (_profiles == null || _profiles.Count == 0)
				return;
			string text = SimpleJson.SimpleJson.SerializeObject(_profiles);
			File.WriteAllText(Path.Combine(_startupDir, _settingsFName), text);
		}

		private void RememberCurrentProfileSettings()
		{
			if (_profiles == null)
				return;
			if (_currentProfileIndex < 0 || _currentProfileIndex >= _profiles.Count)
				return;
			Profile profile = _profiles[_currentProfileIndex];
			profile.Path = PathTB.Text;
			profile.ArticlesArgs = ArticlesTB.Text;
			profile.Clusters1Args = Clustering1TB.Text;
			profile.Clusters2Args = Clustering2TB.Text;
			profile.CategoriesArgs = CategoriesTB.Text;
			profile.BuildArgs = BuildTB.Text;

			profile.BuildRestartOnCrash = BuildRestartCB.IsChecked.Value;
			profile.CategoriesRestartOnCrash = CategoriesRestartCB.IsChecked.Value;

			profile.NotifyOnNoActivity = NotifyOnNoActivityCB.IsChecked.Value;
			profile.NotifyOnCrash = NotifyOnCrashCB.IsChecked.Value;

			profile.FirstBackupOn = BackupFirstTP.Value.Value.ToString();
			float backupVal = 24;
			float.TryParse(BackupEveryTB.Text, out backupVal);
			profile.BackupEveryHours = backupVal;
		}

		private void SelectProfile(string profileName)
		{
			RememberCurrentProfileSettings();
			if (!_profiles.Any(p => p.Name == profileName ))
				return;
			Profile profile = _profiles.First(p => p.Name == profileName);
			PathTB.Text = profile.Path;
			ArticlesTB.Text = profile.ArticlesArgs;
			Clustering1TB.Text = profile.Clusters1Args;
			Clustering2TB.Text = profile.Clusters2Args;
			CategoriesTB.Text = profile.CategoriesArgs;
			BuildTB.Text = profile.BuildArgs;

			BuildRestartCB.IsChecked = profile.BuildRestartOnCrash;
			CategoriesRestartCB.IsChecked = profile.CategoriesRestartOnCrash;

			NotifyOnNoActivityCB.IsChecked = profile.NotifyOnNoActivity;
			NotifyOnCrashCB.IsChecked = profile.NotifyOnCrash;

			DateTime time;
			if (!DateTime.TryParse(profile.FirstBackupOn, out time))
				time = DateTime.Now + TimeSpan.FromHours(1);
			BackupFirstTP.Value = time;
			BackupEveryTB.Text = profile.BackupEveryHours.ToString();
			
			_currentProfileIndex = _profiles.IndexOf(profile);
		}

		private void AddProfile_Click(object sender, RoutedEventArgs e)
		{
			InputBox box = new InputBox();
			box.Title = "Add New Profile";
			box.SetLabel("Profile name:");
			box.SetText("");
			if (box.ShowDialog() == true)
			{
				string profileName = box.GetText();
				if (_profiles.Any(profile => profile.Name == profileName)) {
					MessageBox.Show("Profile with this name already exists. Choose a unique name.");
					return;
				}
				_profiles.Add(new Profile() { Name = profileName });
				int index = ProfilesCombo.Items.Add(profileName);
				ProfilesCombo.SelectedIndex = index;
			}
		}

		private void DeleteProfile_Click(object sender, RoutedEventArgs e)
		{
			if (_profiles.Count == 1) {
				MessageBox.Show("This is the only profile. Since you have to have at least one profile, this one cannot be deleted.");
				return;
			}
			string profileName = (string) ProfilesCombo.SelectedValue;
			_profiles.RemoveAll(profile => profile.Name == profileName);
			ProfilesCombo.Items.Remove(profileName);
			ProfilesCombo.SelectedIndex = 0;
		}

		private void ProfilesCombo_SelectionChanged(object sender, SelectionChangedEventArgs e)
		{
			SelectProfile((string)ProfilesCombo.SelectedValue);
		}
		#endregion
				
		#region ui updates
		private void SetRunning(TextBlock control)
		{
			this.Dispatcher.Invoke((Action)(() => {
				control.Background = new SolidColorBrush(Color.FromArgb(255, 186, 255, 186));
			}));
		}

		private void SetNotRunning(TextBlock control)
		{
			this.Dispatcher.Invoke((Action)(() => {
				control.Background = new SolidColorBrush(Color.FromArgb(255, 226, 226, 226));
			}));
		}

		private void SetEnabled(bool enabled, Control control1, Control control2 = null)
		{
			this.Dispatcher.Invoke((Action)(() => { 
			control1.IsEnabled = enabled;
			if (control2 != null)
				control2.IsEnabled = enabled;
			}));
		}
		#endregion

		// store date to count
		private Counts _counts = new Counts();

		#region articles
		private ConsoleApp _articlesApp = null;
		private Dictionary<string, string> _articlesKeyToOutput = new Dictionary<string, string>();
		private List<string> _articlesLastLog = new List<string>();
		private const string _addedArtCount = "articleAddedCount";
		private bool _articlesDesiredStop = false;

		private void ArticlesStart_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_articlesApp != null) {
					LogEvent("Articles app already running. Not starting another one...");
					return;
				}
				string rootDir = "", args = "";
				this.Dispatcher.Invoke((Action)(() => {
					rootDir = PathTB.Text;
					args = ArticlesTB.Text;
				}));
				if (args.Trim() == "") {
					LogEvent("Articles instance not started. No parameters provided.");
					return;
				}
				// if app is running, then make the web request to close it
				Regex queueReg = new Regex(@"(?<clustering>.*) queue size:\s*(?<queue>\d+)\s*\.\s*Completed:\s*(?<completed>\d+)\s*\.\s*Failed:\s*(?<failed>\d+)\s*");
				Regex addedArtCount = new Regex(@"(?<addedCount>\d+) new articles \(last id: (?<lastId>\d+)\)");

				Directory.SetCurrentDirectory(rootDir);
				
				_articlesDesiredStop = false;
				_articlesApp = new ConsoleApp("ServerArticles.exe", args);
				_articlesApp.ConsoleOutput += (s, consoleArgs) => {
					var line = consoleArgs.Line;
					var lineShort = line.Substring(0, Math.Min(line.Length, 200));
					Match match1 = queueReg.Match(lineShort);
					Match match2 = addedArtCount.Match(lineShort);
					if (match1.Success) {
						string clustering = match1.Groups["clustering"].Value;
						int queue = int.Parse(match1.Groups["queue"].Value);
						int completed = int.Parse(match1.Groups["completed"].Value);
						int failed = int.Parse(match1.Groups["failed"].Value);
						
						lock (_logLock) {
							_articlesKeyToOutput[clustering] = string.Format("{0} queue size: {1:N0}, completed: {2:N0}, failed: {3:N0}", clustering, queue, completed, failed);
						}
						_counts.Add(clustering, 1);
					}
					else if (match2.Success) {
						int added = int.Parse(match2.Groups["addedCount"].Value);
						int lastId = int.Parse(match2.Groups["lastId"].Value);
						_counts.Add(_addedArtCount, added);
						lock (_logLock) {
							_articlesKeyToOutput["lastId"] = string.Format("Last article id: {0:N0}", lastId);
						}
					}
					// only add to log the non-queue size lines
					else
						_articlesLastLog.Add(DateTime.Now.ToString("[yyyy-MM-dd H:mm] ") + line);
					while (_articlesLastLog.Count > _lastLogToKeep)
						_articlesLastLog.RemoveAt(0);
				};
				_articlesApp.Exited += (s, consoleArgs) => {
					LogEvent("Articles instance exited.");
					_articlesApp = null;
					SetEnabled(false, ArticlesStop, ArticlesKill);
					SetNotRunning(ArticlesLog);
					if (_articlesDesiredStop == false && IsNotifyOnCrash())
						SendEmailNotification("Articles service unexpectedly exited.", "");
				};
				_articlesApp.Run();
				LogEvent("Articles instance started.");
				SetRunning(ArticlesLog);
				SetEnabled(true, ArticlesStop, ArticlesKill);
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void ArticlesStop_Click(object sender, RoutedEventArgs e)
		{
			try {
				Regex portReg = new Regex(@".*-port:(?<port>\d+)\s+");
				var args = "";
				this.Dispatcher.Invoke((Action)(() => {
					args = ArticlesTB.Text;
				}));
				Match match = portReg.Match(args);
				if (!match.Success) {
					MessageBox.Show("Unable to find the -port: argument in the articles text box");
					return;
				}
				_articlesDesiredStop = true;
				LogEvent("Calling stop on articles instance...");
				string port = match.Groups["port"].Value;
				string ret = GenLib.Web.Utility.GetWebPageContent(string.Format("http://localhost:{0}/exit", port));
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void ArticlesKill_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_articlesApp == null)
					return;
				if (MessageBox.Show("Are you sure you want to kill the running application? This will likely result in corrupted data.", "Confirm", MessageBoxButton.YesNo, MessageBoxImage.Warning) == MessageBoxResult.Yes) {
					_articlesDesiredStop = true;
					_articlesApp.Stop(forceCloseMillisecondsTimeout: 3000);
					LogEvent("Killed articles instance.");
					SetEnabled(false, ArticlesStop, ArticlesKill);
					SetNotRunning(ArticlesLog);
				}
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void ArticlesLogButton_Click(object sender, RoutedEventArgs e)
		{
			try {
				LogDialog dlg = new LogDialog();
				foreach (var line in _articlesLastLog)
					dlg.LogListView.Items.Add(line);
				var dict = _counts.GetMinuteCounts(_addedArtCount);
				if (dict == null) dict = new Dictionary<string, int>();

				foreach (string item in GetHourlyActivity(false).Split("\n".ToCharArray()))
					dlg.SpeedListView.Items.Add(item.Replace("\n", ""));
				dlg.SpeedListView.Items.Add("Added articles by minute:");
				foreach (var k in dict.Keys.OrderBy(k => k)) {
					dlg.SpeedListView.Items.Add(k + ": " + dict[k]);
				}
				dlg.Show();
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}
		#endregion

		#region clustering #1
		private ConsoleApp _clustering1App = null;
		private Dictionary<string, string> _clustering1KeyToOutput = new Dictionary<string, string>();
		private List<string> _clustering1Log = new List<string>();
		private const string _addedClust1Count = "addedClust1Count";
		private bool _clustering1DesiredStop = false;

		private void Çlustering1Start_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_clustering1App != null) {
					LogEvent("Clustering #1 app already running. Not starting another one...");
					return;
				}
				string rootDir = "", args = "";
				this.Dispatcher.Invoke((Action)(() => {
					rootDir = PathTB.Text;
					args = Clustering1TB.Text;
				}));
				if (args.Trim() == "") {
					LogEvent("Clustering instance #1 not started. No parameters provided.");
					return;
				}
				// if app is running, then make the web request to close it
				Regex addArtreg = new Regex(@".* Add article: id = ""(?<artId>\d+)"", lang\s*=\s*""(?<lang>.*)"",\s*(?<chars>\d+)\s*chars,.*concepts\s*=\s*(?<concepts>\d+)");
				Regex queueReg = new Regex(@"(?<lang>.*) queue size:\s*(?<queue>\d+)\s*\.\s*Completed:\s*(?<completed>\d+)\s*\.\s*Failed:\s*(?<failed>\d+)\s*");

				Directory.SetCurrentDirectory(rootDir);
				_clustering1DesiredStop = false;

				_clustering1App = new ConsoleApp("ServerClusters.exe", args);
				_clustering1App.ConsoleOutput += (s, consoleArgs) => {
					var line = consoleArgs.Line;
					if (line.Length < 3)		// skip the lines \ | / ...
						return;
					var lineShort = line.Substring(0, Math.Min(line.Length, 200));
					Match match1 = addArtreg.Match(lineShort);
					if (match1.Success) {
						int artId = int.Parse(match1.Groups["artId"].Value);
						string lang = match1.Groups["lang"].Value;
						int chars = int.Parse(match1.Groups["chars"].Value);
						int concepts = int.Parse(match1.Groups["concepts"].Value);
						
						lock (_logLock) {
							_clustering1KeyToOutput[_addedClust1Count] = string.Format("Added article id {0:N0}, lang: {1}, chars: {2:N0}, concepts: {3:N0}", artId, lang, chars, concepts);
						}
						_counts.Add(_addedClust1Count, 1);
					}
					Match match2 = queueReg.Match(lineShort);
					if (match2.Success) {
						string lang = match2.Groups["lang"].Value;
						int queue = int.Parse(match2.Groups["queue"].Value);
						int completed = int.Parse(match2.Groups["completed"].Value);
						int failed = int.Parse(match2.Groups["failed"].Value);
						
						lock (_logLock) {
							_clustering1KeyToOutput[lang] = string.Format("{0} queue to articles: {1:N0}, Completed: {2:N0}, Failed: {3:N0}", lang, queue, completed, failed);
						}
					}
					if (!match1.Success && !match2.Success)
						_clustering1Log.Add(DateTime.Now.ToString("[yyyy-MM-dd H:mm] ") + line);					
					while (_clustering1Log.Count > _lastLogToKeep)
						_clustering1Log.RemoveAt(0);
				};
				_clustering1App.Exited += (s, consoleArgs) => {
					LogEvent("Clustering #1 instance exited.");
					_clustering1App = null;
					SetEnabled(false, Clustering1Stop, Clustering1Kill);
					SetNotRunning(Clustering1Log);
					if (_clustering1DesiredStop == false && IsNotifyOnCrash())
						SendEmailNotification("Clustering #1 service unexpectedly exited.", "");
				};
				_clustering1App.Run();
				LogEvent("Clustering #1 instance started.");
				SetEnabled(true, Clustering1Stop, Clustering1Kill);
				SetRunning(Clustering1Log);
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void Clustering1Stop_Click(object sender, RoutedEventArgs e)
		{
			try {
				Regex portReg = new Regex(@".*-port:(?<port>\d+)\s+");
				var args = "";
				this.Dispatcher.Invoke((Action)(() => {
					args = Clustering1TB.Text;
				}));
				Match match = portReg.Match(args);
				if (!match.Success) {
					MessageBox.Show("Unable to find the -port: argument in the clustering #1 text box");
					return;
				}
				_clustering1DesiredStop = true;
				LogEvent("Calling stop on clustering #1 instance...");
				string port = match.Groups["port"].Value;
				string ret = GenLib.Web.Utility.GetWebPageContent(string.Format("http://localhost:{0}/exit", port));
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void Clustering1Kill_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_clustering1App == null)
					return;
				if (MessageBox.Show("Are you sure you want to kill the running application? This will likely result in corrupted data.", "Confirm", MessageBoxButton.YesNo, MessageBoxImage.Warning) == MessageBoxResult.Yes) {
					_clustering1DesiredStop = true;
					_clustering1App.Stop(forceCloseMillisecondsTimeout: 3000);
					LogEvent("Killed clustering #1 instance.");
					SetEnabled(false, Clustering1Stop, Clustering1Kill);
					SetNotRunning(Clustering1Log);
				}
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void Clustering1LogButton_Click(object sender, RoutedEventArgs e)
		{
			try {
				LogDialog dlg = new LogDialog();
				foreach (var line in _clustering1Log)
					dlg.LogListView.Items.Add(line);
				var dict = _counts.GetMinuteCounts(_addedClust1Count);
				if (dict == null) dict = new Dictionary<string, int>();

				foreach (string item in GetHourlyActivity(false).Split("\n".ToCharArray()))
					dlg.SpeedListView.Items.Add(item.Replace("\n", ""));
				dlg.SpeedListView.Items.Add("Articles clustered by minute:");
				foreach (var k in dict.Keys.OrderBy(k => k)) {
					dlg.SpeedListView.Items.Add(k + ": " + dict[k]);
				}
				dlg.Show();
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}
		#endregion

		#region clustering #2
		private ConsoleApp _clustering2App = null;
		private Dictionary<string, string> _clustering2KeyToOutput = new Dictionary<string, string>();
		private List<string> _clustering2Log = new List<string>();
		private const string _addedClust2Count = "addedClust2Count";
		private bool _clustering2DesiredStop = false;

		private void Çlustering2Start_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_clustering2App != null) {
					LogEvent("Clustering #2 app already running. Not starting another one...");
					return;
				}
				string rootDir = "", args = "";
				this.Dispatcher.Invoke((Action)(() => {
					rootDir = PathTB.Text;
					args = Clustering2TB.Text;
				}));
				if (args.Trim() == "") {
					LogEvent("Clustering instance #2 not started. No parameters provided.");
					return;
				}
				// if app is running, then make the web request to close it
				Regex addArtreg = new Regex(@".* Add article: id = ""(?<artId>\d+)"", lang\s*=\s*""(?<lang>.*)"",\s*(?<chars>\d+)\s*chars,.*concepts\s*=\s*(?<concepts>\d+)");
				Regex queueReg = new Regex(@"(?<lang>.*) queue size:\s*(?<queue>\d+)\s*\.\s*Completed:\s*(?<completed>\d+)\s*\.\s*Failed:\s*(?<failed>\d+)\s*");

				Directory.SetCurrentDirectory(rootDir);
				_clustering2DesiredStop = false;

				_clustering2App = new ConsoleApp("ServerClusters.exe", args);
				_clustering2App.ConsoleOutput += (s, consoleArgs) => {
					var line = consoleArgs.Line;
					if (line.Length < 3)		// skip the lines \ | / ...
						return;
					var lineShort = line.Substring(0, Math.Min(line.Length, 200));
					Match match1 = addArtreg.Match(lineShort);
					if (match1.Success) {
						int artId = int.Parse(match1.Groups["artId"].Value);
						string lang = match1.Groups["lang"].Value;
						int chars = int.Parse(match1.Groups["chars"].Value);
						int concepts = int.Parse(match1.Groups["concepts"].Value);
						
						lock (_logLock) {
							_clustering2KeyToOutput[_addedClust2Count] = string.Format("Added article id {0:N0}, lang: {1}, chars: {2:N0}, concepts: {3:N0}", artId, lang, chars, concepts);
						}
						_counts.Add(_addedClust2Count, 1);
					}
					Match match2 = queueReg.Match(lineShort);
					if (match2.Success) {
						string lang = match2.Groups["lang"].Value;
						int queue = int.Parse(match2.Groups["queue"].Value);
						int completed = int.Parse(match2.Groups["completed"].Value);
						int failed = int.Parse(match2.Groups["failed"].Value);
						lock (_logLock) {
							_clustering2KeyToOutput[lang] = string.Format("{0} queue to articles: {1:N0}, Completed: {2:N0}, Failed: {3:N0}", lang, queue, completed, failed);
						}
					}
					if (!match1.Success && !match2.Success)
						_clustering2Log.Add(DateTime.Now.ToString("[yyyy-MM-dd H:mm] ") + line);
										
					while (_clustering2Log.Count > _lastLogToKeep)
						_clustering2Log.RemoveAt(0);
				};
				_clustering2App.Exited += (s, consoleArgs) => {
					LogEvent("Clustering #2 instance exited.");
					_clustering2App = null;
					SetEnabled(false, Clustering2Stop, Clustering2Kill);
					SetNotRunning(Clustering2Log);
					if (_clustering2DesiredStop == false && IsNotifyOnCrash())
						SendEmailNotification("Clustering #2 service unexpectedly exited.", "");
				};
				_clustering2App.Run();
				LogEvent("Clustering #2 instance started.");
				SetEnabled(true, Clustering2Stop, Clustering2Kill);
				SetRunning(Clustering2Log);
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void Clustering2Stop_Click(object sender, RoutedEventArgs e)
		{
			try {
				Regex portReg = new Regex(@".*-port:(?<port>\d+)\s+");
				var args = "";
				this.Dispatcher.Invoke((Action)(() => {
					args = Clustering2TB.Text;
				}));
				Match match = portReg.Match(args);
				if (!match.Success) {
					MessageBox.Show("Unable to find the -port: argument in the clustering #2 text box");
					return;
				}
				_clustering2DesiredStop = true;					
				LogEvent("Calling stop on clustering #2 instance...");
				string port = match.Groups["port"].Value;
				string ret = GenLib.Web.Utility.GetWebPageContent(string.Format("http://localhost:{0}/exit", port));
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void Clustering2Kill_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_clustering2App == null)
					return;
				if (MessageBox.Show("Are you sure you want to kill the running application? This will likely result in corrupted data.", "Confirm", MessageBoxButton.YesNo, MessageBoxImage.Warning) == MessageBoxResult.Yes) {
					_clustering2DesiredStop = true;
					_clustering2App.Stop(forceCloseMillisecondsTimeout: 3000);
					LogEvent("Killed clustering #2 instance.");
					SetEnabled(false, Clustering2Stop, Clustering2Kill);
					SetNotRunning(Clustering2Log);
				}
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void Clustering2LogButton_Click(object sender, RoutedEventArgs e)
		{
			try {
				LogDialog dlg = new LogDialog();
				foreach (var line in _clustering2Log)
					dlg.LogListView.Items.Add(line);
				var dict = _counts.GetMinuteCounts(_addedClust2Count);
				if (dict == null) dict = new Dictionary<string, int>();

				foreach (string item in GetHourlyActivity(false).Split("\n".ToCharArray()))
					dlg.SpeedListView.Items.Add(item.Replace("\n", ""));
				dlg.SpeedListView.Items.Add("Articles clustered by minute:");
				foreach (var k in dict.Keys.OrderBy(k => k)) {
					dlg.SpeedListView.Items.Add(k + ": " + dict[k]);
				}
				dlg.Show();
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}
		#endregion


		#region categories
		private ConsoleApp _categoriesApp = null;
		private List<string> _categoriesLog = new List<string>();
		private const string _categoriesCount = "categoriesCount";
		private bool _categoriesWantedKill = false;

		private void CategoriesStart_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_categoriesApp != null) {
					LogEvent("Categories app already running. Not starting another one...");
					return;
				}
				string rootDir = "", args = "";
				this.Dispatcher.Invoke((Action)(() => {
					rootDir = PathTB.Text;
					args = CategoriesTB.Text;
				}));
				Directory.SetCurrentDirectory(rootDir);
				_categoriesWantedKill = false;
				_categoriesApp = new ConsoleApp("ServerCategories.exe", args);
				_categoriesApp.ConsoleOutput += (s, consoleArgs) => {
					var line = consoleArgs.Line;
					if (line.StartsWith("Classification finished."))
						_counts.Add(_categoriesCount, 1);
					_categoriesLog.Add(DateTime.Now.ToString("[yyyy-MM-dd H:mm] ") + line);

					while (_categoriesLog.Count > _lastLogToKeep)
						_categoriesLog.RemoveAt(0);
				};
				_categoriesApp.Exited += (s, consoleArgs) => {
					LogEvent("Categories instance exited.");
					_categoriesApp = null;
					SetEnabled(false, CategoriesKill);
					SetNotRunning(CategoriesLog);
					// if the build process exited, but it wasn't triggered by our kill button, then it's likely that something crashed.
					bool restart = false;
					this.Dispatcher.Invoke((Action)(() => { restart = CategoriesRestartCB.IsChecked.Value; }));
					if (!_categoriesWantedKill && restart) {
						LogEvent("Categories process ended although we didn't press kill button. Restaring the process...");
						CategoriesStart_Click(sender, e);
					}
				};
				_categoriesApp.Run();
				LogEvent("Categories instance started.");
				SetEnabled(true, CategoriesKill);
				SetRunning(CategoriesLog);
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void CategoriesKill_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_categoriesApp == null)
					return;
				_categoriesApp.Stop(forceCloseMillisecondsTimeout: 3000);
				_categoriesWantedKill = true;
				LogEvent("Killed categories instance.");
				SetEnabled(false, CategoriesKill);
				SetNotRunning(CategoriesLog);
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void CategoriesLogButton_Click(object sender, RoutedEventArgs e)
		{
			try {
				LogDialog dlg = new LogDialog();
				foreach (var line in _categoriesLog)
					dlg.LogListView.Items.Add(line);
				var dict = _counts.GetMinuteCounts(_categoriesCount);
				if (dict == null) dict = new Dictionary<string, int>();

				foreach (string item in GetHourlyActivity(false).Split("\n".ToCharArray()))
					dlg.SpeedListView.Items.Add(item.Replace("\n", ""));
				dlg.SpeedListView.Items.Add("Computed categorizations by minute:");
				foreach (var k in dict.Keys.OrderBy(k => k)) {
					dlg.SpeedListView.Items.Add(k + ": " + dict[k]);
				}
				dlg.Show();
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}
		#endregion

		#region build
		private ConsoleApp _buildApp = null;
		private string _buildLastLog = "";
		private List<string> _buildLog = new List<string>();
		private const string _buildCount = "buildCount";
		private bool _buildWantedKill = false;

		private void BuildStart_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_buildApp != null) {
					LogEvent("Build app already running. Not starting another one...");
					return;
				}
				string rootDir = "", args = "";
				this.Dispatcher.Invoke((Action)(() => {
					rootDir = PathTB.Text;
					args = BuildTB.Text;
				}));
				Directory.SetCurrentDirectory(rootDir);
				_buildWantedKill = false;
				_buildApp = new ConsoleApp("Build.exe", args);
				_buildApp.ConsoleOutput += (s, consoleArgs) => {
					var line = consoleArgs.Line;
					if (line.StartsWith(" from ")) {
						_counts.Add(_buildCount, 1);
						_buildLastLog = "Processed articles: " + line.Replace(" from ", "");
					}
					else if (line.StartsWith("Sleeping for ")) {
						_buildLastLog = DateTime.Now.ToString("[H:mm:ss] ") + line;
					}
					_buildLog.Add(DateTime.Now.ToString("[yyyy-MM-dd H:mm] ") + line);

					while (_buildLog.Count > _lastLogToKeep)
						_buildLog.RemoveAt(0);
				};
				_buildApp.Exited += (s, consoleArgs) => {
					LogEvent("Build instance exited.");
					_buildApp = null;
					SetEnabled(false, BuildKill);
					SetNotRunning(BuildLog);
					// if the build process exited, but it wasn't triggered by our kill button, then it's likely that something crashed.
					bool restart = false;
					this.Dispatcher.Invoke((Action)(() => { restart = BuildRestartCB.IsChecked.Value; }));
					if (!_buildWantedKill && restart) {
						LogEvent("Build process ended although we didn't press kill button. Restaring the process...");
						BuildStart_Click(sender, e);
					}
				};
				_buildApp.Run();
				LogEvent("Build instance started.");
				SetEnabled(true, BuildKill);
				SetRunning(BuildLog);
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void BuildKill_Click(object sender, RoutedEventArgs e)
		{
			try {
				if (_buildApp == null)
					return;
				_buildWantedKill = true;
				_buildApp.Stop(forceCloseMillisecondsTimeout: 3000);
				LogEvent("Killed build instance.");
				SetEnabled(false, BuildKill);
				SetNotRunning(BuildLog);
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}

		private void BuildLogButton_Click(object sender, RoutedEventArgs e)
		{
			try {
				LogDialog dlg = new LogDialog();
				foreach (var line in _buildLog)
					dlg.LogListView.Items.Add(line);
				var dict = _counts.GetMinuteCounts(_buildCount);
				if (dict == null) dict = new Dictionary<string, int>();

				foreach (string item in GetHourlyActivity(false).Split("\n".ToCharArray()))
					dlg.SpeedListView.Items.Add(item.Replace("\n", ""));
				dlg.SpeedListView.Items.Add("Processed files by minute:");
				foreach (var k in dict.Keys.OrderBy(k => k)) {
					dlg.SpeedListView.Items.Add(k + ": " + dict[k]);
				}
				dlg.Show();
			}
			catch (Exception ex) {
				LogEvent("Exception: " + ex.Message);
			}
		}
		#endregion

		#region auto backup
		private System.Timers.Timer _backupTimer = null;
		private System.Timers.Timer _checkIfClosedtimer = null;
		private BackupContent _backupContent;
		private string _autoBackupFolder = "_backups_auto_";
		private string _manualBackupFolder = "_backups_manual_";
				
		private void ButtonStartTimer_Click(object sender, RoutedEventArgs e)
		{
			if (GetRunningAppCount() == 0) {
				MessageBox.Show("There are currently no applications running. In order to use the Auto backup feature you have to run the desired applications.");
				return;
			}
			_backupContent = new BackupContent() { Articles = BackupArticlesCB.IsChecked.Value, Clusters = BackupClustersCB.IsChecked.Value, GeoData = BackupGeoDataCB.IsChecked.Value, Log = BackupPerformanceCB.IsChecked.Value };
			DateTime FirstTime = DateTime.Now.Date + BackupFirstTP.Value.Value.TimeOfDay;
			TimeSpan FirstSpan = FirstTime - DateTime.Now;
			// if the hour is below the current hour then we need to set the first call on the next day
			if (FirstSpan.TotalMilliseconds < 0)
				FirstSpan += TimeSpan.FromDays(1);
			float val;
			if (float.TryParse(BackupEveryTB.Text, out val) == false) {
				MessageBox.Show("'Repeat backup every' field doesn't contain a valid number.");
				return;
			}
			TimeSpan PeriodSpan = TimeSpan.FromHours(val);
			//TimeSpan PeriodSpan = TimeSpan.FromMinutes(val);

			bool currentlyBackingUp = false;
			_backupTimer = new System.Timers.Timer();
			_backupTimer.Interval = (long)FirstSpan.TotalMilliseconds;
			_backupTimer.AutoReset = true;
			_backupTimer.Elapsed += (src, eventArgs) => {
				if (_backupTimer != null)
					_backupTimer.Interval = (long) PeriodSpan.TotalMilliseconds;
				// if we haven't yet finished previous backup then do nothing
				if (currentlyBackingUp)
					return;
				currentlyBackingUp = true;
				StartBackup(_backupContent, _autoBackupFolder);
				currentlyBackingUp = false;
			};
			_backupTimer.Start();
			ButtonStartTimer.IsEnabled = false;
			ButtonStopTimer.IsEnabled = true;
			LogEvent("Auto backup started.");
			LogEvent(String.Format("The first backup will be after {0}:{1}, every next one will be after {2:N1} hours", FirstSpan.Hours, FirstSpan.Minutes, PeriodSpan.TotalHours));
		}

		private void ButtonStopTimer_Click(object sender, RoutedEventArgs e)
		{
			if (_backupTimer != null)
				_backupTimer.Stop();
			_backupTimer = null;
			ButtonStartTimer.IsEnabled = true;
			ButtonStopTimer.IsEnabled = false;
		}

		private int GetRunningAppCount()
		{
			int runningCount = 0;
			if (_articlesApp != null && _articlesApp.State == AppState.Running) runningCount++;
			if (_clustering1App != null && _clustering1App.State == AppState.Running) runningCount++;
			if (_clustering2App != null && _clustering2App.State == AppState.Running) runningCount++;
			return runningCount;
		}

		private RunningApps GetRunningApps()
		{
			RunningApps apps;
			apps.Articles = (_articlesApp != null && _articlesApp.State == AppState.Running);
			apps.Clustering1 = (_clustering1App != null && _clustering1App.State == AppState.Running);
			apps.Clustering2 = (_clustering2App != null && _clustering2App.State == AppState.Running);
			apps.Categories = (_categoriesApp != null && _categoriesApp.State == AppState.Running);
			apps.Build = (_buildApp != null && _buildApp.State == AppState.Running);
			return apps;
		}

		private void StartBackup(BackupContent backupContent, string backupFolder)
		{
			try {
				LogEvent("Started backup...");
				if (_checkIfClosedtimer != null) {
					LogEvent("Previous backup didn't finish yet. Unable to start backup.");
					return;
				}
								
				int origRunningCount = GetRunningAppCount();
				if (origRunningCount == 0) {
					LogEvent("Warning: There are currently no running apps.");
				}
				RunningApps runningApps = GetRunningApps();

				ArticlesStop_Click(this, new RoutedEventArgs());
				Clustering1Stop_Click(this, new RoutedEventArgs());
				Clustering2Stop_Click(this, new RoutedEventArgs());

				string rootDir = "";
				this.Dispatcher.Invoke((Action)(() => { rootDir = PathTB.Text; }));
				_checkIfClosedtimer = new System.Timers.Timer();
				_checkIfClosedtimer.Interval = (long)TimeSpan.FromSeconds(20).TotalMilliseconds;
				_checkIfClosedtimer.AutoReset = true;
				_checkIfClosedtimer.Elapsed += (src, eventArgs) => {
					try {
						int currentRunningCount = GetRunningAppCount();
						// if we've shut down all the apps then make the backup
						if (currentRunningCount == 0) {
							// first, kill the timer so that we don't get elapsed events anymore
							if (_checkIfClosedtimer != null)
								_checkIfClosedtimer.Stop();
							_checkIfClosedtimer = null;

							// backup all the necessary files
							string dateFolder = DateTime.Now.ToString("yyyy-MM-dd HH-mm");
							if (backupContent.Articles) BackupArticles(rootDir, backupFolder, dateFolder);
							if (backupContent.Clusters) BackupClusters(rootDir, backupFolder, dateFolder);
							if (backupContent.GeoData) BackupGeoData(rootDir, backupFolder, dateFolder);
							if (backupContent.Log) BackupLog(rootDir, backupFolder, dateFolder);
							
							// start again all the necessary processes
							if (runningApps.Articles) 
								ArticlesStart_Click(this, new RoutedEventArgs());
							if (runningApps.Clustering1) 
								Çlustering1Start_Click(this, new RoutedEventArgs());
							if (runningApps.Clustering2) 
								Çlustering2Start_Click(this, new RoutedEventArgs());
						}
					}
					catch (Exception ex) {
						LogEvent(ex.Message);
					}
				};
				_checkIfClosedtimer.Start();
			}
			catch (Exception ex) {
				LogEvent(ex.Message);
			}			
		}

		private void BackupArticles(string rootFolder, string backupFolder, string dateFolder)
		{
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder));
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder, dateFolder));

			LogEvent("Backing up IndexArticles...");
			string folderName = "IndexArticles";
			GenLib.IO.CopyDirectory(Path.Combine(rootFolder, folderName), Path.Combine(rootFolder, backupFolder, dateFolder, folderName), true, true);
			LogEvent("Backing up IndexArticles finished.");
		}

		private void BackupClusters(string rootFolder, string backupFolder, string dateFolder)
		{
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder));
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder, dateFolder));

			LogEvent("Backing up IndexClusters");
			string folderName = "IndexClusters";
			HashSet<string> ignoreExtensions = new HashSet<string>() { "bkp" };
			GenLib.IO.CopyDirectory(Path.Combine(rootFolder, folderName), Path.Combine(rootFolder, backupFolder, dateFolder, folderName), true, true, ignoreExtensions);
			LogEvent("Backing up IndexClusters finished");
		}

		private void BackupGeoData(string rootFolder, string backupFolder, string dateFolder)
		{
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder));
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder, dateFolder));

			LogEvent("Backing up IndexGeo...");
			string folderName = "IndexGeo";
			GenLib.IO.CopyDirectory(Path.Combine(rootFolder, folderName), Path.Combine(rootFolder, backupFolder, dateFolder, folderName), true, true);
			LogEvent("Backing up IndexGeo finished.");
		}

		private void BackupLog(string rootFolder, string backupFolder, string dateFolder)
		{
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder));
			Directory.CreateDirectory(Path.Combine(rootFolder, backupFolder, dateFolder));

			LogEvent("Backing up log...");
			string activity = GetHourlyActivity(true);
			
			File.WriteAllText(Path.Combine(rootFolder, backupFolder, dateFolder, "performanceLog.txt"), activity);

			File.WriteAllLines(Path.Combine(rootFolder, backupFolder, dateFolder, "articlesLog.txt"), _articlesLastLog);
			File.WriteAllLines(Path.Combine(rootFolder, backupFolder, dateFolder, "clustering1Log.txt"), _clustering1Log);
			File.WriteAllLines(Path.Combine(rootFolder, backupFolder, dateFolder, "clustering2Log.txt"), _clustering2Log);
			File.WriteAllLines(Path.Combine(rootFolder, backupFolder, dateFolder, "buildLog.txt"), _buildLog);

			LogEvent("Backing up log finished.");
		}
		#endregion

		#region utils
		private void ButtonStartManualBackup_Click(object sender, RoutedEventArgs e)
		{
			BackupContent backupContent = new BackupContent() { Articles = BackupArticlesCB.IsChecked.Value, Clusters = BackupClustersCB.IsChecked.Value, GeoData = BackupGeoDataCB.IsChecked.Value, Log = BackupPerformanceCB.IsChecked.Value };
			StartBackup(backupContent, _manualBackupFolder);

			//if (GetRunningAppCount() > 0) {
			//	MessageBox.Show("Article or clustering instance is currently running. Unable to do a backup now since the data would be corrupted.");
			//	return;
			//}
			//ButtonStartManualBackup.IsEnabled = false;
			//string rootFolder = PathTB.Text;

			//Task.Factory.StartNew(a => {
			//	try {
			//		string backupFolder = DateTime.Now.ToString("yyyy-MM-dd HH-mm");
			//		LogEvent("Starting a manual backup to " + _manualBackupFolder + "\\" + backupFolder + " ...");
			//		BackupArticles(rootFolder, _manualBackupFolder, backupFolder);
			//		BackupClusters(rootFolder, _manualBackupFolder, backupFolder);
			//		BackupGeoData(rootFolder, _manualBackupFolder, backupFolder);
			//		BackupLog(rootFolder, _manualBackupFolder, backupFolder);
			//		LogEvent("Manual backup finished.");
			//	}
			//	catch (Exception ex) {
			//		LogEvent(ex.Message);
			//	}
			//	this.Dispatcher.Invoke((Action)(() => {
			//		ButtonStartManualBackup.IsEnabled = true;
			//	}));
			//}, null);
		}

		private void ButtonDoneToNew_Click(object sender, RoutedEventArgs e)
		{
			ButtonDoneToNew.IsEnabled = false;
			string rootFolder = PathTB.Text;
			Task.Factory.StartNew(a => {
				try {
					string source = Path.Combine(rootFolder, "Raw", "Done");
					string target = Path.Combine(rootFolder, "Raw", "New");
					DirectoryInfo sourceDir = new DirectoryInfo(source);
					FileInfo[] files = sourceDir.GetFiles();
					foreach (FileInfo file in files)
						file.MoveTo(Path.Combine(target, file.Name));
				}
				catch (Exception ex) {
					LogEvent(ex.Message);
				}
				this.Dispatcher.Invoke((Action)(() => {
					ButtonDoneToNew.IsEnabled = true;
				}));
			}, null);
		}
		#endregion

		private void CopyAppLog_Click(object sender, RoutedEventArgs e)
		{
			StringBuilder text = new StringBuilder();
			foreach (var item in LogListView.Items)
				text.Append((string)item + Environment.NewLine);
			Clipboard.SetDataObject(text.ToString());
		}

		private void ButtonStartAll_Click(object sender, RoutedEventArgs e)
		{
			ArticlesStart_Click(this, new RoutedEventArgs());
			Çlustering1Start_Click(this, new RoutedEventArgs());
			Çlustering2Start_Click(this, new RoutedEventArgs());
			CategoriesStart_Click(this, new RoutedEventArgs());
			BuildStart_Click(this, new RoutedEventArgs());
		}

		private void ButtonEndAll_Click(object sender, RoutedEventArgs e)
		{
			ArticlesStop_Click(this, new RoutedEventArgs());
			Clustering1Stop_Click(this, new RoutedEventArgs());
			Clustering2Stop_Click(this, new RoutedEventArgs());
			CategoriesKill_Click(this, new RoutedEventArgs());
			BuildKill_Click(this, new RoutedEventArgs());
		}

		private void ButtonDeleteIndex_Click(object sender, RoutedEventArgs e)
		{
			if (_articlesApp != null || _clustering1App != null || _clustering2App != null) {
				MessageBox.Show("Unable to delete the index files while the articles or clustering apps are running.");
				return;
			}
			if (MessageBox.Show("Are you sure you want to delete the IndexArticles, IndexClusters and IndexGeo folders?", "Confirmation", MessageBoxButton.YesNo) == MessageBoxResult.No)
				return;
			try {
				string rootFolder = PathTB.Text;
				GenLib.IO.DeleteDirectory(Path.Combine(rootFolder, "IndexArticles"));
				GenLib.IO.DeleteDirectory(Path.Combine(rootFolder, "IndexClusters"));
				GenLib.IO.DeleteDirectory(Path.Combine(rootFolder, "IndexGeo"));
			}
			catch (Exception ex) {
				LogEvent(ex.Message);
			}
		}

		#region notifications
		private string GetProfileName()
		{
			string name = "";
			this.Dispatcher.Invoke((Action)(() => {
				name = (string)ProfilesCombo.SelectedValue;
			}));
			return name;
		}

		private bool IsNotifyOnCrash()
		{
			bool notify = false;
			this.Dispatcher.Invoke((Action)(() => {
				notify = NotifyOnCrashCB.IsChecked.Value;
			}));
			return notify;
		}

		private string GetNotifyRecipients()
		{
			string recipients = "";
			this.Dispatcher.Invoke((Action)(() => {
				recipients = NotifyRecipient.Text;
			}));
			return recipients;
		}

		// send notification to the specified recipient
		private void SendEmailNotification(string subject, string content)
		{
			LogEvent("Sending notification with subject: " + subject);
			
			Task.Factory.StartNew(a => {
				try { _emailSender.Send(GetNotifyRecipients(), "[" + GetProfileName() + "] " + subject, content); }
				catch (Exception ex) { LogEvent(ex.Message); }
			}, null);
		}
		#endregion

		#region	logging methods
		private void LogEvent(string text)
		{
			this.Dispatcher.Invoke((Action)(() => {
				LogListView.Items.Add(DateTime.Now.ToString("yyyy-MM-dd H:mm:ss ") + text);
				while (LogListView.Items.Count > 1000) {
					LogListView.Items.RemoveAt(0);
				}
			}));
		}

		private string GetHourlyActivity(bool includeAppLog = true)
		{
			StringBuilder output = new StringBuilder();

			if (includeAppLog) {
				output.AppendLine("===================================");
				output.AppendLine("Application log:");

				this.Dispatcher.Invoke((Action)(() => {
					foreach (var item in LogListView.Items)
						output.AppendLine((string)item);
				}));
			}

			var dict1 = _counts.GetHourCounts(_addedArtCount);
			var dict2 = _counts.GetHourCounts(_addedClust1Count);
			var dict3 = _counts.GetHourCounts(_addedClust2Count);
			var dict4 = _counts.GetHourCounts(_categoriesCount);
			var dict5 = _counts.GetHourCounts(_buildCount);
			var allKeys = dict1.Keys.Union(dict2.Keys).Union(dict3.Keys).Union(dict3.Keys).Union(dict4.Keys);

			output.AppendLine("\n===================================\nActivity of services per hour:");
			output.AppendLine("Date\tAdded arts\tclustered arts #1\tclustered arts #2\tcomputed categorizations\tprocessed gz files");

			foreach (var key in allKeys.OrderBy(k => k)) {
				output.AppendLine(string.Format("{0}h\t{1}\t{2}\t{3}\t{4}\t{5}",
					key,
					dict1.ContainsKey(key) ? dict1[key].ToString("N0") : "?",
					dict2.ContainsKey(key) ? dict2[key].ToString("N0") : "?",
					dict3.ContainsKey(key) ? dict3[key].ToString("N0") : "?",
					dict4.ContainsKey(key) ? dict4[key].ToString("N0") : "?",
					dict5.ContainsKey(key) ? dict5[key].ToString("N0") : "?"));
			}
			return output.ToString();
		}
		#endregion
	}

	struct BackupContent { 
		public bool Articles;
		public bool Clusters;
		public bool GeoData;
		public bool Log;
	}

	struct RunningApps
	{
		public bool Articles;
		public bool Clustering1;
		public bool Clustering2;
		public bool Categories;
		public bool Build;
	}


	public class EmailSender
	{
		public string FromEmail { get; set; }
		private SmtpClient _client = new SmtpClient();
		
		public EmailSender(string fromEmail)
		{
			FromEmail = fromEmail;

			// Connecting to the server and configuring it
			_client.Host = "smtp.gmail.com";
			_client.Port = 587;
			_client.EnableSsl = true;
			// The server requires user's credentials
			// not the default credentials
			_client.UseDefaultCredentials = false;
			// Provide your credentials
			_client.Credentials = new System.Net.NetworkCredential("gleban.test@gmail.com", "gleban-test");
			_client.DeliveryMethod = SmtpDeliveryMethod.Network;
		}
		
		public void Send(string recipientEmails, string subject, string content)
		{
			MailMessage msg = new MailMessage();

			msg.From = new MailAddress(FromEmail, FromEmail);

			// To comma separated list of recipients
			msg.To.Add(recipientEmails);
			
			msg.Subject = subject;

			// You can specify a plain text or HTML contents
			msg.Body = content;
			msg.IsBodyHtml = false;

			// Use SendAsync to send the message asynchronously
			_client.Send(msg);
		}
	}

	public class Counts
	{
		private Dictionary<string, Dictionary<string, int>> _keyToMinuteCountDict;
		private Dictionary<string, Dictionary<string, int>> _keyToHourCountDict;
		private object _lock = new object();
		private string _lastSeenDay = "";

		public Counts()
		{
			_keyToMinuteCountDict = new Dictionary<string, Dictionary<string, int>>();
			_keyToHourCountDict = new Dictionary<string,Dictionary<string,int>>();
		}

		public void Add(string key, int count = 1)
		{
			lock (_lock) {
				// if we started a new day then clear the speed minutes from yesterday
				if (_lastSeenDay != DateTime.Now.ToString("yyyy-MM-dd")) {
					_lastSeenDay = DateTime.Now.ToString("yyyy-MM-dd");
					_keyToMinuteCountDict.Clear();
				}

				// update the minute data
				if (!_keyToMinuteCountDict.ContainsKey(key))
					_keyToMinuteCountDict.Add(key, new Dictionary<string, int>());
				string minuteKey = DateTime.Now.ToString("yyyy-MM-dd H:mm");
				if (!_keyToMinuteCountDict[key].ContainsKey(minuteKey))
					_keyToMinuteCountDict[key][minuteKey] = count;
				else
					_keyToMinuteCountDict[key][minuteKey] += count;

				// update the hour data
				if (!_keyToHourCountDict.ContainsKey(key))
					_keyToHourCountDict.Add(key, new Dictionary<string, int>());
				string hourKey = DateTime.Now.ToString("yyyy-MM-dd HH");
				if (!_keyToHourCountDict[key].ContainsKey(hourKey))
					_keyToHourCountDict[key][hourKey] = count;
				else
					_keyToHourCountDict[key][hourKey] += count;
			}
		}

		#region minute keys
		public string GetMinuteKey(DateTime date)
		{
			return date.ToString("yyyy-MM-dd H:mm");
		}

		public bool HasMinuteKey(string key) {
			return _keyToMinuteCountDict.ContainsKey(key);
		}

		public int ValuesForMinuteKey(string key) {
			if (!HasMinuteKey(key))
				return 0;
			return _keyToMinuteCountDict[key].Keys.Count;
		}

		public Dictionary<string, int> GetMinuteCounts(string key)
		{
			lock (_lock) {
				if (!_keyToMinuteCountDict.ContainsKey(key))
					return new Dictionary<string,int>();
				Dictionary<string, int> copy = new Dictionary<string, int>(_keyToMinuteCountDict[key]);
				return copy;
			}
		}

		public string GetMinuteCount(string key, DateTime date, string defVal = "unknown")
		{
			string dateKey = GetMinuteKey(date);
			lock (_lock) {
				if (!_keyToMinuteCountDict.ContainsKey(key))
					return defVal;
				if (!_keyToMinuteCountDict[key].ContainsKey(dateKey))
					return defVal;
				return String.Format("{0:N0}", _keyToMinuteCountDict[key][dateKey]);
			}
		}

		public int GetMinuteCountInt(string key, DateTime date)
		{
			string dateKey = GetMinuteKey(date);
			lock (_lock) {
				if (!_keyToMinuteCountDict.ContainsKey(key))
					return 0;
				if (!_keyToMinuteCountDict[key].ContainsKey(dateKey))
					return 0;
				return _keyToMinuteCountDict[key][dateKey];
			}
		}
		#endregion

		#region hour keys
		public string GetHourKey(DateTime date)
		{
			return date.ToString("yyyy-MM-dd H");
		}

		public bool HasHourKey(string key)
		{
			return _keyToHourCountDict.ContainsKey(key);
		}

		public int ValuesForHourKey(string key)
		{
			if (!HasHourKey(key))
				return 0;
			return _keyToHourCountDict[key].Keys.Count;
		}

		public Dictionary<string, int> GetHourCounts(string key)
		{
			lock (_lock) {
				if (!_keyToHourCountDict.ContainsKey(key))
					return new Dictionary<string, int>();
				Dictionary<string, int> copy = new Dictionary<string, int>(_keyToHourCountDict[key]);
				return copy;
			}
		}

		public string GetHourCount(string key, DateTime date, string defVal = "unknown")
		{
			string dateKey = GetHourKey(date);
			lock (_lock) {
				if (!_keyToHourCountDict.ContainsKey(key))
					return defVal;
				if (!_keyToHourCountDict[key].ContainsKey(dateKey))
					return defVal;
				return String.Format("{0:N0}", _keyToHourCountDict[key][dateKey]);
			}
		}

		public int GetHourCountInt(string key, DateTime date)
		{
			string dateKey = GetHourKey(date);
			lock (_lock) {
				if (!_keyToMinuteCountDict.ContainsKey(key))
					return 0;
				if (!_keyToMinuteCountDict[key].ContainsKey(dateKey))
					return 0;
				return _keyToMinuteCountDict[key][dateKey];
			}
		}
		#endregion
	}

	#region profiles serialization
	[DataContract]
	public class Profile
	{
		[DataMember]
		public string Name { get; set; }

		[DataMember]
		public string Path { get; set; }

		[DataMember]
		public string ArticlesArgs { get; set; }
		[DataMember]
		public string Clusters1Args { get; set; }
		[DataMember]
		public string Clusters2Args { get; set; }
		[DataMember]
		public string CategoriesArgs { get; set; }
		[DataMember]
		public string BuildArgs { get; set; }

		[DataMember]
		public bool BuildRestartOnCrash { get; set; }
		[DataMember]
		public bool CategoriesRestartOnCrash { get; set; }
		
		[DataMember]
		public string FirstBackupOn { get; set; }
		[DataMember]
		public float BackupEveryHours { get; set; }

		[DataMember]
		public bool NotifyOnNoActivity { get; set; }
		[DataMember]
		public bool NotifyOnCrash { get; set; }

		[DataMember]
		public string NotifyRecipientEmail { get; set; }

		public Profile()
		{
			BuildRestartOnCrash = true;
			CategoriesRestartOnCrash = true;
			Name = "Default";
			BackupEveryHours = 24;
			NotifyOnNoActivity = true;
			NotifyOnCrash = true;
			NotifyRecipientEmail = "gleban@gmail.com";
		}
	}

	[DataContract]
	public class ProfileList : List<Profile>
	{

	}
	#endregion
}
