/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "../NewsMiner/NewsBase.h"
#include "../NewsMiner/NewsSearch.h"
#include "../NewsMiner/NewsConst.h"

#include "../Srv/CategorySrv.h"
#include "../Learning/LrnStoryLinking.h"
#include "../ServerArticles/EventRegBase.h"

#include "../Utils/FileUtils.h"
#include "../NewsMiner/GeoBase.h"
#include "../Trends/ConceptDailyTrends.h"
#include "../NewsMiner/NewsFixDb.h"
#include "../ServerClusters/ClusterUtils.h"
#include "../ServerClusters/Clustering.h"

void WhereIsArticle667714() {
	TClusteringSettings ClustSett;

	TStrV ForwardV;
	PClustering clust = new TClustering(ClustSett, new TMySwSet(), "eng", "e:/Data/NewsMiner/IndexClusters/", ForwardV);
	TIntIntH DaySpanToCntH;
	for (int KeyId = clust->clusters.FFirstKeyId(); clust->clusters.FNextKeyId(KeyId);) {
		PCluster c = clust->clusters[KeyId];
		for (int N = 0; N < c->arts.Len(); N++) {
			TArticleId Id = c->arts[N];
			if (clust->articles[Id]->articleName == "660138") {
				printf("article 667714 is in %s\n", c->clusterName.CStr());
			}
		}
	}
}

void TestEngClustersTimeSpan() 
{
	TClusteringSettings ClustSett;

	TStrV ForwardV;
	PClustering clust = new TClustering(ClustSett, new TMySwSet(), "eng", "e:/Data/NewsMiner/IndexClusters/", ForwardV);
	TIntIntH DaySpanToCntH;
	for (int KeyId = clust->clusters.FFirstKeyId(); clust->clusters.FNextKeyId(KeyId);) {
		PCluster c = clust->clusters[KeyId];
		TFlt TimeMin, TimeMax;
		for (int N = 0; N < c->arts.Len(); N++) {
			TArticleId Id = c->arts[N];
			TFlt Time = clust->articles[Id]->timeStamp;
			if (N == 0) {
				TimeMin = Time; TimeMax = Time; 
			}
			if (Time < TimeMin)
				TimeMin = Time;
			if (Time > TimeMax)
				TimeMax = Time;
		}
		DaySpanToCntH.AddDat((int) (TimeMax - TimeMin)) += 1;
	}
	DaySpanToCntH.SortByKey();
	for (int KeyId = DaySpanToCntH.FFirstKeyId(); DaySpanToCntH.FNextKeyId(KeyId);) {
		printf("Day span: %d, cluster count: %d\n", DaySpanToCntH.GetKey(KeyId), DaySpanToCntH[KeyId]);
	}
}

void PrintClustersThatSpanTheMostTime(const PNewsBase& NewsBase)
{
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;
	PStore StoryStore = NewsBase->StoryStore;

	THash<TInt, TVec<TUInt64> > DaySpanToIdVH;
	PStoreIter StoryIter = StoryStore->GetIter();
	while (StoryIter->Next()) {
		uint64 StoryId = StoryIter->GetRecId();
		PRecSet ArtRecSet = StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasArticle");
		if (ArtRecSet->GetRecs() == 0)
			continue;
		ArtRecSet->SortById();
		TTm Tm1; NewsBase->ArticleStore->GetFieldNmTm(ArtRecSet->GetRecId(0), "DateTime", Tm1);
		TTm Tm2; NewsBase->ArticleStore->GetFieldNmTm(ArtRecSet->GetRecId(ArtRecSet->GetRecs() - 1), "DateTime", Tm2);
		int Diff = (int) TTm::GetDiffDays(Tm1, Tm2) + 1;	// if on the same day then span should be 1
		DaySpanToIdVH.AddDat(Diff).Add(StoryId);
	}
	DaySpanToIdVH.SortByKey(false);
	TVec<TUInt64> Clusters = DaySpanToIdVH[0];
	printf("Clusters of size %d:\n", DaySpanToIdVH.GetKey(0));
	for (int N = 0; N < Clusters.Len(); N++) {
		printf("Cluster: %s\n", StoryStore->GetRecNm(Clusters[N]).CStr());
		PRecSet ArticleSet = StoryStore->GetRec(Clusters[N]).DoJoin(Base, "hasArticle");
		TStrV ArtV;
		for (int A = 0; A < ArticleSet->GetRecs(); A++) {
			ArtV.Add(TUInt64(ArticleSet->GetRecId(A)).GetStr());
		}
		printf("Articles: %s\n\n", TStr::GetStr(ArtV, ",").CStr());
	}

}

void ValidateArticleStoryJoins(const PNewsBase& NewsBase)
{
	printf("validating article - story joins\n");
	PStoreIter Iter;
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;
	
	// do articles point to stories and the stories to the same articles?
	// we don't want Story -> article -> Story2
	Iter = NewsBase->StoryStore->GetIter();
	while (Iter->Next()) {
		uint64 StoryId = Iter->GetRecId();
		PRecSet ArticleSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
		for (int N = 0; N < ArticleSet->GetRecs(); N++) {
			uint64 ArtId = ArticleSet->GetRecId(N);
			PRecSet StoryRecSet = NewsBase->ArticleStore->GetRec(ArtId).DoJoin(NewsBase->Base, "hasStory");
			if (StoryRecSet->GetRecs() > 0) {
				if (StoryRecSet->GetRecId(0) != StoryId)
					printf("Story %I64u contains article %I64u. Article %I64u points to story %I64u.\n", StoryId, ArtId, StoryRecSet->GetRecId(0));
			}
		}
	}

	// does article's hasStory point to a single story?
	Iter = NewsBase->ArticleStore->GetIter();
	while (Iter->Next()) {
		uint64 ArtId = Iter->GetRecId();

		TRec OrigArt = NewsBase->ArticleStore->GetRec(ArtId).DoSingleJoin(Base, "isDuplicateOf");
		PRecSet StorySet = NewsBase->ArticleStore->GetRec(ArtId).DoJoin(Base, "hasStory");
		if (StorySet->GetRecs() > 1)
			printf("Article %I64u is assigned to multiple stories (%d)\n", ArtId, StorySet->GetRecs());
		if (!OrigArt.IsDef() && StorySet->GetRecs() == 0)
			printf("Article %I64u is not a duplicate and is not assigned to any stories\n", ArtId);
	}

	printf("finished\n");
}

void ValidateStoryEventJoins(const PNewsBase& NewsBase)
{
	printf("validating story - event joins\n");
	PStoreIter Iter;
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;

	// do articles point to stories and the stories to the same articles?
	// we don't want Story -> article -> Story2
	Iter = NewsBase->EventStore->GetIter();
	while (Iter->Next()) {
		uint64 EventId = Iter->GetRecId();
		PRecSet StorySet = NewsBase->EventStore->GetRec(EventId).DoJoin(Base, "hasStory");
		for (int N = 0; N < StorySet->GetRecs(); N++) {
			uint64 StoryId = StorySet->GetRecId(N);
			PRecSet EventRecSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasEvent");
			if (EventRecSet->GetRecs() > 0) {
				if (EventRecSet->GetRecId(0) != EventId)
					printf("Event %I64u contains story %I64u. Story %I64u points to event %I64u.\n", EventId, StoryId, EventRecSet->GetRecId(0));
			}
			if (EventRecSet->GetRecs() > 1) {
				printf("Story %I64u points to %d events\n", StoryId, EventRecSet->GetRecs());
			}
		}
	}
	printf("finished\n");
}

void TestArticle9259(const PNewsBase& NewsBase)
{
	PStoreIter Iter;
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;
	Iter = NewsBase->StoryStore->GetIter();
	TRec StoryRec = NewsBase->ArticleStore->GetRec(9259).DoSingleJoin(NewsBase->Base, "hasStory");
	printf("Article -> Story: Story id: %I64u (%s)\n", StoryRec.GetRecId(), NewsBase->StoryStore->GetRecNm(StoryRec.GetRecId()).CStr());
	while (Iter->Next()) {
		uint64 StoryId = Iter->GetRecId();
		PRecSet ArticleSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
		for (int N = 0; N < ArticleSet->GetRecs(); N++) {
			uint64 ArtId = ArticleSet->GetRecId(N);
			if (ArtId == 9259)
				printf("Story -> Article: Story id: %I64u (%s)\n", StoryId, NewsBase->StoryStore->GetRecNm(StoryId).CStr());
		}
	}
	printf("finished\n");
}

void UpdateStoryEventArticleCount(const PNewsBase& NewsBase, const PEventRegBase& EventRegBase)
{
	PStoreIter Iter;
	PStore StoryStore = NewsBase->StoryStore;
	PStore CategoryStore = NewsBase->CategoryStore;
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;
	Iter = StoryStore->GetIter();
	while (Iter->Next()) {
		uint64 StoryId = Iter->GetRecId();
		if (StoryId % 10000 == 0)
			printf("Story %I64u\r", StoryId);
		EventRegBase->UpdateStoryArticleCount(StoryId);
	}
	printf("finished\n");
}

void RemoveCategoriesForNonEnglishStories(const PNewsBase& NewsBase, const PEventRegBase& EventRegBase)
{
	printf("removing categories for non english stories\n");
	PStoreIter Iter;
	PStore StoryStore = NewsBase->StoryStore;
	PStore CategoryStore = NewsBase->CategoryStore;
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;
	Iter = StoryStore->GetIter();
	while (Iter->Next()) {
		uint64 StoryId = Iter->GetRecId();
		if (StoryId % 10000 == 0)
			printf("Story %I64u\r", StoryId);
		TStr Lang = EventRegBase->GetStoryLang(StoryId);
		if (Lang == "eng")
			continue;
		PRecSet CatRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasCategory");
		for (int N = 0; N < CatRecSet->GetRecs(); N++) {
			Index->DeleteJoin(StoryStore, "hasCategory", StoryId, CatRecSet->GetRecId(N));
			Index->DeleteJoin(CategoryStore, "hasStory", CatRecSet->GetRecId(N), StoryId);
		}
	}
	// propagate to the event level
	Iter = NewsBase->EventStore->GetIter();
	while (Iter->Next()) {
		uint64 EventId = Iter->GetRecId();
		if (EventId % 10000 == 0)
			printf("Event %I64u\r", EventId);
		EventRegBase->UpdateEventCategories(EventId);
	}
	printf("finished\n");
}

void UpdateStoryAndEventLocations(const PNewsBase& NewsBase, const PEventRegBase& EventRegBase)
{
	printf("updating story and event locations\n");
	PStoreIter Iter;
	PStore ArticleStore = NewsBase->ArticleStore;
	PStore StoryStore = NewsBase->StoryStore;
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;

	Iter = StoryStore->GetIter();
	while (Iter->Next()) {
		uint64 StoryId = Iter->GetRecId();
		if (StoryId % 10000 == 0)
			printf("Story %I64u\r", StoryId);
		EventRegBase->UpdateStoryConcepts(StoryId);
		EventRegBase->UpdateStoryLocation(StoryId);
	}

	Iter = NewsBase->EventStore->GetIter();
	while (Iter->Next()) {
		uint64 EventId = Iter->GetRecId();
		if (EventId % 10000 == 0)
			printf("Event %I64u\r", EventId);
		EventRegBase->UpdateEventLocation(EventId);
	}

	printf("validating if all duplicated locations were removed\n");
	// check if a single location per story and event
	Iter = StoryStore->GetIter();
	while (Iter->Next()) {
		uint64 StoryId = Iter->GetRecId();
		PRecSet LocSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasLocation");
		if (LocSet->GetRecs() > 1)
			printf("Warning: still multiple locations per story found for story %I64u\n", StoryId);
	}

	Iter = NewsBase->EventStore->GetIter();
	while (Iter->Next()) {
		uint64 EventId = Iter->GetRecId();
		PRecSet LocSet = NewsBase->EventStore->GetRec(EventId).DoJoin(Base, "hasLocation");
		if (LocSet->GetRecs() > 1)
			printf("Warning: still multiple locations per story found for event %I64u\n", EventId);
	}
	printf("finished.\n");
}

void RemoveExtraJoinsFromArticleToStory(const PNewsBase& NewsBase)
{
	PStoreIter Iter;
	PStore ArticleStore = NewsBase->ArticleStore;
	PStore StoryStore = NewsBase->StoryStore;
	PBase Base = NewsBase->Base;
	PIndex Index = NewsBase->Index;

	int ExtraJoins = 0;
	Iter = ArticleStore->GetIter();
	while (Iter->Next()) {
		uint64 ArticleId = Iter->GetRecId();
		if (ArticleId % 10000 == 0)
			printf("Article %I64u (current extra joins: %d)\r", ArticleId, ExtraJoins);
		PRecSet StoryRecSet = ArticleStore->GetRec(ArticleId).DoJoin(Base, "hasStory");
		if (StoryRecSet->GetRecs() > 1) {
			ExtraJoins += StoryRecSet->GetRecs() - 1;
			// find the correct story to which the article is currently assigned to
			uint64 CorrectStoryId = TUInt64::Mx;
			for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
				PRecSet ArticleSet = StoryRecSet->GetRec(N).DoJoin(Base, "hasArticle");
				for (int A = 0; A < ArticleSet->GetRecs(); A++) {
					if (ArticleSet->GetRecId(A) == ArticleId) {
						CorrectStoryId = StoryRecSet->GetRecId(N);		// we've found the correct story id for the article
						break;
					}
				}
				if (CorrectStoryId != TUInt64::Mx)
					break;
			}
			if (CorrectStoryId == TUInt64::Mx) {
				printf("Weird. story id still undefined.\n");
				continue;
			}
			// remove each join from article -> story where story != storyId
			for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
				uint64 StoryId = StoryRecSet->GetRecId(N);
				if (StoryId == CorrectStoryId)
					continue;
				Index->DeleteJoin(ArticleStore, "hasStory", ArticleId, StoryId);
			}
			StoryRecSet = ArticleStore->GetRec(ArticleId).DoJoin(Base, "hasStory");
			if (StoryRecSet->GetRecs() > 1)
				printf("Warning: still multiple stories for same article %I64u\n", ArticleId);
		}
	}
	printf("Found %d extra joins from articles to stories", ExtraJoins);
}

void TestJoinsArticleStory(const PNewsBase& NewsBase)
{
	PStoreIter Iter;
	PStore ArticleStore = NewsBase->ArticleStore;
	PBase Base = NewsBase->Base;

	int ExtraJoins = 0;
	Iter = ArticleStore->GetIter();
	while (Iter->Next()) {
		uint64 ArticleId = Iter->GetRecId();
		if (ArticleId % 10000 == 0)
			printf("Article %I64u (current extra joins: %d)\r", ArticleId, ExtraJoins);
		//TRec OrigArt = ArticleStore->GetRec(ArticleId).DoSingleJoin(Base, "isDuplicateOf");
		//TRec StoryRec = ArticleStore->GetRec(ArticleId).DoSingleJoin(Base, "hasStory");
		//if (StoryRec.IsDef()) {
		//	uint64 StoryId = StoryRec.GetRecId();
		//	PRecSet ArticleSet = StoryRec.DoJoin(Base, "hasArticle");
		//	bool Found = false;
		//	for (int N = 0; N < ArticleSet->GetRecs(); N++) {
		//		if (ArticleSet->GetRecId(N) == ArticleId) {
		//			Found = true;
		//			break;
		//		}
		//	}
		//	if (!Found) {
		//		//printf("Article %I64u has story %I64u, but not reverse\n", ArticleId, StoryId);
		//		Count++;
		//	}
		//}
		PRecSet StoryRecSet = ArticleStore->GetRec(ArticleId).DoJoin(Base, "hasStory");
		if (StoryRecSet->GetRecs() > 1) {
			//printf("Article %I64u has %d stories\n", ArticleId, StoryRecSet->GetRecs());
			ExtraJoins += StoryRecSet->GetRecs() - 1;
		}
	}
	printf("Found %d extra joins from articles to stories", ExtraJoins);
}

// due to a bug in news.cpp we are using storyIds assigned in the xml files. the articles that are non-duplicates then get new story id when
// we get info from clustering. the duplicated articles however keep their story id and don't get assigned a new story id.
// we want to remove storyId join for articles that are duplicates
void RemoveStoryForDuplicateArticles(const PNewsBase& NewsBase, const PEventRegBase& EventRegBase)
{
	PStoreIter Iter;
	PStore ArticleStore = NewsBase->ArticleStore;
	PBase Base = NewsBase->Base;

	Iter = ArticleStore->GetIter();
	int DelJoins = 0;
	TUInt64Set UpdatedStoryH;
	while (Iter->Next()) {
		uint64 ArticleId = Iter->GetRecId();
		TRec OrigArt = ArticleStore->GetRec(ArticleId).DoSingleJoin(Base, "isDuplicateOf");
		// if duplicate then remove story
		if (OrigArt.IsDef()) {
			TRec StoryRec = ArticleStore->GetRec(ArticleId).DoSingleJoin(Base, "hasStory");
			if (StoryRec.IsDef()) {
				DelJoins++;
				printf("Deleted %d joins\r", DelJoins);
				NewsBase->Index->DeleteJoin(ArticleStore, "hasStory", ArticleId, StoryRec.GetRecId());
				NewsBase->Index->DeleteJoin(NewsBase->StoryStore, "hasArticle", StoryRec.GetRecId(), ArticleId);
				UpdatedStoryH.AddKey(StoryRec.GetRecId());
			}
		}
	}
	printf("Done.\n");
	printf("Updating modified stories...\n");
	int Count = UpdatedStoryH.Len(), Index = 0;
	for (int KeyId = UpdatedStoryH.FFirstKeyId(); UpdatedStoryH.FNextKeyId(KeyId);) {
		printf("Updating story %d/%d\r", Index, Count);
		Index++;
		uint64 StoryId = UpdatedStoryH.GetKey(KeyId);
		EventRegBase->UpdateStoryArticleCount(StoryId);
		EventRegBase->UpdateStoryConcepts(StoryId);
		EventRegBase->UpdateStoryLocation(StoryId);
		EventRegBase->UpdateStoryDates(StoryId);
		EventRegBase->UpdateStoryTitleAndSummary(StoryId);
	}
}

void RemoveLocationJoins(const PNewsBase& NewsBase, const PGeoBase GeoBase)
{
	PStoreIter Iter;
	PStore StoryStore = NewsBase->StoryStore;
	PStore ConceptStore = NewsBase->ConceptStore;
	PStore EventStore = NewsBase->EventStore;

	Iter = ConceptStore->GetIter();
	while (Iter->Next()) {
		uint64 ConceptId = Iter->GetRecId();
		TStr ConceptNm = ConceptStore->GetRecNm(ConceptId);
		if (GeoBase->IsCity(ConceptNm) || GeoBase->IsCountry(ConceptNm))
			GeoBase->DecreaseWgt(ConceptNm, TInt::Mx);

		PRecSet StorySet = ConceptStore->GetRec(ConceptId).DoJoin(NewsBase->Base, "hasStoryForLocation");
		for (int N = 0; N < StorySet->GetRecs(); N++) {
			NewsBase->Index->DeleteJoin(StoryStore, "hasLocation", StorySet->GetRecId(N), ConceptId);
			NewsBase->Index->DeleteJoin(ConceptStore, "hasStoryForLocation", ConceptId, StorySet->GetRecId(N));
		}

		PRecSet EventSet = ConceptStore->GetRec(ConceptId).DoJoin(NewsBase->Base, "hasEventForLocation");
		for (int N = 0; N < EventSet->GetRecs(); N++) {
			NewsBase->Index->DeleteJoin(EventStore, "hasLocation", EventSet->GetRecId(N), ConceptId);
			NewsBase->Index->DeleteJoin(ConceptStore, "hasEventForLocation", ConceptId, EventSet->GetRecId(N));
		}
	}
}

void RecomputeEventCategories(const PNewsBase& NewsBase, const PEventRegBase& EventRegBase)
{
	printf("Recomputing event categories...\n");
	PStore EventStore = NewsBase->EventStore;
	PStoreIter Iter = EventStore->GetIter();
	while (Iter->Next()) {
		uint64 EventId = Iter->GetRecId();
		EventRegBase->UpdateEventCategories(EventId);
	}
	printf("Done.\n");
}

void CountArticlesWithGoogleStoryId(const PNewsBase& NewsBase) {
	PStore ArticleStore = NewsBase->ArticleStore;
	PStoreIter Iter = ArticleStore->GetIter();
	int NonNull = 0;
	int Total = 0;
	while (Iter->Next()) {
		uint64 ArticleId = Iter->GetRecId();
		Total++;
		if (!ArticleStore->IsFieldNmNull(ArticleId, "GoogleStoryId"))
			NonNull++;
	}
	printf("Non null: %d/%d (%.2f%%)\n", NonNull, Total, (100*NonNull) / (float) Total);
}

// evaluate how accurately our clustering method groups articles compared to google clustering
void EvaluateClusterToGoogleAgreement(const PNewsBase& NewsBase) {
	
	PStore ArticleStore = NewsBase->ArticleStore;
	PStoreIter Iter = ArticleStore->GetIter();
	THash<TStr, TUInt64V> GoogleIdToArtVH;
	// build a hash of google cluster it -> articles with this google cluster id
	while (Iter->Next()) {
		uint64 ArticleId = Iter->GetRecId();
		if (!ArticleStore->IsFieldNmNull(ArticleId, "GoogleStoryId")) {
			TStr GoogleId = ArticleStore->GetFieldNmStr(ArticleId, "GoogleStoryId");
			GoogleIdToArtVH.AddDat(GoogleId).Add(ArticleId);
		}
	}
	
	PMom PrecMom = TMom::New();
	PMom RecMom = TMom::New();
	PMom F1Mom = TMom::New();
	// go over each google cluster id and compute precision and recall for it
	for (int KeyId = GoogleIdToArtVH.FFirstKeyId(); GoogleIdToArtVH.FNextKeyId(KeyId); ) {
		TUInt64V ArtV = GoogleIdToArtVH[KeyId];
		// some of the articles in ArtV are likely duplicates. find the original article for them and use those, since duplicates are not assigned to a cluster
		PRecSet ArtSet = TRecSet::New(NewsBase->ArticleStore, ArtV);
		PRecSet OrigArtSet = ArtSet->DoJoin(NewsBase->Base, "isDuplicateOf");
		// extend the known set with the orginal articles. when going through the list ignore the ones that are duplicates (=don't have a story)
		for (int N=0; N < OrigArtSet->GetRecs(); N++) 
			ArtV.AddUnique(OrigArtSet->GetRecId(N));

		THash<TUInt64, TInt> StoryIdToCntH;
		int UniqueArtsInCluster = 0;
		for (int N=0; N < ArtV.Len(); N++) {
			TRec StoryRec = ArticleStore->GetRec(ArtV[N]).DoSingleJoin(NewsBase->Base, "hasStory");
			// if an article doesn't have a story then it's a duplicate. since we've already added it's original to the list we can just continue
			if (!StoryRec.IsDef()) 
				continue;
			UniqueArtsInCluster++;
			StoryIdToCntH.AddDat(StoryRec.GetRecId()) += 1;
		}
		if (StoryIdToCntH.Len() == 0)
			continue;
		StoryIdToCntH.SortByDat(false);
		int KeyId = StoryIdToCntH.FFirstKeyId(); StoryIdToCntH.FNextKeyId(KeyId);
		// TopStoryId is the story that has the highest number of articles for this google cluster
		uint64 TopStoryId = StoryIdToCntH.GetKey(KeyId);
		// find all the articles for the TopStory that have a google story id. only those articles can be used when computing the recall
		// that is because we have in our system also other articles not seen by google and for those we don't know if google would assign them to this story or not
		PRecSet ArtsForStory = NewsBase->StoryStore->GetRec(TopStoryId).DoJoin(NewsBase->Base, "hasArticle");
		int ArtsWithGoogleId = 0;
		for (int N = 0; N < ArtsForStory->GetRecs(); N++) {
			if (!NewsBase->ArticleStore->IsFieldNmNull(ArtsForStory->GetRecId(N), "GoogleStoryId"))
				ArtsWithGoogleId++;
		}
		
		int TP = StoryIdToCntH[KeyId];
		int FP = UniqueArtsInCluster - TP;

		double Prec = TP / (double) UniqueArtsInCluster;
		double Rec = TP / (double) ArtsWithGoogleId;
		double F1 = 2 * Prec * Rec / (Prec + Rec);
		if (Prec + Rec < 1e-5) 
			F1 = 0;

		// prevent division by zero
		if (ArtsWithGoogleId == 0 || UniqueArtsInCluster == 0)
			continue;

		PrecMom->Add(Prec);
		RecMom->Add(Rec);
		F1Mom->Add(F1);
	}

	PrecMom->Def(); RecMom->Def(); F1Mom->Def();
	// average over all precision and recall
	double PrecAvg = PrecMom->GetMean(), PrecStDev = PrecMom->GetSDev();
	double RecAvg = RecMom->GetMean(), RecStDev = RecMom->GetSDev();
	double F1Avg = F1Mom->GetMean(), F1StDev = F1Mom->GetSDev();
	printf("Precision: %.4f (+-%.4f)\n", PrecAvg, PrecStDev);
	printf("Recall: %.4f (+-%.4f)\n", RecAvg, RecStDev);
	printf("F1: %.4f (+-%.4f)\n", F1Avg, F1StDev);
}

void FixInvalidMediods(const PNewsBase& NewsBase) {
	PStoreIter Iter = NewsBase->StoryStore->GetIter();
	while (Iter->Next()) {
		const uint64 StoryId = Iter->GetRecId();
		const uint64 ArticleId = NewsBase->StoryStore->GetFieldNmUInt64(StoryId, "hasMedoidId");
		if (ArticleId == TUInt64::Mx)
			continue;
		if (NewsBase->ArticleStore->IsRecId(ArticleId))
			continue;
		// we have a bad article id here for a story
		printf("removing bad medoid: %I64u -> %I64u\n", StoryId, ArticleId);
		NewsBase->StoryStore->DelJoin("hasMedoid", StoryId, ArticleId);
	}
}

// remove meta information for all events and stories that no longer have any articles
void ClearStoryAndEventInfoForEmpty(const PNewsBase& NewsBase) {
	PStoreIter Iter = NewsBase->StoryStore->GetIter();
	while (Iter->Next()) {
		const uint64 StoryId = Iter->GetRecId();
		PRecSet ArticleSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasArticle");
		if (ArticleSet->GetRecs() == 0) {
			printf("Clearing story %I64u\n", StoryId);
			NewsBase->ClearStoryInformation(StoryId);
		}
	}
}

// set cosine similarity for the article duplicates to 0. strangely we found that they commonly have it set to 1
void SetCosSimForDuplicatesToZero(const PNewsBase& NewsBase)
{
	PStoreIter Iter = NewsBase->ArticleStore->GetIter();
	int DupsWithBadCosSim = 0;
	int TotalDups = 0;
	while (Iter->Next()) {
		const uint64 ArticleId = Iter->GetRecId();
		if (NewsBase->ArticleStore->GetFieldNmUInt64(ArticleId, "isDuplicateOfId") == TUInt64::Mx)
			continue;
		TotalDups++;
		const float CosSim = NewsBase->ArticleStore->GetFieldNmFlt(ArticleId, "CosToCentroid");
		if (CosSim > 0.01) {
			DupsWithBadCosSim++;
			NewsBase->ArticleStore->SetFieldNmFlt(ArticleId, "CosToCentroid", 0);
		}
		if (TotalDups % 100 == 0)
			printf("Bad dups: %d/%d\n", DupsWithBadCosSim, TotalDups);
	}

}

// count the number of duplicates that are also assigned to a cluster (value should be 0)
void CountDupsInACluster(const PNewsBase& NewsBase)
{
	PStoreIter Iter = NewsBase->ArticleStore->GetIter();
	int DupsInCluster = 0;
	int TotalDups = 0;
	while (Iter->Next()) {
		const uint64 ArticleId = Iter->GetRecId();
		if (NewsBase->ArticleStore->GetFieldNmUInt64(ArticleId, "isDuplicateOfId") == TUInt64::Mx)
			continue;
		TotalDups++;

		if (NewsBase->ArticleStore->GetFieldNmUInt64(ArticleId, "hasStoryId") != TUInt64::Mx) {
			DupsInCluster++;
		}
		if (TotalDups % 100 == 0)
			printf("Bad dups: %d/%d\n", DupsInCluster, TotalDups);
	}

}

void PrintStoriesWithLargestTimeSpan(const PNewsBase& NewsBase)
{
	PStoreIter Iter = NewsBase->StoryStore->GetIter();
	PStore ArticleStore = NewsBase->ArticleStore;

	THash<TUInt64, TUInt64> StoryIdToDaySpanH;
	while (Iter->Next()) {
		const uint64 StoryId = Iter->GetRecId();
		PRecSet ArticleSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasArticle");
		if (ArticleSet->GetRecs() == 0)
			continue;
		TTm Oldest; TTm Newest;
		for (int N = 0; N < ArticleSet->GetRecs(); N++) {
			const uint64 ArticleId = ArticleSet->GetRecId(N);
			TTm ArtDate;
			ArticleStore->GetFieldNmTm(ArticleId, "DateTime", ArtDate);
			if (ArtDate < Oldest || N == 0)
				Oldest = ArtDate;
			if (ArtDate > Newest || N == 0)
				Newest = ArtDate;
		}
		uint64 Days = TTm::GetDiffDays(Oldest, Newest);
		StoryIdToDaySpanH.AddDat(StoryId, Days);
	}
	StoryIdToDaySpanH.SortByDat(false);
	int printed = 0;
	for (int KeyId = StoryIdToDaySpanH.FFirstKeyId(); StoryIdToDaySpanH.FNextKeyId(KeyId);) {
		uint64 StoryId = StoryIdToDaySpanH.GetKey(KeyId);
		uint Days = StoryIdToDaySpanH[KeyId];
		printf("Story %I64u (%s) - days: %I64u\n", StoryId, NewsBase->StoryStore->GetRecNm(StoryId).CStr(), Days);
		printed++;
	}
}

int main(int argc, char* argv[]){

	try {
		TQm::TEnv::Init();
		// create environment
		Env = ::TEnv(argc, argv, TNotify::StdNotify);
		// command line parameters
		Env.PrepArgs("News miner server", 0);

		// store filename
		TStr NewsStoreDefFNm = Env.GetIfArgPrefixStr("-newsDef:", "./settings/news.def", "News schema Definition File");
		TStr GeoStoreDefFNm = Env.GetIfArgPrefixStr("-geoDef:", "./settings/geo.def", "Geo info schema Definition File");

		// paths
		const TStr IndexFPath = Env.GetIfArgPrefixStr("-folderIndex:", "./IndexArticles/", "Index Folder");
		//const TStr EventFPath = Env.GetIfArgPrefixStr("-folderEvents:", "./EventEvents/", "Folder for event store");
		const TStr TempFPath = Env.GetIfArgPrefixStr("-temp:", "./temp/", "Temporary Folder");
		const TStr GeoFPath = Env.GetIfArgPrefixStr("-geoIndex:", "./IndexGeo/", "GeoInfo Index Folder");
		const TStr DbsPath = Env.GetIfArgPrefixStr("-dbsPath:", "./dbs/Articles/", "Folder with databases and models");

		const TStr GeoCountryFNm = Env.GetIfArgPrefixStr("-geoCountryFNm:", "countryInfo.json", "GeoInfo country data file");
		const TStr GeoCityFNm = Env.GetIfArgPrefixStr("-geoCityFNm:", "cityInfo.json", "GeoInfo city data file");

		// adding articles
		bool AddArticleP = Env.GetIfArgPrefixBool("-add:", true, "Expose-Add-Functions");

		// is it ok if geo data files are missing?
		const bool OkIfGeoDataMissingP = Env.GetIfArgPrefixBool("-okIfGeoDataMissing:", false, "Is it ok if geo data is missing");

		// clustering
		const TStrV StoryUrlV = Env.GetIfArgPrefixStrV("-addstory:", "Article-story service [lang1,lang2,...langN@url]");
		const bool ClusterArticlesP = Env.GetIfArgPrefixBool("-clusterArticles:", true, "Perform article clustering (true|false)");

		// categories for stories
		const TStr CategorySrvUrl = Env.GetIfArgPrefixStr("-addcategory:", "", "Story category service [e.g. http://localhost:8099/categorySrv]");

		// index sizes
		const int IndexCacheSizeMB = Env.GetIfArgPrefixInt("-indexcache:", 4000, "Cache Size (in MB)");
		const int ArticleCacheSizeMB = Env.GetIfArgPrefixInt("-articlecache:", 4000, "Cache Size (in MB)");

		// languages we want to consider
		const TStr Langs = Env.GetIfArgPrefixStr("-langs:", "eng,deu,spa,zho,slv", "Comma separated list of languages to consider (eng,deu,spa,zho,slv)");

		// settings
		const TStr SettingsFNm = Env.GetIfArgPrefixStr("-settingsArticles:", "./settings/settingsArticles.json", "File with settings related to event registry");

		if (Env.IsEndOfRun()) 
		{ 
			printf("not enough arguments specified.");
			return 0; 
		}

		// prepare unicode base
		if (!TUnicodeDef::IsDef()) { TUnicodeDef::Load("./UnicodeDef.Bin"); }
		// set the languages we wish to consider (in autosuggest, and other operations)
		TStrV LangV; Langs.SplitOnAllCh(',', LangV);
		NewsConst.SetLangs(LangV);

		// calculate cache-size in bytes
		const int64 IndexCacheSize = int64(IndexCacheSizeMB) * int64(TInt::Mega);
		const int64 ArticleCacheSize = int64(ArticleCacheSizeMB) * int64(TInt::Mega);

		PNewsFixDb NewsFixDb = TNewsFixDb::New(DbsPath, true);

		PGeoBase GeoBase = TGeoBase::Load(GeoFPath, faRdOnly, NewsConst.LangV);
		PNewsBase NewsBase = TNewsBase::Load(GeoBase, NewsFixDb, IndexFPath, AddArticleP ? faUpdate : faRdOnly, IndexCacheSize, ArticleCacheSize, SettingsFNm, DbsPath, TempFPath);
		
			
		PNewsSrvUtil NewsSrvUtil = TNewsSrvUtil::New(NewsBase);

		// news search methods
		PLogger ConceptFolderLogger = TLogger::New(3);
		ConceptFolderLogger->AddLogger(new TFileNotify("./Logs/Log-ConceptFolders.txt", true, true));
		PConceptFolderBase ConceptFolderBase = TConceptFolderBase::New(NewsBase, SettingsFNm);

		PNewsSearch NewsSearch = TNewsSearch::New(NewsBase, ConceptFolderBase);
				
		PConceptDailyTrends DailyTrends = TConceptDailyTrends::New(NewsBase, NewsSearch, NewsSrvUtil, IndexFPath + "TrendsDaily.bin");
		//DailyTrends->SetCustomParams("D0,D1", "D2,D3,D4,D5,D6,D7,D8,D9,D10,D11,D12,D13,D14,D15", 50, 100, 100);
		//DailyTrends->UpdateTrends();

		// load event store
		PLogger EventRegLogger = TLogger::New(3);
		//EventRegLogger->AddLogger(new TFileNotify("./Logs/Log-EventRegistry.txt", true, true));

		PCategoryReq CategoryReq = TCategoryReq::New(CategorySrvUrl, IndexFPath + "Req-StoriesToCategorySrv.dat");
		
		PEventRegBase EventRegBase = TEventRegBase::New(NewsBase, NewsSearch, CategoryReq, DbsPath, SettingsFNm);
		
		//RemoveLocationJoins(NewsBase, GeoBase);
		//RecomputeEventCategories(NewsBase, EventRegBase);
		//CountArticlesWithGoogleStoryId(NewsBase);
		//EvaluateClusterToGoogleAgreement(NewsBase);
		//RemoveStoryForDuplicateArticles(NewsBase, EventRegBase);
		//TestJoinsArticleStory(NewsBase);
		//RemoveExtraJoinsFromArticleToStory(NewsBase);
		//UpdateStoryAndEventLocations(NewsBase, EventRegBase);
		//RemoveCategoriesForNonEnglishStories(NewsBase, EventRegBase);
		//UpdateStoryEventArticleCount(NewsBase, EventRegBase);
		//TestArticle9259(NewsBase);
		//ValidateArticleStoryJoins(NewsBase);
		//ValidateStoryEventJoins(NewsBase);
		//TestEngClustersTimeSpan();
		//PrintClustersThatSpanTheMostTime(NewsBase);
		//WhereIsArticle667714();
		//FixInvalidMediods(NewsBase);
		//ClearStoryAndEventInfoForEmpty(NewsBase);
		//CountDupsInACluster(NewsBase);
		PrintStoriesWithLargestTimeSpan(NewsBase);

		printf("Closing...");
		return 0;
	}
	catch (PExcept E)
	{
		ErrNotify(E->GetMsgStr());
		ErrNotify(E->GetLocStr());
		return 1;
	}
	catch (...){
		ErrNotify("Unrecognized exception occured.");
		return 1;
	}
}
