/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ClusterEvaluation.h"

// //////////
// TAcc
TAcc::TAcc() 
{
	Init();
}

void TAcc::Init()
{
	PrecMom = TMom::New();
	RecMom = TMom::New();
	F1Mom = TMom::New();
}

void TAcc::Add(const double& Prec, const double& Rec, const double& Wgt)
{
	PrecMom->Add(Prec, Wgt);
	RecMom->Add(Rec, Wgt);
	double F1 = 2 * Prec * Rec / (Prec + Rec);
	if (Prec + Rec < 1e-5)
		F1 = 0;
	F1Mom->Add(F1, Wgt);
}

PJsonVal TAcc::GetJson() const
{
	PJsonVal Json = TJsonVal::NewObj();
	if (!PrecMom->IsUsable() || !RecMom->IsUsable() || !F1Mom->IsUsable())
		return Json;
	Json->AddToObj("count", PrecMom->GetVals());
	Json->AddToObj("prec", PrecMom->GetMean());
	Json->AddToObj("precQuart1", PrecMom->GetQuart1());
	Json->AddToObj("precQuart3", PrecMom->GetQuart3());
	Json->AddToObj("precDec1", PrecMom->GetDecile(1));
	Json->AddToObj("precDec9", PrecMom->GetDecile(9));
	Json->AddToObj("precQuart3", PrecMom->GetQuart3());
	Json->AddToObj("precStDev", PrecMom->GetSDev());
	Json->AddToObj("rec", RecMom->GetMean());
	Json->AddToObj("recQuart1", RecMom->GetQuart1());
	Json->AddToObj("recQuart3", RecMom->GetQuart3());
	Json->AddToObj("recDec1", RecMom->GetDecile(1));
	Json->AddToObj("recDec9", RecMom->GetDecile(9));
	Json->AddToObj("recStDev", RecMom->GetSDev());
	Json->AddToObj("f1", F1Mom->GetMean());
	Json->AddToObj("f1Quart1", F1Mom->GetQuart1());
	Json->AddToObj("f1Quart3", F1Mom->GetQuart3());
	Json->AddToObj("f1Dec1", F1Mom->GetDecile(1));
	Json->AddToObj("f1Dec9", F1Mom->GetDecile(9));
	Json->AddToObj("f1StDev", F1Mom->GetSDev());
	return Json;
}

// ////////////
// TClusterEvaluation
void TClusterEvaluation::RunEvaluation()
{
	TotalAcc.Init();
	DailyAccH.Clr();

	PStore ArticleStore = NewsBase->ArticleStore;
	PStoreIter Iter = ArticleStore->GetIter();
	THash<TStr, TUInt64V> GoogleIdToArtVH;
	TUInt64Set SeenArticleIdH;
	TNotify::StdNotify->OnStatus(TTm::GetCurLocTm().GetWebLogTimeStr() + " Iterating over individual articles...");
	int ArtsWithoutGoogleCluster = 0;
	int NonDupArtsWithoutGoogleCluster = 0;
	int NonDupArtsWithGoogleCluster = 0;
	int Duplicates = 0;
	int TotalArts = 0;
	// build a hash of google cluster it -> articles with this google cluster id
	while (Iter->Next()) {
		TotalArts++;
		uint64 ArticleId = Iter->GetRecId();
		if (ArticleId % 10000 == 0)
			TNotify::StdNotify->OnStatusFmt("Checking article %I64u\r", ArticleId);
		// if article is a duplicate then change the id to the original article id
		TRec OrigArt = ArticleStore->GetRec(ArticleId).DoSingleJoin(NewsBase->Base, "isDuplicateOf");
		if (OrigArt.IsDef()) {
			ArticleId = OrigArt.GetRecId();
			Duplicates++;
		}
		// count each article only once
		if (SeenArticleIdH.IsKey(ArticleId))
			continue;
		SeenArticleIdH.AddKey(ArticleId);
		if (ArticleStore->GetRec(ArticleId).DoSingleJoin(NewsBase->Base, "isDuplicateOf").IsDef() != false)
			NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Article that was supposed to be a non-duplicate is also a duplicate.");
		if (!ArticleStore->IsFieldNmNull(ArticleId, "GoogleStoryId")) {
			TStr GoogleId = ArticleStore->GetFieldNmStr(ArticleId, "GoogleStoryId");
			GoogleIdToArtVH.AddDat(GoogleId).Add(ArticleId);
			NonDupArtsWithGoogleCluster++;
		}
		else
			NonDupArtsWithoutGoogleCluster++;
	}

	TNotify::StdNotify->OnStatusFmt("Total article count: %d", TotalArts);
	TNotify::StdNotify->OnStatusFmt("Found %d (%.1f %%) duplicated articles", Duplicates, 100 * Duplicates / (double) MAX(1, TotalArts));
	TNotify::StdNotify->OnStatusFmt("Found %d (%.1f %%) non-duplicated articles without google cluster id", NonDupArtsWithoutGoogleCluster, 100 * NonDupArtsWithoutGoogleCluster / (double) MAX(1, (NonDupArtsWithGoogleCluster + NonDupArtsWithoutGoogleCluster)));
	
	const bool OptimizeF1 = true;
	TNotify::StdNotify->OnStatus(TTm::GetCurLocTm().GetWebLogTimeStr() + " Computing precission/recall");
	// go over each google cluster id and compute precision and recall for it
	for (int KeyId = GoogleIdToArtVH.FFirstKeyId(); GoogleIdToArtVH.FNextKeyId(KeyId);) {
		TStr GoogleClusterId = GoogleIdToArtVH.GetKey(KeyId);
		// take the articles assigned to the same google cluster
		TUInt64V ArtV = GoogleIdToArtVH[KeyId];
		THash<TUInt64, TInt> StoryIdToCntH;
		int GoogleArticleCnt = 0;
		for (int N = 0; N < ArtV.Len(); N++) {
			const uint64 ArticleId = ArtV[N];
			if (ArticleStore->GetRec(ArticleId).DoSingleJoin(NewsBase->Base, "isDuplicateOf").IsDef() != false)
				NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Expected to get articles that are not duplicates");
			TRec StoryRec = ArticleStore->GetRec(ArticleId).DoSingleJoin(NewsBase->Base, "hasStory");
			// it's possible that recent articles don't yet have a story assigned to it
			if (!StoryRec.IsDef()) {
				NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Weird. Article %I64u is not a duplicate nor assigned to a story.", ArticleId);
				continue;
			}
			GoogleArticleCnt++;
			StoryIdToCntH.AddDat(StoryRec.GetRecId()) += 1;

			// DEBUG:
			bool found = false;
			PRecSet ArticleSet = NewsBase->StoryStore->GetRec(StoryRec.GetRecId()).DoJoin(NewsBase->Base, "hasArticle");
			for (int A = 0; A < ArticleSet->GetRecs(); A++) {
				uint64 StoryArtId = ArticleSet->GetRecId(A);
				if (StoryArtId == ArticleId)
					found = true;
			}
			if (!found) 
				NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Weird. Nonduplicate article %I64u is assigned to story %I64u, but the story doesn't have the article assigned to it", ArticleId, StoryRec.GetRecId());
		}
		if (StoryIdToCntH.Len() == 0)
			continue;
		
		uint64 TopStoryId = TUInt64::Mx;
		
		if (OptimizeF1) {
			// go over all stories in the StoryIdToCntH and compute F1 measure of the story to the GoogleClusterId
			double TopF1 = -1;
			for (int TestKeyId = StoryIdToCntH.FFirstKeyId(); StoryIdToCntH.FNextKeyId(TestKeyId);) {
				// find all the articles for the TopStory that have a google story id. only those articles can be used when computing the recall
				// that is because we have in our system also other articles not seen by google and for those we don't know if google would assign them to this story or not
				uint64 TestStoryId = StoryIdToCntH.GetKey(TestKeyId);
				PRecSet ArtsForStory = NewsBase->StoryStore->GetRec(TestStoryId).DoJoin(NewsBase->Base, "hasArticle");
				int StoryArticleCnt = 0;
				for (int N = 0; N < ArtsForStory->GetRecs(); N++) {
					uint64 ArticleId = ArtsForStory->GetRecId(N);
					if (ArticleStore->IsFieldNmNull(ArticleId, "GoogleStoryId"))
						continue;
					if (ArtsForStory->GetRec(N).DoSingleJoin(NewsBase->Base, "isDuplicateOf").IsDef())
						continue;
					StoryArticleCnt++;
				}

				int TP = StoryIdToCntH[TestKeyId];
				double Prec = TP / (double) GoogleArticleCnt;
				double Rec = TP / (double) StoryArticleCnt;
				double F1 = 2 * Prec * Rec / (Prec + Rec);
				if (Prec > 1.001)
					NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Weird.For cluster %s precision is invalid(%.3f)", GoogleClusterId.CStr(), Prec);
				if (Rec > 1.001)
					NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Weird. For cluster %s recall is invalid (%.3f)", GoogleClusterId.CStr(), Rec);
				
				// DEBUG:
				if (Prec > 1.001 || Rec > 1.001) {
					PRecSet ArtsForStory = NewsBase->StoryStore->GetRec(TestStoryId).DoJoin(NewsBase->Base, "hasArticle");
					int StoryArticleCnt = 0;
					for (int N = 0; N < ArtsForStory->GetRecs(); N++) {
						uint64 ArticleId = ArtsForStory->GetRecId(N);
						TRec StoryRec = ArticleStore->GetRec(ArticleId).DoSingleJoin(NewsBase->Base, "hasStory");
						if (ArticleStore->IsFieldNmNull(ArticleId, "GoogleStoryId"))
							continue;
						if (ArtsForStory->GetRec(N).DoSingleJoin(NewsBase->Base, "isDuplicateOf").IsDef())
							continue;
						StoryArticleCnt++;
					}
				}

				if (F1 > TopF1) { TopF1 = F1; TopStoryId = TestStoryId; }
			}
		}
		// if we would take the story with the highest count then we would be selecting the cluster that would give us the greatest precission
		else {
			StoryIdToCntH.SortByDat(false);
			int TestKeyId = StoryIdToCntH.FFirstKeyId(); StoryIdToCntH.FNextKeyId(TestKeyId);
			// TopStoryId is the story that has the highest number of articles for this google cluster
			TopStoryId = StoryIdToCntH.GetKey(TestKeyId);
		}

		if (TopStoryId == TUInt64::Mx) {
			NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Weird. Invalid value for story id chosen for cluster %s", GoogleClusterId.CStr());
			continue;
		}

		// find all the articles for the TopStory that have a google story id. only those articles can be used when computing the recall
		// that is because we have in our system also other articles not seen by google and for those we don't know if google would assign them to this story or not
		PRecSet ArtsForStory = NewsBase->StoryStore->GetRec(TopStoryId).DoJoin(NewsBase->Base, "hasArticle");
		int StoryArticleCnt = 0;
		for (int N = 0; N < ArtsForStory->GetRecs(); N++) {
			if (ArticleStore->IsFieldNmNull(ArtsForStory->GetRecId(N), "GoogleStoryId"))
				continue;
			if (ArtsForStory->GetRec(N).DoSingleJoin(NewsBase->Base, "isDuplicateOf").IsDef())
				continue;
			StoryArticleCnt++;
		}

		int TP = StoryIdToCntH.GetDat(TopStoryId);
		double Prec = TP / (double) GoogleArticleCnt;
		double Rec = TP / (double) StoryArticleCnt;
		if (Prec > 1.001)
			NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Weird. For cluster %s precision is invalid (%.3f)", GoogleClusterId.CStr(), Prec);
		if (Rec > 1.001)
			NewsBase->ErrorLogFile->NotifyErr("TClusterEvaluation::RunEvaluation. Weird. For cluster %s recall is invalid (%.3f)", GoogleClusterId.CStr(), Rec);
		double F1 = 2 * Prec * Rec / (Prec + Rec);
		if (Prec + Rec < 1e-5)
			F1 = 0;

		// prevent division by zero
		if (StoryArticleCnt == 0 || GoogleArticleCnt == 0)
			continue;

		// update total score
		TotalAcc.Add(Prec, Rec, GoogleArticleCnt);		// weight the precisions by the size of the google's cluster (GoogleIdToArtVH.GetKey(KeyId))
		uint64 FirstArtId = ArtV[0];
		if (NewsBase->ArticleStore->IsRecId(FirstArtId)) {
			TStr Date = NewsBase->ArticleStore->GetFieldNmStr(FirstArtId, "Date");
			// update daily score
			DailyAccH.AddDat(Date).Add(Prec, Rec, GoogleArticleCnt);
		}
	}

	// def
	TotalAcc.Def();
	for (int KeyId = DailyAccH.FFirstKeyId(); DailyAccH.FNextKeyId(KeyId);)
		DailyAccH[KeyId].Def();
}