/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "LrnStoryLocation.h"
#include "../ServerArticles/EventRegBase.h"

TLrnStoryLocationExample::TLrnStoryLocationExample() :
	Username(""), DateTime(""), 
	StoryUri(""), ConceptUri(""), 
	StoryArticles(-1),
	NormalizedConceptWgt(-1), RatioOfArticlesWithConcept(-1), ConceptWgtToSum(-1), 
	NormalizedExplicitConceptWgt(-1), RatioOfArticlesWithExplicitConcept(-1), ExplicitConceptWgtToSum(-1), 
	IsStoryLocation(false)
{

}

TLrnStoryLocationExample::TLrnStoryLocationExample(PJsonVal ExampleJson)
{
	Username = ExampleJson->GetObjStr("_username", "");
	DateTime = ExampleJson->GetObjStr("_datetime", "");

	StoryUri = ExampleJson->GetObjStr("_storyUri", "");
	ConceptUri = ExampleJson->GetObjStr("_conceptUri", "");

	StoryArticles = ExampleJson->GetObjNum("storyArticles", 0);

	NormalizedConceptWgt = ExampleJson->GetObjNum("normalizedConceptWgt", -1);
	RatioOfArticlesWithConcept = ExampleJson->GetObjNum("ratioOfArticlesWithConcept", -1);
	ConceptWgtToSum = ExampleJson->GetObjNum("conceptWgtToSum", -1);

	NormalizedExplicitConceptWgt = ExampleJson->GetObjNum("normalizedExplicitConceptWgt", -1);
	RatioOfArticlesWithExplicitConcept = ExampleJson->GetObjNum("ratioOfArticlesWithExplicitConcept", -1);
	ExplicitConceptWgtToSum = ExampleJson->GetObjNum("explicitConceptWgtToSum", -1);

	IsStoryLocation = ExampleJson->GetObjBool("isStoryLocation", false);
}

PJsonVal TLrnStoryLocationExample::GetJson() const
{
	PJsonVal ExampleJson = TJsonVal::NewObj();
	ExampleJson->AddToObj("_username", Username);
	ExampleJson->AddToObj("_datetime", DateTime);

	ExampleJson->AddToObj("_storyUri", StoryUri);
	ExampleJson->AddToObj("_conceptUri", ConceptUri);

	ExampleJson->AddToObj("storyArticles", StoryArticles);

	ExampleJson->AddToObj("normalizedConceptWgt", NormalizedConceptWgt);
	ExampleJson->AddToObj("ratioOfArticlesWithConcept", RatioOfArticlesWithConcept);
	ExampleJson->AddToObj("conceptWgtToSum", ConceptWgtToSum);

	ExampleJson->AddToObj("normalizedExplicitConceptWgt", NormalizedExplicitConceptWgt);
	ExampleJson->AddToObj("ratioOfArticlesWithExplicitConcept", RatioOfArticlesWithExplicitConcept);
	ExampleJson->AddToObj("explicitConceptWgtToSum", ExplicitConceptWgtToSum);

	ExampleJson->AddToObj("isStoryLocation", IsStoryLocation);

	return ExampleJson;
}

//////////////////////////////////////
/// TLrnStoryLocationExampleTable

// load example table from disk
bool TLrnStoryLocationExampleTable::LoadTable(const TStr& FileNm)
{
	if (TFile::Exists(FileNm)) {
		TFIn FIn(FileNm);
		TStr LnStr;
		while (FIn.GetNextLn(LnStr)) {
			try {
				PJsonVal ExampleJson = TJsonVal::GetValFromStr(LnStr);
				ExampleV.Add(TLrnStoryLocationExample(ExampleJson));
			}
			catch (PExcept E) {
				TNotify::StdNotify->OnStatus("TLrnStoryLocationExampleTable::LoadTable. Error while loading a learning example. Error: " + E->GetMsgStr());
				NewsBase->ErrorLogFile->NotifyErr("TLrnStoryLocationExampleTable::LoadTable. Error while loading a learning example", E);
			}
			catch (...) {}
		}
		return true;
	}
	else
		return false;
}

void TLrnStoryLocationExampleTable::SaveTable(const TStr& FileNm) const
{
	TFOut FOut(FileNm);
	for (int N = 0; N < ExampleV.Len(); N++) {
		PJsonVal ExampleJson = ExampleV[N].GetJson();
		const TStr LnStr = TJsonVal::GetStrFromVal(ExampleJson);
		FOut.PutStrLn(LnStr);
	}
}

void TLrnStoryLocationExampleTable::AddExample(const TLrnStoryLocationExample& Example)
{
	ExampleV.Add(Example);
	Changed = true;
}


//////////////////////////////////////
/// TLrnStoryLocation


TLrnStoryLocation::TLrnStoryLocation(const PNewsBase& _NewsBase, const TStr& _Path) : NewsBase(_NewsBase), Path(_Path), ExampleTable(NULL)
{
	ModelStoryLocation = NULL;
	// load data on startup
	bool ok = LoadLearningData();
	TStr FName = NewsBase->CustomDataPath + "LrnStoryLocation.bin";
	if (TFile::Exists(FName)) {
		try {
			TFIn FIn(FName);
			ModelStoryLocation = new TSvm::TLinModel(FIn);
		}
		catch (...) {
			ModelStoryLocation = NULL;
			TNotify::StdNotify->OnStatus("TLrnStoryLocation. Error while loading story location model from LrnStoryLocation.bin.");
		}
	}
}

TLrnStoryLocation::~TLrnStoryLocation()
{
	// save data on exit
	Save();
	if (!ExampleTable.Empty() && ExampleTable->HasTableChanged())
		SaveLearningData();
	if (!ExampleTable.Empty())
		TLrnStoryLocationModel::SaveFtrVecData(Path, ExampleTable, NewsBase);
	if (ModelStoryLocation != NULL)
		delete ModelStoryLocation;
	ModelStoryLocation = NULL;
}

void TLrnStoryLocation::Save()
{
	try {
		TFOut FOut(NewsBase->CustomDataPath + "LrnStoryLocation.bin");
		if (ModelStoryLocation != NULL) {
			ModelStoryLocation->Save(FOut);
		}
	}
	catch (...) { 
		TNotify::StdNotify->OnStatus("TLrnStoryLocation::Save. Error while saving LrnStoryLocation.bin.");
	}
}

// load example table from file
bool TLrnStoryLocation::LoadLearningData()
{
	TStr FileNm = Path + "storyLocation.json";
	try {
		ExampleTable = TLrnStoryLocationExampleTable::New(NewsBase);
		return ExampleTable->LoadTable(FileNm);
	}
	catch (PExcept E) {
		TNotify::StdNotify->OnStatus("TLrnStoryLocation::LoadLearningData. Error while trying to load a learning data table. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TLrnStoryLocation::LoadLearningData. Error while trying to load a learning data table", E);
		return false;
	}
	catch (...) { 
		TNotify::StdNotify->OnStatus("TLrnStoryLocation::LoadLearningData. Error while loading example table from storyLocation.json.");
		return false;
	}
}

// save the example table to a file
void TLrnStoryLocation::SaveLearningData()
{
	try {
		TStr FileNm = Path + "storyLocation.json";
		if (!ExampleTable.Empty()) {
			ExampleTable->SaveTable(FileNm);
			TLrnStoryLocationModel::SaveFtrVecData(Path, ExampleTable, NewsBase);
		}
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("Exception while saving learning data.");
		NewsBase->ErrorLogFile->NotifyErr("TLrnStoryLocation::SaveLearningData. Exception while saving learning data.");
	}
}

// create a learning example from the info about a location in a particular story
void TLrnStoryLocation::MakeLearningExample(const PEventRegBase& EventRegBase, const TStr& StoryUri, const TStr& LocationConceptUri, const TStoryLocationsInfo& StoryLocationInfo, TLrnStoryLocationExample& Example) const
{
	const uint64 StoryId = EventRegBase->StoryStore->GetRecId(StoryUri);
	const uint64 ConceptId = EventRegBase->ConceptStore->GetRecId(LocationConceptUri);

	const int Articles = StoryLocationInfo.Articles;
	
	const double ConceptWgt = StoryLocationInfo.ConceptIdToWgtH.GetDatOrDef(ConceptId, 0);
	const double ExplicitConceptWgt = StoryLocationInfo.ExplicitConceptIdToWgtH.GetDatOrDef(ConceptId, 0);		// if there is no mention of the concept then we treat the value as 0

	Example.ConceptUri = EventRegBase->ConceptStore->GetRecNm(ConceptId);
	Example.StoryUri = EventRegBase->StoryStore->GetRecNm(StoryId);

	Example.StoryArticles = StoryLocationInfo.Articles;
		
	Example.NormalizedConceptWgt = StoryLocationInfo.SumCosSim > TPredictionModel::Eps ? ConceptWgt / StoryLocationInfo.SumCosSim : -1;		// if there are no concepts then we treat the values as missing
	Example.RatioOfArticlesWithConcept = Articles > 0 ? StoryLocationInfo.ConceptIdToCntH.GetDatOrDef(ConceptId, 0) / float(Articles) : -1;
	Example.ConceptWgtToSum = StoryLocationInfo.SumConceptWgts > TPredictionModel::Eps ? ConceptWgt / StoryLocationInfo.SumConceptWgts : -1;

	Example.NormalizedExplicitConceptWgt = StoryLocationInfo.SumCosSim > TPredictionModel::Eps ? ExplicitConceptWgt / StoryLocationInfo.SumCosSim : -1;
	Example.RatioOfArticlesWithExplicitConcept = Articles > 0 ? StoryLocationInfo.ExplicitConceptIdToCntH.GetDatOrDef(ConceptId, 0) / float(Articles) : -1;
	Example.ExplicitConceptWgtToSum = StoryLocationInfo.SumCosSimForExplicitArticles > TPredictionModel::Eps ? ExplicitConceptWgt / StoryLocationInfo.SumCosSimForExplicitArticles : -1;
}


// check the articles assigned to the story, create learning example for the locations in the story and find the location
// where the story occured. if not location is the right one we return false
bool TLrnStoryLocation::PredictStoryLocation(const PEventRegBase& EventRegBase, const TStr& StoryUri, uint64& LocationConceptId, double& LocationProb) const
{
	if (ModelStoryLocation == NULL)
		return false;
	
	const uint64 StoryId = EventRegBase->StoryStore->GetRecId(StoryUri);
	
	TStoryLocationsInfo StoryLocationInfo;
	EventRegBase->GetStoryLocationsInfo(StoryId, StoryLocationInfo);
		
	TFltV FtrV; TStrV LabelV;
	double MaxProb = -1; TUInt64 MaxLocation = TUInt64::Mx; TFltV MaxFtrV;
	for (int KeyId = StoryLocationInfo.CandidateLocationH.FFirstKeyId(); StoryLocationInfo.CandidateLocationH.FNextKeyId(KeyId);) {
		const uint64 LocationId = StoryLocationInfo.CandidateLocationH.GetKey(KeyId);
		const TStr LocationUri = EventRegBase->ConceptStore->GetRecNm(LocationId);
		TLrnStoryLocationExample Example;
		MakeLearningExample(EventRegBase, StoryUri, LocationUri, StoryLocationInfo, Example);
		
		TLrnStoryLocationModel::MakeFtrV(Example, FtrV, LabelV);
		const double Prob = ModelStoryLocation->Predict(FtrV);
		if (Prob > MaxProb) {
			MaxProb = Prob; MaxLocation = LocationId; MaxFtrV = FtrV;
		}
	}
	if (MaxProb > 0.5) {
		LocationConceptId = MaxLocation;
		LocationProb = MaxProb;
		return true;
	}
	return false;
}

PJsonVal TLrnStoryLocation::BuildModel(const bool& GenerateReport)
{
	try {
		if (ModelStoryLocation != NULL)
			delete ModelStoryLocation;
		ModelStoryLocation = NULL;
		// prepare training set
		if (ExampleTable.Empty() || ExampleTable->Len() == 0) {
			TNotify::StdNotify->OnStatus("TCrossLingStoryLink::BuildModel. Unable to build a story-linking model. Example table is empty.");
			NewsBase->ErrorLogFile->NotifyErr("TCrossLingStoryLink::BuildModel. Unable to build a story-linking model. Example table is empty.");
			return TJsonVal::NewObj("error", TStr("Unable to build a model. Example table is empty."));
		}

		TVec<TFltV> ExamplesV; TFltV LabelsV;
		TLrnStoryLocationModel::GetSVMExampleTable(ExampleTable, ExamplesV, LabelsV);
		TSvm::TLinModel Model = TSvm::SolveClassify<TVec<TFltV>>(ExamplesV, ExamplesV[0].Len(), ExamplesV.Len(), LabelsV, 1.0, 1.0, 10 * 1000, 10000, 1e-6, 10000);
		ModelStoryLocation = new TSvm::TLinModel(Model);

		if (GenerateReport)
			return TCrossLingModel::EvaluateSVMAccuracy(ExamplesV, LabelsV);
		else
			return TJsonVal::NewObj();
	}
	catch (PExcept E) {
		TNotify::StdNotify->OnStatus("TCrossLingStoryLink::BuildModel. Error while building a story-linking model. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TCrossLingStoryLink::BuildModel. Error while building a story-linking model", E);
		return TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("Unknown exception happened while building a crosslingual linking model.");
		NewsBase->ErrorLogFile->NotifyErr("TCrossLingStoryLink::BuildModel. Unknown exception happened while building a crosslingual linking model.");
		return TJsonVal::NewObj("error", TStr("Unknown exception happened while building a crosslingual linking model."));
	}
}

// do we already have information about a story par StoryId1, StoryId2 ?
bool TLrnStoryLocation::IsKnownExample(const TStr& StoryUri, const TStr& ConceptUri) const
{
	if (ExampleTable.Empty())
		return false;
	int Examples = ExampleTable->Len();
	for (int N = 0; N < Examples; N++) {
		const TLrnStoryLocationExample& Example = ExampleTable->GetExample(N);
		if (Example.StoryUri == StoryUri && Example.ConceptUri == ConceptUri)
			return true;
	}
	return false;
}

void TLrnStoryLocation::RemoveExamples(const TStr& StoryUri)
{
	if (ExampleTable.Empty())
		return;
	int Examples = ExampleTable->Len();
	for (int N = Examples - 1; N >= 0; N--) {
		const TLrnStoryLocationExample& Example = ExampleTable->GetExample(N);
		if (Example.StoryUri == StoryUri)
			ExampleTable->RemoveExample(N);
	}
}

// add learning examples for the story StoryUri. The correct location concept is LocationConceptUri, all other locations that can be found in the articles are incorrect
void TLrnStoryLocation::AddLearningExamples(const PEventRegBase& EventRegBase, const TStr& StoryUri, const TStr& LocationConceptUri, const TStr& Username)
{
	if (!EventRegBase->StoryStore->IsRecNm(StoryUri))
		return;

	// remove existing examples related to the specified story
	RemoveExamples(StoryUri);

	const uint64 StoryId = EventRegBase->StoryStore->GetRecId(StoryUri);
	TStoryLocationsInfo StoryLocationInfo;
	EventRegBase->GetStoryLocationsInfo(StoryId, StoryLocationInfo);

	for (int KeyId = StoryLocationInfo.CandidateLocationH.FFirstKeyId(); StoryLocationInfo.CandidateLocationH.FNextKeyId(KeyId);) {
		TLrnStoryLocationExample Example;
		// populate example features
		const TStr ConceptUri = EventRegBase->ConceptStore->GetRecNm(StoryLocationInfo.CandidateLocationH.GetKey(KeyId));
		MakeLearningExample(EventRegBase, StoryUri, ConceptUri, StoryLocationInfo, Example);
		Example.Username = Username;
		TTm Tm = TTm::GetCurLocTm();
		Example.DateTime = Tm.GetStr(false);
		// set the correct class value
		Example.IsStoryLocation = (ConceptUri == LocationConceptUri);
		// add learning example
		ExampleTable->AddExample(Example);
	}
}

void TLrnStoryLocation::RebuildLearningTable(const PEventRegBase& EventRegBase)
{
	PStoryLocationExampleTable OldExampleTable = ExampleTable;
	ExampleTable = TLrnStoryLocationExampleTable::New(NewsBase);
	if (!ExampleTable.Empty()) {
		for (int ExN = 0; ExN < OldExampleTable->Len(); ExN++) {
			const TLrnStoryLocationExample& Example = OldExampleTable->GetExample(ExN);
			if (Example.IsStoryLocation)
				AddLearningExamples(EventRegBase, Example.StoryUri, Example.ConceptUri, Example.Username);
		}
	}
}

//////////////////////////////////////
/// TLrnStoryLocationModel
// creating SVM example table, normalization methods, testing svm accuracy, ...

// creates an example table where each row is an example and columns are features
// normalizes each feature in ExampleV to have mean = 0 and stdev = 1
// assumptions: the last item in ExampleV represents a dummy feature - it's value will not be normalized
void TLrnStoryLocationModel::GetSVMExampleTable(const PStoryLocationExampleTable& ExampleTable, TVec<TFltV>& ExampleV, TFltV& LabelV)
{
	int Examples = ExampleTable->Len();
	ExampleV.Clr(); LabelV.Clr();
	if (Examples == 0)
		return;

	for (int N = 0; N < Examples; N++) {
		const TLrnStoryLocationExample& Example = ExampleTable->GetExample(N);
		TFltV FtrV; TStrV FtrNmV;
		MakeFtrV(Example, FtrV, FtrNmV);
		ExampleV.Add(FtrV);
		LabelV.Add(Example.IsStoryLocation ? 1 : -1, true);
	}
	int Features = ExampleV[0].Len();
	// normalize each feature to mean = 0, dev = 1
	// NOTE: don't normalize the last (dummy) feature since it serves as the free coeficient
	for (int FtrN = 0; FtrN < Features - 1; FtrN++)
		TPredictionModel::NormalizeColumn(ExampleV, FtrN);
	// change -1s to 0s to make them like redundant
	for (int N = 0; N < Examples; N++) {
		for (int FtrN = 0; FtrN < Features - 1; FtrN++) {
			if (TPredictionModel::IsMissing(ExampleV[N][FtrN]))
				ExampleV[N][FtrN] = 0;
		}
	}
}

// make learning features from the given example
void TLrnStoryLocationModel::MakeFtrV(const TLrnStoryLocationExample& Example, TFltV& FtrV, TStrV& FtrNmV)
{
	FtrV.Clr(); FtrNmV.Clr();

	//NormalizedConceptWgt\t\t\tCoreMaxTimeStDev\tCoreAvgTimeStDev\tCoreTimeDiff\tExtractedDateJacSim\tMatchedDateMentions\tCategorySim\tHasSameLoc\tSameLang\tShouldMerge");
	FtrV.Add(Example.NormalizedConceptWgt);	 FtrNmV.Add("NormalizedConceptWgt");			// sum(ConceptWgt) / float(Articles)
	FtrV.Add(Example.RatioOfArticlesWithConcept); FtrNmV.Add("RatioOfArticlesWithConcept");	// ConceptIdToCntH.GetDat(ConceptId) / float(Articles);
	FtrV.Add(Example.ConceptWgtToSum); FtrNmV.Add("ConceptWgtToSum");						// ConceptWgt / SumConceptWgts
	
	FtrV.Add(Example.NormalizedExplicitConceptWgt);	 FtrNmV.Add("NormalizedExplicitConceptWgt");
	FtrV.Add(Example.RatioOfArticlesWithExplicitConcept); FtrNmV.Add("RatioOfArticlesWithExplicitConcept");
	FtrV.Add(Example.ExplicitConceptWgtToSum); FtrNmV.Add("ExplicitConceptWgtToSum");

	FtrV.Add(1); FtrNmV.Add("DummyVar");
}

// save the examples using the computed features to be used when learning
void TLrnStoryLocationModel::SaveFtrVecData(const TStr& Path, const PStoryLocationExampleTable& ExampleTable, const PNewsBase& NewsBase)
{
	try {
		TStr FileNm = Path + "storyLocationFtrVec.tab";
		TFOut FOut(FileNm);
		if (!ExampleTable.Empty()) {
			for (int ExN = 0; ExN < ExampleTable->Len(); ExN++) {
				const TLrnStoryLocationExample& Example = ExampleTable->GetExample(ExN);
				TFltV FtrV; TStrV FtrNmV;
				MakeFtrV(Example, FtrV, FtrNmV);
				if (ExN == 0) {
					// add feature names
					for (int N = 0; N < FtrNmV.Len(); N++)
						FOut.PutStr(FtrNmV[N] + "\t");
					// add class name
					FOut.PutStrLn("IsStoryLocation");
					// line specifying continuous variables and discrete class
					for (int N = 0; N < FtrV.Len(); N++)
						FOut.PutStr("c\t");	// attrs
					FOut.PutStrLn("d");		// class
					// the line saying which attr is the class
					for (int N = 0; N < FtrNmV.Len(); N++)
						FOut.PutStr("\t");
					FOut.PutStrLn("c");
				}
				for (int N = 0; N < FtrV.Len(); N++)
					FOut.PutStrFmt("%.3f\t", FtrV[N]);
				FOut.PutStrLn(Example.IsStoryLocation ? "yes" : "no");
			}
		}
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("Exception while saving learning data with features.");
		NewsBase->ErrorLogFile->NotifyErr("TLrnStoryLocationModel::SaveFtrVecData. Exception while saving learning data with features.");
	}
}

PJsonVal TLrnStoryLocationModel::EvaluateSVMAccuracy(const TVec<TFltV>& ExampleV, const TFltV& LabelV)
{
	TFltV AccV;
	int Folds = 5;
	for (int FoldN = 0; FoldN < Folds; FoldN++) {
		TVec<TFltV> TrainV; TFltV TrainLabelV;
		TVec<TFltV> TestV; TFltV TestLabelV;
		for (int N = 0; N < ExampleV.Len(); N++) {
			if ((N + FoldN) % Folds == 0) { TestV.Add(ExampleV[N]); TestLabelV.Add(LabelV[N]); }
			else { TrainV.Add(ExampleV[N]); TrainLabelV.Add(LabelV[N]); }
		}
		TSvm::TLinModel Model = TSvm::SolveClassify<TVec<TFltV>>(TrainV, TrainV[0].Len(), TrainV.Len(), TrainLabelV, 1.0, 1.0, 10 * 1000, 10000, 1e-6, 10000);
		int Correct = 0;
		for (int N = 0; N < TestV.Len(); N++) {
			double Val = Model.Predict(TestV[N]);
			double Expected = TestLabelV[N];
			if (Val * Expected > 0) Correct++;		// if both Val and Expected have the same sign then it means that the prediction is correct
		}
		TFlt Acc = 100 * Correct / float(TestV.Len());
		AccV.Add(Acc);
		TNotify::StdNotify->OnStatusFmt("Accuracy on fold %d: %.3f%%", FoldN + 1, Acc.Val);
	}

	TFlt AvgAcc = TLinAlg::SumVec(AccV) / float(AccV.Len());
	TNotify::StdNotify->OnStatusFmt("Average accuracy: %.3f%%", AvgAcc.Val);
	return TJsonVal::NewObj("acc", AvgAcc);
}

