// the purpose of the classes here is to determine if a particular story has a location where it happened and if it does, then which is the correct one (in cases when multiple locations are mentioned)
// using the model we find what is the most likely location. if the predicted probability is above 0.5 then we use it as story location

#ifndef STORYLOCATION_H
#define STORYLOCATION_H

#include "../NewsMiner/NewsBase.h"
#include "PredictionModel.h"

// forward declaration
class TEventRegBase;
typedef TPt<TEventRegBase> PEventRegBase;

class TStoryLocationsInfo;

class TLrnStoryLocation;
typedef TPt<TLrnStoryLocation> PStoryLocation;

class TLrnStoryLocationExample {
public:
	TStr Username;
	TStr DateTime;

	TStr StoryUri;
	TStr ConceptUri;

	TInt StoryArticles;

	TFlt NormalizedConceptWgt;
	TFlt RatioOfArticlesWithConcept;
	TFlt ConceptWgtToSum;

	TFlt NormalizedExplicitConceptWgt;
	TFlt RatioOfArticlesWithExplicitConcept;
	TFlt ExplicitConceptWgtToSum;

	TBool IsStoryLocation;

	TLrnStoryLocationExample();
	TLrnStoryLocationExample(PJsonVal ExampleJson);
	PJsonVal GetJson() const;
};

class TLrnStoryLocationExampleTable;
typedef TPt<TLrnStoryLocationExampleTable> PStoryLocationExampleTable;
class TLrnStoryLocationExampleTable {
public:
	friend class TPt<TLrnStoryLocationExampleTable>;
private:
	TCRef CRef;
	PNewsBase NewsBase;
	TVec<TLrnStoryLocationExample> ExampleV;

	bool Changed;
	int FtrCount;
	TLrnStoryLocationExampleTable(const PNewsBase& _NewsBase) : NewsBase(_NewsBase) {
		FtrCount = 16;
		Changed = false;
	}
public:
	static PStoryLocationExampleTable New(const PNewsBase& NewsBase) {
		return new TLrnStoryLocationExampleTable(NewsBase);
	}

	bool HasTableChanged() { return Changed; }
	bool LoadTable(const TStr& FileNm);
	void SaveTable(const TStr& FileNm) const;

	void AddExample(const TLrnStoryLocationExample& Example);
	void RemoveExample(const int& ValN) { EAssertR(ValN < Len() && ValN >= 0, "Example table: Index outside of valid range.");  ExampleV.Del(ValN); }
	int Len() { return ExampleV.Len(); }
	const TLrnStoryLocationExample& GetExample(const int& ValN) { EAssertR(ValN < Len() && ValN >= 0, "Example table: Index outside of valid range.");  return ExampleV[ValN]; }

};

//////////////////////////////////////
class TLrnStoryLocation {
private:
	TCRef CRef;
public:
	friend class TPt<TLrnStoryLocation>;

private:
	PNewsBase NewsBase;
	PStoryLocationExampleTable ExampleTable;
	TStr Path;

	TSvm::TLinModel* ModelStoryLocation; 	// model that decides which of the locations found in the story is the most likely candidate for the story location
	
	TLrnStoryLocation(const PNewsBase& _NewsBase, const TStr& _Path);
	~TLrnStoryLocation();
	void Save();

	void MakeLearningExample(const PEventRegBase& EventRegBase, const TStr& StoryUri, const TStr& LocationConceptUri, const TStoryLocationsInfo& StoryLocationInfo, TLrnStoryLocationExample& Example) const;

public:
	static PStoryLocation New(const PNewsBase& NewsBase, const TStr& Path) { return new TLrnStoryLocation(NewsBase, Path); }

	//
	bool LoadLearningData();
	void SaveLearningData();

	PJsonVal BuildModel(const bool& GenerateReport = false);
	void AddLearningExamples(const PEventRegBase& EventRegBase, const TStr& StoryUri, const TStr& LocationConceptUri, const TStr& Username);
	void RebuildLearningTable(const PEventRegBase& EventRegBase);

	//double GetLocationProbability(const PEventRegBase& EventRegBase, const TStr& StoryUri, const TStr& LocationConceptUri);
	bool PredictStoryLocation(const PEventRegBase& EventRegBase, const TStr& StoryUri, uint64& LocationConceptId, double& LocationProb) const;

	bool IsKnownExample(const TStr& StoryUri, const TStr& ConceptUri) const;
	void RemoveExamples(const TStr& StoryUri);
	TStr GetExampleTablePath() const { return TStr(Path); }

	int GetExamplesCount() const { return ExampleTable.Empty() ? 0 : ExampleTable->Len(); }
};

class TLrnStoryLocationModel : TPredictionModel
{
public:
	static bool IsMissing(const double& Val);

	static void GetSVMExampleTable(const PStoryLocationExampleTable& ExampleTable, TVec<TFltV>& ExampleV, TFltV& LabelV);	
	static void MakeFtrV(const TLrnStoryLocationExample& Example, TFltV& FtrV, TStrV& FtrNmV);	
	static void SaveFtrVecData(const TStr& Path, const PStoryLocationExampleTable& ExampleTable, const PNewsBase& NewsBase);
	static PJsonVal EvaluateSVMAccuracy(const TVec<TFltV>& ExampleV, const TFltV& LabelV);
};

#endif