/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "CategoryBase.h"


///////////////////////////////////////////////
// base class for managing the category store - creating missing categories, finding subcategories, ...
TCategoryBase::TCategoryBase(const PBase& _Base) : Base(_Base)
{
	CategoryStore = Base->GetStoreByStoreNm("Category");
	DmozPrefix = "http://www.dmoz.org";
}

void TCategoryBase::CreateDMozCats(const TStr& Category)
{
	int N = 0; 
	// if category is a dmoz category then move index to the start of the category name
	if (Category.IsPrefix(DmozPrefix))
		N = DmozPrefix.Len();
	uint64 SuperCatRecId = TUInt64::Mx;
	TStr Cat = "";
	const int CatLen = Category.Len();
	for (; N <= CatLen; N++) {
		if (N == CatLen || Category[N] == '/') {
			Cat = Category.GetSubStr(0, N - 1);
			if (!CategoryStore->IsRecNm(Cat)) {
				PJsonVal CatJson = TJsonVal::NewObj("URI", Cat);
				TStr Label = Cat.TrimLeft(GetDMozUriPrefix());
				Label = Label.TrimLeft("/");
				Label.ChangeChAll('_', ' ');
				CatJson->AddToObj("Label", Label);
				uint64 CatRecId = CategoryStore->AddRec(CatJson);
				if (SuperCatRecId != TUInt64::Mx)
					CategoryStore->AddJoin("hasSupercategory", CatRecId, SuperCatRecId);
				SuperCatRecId = CatRecId;
			}
			else
				SuperCatRecId = CategoryStore->GetRecId(Cat);
		}
	}
}

TUInt64Set TCategoryBase::GetAllSubCategories(const TUInt64V& CategoryRecV)
{
	TUInt64Set FullRetSet(CategoryRecV);
	PRecSet CatSet = TRecSet::New(CategoryStore, CategoryRecV);
	while (CatSet->GetRecs() > 0) {
		CatSet = CatSet->DoJoin(Base, "hasSubcategory");
		TUInt64V CatV; CatSet->GetRecIdV(CatV);
		FullRetSet.AddKeyV(CatV);
	}
	return FullRetSet;
}

TUInt64Set TCategoryBase::GetAllSubCategories(const uint64& CategoryRecId)
{
	TUInt64Set FullRetSet;
	FullRetSet.AddKey(CategoryRecId);
	PRecSet CatSet = TRecSet::New(CategoryStore, CategoryRecId);
	while (CatSet->GetRecs() > 0) {
		CatSet = CatSet->DoJoin(Base, "hasSubcategory");
		TUInt64V CatV; CatSet->GetRecIdV(CatV);
		FullRetSet.AddKeyV(CatV);
	}
	return FullRetSet;
}

PRecSet TCategoryBase::GetChildCategories(const TStr& CategoryNm)
{
	if (!CategoryStore->IsRecNm(CategoryNm))
		return TRecSet::New(CategoryStore);
	uint64 CategoryRecId = CategoryStore->GetRecId(CategoryNm);
	return GetChildCategories(CategoryRecId);
}

PRecSet TCategoryBase::GetChildCategories(const uint64& CategoryRecId)
{
	AssertR(CategoryStore->IsRecId(CategoryRecId), "Category record id is not valid");
	TRec CatRec(CategoryStore, CategoryRecId);
	return CatRec.DoJoin(Base, "hasSubcategory");
}

// return the similarity between two categories
// "/A/B/C" and "/A/X" have A in common so it should return 1 / MIN(2,3) = 0.5
double TCategoryBase::GetCategorySimilarity(const uint64& CatRecId1, const uint64& CatRecId2)
{
	TUInt64V Cats1V, Cats2V;
	uint64 CatId1 = CatRecId1, CatId2 = CatRecId2;
	while (CatId1 != TUInt64::Mx) {
		// debug:
		//TStr Label = CategoryStore->GetFieldNmStr(CatId1, "Label");
		// ignore the root category
		if (CatId1 != 0)
			Cats1V.Add(CatId1);
		TRec ParentCat = CategoryStore->GetRec(CatId1).DoSingleJoin(Base, "hasSupercategory");
		CatId1 = ParentCat.GetRecId();
	}
	while (CatId2 != TUInt64::Mx) {
		// debug:
		//TStr Label = CategoryStore->GetFieldNmStr(CatId2, "Label");
		// ignore the root category
		if (CatId2 != 0)
			Cats2V.Add(CatId2);
		TRec ParentCat = CategoryStore->GetRec(CatId2).DoSingleJoin(Base, "hasSupercategory");
		CatId2 = ParentCat.GetRecId();
	}
	if (Cats1V.Len() == 0 || Cats2V.Len() == 0)
		return 0;

	Cats1V.Reverse();
	Cats2V.Reverse();
	// find the common subvector length
	int LastMatchIndex = -1;
	for (int N = 0; N < Cats1V.Len() && N < Cats2V.Len(); N++) {
		if (Cats1V[N] != Cats2V[N])
			break;
		LastMatchIndex = N;
	}
	return (LastMatchIndex + 1) / (double) MIN(Cats1V.Len(), Cats2V.Len());
}


PJsonVal TCategoryBase::Category(const uint64& CategoryRecId)
{
	const TStr Uri = CategoryStore->GetRecNm(CategoryRecId);
	const TStr Label = CategoryStore->GetFieldNmStr(CategoryRecId, "Label");
	PJsonVal CatJson = TJsonVal::NewObj("uri", Uri);
	CatJson->AddToObj("label", Label);
	return CatJson;
}

PJsonVal TCategoryBase::Categories(const PRecSet& CategoryRecSet)
{
	if (CategoryRecSet->GetRecs() == 0)
		return NULL;

	PJsonVal CatArr = TJsonVal::NewArr();
	for (int N = 0; N < CategoryRecSet->GetRecs(); N++)
		CatArr->AddToArr(Category(CategoryRecSet->GetRecId(N)));
	return CatArr;
}
