/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ConceptFolderBase.h"

void TConceptFolderSettings::LoadSettings(const PJsonVal& Settings)
{
	MinSecondsBetweenArticleUpdates = (int) Settings->GetObjNum("MinSecondsBetweenArticleUpdates", 60);
	MinSecondsBetweenEventUpdates = (int) Settings->GetObjNum("MinSecondsBetweenEventUpdates", 60);

	MaxArticleScore = (int) Settings->GetObjNum("MaxArticleScore", 1000);
	MaxEventScore = (int) Settings->GetObjNum("MaxEventScore", 1000);
	//_MaxScoreForValidLocation = Settings->GetObjNum("MaxScoreForValidLocation", 150);
	MaxScoreForArticlePopularity = (int) Settings->GetObjNum("MaxScoreForArticlePopularity", 150);
	DefSourceWgt = (int) Settings->GetObjNum("DefSourceWgt", 100);
	ArtPopReadScore = Settings->GetObjNum("ArtPopReadScore", 1.0);
	ArtPopLikeScore = Settings->GetObjNum("ArtPopLikeScore", 10.0);
	ArtPopStoreScore = Settings->GetObjNum("ArtPopStoreScore", 10.0);
	VerbosityLevel = (int) Settings->GetObjNum("VerbosityLevel", 3);
}

TConceptFolderBase::TConceptFolderBase(PNewsBase _NewsBase, const TStr& SettingsFNm) : NewsBase(_NewsBase)
{
	Logger = TLogger::New(3);
	Logger->AddLogger(new TFileNotify("./Logs/Log-ConceptFolders.txt", true, true));

	Base = NewsBase->Base;
	IndexVoc = Base->GetIndexVoc();
	Index = Base->GetIndex();
	
	ArticleStore = NewsBase->ArticleStore;
	ConceptStore = NewsBase->ConceptStore;
	StoryStore = NewsBase->StoryStore;
	EventStore = NewsBase->EventStore;
	ConceptFolderStore = NewsBase->ConceptFolderStore;
	UserStore = NewsBase->UserStore;

	// load default settings
	SettingsFName = SettingsFNm;
	ReloadSettings();

	// make sure we have the default user who is the owner of the concept folders
	if (! UserStore->IsRecNm(AdminUserUri)) {
		AddUser(AdminUserUri, "Administrator");
	}
	// load concept folder related data
	Load();
}

TConceptFolderBase::~TConceptFolderBase()
{
	Save();
}

void TConceptFolderBase::Load()
{
	TStr FName = Base->GetFPath() + "CustomData/ConceptFolderData.bin";
	if (TFile::Exists(FName)) {
		TFIn FIn(FName);
		AvailableFolderUriV.Load(FIn);
	}
}
void TConceptFolderBase::Save()
{
	TStr FName = Base->GetFPath() + "CustomData/ConceptFolderData.bin";
	TFOut FOut(FName);
	AvailableFolderUriV.Save(FOut);
}

// read settings from a json configuration file
void TConceptFolderBase::ReloadSettings()
{
	const TStr SettingsStr = TFile::Exists(SettingsFName) ? TStr::LoadTxt(SettingsFName) : "{}";
	const PJsonVal AllSettings = TJsonVal::GetValFromStr(SettingsStr);
	const PJsonVal CFSettings = AllSettings->IsObjKey("conceptFolders") ? AllSettings->GetObjKey("conceptFolders") : TJsonVal::NewObj();

	Settings.LoadSettings(CFSettings);
	Logger->SetVerbosityLevel(Settings.VerbosityLevel);
}

uint64 TConceptFolderBase::AddUser(const TStr& UserUri, const TStr& UserName)
{
	PJsonVal Json = TJsonVal::NewObj("URI", UserUri);
	Json->AddToObj("Name", UserName);
	const uint64 UserId = UserStore->AddRec(Json);
	return UserId;
}

// create a new concept folder and return new concept uri
// NOTE: concept uri and concept folder uri will be the same.
TStr TConceptFolderBase::AddConceptFolder(const TStr& UserUri, const PConceptFolder& Folder, const TStr& Label, const TStr& Description)
{
	const TStr Uri = CreateConceptFolderUri();

	PJsonVal Json = TJsonVal::NewObj("URI", Uri);
	// set label for concept
	for (int N = 0; N < NewsConst.LangCoreV.Len(); N++)
		NewsBase->ConceptLabel->SetConceptLabel(Uri, NewsConst.LangCoreV[N], Label);
	Json->AddToObj("Type", NewsConst.TypeConceptFolder);
	Json->AddToObj("Description", Description);

	const uint64 ConceptId = ConceptStore->AddRec(Json);

	// set the uri for the new concept folder
	Folder->Uri = Uri;
	const uint64 ConceptFolderId = ConceptFolderStore->AddRec(Folder->GetJsonForStore());

	EAssertR(UserStore->IsRecNm(UserUri), TStr::Fmt("User with uri '%s' is not known.", UserUri.CStr()));
	const uint64 UserId = UserStore->GetRecId(UserUri);

	ConceptFolderStore->AddJoin("hasUser", ConceptFolderId, UserId);
	
	// create joins for all concept folders (concept folder <--> concept)
	ConceptFolderStore->DelJoins("hasConcept", ConceptFolderId);
	for (int N = 0; N < Folder->ConceptIdV.Len(); N++) {
		if (!ConceptStore->IsRecId((uint64) Folder->ConceptIdV[N]))
			continue;
		if (Folder->IsConceptFixed(Folder->ConceptIdV[N]))
			ConceptFolderStore->AddJoin("hasConcept", ConceptFolderId, (uint64) Folder->ConceptIdV[N]);
	}
	return Uri;
}

// update name, visibility and autoAddArticles for a concept folder
bool TConceptFolderBase::UpdateConceptFolder(const PConceptFolder& Folder, TStr& ErrorInfo)
{
	if (!ConceptFolderStore->IsRecNm(Folder->Uri)) {
		ErrorInfo = "Concept uri does not represent a known concept folder"; return false;
	}
	if (Folder->ConceptIdV.Len() != Folder->ConceptWgtV.Len()) {
		ErrorInfo = "Concept uri and concept weight arrays are of different lengths. Update was canceled"; return false;
	}
	if (Folder->SourceIdV.Len() != Folder->SourceWgtV.Len()) {
		ErrorInfo = "Source uri and source weight arrays are of different lengths. Update was canceled"; return false;
	}
	if (Folder->CategoryIdV.Len() != Folder->CategoryWgtV.Len()) {
		ErrorInfo = "Category uri and category weight arrays are of different lengths. Update was canceled"; return false;
	}
	if (Folder->LocationIdV.Len() != Folder->LocationWgtV.Len()) {
		ErrorInfo = "Location uri and location weight arrays are of different lengths. Update was canceled"; return false;
	}
	if (Folder->KeywordV.Len() != Folder->KeywordWgtV.Len()) {
		ErrorInfo = "Keyword and keyword weight arrays are of different lengths. Update was canceled"; return false;
	}

	// since we've updated the definition of the folder we also have to remove the articles and events associated with the folder and set counts to 0
	if (Folder->AutoAddArticles == true) {
		ClearConceptFolderContent(Folder, false);
	}

	// call AddRec instead of UpdateRec since we have URI inside and not $id
	ConceptFolderStore->AddRec(Folder->GetJsonForStore());

	// we have to manually check the joins to concept store and update them based on the current concepts in folder. those joins don't update automatically
	const uint64 ConceptFolderId = GetFolderId(Folder->Uri);
	ConceptFolderStore->DelJoins("hasConcept", ConceptFolderId);
	for (int N = 0; N < Folder->ConceptIdV.Len(); N++) {
		if (!ConceptStore->IsRecId((uint64) Folder->ConceptIdV[N]))
			continue;
		if (Folder->IsConceptFixed(Folder->ConceptIdV[N]))
			ConceptFolderStore->AddJoin("hasConcept", ConceptFolderId, (uint64) Folder->ConceptIdV[N]);
	}
	
	return true;
}

// remove all articles and events that have been assigned to the concept folder
void TConceptFolderBase::ClearConceptFolderContent(const PConceptFolder& Folder, const bool& SaveFolderChanges) {
	Folder->LastUpdatedArticleId = 0;
	Folder->TimeOfLastArticleUpdate = TTm::GetTmFromMSecs(0);
	Folder->TimeOfLastEventUpdate = TTm::GetTmFromMSecs(0);
	Folder->ArticlesInFolder = 0;
	const uint64 FolderId = ConceptFolderStore->GetRecId(Folder->Uri);
	ConceptFolderStore->DelJoins("hasArticle", FolderId);
	ConceptFolderStore->DelJoins("hasEvent", FolderId);

	if (SaveFolderChanges)
		ConceptFolderStore->AddRec(Folder->GetJsonForStore());
}

// delete an existing concept folder
void TConceptFolderBase::DeleteConceptFolder(const TStr& ConceptFolderUri)
{
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri))
		return;
	// put an empty folder in place of existing ConceptFolderUri
	PConceptFolder Folder = TConceptFolder::New();
	Folder->Uri = ConceptFolderUri;
	TStr ErrorInfo;
	bool Ok = UpdateConceptFolder(Folder, ErrorInfo);
	
	const uint64 ConceptFolderId = GetFolderId(ConceptFolderUri);
	const uint64 ConceptId = GetConceptId(ConceptFolderUri);
	const uint64 UserId = GetUserId(ConceptFolderUri);

	// keep the concept folder <--> concept in the index. the next use of this concept folder uri will reuse it
	// just clear the label for concept
	if (ConceptStore->IsRecId(ConceptId)) {
		NewsBase->ConceptLabel->RemoveConcept(ConceptFolderUri);
		ConceptStore->SetFieldNmStr(ConceptId, "Description", "");
	}

	// delete concept folder <--> user
	ConceptFolderStore->DelJoins("hasUser", ConceptFolderId);
	
	// delete all joins article, story, event <--> concept folder
	ConceptFolderStore->DelJoins("hasArticle", ConceptFolderId);
	ConceptFolderStore->DelJoins("hasStory", ConceptFolderId);
	ConceptFolderStore->DelJoins("hasEvent", ConceptFolderId);

	// mark the concept folder as 
	AddAvailableFolderUri(ConceptFolderUri);
}

// ////////////////////////
// is concept with conceptId a concept folder?
bool TConceptFolderBase::IsConceptIdOfConceptFolder(const uint64& ConceptId)
{
	if (!ConceptStore->IsRecId(ConceptId))
		return false;
	const TStr Type = ConceptStore->GetFieldNmStr(ConceptId, "Type");
	return NewsConst.ConceptFolderSet.IsKey(Type);
}

bool TConceptFolderBase::IsFolder(const TStr& ConceptFolderUri)
{
	return ConceptFolderStore->IsRecNm(ConceptFolderUri);
}

bool TConceptFolderBase::IsConcept(const TStr& ConceptFolderUri)
{
	return ConceptStore->IsRecNm(ConceptFolderUri);
}

uint64 TConceptFolderBase::GetConceptId(const TStr& ConceptFolderUri)
{
	return NewsBase->ConceptStore->IsRecNm(ConceptFolderUri) ? 
		(uint64) NewsBase->ConceptStore->GetRecId(ConceptFolderUri) : (uint64) TUInt64::Mx;
}

uint64 TConceptFolderBase::GetFolderId(const TStr& ConceptFolderUri)
{
	return NewsBase->ConceptFolderStore->IsRecNm(ConceptFolderUri) ? 
		(uint64) NewsBase->ConceptFolderStore->GetRecId(ConceptFolderUri) : (uint64) TUInt64::Mx;
}

// return the owner of the concept folder
uint64 TConceptFolderBase::GetUserId(const TStr& ConceptFolderUri)
{
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri))
		return TUInt64::Mx;

	const uint64 FolderId = ConceptFolderStore->GetRecId(ConceptFolderUri);
	const uint64 UserId = ConceptFolderStore->GetFieldNmUInt64(FolderId, "hasUserId");
	return UserId;
}

// ////////////////////////
// add to index new articles that belong to the concept folder
void TConceptFolderBase::UpdateConceptFolderArticles(const TStr& ConceptFolderUri)
{
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri)) {
		NotifyVerboseFmt(0, "ERROR: UpdateConceptFolderArticles. Trying to clear concept folder for invalid concept '%s'", ConceptFolderUri.CStr());
		return;
	}

	const PConceptFolder ConceptFolder = TConceptFolder::New(ConceptFolderUri, ConceptFolderStore);

	// only update concept folder if AutoAddArticles == true
	if (ConceptFolder->AutoAddArticles != true)
		return;

	// update the index only after enough time has passed
	TTm TmNow = TTm::GetCurLocTm();
	uint64 DiffSeconds = TTm::GetDiffSecs(TmNow, ConceptFolder->TimeOfLastArticleUpdate);
	if (DiffSeconds < Settings.MinSecondsBetweenArticleUpdates) {
		NotifyVerboseFmt(2, "UpdateConceptFolderArticles: Concept folder '%s' is not updated since only %I64u have passed passed since last update.", ConceptFolderUri.CStr(), DiffSeconds);
		return;
	}
	
	TTmStopWatch StopWatch(true);
	
	NotifyVerboseFmt(2, "Updating concept folder '%s'", ConceptFolderUri.CStr());
	// find new articles that we have to add to the index
	PRecSet Articles = GetSuggestedArticlesForFolder(ConceptFolder, ConceptFolder->LastUpdatedArticleId);
	int ArticlesSinceLastUpdate = Articles->GetRecs();

	TWPt<TStore> ArticleStore = Base->GetStoreByStoreNm("Article");
	const uint64 NewestArticleId = ArticleStore->GetLastRecId();
	ConceptFolder->LastUpdatedArticleId = NewestArticleId > 0 ? NewestArticleId+1 : 0;
	ConceptFolder->TimeOfLastArticleUpdate = TmNow;
	ConceptFolder->ArticlesInFolder += Articles->GetRecs();
	// save changes 
	ConceptFolderStore->AddRec(ConceptFolder->GetJsonForStore());

	// add articles to index
	const uint64 ConceptId = GetFolderId(ConceptFolderUri);
	const int Recs = Articles->GetRecs();
	if (ConceptFolderStore->IsJoinNm("hasArticle")) {
		for (int N = 0; N < Recs; N++) {
			// index concept->article join
			const int RecFq = Articles->GetRecFq(N);
			ConceptFolderStore->AddJoin("hasArticle", ConceptId, Articles->GetRecId(N), RecFq);
		}
	}
	
	NotifyVerboseFmt(2, "UpdateConceptFolderArticles on folder '%s' took %d ms. Added articles to folder: %d. Seconds since last update: %I64u", ConceptFolderUri.CStr(), StopWatch.GetMSecInt(), ArticlesSinceLastUpdate, DiffSeconds);
}

// ////////////////////////
// add to index new articles that belong to the concept folder
void TConceptFolderBase::UpdateConceptFolderEvents(const TStr& ConceptFolderUri)
{
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri)) {
		NotifyVerboseFmt(0, "ERROR: UpdateConceptFolderEvents. Trying to clear concept folder for invalid concept '%s'", ConceptFolderUri.CStr());
		return;
	}

	const PConceptFolder ConceptFolder = TConceptFolder::New(ConceptFolderUri, ConceptFolderStore);

	// only update concept folder if AutoAddArticles == true
	if (ConceptFolder->AutoAddArticles != true)
		return;

	// update the index only after enough time has passed
	TTm TmNow = TTm::GetCurLocTm();
	uint64 DiffSeconds = TTm::GetDiffSecs(TmNow, ConceptFolder->TimeOfLastEventUpdate);
	if (DiffSeconds < Settings.MinSecondsBetweenEventUpdates) {
		NotifyVerboseFmt(2, "UpdateConceptFolderEvents: Concept folder '%s' is not updated since only %I64u seconds have passed since last update.", ConceptFolderUri.CStr(), DiffSeconds);
		return;
	}

	TTmStopWatch StopWatch(true);
	NotifyVerboseFmt(2, "Updating concept folder '%s'", ConceptFolderUri.CStr());
			
	ConceptFolder->TimeOfLastEventUpdate = TmNow;
	// save changes 
	ConceptFolderStore->AddRec(ConceptFolder->GetJsonForStore());

	// find new articles that we have to add to the index
	const PRecSet EventSet = GetSuggestedEventsForFolder(ConceptFolder, 0);
	// add events to index
	const uint64 FolderId = GetFolderId(ConceptFolderUri);
	const int Recs = EventSet->GetRecs();
	if (ConceptFolderStore->IsJoinNm("hasEvent")) {
		ConceptFolderStore->DelJoins("hasEvent", FolderId);
		for (int N = 0; N < Recs; N++) {
			// index folder->article join
			const int RecFq = EventSet->GetRecFq(N);
			ConceptFolderStore->AddJoin("hasEvent", FolderId, EventSet->GetRecId(N), RecFq);
		}
	}
	
	NotifyVerboseFmt(2, "UpdateConceptFolderEvents on folder '%s' took %d ms. Matching events: %d. Seconds since last update: %I64u", ConceptFolderUri.CStr(), StopWatch.GetMSecInt(), EventSet->GetRecs(), DiffSeconds);
}

// return a set of articles that are similar to the definition of the folder FolderRecId. 
// the returned articles are sorted based on their score (how well they match the folder)
PRecSet TConceptFolderBase::GetSuggestedArticlesForFolder(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, const bool& RemoveDuplicateArticles)
{
	TArticleSetH ArticleWgtH;
	// find suggested articles based on concepts
	GetSuggestedArticlesBasedOnConcepts(ConceptFolder, MinArticleId, ArticleWgtH);
	if (ConceptFolder->RestrictToSetConcepts && ArticleWgtH.Len() == 0)
		return TRecSet::New(ArticleStore);	// if no concepts match the criteria and RestrictToSetConcepts = true then return empty set

	// update scores based on article sources the user is interested in
	UpdateArticlesBasedOnSources(ConceptFolder, MinArticleId, ArticleWgtH);
	// update scores based on article categories
	UpdateArticlesBasedOnCategories(ConceptFolder, MinArticleId, ArticleWgtH);
	// update scores based on locations of sources of article
	UpdateArticlesBasedOnLocations(ConceptFolder, MinArticleId, ArticleWgtH);
	// update scores based on specified keywords
	UpdateArticlesBasedOnKeywords(ConceptFolder, MinArticleId, ArticleWgtH);

	// update scores based on article source importance
	//UpdateArticlesBasedOnSourceProminence(ConceptFolder, MinArticleId, ArticleWgtH);
	// update scores based on article popularity (nr of reads, likes, stores)
	//UpdateArticlesBasedOnArticlePopularity(ConceptFolder, MinArticleId, Settings.MaxScoreForArticlePopularity, ArticleWgtH);

	// add other reranking functionality

	// remove article with too low score
	const int MinArticleScore = ConceptFolder->ArticleTreshWgt;
	if (MinArticleScore > 0) {
		TArticleSetH FilteredArticleWgtH;
		for (int N = ArticleWgtH.FFirstKeyId(); ArticleWgtH.FNextKeyId(N);) {
			const int Score = ArticleWgtH[N];
			if (Score >= MinArticleScore)
				FilteredArticleWgtH.AddDat(ArticleWgtH.GetKey(N), Score);
		}
		ArticleWgtH = FilteredArticleWgtH;
	}

	// remove articles in wrong languages
	if (ConceptFolder->LangV.Len() > 0) {
		TStrSet LangH(ConceptFolder->LangV);
		TArticleSetH FilteredArticleWgtH;
		for (int N = ArticleWgtH.FFirstKeyId(); ArticleWgtH.FNextKeyId(N);) {
			if (LangH.IsKey(ArticleStore->GetFieldNmStr(ArticleWgtH.GetKey(N), "Language")))
				FilteredArticleWgtH.AddDat(ArticleWgtH.GetKey(N), ArticleWgtH[N]);
		}
		ArticleWgtH = FilteredArticleWgtH;
	}

	if (ArticleWgtH.Len() == 0)
		return TRecSet::New(ArticleStore);	// make sure we don't have a null recset

	// remove duplicates if needed
	if (RemoveDuplicateArticles) {
		RemoveDuplicates(ArticleWgtH);
	}
	TUInt64IntKdV ArticleKdV;
	ArticleWgtH.GetKeyDatKdV(ArticleKdV);
	return TRecSet::New(ArticleStore, ArticleKdV, true);
}

void TConceptFolderBase::GetSuggestedArticlesBasedOnConcepts(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	TTmStopWatch StopWatch(true);

	if (Settings.VerbosityLevel >= 3) {
		NotifyVerbose(3, "Folder definition:");
		PrintFolderDefinition(ConceptFolder);
	}

	TUInt64FltKdV ConceptKdV; ConceptFolder->GetConceptKdV(ConceptKdV, true);
	
	// for the concepts in the folder that are concept folders update the articles in the folder
	for (int ConceptN = 0; ConceptN < ConceptKdV.Len(); ConceptN++) {
		if (IsConceptIdOfConceptFolder(ConceptKdV[ConceptN].Key)) {
			const TStr FolderUri = ConceptStore->GetRecNm(ConceptKdV[ConceptN].Key);
			UpdateConceptFolderArticles(FolderUri);
		}
	}

	// take only 20 most important concepts
	ConceptKdV.Trunc(20);
	
	if (ConceptKdV.Len() > 0) {
		GetSuggestedArticlesForConcepts(ConceptKdV, MinArticleId, -1, ArticleWgtH);
		NotifyVerboseFmt(2, "GetSuggestedArticlesBasedOnConcepts found %d matching articles", ArticleWgtH.Len());
	}
	// if there are no conditions specified return the latest articles
	else {
		NotifyVerbose(2, "GetSuggestedArticlesBasedOnConcepts - no concepts were specified therefore we return 0 matching articles");
	}
	NotifyVerboseFmt(2, "GetSuggestedArticlesBasedOnConcepts took %d ms", StopWatch.GetMSecInt());
}

// return a recset with articles that are annotated with concepts in HasConceptV. Articles are sorted based on how well they match HasConceptV
// GetRecFq will return a score of the article in range [0, _MaxArticleScore(=1000)]
void TConceptFolderBase::GetSuggestedArticlesForConcepts(const TUInt64FltKdV& HasConceptV, const uint64& MinArticleId, const int& MaxArticleCount, TArticleSetH& ArticleWgtH)
{
	//LogToFile("--------------\nComputing suggested articles for concepts. Concept count = " + TInt(HasConceptV.Len()).GetStr());
	TUInt64 TotalArticles = ArticleStore->GetRecs();
	// in this loop we only process normal concepts - not folder concepts
	for (int ConceptN = 0; ConceptN < HasConceptV.Len(); ConceptN++) {
		TTmStopWatch StopWatch(true);
		TUInt64 ConceptId = HasConceptV[ConceptN].Key;
		// concept folders are processed in a different loop
		if (IsConceptIdOfConceptFolder(ConceptId))
			continue;

		TFlt ConceptWgt = HasConceptV[ConceptN].Dat;
		PRecSet RecSet = ConceptStore->GetRec(ConceptId).DoJoin(Base, "hasArticle");
		TFlt IDFConcept = log(TotalArticles / (double) (RecSet->GetRecs() + 1)) / 2;
		IDFConcept = MIN((double) IDFConcept, (double) 5.0);		// normalize IDF to interval between 0 - 5
		
		// take only articles after MinArticleId
		for (int RecN = RecSet->GetRecs() - 1; RecN >= 0; RecN--) {
			const uint64 ArticleId = RecSet->GetRecId(RecN);
			if (ArticleId <= MinArticleId) 
				break;					// ignore articles that are more than x days old
			const int Wgt = (int) (RecSet->GetRecFq(RecN) * ConceptWgt * IDFConcept);
			ArticleWgtH.AddDat(ArticleId) += Wgt;
		}
		if (Settings.VerbosityLevel >= 3)
			NotifyVerboseFmt(3, "GetSuggestedArticlesForConcepts took %d ms for %d articles with concept %s", StopWatch.GetMSecInt(), RecSet->GetRecs(), ConceptStore->GetRecNm(ConceptId).CStr());
	}
	
	// here we only process concept folders. instead of adding Wgt to ArticlesForInterestH we have to here just use the max of existing Wgt and new wgt
	// this is needed because otherwise we could get very high values for some articles since the same concept would be taken into account multiple times
	// once in the concept folder and once as a separate concept
	for (int ConceptN = 0; ConceptN < HasConceptV.Len(); ConceptN++) {
		TUInt64 ConceptId = HasConceptV[ConceptN].Key;
		if (!IsConceptIdOfConceptFolder(ConceptId))
			continue;
		const TStr ConceptUri = ConceptStore->GetRecNm(ConceptId);
		// update concept folder
		UpdateConceptFolderArticles(ConceptUri);
		
		const double ConceptWgt = HasConceptV[ConceptN].Dat;
		PRecSet RecSet = ConceptStore->GetRec(ConceptId).DoJoin(Base, "hasArticle");
		NotifyVerboseFmt(2, "Concept folder: %s (wgt = %.3f)", ConceptStore->GetFieldNmStr(ConceptId, "URI").CStr(), ConceptWgt);
		int Recs = RecSet->GetRecs();
		for (int RecN = 0; RecN < Recs; RecN++) {
			const uint64 ArticleId = RecSet->GetRecId(RecN);
			if (ArticleId < MinArticleId) continue;					// ignore articles that are more than x days old
			const int RecFq = RecSet->GetRecFq(RecN);			// for concept folders, RecFq's are in range [0, _MaxArticleScore]
			const int Wgt = (int) (RecFq * ConceptWgt * 0.1);	// 0.1 is to compensate for the fact that ConceptWgts are [0, 10]
			// here we use a different function - just take the maximum of the current article score and the new score
			ArticleWgtH.AddDat(ArticleId, ArticleWgtH.IsKey(ArticleId) ? MAX(ArticleWgtH.GetDat(ArticleId), Wgt) : Wgt);
		}
	}

	const int MaxScore = Settings.MaxArticleScore;
	for (int KeyId = ArticleWgtH.FFirstKeyId(); ArticleWgtH.FNextKeyId(KeyId);) {
		if (ArticleWgtH[KeyId] > MaxScore)
			ArticleWgtH[KeyId] = MaxScore;
	}
}

void TConceptFolderBase::UpdateArticlesBasedOnSources(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	// find articles based on valid sources for folder
	if (ConceptFolder->SourceIdV.Len() == 0)
		return;
	TTmStopWatch StopWatch(true);
	// find articles from the valid sources, score them based on source prominence between interval 0 - MaxScoreForValidSource
	TArticleSetH SourceArticleWgtH;
	GetArticlesForSources(ConceptFolder, MinArticleId, SourceArticleWgtH);

	// if we specified which are the valid sources then remove the articles that belong to different sources
	// NOTE: an exception is when we have empty article set (maybe no specified concepts)
	if (ConceptFolder->RestrictToSetSources && ArticleWgtH.Len() > 0)
		ArticleWgtH = IntersectHashSets(ArticleWgtH, SourceArticleWgtH);
	else
		ArticleWgtH = MergeHashSets(ArticleWgtH, SourceArticleWgtH);
	NotifyVerboseFmt(2, "UpdateArticlesBasedOnSources on %d articles took %d ms", SourceArticleWgtH.Len(), StopWatch.GetMSecInt());
}

void TConceptFolderBase::UpdateArticlesBasedOnCategories(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	// find articles based on valid categories for folder
	if (ConceptFolder->CategoryIdV.Len() == 0)
		return;
	TTmStopWatch StopWatch(true);
	// find articles from the valid sources, score them based on source prominence between interval 0 - MaxScoreForValidSource
	TArticleSetH CategoryArticleWgtH;
	GetArticlesForCategories(ConceptFolder, MinArticleId, CategoryArticleWgtH);
		
	if (ConceptFolder->RestrictToSetCategories && ArticleWgtH.Len() > 0)
		ArticleWgtH = IntersectHashSets(ArticleWgtH, CategoryArticleWgtH);
	else
		ArticleWgtH = MergeHashSets(ArticleWgtH, CategoryArticleWgtH);
	NotifyVerboseFmt(2, "UpdateArticlesBasedOnCategories on %d articles took %d ms", CategoryArticleWgtH.Len(), StopWatch.GetMSecInt());
}

void TConceptFolderBase::UpdateArticlesBasedOnLocations(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	if (ConceptFolder->LocationIdV.Len() == 0)
		return;
	TTmStopWatch StopWatch(true);
	// find articles from the sources, score them based on source prominence between interval 0 - MaxScoreForValidSource
	TArticleSetH LocationArticleWgtH;
	GetArticlesForLocations(ConceptFolder, MinArticleId, LocationArticleWgtH);

	// if we specified which are the sources then remove the articles that belong to different sources		
	if (ConceptFolder->RestrictToSetLocations)
		ArticleWgtH = IntersectHashSets(ArticleWgtH, LocationArticleWgtH);		// check if we need to switch the order in order to keep recFqs
	// for the articles that are from the sources increase their rating based on the source importance
	else
		ArticleWgtH = MergeHashSets(ArticleWgtH, LocationArticleWgtH);
	NotifyVerboseFmt(2, "UpdateArticlesBasedOnLocation on %d articles took %d ms", LocationArticleWgtH.Len(), StopWatch.GetMSecInt());
}

void TConceptFolderBase::UpdateArticlesBasedOnKeywords(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	if (ConceptFolder->LocationIdV.Len() == 0)
		return;
	TTmStopWatch StopWatch(true);
	// find articles containing the specified keywords, score them based on how many specified keywords appear in each of the articles
	TArticleSetH KeywordArticleWgtH;
	GetArticlesForKeywords(ConceptFolder, MinArticleId, KeywordArticleWgtH);

	// if we specified which are the sources then remove the articles that belong to different sources		
	if (ConceptFolder->RestrictToSetKeywords)
		ArticleWgtH = IntersectHashSets(ArticleWgtH, KeywordArticleWgtH);		// check if we need to switch the order in order to keep recFqs
	// for the articles that are from the sources increase their rating based on the source importance
	else
		ArticleWgtH = MergeHashSets(ArticleWgtH, KeywordArticleWgtH);
	NotifyVerboseFmt(2, "UpdateArticlesBasedOnKeywords on %d articles took %d ms", KeywordArticleWgtH.Len(), StopWatch.GetMSecInt());
}

// change the article scores (Fq) in ArticleSet based on the prominence of the source
void TConceptFolderBase::UpdateArticlesBasedOnSourceProminence(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	TTmStopWatch StopWatch(true);
	for (int i = ArticleWgtH.FFirstKeyId(); ArticleWgtH.FNextKeyId(i);) {
		const uint64 ArticleRecId = ArticleWgtH.GetKey(i);
		const TRec SourceRec = ArticleStore->GetRec(ArticleRecId).DoSingleJoin(Base, "hasSource");
		const int SourceImportance = NewsBase->SourceStore->GetFieldNmInt(SourceRec.GetRecId(), "Importance");
		if (SourceImportance > 0) 
			ArticleWgtH.AddDat(ArticleRecId) += SourceImportance;
	}
	NotifyVerboseFmt(2, "UpdateArticlesBasedOnSourceProminence on %d articles took %d ms", ArticleWgtH.Len(), StopWatch.GetMSecInt());
}

void TConceptFolderBase::UpdateArticlesBasedOnArticlePopularity(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, const int& MaxScoreForArticlePopularity, TArticleSetH& ArticleWgtH)
{
	TTmStopWatch StopWatch(true);
	for (int i = ArticleWgtH.FFirstKeyId(); ArticleWgtH.FNextKeyId(i);) {
		const uint64 ArticleRecId = ArticleWgtH.GetKey(i);
		const int ReadCount = ArticleStore->GetFieldNmInt(ArticleRecId, "ReadCount");
		const int LikedCount = ArticleStore->GetFieldNmInt(ArticleRecId, "LikedCount");
		const int StoredCount = ArticleStore->GetFieldNmInt(ArticleRecId, "StoredCount");

		int PopularityScore = 10 * (int) (log(Settings.ArtPopReadScore * ReadCount + Settings.ArtPopLikeScore*LikedCount + Settings.ArtPopStoreScore*StoredCount) / log(2.0));
		PopularityScore = MIN(PopularityScore, MaxScoreForArticlePopularity);
		ArticleWgtH.AddDat(ArticleRecId) += PopularityScore;
	}
	NotifyVerboseFmt(2, "UpdateArticlesBasedOnArticlePopularity on %d articles took %d ms", ArticleWgtH.Len(), StopWatch.GetMSecInt());
}

// get all the articles from the sources that are specified in the SourceIdV vector
void TConceptFolderBase::GetArticlesForSources(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, THash<TUInt64, TInt>& ArticleWgtH)
{
	TTmStopWatch StopWatch(true);
	for (int N = 0; N < ConceptFolder->SourceIdV.Len(); N++) {
		const uint64 SourceId = ConceptFolder->SourceIdV[N];
		if (!NewsBase->SourceStore->IsRecId(SourceId))
			continue;
		const int SourceWgt = (int) ConceptFolder->SourceWgtV[N].Val;
		const PRecSet ArticleSet = NewsBase->SourceStore->GetRec(SourceId).DoJoin(Base, "hasArticle");
		for (int N = 0; N < ArticleSet->GetRecs(); N++) {
			const uint64 RecId = ArticleSet->GetRecId(N);
			if (RecId >= MinArticleId)
				ArticleWgtH.AddDat(RecId, SourceWgt);
		}
	}
	NotifyVerboseFmt(2, "GetArticlesForSources on %d articles took %d ms", ArticleWgtH.Len(), StopWatch.GetMSecInt());
}

// get all the articles from the sources that are specified in the SourceIdV vector
void TConceptFolderBase::GetArticlesForCategories(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, THash<TUInt64, TInt>& ArticleWgtH)
{
	TTmStopWatch StopWatch(true);
	for (int CatN = 0; CatN < ConceptFolder->CategoryIdV.Len(); CatN++) {
		const uint64 CategoryId = ConceptFolder->CategoryIdV[CatN];
		const double CategoryWgt = ConceptFolder->CategoryWgtV[CatN];
		if (!NewsBase->CategoryStore->IsRecId(CategoryId))
			continue;
		TUInt64Set CatRecIdSet;
		// add to CatRecIdV all categories that should be included in the search
		TUInt64V CatRecIdV; CatRecIdV.Add(CategoryId);
		if (ConceptFolder->IncludeSubCategories) {
			CatRecIdSet = NewsBase->CategoryBase->GetAllSubCategories(CatRecIdV);
			CatRecIdSet.GetKeyV(CatRecIdV);
		}
			
		// category rec set
		PRecSet CatRecSet = TRecSet::New(NewsBase->CategoryStore, CatRecIdV);
		// join articles for those categories
		PRecSet ArticleSet = CatRecSet->DoJoin(Base, "hasArticle");
		for (int N = 0; N < ArticleSet->GetRecs(); N++) {
			const uint64 ArticleId = ArticleSet->GetRecId(N);
			if (ArticleId >= MinArticleId) {
				int Wgt = (int) (CategoryWgt * ArticleSet->GetRecFq(N) * 0.01);		// article category fq is in range [0..100]. resulting wgt should be in [0..CategoryWgt]
				if (!ArticleWgtH.IsKey(ArticleId) || ArticleWgtH.GetDat(ArticleId) < Wgt)
					ArticleWgtH.AddDat(ArticleId, Wgt);
			}
		}
	}
	NotifyVerboseFmt(2, "GetArticlesForCategories on %d articles took %d ms", ArticleWgtH.Len(), StopWatch.GetMSecInt());
}

// get all the articles from the locations that are specified in the locations vector
void TConceptFolderBase::GetArticlesForLocations(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	TTmStopWatch StopWatch(true);
	const int CityCountryKeyId = IndexVoc->GetKeyId(NewsBase->SourceStoreId, "CityCountry");

	for (int LocationN = 0; LocationN < ConceptFolder->LocationIdV.Len(); LocationN++) {
		// if not valid word id continue
		if (!IndexVoc->IsWordId(CityCountryKeyId, ConceptFolder->LocationIdV[LocationN]))
			continue;
		const TStr Location = IndexVoc->GetWordStr(CityCountryKeyId, ConceptFolder->LocationIdV[LocationN]);
		const int LocationWgt = (int) ConceptFolder->LocationWgtV[LocationN];
		TQueryItemV QueryItemV;
		//QueryItemV.Add(TQueryItem(Base, "Source", "Country", Location));
		QueryItemV.Add(TQueryItem(Base, "Source", "CityCountry", Location));
		PQuery Query = TQuery::New(Base, TQueryItem(oqitOr, QueryItemV));
		PRecSet SourcesSet = Base->Search(Query);
		PRecSet ArticleSet = SourcesSet->DoJoin(Base, "hasArticle");
		ArticleSet->FilterByRecId(MinArticleId, TUInt64::Mx);
		const int Recs = ArticleSet->GetRecs();
		for (int RecN = 0; RecN < Recs; RecN++) {
			uint64 RecId = ArticleSet->GetRecId(RecN);
			ArticleWgtH.AddDat(RecId) += LocationWgt;
		}
	}
	NotifyVerboseFmt(2, "GetArticlesForLocations on %d articles took %d ms", ArticleWgtH.Len(), StopWatch.GetMSecInt());
}

// get all the articles from the locations that are specified in the locations vector
void TConceptFolderBase::GetArticlesForKeywords(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH)
{
	int FieldKeyId = ArticleStore->GetFieldKeyId(ArticleStore->GetFieldId("Body"));
	if (!IndexVoc->IsKeyId(FieldKeyId)) {
		NotifyVerboseFmt(0, "Article body does not have the body key. Unable to do keyword search");
		return;
	}
	if (ConceptFolder->KeywordV.Len() != ConceptFolder->KeywordWgtV.Len()) {
		NotifyVerboseFmt(0, "Folder %s has different sizes for keyword (%d) and keyword wgt (%d) arrays", ConceptFolder->Uri.CStr(), ConceptFolder->KeywordV.Len(), ConceptFolder->KeywordWgtV.Len());
		return;
	}
	TTmStopWatch StopWatch(true);

	for (int KeywordN = 0; KeywordN < ConceptFolder->KeywordV.Len(); KeywordN++) {
		const int KeywordWgt = (int) ConceptFolder->KeywordWgtV[KeywordN];
		TQueryItem KwQuery(Base, FieldKeyId, ConceptFolder->KeywordV[KeywordN]);
		TPair<TBool, PRecSet> Results = Index->Search(Base, KwQuery, Index->GetDefMerger());
		PRecSet ArticleSet = Results.Val2;
		if (MinArticleId > 0)
			ArticleSet->FilterByRecId(MinArticleId, TUInt64::Mx);
		const int Recs = ArticleSet->GetRecs();
		for (int RecN = 0; RecN < Recs; RecN++) {
			uint64 RecId = ArticleSet->GetRecId(RecN);
			ArticleWgtH.AddDat(RecId) += KeywordWgt;
		}
	}
	NotifyVerboseFmt(2, "GetArticlesForKeywords on %d articles took %d ms", ArticleWgtH.Len(), StopWatch.GetMSecInt());
}

// ////////////////////////
// get suggested events

// return a set of events that are similar to the definition of the folder FolderRecId. 
// the returned events are sorted based on their score (how well they match the folder)
PRecSet TConceptFolderBase::GetSuggestedEventsForFolder(const PConceptFolder& ConceptFolder, const uint64& MinEventId)
{
	TEventSetH EventWgtH;
	// find suggested events based on concepts
	GetSuggestedEventsBasedOnConcepts(ConceptFolder, MinEventId, EventWgtH);
	if (ConceptFolder->RestrictToSetConcepts && EventWgtH.Len() == 0)
		return TRecSet::New(StoryStore);		// if no concepts match the criteria and RestrictToSetConcepts = true then return empty set

	// update scores based on article sources the user is interested in
	UpdateEventsBasedOnSources(ConceptFolder, MinEventId, EventWgtH);
	// update scores based on event categories
	UpdateEventsBasedOnCategories(ConceptFolder, MinEventId, EventWgtH);
	// update scores based on specified keywords
	UpdateEventsBasedOnKeywords(ConceptFolder, MinEventId, EventWgtH);

	// add other reranking functionality

	// remove event with too low score
	const int MinEventScore = ConceptFolder->EventTreshWgt;
	if (MinEventScore > 0) {
		TEventSetH FilteredEventWgtH;
		for (int N = EventWgtH.FFirstKeyId(); EventWgtH.FNextKeyId(N);) {
			const int Score = EventWgtH[N];
			if (Score >= MinEventScore)
				FilteredEventWgtH.AddDat(EventWgtH.GetKey(N), Score);
		}
		EventWgtH = FilteredEventWgtH;
	}

	// remove events that are not covered by the specified languages
	if (ConceptFolder->LangV.Len() > 0) {
		TStrSet LangH(ConceptFolder->LangV);
		TEventSetH FilteredEventWgtH;
		for (int N = EventWgtH.FFirstKeyId(); EventWgtH.FNextKeyId(N);) {
			PRecSet StorySet = EventStore->GetRec(EventWgtH.GetKey(N)).DoJoin(Base, "hasStory");
			bool FoundLang = false;
			for (int L = 0; L < StorySet->GetRecs(); L++) {
				if (LangH.IsKey(StoryStore->GetFieldNmStr(StorySet->GetRecId(L), "Language"))) {
					FoundLang = true;
					break;
				}
			}
			if (FoundLang)
				FilteredEventWgtH.AddDat(EventWgtH.GetKey(N), EventWgtH[N]);
		}
		EventWgtH = FilteredEventWgtH;
	}


	if (EventWgtH.Len() == 0)
		return TRecSet::New(EventStore);	// make sure we don't have a null recset

	TUInt64IntKdV EventKdV;
	EventWgtH.GetKeyDatKdV(EventKdV);
	return TRecSet::New(EventStore, EventKdV, true);
}

void TConceptFolderBase::GetSuggestedEventsBasedOnConcepts(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH)
{
	TTmStopWatch StopWatch(true);

	if (Settings.VerbosityLevel >= 3) {
		NotifyVerbose(3, "Folder definition:");
		PrintFolderDefinition(ConceptFolder);
	}

	TUInt64FltKdV ConceptKdV; ConceptFolder->GetConceptKdV(ConceptKdV, true);

	// for the concepts in the folder that are concept folders update the events in the folder
	for (int ConceptN = 0; ConceptN < ConceptKdV.Len(); ConceptN++) {
		if (IsConceptIdOfConceptFolder(ConceptKdV[ConceptN].Key)) {
			const TStr FolderUri = ConceptStore->GetRecNm(ConceptKdV[ConceptN].Key);
			UpdateConceptFolderEvents(FolderUri);
		}
	}

	// take only 20 most important concepts
	ConceptKdV.Trunc(20);

	if (ConceptKdV.Len() > 0) {
		GetSuggestedEventsForConcepts(ConceptKdV, MinEventId, -1, EventWgtH);
		NotifyVerboseFmt(2, "GetSuggestedEventsBasedOnConcepts found %d matching events", EventWgtH.Len());
	}
	// if there are no conditions specified return the latest events
	else {
		NotifyVerbose(2, "GetSuggestedEventsBasedOnConcepts - no concepts were specified therefore we return 0 matching events");
	}
	NotifyVerboseFmt(2, "GetSuggestedEventsBasedOnConcepts took %d ms", StopWatch.GetMSecInt());
}

// return a recset with events that are annotated with concepts in HasConceptV. Events are sorted based on how well they match HasConceptV
// GetRecFq will return a score of the event in range [0, _MaxArticleScore(=1000)]
void TConceptFolderBase::GetSuggestedEventsForConcepts(const TUInt64FltKdV& HasConceptV, const uint64& MinEventId, const int& MaxEventCount, TEventSetH& EventWgtH)
{
	//LogToFile("--------------\nComputing suggested events for concepts. Concept count = " + TInt(HasConceptV.Len()).GetStr());
	const uint64 TotalArticles = NewsBase->NewsCounts->GetArticleCount();
	// in this loop we only process normal concepts - not folder concepts
	for (int ConceptN = 0; ConceptN < HasConceptV.Len(); ConceptN++) {
		TTmStopWatch StopWatch(true);
		const uint64 ConceptId = HasConceptV[ConceptN].Key;
		// concept folders are processed in a different loop
		if (IsConceptIdOfConceptFolder(ConceptId))
			continue;

		const double ConceptWgt = HasConceptV[ConceptN].Dat;
		const uint64 ConceptFq = MAX(1, NewsBase->NewsCounts->GetConceptFq(ConceptId));
		double IDFConcept = log(TotalArticles / (double) ConceptFq) / 2;
		IDFConcept = MIN((double) IDFConcept, (double) 5.0);		// normalize IDF to interval between 0 - 5

		// take only articles after MinEventId
		const PRecSet RecSet = ConceptStore->GetRec(ConceptId).DoJoin(Base, "hasEvent");
		for (int RecN = RecSet->GetRecs() - 1; RecN >= 0; RecN--) {
			const uint64 EventId = RecSet->GetRecId(RecN);
			if (EventId < MinEventId)
				break;
			const int Wgt = (int) (RecSet->GetRecFq(RecN) * ConceptWgt * IDFConcept / 20);
			EventWgtH.AddDat(EventId) += Wgt;
		}
		if (Settings.VerbosityLevel >= 3)
			NotifyVerboseFmt(3, "GetSuggestedEventsForConcepts took %d ms for %d events with concept %s", StopWatch.GetMSecInt(), RecSet->GetRecs(), ConceptStore->GetRecNm(ConceptId).CStr());
	}

	// here we only process concept folders. instead of adding Wgt to ArticlesForInterestH we have to here just use the max of existing Wgt and new wgt
	// this is needed because otherwise we could get very high values for some articles since the same concept would be taken into account multiple times
	// once in the concept folder and once as a separate concept
	for (int ConceptN = 0; ConceptN < HasConceptV.Len(); ConceptN++) {
		const uint64 ConceptId = HasConceptV[ConceptN].Key;
		if (!IsConceptIdOfConceptFolder(ConceptId))
			continue;
		const TStr ConceptUri = ConceptStore->GetRecNm(ConceptId);
		// update concept folder
		UpdateConceptFolderEvents(ConceptUri);

		const double ConceptWgt = HasConceptV[ConceptN].Dat;
		PRecSet RecSet = ConceptStore->GetRec(ConceptId).DoJoin(Base, "hasEvent");
		NotifyVerboseFmt(2, "Concept folder: %s (wgt = %.3f)", ConceptStore->GetFieldNmStr(ConceptId, "URI").CStr(), ConceptWgt);
		int Recs = RecSet->GetRecs();
		for (int RecN = 0; RecN < Recs; RecN++) {
			const uint64 EventId = RecSet->GetRecId(RecN);
			if (EventId < MinEventId) continue;					// ignore too old events
			const int RecFq = RecSet->GetRecFq(RecN);			// for concept folders, RecFq's are in range [0, _MaxArticleScore]
			const int Wgt = (int) (RecFq * ConceptWgt * 0.1);	// 0.1 is to compensate for the fact that ConceptWgts are [0, 10]
			// here we use a different function - just take the maximum of the current article score and the new score
			EventWgtH.AddDat(EventId, EventWgtH.IsKey(EventId) ? MAX(EventWgtH.GetDat(EventId), Wgt) : Wgt);
		}
	}

	const int MaxScore = Settings.MaxEventScore;
	for (int KeyId = EventWgtH.FFirstKeyId(); EventWgtH.FNextKeyId(KeyId);) {
		if (EventWgtH[KeyId] > MaxScore)
			EventWgtH[KeyId] = MaxScore;
	}
}

// compute events based on news sources. Merge or intersect them with events in EventWgtH
void TConceptFolderBase::UpdateEventsBasedOnSources(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TArticleSetH& EventWgtH)
{
	// find articles based on valid sources for folder
	if (ConceptFolder->SourceIdV.Len() == 0)
		return;
	TTmStopWatch StopWatch(true);
	// find articles from the valid sources, score them based on source prominence between interval 0 - MaxScoreForValidSource
	TArticleSetH SourceEventWgtH;
	GetEventsForSources(ConceptFolder, MinEventId, SourceEventWgtH);

	// if we specified which are the valid sources then remove the articles that belong to different sources
	// NOTE: an exception is when we have empty article set (maybe no specified concepts)
	if (ConceptFolder->RestrictToSetSources && EventWgtH.Len() > 0)
		EventWgtH = IntersectHashSets(EventWgtH, SourceEventWgtH);
	else
		EventWgtH = MergeHashSets(EventWgtH, SourceEventWgtH);
	NotifyVerboseFmt(2, "UpdateEventsBasedOnSources on %d articles took %d ms", SourceEventWgtH.Len(), StopWatch.GetMSecInt());
}

// compute events based on categories. Merge or intersect them with events in EventWgtH
void TConceptFolderBase::UpdateEventsBasedOnCategories(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TArticleSetH& EventWgtH)
{
	// find articles based on valid categories for folder
	if (ConceptFolder->CategoryIdV.Len() == 0)
		return;
	TTmStopWatch StopWatch(true);
	// find articles from the valid sources, score them based on source prominence between interval 0 - MaxScoreForValidSource
	TArticleSetH CategoryEventWgtH;
	GetEventsForCategories(ConceptFolder, MinEventId, CategoryEventWgtH);

	if (ConceptFolder->RestrictToSetCategories && EventWgtH.Len() > 0)
		EventWgtH = IntersectHashSets(EventWgtH, CategoryEventWgtH);
	else
		EventWgtH = MergeHashSets(EventWgtH, CategoryEventWgtH);
	NotifyVerboseFmt(2, "UpdateEventsBasedOnCategories on %d articles took %d ms", CategoryEventWgtH.Len(), StopWatch.GetMSecInt());
}

// compute events based on keywords. Merge or intersect them with events in EventWgtH
void TConceptFolderBase::UpdateEventsBasedOnKeywords(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TArticleSetH& EventWgtH)
{
	if (ConceptFolder->LocationIdV.Len() == 0)
		return;
	TTmStopWatch StopWatch(true);
	// find articles containing the specified keywords, score them based on how many specified keywords appear in each of the articles
	TArticleSetH KeywordEventWgtH;
	GetEventsForKeywords(ConceptFolder, MinEventId, KeywordEventWgtH);

	// if we specified which are the sources then remove the articles that belong to different sources		
	if (ConceptFolder->RestrictToSetKeywords)
		EventWgtH = IntersectHashSets(EventWgtH, KeywordEventWgtH);		// check if we need to switch the order in order to keep recFqs
	// for the articles that are from the sources increase their rating based on the source importance
	else
		EventWgtH = MergeHashSets(EventWgtH, KeywordEventWgtH);
	NotifyVerboseFmt(2, "UpdateEventsBasedOnKeywords on %d articles took %d ms", KeywordEventWgtH.Len(), StopWatch.GetMSecInt());
}

// get events that are covered by specific news sources
void TConceptFolderBase::GetEventsForSources(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH)
{
	TTmStopWatch StopWatch(true);
	for (int N = 0; N < ConceptFolder->SourceIdV.Len(); N++) {
		const uint64 SourceId = ConceptFolder->SourceIdV[N];
		if (!NewsBase->SourceStore->IsRecId(SourceId))
			continue;
		const int SourceWgt = (int) ConceptFolder->SourceWgtV[N].Val;
		const PRecSet ArticleSet = NewsBase->SourceStore->GetRec(SourceId).DoJoin(Base, "hasArticle");
		const PRecSet StorySet = ArticleSet->DoJoin(Base, "hasStory");
		const PRecSet EventSet = StorySet->DoJoin(Base, "hasEvent");
		for (int N = 0; N < EventSet->GetRecs(); N++) {
			const uint64 RecId = EventSet->GetRecId(N);
			if (RecId >= MinEventId)
				EventWgtH.AddDat(RecId, SourceWgt);
		}
	}
	NotifyVerboseFmt(2, "GetEventsForSources on %d articles took %d ms", EventWgtH.Len(), StopWatch.GetMSecInt());
}

// get events that are assigned to the specified categories
void TConceptFolderBase::GetEventsForCategories(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH)
{
	TTmStopWatch StopWatch(true);
	for (int CatN = 0; CatN < ConceptFolder->CategoryIdV.Len(); CatN++) {
		const uint64 CategoryId = ConceptFolder->CategoryIdV[CatN];
		const double CategoryWgt = ConceptFolder->CategoryWgtV[CatN];
		if (!NewsBase->CategoryStore->IsRecId(CategoryId))
			continue;
		TUInt64Set CatRecIdSet;
		// add to CatRecIdV all categories that should be included in the search
		TUInt64V CatRecIdV; CatRecIdV.Add(CategoryId);
		if (ConceptFolder->IncludeSubCategories) {
			CatRecIdSet = NewsBase->CategoryBase->GetAllSubCategories(CatRecIdV);
			CatRecIdSet.GetKeyV(CatRecIdV);
		}

		// category rec set
		PRecSet CatRecSet = TRecSet::New(NewsBase->CategoryStore, CatRecIdV);
		// join articles for those categories
		PRecSet EventSet = CatRecSet->DoJoin(Base, "hasEvent");
		for (int N = 0; N < EventSet->GetRecs(); N++) {
			const uint64 EventId = EventSet->GetRecId(N);
			if (EventId >= MinEventId) {
				int Wgt = (int) (CategoryWgt * EventSet->GetRecFq(N) * 0.01);		// article category fq is in range [0..100]. resulting wgt should be in [0..CategoryWgt]
				if (!EventWgtH.IsKey(EventId) || EventWgtH.GetDat(EventId) < Wgt)
					EventWgtH.AddDat(EventId, Wgt);
			}
		}
	}
	NotifyVerboseFmt(2, "GetEventsForCategories on %d articles took %d ms", EventWgtH.Len(), StopWatch.GetMSecInt());
}

// get events that contain articles that contain the specified keywords
void TConceptFolderBase::GetEventsForKeywords(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH)
{
	int FieldKeyId = ArticleStore->GetFieldKeyId(ArticleStore->GetFieldId("Body"));
	if (!IndexVoc->IsKeyId(FieldKeyId)) {
		NotifyVerboseFmt(0, "Article body does not have the body key. Unable to do keyword search");
		return;
	}
	if (ConceptFolder->KeywordV.Len() != ConceptFolder->KeywordWgtV.Len()) {
		NotifyVerboseFmt(0, "Folder %s has different sizes for keyword (%d) and keyword wgt (%d) arrays", ConceptFolder->Uri.CStr(), ConceptFolder->KeywordV.Len(), ConceptFolder->KeywordWgtV.Len());
		return;
	}
	TTmStopWatch StopWatch(true);

	for (int KeywordN = 0; KeywordN < ConceptFolder->KeywordV.Len(); KeywordN++) {
		const int KeywordWgt = (int) ConceptFolder->KeywordWgtV[KeywordN];
		TQueryItem KwQuery(Base, FieldKeyId, ConceptFolder->KeywordV[KeywordN]);
		TPair<TBool, PRecSet> Results = Index->Search(Base, KwQuery, Index->GetDefMerger());
		const PRecSet ArticleSet = Results.Val2;
		const PRecSet StorySet = ArticleSet->DoJoin(Base, "hasStory");
		const PRecSet EventSet = StorySet->DoJoin(Base, "hasEvent");
		if (MinEventId > 0)
			EventSet->FilterByRecId(MinEventId, TUInt64::Mx);
		const int Recs = EventSet->GetRecs();
		for (int RecN = 0; RecN < Recs; RecN++) {
			const uint64 RecId = EventSet->GetRecId(RecN);
			EventWgtH.AddDat(RecId) += KeywordWgt;
		}
	}
	NotifyVerboseFmt(2, "GetEventsForKeywords on %d articles took %d ms", EventWgtH.Len(), StopWatch.GetMSecInt());
}


// ////////////////////////
// curated methods

// add an article to a concept folder. called when concept folder has autoAddArticles = false
void TConceptFolderBase::AddCuratedArticleToConceptFolder(const TStr& ConceptFolderUri, const uint64& ArticleRecId)
{
	if (!ArticleStore->IsRecId(ArticleRecId))
		return;
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri))
		return;
	ConceptFolderStore->AddJoin("hasArticle", ConceptFolderStore->GetRecId(ConceptFolderUri), ArticleRecId, Settings.MaxArticleScore);
}

// remove the article from a concept folder
void TConceptFolderBase::RemoveCuratedArticleFromConceptFolder(const TStr& ConceptFolderUri, const uint64& ArticleRecId)
{
	if (!ArticleStore->IsRecId(ArticleRecId))
		return;
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri))
		return;
	ConceptFolderStore->DelJoin("hasArticle", ConceptFolderStore->GetRecId(ConceptFolderUri), ArticleRecId, Settings.MaxArticleScore);
}

// add an article to a concept folder. called when concept folder has autoAddArticles = false
void TConceptFolderBase::AddCuratedEventToConceptFolder(const TStr& ConceptFolderUri, const uint64& EventRecId)
{
	if (!EventStore->IsRecId(EventRecId))
		return;
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri))
		return;
	ConceptFolderStore->AddJoin("hasEvent", ConceptFolderStore->GetRecId(ConceptFolderUri), EventRecId, Settings.MaxEventScore);
}

// remove the article from a concept folder
void TConceptFolderBase::RemoveCuratedEventFromConceptFolder(const TStr& ConceptFolderUri, const uint64& EventRecId)
{
	if (!EventStore->IsRecId(EventRecId))
		return;
	if (!ConceptFolderStore->IsRecNm(ConceptFolderUri))
		return;
	ConceptFolderStore->DelJoin("hasEvent", ConceptFolderStore->GetRecId(ConceptFolderUri), EventRecId, Settings.MaxEventScore);
}

// ////////////////////////////////
// helper methods
void TConceptFolderBase::RemoveDuplicates(TArticleSetH& ArticleWgtH)
{
	TTmStopWatch StopWatch(true);
	TArticleSetH OutArticleWgtH;
	for (int i = ArticleWgtH.FFirstKeyId(); ArticleWgtH.FNextKeyId(i);) {
		const uint64 ArticleRecId = ArticleWgtH.GetKey(i);
		if (ArticleStore->GetFieldNmUInt64(ArticleRecId, "isDuplicateOfId") != TUInt64::Mx)
			continue;
		OutArticleWgtH.AddDat(ArticleRecId, ArticleWgtH[i]);
	}
	NotifyVerboseFmt(2, "RemoveDuplicates removed %d of %d articles and took %d ms", ArticleWgtH.Len() - OutArticleWgtH.Len(), ArticleWgtH.Len(), StopWatch.GetMSecInt());
	ArticleWgtH = OutArticleWgtH;
}

// after a folder is deleted add the uri to the list of available uris
void TConceptFolderBase::AddAvailableFolderUri(const TStr& ConceptFolderUri)
{
	AvailableFolderUriV.Add(ConceptFolderUri);
}

// if there an available uri, get it, otherwise return empty
TStr TConceptFolderBase::GetAvailableFolderUri()
{
	if (AvailableFolderUriV.Len() == 0)
		return "";
	TStr Uri = AvailableFolderUriV[AvailableFolderUriV.Len() - 1];
	AvailableFolderUriV.DelLast();
	return Uri;
}

// create a uniqe id for a concept folder. based on existing concept folder store size and folder name
TStr TConceptFolderBase::CreateConceptFolderUri()
{
	TStr Uri = GetAvailableFolderUri();
	if (Uri == "")  {
		uint64 NewRecId = 0;
		if (ConceptFolderStore->GetRecs() > 0)
			NewRecId = NewsBase->ConceptFolderStore->GetLastRecId() + 1;
		Uri = TStr::Fmt("%s-%I64u", NewsConst.ConceptFolderUriPrefix.CStr(), NewRecId);
	}
	return Uri;
}

// return the id of the article that was published APPROXIMATELY X days ago
uint64 TConceptFolderBase::GetEventIdForTime(const int DaysAgo)
{
	//TTm TimeSomeDaysAgo = TTm::GetCurLocTm();
	//TimeSomeDaysAgo.AddDays(-DaysAgo);
	//uint64 RecId = StoryStore->GetRecs();
	//if (RecId == 0) return 0;

	//int RecStep = 10000;
	//while (RecId > RecStep) {		// we have uint so don't try to go below zero
	//	RecId -= RecStep;
	//	if (RecId > 0 && ArticleStore->IsRecId(RecId)) {
	//		TTm ArticleTime; ArticleStore->GetFieldNmTm(RecId, "DateTime", ArticleTime);
	//		if (ArticleTime < TimeSomeDaysAgo)
	//			return RecId;
	//	}
	//}
	return 0;
}

// return the id of the article that was published APPROXIMATELY X days ago
uint64 TConceptFolderBase::GetArticleIdForTime(const int DaysAgo)
{
	TTm TimeSomeDaysAgo = TTm::GetCurLocTm();
	TimeSomeDaysAgo.AddDays(-DaysAgo);
	uint64 RecId = ArticleStore->GetRecs();
	if (RecId == 0) return 0;

	int RecStep = 10000;
	while (RecId > RecStep) {		// we have uint so don't try to go below zero
		RecId -= RecStep;
		if (RecId > 0 && ArticleStore->IsRecId(RecId)) {
			TTm ArticleTime; ArticleStore->GetFieldNmTm(RecId, "DateTime", ArticleTime);
			if (ArticleTime < TimeSomeDaysAgo)
				return RecId;
		}
	}
	return 0;
}


// add/replace id vectors with uri vectors. used when sending the concept folder to clients
void TConceptFolderBase::AddUrisForFolderJson(const PJsonVal& JsonData, const bool& RemoveIds)
{
	if (JsonData->IsObjKey("conceptIdV")) {
		PJsonVal ConceptIdV = JsonData->GetObjKey("conceptIdV");
		PJsonVal ConceptUriV = TJsonVal::NewArr();
		for (int N = 0; N < ConceptIdV->GetArrVals(); N++)  {
			const uint64 ConceptId = (uint64) ConceptIdV->GetArrVal(N)->GetNum();
			EAssertR(ConceptStore->IsRecId(ConceptId), "ERROR: AddUrisForFolderJson: Concept id is invalid");
			ConceptUriV->AddToArr(TJsonVal::NewStr(ConceptStore->GetRecNm(ConceptId)));
		}
		JsonData->AddToObj("conceptUriV", ConceptUriV);
	}
	if (JsonData->IsObjKey("fixedConceptIdV")) {
		PJsonVal ConceptIdV = JsonData->GetObjKey("fixedConceptIdV");
		PJsonVal ConceptUriV = TJsonVal::NewArr();
		for (int N = 0; N < ConceptIdV->GetArrVals(); N++) {
			const uint64 ConceptId = (uint64) ConceptIdV->GetArrVal(N)->GetNum();
			EAssertR(ConceptStore->IsRecId(ConceptId), "ERROR: AddUrisForFolderJson: Concept id is invalid");
			ConceptUriV->AddToArr(TJsonVal::NewStr(ConceptStore->GetRecNm(ConceptId)));
		}
		JsonData->AddToObj("fixedConceptUriV", ConceptUriV);
	}
	if (JsonData->IsObjKey("sourceIdV")) {
		PJsonVal SourceIdV = JsonData->GetObjKey("sourceIdV");
		PJsonVal SourceUriV = TJsonVal::NewArr();
		for (int N = 0; N < SourceIdV->GetArrVals(); N++) {
			const uint64 SourceId = (uint64) SourceIdV->GetArrVal(N)->GetNum();
			EAssertR(NewsBase->SourceStore->IsRecId(SourceId), "ERROR: AddUrisForFolderJson: Source id is invalid");
			SourceUriV->AddToArr(TJsonVal::NewStr(NewsBase->SourceStore->GetRecNm(SourceId)));
		}
		JsonData->AddToObj("sourceUriV", SourceUriV);
	}
	if (JsonData->IsObjKey("categoryIdV")) {
		PJsonVal CategoryIdV = JsonData->GetObjKey("categoryIdV");
		PJsonVal CategoryUriV = TJsonVal::NewArr();
		for (int N = 0; N < CategoryIdV->GetArrVals(); N++) {
			const uint64 CategoryId = (uint64) CategoryIdV->GetArrVal(N)->GetNum();
			EAssertR(NewsBase->CategoryStore->IsRecId(CategoryId), "ERROR: AddUrisForFolderJson: Category id is invalid");
			CategoryUriV->AddToArr(TJsonVal::NewStr(NewsBase->CategoryStore->GetRecNm(CategoryId)));
		}
		JsonData->AddToObj("categoryUriV", CategoryUriV);
	}
	if (JsonData->IsObjKey("locationIdV")) {
		int CityCountryKeyId = IndexVoc->GetKeyId(NewsBase->SourceStoreId, "CityCountry");
		PJsonVal LocationIdV = JsonData->GetObjKey("locationIdV");
		PJsonVal LocationUriV = TJsonVal::NewArr();
		for (int N = 0; N < LocationIdV->GetArrVals(); N++) {
			const uint64 LocationId = (uint64) LocationIdV->GetArrVal(N)->GetNum();
			TStr Loc = IndexVoc->GetWordStr(CityCountryKeyId, LocationId);
			LocationUriV->AddToArr(TJsonVal::NewStr(Loc));
		}
		JsonData->AddToObj("locationUriV", LocationUriV);
	}
	
	if (RemoveIds) {
		JsonData->DelObjKey("conceptIdV");
		JsonData->DelObjKey("fixedConceptIdV");
		JsonData->DelObjKey("sourceIdV");
		JsonData->DelObjKey("categoryIdV");
		JsonData->DelObjKey("locationIdV");
	}
}

// add labels for concepts and sources of the concept folder
void TConceptFolderBase::AddLabelsForFolderJson(const PJsonVal& JsonData)
{
	if (JsonData->IsObjKey("conceptUriV")) {
		PJsonVal ConceptsJson = JsonData->GetObjKey("conceptUriV");
		TStrV LabelV;
		TStrV TypeV;
		for (int N = 0; N < ConceptsJson->GetArrVals(); N++) {
			const TStr ConceptUri = ConceptsJson->GetArrVal(N)->GetStr();
			if (ConceptStore->IsRecNm(ConceptUri)) {
				const TStr Label = NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, "eng");
				const TStr Type = ConceptStore->GetFieldNmStr(ConceptStore->GetRecId(ConceptUri), "Type");
				LabelV.Add(Label);
				TypeV.Add(Type);
			}
		}
		JsonData->AddToObj("conceptLabelV", TJsonVal::NewArr(LabelV));
		JsonData->AddToObj("conceptTypeV", TJsonVal::NewArr(TypeV));
	}

	if (JsonData->IsObjKey("sourceUriV")) {
		PJsonVal SourcesJson = JsonData->GetObjKey("sourceUriV");
		TStrV SourceLabels;
		for (int N = 0; N < SourcesJson->GetArrVals(); N++) {
			const TStr SourceUri = SourcesJson->GetArrVal(N)->GetStr();
			if (NewsBase->SourceStore->IsRecNm(SourceUri))
				SourceLabels.Add(NewsBase->SourceStore->GetFieldNmStr(NewsBase->SourceStore->GetRecId(SourceUri), "Title"));
		}
		JsonData->AddToObj("sourceLabelV", TJsonVal::NewArr(SourceLabels));
	}

	if (JsonData->IsObjKey("categoryUriV")) {
		PJsonVal CategoryJson = JsonData->GetObjKey("categoryUriV");
		TStrV CategoryLabels;
		for (int N = 0; N < CategoryJson->GetArrVals(); N++) {
			const TStr CategoryUri = CategoryJson->GetArrVal(N)->GetStr();
			if (NewsBase->CategoryStore->IsRecNm(CategoryUri))
				CategoryLabels.Add(NewsBase->CategoryStore->GetFieldNmStr(NewsBase->CategoryStore->GetRecId(CategoryUri), "Label"));
		}
		JsonData->AddToObj("categoryLabelV", TJsonVal::NewArr(CategoryLabels));
	}
}

// print some info about the articles in the rec set
void TConceptFolderBase::PrintArticleRecSet(const PRecSet& RecSet, const int& PrintCount)
{
	NotifyVerboseFmt(3, "Printing Article Rec Set. Recs = %d", RecSet->GetRecs());
	for (int RecN = 0; RecN < PrintCount && RecN < RecSet->GetRecs(); RecN++) {
		uint64 RecId = RecSet->GetRecId(RecN);
		if (!ArticleStore->IsRecId(RecId))
			continue;
		NotifyVerboseFmt(3, "RecWgt = %d, RecId = %I64u, title = %s",
			RecSet->GetRecFq(RecN),
			RecSet->GetRecId(RecN),
			ArticleStore->GetFieldNmStr(RecSet->GetRecId(RecN), "Title").CStr());
	}
}

// print article annotations
void TConceptFolderBase::PrintArticleAnnotations(const uint64& ArticleRecId, const int& NumOfConcepts)
{
	PRecSet ConceptSet = ArticleStore->GetRec(ArticleRecId).DoJoin(Base, "hasConcept");
	ConceptSet->SortByFq(false);
	NotifyVerbose(3, "------------------");
	NotifyVerbose(3, "Article annotations:");
	for (int ConceptN = 0; ConceptN < ConceptSet->GetRecs() && ConceptN < NumOfConcepts; ConceptN++) {
		const uint64 ConceptRecId = ConceptSet->GetRecId(ConceptN);
		const int Wgt = ConceptSet->GetRecFq(ConceptN);		// fq is int but we treat is as float
		NotifyVerboseFmt(3, "%d %s", Wgt, ConceptStore->GetRecNm(ConceptRecId).CStr());
	}
}

// print the concepts in the folder
void TConceptFolderBase::PrintFolderDefinition(const PConceptFolder& ConceptFolder, const int& NumOfConcepts)
{
	TUInt64FltKdV ConceptKdV;
	ConceptFolder->GetConceptKdV(ConceptKdV, true);
	
	NotifyVerbose(3, "------------------");
	NotifyVerbose(3, "Folder Definition:");
	NotifyVerboseFmt(3, "Number of concepts: %d", ConceptFolder->ConceptIdV.Len());
	NotifyVerbose(3, "Folder concepts:");
	for (int N = 0; N < ConceptKdV.Len(); N++)
		NotifyVerboseFmt(3, "%.2f %s (fixed=%s)", ConceptKdV[N].Dat, ConceptStore->GetFieldNmStr(ConceptKdV[N].Key, "URI").CStr(), TBool::GetStr(ConceptFolder->IsConceptFixed(ConceptKdV[N].Key)).CStr());
}

THash<TUInt64, TInt> TConceptFolderBase::MergeHashSets(THash<TUInt64, TInt>& H1, THash<TUInt64, TInt>& H2)
{
	THash<TUInt64, TInt> DestH, SrcH;
	if (H1.Len() > H2.Len()) {
		DestH = H1; SrcH = H2;
	}
	else {
		DestH = H2; SrcH = H1;
	}
	for (int i = SrcH.FFirstKeyId(); SrcH.FNextKeyId(i);) {
		TUInt64 Key = SrcH.GetKey(i);
		DestH.AddDat(Key, DestH.IsKey(Key) ? (uint64) (DestH.GetDat(Key) + SrcH.GetDat(Key)) : (uint64) SrcH.GetDat(Key));
	}
	return DestH;
}

// compute a new hash with only keys that occur in H1 and H2
// the value should be H1[key] + H2[key]
THash<TUInt64, TInt> TConceptFolderBase::IntersectHashSets(THash<TUInt64, TInt>& H1, THash<TUInt64, TInt>& H2)
{
	THash<TUInt64, TInt> SmallH, LargeH;
	if (H1.Len() > H2.Len()) {
		LargeH = H1; SmallH = H2;
	}
	else {
		LargeH = H2; SmallH = H1;
	}
	THash<TUInt64, TInt> DestH;
	for (int i = SmallH.FFirstKeyId(); SmallH.FNextKeyId(i);) {
		TUInt64 Key = SmallH.GetKey(i);
		if (LargeH.IsKey(Key))
			DestH.AddDat(Key, SmallH.GetDat(Key) + LargeH.GetDat(Key));
	}
	return DestH;
}


// notifications
void TConceptFolderBase::NotifyVerbose(const int VerbosityLevel, const char *Str)
{
	if (Logger != NULL && VerbosityLevel >= VerbosityLevel)
		Logger->NotifyVerbose(VerbosityLevel, Str);
}

void TConceptFolderBase::NotifyVerboseFmt(const int VerbosityLevel, const char *FmtStr, ...)
{
	if (Logger != NULL && VerbosityLevel >= VerbosityLevel) {
		va_list valist; va_start(valist, FmtStr);
		const int RetVal = vsnprintf(NotifyBuf, 100 * 1024 - 2, FmtStr, valist);
		if (RetVal < 0) return;
		Logger->NotifyVerbose(VerbosityLevel, NotifyBuf);
		va_end(valist);
	}
}