#ifndef CONCEPTFOLDERBASE_H
#define CONCEPTFOLDERBASE_H

#include <base.h>
#include <qminer.h>
#include <logger.h>
#include "NewsConst.h"
#include "ConceptFolder.h"
#include "NewsBase.h"

class TConceptFolderSettings
{
public: 
	// the minimum number of seconds that have to pass before we update articles and events assigned to the concept folder
	int MinSecondsBetweenArticleUpdates;
	int MinSecondsBetweenEventUpdates;

	int MaxArticleScore;
	int MaxEventScore;
	//const int _MaxScoreForValidSource = 150;
	//int _MaxScoreForValidLocation;
	//const int _MaxScoreForSourceProminence = 150;
	int DefSourceWgt;
	int MaxScoreForArticlePopularity;
	double ArtPopReadScore;
	double ArtPopLikeScore;
	double ArtPopStoreScore;
	int VerbosityLevel;

	void LoadSettings(const PJsonVal& Settings);
};

typedef THash<TUInt64, TInt> TArticleSetH;
typedef THash<TUInt64, TInt> TEventSetH;

class TConceptFolderBase;
typedef TPt<TConceptFolderBase> PConceptFolderBase;
class TConceptFolderBase {
public:
	friend class TPt<TConceptFolderBase>;

private:
	TCRef CRef;

	PNewsBase NewsBase;
	// index to base
	PBase Base;
	// pointers to the index and stores
	PIndexVoc IndexVoc;
	PIndex Index;

	PStore ArticleStore;
	PStore ConceptStore;
	PStore ConceptFolderStore;
	PStore StoryStore;
	PStore EventStore;
	PStore UserStore;
	
	TConceptFolderSettings Settings;
	char NotifyBuf[100 * 1024];
	PLogger Logger;
	TStr SettingsFName;

	TStrV AvailableFolderUriV;
	// for loading existing index and stores for update or read-only
	TConceptFolderBase(PNewsBase _NewsBase, const TStr& SettingsFNm);

public:
	
	// for creating new index and stores or updating existing
	static PConceptFolderBase New(PNewsBase NewsBase, const TStr& SettingsFNm) {
		return new TConceptFolderBase(NewsBase, SettingsFNm);
	}
	~TConceptFolderBase();

	// add user
	uint64 AddUser(const TStr& UserUri, const TStr& UserName);

	// concept folder management
	TStr AddConceptFolder(const TStr& UserUri, const PConceptFolder& Folder, const TStr& Label, const TStr& Description);
	bool UpdateConceptFolder(const PConceptFolder& Folder, TStr& ErrorInfo);
	void DeleteConceptFolder(const TStr& ConceptFolderUri);

	// getters
	/*TStr GetConceptFolderUri(const uint64& ConceptId);
	uint64 GetConceptId(const TStr& ConceptFolderUri);*/
	
	void UpdateConceptFolderArticles(const TStr& ConceptFolderUri);  // add to index new articles that belong to the concept folder
	void UpdateConceptFolderEvents(const TStr& ConceptFolderUri);  // add to index new articles that belong to the concept folder

	// article methods
	PRecSet GetSuggestedArticlesForFolder(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, const bool& RemoveDuplicateArticles = false);
	void GetSuggestedArticlesBasedOnConcepts(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void GetArticlesForSources(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void GetArticlesForCategories(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, THash<TUInt64, TInt>& ArticleWgtH);
	void GetArticlesForLocations(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void GetArticlesForKeywords(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);

	// story methods
	PRecSet GetSuggestedEventsForFolder(const PConceptFolder& ConceptFolder, const uint64& MinEventId);
	void GetSuggestedEventsBasedOnConcepts(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH);
	void GetEventsForSources(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH);
	void GetEventsForCategories(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH);
	void GetEventsForKeywords(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH);

	void RemoveDuplicates(TArticleSetH& ArticleWgtH);
	
	void AddCuratedArticleToConceptFolder(const TStr& ConceptFolderUri, const uint64& ArticleRecId);
	void RemoveCuratedArticleFromConceptFolder(const TStr& ConceptFolderUri, const uint64& ArticleRecId);
	void AddCuratedEventToConceptFolder(const TStr& ConceptFolderUri, const uint64& EventRecId);
	void RemoveCuratedEventFromConceptFolder(const TStr& ConceptFolderUri, const uint64& EventRecId);

	// remove all articles and events that have been assigned to the concept folder
	void ClearConceptFolderContent(const PConceptFolder& Folder, const bool& SaveFolderChanges = false);
	
	//// helper functions
	bool IsConceptIdOfConceptFolder(const uint64& ConceptId);			// is concept with id ConceptId a concept folder?
	bool IsFolder(const TStr& ConceptFolderUri);
	bool IsConcept(const TStr& ConceptFolderUri);

	inline uint64 GetConceptId(const TStr& ConceptFolderUri);
	inline uint64 GetFolderId(const TStr& ConceptFolderUri);
	uint64 GetUserId(const TStr& ConceptFolderUri);

	uint64 GetArticleIdForTime(const int DaysAgo);
	uint64 GetEventIdForTime(const int DaysAgo);
	void AddLabelsForFolderJson(const PJsonVal& JsonData);
	void AddUrisForFolderJson(const PJsonVal& JsonData, const bool& RemoveIds = true);
	
	void PrintFolderDefinition(const PConceptFolder& ConceptFolder, const int& NumOfConcepts = 5);
	void PrintArticleAnnotations(const uint64& ArticleRecId, const int& NumOfConcepts = 5);
	void PrintArticleRecSet(const PRecSet& RecSet, const int& PrintCount);

	void NotifyVerbose(const int VerbosityLevel, const char *Str);
	void NotifyVerboseFmt(const int VerbosityLevel, const char *FmtStr, ...);

	void ReloadSettings();
	int GetVerbosityLevel() { return Settings.VerbosityLevel; }

private:
	void Load();
	void Save();

	void AddAvailableFolderUri(const TStr& ConceptFolderUri);
	TStr GetAvailableFolderUri();
	TStr CreateConceptFolderUri();

	void GetSuggestedArticlesForConcepts(const TUInt64FltKdV& HasConceptV, const uint64& MinArticleId, const int& MaxArticleCount, TArticleSetH& ArticleWgtH);
	void GetSuggestedEventsForConcepts(const TUInt64FltKdV& HasConceptV, const uint64& MinEventId, const int& MaxEventCount, TArticleSetH& ArticleWgtH);
	
	void UpdateArticlesBasedOnSources(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void UpdateArticlesBasedOnCategories(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void UpdateArticlesBasedOnLocations(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void UpdateArticlesBasedOnKeywords(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void UpdateArticlesBasedOnSourceProminence(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, TArticleSetH& ArticleWgtH);
	void UpdateArticlesBasedOnArticlePopularity(const PConceptFolder& ConceptFolder, const uint64& MinArticleId, const int& MaxScoreForArticlePopularity, TArticleSetH& ArticleWgtH);

	void UpdateEventsBasedOnSources(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH);
	void UpdateEventsBasedOnCategories(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH);
	void UpdateEventsBasedOnKeywords(const PConceptFolder& ConceptFolder, const uint64& MinEventId, TEventSetH& EventWgtH);
	
	THash<TUInt64, TInt> MergeHashSets(THash<TUInt64, TInt>& H1, THash<TUInt64, TInt>& H2);
	THash<TUInt64, TInt> IntersectHashSets(THash<TUInt64, TInt>& H1, THash<TUInt64, TInt>& H2);
};

#endif