/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ConceptLabel.h"
#include "../NewsMiner/NewsConst.h"

TConceptLabel::TConceptLabel(const TStr& _DbsPath) : DbsPath(_DbsPath)
{
	LabelsChanged = false;
	
	// vector of fallback languages - if label in desired language is not found, we try for each lang in this list first
	FallbackLangV.Add("eng"); 
	FallbackLangV.Add("spa"); 
	FallbackLangV.Add("deu");

	if (TFile::Exists(DbsPath + "ConceptLabel.json") && TFile::Exists(DbsPath + "ConceptLabel.bin") &&
		TFile::GetLastWriteTm(DbsPath + "ConceptLabel.json") > TFile::GetLastWriteTm(DbsPath + "ConceptLabel.bin"))
		LoadFromJson();
	else if (!TFile::Exists(DbsPath + "ConceptLabel.bin"))
		LoadFromJson();
	else
		Load();
}

TConceptLabel::~TConceptLabel()
{
	if (LabelsChanged) {
		SaveToJson();
		Save();
	}
}

void TConceptLabel::Load()
{
	if (TFile::Exists(DbsPath + "ConceptLabel.bin")) {
		TFIn FIn(DbsPath + "ConceptLabel.bin");
		ConceptUriToLangValH.Load(FIn);
	}
}

void TConceptLabel::Save() const
{
	TFOut FOut(DbsPath + "ConceptLabel.bin");
	ConceptUriToLangValH.Save(FOut);
}

bool TConceptLabel::HasConcept(const TStr& ConceptUri) const
{
	return ConceptUriToLangValH.IsKey(ConceptUri);
}

bool TConceptLabel::HasConceptLabel(const TStr& ConceptUri, const TStr& Lang) const
{
	return ConceptUriToLangValH.IsKey(ConceptUri) && ConceptUriToLangValH.GetDat(ConceptUri).IsKey(Lang);
}

void TConceptLabel::SetConceptLabel(const TStr& ConceptUri, const TStr& Lang, const TStr& Label)
{
	// don't set the label if the label for the concept would already be the same
	// this also prevents setting the labels for "Brad Pitt" in X lang if it's already set for eng
	if (GetConceptLabel(ConceptUri, Lang) == Label)
		return;
	LabelsChanged = true;
	ConceptUriToLangValH.AddDat(ConceptUri).AddDat(Lang, Label);
}

TStr TConceptLabel::GetConceptLabel(const TStr& ConceptUri, const TStr& Lang)
{
	if (!ConceptUriToLangValH.IsKey(ConceptUri))
		return "";
	const THash<TStr, TStr> LangToLabelH = ConceptUriToLangValH.GetDat(ConceptUri);
	// we have the concept in the desired language
	if (LangToLabelH.IsKey(Lang))
		return LangToLabelH.GetDat(Lang);
	// we have the concept in one of the fallback languages
	for (int N = 0; N < FallbackLangV.Len(); N++) {
		if (LangToLabelH.IsKey(FallbackLangV[N]))
			return LangToLabelH.GetDat(FallbackLangV[N]);
	}
	// return the label in ANY language
	if (LangToLabelH.Len() > 0) {
		int KeyId = LangToLabelH.FFirstKeyId(); LangToLabelH.FNextKeyId(KeyId);
		return LangToLabelH[KeyId];
	}
	// no label - just return ""
	return "";
}

const THash<TStr, TStr> TConceptLabel::GetConceptLabels(const TStr& ConceptUri) const
{
	if (ConceptUriToLangValH.IsKey(ConceptUri))
		return ConceptUriToLangValH.GetDat(ConceptUri);
	THash<TStr, TStr> EmptyH;
	return EmptyH;
}

void TConceptLabel::RemoveConcept(const TStr& ConceptUri)
{
	ConceptUriToLangValH.DelIfKey(ConceptUri);
}

// load concept label information from json file
void TConceptLabel::LoadFromJson()
{
	if (TFile::Exists(DbsPath + "ConceptLabel.json")) {
		TFIn FIn(DbsPath + "ConceptLabel.json");
		TStr Ln;
		while (FIn.GetNextLn(Ln)) {
			try {
				PJsonVal Json = TJsonVal::GetValFromStr(Ln);
				if (!Json->IsObj() || !Json->IsObjKey("conceptUri"))
					continue;
				const TStr ConceptUri = Json->IsObjKey("conceptUri");
				THash<TStr, TStr> LangToLabelH;
				int Keys = Json->GetObjKeys();
				for (int N = 0; N < Keys; N++) {
					const TStr Key = Json->GetObjKey(N);
					if (Key == "conceptUri")
						continue;
					const TStr Val = Json->GetObjStr(Key);
					LangToLabelH.AddDat(Key, Val);
				}
				if (LangToLabelH.Len() > 0)
					ConceptUriToLangValH.AddDat(ConceptUri, LangToLabelH);
			}
			catch (PExcept Ex) {
				TNotify::StdNotify->OnStatusFmt("TConceptLabel: Exception while loading concepts from ConceptLabel.json. Error: %s", Ex->GetMsgStr().CStr());
			}
		}
	}
}

// save concept label to json file
void TConceptLabel::SaveToJson() const
{
	// save json file for concepts
	TFOut FOut(DbsPath + "ConceptLabel.json");
	for (int KeyId = ConceptUriToLangValH.FFirstKeyId(); ConceptUriToLangValH.FNextKeyId(KeyId); ) {
		const TStr ConceptUri = ConceptUriToLangValH.GetKey(KeyId);
		const THash<TStr, TStr>& LangToLabelH = ConceptUriToLangValH[KeyId];
		PJsonVal Json = TJsonVal::NewObj("conceptUri", ConceptUri);
		for (int LangId = LangToLabelH.FFirstKeyId(); LangToLabelH.FNextKeyId(LangId);) {
			const TStr Lang = LangToLabelH.GetKey(LangId);
			const TStr Label = LangToLabelH[LangId];
			Json->AddToObj(Lang, Label);
		}
		const TStr JsonStr = TJsonVal::GetStrFromVal(Json);
		FOut.PutStr(JsonStr + "\n");
	}
}