/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "GeoBase.h"
//#include "../Utils/FileUtils.h"
//#include "ConceptFix.h"
#include "NewsConst.h"
#include "NewsBase.h"

const int GeoCacheSizeMB = 300;

// create new geo base
TGeoBase::TGeoBase(const TStr& _FPath, const TStr& StoreDefFNm, const PConceptLabel _ConceptLabel) : ConceptLabel(_ConceptLabel)
{
	FPath = _FPath;

	// initialize store
	PJsonVal StoreDefVal = TJsonVal::GetValFromStr(TStr::LoadTxt(StoreDefFNm));
	Base = NewBase(FPath, StoreDefVal, GeoCacheSizeMB*TInt::Mega, GeoCacheSizeMB*TInt::Mega, TStrUInt64H(), false);

	Init();
	// finish initalization
	Base->Init();
	SaveBase(Base);
}

// load geo base
TGeoBase::TGeoBase(const TStr& _FPath, const TFAccess& FAccess, const PConceptLabel _ConceptLabel) : ConceptLabel(_ConceptLabel)
{
	FPath = _FPath;
	Base = LoadBase(FPath, FAccess, GeoCacheSizeMB*TInt::Mega, GeoCacheSizeMB*TInt::Mega, TStrUInt64H(), false);
	
	Init();
	// finish initalization
	Base->Init();

	if (TFile::Exists(FPath + "GeoBase.dat")) {
		TFIn FIn(FPath + "GeoBase.dat");
		CityWgtH.Load(FIn);
		CountryWgtH.Load(FIn);
	}
}

void TGeoBase::Init()
{
	// prepare shortcuts
	IndexVoc = Base->GetIndexVoc();
	Index = Base->GetIndex();

	EAssert(Base->IsStoreNm("Country"));
    CountryStore = Base->GetStoreByStoreNm("Country");
	CountryStoreId = CountryStore->GetStoreId();

	EAssert(Base->IsStoreNm("City"));
    CityStore = Base->GetStoreByStoreNm("City");
	CityStoreId = CityStore->GetStoreId();

	EAssert(Base->IsStoreNm("Language"));
    LanguageStore = Base->GetStoreByStoreNm("Language");
	LanguageStoreId = LanguageStore->GetStoreId();

	CountryWikiKeyId = CountryStore->GetFieldKeyId(CountryStore->GetFieldId("WikiUri"));
	CityWikiKeyId = CityStore->GetFieldKeyId(CityStore->GetFieldId("WikiUri"));
}

TGeoBase::~TGeoBase()
{
	// save the concept weights
	const TStr File = FPath + "GeoBase.dat";
	TFOut FOut(File);
	CityWgtH.Save(FOut);
	CountryWgtH.Save(FOut);

	// save the autosuggest data
	const TStr FileSugg = FPath + "AutoSuggestCache.dat";
	TFOut FOutSugg(FileSugg);
	SortByPopularity.Save(FOutSugg);
	NextLangInd.Save(FOutSugg);
	CountryData.Save(FOutSugg);
	CityData.Save(FOutSugg);
	CityPrefixToIdWgtHH.Save(FOutSugg);
	CountryPrefixToIdWgtHH.Save(FOutSugg);
}

// import data from json files into the city and country stores
void TGeoBase::ImportData(const PNewsFixDb& NewsFixDb, const TStr& CountryInfoJsonFNm, const TStr& CityInfoJsonFNm)
{
	TTmStopWatch Timer; Timer.Start();
	TQm::TEnv::Logger->OnStatus("Importing the geo info data...");
	if (TFile::Exists(CountryInfoJsonFNm)) {
		TFIn CountryJsonF(CountryInfoJsonFNm);
		TStr CountryJsonStr;
		uchar CountryStoreId = CountryStore->GetStoreId();
		TVec<TPair<TStr, TStr> > HasNeighborV;
		while (CountryJsonF.GetNextLn(CountryJsonStr)) {
			PJsonVal CountryJson = TJsonVal::GetValFromStr(CountryJsonStr);
			// take into account desired concept prefix + potential concept rewrites
			TStr WikiUri = CountryJson->GetObjStr("WikiUri", "");
			if (WikiUri != "") {
				// find the correct wiki url
				if (NewsFixDb->GetConceptRewrite(WikiUri))
					TNotify::StdNotify->OnStatusFmt("Corrected country url to %s", WikiUri.CStr());
				CountryJson->AddToObj("WikiUri", WikiUri);

				// set the labels if not set yet
				for (int N = 0; N < CountryJson->GetObjKeys(); N++) {
					const TStr Key = CountryJson->GetObjKey(N);
					if (Key.IsPrefix("Label")) {
						const TStr Val = CountryJson->GetObjStr(Key);
						const TStr Lang = Key.GetSubStr(5).GetLc();
						if (!ConceptLabel->HasConceptLabel(WikiUri, Lang))
							ConceptLabel->SetConceptLabel(WikiUri, Lang, Val);
					}
				}
			}
			
			if (CountryJson->IsObjKey("hasNeighbor")) {
				TStr Geo1 = CountryJson->GetObjStr("GeoNameId", "");
				PJsonVal Arr = CountryJson->GetObjKey("hasNeighbor");
				for (int N=0; N < Arr->GetArrVals(); N++) {
					TStr Geo2 = Arr->GetArrVal(N)->GetObjStr("GeoNameId", "");
					if (Geo1 < Geo2)
						HasNeighborV.Add(TPair<TStr, TStr> (Geo1, Geo2));
				}
			}
			
			// we need to remove the join from the json otherwise we get error "missing Name"
			CountryJson->DelObjKey("hasNeighbor");
			Base->AddRec(CountryStoreId, CountryJson);
		}
		// create neighbor joins
		for (int N=0; N < HasNeighborV.Len(); N++) {
			TStr Geo1 = HasNeighborV[N].Val1; TStr Geo2 = HasNeighborV[N].Val2;
			CountryStore->AddJoin("hasNeighbor", CountryStore->GetRecId(Geo1), CountryStore->GetRecId(Geo2));
		}
	}
	else {
		TNotify::StdNotify->OnStatusFmt("The file %s with country info was not found", CountryInfoJsonFNm.CStr());
	}

	if (TFile::Exists(CityInfoJsonFNm)) {
		TFIn CityJsonF(CityInfoJsonFNm);
		TStr CityJsonStr;
		uchar CityStoreId = CityStore->GetStoreId();
		while (CityJsonF.GetNextLn(CityJsonStr)) {
			PJsonVal CityJson = TJsonVal::GetValFromStr(CityJsonStr);
			
			// take into account desired concept prefix + potential concept rewrites
			TStr WikiUri = CityJson->GetObjStr("WikiUri", "");
			if (WikiUri != "") {
				// find the correct wiki url
				if (NewsFixDb->GetConceptRewrite(WikiUri))
					TNotify::StdNotify->OnStatusFmt("Corrected city url to %s", WikiUri.CStr());
				// prepend the desired prefix
				CityJson->AddToObj("WikiUri", WikiUri);

				// set the labels if not set yet
				for (int N = 0; N < CityJson->GetObjKeys(); N++) {
					const TStr Key = CityJson->GetObjKey(N);
					if (Key.IsPrefix("Label")) {
						const TStr Val = CityJson->GetObjStr(Key);
						const TStr Lang = Key.GetSubStr(5).GetLc();
						if (!ConceptLabel->HasConceptLabel(WikiUri, Lang))
							ConceptLabel->SetConceptLabel(WikiUri, Lang, Val);
					}
				}
			}
						
			Base->AddRec(CityStoreId, CityJson);
		}
	}
	else {
		TNotify::StdNotify->OnStatusFmt("The file %s with city info was not found", CityInfoJsonFNm.CStr());
	}
	Timer.Stop();
	TNotify::StdNotify->OnStatusFmt("Done. %.0f ms needed.", Timer.GetMSec());
}

bool TGeoBase::LoadPrefixes(const bool& _SortByPopularity)
{
	try {
		if (TFile::Exists(FPath + "AutoSuggestCache.dat")) {
			TFIn FIn(FPath + "AutoSuggestCache.dat");
			SortByPopularity.Load(FIn);
			if (_SortByPopularity != SortByPopularity) 
				return false;
			NextLangInd.Load(FIn);
			CountryData.Load(FIn);
			CityData.Load(FIn);
			CityPrefixToIdWgtHH.Load(FIn);
			CountryPrefixToIdWgtHH.Load(FIn);
			return true;
		}
	}
	catch (PExcept E) {
		TNotify::StdNotify->OnStatusFmt("TGeoBase::LoadPrefixes: %s", E->GetMsgStr().CStr());
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("TGeoBase::LoadPrefixes: General exception");
	}
	return false;
}

void TGeoBase::LoadOrUpdatePrefixes(const bool& _SortByPopularity)
{
	bool LoadedOk = LoadPrefixes(_SortByPopularity);
	if (LoadedOk)
		return;

	// the first time we want to update prefixes for all languages
	SortByPopularity = _SortByPopularity;
	UpdateDataCounts();
	TTmStopWatch StopWatch(true);
	TNotify::StdNotify->OnStatus("Updating geographic name suggestions for all languages...");
	for (int N = 0; N < NewsConst.LangAllV.Len(); N++)
		UpdatePrefixes(false);
	const double Tm = StopWatch.GetSec();
	TNotify::StdNotify->OnStatusFmt("Update took: %.2fsec", Tm);
}

void TGeoBase::SetSorting(const bool& _SortByPopularity)
{
	bool Changed = SortByPopularity != _SortByPopularity;
	SortByPopularity = _SortByPopularity;
	if (Changed)
		UpdateDataCounts();
}

void TGeoBase::UpdateDataCounts()
{
	CountryData.Clr();
	if (SortByPopularity && CountryWgtH.Len() > 0) {
		CountryData = THash<TUInt64, TInt>(CountryWgtH);
		// for the cities that don't have any results add 0 - otherwise we don't see them in the autosuggests
		for (uint64 N = 0; N < CountryStore->GetRecs(); N++)
			CountryData.AddDat(N);
	}
	else {
		for (uint64 N = 0; N < CountryStore->GetRecs(); N++)
			CountryData.AddDat(N, CountryStore->GetFieldNmInt(N, "Population"));
	}
	CountryData.SortByDat(false);

	CityData.Clr();
	if (SortByPopularity) {
		CityData = THash<TUInt64, TInt>(CityWgtH);
		// for the cities that don't have any results add 0 - otherwise we don't see them in the autosuggests
		for (uint64 N = 0; N < CityStore->GetRecs(); N++)
			CityData.AddDat(N);
	}
	else {
		for (uint64 N = 0; N < CityStore->GetRecs(); N++)
			CityData.AddDat(N, CityStore->GetFieldNmInt(N, "Population"));
	}
	CityData.SortByDat(false);
}

// update prefixes (for a single language at a time) that are used for autosuggest
void TGeoBase::UpdatePrefixes(const bool& Report)
{
	if (NextLangInd >= NewsConst.LangAllV.Len()) {
		UpdateDataCounts();
		NextLangInd++;
		return;
	}

	if (Report)
		TNotify::StdNotify->OnStatus("Updating suggestions for geographical concepts...");

	TTmStopWatch StopWatch(true);
	
	const TStr Lang = NewsConst.LangAllV[NextLangInd];
	NextLangInd = (NextLangInd + 1) % (NewsConst.LangAllV.Len()+1);
	CountryPrefixToIdWgtHH.AddDat(Lang, PrefToSugg());		// add empty set - needed for next updates for the same language (otherwise we would just increase counts)
		
	for (int CountryKeyId = CountryData.FFirstKeyId(); CountryData.FNextKeyId(CountryKeyId);) {
		const uint64 CountryRecId = CountryData.GetKey(CountryKeyId);
		const int Val = CountryData[CountryKeyId];
		const TStr CountryUri = CountryStore->GetFieldNmStr(CountryRecId, "WikiUri");
		
		TStr Name = ConceptLabel->GetConceptLabel(CountryUri, Lang);
		Name = Name.GetLc();
		Name.ChangeCh('.', ' ');
		Name.ChangeCh(',', ' ');
			
		PrefToSugg& CountryPrefixToIdWgtH = CountryPrefixToIdWgtHH.AddDat(Lang);
		TStrV WordParts; Name.SplitOnAllCh(' ', WordParts);
		for (int Parts = 0; Parts < WordParts.Len(); Parts++) {
			TStr Sub1 = WordParts[Parts].GetSubStr(0, 0);
			TStr Sub2 = WordParts[Parts].GetSubStr(0, 1);
			CountryPrefixToIdWgtH.AddDat(Sub1).Add(TUInt64IntKd(CountryRecId, Val));
			if (Sub2.Len() == 2)
				CountryPrefixToIdWgtH.AddDat(Sub2).Add(TUInt64IntKd(CountryRecId, Val));
		}
	}
		
	for (int CityKeyId = CityData.FFirstKeyId(); CityData.FNextKeyId(CityKeyId);) {
		const uint64 CityRecId = CityData.GetKey(CityKeyId);
		const int Val = CityData[CityKeyId];
		const TStr CityUri = CityStore->GetFieldNmStr(CityRecId, "WikiUri");

		TStr Name = ConceptLabel->GetConceptLabel(CityUri, Lang);
		Name = Name.GetLc();
		Name.ChangeCh('.', ' ');
		Name.ChangeCh(',', ' ');

		PrefToSugg& CityPrefixToIdWgtH = CityPrefixToIdWgtHH.AddDat(Lang);
		TStrV WordParts; Name.SplitOnAllCh(' ', WordParts);
		for (int Parts = 0; Parts < WordParts.Len(); Parts++) {
			TStr Sub1 = WordParts[Parts].GetSubStr(0, 0);
			TStr Sub2 = WordParts[Parts].GetSubStr(0, 1);
			CityPrefixToIdWgtH.AddDat(Sub1).Add(TUInt64IntKd(CityRecId, Val));
			if (Sub2.Len() == 2)
				CityPrefixToIdWgtH.AddDat(Sub2).Add(TUInt64IntKd(CityRecId, Val));
		}
	}
		
	if (Report)
		TNotify::StdNotify->OnStatusFmt("Update country & city prefixes took: %.2fms", StopWatch.GetMSec());
}

void TGeoBase::SuggestCountries(const TStr& Prefix, const TStr& Lang, const int& MaxCount, TUInt64IntKdV& Suggestions)
{
	Suggestions.Clr();
	
	int AddedCount = 0;
	if (CountryPrefixToIdWgtHH.IsKey(Lang)) {
		PrefToSugg& PrefixToIdWgtH = CountryPrefixToIdWgtHH.GetDat(Lang);
		TStr PrefixLc = Prefix.GetLc();
		TStr PrefixKey = PrefixLc.GetSubStr(0, 1);
		if (PrefixToIdWgtH.IsKey(PrefixKey)) {
			TVec<TUInt64IntKd>& SuggV = PrefixToIdWgtH.GetDat(PrefixKey);
			int Count = SuggV.Len();
			for (int N=0; N < Count; N++) {
				const uint64 CountryRecId = SuggV[N].Key;
				const TStr CountryUri = CountryStore->GetFieldNmStr(CountryRecId, "WikiUri");
				TStr Label = ConceptLabel->GetConceptLabel(CountryUri, Lang);
				TStr LabelLc = Label.GetLc();
				if (!LabelLc.IsPrefix(PrefixLc) && LabelLc.SearchStr(" " + PrefixLc, 0) == -1)
					continue;
				if (Suggestions.IsIn(SuggV[N]))
					continue;
				Suggestions.Add(SuggV[N]);
				AddedCount++;
				if (AddedCount >= MaxCount)
					break;
			}
		}
	}
}

void TGeoBase::SuggestCities(const TStr& Prefix, const TStr& Lang, const int& MaxCount, TUInt64IntKdV& Suggestions)
{
	Suggestions.Clr();
	
	int AddedCount = 0;
	if (CityPrefixToIdWgtHH.IsKey(Lang)) {
		PrefToSugg& PrefixToIdWgtH = CityPrefixToIdWgtHH.GetDat(Lang);
		TStr PrefixLc = Prefix.GetLc();
		TStr PrefixKey = PrefixLc.GetSubStr(0, 1);
		if (PrefixToIdWgtH.IsKey(PrefixKey)) {
			TVec<TUInt64IntKd>& SuggV = PrefixToIdWgtH.GetDat(PrefixKey);
			int Count = SuggV.Len();
			for (int N=0; N < Count; N++) {
				const uint64 CityRecId = SuggV[N].Key;
				const TStr CityUri = CityStore->GetFieldNmStr(CityRecId, "WikiUri");
				TStr Label = ConceptLabel->GetConceptLabel(CityUri, Lang);
				TStr LabelLc = Label.GetLc();
				if (!LabelLc.IsPrefix(PrefixLc) && LabelLc.SearchStr(" " + PrefixLc, 0) == -1)
					continue;
				if (Suggestions.IsIn(SuggV[N]))
					continue;
				Suggestions.Add(SuggV[N]);
				AddedCount++;
				if (AddedCount >= MaxCount)
					break;
			}
		}
	}
}

TStr TGeoBase::GetCityForGeoId(const TStr& GeoId) const
{
	if (! CityStore->IsRecNm(GeoId))
		return "";
	uint64 RecId = CityStore->GetRecId(GeoId);
	return CityStore->GetFieldNmStr(RecId, "WikiUri");
}

TStr TGeoBase::GetCountryForGeoId(const TStr& GeoId) const
{
	if (! CountryStore->IsRecNm(GeoId))
		return "";
	uint64 RecId = CountryStore->GetRecId(GeoId);
	return CountryStore->GetFieldNmStr(RecId, "WikiUri");
}

TStr TGeoBase::GetCountryForCity(const TStr& WikiUri) const
{
	TStr CityGeoId = GetGeoIdForCity(WikiUri);
	if (CityGeoId == "")
		return "";
	TRec CountryRec = CityStore->GetRec(CityGeoId).DoSingleJoin(Base, "hasCountry");
	if (CountryRec.IsDef())
		return CountryStore->GetFieldNmStr(CountryRec.GetRecId(), "WikiUri");
	return "";	
}

TStr TGeoBase::GetGeoIdForCity(const TStr& WikiUri) const
{
	if (! IsCity(WikiUri)) 
		return "";
	uint64 WId = IndexVoc->GetWordId(CityWikiKeyId, WikiUri);
	TQueryItem Query = TQueryItem(Base, CityWikiKeyId, WId, oqctEqual);
	PRecSet Items = Base->Search(TQuery::New(Base, Query));
	if (Items->GetRecs() == 1)
		return CityStore->GetRecNm(Items->GetRecId(0));
	else if (Items->GetRecs() > 1)
		TNotify::StdNotify->OnStatusFmt("Unexpectedly, there are multiple records having the same wiki uri %s", WikiUri.CStr());
	return "";
}

TStr TGeoBase::GetGeoIdForCountry(const TStr& WikiUri) const
{
	if (! IsCountry(WikiUri)) 
		return "";
	uint64 WId = IndexVoc->GetWordId(CountryWikiKeyId, WikiUri);
	TQueryItem Query = TQueryItem(Base, CountryWikiKeyId, WId, oqctEqual);
	PRecSet Items = Base->Search(TQuery::New(Base, Query));
	if (Items->GetRecs() == 1)
		return CountryStore->GetRecNm(Items->GetRecId(0));
	else if (Items->GetRecs() > 1)
		TNotify::StdNotify->OnStatusFmt("Unexpectedly, there are multiple records having the same wiki uri %s", WikiUri.CStr());
	return "";
}

bool TGeoBase::GetCitiesForCountry(const TStr& WikiUri, TStrV& CityWikiUriV) const
{
	if (!IsCountry(WikiUri))
		return false;
	CityWikiUriV.Clr();
	TStr CountryGeoId = GetGeoIdForCountry(WikiUri);
	PRecSet CitySet = CountryStore->GetRec(CountryGeoId).DoJoin(Base, "hasCity");
	for (int N = 0; N < CitySet->GetRecs(); N++) {
		TStr CityWikiUri = CityStore->GetFieldNmStr(CitySet->GetRecId(N), "WikiUri");
		if (CityWikiUri != "")
			CityWikiUriV.Add(CityWikiUri);
	}
	return true;
}

bool TGeoBase::IsCity(const TStr& WikiUri) const
{
	return IndexVoc->IsWordStr(CityWikiKeyId, WikiUri);
}

bool TGeoBase::IsCountry(const TStr& WikiUri) const
{
	return IndexVoc->IsWordStr(CountryWikiKeyId, WikiUri);
}

int TGeoBase::GetWgt(const TStr& WikiUri)
{
	if (IsCity(WikiUri)) {
		const TStr CityNm = GetGeoIdForCity(WikiUri);
		if (!CityStore->IsRecNm(CityNm))
			return 0;
		const uint64 CityId = CityStore->GetRecId(CityNm);
		return CityWgtH.GetDatOrDef(CityId, 0);
	}
	else if (IsCountry(WikiUri)) {
		TStr CountryNm = GetGeoIdForCountry(WikiUri);
		if (!CountryStore->IsRecNm(CountryNm))
			return 0;
		uint64 CountryId = CountryStore->GetRecId(CountryNm);
		return CountryWgtH.GetDatOrDef(CountryId, 0);
	}
	else
		return 0;
}

bool TGeoBase::IncreaseWgt(const TStr& WikiUri, const int Wgt)
{
	if (IsCity(WikiUri)) {
		const TStr CityNm = GetGeoIdForCity(WikiUri);
		if (!CityStore->IsRecNm(CityNm))
			return false;
		const uint64 CityId = CityStore->GetRecId(CityNm);
		CityWgtH.AddDat(CityId) += Wgt;
		return true;
	}
	else if (IsCountry(WikiUri)) {
		const TStr CountryNm = GetGeoIdForCountry(WikiUri);
		if (!CountryStore->IsRecNm(CountryNm))
			return false;
		const uint64 CountryId = CountryStore->GetRecId(CountryNm);
		CountryWgtH.AddDat(CountryId) += Wgt;
		return true;
	}
	else 
		return false;
}

bool TGeoBase::DecreaseWgt(const TStr& WikiUri, const int Wgt)
{
	if (IsCity(WikiUri)) {
		const TStr CityNm = GetGeoIdForCity(WikiUri);
		if (!CityStore->IsRecNm(CityNm))
			return false;
		const uint64 CityId = CityStore->GetRecId(CityNm);
		CityWgtH.AddDat(CityId, MAX(0, CityWgtH.IsKey(CityId) ? CityWgtH.GetDat(CityId) - Wgt : -Wgt));
		return true;
	}
	else if (IsCountry(WikiUri)) {
		const TStr CountryNm = GetGeoIdForCountry(WikiUri);
		if (!CountryStore->IsRecNm(CountryNm))
			return false;
		const uint64 CountryId = CountryStore->GetRecId(CountryNm);
		CountryWgtH.AddDat(CountryId, MAX(0, CountryWgtH.IsKey(CountryId) ? CountryWgtH.GetDat(CountryId) - Wgt : -Wgt));
		return true;
	}
	else
		return false;
}

PJsonVal TGeoBase::GetInfoForCountry(const TStr& WikiUri, const TStrV& LabelLanguageV, const bool& IncludeUri, const bool& IncludeLocation, const bool& IncludeDetailedInfo)
{
	const TStr GeoId = GetGeoIdForCountry(WikiUri);
	PJsonVal LocJson = TJsonVal::NewObj();
	LocJson->AddToObj("wikiUri", WikiUri);

	if (!CountryStore->IsRecNm(GeoId)) {
		TNotify::StdNotify->OnStatusFmt("Country '%s' is not recognized as a known country. Unable to provide country info.", WikiUri.CStr());
		return LocJson;
	}
	const uint64 RecId = CountryStore->GetRecId(GeoId);

	// add also labels for the concept in other languages
	for (int N = 0; N < LabelLanguageV.Len(); N++) {
		TStr Lang = LabelLanguageV[N] != "" ? LabelLanguageV[N] : "eng";	// geo stores don't have the Label field so we map "" -> lang
		const TStr Label = ConceptLabel->GetConceptLabel(WikiUri, Lang);
		LocJson->AddToObj(NewsConst.GetJsonNmForLang(LabelLanguageV[N]), Label);
	}

	if (IncludeUri)
		LocJson->AddToObj("wikiUri", WikiUri);

	if (IncludeLocation && ! CountryStore->IsFieldNmNull(RecId, "Location")) {
		const TFltPr& LonLat = CountryStore->GetFieldNmFltPr(RecId, "Location");
		LocJson->AddToObj("lat", LonLat.Val1);
		LocJson->AddToObj("long", LonLat.Val2);
	}
	
	if (IncludeDetailedInfo) {
		LocJson->AddToObj("area", CountryStore->GetFieldNmInt(RecId, "Area"));
		LocJson->AddToObj("population", CountryStore->GetFieldNmInt(RecId, "Population"));
		LocJson->AddToObj("code2", CountryStore->GetFieldNmInt(RecId, "Code2"));
		LocJson->AddToObj("code3", CountryStore->GetFieldNmInt(RecId, "Code3"));
		LocJson->AddToObj("webExt", CountryStore->GetFieldNmInt(RecId, "WebExt"));
		LocJson->AddToObj("currencyName", CountryStore->GetFieldNmInt(RecId, "CurrencyName"));
		LocJson->AddToObj("continent", CountryStore->GetFieldNmInt(RecId, "Continent"));
		PRecSet LangRecSet = CountryStore->GetRec(RecId).DoJoin(Base, "hasLanguage");
		PJsonVal LangArr = TJsonVal::NewArr();
		for (int N=0; N < LangRecSet->GetRecs(); N++) {
			LangArr->AddToArr(TJsonVal::NewStr(LanguageStore->GetFieldNmStr(LangRecSet->GetRecId(N), "Name")));
		}
		LocJson->AddToObj("languages", LangArr);
	}
	return LocJson;
}
PJsonVal TGeoBase::GetInfoForCity(const TStr& WikiUri, const TStrV& LabelLanguageV, const bool& IncludeUri, const bool& IncludeLocation, const bool& IncludePopulation, const bool& IncludeCountryInfo)
{
	const TStr GeoId = GetGeoIdForCity(WikiUri);
	PJsonVal LocJson = TJsonVal::NewObj();
	if (!CityStore->IsRecNm(GeoId)) {
		TNotify::StdNotify->OnStatusFmt("City '%' is not recognized as a known city uri. Unable to provide city info.", WikiUri.CStr());
		return LocJson;
	}
	const uint64 RecId = CityStore->GetRecId(GeoId);

	// add also labels for the concept in other languages
	for (int N = 0; N < LabelLanguageV.Len(); N++) {
		const TStr Lang = LabelLanguageV[N] != "" ? LabelLanguageV[N] : "eng";	// geo stores don't have the Label field so we map "" -> lang
		const TStr Label = ConceptLabel->GetConceptLabel(WikiUri, Lang);
		LocJson->AddToObj(NewsConst.GetJsonNmForLang(LabelLanguageV[N]), Label);
	}

	if (IncludeUri)
		LocJson->AddToObj("wikiUri", WikiUri);

	if (IncludePopulation)
		LocJson->AddToObj("population", CityStore->GetFieldNmInt(RecId, "Population"));

	if (IncludeLocation && ! CityStore->IsFieldNmNull(RecId, "Location")) {
		const TFltPr& LonLat = CityStore->GetFieldNmFltPr(RecId, "Location");
		LocJson->AddToObj("lat", LonLat.Val1);
		LocJson->AddToObj("long", LonLat.Val2);
	}
	// add also country info
	if (IncludeCountryInfo) {
		const TRec CountryRec = CityStore->GetRec(GeoId).DoSingleJoin(Base, "hasCountry");
		if (CountryRec.IsDef()) {
			const TStr CountryWikiUri = CountryStore->GetFieldNmStr(CountryRec.GetRecId(), "WikiUri");
			LocJson->AddToObj("country", GetInfoForCountry(CountryWikiUri, LabelLanguageV, IncludeUri, IncludeLocation));
		}
	}
	return LocJson;
}

void TGeoBase::ImportGeoConcepts(const TNewsBase* NewsBase)
{
	TVec<PStore> StoreV; StoreV.Add(CountryStore); StoreV.Add(CityStore);
	for (int StoreN = 0; StoreN < StoreV.Len(); StoreN++) {
		const PStore Store = StoreV[StoreN];
		const uint64 Count = Store->GetRecs();
		for (uint64 N = 0; N < Count; N++) {
			const TStr Uri = Store->GetFieldNmStr(N, "WikiUri");
			if (Uri != "" && !NewsBase->ConceptStore->IsRecNm(Uri)) {
				PJsonVal Json = TJsonVal::NewObj("URI", Uri);
				Json->AddToObj("Type", NewsConst.TypeLocation);
				NewsBase->ConceptStore->AddRec(Json);
			}
		}
	}
}

// ///////////
// AUTO SUGGEST SERVICE

// suggest countries and/or cities for given prefix in desired languages
// epected params: prefix, lang+, source?
TStr TGeoAutoSuggestSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv) 
{
	const TStr Action = GetFldVal(FldNmValPrV, "action", "");
	if (Action == "updatePrefixes") {
		const TStr SortBy = GetFldVal(FldNmValPrV, "sortBy", "popularity");		// popularity or population
		GeoBase->SetSorting(SortBy == "popularity");
		GeoBase->UpdatePrefixes();
		return TJsonVal::GetStrFromVal(TJsonVal::NewObj("info", TStr("Locations updated.")));
	}

	const TStr PrefixStr = GetFldVal(FldNmValPrV, "prefix");		
	const TStr PrefixStrLc = PrefixStr.GetLc();
	
	PJsonVal SuggJson = TJsonVal::NewObj();
	if (PrefixStrLc.Len() == 0)
		return TJsonVal::GetStrFromVal(SuggJson);
	
	// get languages to provide suggestions
	const TStr Lang = GetFldVal(FldNmValPrV, "lang", "eng");
	
	// country, city or both
	TStrV SourceV; GetFldValV(FldNmValPrV, "source", SourceV);
	if (SourceV.Len() == 0) {
		SourceV.Add("city"); SourceV.Add("country");
	}
	TInt Count = GetFldInt(FldNmValPrV, "count", 20);
	bool IgnoreNonWiki = GetFldVal(FldNmValPrV, "ignoreNonWiki", "true").GetLc() == "true";

	PJsonVal Arr = TJsonVal::NewArr();

	if (SourceV.IsIn("country")) {
		//TVec<TUInt64IntKdV> SuggestionsV;
		
		//EAssertR(SuggestionsV.Len() == LangV.Len(), "The country suggestions list did not contain the requested number of languages");
		PJsonVal CountryJsonObj = TJsonVal::NewObj();
		TUInt64IntKdV Suggestions;
		GeoBase->SuggestCountries(PrefixStr, Lang, 2*Count, Suggestions);
			
		int Added = 0;
		for (int SuggN = 0; SuggN < Suggestions.Len() && Added < Count; SuggN++) {
			TUInt64& CountryRecId = Suggestions[SuggN].Key;
			TStr WikiUri = GeoBase->CountryStore->GetFieldNmStr(CountryRecId, "WikiUri");
			if (WikiUri == "" && IgnoreNonWiki)
				continue;
			TInt& Population = Suggestions[SuggN].Dat;
			PJsonVal Item = TJsonVal::NewObj("pop", Population);
			Item->AddToObj("id", (int) CountryRecId);
			Item->AddToObj("uri", GeoBase->CountryStore->GetRecNm(CountryRecId));
			Item->AddToObj("wikiUri", WikiUri);
			Item->AddToObj("label", GeoBase->ConceptLabel->GetConceptLabel(WikiUri, Lang));
			Item->AddToObj("type", TStr("country"));
			Arr->AddToArr(Item);
			Added++;
		}
	}

	if (SourceV.IsIn("city")) {
		TUInt64IntKdV Suggestions;
		GeoBase->SuggestCities(PrefixStr, Lang, 2*Count, Suggestions);
			
		int Added = 0;
		for (int SuggN=0; SuggN < Suggestions.Len() && Added < Count; SuggN++) {
			TUInt64& CityRecId = Suggestions[SuggN].Key;
			TStr WikiUri = GeoBase->CityStore->GetFieldNmStr(CityRecId, "WikiUri");
			if (WikiUri == "" && IgnoreNonWiki)
				continue;
			TInt& Population = Suggestions[SuggN].Dat;
			PJsonVal Item = TJsonVal::NewObj("pop", (int) Population);
			Item->AddToObj("id", (int) CityRecId);
			Item->AddToObj("uri",  GeoBase->CityStore->GetRecNm(CityRecId));
			Item->AddToObj("wikiUri",  WikiUri);
			TRec CountryRec = GeoBase->CityStore->GetRec(CityRecId).DoSingleJoin(GeoBase->Base, "hasCountry");
			TStr Label = GeoBase->ConceptLabel->GetConceptLabel(WikiUri, Lang);
			if (CountryRec.IsDef()) {
				const TStr CountryWikiUri = GeoBase->CountryStore->GetFieldNmStr(CountryRec.GetRecId(), "WikiUri");
				Label += ", " + GeoBase->ConceptLabel->GetConceptLabel(CountryWikiUri, Lang);
			}
			Item->AddToObj("label", Label);
			Item->AddToObj("type", TStr("city"));
			Arr->AddToArr(Item);
			Added++;
		}
	}
	
	return TJsonVal::GetStrFromVal(Arr);
}

