/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "NewsArticle.h"
#include "../Utils/DateExtractor.h"

///////////////////////////////
// News-Concept
TNewsConcept::TNewsConcept(const TStr& _Uri, const TStr& _Label, const TStr& _Type) :
Uri(_Uri), Label(_Label), Type(_Type), WgtP(false) { }

TNewsConcept::TNewsConcept(const TStr& _Uri, const TStr& _Label, const TStr& _Type,
	const double& _Wgt) : Uri(_Uri), Label(_Label), Type(_Type), WgtP(true), Wgt(_Wgt) { }

TNewsConcept::TNewsConcept(TSIn& SIn) :
Uri(SIn), Label(SIn), Type(SIn), Desc(SIn), WgtP(SIn), Wgt(SIn) { }

void TNewsConcept::Save(TSOut& SOut) const {
	Uri.Save(SOut);
	Label.Save(SOut);
	Type.Save(SOut);
	Desc.Save(SOut);
	WgtP.Save(SOut);
	Wgt.Save(SOut);
}

TNewsConcept::TNewsConcept(const PXmlTok& ConceptTok) :
Uri(ConceptTok->GetStrArgVal("uri")), Label(ConceptTok->GetStrArgVal("label")),
Type(ConceptTok->GetStrArgVal("type")), Desc(ConceptTok->GetStrArgVal("desc")),
WgtP(ConceptTok->IsArg("wgt")), Wgt(ConceptTok->GetFltArgVal("wgt", 0.0)) { }

PXmlTok TNewsConcept::GetXmlTok() const {
	PXmlTok ConceptTok = TXmlTok::New("concept");
	ConceptTok->AddArg("uri", GetUri());
	ConceptTok->AddArg("label", GetLabel());
	ConceptTok->AddArg("type", GetType());
	ConceptTok->AddArg("desc", GetDesc());
	if (IsWgt()) { ConceptTok->AddArg("wgt", GetWgt()); }
	return ConceptTok;
}

uint64 TNewsConcept::GetMemUsed() const {
	uint64 MemUsed = 0;
	MemUsed += (uint64) Uri.GetMemUsed();
	MemUsed += (uint64) Label.GetMemUsed();
	MemUsed += (uint64) Type.GetMemUsed();
	MemUsed += (uint64) Desc.GetMemUsed();
	MemUsed += (uint64) WgtP.GetMemUsed();
	MemUsed += (uint64) Wgt.GetMemUsed();
	return MemUsed;
}

///////////////////////////////
// News-Article
TNewsArticle::TNewsArticle(TSIn& SIn) : SourceUri(SIn), SourceTitle(SIn),
SourceType(SIn), SourceLocP(SIn), SourceLoc(SIn), SourceCountry(SIn), SourceCity(SIn),
ArticleLang(SIn), ArticleUri(SIn), ArticlePubTm(SIn), ArticleCrawlTm(SIn),
ArticleTitle(SIn), ArticleBody(SIn), ArticleKeyWd(SIn), ArticleClustId(SIn), ArticleGoogleClustId(SIn),
AuxKdV(SIn), ImageUriV(SIn), ConceptV(SIn), SimArticleV(SIn), LocationConceptN(SIn), DatesJson(SIn), CategoryV(SIn) {}

void TNewsArticle::Save(TSOut& SOut) const {
	SourceUri.Save(SOut); SourceTitle.Save(SOut); SourceType.Save(SOut);
	SourceLocP.Save(SOut); SourceLoc.Save(SOut); SourceCountry.Save(SOut); SourceCity.Save(SOut);
	ArticleLang.Save(SOut); ArticleUri.Save(SOut); ArticlePubTm.Save(SOut); ArticleCrawlTm.Save(SOut);
	ArticleTitle.Save(SOut); ArticleBody.Save(SOut); ArticleKeyWd.Save(SOut); ArticleClustId.Save(SOut); ArticleGoogleClustId.Save(SOut);
	AuxKdV.Save(SOut); ImageUriV.Save(SOut); ConceptV.Save(SOut); SimArticleV.Save(SOut); LocationConceptN.Save(SOut); DatesJson.Save(SOut); CategoryV.Save(SOut);
}

TNewsArticle::TNewsArticle(const PXmlTok& ArticleTok, const PTokenizer& SearchTok, const PTokenizer& FtrTok)
{
	// parse source
	PXmlTok SourceTok = ArticleTok->GetTagTok("source");
	SourceUri = SourceTok->GetTagVal("uri", false);
	SourceTitle = SourceTok->GetTagVal("title", false);
	SourceType = SourceTok->GetTagVal("type", false);
	if (SourceTok->IsSubTag("location")) {
		SourceLocP = true;
		PXmlTok LocTok = SourceTok->GetTagTok("location");
		SourceLoc.Val1 = LocTok->GetTagVal("latitude", false).GetFlt();
		SourceLoc.Val2 = LocTok->GetTagVal("longitude", false).GetFlt();
		SourceCountry = LocTok->GetTagVal("country", false);
		SourceCity = LocTok->GetTagVal("city", false);
	}
	else {
		SourceLocP = false;
	}
	// parse article
	ArticleLang = ArticleTok->GetTagVal("lang", false);
	ArticleUri = ArticleTok->GetTagVal("uri", false);
	ArticlePubTm = ArticleTok->IsTagTok("publish-date") ?
		TTm::GetTmFromWebLogDateTimeStr(ArticleTok->GetTagVal("publish-date", false)) : TTm();
	ArticleCrawlTm = TTm::GetTmFromWebLogDateTimeStr(ArticleTok->GetTagVal("retrieve-date", false));
	ArticleTitle = ArticleTok->GetTagVal("title", false);
	ArticleBody = ArticleTok->GetTagVal("body", false);
	// prepare search and feature tokens
	/*if (!FtrTok.Empty()) {
	TChA AllChA = ArticleTitle; AllChA += '\n'; AllChA += ArticleBody;
	FtrTok->GetTokens(AllChA, ArticleKeyWd);
	}*/
	ArticleClustId = ArticleTok->GetTagVal("cluster", false);
	ArticleGoogleClustId = ArticleTok->IsTagTok("google_story_id") ? ArticleTok->GetTagVal("google_story_id", false) : "";
	// parse images
	if (ArticleTok->IsTagTok("img")) {
		ArticleTok->GetTagValV("img", false, ImageUriV);
	}
	// parse concepts
	TXmlTokV ConceptTokV; ArticleTok->GetTagTokV("concepts|concept", ConceptTokV);
	for (int ConceptN = 0; ConceptN < ConceptTokV.Len(); ConceptN++) {
		ConceptV.Add(TNewsConcept(ConceptTokV[ConceptN]));
	}
	LocationConceptN = ArticleTok->GetTagVal("locationConceptN", false).GetInt(TInt::Mx);
}

PXmlTok TNewsArticle::GetXmlTok() const
{
	PXmlTok TopTok = TXmlTok::New("article");
	// dump source
	PXmlTok SourceTok = TXmlTok::New("source");
	SourceTok->AddSubTok(TXmlTok::New("uri", SourceUri));
	SourceTok->AddSubTok(TXmlTok::New("title", SourceTitle));
	SourceTok->AddSubTok(TXmlTok::New("type", SourceType));
	if (SourceLocP) {
		PXmlTok LocTok = TXmlTok::New("location");
		LocTok->AddSubTok(TXmlTok::New("latitude", SourceLoc.Val1.GetStr()));
		LocTok->AddSubTok(TXmlTok::New("longitude", SourceLoc.Val2.GetStr()));
		LocTok->AddSubTok(TXmlTok::New("country", SourceCountry));
		LocTok->AddSubTok(TXmlTok::New("city", SourceCity));
		SourceTok->AddSubTok(LocTok);
	}
	TopTok->AddSubTok(SourceTok);
	// dump article
	TopTok->AddSubTok(TXmlTok::New("lang", ArticleLang));
	TopTok->AddSubTok(TXmlTok::New("uri", ArticleUri));
	if (ArticlePubTm.IsDef()) {
		TopTok->AddSubTok(TXmlTok::New("publish-date", ArticlePubTm.GetWebLogDateTimeStr()));
	}
	TopTok->AddSubTok(TXmlTok::New("retrieve-date", ArticleCrawlTm.GetWebLogDateTimeStr()));
	TopTok->AddSubTok(TXmlTok::New("title", ArticleTitle));
	TopTok->AddSubTok(TXmlTok::New("body", ArticleBody));
	TopTok->AddSubTok(TXmlTok::New("cluster", ArticleClustId));
	TopTok->AddSubTok(TXmlTok::New("google_story_id", ArticleGoogleClustId));
	// dump images
	for (int ImageUriN = 0; ImageUriN < ImageUriV.Len(); ImageUriN++) {
		TopTok->AddSubTok(TXmlTok::New("img", ImageUriV[ImageUriN]));
	}
	// dump concepts
	if (!ConceptV.Empty()) {
		PXmlTok ConceptsTok = TXmlTok::New("concepts");
		for (int ConceptN = 0; ConceptN < ConceptV.Len(); ConceptN++) {
			ConceptsTok->AddSubTok(ConceptV[ConceptN].GetXmlTok());
		}
		TopTok->AddSubTok(ConceptsTok);
	}
	TopTok->AddSubTok(TXmlTok::New("locationConceptN", LocationConceptN.GetStr()));
	return TopTok;
}

PJsonVal TNewsArticle::GetJson(const PNewsFixDb NewsFixDb) const
{
	PJsonVal ArticleVal = TJsonVal::NewObj();

	// basic article parameters
	ArticleVal->AddToObj("URI", ArticleUri);
	ArticleVal->AddToObj("Language", ArticleLang);

	// remember newsfeed id
	for (int N = 0; N < AuxKdV.Len(); N++) {
		if (AuxKdV[N].Key == "NewsfeedId")
			ArticleVal->AddToObj("NewsfeedId", AuxKdV[N].Dat);
	}

	// published date is the one provided by publisher. should be best to use except in cases when the clock of the publisher is totally mixed up
	// check the diff between the dates. if smaller than 5 days then use publisher date, otherwise crawl date
	if (!ArticleCrawlTm.IsDef() && !ArticlePubTm.IsDef()) {
		TNotify::StdNotify->OnStatus("Warning: Received an article without a valid crawl or published time. Article was ignored.");
		return TJsonVal::NewObj();
	}
	// if no publisher time just use crawl time
	if (!ArticlePubTm.IsDef()) {
		ArticleVal->AddToObj("DateTime", ArticleCrawlTm.GetWebLogDateTimeStr(true, "T", false));
		ArticleVal->AddToObj("Date", ArticleCrawlTm.GetWebLogDateStr());
	}
	// no crawl time (rarely - in articles from STA, for example) - just use the publisher time
	else if (!ArticleCrawlTm.IsDef()) {
		ArticleVal->AddToObj("DateTime", ArticlePubTm.GetWebLogDateTimeStr(true, "T", false));
		ArticleVal->AddToObj("Date", ArticlePubTm.GetWebLogDateStr());
	}
	else {
		uint64 DiffHrs = TTm::GetDiffHrs(ArticlePubTm, ArticleCrawlTm);
		// use publisher time
		if (DiffHrs < 5 * 24 && ArticlePubTm < ArticleCrawlTm) {
			ArticleVal->AddToObj("DateTime", ArticlePubTm.GetWebLogDateTimeStr(true, "T", false));
			ArticleVal->AddToObj("Date", ArticlePubTm.GetWebLogDateStr());
		}
		// use crawl time
		else {
			ArticleVal->AddToObj("DateTime", ArticleCrawlTm.GetWebLogDateTimeStr(true, "T", false));
			ArticleVal->AddToObj("Date", ArticleCrawlTm.GetWebLogDateStr());
		}
	}

	if (ArticlePubTm.IsDef())
		ArticleVal->AddToObj("DateTimePub",  ArticlePubTm.GetWebLogDateTimeStr(true, "T", false));
	if (ArticleCrawlTm.IsDef())
		ArticleVal->AddToObj("DateTimeCrawl", ArticleCrawlTm.GetWebLogDateTimeStr(true, "T", false));

	ArticleVal->AddToObj("Title", ArticleTitle);
	ArticleVal->AddToObj("Body", ArticleBody);
	ArticleVal->AddToObj("BodyLen", ArticleBody.Len());
	//ArticleVal->AddToObj("Keywords", TJsonVal::NewArr(ArticleKeyWd));
	ArticleVal->AddToObj("ExtractedDates", DatesJson);

	TExtractedDateColl ArticleDateColl(DatesJson);
	// it's possible that we have an article with tons of dates - in such cases make sure we only use the first 5 dates
	TJsonValV DateV;
	for (int N = 0; DateV.Len() < 5 && N < ArticleDateColl.Len(); N++) {
		TExtractedDate ExtractedDate = ArticleDateColl.GetExtractedDate(N);
		if (ExtractedDate.IsDateAmbiguous() || !ExtractedDate.IsDateComplete())
			continue;
		PJsonVal DateJson = TJsonVal::NewObj("Date", ExtractedDate.GetDate());
		DateJson->AddToObj("$fq", ExtractedDate.GetFreq());
		DateV.Add(DateJson);
	}
	ArticleVal->AddToObj("hasDate", TJsonVal::NewArr(DateV));

	// article source - use source info from newsfixdb if existing
	PJsonVal SourceVal = NewsFixDb->HasSourceInfo(SourceUri) ? NewsFixDb->GetSourceInfo(SourceUri) : TJsonVal::NewObj();
	SourceVal->AddToObj("URI", SourceUri);
	if (!SourceVal->IsObjKey("Title"))
		SourceVal->AddToObj("Title", SourceTitle);
	if (!SourceVal->IsObjKey("Type"))
		SourceVal->AddToObj("Type", SourceType);
	if (SourceLocP) {
		if (!SourceVal->IsObjKey("Location"))
			SourceVal->AddToObj("Location", TJsonVal::NewArr(SourceLoc));
		if (!SourceVal->IsObjKey("Country"))
			SourceVal->AddToObj("Country", SourceCountry);
		if (!SourceVal->IsObjKey("City"))
			SourceVal->AddToObj("City", SourceCity);
		if (SourceVal->IsObjKey("City") && SourceVal->IsObjKey("Country"))
			SourceVal->AddToObj("CityCountry", SourceVal->GetObjStr("City") + ", " + SourceVal->GetObjStr("Country"));
	}
	ArticleVal->AddToObj("hasSource", SourceVal);

	if (ArticleGoogleClustId != "")
		ArticleVal->AddToObj("GoogleStoryId", ArticleGoogleClustId);

	// auxilary stuff
	for (int AuxN = 0; AuxN < AuxKdV.Len(); AuxN++) {
		ArticleVal->AddToObj(AuxKdV[AuxN].Key, AuxKdV[AuxN].Dat);
	}
	// article image
	if (ImageUriV.Len() > 0)
		ArticleVal->AddToObj("hasImage", TJsonVal::NewObj("URI", ImageUriV[0]));

	// add article categories
	if (CategoryV.Len() > 0) {
		TJsonValV CatJsonV;
		for (int N = 0; N < CategoryV.Len(); N++) {
			PJsonVal CatJson = TJsonVal::NewObj("URI", CategoryV[N].Val1);
			CatJson->AddToObj("$fq", (int) 100 * CategoryV[N].Val2);
			CatJsonV.Add(CatJson);
		}
		ArticleVal->AddToObj("hasCategory", CatJsonV);
	}

	// article concepts
	// remember for each concept uri it's corresponding index in ConceptsVal
	// sometimes the same concept is added more than once. this can result in wgts > 5 for a concept. 
	// we make sure it is added only once and just take the higher wgt
	TStrIntH ConceptUriToIndexH;
	PJsonVal ConceptsVal = TJsonVal::NewArr();
	for (int ConceptN = 0; ConceptN < ConceptV.Len(); ConceptN++) {
		const double OrigWgt = ConceptV[ConceptN].GetWgt();
		WAssert(OrigWgt >= -0.1 && OrigWgt <= 1.1, TStr::Fmt("Concept weight %.3f is out of the [0,1] range", OrigWgt));
		const int CptWgt = 1 + TFlt::Round(ConceptV[ConceptN].GetWgt() * 4);

		PJsonVal ConceptVal = TJsonVal::NewObj();
		TStr ConceptUri = ConceptV[ConceptN].GetUri();

		TStr Label = ConceptV[ConceptN].GetLabel();
		TStr Type = ConceptV[ConceptN].GetType();
		// correct the url, type and label for the given concept - if we have info from the rules files
		if (!NewsFixDb.Empty()) {
			NewsFixDb->CorrectConceptInfo(ConceptUri, Type, Label, ArticleLang);
			// after we (possibly) corrected the url, we check if this url is maybe ignored
			if (NewsFixDb->HasConceptIgnore(ConceptUri))
				continue;
		}

		// once we are sure we have the correct uri for the concept we can check if we've already seen it. if yes, just use the larger wgt
		if (ConceptUriToIndexH.IsKey(ConceptUri)) {
			const int ConcInd = ConceptUriToIndexH.GetDat(ConceptUri);
			PJsonVal ExistingConc = ConceptsVal->GetArrVal(ConcInd);
			int ExistingWgt = (int) ExistingConc->GetObjNum("$fq");
			ExistingConc->AddToObj("$fq", MAX(ExistingWgt, CptWgt));
			continue;
		}

		ConceptVal->AddToObj("URI", ConceptUri);
		ConceptVal->AddToObj("Label", Label);		
		ConceptVal->AddToObj("Type", Type);
		// set description field only if not "", otherwise we will rewrite the existing description
		if (ConceptV[ConceptN].GetDesc() != "")
			ConceptVal->AddToObj("Description", ConceptV[ConceptN].GetDesc());
		ConceptVal->AddToObj("$fq", CptWgt);
		// remember the index where we'll put the concept
		ConceptUriToIndexH.AddDat(ConceptUri, ConceptsVal->GetArrVals());

		ConceptsVal->AddToArr(ConceptVal);
	}
	if (ConceptsVal->GetArrVals() > 0) {
		ArticleVal->AddToObj("hasConcept", ConceptsVal);
	}
	// save the article location if set
	if (LocationConceptN != TInt::Mx && LocationConceptN < ConceptV.Len()) {
		PJsonVal ConceptVal = TJsonVal::NewObj("URI", ConceptV[LocationConceptN].GetUri());
		ConceptVal->AddToObj("Label", ConceptV[LocationConceptN].GetLabel());		
		ConceptVal->AddToObj("Type", ConceptV[LocationConceptN].GetType());
		ArticleVal->AddToObj("hasLocation", ConceptVal);
	}
	return ArticleVal;
}

uint64 TNewsArticle::GetMemUsed() const {
	uint64 MemUsed = 0;
	MemUsed += (uint64) SourceUri.GetMemUsed();
	MemUsed += (uint64) SourceTitle.GetMemUsed();
	MemUsed += (uint64) SourceType.GetMemUsed();
	MemUsed += (uint64) SourceLocP.GetMemUsed();
	MemUsed += (uint64) SourceLoc.GetMemUsed();
	MemUsed += (uint64) SourceCountry.GetMemUsed();
	MemUsed += (uint64) SourceCity.GetMemUsed();
	MemUsed += (uint64) ArticleLang.GetMemUsed();
	MemUsed += (uint64) ArticleUri.GetMemUsed();
	MemUsed += (uint64) ArticlePubTm.GetMemUsed();
	MemUsed += (uint64) ArticleCrawlTm.GetMemUsed();
	MemUsed += (uint64) ArticleTitle.GetMemUsed();
	MemUsed += (uint64) ArticleBody.GetMemUsed();
	MemUsed += (uint64) ArticleKeyWd.GetMemUsed();
	MemUsed += (uint64) ArticleClustId.GetMemUsed();
	MemUsed += (uint64) ArticleGoogleClustId.GetMemUsed();
	MemUsed += (uint64) ImageUriV.GetMemUsed();
	MemUsed += (uint64) ConceptV.GetMemUsed();
	return MemUsed;
}

void TNewsArticle::AddCategory(const TStr& CatName, const double& Wgt)
{
	for (int N = 0; N < CategoryV.Len(); N++) {
		if (CategoryV[N].Val1 == CatName) {
			CategoryV[N].Val2 = MAX(CategoryV[N].Val2.Val, Wgt);
			return;
		}
	}
	CategoryV.Add(TStrFltPr(CatName, Wgt));
}

///////////////////////////////
// News-Source-Types
TStr TNewsSourceType::MainstreamNews = "MAINSTREAM_NEWS";
TStr TNewsSourceType::Weblog = "WEBLOG";
TStr TNewsSourceType::MicroBlog = "MICRO_BLOG";
