/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "NewsBase.h"
#include "NewsFixDb.h"
#include "../Utils/DateExtractor.h"

///////////////////////////////
// News-Notifier
//PNotify TNewsNotify::Notify = TStdNotify::New();


///////////////////////////////
// News-Base
TNewsBase::TNewsBase(const PGeoBase _GeoBase, const PNewsFixDb _NewsFixDb, const PConceptLabel _ConceptLabel, const TStr& StoreDefFNm, const TStr& FPath, const int64& IndexCacheSize, const int64& ArticleCacheSize,
	const TStr& _SettingsFNm, const TStr& _DbsPath, const TStr& TempFPath) : GeoBase(_GeoBase), NewsFixDb(_NewsFixDb), ConceptLabel(_ConceptLabel), SettingsFNm(_SettingsFNm), DbsPath(_DbsPath) {
	
	NewsBaseFix = TNewsBaseFix::New(this);

	// initialize store
	PJsonVal StoreDefVal = TJsonVal::GetValFromStr(TStr::LoadTxt(StoreDefFNm));
	// prepare base
	Base = NewBase(FPath, StoreDefVal, IndexCacheSize, ArticleCacheSize, TStrUInt64H(), false);

	// initialize news base
	Init();
	// finish initalization
	Base->Init(TempFPath);
	SaveBase(Base);
}

TNewsBase::TNewsBase(const PGeoBase _GeoBase, const PNewsFixDb _NewsFixDb, const PConceptLabel _ConceptLabel, const TStr& FPath, const TFAccess& FAccess, const int64& IndexCacheSize, const int64& ArticleCacheSize,
	const TStr& _SettingsFNm, const TStr& _DbsPath, const TStr& TempFPath) : 
	GeoBase(_GeoBase), NewsFixDb(_NewsFixDb), ConceptLabel(_ConceptLabel), SettingsFNm(_SettingsFNm), DbsPath(_DbsPath) {
	
	NewsBaseFix = TNewsBaseFix::New(this);

	// prepare base
	Base = LoadBase(FPath, FAccess, IndexCacheSize, ArticleCacheSize, TStrUInt64H(), false);
	// load newsbase
	if (TFile::Exists(FPath + "CustomData/NewsBase.dat")) {
		TFIn FIn(FPath + "CustomData/NewsBase.dat");
		NewsfeedIdH.Load(FIn);
	}
	// initialize news base
	Init();
	// finish initalization
	Base->Init(TempFPath);
}

void TNewsBase::Init() {
	Timer = NULL;
	CustomDataPath = Base->GetIndex()->GetIndexFPath() + "CustomData/";
		
	// set up the files where we will save the info and error logs
	InfoLogFile = TLogger::New(3);
	InfoLogFile->AddLogger(new TFileNotify("./Logs/Log-EventRegistry-Info.txt", true, true));
	ActivityLogFile = TLogger::New(3);
	ActivityLogFile->AddLogger(new TFileNotify("./Logs/Log-EventRegistry-Activity.txt", true, true));
	ErrorLogFile = TLogger::New(3);
	ErrorLogFile->AddLogger(new TFileNotify("./Logs/Log-EventRegistry-Errors.txt", true, true));

	InfoLogFile->NotifyVerbose(0, TStr("Process started.").CStr());

	ReloadSettings();

	IndexVoc = Base->GetIndexVoc();
	Index = Base->GetIndex();
	// prepare article store
	EAssertR(Base->IsStoreNm("Article"), "Article store does not exist");
	ArticleStore = Base->GetStoreByStoreNm("Article");
	ArticleStoreId = ArticleStore->GetStoreId();
	// prepare source store
	EAssertR(Base->IsStoreNm("Source"), "Source store does not exist");
	SourceStore = Base->GetStoreByStoreNm("Source");
	SourceStoreId = SourceStore->GetStoreId();
	// prepare concept store
	EAssertR(Base->IsStoreNm("Concept"), "Concept store does not exist");
	ConceptStore = Base->GetStoreByStoreNm("Concept");
	ConceptStoreId = ConceptStore->GetStoreId();
	// prepare concept class store
	if (Base->IsStoreNm("ConceptClass")) {
		EAssertR(Base->IsStoreNm("ConceptClass"), "ConceptClass store does not exist");
		ConceptClassStore = Base->GetStoreByStoreNm("ConceptClass");
		ConceptClassStoreId = ConceptClassStore->GetStoreId();
	}
	// prepare concept folder store
	if (Base->IsStoreNm("ConceptFolder")) {
		EAssertR(Base->IsStoreNm("ConceptFolder"), "ConceptFolder store does not exist");
		ConceptFolderStore = Base->GetStoreByStoreNm("ConceptFolder");
		ConceptFolderStoreId = ConceptFolderStore->GetStoreId();
	}
	// prepare user store
	if (Base->IsStoreNm("User")) {
		EAssertR(Base->IsStoreNm("User"), "User store does not exist");
		UserStore = Base->GetStoreByStoreNm("User");
		UserStoreId = UserStore->GetStoreId();
	}
	// prepare story store
	EAssertR(Base->IsStoreNm("Story"), "Story store does not exist");
	StoryStore = Base->GetStoreByStoreNm("Story");
	StoryStoreId = StoryStore->GetStoreId();
	// prepare event store
	if (Base->IsStoreNm("Event")) {
		EventStore = Base->GetStoreByStoreNm("Event");
		EventStoreId = EventStore->GetStoreId();
	}
	// prepare image store
	EAssertR(Base->IsStoreNm("Image"), "Image store does not exist");
	ImageStore = Base->GetStoreByStoreNm("Image");
	ImageStoreId = ImageStore->GetStoreId();
	// prepare category store
	if (Base->IsStoreNm("Category")) {
		CategoryStore = Base->GetStoreByStoreNm("Category");
		CategoryStoreId = CategoryStore->GetStoreId();
	}
	else
		ErrorLogFile->NotifyErr("Warning: Category store was not found in the data.");

	// prepare date store
	if (Base->IsStoreNm("Date")) {
		DateStore = Base->GetStoreByStoreNm("Date");
		DateStoreId = DateStore->GetStoreId();
	}
	else
		ErrorLogFile->NotifyErr("Warning: Date store was not found in the data.");

	// prepare tag store
	if (Base->IsStoreNm("Tag")) {
		TagStore = Base->GetStoreByStoreNm("Tag");
		TagStoreId = TagStore->GetStoreId();
	}
	else
		ErrorLogFile->NotifyErr("Warning: Tag store was not found in the data.");
		
	// load news counts
	NewsCounts = TNewsCounts::New(Base->GetFPath() + "CustomData/NewsCounters.bin");
	const bool NewsCountsLoaded = NewsCounts->Load();

	// load news activity
	NewsActivity = TNewsActivity::New(Base->GetFPath() + "CustomData/NewsActivity.bin", SettingsFNm);
	NewsActivity->Load();

	// load validated story links
	ValidatedStoryLinks = TValidatedStoryLinks::New(Base->GetFPath() + "CustomData/ValidatedStoryLinks.bin");
	ValidatedStoryLinks->Load();

	// load validated article links
	ValidatedArticleLinks = TValidatedArticleLinks::New(Base->GetFPath() + "CustomData/ValidatedArticleLinks.bin");
	ValidatedArticleLinks->Load();

	// add article trigger
	NewArticleTrigger = TNewsArticleTrigger::New(this);
	ArticleStore->AddTrigger(NewArticleTrigger);

	CategoryBase = TCategoryBase::New(Base);

	// set up trends manager and the trigger for it
	TrendsManager = TConceptTrendsManager::New(Base->GetFPath() + "CustomData/TrendsManager.bin", SettingsFNm);
	bool TrendsLoaded = TrendsManager->Load();
		
	DailyTrendInfos = TDailyTrendInfos::New(this, Base->GetFPath() + "CustomData/DailyTrendsInfos.bin");
	DailyTrendInfos->Load();
	TTrendParams TrendParams;
	DailyTrendInfos->AddTrendParams(TrendParams, "default");
	
	// add story trigger
	NewStoryTrigger = TNewStoryTrigger::New(this, InfoLogFile, ErrorLogFile);
	StoryStore->AddTrigger(NewStoryTrigger);

	// if we haven't successfully parsed the counts or trends file then rebuild it
	if (NewsCountsLoaded == false || TrendsLoaded == false)
		RebuildNewsCountsAndTrends();
}

void TNewsBase::Save() { 
	const TStr FName = Base->GetFPath() + +"CustomData/NewsBase.dat";
	TFOut FOut(FName);
	NewsfeedIdH.Defrag();	// we might have removed some items if old articles were deleted
	NewsfeedIdH.Save(FOut);
}

PNewsBase TNewsBase::New(const PGeoBase GeoBase, const PNewsFixDb NewsFixDb, const PConceptLabel ConceptLabel, const TStr& StoreDefFNm, const TStr& FPath,
	const int64& IndexCacheSize, const int64& ArticleCacheSize, const TStr& SettingsFNm, const TStr& DbsPath, const TStr& TempFPath) {
			
	return new TNewsBase(GeoBase, NewsFixDb, ConceptLabel, StoreDefFNm, FPath, IndexCacheSize, ArticleCacheSize, SettingsFNm, DbsPath, TempFPath);
}

PNewsBase TNewsBase::Load(const PGeoBase GeoBase, const PNewsFixDb NewsFixDb, const PConceptLabel ConceptLabel, const TStr& FPath, const TFAccess& FAccess,
	const int64& IndexCacheSize, const int64& ArticleCacheSize, const TStr& SettingsFNm, const TStr& DbsPath, const TStr& TempFPath) {

	return new TNewsBase(GeoBase, NewsFixDb, ConceptLabel, FPath, FAccess, IndexCacheSize, ArticleCacheSize, SettingsFNm, DbsPath, TempFPath);
}

void TNewsBase::ReloadSettings()
{
	const TStr SettingsStr = TFile::Exists(SettingsFNm) ? TStr::LoadTxt(SettingsFNm) : "{}";
	const PJsonVal AllSettings = TJsonVal::GetValFromStr(SettingsStr);
	const PJsonVal NewsBaseSettings = AllSettings->IsObjKey("newsBase") ? AllSettings->GetObjKey("newsBase") : TJsonVal::NewObj();

	ClearInfoForEmptyStoriesAndEvents = NewsBaseSettings->GetObjBool("ClearInfoForEmptyStoriesAndEvents", true);
	GarbageCollectionTimerMins = NewsBaseSettings->GetObjInt("GarbageCollectionTimerMins", 10);
	const int InfoVerbosity = NewsBaseSettings->GetObjInt("InfoVerbosityLevel", 3);
	const int ActivityVerbosity = NewsBaseSettings->GetObjInt("ActivityVerbosityLevel", 3);
	const int ErrorVerbosity = NewsBaseSettings->GetObjInt("ErrorVerbosityLevel", 3);
	InfoLogFile->SetVerbosityLevel(InfoVerbosity);
	ActivityLogFile->SetVerbosityLevel(ActivityVerbosity);
	ErrorLogFile->SetVerbosityLevel(ErrorVerbosity);

	if (GarbageCollectionTimerMins > 0)
		Timer = new TTmTimer(GarbageCollectionTimerMins * 60 * 1000);
	else if (Timer != NULL) {
		delete Timer;
		Timer = NULL;
	}
}

TStr TNewsBase::ParseSourceUri(const TStr& InUri, const bool& KeepWwwP) {
	// if empty, do nothing
	if (InUri.Empty()) { return InUri; }
	// extract the basic url
	PUrl Url = TUrl::New(InUri);
	if (!Url->IsOk()) { return ""; }
	TChA OutUri = Url->GetHostNm();
	if (!OutUri.IsPrefix("www.") || !KeepWwwP) {
		// clean URLs only if they do not start with www. prefix
		if (OutUri.CountCh('.') > 1) { 
			// remove any subdomains
			int FirstDot = OutUri.SearchChBack('.');
			int SecondDot = OutUri.SearchChBack('.', FirstDot - 1);
			if (OutUri.Len() - SecondDot > 7) {			
				OutUri = OutUri.GetSubStr(SecondDot + 1, OutUri.Len() - 1);
			}
		}
	}
	return OutUri.ToLc();
}

// is the source enabled?
// if NewsFixDb has the source and is set to disabled return false
// if source is in the sourcestore and is disabled return false
// return true
bool TNewsBase::IsSourceEnabled(const TStr& SourceUri)
{
	if (NewsFixDb->HasSourceInfo(SourceUri)) {
		PJsonVal SourceJson = NewsFixDb->GetSourceInfo(SourceUri);
		if (SourceJson->IsObjKey("Enabled") && SourceJson->GetObjBool("Enabled") == false)
			return false;
	}
	if (SourceStore->IsRecNm(SourceUri)) {
		const uint64 SourceId = SourceStore->GetRecId(SourceUri);
		if (SourceStore->IsFieldNm("Enabled") && SourceStore->GetFieldNmBool(SourceId, "Enabled") == false)
			return false;
	}
	return true;
}

TStr TNewsBase::ParseSourceTitle(const TStr& InTitle) {
	TChA TitleChA = InTitle;
	// remove any sub-titles
	TChA FilterChV = "-|/:>";
	for (int FilterChN = 0; FilterChN < FilterChV.Len(); FilterChN++) {
		const char FilterCh = FilterChV[FilterChN];
		if (TitleChA.IsChIn(FilterCh)) {
			int ChStart = TitleChA.SearchCh(FilterCh);
			if (ChStart > 0) { TitleChA = TitleChA.GetSubStr(0, ChStart-1); }
		}
	}
	// trime white space
	return TitleChA.ToTrunc();
}

bool TNewsBase::HasLocationForArticleSource(const uint64& ArticleId) const
{
	if (!ArticleStore->IsRecId(ArticleId))
		return false;
	TRec ArticleRec = ArticleStore->GetRec(ArticleId);
	TRec SourceRec = ArticleRec.DoSingleJoin(Base, "hasSource");
	if (SourceRec.IsDef() && !SourceStore->IsFieldNmNull(SourceRec.GetRecId(), "Location"))
		return true;
	return false;
}

uint64 TNewsBase::AddArticle(const TNewsArticle& Article) 
{
	// check if the source is enabled - if not then ignore the article
	const TStr SourceUri = Article.SourceUri;
	if (NewsFixDb->HasSourceInfo(SourceUri)) {
		PJsonVal SourceJson = NewsFixDb->GetSourceInfo(SourceUri);
		if (SourceJson->IsObjKey("Enabled") && SourceJson->GetObjBool("Enabled") == false)
			return TUInt64::Mx;
	}
	if (SourceStore->IsRecNm(SourceUri)) {
		const uint64 SourceId = SourceStore->GetRecId(SourceUri);
		if (SourceStore->IsFieldNm("Enabled") && SourceStore->GetFieldNmBool(SourceId, "Enabled") == false)
			return TUInt64::Mx;
	}

	// get article json
	PJsonVal ArticleJson = Article.GetJson(NewsFixDb);
	
	// before we add an article we need to check if we have some categories for it. if yes, make sure we create the whole category heirarchy in the CategoryStore
	for (int N = 0; N < Article.CategoryV.Len(); N++)
		CategoryBase->CreateDMozCats(Article.CategoryV[N].Val1);

	// add the article
	const uint64 ArticleId = AddArticle(ArticleJson);
	if (ArticleId == TUInt64::Mx)
		return ArticleId;
	
	// ///////
	// register similar articles
	for (int SimArticleN = 0; SimArticleN < Article.SimArticleV.Len(); SimArticleN++) {
		const TStrStrIntTr& SimArticle = Article.SimArticleV[SimArticleN];
		const TStr SimLang = SimArticle.Val2;
		const TStr& NewsfeedId = SimArticle.Val1;
		const uint64 SimArticleId = GetArticleIdForNewsfeedId(NewsfeedId);
		const int Wgt = SimArticle.Val3;
		if (SimArticleId != TUInt64::Mx) {
			const TStr JoinNm = "similar_" + SimLang;
			const TStr JoinNm2 = "similar_" + Article.ArticleLang;
			if (!ArticleStore->IsJoinNm(JoinNm))
				ArticleStore->AddJoinDesc(TJoinDesc(JoinNm, ArticleStore->GetStoreId(), ArticleStore->GetStoreId(), Base->GetIndexVoc()));
			if (!ArticleStore->IsJoinNm(JoinNm2))
				ArticleStore->AddJoinDesc(TJoinDesc(JoinNm2, ArticleStore->GetStoreId(), ArticleStore->GetStoreId(), Base->GetIndexVoc()));

			Index->IndexJoin(ArticleStore, JoinNm, ArticleId, SimArticleId, Wgt);
			Index->IndexJoin(ArticleStore, JoinNm2, SimArticleId, ArticleId, Wgt);
		}
		else {
			// unkown article
		}
	}
	return ArticleId;
}

// article can be added directly as a json object
// WARNING: this will not add joins to similar articles - these are only added if article is added as TNewsArticle object
uint64 TNewsBase::AddArticle(const PJsonVal& ArticleVal)
{
	// ignore articles that are not objects or don't contain URI
	if (!ArticleVal->IsObj())
		throw TExcept::New("Add article failed. The given data is not a valid JSON object.");
	if (!ArticleVal->IsObjKey("URI"))
		throw TExcept::New("Add article failed. The article did not contain URI.");
	// ignore articles that are already indexed
	if (ArticleStore->IsRecNm(ArticleVal->GetObjStr("URI"))) 
		return TUInt64::Mx;

	// shall we call garbage collection
	const int GargageThreshold = 10000;
	if (Timer != NULL && Timer->IsTimeUp() || NewArticles >= GargageThreshold) {
		Base->GarbageCollect();
		Timer->Restart(); NewArticles = 0;
	}

	NewArticles++;
	
	// 
	// save concept labels (not stored in the stores) and remove the concept types for known concepts
	TStr ArticleLang = ArticleVal->GetObjStr("Language", "");
	// first clear concept types for the known concepts. we do this in order not to override the known (correct) concept types
	if (ArticleVal->IsObjKey("hasConcept")) {
		PJsonVal ConceptArr = ArticleVal->GetObjKey("hasConcept");
		EAssertR(ConceptArr->IsArr(), "Article's hasConcept did not contain an array.");
		int Cnt = ConceptArr->GetArrVals();
		for (int N = 0; N < Cnt; N++) {
			PJsonVal ConceptJson = ConceptArr->GetArrVal(N);
			const TStr ConceptUri = ConceptJson->GetObjStr("URI", "");
			if (ConceptStore->IsRecNm(ConceptUri)) {
				ConceptJson->DelObjKey("Type");
				ConceptJson->DelObjKey("Description");
			}
			const TStr Label = ConceptJson->GetObjStr("Label", "");
			// ConceptLabel can contain fixed concept names. The names we get from articles can contain bad characters. in order not to screw up the names
			// we only update the label if we don't already have the label for the given language
			if (Label != "" && !ConceptLabel->HasConceptLabel(ConceptUri, ArticleLang))
				ConceptLabel->SetConceptLabel(ConceptUri, ArticleLang, Label);
		}
	}

	const uint64 ArticleId = ArticleStore->AddRec(ArticleVal);
	EAssertR(ArticleId != TUInt64::Mx, "Adding new article failed. Got invalid article id.");
	
	// update article activity
	if (ArticleId != TUInt64::Mx)
		NewsActivity->AddArticleActivity(ArticleId);
	
	return ArticleId;
}

// delete the article info 
// used when the article is removed from the store (when too old) or when we want for some reason to hide the article from any kind of search (title, concept, body, category, ....)
// for example, used when we decide to remove a particular news source that was currently enabled
void TNewsBase::DeleteArticleInfo(const uint64& ArticleId)
{
	if (!ArticleStore->IsRecId(ArticleId)) {
		ErrorLogFile->NotifyInfo("NewsBase::DeleteArticleInfo. Trying to remove a non-existing article %I64u.", ArticleId);
		return;
	}

	InfoLogFile->NotifyInfo("NewsBase::DeleteArticleInfo. Removing article %I64u.", ArticleId);

	// delete newsfeed id -> miner id map entry
	const TStr NewsfeedId = ArticleStore->GetFieldNmStr(ArticleId, "NewsfeedId");
	NewsfeedIdH.DelIfKey(NewsfeedId);

	const TRec ArtRec = ArticleStore->GetRec(ArticleId);
	const TStr Date = ArticleStore->GetFieldNmStr(ArticleId, "Date");
	const TStr Lang = ArticleStore->GetFieldNmStr(ArticleId, "Language");
	if (Date == "" && Lang == "") {
		ConsoleLog->OnStatusFmt("Warning: Deleting article info on an article (%I64u) that has Date and Lang set to empty. It's probably an article that we've already deleted in the past.", ArticleId);
		ErrorLogFile->NotifyInfo("Warning: Deleting article info on an article (%I64u) that has Date and Lang set to empty. It's probably an article that we've already deleted in the past.", ArticleId);
	}

	// 
	// decrease counts 

	// decrease concept frequencies
	const PRecSet ConceptSet = ArtRec.DoJoin(Base, "hasConcept");
	for (int ConceptN = 0; ConceptN < ConceptSet->GetRecs(); ConceptN++) {
		const uint64 ConceptRecId = ConceptSet->GetRecId(ConceptN);
		NewsCounts->AddConceptFq(ConceptRecId, -1);	// decrease fq for concept
		if (ConceptStore->GetFieldNmStr(ConceptRecId, "Type") == NewsConst.TypeLocation)
			GeoBase->DecreaseWgt(ConceptStore->GetRecNm(ConceptRecId));
	}

	// decrease source frequencies
	TRec SourceRec = ArtRec.DoSingleJoin(Base, "hasSource");
	NewsCounts->AddSourceFq(SourceRec.GetRecId(), -1);

	// decrease total article count
	NewsCounts->AddArticleCount(-1);

	// decrease annotated counts if necessary
	if (ConceptSet->GetRecs() > 0) {
		NewsCounts->AddAnnArticleCount(-1);
		NewsCounts->AddAnnArticleCount(Date, -1);
	}
	// decrease article count for a date
	NewsCounts->AddArticleCount(Date, -1);
	
	// 
	// remove the similar_xxx joins for the deleted articles. they are manually added so have to be manually deleted
	TStr ThisNm = "similar_" + ArticleStore->GetFieldNmStr(ArticleId, "Language");
	if (ArticleStore->IsJoinNm(ThisNm)) {
		for (int LangN = 0; LangN < NewsConst.LangSimilarV.Len(); LangN++) {
			const TStr ThatNm = "similar_" + NewsConst.LangSimilarV[LangN];
			// make sure have the join
			if (!ArticleStore->IsJoinNm(ThatNm))
				continue;
			const PRecSet SimArts = ArtRec.DoJoin(Base, ThatNm);
			for (int ArtN = 0; ArtN < SimArts->GetRecs(); ArtN++) {
				const uint64 SimArt = SimArts->GetRecId(ArtN);
				Index->DeleteJoin(ArticleStore, ThisNm, ArticleId, SimArt);
				Index->DeleteJoin(ArticleStore, ThatNm, SimArt, ArticleId);
			}
		}
	}

	// 
	// if ArticleId is the original article that has one or more duplicates then we need to update the duplicate joins
	PRecSet DuplArtRecSet = ArtRec.DoJoin(Base, "hasDuplicate");
	if (DuplArtRecSet->GetRecs() > 0) {
		DuplArtRecSet->SortByField(true, ArticleStore->GetFieldId("DateTime"));		// sort duplicated articles by date
		const uint64 NewOrigArtId = DuplArtRecSet->GetRecId(0);						// id of the newest article will now become the original article and other duplicates should point to it
		ArticleStore->DelJoin("hasDuplicate", ArticleId, NewOrigArtId);
		for (int N = 1; N < DuplArtRecSet->GetRecs(); N++) {
			const uint64 DuplArtId = DuplArtRecSet->GetRecId(N);
			ArticleStore->DelJoin("hasDuplicate", ArticleId, DuplArtId);		// delete existing join ArticleId -> DuplArtId
			ArticleStore->AddJoin("hasDuplicate", NewOrigArtId, DuplArtId);		// ccreate join NewOrigArtId -> DuplArtId
		}
	}

	const uint64 StoryId = ArticleStore->GetFieldNmUInt64(ArticleId, "hasStoryId");

	// 
	// remove the joins
	ArticleStore->DelJoins("hasSource", ArticleId);
	ArticleStore->DelJoins("hasConcept", ArticleId);
	ArticleStore->DelJoins("hasStory", ArticleId);
	ArticleStore->DelJoins("hasImage", ArticleId);
	ArticleStore->DelJoins("hasLocation", ArticleId);
	ArticleStore->DelJoins("hasCategory", ArticleId);
	ArticleStore->DelJoins("hasDate", ArticleId);
	ArticleStore->DelJoins("isDuplicateOf", ArticleId);

	//
	// set empty fields to the article so that searching by title, body, date, ... won't find this article
	ArticleStore->SetFieldNmStr(ArticleId, "Title", "");
	ArticleStore->SetFieldNmStr(ArticleId, "Body", "");
	ArticleStore->SetFieldNmStr(ArticleId, "Language", "");
	ArticleStore->SetFieldNmStr(ArticleId, "Date", "");
	ArticleStore->SetFieldNmStr(ArticleId, "NewsfeedId", "");
	ArticleStore->SetFieldNmStr(ArticleId, "GoogleStoryId", "");
	ArticleStore->SetFieldNmStr(ArticleId, "ExtractedDates", "");
	ArticleStore->SetFieldNmFlt(ArticleId, "CosToCentroid", 0);

	if (ClearInfoForEmptyStoriesAndEvents && StoryStore->IsRecId(StoryId)) {
		PRecSet ArticleSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
		if (ArticleSet->GetRecs() == 0)
			ClearStoryInformation(StoryId);
	}
}

// remove story information for a story that now has no articles anymore
void TNewsBase::ClearStoryInformation(const uint64& StoryId)
{
	if (!StoryStore->IsRecId(StoryId))
		return;
	InfoLogFile->NotifyInfo("NewsBase::ClearStoryInformation. Clearing info about story %I64u.", StoryId);

	StoryStore->SetFieldNmStr(StoryId, "Language", "");
	StoryStore->SetFieldNmFlt(StoryId, "AvgCosToCentroid", -1);
	StoryStore->SetFieldNmFlt(StoryId, "Variance", -1);
	
	StoryStore->SetFieldNmInt(StoryId, "ChangeCount", 0);
	StoryStore->SetFieldNmInt(StoryId, "ArticleCount", 0);
	StoryStore->SetFieldNmStr(StoryId, "ArticleCountKey", "0");

	StoryStore->SetFieldNmFlt(StoryId, "Sentiment", 0);
	StoryStore->SetFieldNmStr(StoryId, "CommonDates", "");
	StoryStore->SetFieldNmStr(StoryId, "ExtractedDate", "");
	StoryStore->SetFieldNmStr(StoryId, "ExtractedDateEnd", "");
	StoryStore->SetFieldNmStr(StoryId, "AverageDate", "");
	
	StoryStore->SetFieldNmStr(StoryId, "StoryDate", "");
	StoryStore->SetFieldNmStr(StoryId, "StoryDateEnd", "");

	StoryStore->SetFieldNmStr(StoryId, "Title", "");
	StoryStore->SetFieldNmStr(StoryId, "Summary", "");

	StoryStore->SetFieldNmInt(StoryId, "StoryFlags", 0);

	StoryStore->DelJoins("hasArticle", StoryId);
	StoryStore->DelJoins("hasMedoid", StoryId);
	StoryStore->DelJoins("hasConcept", StoryId);
	StoryStore->DelJoins("hasLocation", StoryId);
	StoryStore->DelJoins("hasCategory", StoryId);

	const uint64 EventId = StoryStore->GetFieldNmUInt64(StoryId, "hasEventId");
	if (EventStore->IsRecId(EventId)) {
		PRecSet StorySet = EventStore->GetRec(EventId).DoJoin(Base, "hasStory");
		if (StorySet->GetRecs() == 1)
			ClearEventInformation(EventId);
		StoryStore->DelJoin("hasEvent", StoryId, EventId);
	}
}

// remove event information for an event that now has no articles anymore
void TNewsBase::ClearEventInformation(const uint64& EventId)
{
	if (!EventStore->IsRecId(EventId))
		return;
	InfoLogFile->NotifyInfo("NewsBase::ClearEventInformation. Clearing info about event %I64u.", EventId);
	EventStore->SetFieldNmInt(EventId, "ArticleCount", 0);
	EventStore->SetFieldNmStr(EventId, "ArticleCountKey", "0");

	EventStore->SetFieldNmFlt(EventId, "Sentiment", 0);

	EventStore->SetFieldNmStr(EventId, "EventDate", "");
	EventStore->SetFieldNmStr(EventId, "EventDateEnd", "");

	EventStore->DelJoins("hasConcept", EventId);
	EventStore->DelJoins("hasLocation", EventId);
	EventStore->DelJoins("hasCategory", EventId);
}

PRecSet TNewsBase::GetSimilarArticles(const uint64& ArticleId, const TStr& Lang) const
{
	if (!ArticleStore->IsRecId(ArticleId))
		return TRecSet::New();
	const TStr JoinNm = "similar_" + Lang;
	if (!ArticleStore->IsJoinNm(JoinNm))
		return TRecSet::New(ArticleStore);
	PRecSet ArtSet = ArticleStore->GetRec(ArticleId).DoJoin(Base, JoinNm);
	return ArtSet;
}

TStr TNewsBase::GetConceptLabel(const uint64& ConceptId, const TStr& Lang) const
{
	AssertR(ConceptStore->IsRecId(ConceptId), "Invalid concept id specified");
	TStr ConceptUri = ConceptStore->GetRecNm(ConceptId);
	return ConceptLabel->GetConceptLabel(ConceptUri, Lang);
}

TStr TNewsBase::GetConceptLabel(const TStr& ConceptUri, const TStr& Lang) const
{ 
	return ConceptLabel->GetConceptLabel(ConceptUri, Lang);
}

//void TNewsBase::DeleteStoryInfo(const uint64& StoryId)
//{
//	if (!StoryStore->IsRecId(StoryId))
//		return;
//
//	const TRec StoryRec = StoryStore->GetRec(StoryId);
//	const PRecSet ArticleSet = StoryRec.DoJoin(Base, "hasArticle");
//	if (ArticleSet->GetRecs() > 0) {
//		ConsoleLog->OnStatusFmt("Warning: The story that we are deleting (%I64u) still has some articles (%d) assigned to it. This should not happen.", StoryId, ArticleSet->GetRecs());
//	}
//	
//	StoryStore->DelJoins("hasArticle", StoryId);
//	StoryStore->DelJoins("hasMedoid", StoryId);
//	StoryStore->DelJoins("hasConcept", StoryId);
//	StoryStore->DelJoins("hasLocation", StoryId);
//	StoryStore->DelJoins("hasCategory", StoryId);
//	
//	TRec EventRec = StoryRec.DoSingleJoin(Base, "hasEvent");
//	if (EventRec.IsDef()) {
//		PRecSet StorySet = EventRec.DoJoin(Base, "hasStory");
//		// if this story is the only one assigned to the event then we also remove all info about the event
//		if (StorySet->GetRecs() == 1)
//			DeleteEventInfo(EventRec.GetRecId());
//	}
//	else {
//		ConsoleLog->OnStatusFmt("Warning: The story (%I64u) is not assigned to an event. This should not happen.", StoryId);
//	}
//
//	// set empty fields to the article so that searching by title, body, date, ... won't find this article
//	PJsonVal StoryJson = TJsonVal::NewObj("URI", StoryStore->GetRecNm(StoryId));
//	StoryJson->AddToObj("Language", "");
//	StoryJson->AddToObj("AvgCosToCentroid", -1);
//	StoryJson->AddToObj("Variance", -1);
//	StoryJson->AddToObj("ChangeCount", 0);
//	StoryJson->AddToObj("ArticleCount", 0);
//	StoryJson->AddToObj("ArticleCountKey", "0");
//	StoryJson->AddToObj("CommonDates", "");
//	StoryJson->AddToObj("ExtractedDate", "");
//	StoryJson->AddToObj("ExtractedDateEnd", "");
//	StoryJson->AddToObj("AverageDate", "");
//	StoryJson->AddToObj("StoryDate", "");
//	StoryJson->AddToObj("StoryDateEnd", "");
//	/*StoryJson->AddToObj("Title", "");
//	StoryJson->AddToObj("Summary", "");*/
//	StoryStore->AddRec(StoryJson);
//}
//
//void TNewsBase::DeleteEventInfo(const uint64& EventId)
//{
//	if (!EventStore->IsRecId(EventId))
//		return;
//}

uint64 TNewsBase::AddStory(const TStr& Story, const TStr& Lang) {
	PJsonVal StoryVal = TJsonVal::NewObj("URI", Story);
	StoryVal->AddToObj("Language", Lang);
	TTm Tm = TTm::GetCurLocTm();
	StoryVal->AddToObj("LastUpdate", Tm.GetWebLogDateTimeStr(true, "T", false));
	return StoryStore->AddRec(StoryVal);
}

uint64 TNewsBase::AddStory(const uint64& ArticleId, const TStr& Story) {
	// make sure we get a valid article
	if (!ArticleStore->IsRecId(ArticleId)) 
		return TUInt64::Mx;
	
	// append given story id with language
	const TStr ArticleLang = ArticleStore->GetFieldNmStr(ArticleId, "Language");
	// check if known story or new
	const bool ExistingStory = StoryStore->IsRecNm(Story);
	const uint64 StoryId = ExistingStory ?	StoryStore->GetRecId(Story) : AddStory(Story, ArticleLang);
	if (!ExistingStory) {
		const TStr ArticleDate = ArticleStore->GetFieldNmStr(ArticleId, "Date");
		// use the article date as the date of the story when computing the counts
		NewsCounts->AddStoryCount(ArticleDate);
	}

	const TRec OldArticleStoryRec = ArticleStore->GetRec(ArticleId).DoSingleJoin(Base, "hasStory");
	// if the story to which the article is already assigned is the same just exit - don't do unnecesary deletejoins and addjoins
	if (StoryId == OldArticleStoryRec.GetRecId())
		return StoryId;
	// delete existing story <-> article
	ArticleStore->DelJoins("hasStory", ArticleId);
	// add new story <-> article join
	ArticleStore->AddJoin("hasStory", ArticleId, StoryId);
	return StoryId;
}

void TNewsBase::PutMedoid(const uint64& MedoidId, const TStr& Story) 
{
	if (!ArticleStore->IsRecId(MedoidId)) 
		return;
	const TStr ArticleLang = ArticleStore->GetFieldNmStr(MedoidId, "Language");
	// check if known story or new
	const uint64 StoryId = StoryStore->IsRecNm(Story) ?	StoryStore->GetRecId(Story) : AddStory(Story, ArticleLang);
	StoryStore->AddJoin("hasMedoid", StoryId, MedoidId);
}

// return medoid article id or the article that has the highest cosine similarity to center (if medoid not set)
uint64 TNewsBase::GetCentralArticle(const uint64& StoryId) const
{
	if (!StoryStore->IsRecId(StoryId))
		return TUInt64::Mx;
	uint64 ArticleId = StoryStore->GetFieldNmUInt64(StoryId, "hasMedoidId");
	if (ArticleStore->IsRecId(ArticleId))
		return ArticleId;
	PRecSet ArtSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
	ArtSet->SortByFq(false);
	if (ArtSet->GetRecs() > 0)
		return ArtSet->GetRecId(0);
	return TUInt64::Mx;
}

void TNewsBase::AddNewsfeedId(const uint64& ArticleId, const TStr& NewsfeedId) 
{
	EAssertR(ArticleId != TUInt64::Mx, "AddNewsfeedId: Invalid article id.");
	NewsfeedIdH.AddDat(NewsfeedId, ArticleId); 
}

void TNewsBase::RemoveNewsfeedId(const TStr& NewsfeedId)
{
	NewsfeedIdH.DelIfKey(NewsfeedId);
}

uint64 TNewsBase::GetArticleIdForNewsfeedId(const TStr& NewsfeedId) const 
{
	return NewsfeedIdH.IsKey(NewsfeedId) ? NewsfeedIdH.GetDat(NewsfeedId) : TUInt64::Mx;
}

// rebuild the source and concept frequency by making joins for each source and concept
// slow but needed if the file is missing or invalid
void TNewsBase::RebuildNewsCountsAndTrends()
{
	ConsoleLog->OnStatus("Recalculating frequencies for sources and concepts. This could take a while...");
	// clear all counts
	NewsCounts->ClearCounts();
	// clear all geo weights - they will be rebuilt when trigger will be called
	GeoBase->ClearWeights();
	TrendsManager->Init();

	// simulate adding new articles by calling the new article trigger
	const uint64 Articles = Base->GetStoreByStoreNm("Article")->GetRecs();
	PStoreIter Iter = ArticleStore->GetIter();
	while (Iter->Next()) {
		const uint64 ArticleId = Iter->GetRecId();
		if (ArticleId % 1000 == 0)
			ConsoleLog->OnStatusFmt("Processed %I64u/%I64u articles.\r", ArticleId, Articles);
		const TRec ArticleRec = ArticleStore->GetRec(ArticleId);
		NewArticleTrigger->OnAdd(ArticleRec);
	}

	const uint64 Stories = Base->GetStoreByStoreNm("Story")->GetRecs();
	Iter = StoryStore->GetIter();
	while (Iter->Next()) {
		const uint64 StoryId = Iter->GetRecId();
		if (StoryId % 1000 == 0)
			ConsoleLog->OnStatusFmt("Processed %I64u/%I64u stories.\r", StoryId, Stories);
		const TStr Date = StoryStore->GetFieldNmStr(StoryId, "AverageDate");
		NewsCounts->AddStoryCount(Date);
	}
	// rebuilding is usually quite slow so we want to make sure we save the recomputed data
	NewsCounts->Save();
	TrendsManager->Save();
	ConsoleLog->OnStatus("Recalculation finished.");
}

// compute intersection of two rec sets
// if the Dest recset is NULL then simply return the Source recset
void TNewsBase::IntersectRecSets(PRecSet& Dest, const PRecSet& Source, const bool& AddSourceFqs)
{
	if (Dest.Empty() || Source.Empty())
		Dest = Source;
	else {
		// intersect two rec sets. it's complicated only beacuse we want to add fqs also from each recset
		THash<TUInt64, TInt> DestRecIdFqH, FinalRecIdFqH;
		Dest->GetRecIdFqH(DestRecIdFqH);
		const int SourceRecs = Source->GetRecs();
		for (int RecN = 0; RecN < SourceRecs; RecN++) {
			uint64 RecId = Source->GetRecId(RecN);
			if (DestRecIdFqH.IsKey(RecId)) {
				const int DestFq = DestRecIdFqH.GetDat(RecId);
				FinalRecIdFqH.AddDat(RecId, DestFq + (AddSourceFqs ? Source->GetRecFq(RecN) : 0));
			}
		}
		TUInt64IntKdV FinalKdV; FinalRecIdFqH.GetKeyDatKdV(FinalKdV);
		Dest = TRecSet::New(Source->GetStore(), FinalKdV, true);
	}
}

void TNewsBase::RemoveRecSets(PRecSet& Dest, const PRecSet& RecsToRemove)
{
	if (Dest.Empty() || RecsToRemove.Empty() || RecsToRemove->GetRecs() == 0)
		return;
	THash<TUInt64, TInt> RecIdFqH;
	int Recs = Dest->GetRecs();
	for (int N = 0; N < Recs; N++)  {
		RecIdFqH.AddDat(Dest->GetRecId(N), Dest->GetRecFq(N));
	}
	Recs = RecsToRemove->GetRecs();
	for (int N = 0; N < Recs; N++) {
		RecIdFqH.DelIfKey(RecsToRemove->GetRecId(N));
	}
	TUInt64IntKdV RecIdFqV; RecIdFqH.GetKeyDatKdV(RecIdFqV);
	Dest = TRecSet::New(Dest->GetStore(), RecIdFqV, true);
}


bool TNewsBase::GetStoryFlag(const uint64& StoryId, const EnumStoryFlag& Flag) const
{
	EAssertR(StoryStore->IsRecId(StoryId), "StoryId is not a valid record id in StoryStore");
	const int StoryFlag = StoryStore->GetFieldNmInt(StoryId, "StoryFlags");
	return (StoryFlag & Flag) != 0;
}

void TNewsBase::SetStoryFlag(const uint64& StoryId, const EnumStoryFlag& Flag, const bool& Val)
{
	EAssertR(StoryStore->IsRecId(StoryId), "StoryId is not a valid record id in StoryStore");
	int StoryFlag = StoryStore->GetFieldNmInt(StoryId, "StoryFlags");
	if (Val) StoryFlag |= Flag;
	else StoryFlag &= ~Flag;
	StoryStore->SetFieldNmInt(StoryId, "StoryFlags", StoryFlag);
}

// helper method. sets FieldNm to false for all stories in StoryRecSet. If Lang != "" then only to the stories that are in language Lang
// info: when custom data is set to events we set the data to the top underlying story and the changes then propagate to the event
// if in time the quality of the stories changes, another story can become top story. If we would then re-edit the story info we would actually
// set the new event info to a different story and we would have multiple stories with CustomXXX set. This should not happen since it brings
// in ambiguity which custom fields to take. Before setting the custom values we therefore clear the CustomXXX field for stories in the event.
void TNewsBase::ClearFlagForStories(const PRecSet& StoryRecSet, const EnumStoryFlag& StoryFlag, const TStr& Lang)
{
	//PJsonVal StoryJson = TJsonVal::NewObj(FieldNm, TJsonVal::NewBool(false));
	for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
		uint64 StoryId = StoryRecSet->GetRecId(N);
		if (!StoryStore->IsRecId(StoryId))
			continue;
		if (Lang != "" && Lang != StoryStore->GetFieldNmStr(StoryId, "Language"))
			continue;
		SetStoryFlag(StoryId, StoryFlag, false);
	}
}

// does any of the specified stories have the specified flag set
// if Lang is set then just check those stories that match the language
bool TNewsBase::IsStoryFlagSet(const PRecSet& StoryRecSet, const EnumStoryFlag& StoryFlag, const TStr& Lang)
{
	for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
		const uint64 StoryId = StoryRecSet->GetRecId(N);
		if (!StoryStore->IsRecId(StoryId))
			continue;
		if (Lang != "" && Lang != StoryStore->GetFieldNmStr(StoryId, "Language"))
			continue;
		bool IsSet = GetStoryFlag(StoryId, StoryFlag);
		if (IsSet)
			return true;
	}
	return false;
}

// return the stored quality for a given story. 
int TNewsBase::GetStoryQuality(const uint64& StoryId)
{
	EAssertR(StoryStore->IsRecId(StoryId), "StoryId is not a valid record id in StoryStore");
	const int StoryFq = StoryStore->GetFieldNmInt(StoryId, "hasEventFq");
	return StoryFq;
}

// return the story id of the story with the highest score in the event. If Lang is provided we return the highest-scoring story in the given language
uint64 TNewsBase::GetBestStory(const uint64& EventId, const TStr& Lang) const
{
	if (!EventStore->IsRecId(EventId))
		return TUInt64::Mx;
	PRecSet StorySet = EventStore->GetRec(EventId).DoJoin(Base, "hasStory");
	StorySet->SortByFq(false);
	for (int N = 0; N < StorySet->GetRecs(); N++) {
		if (Lang == "" || StoryStore->GetFieldNmStr(StorySet->GetRecId(N), "Language") == Lang)
			return StorySet->GetRecId(N);
	}
	return TUInt64::Mx;
}

///////////////////////////////
// NewBase Fix

// find what's currently the correct uri for a concept uri
// if concept is not known return false
//bool TNewsBaseFix::GetConceptUri(const TStr& Url, TStr& Uri) const
//{
//	TStr Type; 
//	if (GetConceptType(Url, Type)) {
//		Uri = Type + ":" + Url;
//		return true;
//	}
//	return false;
//}


// remove the concept from all articles and index
void TNewsBaseFix::RemoveInvalidConcept(const TStr& ConceptUri)
{
	if (!NewsBase->ConceptStore->IsRecNm(ConceptUri))
		return;
	Notify->OnStatusFmt("Removing invalid concept %s", ConceptUri.CStr());
	uint64 ConceptRecId = NewsBase->ConceptStore->GetRecId(ConceptUri);
	// don't use the old concept in the autosuggest anymore
	NewsBase->NewsCounts->RemoveConcept(ConceptRecId);

	// remove old joins
	TRec ConceptRec(NewsBase->ConceptStore, ConceptRecId);
	PRecSet ArticleSet = ConceptRec.DoJoin(NewsBase->Base, "hasArticle");
	Notify->OnStatusFmt("\t%d articles were removed from concept %s", ArticleSet->GetRecs(), ConceptUri.CStr());
	for (int ArticleN = 0; ArticleN < ArticleSet->GetRecs(); ArticleN++)
		NewsBase->ConceptStore->DelJoin("hasArticle", ConceptRecId, ArticleSet->GetRecId(ArticleN));

	if (NewsBase->ConceptStore->IsJoinNm("hasStory")) {
		PRecSet StorySet = ConceptRec.DoJoin(NewsBase->Base, "hasStory");
		for (int StoryN = 0; StoryN < StorySet->GetRecs(); StoryN++)
			NewsBase->ConceptStore->DelJoin("hasStory", ConceptRecId, StorySet->GetRecId(StoryN));
	}

	if (NewsBase->ConceptStore->IsJoinNm("hasEvent")) {
		PRecSet EventSet = ConceptRec.DoJoin(NewsBase->Base, "hasEvent");
		for (int EventN = 0; EventN < EventSet->GetRecs(); EventN++)
			NewsBase->ConceptStore->DelJoin("hasEvent", ConceptRecId, EventSet->GetRecId(EventN));
	}

	if (NewsBase->ConceptStore->IsJoinNm("hasStoryForLocation")) {
		PRecSet StorySet = ConceptRec.DoJoin(NewsBase->Base, "hasStoryForLocation");
		for (int StoryN = 0; StoryN < StorySet->GetRecs(); StoryN++)
			NewsBase->ConceptStore->DelJoin("hasStoryForLocation", ConceptRecId, StorySet->GetRecId(StoryN));
	}

	if (NewsBase->ConceptStore->IsJoinNm("hasEventForLocation")) {
		PRecSet EventSet = ConceptRec.DoJoin(NewsBase->Base, "hasEventForLocation");
		for (int EventN = 0; EventN < EventSet->GetRecs(); EventN++)
			NewsBase->ConceptStore->DelJoin("hasEventForLocation", ConceptRecId, EventSet->GetRecId(EventN));
	}

	// remove the old concept from all the places in index
	NewsBase->ConceptStore->SetFieldNmStr(ConceptRecId, "Label", "");
	NewsBase->ConceptStore->SetFieldNmStr(ConceptRecId, "Type", "");
	NewsBase->ConceptStore->SetFieldNmStr(ConceptRecId, "Description", "");
}

// remove concept BadUri and replace it with (GoodUri, Label, Description)
void TNewsBaseFix::RewriteConceptUri(const TStr& BadUri, const TStr& GoodUri)
{
	if (!NewsBase->ConceptStore->IsRecNm(BadUri) || BadUri == GoodUri)
		return;
	if (!NewsBase->ConceptStore->IsRecNm(GoodUri)) {
		Notify->OnStatusFmt("ERROR: Unable to rewrite concept %s to %s. Target concept does not exist", BadUri.CStr(), GoodUri.CStr());
		return;
	}

	Notify->OnStatusFmt("Rewriting: %s --> %s", BadUri.CStr(), GoodUri.CStr());
	uint64 OldConceptRecId = NewsBase->ConceptStore->GetRecId(BadUri);
	uint64 NewConceptRecId = NewsBase->ConceptStore->GetRecId(GoodUri);

	// copy concept frequency to the new concept
	uint64 OldFq = NewsBase->NewsCounts->GetConceptFq(OldConceptRecId);
	NewsBase->NewsCounts->AddConceptFq(NewConceptRecId, OldFq);

	// find all articles that are annotated with the old concept and fix them
	const TRec& BadRec = NewsBase->ConceptStore->GetRec(BadUri);
	PRecSet ArticleSet = BadRec.DoJoin(NewsBase->Base, "hasArticle");
	Notify->OnStatusFmt("\t%d articles were reassigned to the new concept", ArticleSet->GetRecs());
	for (int ArticleN = 0; ArticleN < ArticleSet->GetRecs(); ArticleN++)
		NewsBase->ConceptStore->AddJoin("hasArticle", NewConceptRecId, ArticleSet->GetRecId(ArticleN), ArticleSet->GetRecFq(ArticleN));

	if (NewsBase->ConceptStore->IsJoinNm("hasStory")) {
		PRecSet StorySet = BadRec.DoJoin(NewsBase->Base, "hasStory");
		for (int StoryN = 0; StoryN < StorySet->GetRecs(); StoryN++)
			NewsBase->ConceptStore->AddJoin("hasStory", NewConceptRecId, StorySet->GetRecId(StoryN), StorySet->GetRecFq(StoryN));
	}

	if (NewsBase->ConceptStore->IsJoinNm("hasEvent")) {
		PRecSet EventSet = BadRec.DoJoin(NewsBase->Base, "hasEvent");
		for (int EventN = 0; EventN < EventSet->GetRecs(); EventN++)
			NewsBase->ConceptStore->AddJoin("hasEvent", NewConceptRecId, EventSet->GetRecId(EventN), EventSet->GetRecFq(EventN));
	}

	if (NewsBase->ConceptStore->IsJoinNm("hasStoryForLocation")) {
		PRecSet StorySet = BadRec.DoJoin(NewsBase->Base, "hasStoryForLocation");
		for (int StoryN = 0; StoryN < StorySet->GetRecs(); StoryN++)
			NewsBase->ConceptStore->AddJoin("hasStoryForLocation", NewConceptRecId, StorySet->GetRecId(StoryN), StorySet->GetRecFq(StoryN));
	}

	if (NewsBase->ConceptStore->IsJoinNm("hasEventForLocation")) {
		PRecSet EventSet = BadRec.DoJoin(NewsBase->Base, "hasEventForLocation");
		for (int EventN = 0; EventN < EventSet->GetRecs(); EventN++)
			NewsBase->ConceptStore->AddJoin("hasEventForLocation", NewConceptRecId, EventSet->GetRecId(EventN), EventSet->GetRecFq(EventN));
	}

	// remove everything from the old concept
	RemoveInvalidConcept(BadUri);
}

void TNewsBaseFix::ApplyConceptFixesOnNewsBase()
{
	Notify = TFileNotify::New("./Logs/ConceptFixLog.txt", false, true);

	// first rewrite wrong concept uris to correct ones (like /wiki/Republic_of_China to /wiki/China
	for (int KeyId = NewsFixDb->ConceptUriToInfoHH.FFirstKeyId(); NewsFixDb->ConceptUriToInfoHH.FNextKeyId(KeyId);) {
		const TStr Uri = NewsFixDb->ConceptUriToInfoHH.GetKey(KeyId);
		const THash<TStr, TStr> ConceptInfoH = NewsFixDb->ConceptUriToInfoHH[KeyId];
		if (ConceptInfoH.IsKey("uriTo")) {
			RewriteConceptUri(Uri, ConceptInfoH.GetDat("uriTo"));
		}
		if (ConceptInfoH.IsKey("type"))
			SetConceptType(Uri, ConceptInfoH.GetDat("type"));
		THash<TStr, TStr> ConceptLabelsH = NewsFixDb->GetConceptLabels(Uri);
		for (int LangId = ConceptLabelsH.FFirstKeyId(); ConceptLabelsH.FNextKeyId(LangId);) {
			const TStr Lang = ConceptLabelsH.GetKey(LangId);
			SetConceptLabel(Uri, Lang, ConceptLabelsH[LangId]);
		}
		if (ConceptInfoH.IsKey("img"))
			SetConceptImage(Uri, ConceptInfoH.GetDat("img"), true);
		if (ConceptInfoH.IsKey("disabled") && ConceptInfoH.GetDat("disabled") == "true")
			RemoveInvalidConcept(Uri);
	}

	Notify = ConsoleLog;
}

void TNewsBaseFix::ApplySourceFixesOnNewsBase() 
{
	Notify = TFileNotify::New("./Logs/SourceFixLog.txt", false, true);
	int Sources = NewsFixDb->GetSourceInfos();

	for (int N = 0; N < Sources; N++) {
		Notify->OnStatus("--- Applying source updates ---");
		PJsonVal SourceJson = NewsFixDb->GetSourceInfo(N);
		if (SourceJson.Empty() || SourceJson->IsObj() == false) {
			Notify->OnStatusFmt("Warning: Obtaining %d-th source info returned invalid/empty json", N);
			continue;
		}
		if (!SourceJson->IsObjKey("URI")) {
			Notify->OnStatusFmt("Warning: Skipping source info without URI: %s", TJsonVal::GetStrFromVal(SourceJson).CStr());
			continue;
		}
		
		const TStr Uri = SourceJson->GetObjStr("URI", "");
		Notify->OnStatusFmt("Updating source: %s", Uri.CStr());

		uint64 ExistingSourceId = NewsBase->SourceStore->IsRecNm(Uri) ? NewsBase->SourceStore->GetRecId(Uri) : TUInt64::Mx;
		if (ExistingSourceId != TUInt64::Mx && SourceJson->IsObjKey("hasImage") && SourceJson->GetObjKey("hasImage")->IsArr()) {
			NewsBase->SourceStore->DelJoins("hasImage", ExistingSourceId);	// remove old images first
			PJsonVal ImageArr = SourceJson->GetObjKey("hasImage");
			for (int N = 0; N < ImageArr->GetArrVals(); N++) {
				const uint64 ImageId = NewsBase->ImageStore->AddRec(ImageArr->GetArrVal(N));
				NewsBase->SourceStore->AddJoin("hasImage", ExistingSourceId, ImageId);
			}
		}
		if (ExistingSourceId != TUInt64::Mx && SourceJson->IsObjKey("hasThumbImage") && SourceJson->GetObjKey("hasThumbImage")->IsArr()) {
			NewsBase->SourceStore->DelJoins("hasThumbImage", ExistingSourceId);	// remove old thumb images first
			PJsonVal ImageArr = SourceJson->GetObjKey("hasThumbImage");
			for (int N = 0; N < ImageArr->GetArrVals(); N++) {
				const uint64 ImageId = NewsBase->ImageStore->AddRec(ImageArr->GetArrVal(N));
				NewsBase->SourceStore->AddJoin("hasThumbImage", ExistingSourceId, ImageId);
			}
		}
		if (ExistingSourceId != TUInt64::Mx && SourceJson->IsObjKey("hasTag") && SourceJson->GetObjKey("hasTag")->IsArr()) {
			NewsBase->SourceStore->DelJoins("hasTag", ExistingSourceId);	// remove old tags first
			PJsonVal TagArr = SourceJson->GetObjKey("hasTag");
			for (int N = 0; N < TagArr->GetArrVals(); N++) {
				const uint64 TagId = NewsBase->ImageStore->AddRec(TagArr->GetArrVal(N));
				NewsBase->SourceStore->AddJoin("hasTag", ExistingSourceId, TagId);
			}
		}
		// 
		// add/update the rest of the source content to the store

		// we do need title so try to get it
		if (!SourceJson->IsObjKey("Title") && ExistingSourceId != TUInt64::Mx)
			SourceJson->AddToObj("Title", NewsBase->SourceStore->GetFieldNmStr(ExistingSourceId, "Title"));
		if (!SourceJson->IsObjKey("Title"))
			SourceJson->AddToObj("Title", Uri);
		const uint64 SourceId = NewsBase->SourceStore->AddRec(SourceJson);

		if (!NewsBase->SourceStore->IsRecId(SourceId)) {
			Notify->OnStatusFmt("Warning: We added source info %s, but the source is not available in the source store", TJsonVal::GetStrFromVal(SourceJson).CStr());
			continue;
		}
		
		// if the source is disabled then remove all the articles for it
		if (ExistingSourceId != TUInt64::Mx && NewsBase->IsSourceEnabled(Uri) == false) {
			PRecSet ArticleSet = NewsBase->SourceStore->GetRec(SourceId).DoJoin(NewsBase->Base, "hasArticle");
			if (ArticleSet->GetRecs() > 0)
				Notify->OnStatusFmt("Source is disabled so we are removing %d articles from the source", ArticleSet->GetRecs());
			for (int ArtN = 0; ArtN < ArticleSet->GetRecs(); ArtN++) {
				if (ArtN % 10 == 0)
					Notify->OnStatusFmt("Removing article %d/%d", ArtN, ArticleSet->GetRecs());
				NewsBase->DeleteArticleInfo(ArticleSet->GetRecId(ArtN));
			}
		}
	}
	Notify = ConsoleLog;
}

bool TNewsBaseFix::GetConceptType(const TStr& ConceptUri, TStr& Type) const
{
	if (NewsFixDb->GetConceptType(ConceptUri, Type)) {
		return true;
	}
	for (int TypeN = 0; TypeN < NewsConst.ConceptTypeV.Len(); TypeN++) {
		if (NewsBase->ConceptStore->IsRecNm(ConceptUri)) {
			uint64 RecId = NewsBase->ConceptStore->GetRecId(ConceptUri);
			Type = NewsBase->ConceptStore->GetFieldNmStr(RecId, "Type");
			return true;
		}
	}
	return false;
}

PJsonVal TNewsBaseFix::SetConceptImage(const TStr& ConceptUri, const TStr& ImageUrl, const bool& SetAsPrimary, const bool& SetPrimaryIfFirst)
{
	if (!NewsBase->ConceptStore->IsRecNm(ConceptUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Concept uri %s is not a known concept. It is possible that a rewrite rule that hasn't been applied yet. Unable to set it's image.", ConceptUri.CStr()));
	if (ImageUrl == "")
		return TJsonVal::NewObj("error", TStr("Empty string is not a valid url for an image."));

	// create a record for the image
	const uint64 ImageRecId = NewsBase->ImageStore->AddRec(TJsonVal::NewObj("URI", ImageUrl));

	// create a join from concept -> image
	const uint64 ConceptId = NewsBase->ConceptStore->GetRecId(ConceptUri);
	PRecSet ExistingImages = NewsBase->ConceptStore->GetRec(ConceptId).DoJoin(NewsBase->Base, "hasImage");
	NewsBase->ConceptStore->AddJoin("hasImage", ConceptId, ImageRecId, 1);	// concept -> image
	if (SetAsPrimary || (SetPrimaryIfFirst && ExistingImages->GetRecs() == 0))	// if this is the first set image then set it as primary
		NewsBase->ConceptStore->AddJoin("hasPrimaryImage", ConceptId, ImageRecId);

	return TJsonVal::NewObj("info", TStr("Concept image was set."));
}

PJsonVal TNewsBaseFix::SetConceptDescription(const TStr& ConceptUri, const TStr& Description)
{
	if (!NewsBase->ConceptStore->IsRecNm(ConceptUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Concept uri %s is not a known concept. It is possible that a rewrite rule that hasn't been applied yet. Unable to set it's description.", ConceptUri.CStr()));
	const uint64 ConceptId = NewsBase->ConceptStore->GetRecId(ConceptUri);
	NewsBase->ConceptStore->SetFieldNmStr(ConceptId, "Description", Description);

	return TJsonVal::NewObj("info", TStr("Concept description was set."));
}

void TNewsBaseFix::SetConceptType(const TStr& ConceptUri, const TStr& Type)
{
	if (!NewsBase->ConceptStore->IsRecNm(ConceptUri))
		return;
	const uint64 ConceptId = NewsBase->ConceptStore->GetRecId(ConceptUri);
	NewsBase->ConceptStore->SetFieldNmStr(ConceptId, "Type", Type);
}

void TNewsBaseFix::SetConceptLabel(const TStr& ConceptUri, const TStr& Lang, const TStr& NewLabel)
{
	if (!NewsBase->ConceptStore->IsRecNm(ConceptUri))
		return;
	const TStr ExistingLabel = NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, Lang);
	Notify->OnStatusFmt("Changing %s label for concept %s. Changing %s --> %s", Lang.CStr(), ConceptUri.CStr(), ExistingLabel.CStr(), NewLabel.CStr());
	NewsBase->ConceptLabel->SetConceptLabel(ConceptUri, Lang, NewLabel);
}

void TNewsBaseFix::SetMissingLabelsFromWikiUri(const TStr& Lang)
{
	for (uint64 N = 0; N < NewsBase->ConceptStore->GetRecs(); N++) {
		const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(N);
		if (!NewsBase->ConceptLabel->HasConceptLabel(ConceptUri, Lang))  {
			TStr Prefix, NewLabel; ConceptUri.SplitOnLastCh(Prefix, '/', NewLabel);
			if (NewLabel.Len() == 0)
				continue;
			NewLabel.ChangeChAll('_', ' ');
			Notify->OnStatusFmt("Setting label for %s to %s", ConceptUri.CStr(), NewLabel.CStr());
			NewsBase->ConceptLabel->SetConceptLabel(ConceptUri, Lang, NewLabel);
		}
	}
}

///////////////////////////////
// New Article Trigger

// updates news counts 
// remembers the Newsfeed articleId <-> miner article id
void TNewsArticleTrigger::OnAdd(const TRec& ArticleRec) 
{
	EAssert(ArticleRec.GetStoreId() == NewsBase->ArticleStoreId);
	const PStore Store = NewsBase->ArticleStore;
	const PStore ConceptStore = NewsBase->ConceptStore;
	const uint64 RecId = ArticleRec.GetRecId();
	
	const TStr NewsfeedId = Store->GetFieldNmStr(ArticleRec.GetRecId(), "NewsfeedId");
	NewsBase->AddNewsfeedId(ArticleRec.GetRecId(), NewsfeedId);

	const TStr Date = Store->GetFieldNmStr(ArticleRec.GetRecId(), "Date");
	TTm Tm; ArticleRec.GetStore()->GetFieldNmTm(ArticleRec.GetRecId(), "DateTime", Tm);
	const int MinWgt = NewsBase->TrendsManager->MinConceptWgt;
	TConceptIdV ConceptV;
	// update concept counts
	PRecSet ConceptSet = ArticleRec.DoJoin(NewsBase->Base, "hasConcept");
	for (int ConceptN = 0; ConceptN < ConceptSet->GetRecs(); ConceptN++) {
		const uint64 ConceptRecId = ConceptSet->GetRecId(ConceptN);
		NewsBase->NewsCounts->AddConceptFq(ConceptRecId);	// increase fq for concept
		// increase weights for the locations that we found in the article
		if (ConceptStore->GetFieldNmStr(ConceptRecId, "Type") == NewsConst.TypeLocation)
			NewsBase->GeoBase->IncreaseWgt(ConceptStore->GetRecNm(ConceptRecId));

		if (ConceptSet->GetRecFq(ConceptN) >= MinWgt)
			ConceptV.Add((int) ConceptSet->GetRecId(ConceptN));
	}

	// update the trends for the concepts in the article
	// skip this if we know the article is a duplicate (might happen if we load data from the dumps)
	bool ArticleIsDuplicate = (NewsBase->ArticleStore->GetFieldNmUInt64(RecId, "isDuplicateOfId") != TUInt64::Mx);
	if (ConceptV.Len() > 0 && ArticleIsDuplicate == false)
		NewsBase->TrendsManager->AddConceptV(Tm, ConceptV);

	// update source counts
	TRec SourceRec = ArticleRec.DoSingleJoin(NewsBase->Base, "hasSource");
	const uint64 SourceRecId = SourceRec.GetRecId();
	NewsBase->NewsCounts->AddSourceFq(SourceRecId);

	// add article count
	NewsBase->NewsCounts->AddArticleCount();
	// increase annotated counts if necessary
	if (ConceptSet->GetRecs() > 0) {
		NewsBase->NewsCounts->AddAnnArticleCount();
		NewsBase->NewsCounts->AddAnnArticleCount(Date);
	}

	// add article count for date
	NewsBase->NewsCounts->AddArticleCount(Date);
}

// called when the articles is deleted from the ArticleStore (when too old)
void TNewsArticleTrigger::OnDelete(const TRec& ArticleRec) 
{
	const PStore Store = NewsBase->ArticleStore;
	const TStr Date = Store->GetFieldNmStr(ArticleRec.GetRecId(), "Date");
	const TStr Lang = Store->GetFieldNmStr(ArticleRec.GetRecId(), "Language");
	
	// if we've set date and lang to empty string then the article was already removed (maybe by disabling a source)
	// in that case we should not call DeleteArticleInfo again since it would decrease the counts twice
	if (Date == "" && Lang == "")
		return;
	
	if (NewsBase != NULL)
		NewsBase->DeleteArticleInfo(ArticleRec.GetRecId());
}

// new story trigger is used to create a new event whenever a new story is created
// for the new event we create hasStory join to the story and also mark the 
// story as primary story for the event (using the hasPrimaryStory join)

void TNewStoryTrigger::OnAdd(const TRec& Rec)
{
	// assign the story to this event
	const TStr StoryName = Rec.GetRecNm();

	// create a new event and assign cluster to it
	TUInt64 EventCount = NewsBase->EventStore->GetRecs();
	PJsonVal EventJson = TJsonVal::NewObj("URI", EventCount.GetStr());

	uint64 EventRecId = NewsBase->EventStore->AddRec(EventJson);
	NewsBase->StoryStore->AddJoin("hasEvent", Rec.GetRecId(), EventRecId, 100);
	NewsBase->StoryStore->AddJoin("hasPrimaryEvent", Rec.GetRecId(), EventRecId, 100);

	if (InfoLogFile != NULL)
		InfoLogFile->NotifyVerboseFmt(3, "Created a new event (id %I64u) for a new story (id %I64u, %s)", EventRecId, Rec.GetRecId(), StoryName.CStr());
}