/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include <base.h>
#include "NewsConst.h"

TNewsConst::TNewsConst()
{
	Eps = 0.00001;

	TypeLocation = "loc";
	TypePerson = "person";
	TypeOrganization = "org";
	TypeConceptFolder = "topicPage";
	TypeKeyword = "keyword";
	TypeWiki = "wiki";

	ConceptFolderUriPrefix = "topic-page";

	ClassTypeCustom = "custom";
	ClassTypeDBPedia = "dbpedia";

	EntitySet.AddKey(TypePerson);
	EntitySet.AddKey(TypeOrganization);
	EntitySet.AddKey(TypeLocation);
	
	KeywordSet.AddKey(TypeKeyword);
	
	WikiSet.AddKey(TypeWiki);

	ConceptFolderSet.AddKey(TypeConceptFolder);

	TStrV Vec; 
	EntitySet.GetKeyV(Vec); ConceptSet.AddKeyV(Vec);
	KeywordSet.GetKeyV(Vec); ConceptSet.AddKeyV(Vec);
	WikiSet.GetKeyV(Vec); ConceptSet.AddKeyV(Vec);
	ConceptFolderSet.GetKeyV(Vec); ConceptSet.AddKeyV(Vec);
	
	for (int KeyId = EntitySet.FFirstKeyId(); EntitySet.FNextKeyId(KeyId); )
		ConceptTypeV.Add(EntitySet.GetKey(KeyId)); 
	for (int KeyId = KeywordSet.FFirstKeyId(); KeywordSet.FNextKeyId(KeyId); )
		ConceptTypeV.Add(KeywordSet.GetKey(KeyId)); 
	for (int KeyId = WikiSet.FFirstKeyId(); WikiSet.FNextKeyId(KeyId); )
		ConceptTypeV.Add(WikiSet.GetKey(KeyId)); 
	for (int KeyId = ConceptFolderSet.FFirstKeyId(); ConceptFolderSet.FNextKeyId(KeyId);)
		ConceptTypeV.Add(ConceptFolderSet.GetKey(KeyId));

	LangCoreV.Add("eng");
	LangCoreV.Add("spa");
	LangCoreV.Add("zho");
	LangCoreV.Add("deu");
}

void TNewsConst::SetLangInfo(const PJsonVal& LangInfo)
{
	if (LangInfo->IsObjKey("coreLangs") && LangInfo->GetObjKey("coreLangs")->IsArr()) {
		LangCoreV.Clr();
		LangInfo->GetObjStrV("coreLangs", LangCoreV);
	}

	if (LangInfo->IsObjKey("similarLangs") && LangInfo->GetObjKey("similarLangs")->IsArr()) {
		LangSimilarV.Clr();
		LangInfo->GetObjStrV("similarLangs", LangSimilarV);
	}
	
	if (LangInfo->IsObjKey("allLangs") && LangInfo->GetObjKey("allLangs")->IsArr()) {
		LangSimilarV.Clr();
		LangInfo->GetObjStrV("allLangs", LangAllV);
	}
}

// is the concept person, location or organization?
bool TNewsConst::IsEntity(const TStr& Type) const
{
	return EntitySet.IsKey(Type);
}

bool TNewsConst::IsKeywordOrWiki(const TStr& Type) const
{
	return (KeywordSet.IsKey(Type) || WikiSet.IsKey(Type));
}

EnumConceptType TNewsConst::GetValidTypes(const TStrV& TypeV, const EnumConceptType& TypeForEmptyTypeV) const
{
	EnumConceptType ValidTypes = ctNone;
	if (TypeV.Len() == 0)
		ValidTypes = TypeForEmptyTypeV;
	else {
		if (TypeV.IsIn("concepts")) ValidTypes = EnumConceptType(ValidTypes | ctEntities | ctWiki | ctKeyword);
		if (TypeV.IsIn("entities")) ValidTypes = EnumConceptType(ValidTypes | ctEntities);
		
		if (TypeV.IsIn("person")) ValidTypes = EnumConceptType(ValidTypes | ctPerson);
		if (TypeV.IsIn("org")) ValidTypes = EnumConceptType(ValidTypes | ctOrg);
		if (TypeV.IsIn("loc")) ValidTypes = EnumConceptType(ValidTypes | ctLoc);

		if (TypeV.IsIn("keyword")) ValidTypes = EnumConceptType(ValidTypes | ctKeyword);
		if (TypeV.IsIn("wiki")) ValidTypes = EnumConceptType(ValidTypes | ctWiki);
		if (TypeV.IsIn("conceptFolder")) ValidTypes = EnumConceptType(ValidTypes | ctConceptFolder);
	}
	return ValidTypes;
}

bool TNewsConst::IsValidType(const TStr& Type, const EnumConceptType& ValidTypes) const
{
	if (ValidTypes == ctAll)
		return true;
	if ((Type == TypePerson && ValidTypes & ctPerson) ||
		(Type == TypeLocation && ValidTypes & ctLoc) ||
		(Type == TypeOrganization && ValidTypes & ctOrg) ||

		(KeywordSet.IsKey(Type) && ValidTypes & ctKeyword) ||
		(WikiSet.IsKey(Type) && ValidTypes & ctWiki) ||
		(ConceptFolderSet.IsKey(Type) && ValidTypes & ctConceptFolder))
		return true;
	else
		return false;
}

TStr TNewsConst::GetTypeLabel(const TStr& Type) const
{
	if (Type == TypeLocation) return "Location";
	if (Type == TypePerson) return "Person";
	if (Type == TypeOrganization) return "Organization";
	if (Type == TypeConceptFolder) return "Concept Folder";
	if (Type == TypeWiki) return "Wiki";
	if (Type == TypeKeyword) return "Keyword";
	return Type;
}

TStr TNewsConst::GetPluralTypeLabel(const TStr& Type) const
{
	if (Type == TypeLocation) return "Locations";
	if (Type == TypePerson) return "People";
	if (Type == TypeOrganization) return "Organizations";
	if (Type == TypeConceptFolder) return "Concept Folders";
	if (Type == TypeWiki) return "Wikis";
	if (Type == TypeKeyword) return "Keywords";
	return Type;
}

TStr TNewsConst::GetJsonNmForLang(const TStr& Lang) const
{
	return "label" + Lang.GetCap();
}

TStr TNewsConst::GetLangForJsonNm(const TStr& JsonNm) const
{
	if (JsonNm.IsPrefix("label"))
		return JsonNm.GetSubStr(5);
	return JsonNm;
}

bool TNewsConst::IsJsonNmForLang(const TStr& JsonNm) const
{
	return JsonNm.IsPrefix("label") && JsonNm.Len() == 5 + 3;
}

TNewsConst NewsConst;
PNotify ConsoleLog = TStdNotify::New();