/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "NewsFixDb.h"
#include "NewsConst.h"

// ///////////////////////
// TNewsFix

#define CONCEPTINFO_JSON "conceptInfo.json"
#define CONCEPTINFO_BIN "conceptInfo.bin"
#define SOURCEINFO_JSON "sourceInfo.json"

TNewsFixDb::TNewsFixDb()
{
	EditedConceptData = false;
	EditedSourceData = false;
}

TNewsFixDb::~TNewsFixDb()
{
	if (EditedConceptData || !TFile::Exists(Path + CONCEPTINFO_BIN))
		SaveConceptFixes();
	if (EditedSourceData)
		SaveSourceFixes();
}

void TNewsFixDb::LoadConceptFixes() {
	if (Path == "")
		TNotify::StdNotify->OnStatus("WARNING: There is no path specified when loading the concept fixes. Call SetDbsPath() first!");
	TNotify::StdNotify->OnStatus("Loading news fixes...");
	
	ConceptUriToInfoHH.Clr();
	
	enum FileToUseEnum { undef, bin, json };
	FileToUseEnum FileToUse = FileToUseEnum::undef;
	if (TFile::Exists(Path + CONCEPTINFO_JSON))
		FileToUse = FileToUseEnum::json;
	if (TFile::Exists(Path + CONCEPTINFO_BIN))
		FileToUse = FileToUseEnum::bin;
	
	// check if the txt file was edited manually after the bin was last saved. if yes, then ignore bin and load from json
	if (TFile::Exists(Path + CONCEPTINFO_BIN) && TFile::Exists(Path + CONCEPTINFO_JSON) &&
		TFile::GetLastWriteTm(Path + CONCEPTINFO_JSON) > TFile::GetLastWriteTm(Path + CONCEPTINFO_BIN))
		FileToUse = FileToUseEnum::json;

	bool BinLoadedOk = true;
	if (FileToUse == FileToUseEnum::bin) {
		BinLoadedOk = LoadConceptInfoBin();
	}
	// if we failed to read the bin file then also load the json file
	if (FileToUse == FileToUseEnum::json || BinLoadedOk == false) {
		TNotify::StdNotify->OnStatus("Reading entity types from 'conceptInfo.json'");
		TFIn FIn(Path + CONCEPTINFO_JSON);
		TStr Ln;
		while (FIn.GetNextLn(Ln)) {
			try {
				PJsonVal Json = TJsonVal::GetValFromStr(Ln);
				if (!Json->IsObj() || !Json->IsObjKey("uri"))
					continue;

				bool Disabled = Json->GetObjBool("disabled", false);
				TStr Type = Json->GetObjStr("type", "");
				TStr Uri = Json->GetObjStr("uri", "");
				TStr UriTo = Json->GetObjStr("uriTo", "");
				TStr Img = Json->GetObjStr("img", "");

				THash<TStr, TStr> ConceptInfoH;
				if (Type != "")
					ConceptInfoH.AddDat("type", Type);
				if (Disabled)
					ConceptInfoH.AddDat("disabled", "true");
				if (UriTo != "") {
					ConceptInfoH.AddDat("uriTo", UriTo);
					ConceptInfoH.AddDat("disabled", "true");
				}
				if (Img != "")
					ConceptInfoH.AddDat("img", Img);

				THash<TStr, TStr> LangToLabelH;
				for (int N = 0; N < Json->GetObjKeys(); N++) {
					const TStr Key = Json->GetObjKey(N);
					if (NewsConst.IsJsonNmForLang(Key)) {
						ConceptInfoH.AddDat(Key, Json->GetObjStr(Key));
					}
				}
				if (ConceptInfoH.Len() >= 1)
					ConceptUriToInfoHH.AddDat(Uri, ConceptInfoH);
			}
			catch (PExcept Ex) {
				TNotify::StdNotify->OnStatusFmt("Exception while loading concept info from conceptInfo.json. Error: %s", Ex->GetMsgStr().CStr());
			}
		}
	}
	
	EditedConceptData = false;
	TNotify::StdNotify->OnStatus("Loading concept fixes finished.");
}

bool TNewsFixDb::LoadConceptInfoBin()
{
	try {
		TNotify::StdNotify->OnStatus("Reading entity types from 'conceptInfo.bin'");
		TFIn FIn(Path + CONCEPTINFO_BIN);
		ConceptUriToInfoHH.Load(FIn);
		return true;
	}
	catch (PExcept Ex) {
		TNotify::StdNotify->OnStatusFmt("Exception while loading concept info from conceptInfo.bin. Error: %s", Ex->GetMsgStr().CStr());
		return false;
	}
}

void TNewsFixDb::LoadSourceFixes()
{
	if (TFile::Exists(Path + SOURCEINFO_JSON)) {
		TFIn FIn(Path + SOURCEINFO_JSON);
		TStr Ln;
		while (FIn.GetNextLn(Ln)) {
			try {
				PJsonVal Json = TJsonVal::GetValFromStr(Ln);
				if (!Json->IsObj() || !Json->IsObjKey("URI"))
					continue;
				SourceUriToJsonH.AddDat(Json->GetObjStr("URI"), Json);
			}
			catch (PExcept Ex) {
				TNotify::StdNotify->OnStatusFmt("Exception while loading source info from sourceInfo.json. Error: %s", Ex->GetMsgStr().CStr());
			}
		}
		TNotify::StdNotify->OnStatus("Loading source fixes finished.");
	}
	
	EditedSourceData = false;
}

void TNewsFixDb::SaveConceptFixes()
{
	// save bin file for concepts
	ConceptUriToInfoHH.Defrag();
	ConceptUriToInfoHH.SortByKey();
	
	// save json file for concepts
	TFOut FOut3(Path + CONCEPTINFO_JSON);
	for (int ConceptKeyId = ConceptUriToInfoHH.FFirstKeyId(); ConceptUriToInfoHH.FNextKeyId(ConceptKeyId);) {
		const TStr Uri = ConceptUriToInfoHH.GetKey(ConceptKeyId);
		THash<TStr, TStr>& ConceptInfoH = ConceptUriToInfoHH[ConceptKeyId];
		ConceptInfoH.Defrag();
		
		PJsonVal Json = TJsonVal::NewObj("uri", Uri);
		for (int KeyId = ConceptInfoH.FFirstKeyId(); ConceptInfoH.FNextKeyId(KeyId);) {
			const TStr Key = ConceptInfoH.GetKey(KeyId);
			const TStr Val = ConceptInfoH[KeyId];
			if (Key == "disabled") {
				if (Val == "true")
					Json->AddToObj("disabled", true);
			}
			else
				Json->AddToObj(Key, Val);
		}
		FOut3.PutStr(Json->SaveStr() + "\n");
	}
	
	TFOut FOut(Path + CONCEPTINFO_BIN);
	ConceptUriToInfoHH.Save(FOut);

	EditedConceptData = false;
}

void TNewsFixDb::SaveSourceFixes()
{
	// save edited sources info
	SourceUriToJsonH.Defrag();
	//SourceUriToJsonH.SortByKey();	// cannot sort by key since Pjsonval is not sortable 
	TStrV SourceUriV; SourceUriToJsonH.GetKeyV(SourceUriV);
	SourceUriV.Sort();
	TFOut FOut(Path + SOURCEINFO_JSON);
	for (int N = 0; N < SourceUriV.Len(); N++) {
		const TStr SourceUri = SourceUriV[N];
		PJsonVal Json = SourceUriToJsonH.GetDat(SourceUri);
		const TStr JsonStr = TJsonVal::GetStrFromVal(Json);
		FOut.PutStr(JsonStr + "\n");
	}
	EditedSourceData = false;
}

// if we have information about the concept then find the corrected concept uri, concept type and label
// parameter lang is used to determine if we have the corrected label
void TNewsFixDb::CorrectConceptInfo(TStr& ConceptUri, TStr& ConceptType, TStr& ConceptLabel, const TStr& Lang) const
{
	// first try to see if we need to find the corrected concept uri
	if (!ConceptUriToInfoHH.IsKey(ConceptUri))
		return;
	const THash<TStr, TStr>& ConceptInfoH = ConceptUriToInfoHH.GetDat(ConceptUri);
	if (ConceptInfoH.IsKey("uriTo"))
		ConceptUri = ConceptInfoH.GetDat("uriTo");

	// fix the concept type, if we have info
	if (ConceptInfoH.IsKey("type"))
		ConceptType = ConceptInfoH.GetDat("type");
	// fix the label
	const TStr JsonNm = NewsConst.GetJsonNmForLang(Lang);
	if (ConceptInfoH.IsKey(JsonNm))
		ConceptLabel = ConceptInfoH.GetDat(JsonNm);
}

void TNewsFixDb::SetConceptIgnore(const TStr& ConceptUri)
{
	ConceptUriToInfoHH.AddDat(ConceptUri).AddDat("disabled", "true");
	EditedConceptData = true;
}

bool TNewsFixDb::HasConceptIgnore(const TStr& ConceptUri) const { 
	return ConceptUriToInfoHH.IsKey(ConceptUri) && ConceptUriToInfoHH.GetDat(ConceptUri).IsKey("disabled");
}

void TNewsFixDb::RemoveConceptIgnore(const TStr& ConceptUri)
{
	ConceptUriToInfoHH.AddDat(ConceptUri).DelIfKey("disabled");
	EditedConceptData = true;
}

void TNewsFixDb::SetConceptRewrite(const TStr& FromUri, const TStr& ToUri)
{
	ConceptUriToInfoHH.AddDat(FromUri).AddDat("uriTo", ToUri);
	// ignore the from uri - remove it from future suggestions
	ConceptUriToInfoHH.AddDat(FromUri).AddDat("disabled", "true");
	// make sure the to uri is not ignored
	ConceptUriToInfoHH.AddDat(ToUri).DelIfKey("disabled");
	EditedConceptData = true;
}

bool TNewsFixDb::GetConceptRewrite(TStr& Uri) const 
{ 
	if (ConceptUriToInfoHH.IsKey(Uri) && ConceptUriToInfoHH.GetDat(Uri).IsKey("uriTo")) {
		Uri = ConceptUriToInfoHH.GetDat(Uri).GetDat("uriTo");
		return true;
	}
	return false;
}

bool TNewsFixDb::GetConceptRewrite(const TStr& FromUri, TStr& ToUri) const 
{ 
	if (ConceptUriToInfoHH.IsKey(FromUri) && ConceptUriToInfoHH.GetDat(FromUri).IsKey("uriTo")) {
		ToUri = ConceptUriToInfoHH.GetDat(FromUri).GetDat("uriTo");
		return true;
	}
	return false;
}

bool TNewsFixDb::HasConceptRewrite(const TStr& Uri) const { 
	return (ConceptUriToInfoHH.IsKey(Uri) && ConceptUriToInfoHH.GetDat(Uri).IsKey("uriTo"));
}

void TNewsFixDb::SetConceptType(const TStr& Uri, const TStr& Type)
{
	if (! NewsConst.ConceptTypeV.IsIn(Type)) {
		TNotify::StdNotify->OnStatusFmt("'%s' is not a valid value for concept type. Unable to set concept type.", Type.CStr());
		return;
	}
	ConceptUriToInfoHH.AddDat(Uri).AddDat("type", Type);
	EditedConceptData = true;
}

bool TNewsFixDb::HasConceptType(const TStr& Uri) const 
{ 
	return (ConceptUriToInfoHH.IsKey(Uri) && ConceptUriToInfoHH.GetDat(Uri).IsKey("type"));
}

bool TNewsFixDb::GetConceptType(const TStr& Uri, TStr& Type) const 
{ 
	if (ConceptUriToInfoHH.IsKey(Uri) && ConceptUriToInfoHH.GetDat(Uri).IsKey("type")) {
		Type = ConceptUriToInfoHH.GetDat(Uri).GetDat("type");
		return true;
	}
	return false;
}

void TNewsFixDb::SetConceptLabel(const TStr& Uri, const TStr& Lang, const TStr& Label)
{
	const TStr JsonNm = NewsConst.GetJsonNmForLang(Lang);
	// make sure we don't keep any tabs in the label since we serialize this in tab separated file
	TStr CleanLabel = Label; CleanLabel.ChangeChAll('\t', ' ');
	ConceptUriToInfoHH.AddDat(Uri).AddDat(JsonNm, CleanLabel);
	EditedConceptData = true;
}

bool TNewsFixDb::GetConceptLabel(const TStr& Uri, const TStr& Lang, TStr& Label) const
{
	const TStr JsonNm = NewsConst.GetJsonNmForLang(Lang);
	if (ConceptUriToInfoHH.IsKey(Uri) && ConceptUriToInfoHH.GetDat(Uri).IsKey(JsonNm)) {
		Label = ConceptUriToInfoHH.GetDat(Uri).GetDat(JsonNm);
		return true;
	}
	return false;
}

void TNewsFixDb::RemoveConceptLabel(const TStr& Uri, const TStr& Lang)
{
	if (ConceptUriToInfoHH.IsKey(Uri))
		ConceptUriToInfoHH.GetDat(Uri).DelIfKey(Lang);
	EditedConceptData = true;
}

const THash<TStr, TStr> TNewsFixDb::GetConceptLabels(const TStr& Uri) const
{
	THash<TStr, TStr> ConceptLabelH;
	if (ConceptUriToInfoHH.IsKey(Uri)) {
		const THash<TStr, TStr>& ConceptInfoH = ConceptUriToInfoHH.GetDat(Uri);
		for (int KeyId = ConceptInfoH.FFirstKeyId(); ConceptInfoH.FNextKeyId(KeyId);) {
			const TStr Key = ConceptInfoH.GetKey(KeyId);
			if (NewsConst.IsJsonNmForLang(Key)) {
				const TStr Lang = NewsConst.GetLangForJsonNm(Key);
				ConceptLabelH.AddDat(Lang, ConceptInfoH[KeyId]);
			}
		}
	}
	return ConceptLabelH;
}

void TNewsFixDb::SetConceptImage(const TStr& Uri, const TStr& Img)
{
	ConceptUriToInfoHH.AddDat(Uri).AddDat("img", Img);
	EditedConceptData = true;
}

bool TNewsFixDb::GetConceptImage(const TStr& Uri, TStr& Img) const
{
	if (ConceptUriToInfoHH.IsKey(Uri) && ConceptUriToInfoHH.GetDat(Uri).IsKey("img")) {
		Img = ConceptUriToInfoHH.GetDat(Uri).GetDat("img");
		return true;
	}
	return false;
}

//
// source methods
// 
PJsonVal TNewsFixDb::GetSourceJson(const TStr& Uri)
{
	if (SourceUriToJsonH.IsKey(Uri))
		return SourceUriToJsonH.GetDat(Uri);
	PJsonVal Json = TJsonVal::NewObj("URI", Uri);
	SourceUriToJsonH.AddDat(Uri, Json);
	return Json;
}

void TNewsFixDb::SetSourceTitle(const TStr& Uri, const TStr& Title)
{
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("Title", Title);
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceDescription(const TStr& Uri, const TStr& Desc)
{
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("Description", Desc);
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceEnabled(const TStr& Uri, const bool& Enabled)
{
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("Enabled", Enabled);
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceImportance(const TStr& Uri, const int& Importance)
{
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("Importance", Importance);
	EditedSourceData = true;
}


void TNewsFixDb::SetSourceCity(const TStr& Uri, const TStr& City)
{
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("City", City);
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceCountry(const TStr& Uri, const TStr& Country)
{
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("Country", Country);
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceLatLong(const TStr& Uri, const double& Lat, const double& Long)
{
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("Location", Lat, Long);
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceTags(const TStr& Uri, const TStrV& TagV)
{
	TJsonValV JsonTagV;
	for (int N = 0; N < TagV.Len(); N++) {
		PJsonVal Tag = TJsonVal::NewObj("URI", TagV[N]);
		JsonTagV.Add(Tag);
	}
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("hasTag", TJsonVal::NewArr(JsonTagV));
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceImage(const TStr& Uri, const TStr& ImageUrl)
{
	PJsonVal ImageJson = TJsonVal::NewObj("URI", ImageUrl);
	TJsonValV ImageJsonV; ImageJsonV.Add(ImageJson);
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("hasImage", TJsonVal::NewArr(ImageJsonV));
	EditedSourceData = true;
}

void TNewsFixDb::SetSourceThumbImage(const TStr& Uri, const TStr& ImageUrl)
{
	PJsonVal ImageJson = TJsonVal::NewObj("URI", ImageUrl);
	TJsonValV ImageJsonV; ImageJsonV.Add(ImageJson);
	PJsonVal Json = GetSourceJson(Uri);
	Json->AddToObj("hasThumbImage", TJsonVal::NewArr(ImageJsonV));
	EditedSourceData = true;
}
