/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "NewsLoad.h"
//#include "NewsConst.h"

///////////////////////////////
// News-Notifier
PNotify TNewsNotify::Notify = TStdNotify::New();

// TTask class
// ////////////////
TTask::TTask(TNewsProc* _NewsProc) {
	uv_mutex_init(&StatusMutex);
	uv_mutex_init(&DataMutex);
	uv_mutex_init(&BufferMutex);

	NewsProc = _NewsProc;
	Status = waitingForData;
}

TTaskStatus TTask::GetStatus() {
	while (uv_mutex_trylock(&StatusMutex) != 0)
		Sleep(10);
	TTaskStatus _Status = Status;
	uv_mutex_unlock(&StatusMutex);
	return _Status;
}

void TTask::SetStatus(const TTaskStatus& _Status) {
	while (uv_mutex_trylock(&StatusMutex) != 0)
		Sleep(10);
	Status = _Status;
	uv_mutex_unlock(&StatusMutex);
}

void TTask::SetNewTaskData(const TStr& _FileName, const uint64& _RequestId)
{
	while (uv_mutex_trylock(&DataMutex) != 0)
		Sleep(10);

	FileName = _FileName;
	RequestId = _RequestId;

	Buffer.Clr();

	uv_mutex_unlock(&DataMutex);

	SetStatus(readyToParse);
}

void TTask::GetTaskData(TStr& _FileName, uint64& _RequestId)
{
	while (uv_mutex_trylock(&DataMutex) != 0)
		Sleep(10);

	_FileName = FileName;
	_RequestId = RequestId;

	uv_mutex_unlock(&DataMutex);
}

void TTask::SetBuffer(const TMem& _Buffer)
{
	LockBufferMutex();
	Buffer = _Buffer;
	UnlockBufferMutex();
}

void TTask::LockBufferMutex()
{
	while (uv_mutex_trylock(&BufferMutex) != 0)
		Sleep(10);
}

TMem TTask::GetBuffer()
{
	return Buffer;
}

void TTask::UnlockBufferMutex()
{
	uv_mutex_unlock(&BufferMutex);
}

///////////////////////////////
// News-Processing
TNewsProc::TNewsProc(const TStr& _DbsPath, const PJsonVal& NewsProcSett, const PNewsProcFilter& _Filter, const TStr& _AddArticleUrl, const TNewsProcSource& _Source) :
	DbsPath(_DbsPath), Filter(_Filter), AddArticleUrl(_AddArticleUrl), Source(_Source), ReStartsWithLoc("(- | -| \\(|\\. | \\.)")
{
	RedirectDynamicFNm = "redirectsDynamic";
	InfoLogFile = TLogger::New(NewsProcSett->GetObjInt("VerbosityLevel", 1));
	InfoLogFile->AddLogger(new TFileNotify("./Logs/Log-Build.txt", true, true));
		
	IgnoreIfNoAnnotations = false;
	SleepSec = 5;
	SleepBinaryMSec = 0;
	SkipLast = 0;
	DateExtractor = NULL;
	uv_rwlock_init(&HashLock);

	//if (ThreadCount > 1) {
	//	ThreadV.Gen(ThreadCount);
	//	TaskV.Gen(ThreadCount);
	//	for (int N = 0; N < ThreadCount; N++) {
	//		TaskV[N] = new TTask(this);
	//		uv_thread_create(&ThreadV[N], TNewsProc::ProcessFile, TaskV[N]);
	//		//uv_thread_join(&ThreadV[N]);
	//	}
	//}
}

// the static method that is called by each thread
void TNewsProc::ProcessFile(void *arg)
{
	TTask* Task = (TTask *) arg;
	while (true) {
		TTaskStatus Status = Task->GetStatus();

		if (Status == readyToParse) {
			int Rand = rand() % 200;
			Sleep(Rand);
			TStr FileName;
			uint64 RequestId;
			Task->GetTaskData(FileName, RequestId);

			// process the new file
			TMem Buffer;
			Task->NewsProc->ParseNewsFeedFile(FileName, Buffer);

			// set buffer to send
			Task->SetBuffer(Buffer);

			// set the status to ready to send
			Task->SetStatus(outputReady);
		}
		else
			Sleep(50);
	}
}

bool TNewsProc::EnumerateFiles(const TStr FPath, const TStrV& FileExV, TVec<TPair<TUInt, TStr> >& FTmFNmV)
{
	FTmFNmV.Clr();

	RE2::Options ReOptions;	ReOptions.set_case_sensitive(false);
	RE2 Re("\\d{4}-\\d{2}-\\d{2}T\\d{2}-\\d{2}-\\d{2}Z", ReOptions);

	TStrV TInPathV; TInPathV.Add(FPath);
	TFFile File(TInPathV, FileExV, "", false);

	TStr FileNm;
	while (File.Next(FileNm)) {
		try {
			StringPiece Matches[1];
			const StringPiece Text(FileNm.CStr());
			if (Re.Match(Text, 0, Text.size(), RE2::UNANCHORED, Matches, 1)) {
				int MatchStart = (int) (Matches[0].begin() - Text.begin());
				int MatchEnd = (int) (Matches[0].end() - Text.begin());
				TStr DateStr = Text.substr(MatchStart, MatchEnd - MatchStart).ToString().c_str();
				TTm Date = TTm::GetTmFromWebLogDateTimeStr(DateStr, '-', '-', 'Z', 'T');
				uint Sec = Date.GetSecTm().GetAbsSecs();
				//const uint64 CrawlTm = TFile::GetCreateTm(FileNm);
				FTmFNmV.Add(TPair<TUInt, TStr>(Sec, FileNm));
			}
			// if we weren't able to find the date in the name, then just add the file
			else
				FTmFNmV.Add(TPair<TUInt, TStr>(FTmFNmV.Len(), FileNm));
		}
		catch (PExcept E) {
			TNewsNotify::Notify->OnStatus("Error: " + E->GetMsgStr());
		}
		catch (...) {
			TNewsNotify::Notify->OnStatus("Unrecognized exception occured.");
		}
	}

	FTmFNmV.Sort(true);
	return true;
}

//void TNewsProc::StartMT(const TStr& InXmlFPath, const TStr& OutXmlFPath, const TStr& DoneBinFPath)
//{
//	TZipIn::IsZipExt(".zip");	// static variable that needs to be initialized before we start multiple threads
//	TStrV FileExV; FileExV.Add(".gz"); FileExV.Add(".xml");
//	TVec<TPair<TUInt, TStr> > FTmFNmV;
//	uint64 RequestId = 0;
//	int NextFileToEnqueue = 0;
//	uint64 NextExpectedRequestId = 0;
//	
//	while (true) {
//		// ////
//		// if we don't have any files left re-read the folder for new files
//		if (NextFileToEnqueue >= FTmFNmV.Len()) {
//			
//			NextFileToEnqueue = 0;
//			FTmFNmV.Clr();
//
//			TNewsNotify::Notify->OnStatus("Enumerating .gz files...");
//			TTmStopWatch StopWatch(true);
//			// prepare a list of crawled files
//			EnumerateFiles(InXmlFPath, FileExV, FTmFNmV);
//			TNewsNotify::Notify->OnStatusFmt("Enumerating %d files took %.2f seconds.", FTmFNmV.Len(), StopWatch.GetSec());
//
//			FTmFNmV.SortCmp(TCmpPairByVal2<TUInt, TStr>(true));
//			
//			// sleep for few seconds, to make sure all is crawled and saved
//			if (FTmFNmV.Len() < 10) {
//				TNewsNotify::Notify->OnStatusFmt("Sleeping for %d seconds.", SleepSec);
//				TSysProc::Sleep(SleepSec * 1000);
//			}
//		}
//
//		// ////
//		// check if new data needs to be assigned
//		for (int N = 0; N < TaskV.Len() && NextFileToEnqueue < FTmFNmV.Len(); N++) {
//			if (TaskV[N]->GetStatus() == waitingForData) {
//				TaskV[N]->SetNewTaskData(FTmFNmV[NextFileToEnqueue].Val2, RequestId);
//				NextFileToEnqueue++;
//				RequestId++;
//
//				// save wiki redirect files every 1000 processed files
//				if (RequestId % 1000 == 0)
//					SaveWikiRedirectFile();
//			}
//		}
//		
//		Sleep(50);
//
//		// ////
//		// repeatedly go over the tasks while we have some output ready
//		bool Change = true;
//		while (Change == true) {
//			Change = false;
//			for (int N = 0; N < TaskV.Len(); N++) {
//				TTaskStatus Status = TaskV[N]->GetStatus();
//				if (Status != outputReady)
//					continue;
//				// if we are here then we have a ready data to send. we need to check if the request id is the next in line
//				TStr FileName; uint64 RequestId;
//				TaskV[N]->GetTaskData(FileName, RequestId);
//				if (RequestId != NextExpectedRequestId)
//					continue;
//				
//				// send data
//				TaskV[N]->LockBufferMutex();
//				TMem Buffer = TaskV[N]->GetBuffer();
//				TaskV[N]->UnlockBufferMutex();
//
//				TNewsNotify::Notify->OnStatusFmt("Sending file nr. %I64u: %s.", RequestId, FileName.CStr());
//				if (Buffer.Len() > 0)
//					SendBuffer(Buffer);
//				else 
//					TNewsNotify::Notify->OnStatusFmt("Warning. The processed data for file %s was empty.", FileName.CStr());
//					
//				// send the data
//				Change = true;
//				TaskV[N]->SetStatus(waitingForData);
//				// move the file
//				TStr DstFNm = OutXmlFPath + FileName.GetFBase();
//				TFile::Copy(FileName, DstFNm); TFile::Del(FileName, false);
//
//				// increase the request id of the next expected package
//				NextExpectedRequestId++;
//				//TNewsNotify::Notify->OnStatusFmt("Processed file nr. %I64u: %s.", RequestId, FileName.CStr());
//			}
//		}
//	}
//}

void TNewsProc::StartOnline(const TStr& InXmlFPath, const TStr& OutXmlFPath)
{
	TStrV FileExV; FileExV.Add(".gz"); FileExV.Add(".xml");
	TVec<TPair<TUInt, TStr> > FTmFNmV;
	int SentFiles = 0;
	int SentArticles = 0;
	
	while (true) {
		// (re-)read the folder for new files
		TNewsNotify::Notify->OnStatus("Enumerating files...");
		TTmStopWatch StopWatch(true);
		// prepare a list of crawled files
		EnumerateFiles(InXmlFPath, FileExV, FTmFNmV);
		TNewsNotify::Notify->OnStatusFmt("Enumerating %d files took %.2f seconds.", FTmFNmV.Len(), StopWatch.GetSec());

		// sleep for few seconds, to make sure all is crawled and saved
		if (FTmFNmV.Len() < 10) {
			TNewsNotify::Notify->OnStatusFmt("Sleeping for %d seconds.", SleepSec);
			TSysProc::Sleep(SleepSec * 1000);
		}

		for (int N = 0; N < FTmFNmV.Len(); N++) {
			const TStr FileName = FTmFNmV[N].Val2;
			try {
				TMem Buffer;
				int Articles = ParseNewsFeedFile(FileName, Buffer);
				SentArticles += Articles;
				TNewsNotify::Notify->OnStatusFmt("[%s] Sending file nr. %d: %s (%d articles, total sent: %d).", TTm::GetCurLocTm().GetWebLogTimeStr().CStr(), ++SentFiles, FileName.CStr(), Articles, SentArticles);
				if (Buffer.Len() > 0) {
					// store the data into the output buffer
					TMem EncodedBuffer;
					TNewsMemEncoding::EncodeMem(Buffer, EncodedBuffer);
					SendBuffer(EncodedBuffer);
				}
				else
					TNewsNotify::Notify->OnStatusFmt("Warning. The processed data for file %s was empty.", FileName.CStr());
				// move the file
				TStr DstFNm = OutXmlFPath + FileName.GetFBase();
				TFile::Move(FileName, DstFNm, false);
				
				// save wiki redirect files every 1000 processed files
				if (SentFiles % 1000 == 0)
					SaveWikiRedirectFile();
			}
			catch (PExcept Except) {
				InfoLogFile->NotifyErr(TStr::Fmt("Exception while parsing file %s.", FileName.CStr()).CStr(), Except);
				printf("Exception while parsing file %s: %s\n", FileName.CStr(), Except->GetMsgStr().CStr());
			}
			catch (...) {
				InfoLogFile->NotifyErr("Unhandled exception.");
				printf("Unhandled exception.\n");
			}
		}
	}
}

void TNewsProc::StartParsing(const TStr& InXmlFPath, const TStr& OutXmlFPath, const TStr& OutBinFPath)
{
	TStrV FileExV; FileExV.Add(".gz"); FileExV.Add(".xml");
	TVec<TPair<TUInt, TStr> > FTmFNmV;
	int SentFiles = 0;
	int SentArticles = 0;

	while (true) {
		// (re-)read the folder for new files
		TNewsNotify::Notify->OnStatus("Enumerating files...");
		TTmStopWatch StopWatch(true);
		// prepare a list of crawled files
		EnumerateFiles(InXmlFPath, FileExV, FTmFNmV);
		TNewsNotify::Notify->OnStatusFmt("Enumerating %d files took %.2f seconds.", FTmFNmV.Len(), StopWatch.GetSec());

		// sleep for few seconds, to make sure all is crawled and saved
		if (FTmFNmV.Len() < 10) {
			TNewsNotify::Notify->OnStatusFmt("Sleeping for %d seconds.", SleepSec);
			TSysProc::Sleep(SleepSec * 1000);
		}

		for (int N = 0; N < FTmFNmV.Len(); N++) {
			const TStr FileName = FTmFNmV[N].Val2;
			TStr DestFNm = OutXmlFPath + FileName.GetFBase();

			try {
				HANDLE hFile = NULL;
				TStr BinFNm = OutBinFPath + FileName.GetFMid() + ".bin";
				if (TFile::Exists(BinFNm)) {
					TNewsNotify::Notify->OnStatusFmt("[%s] Skipping file nr. %d: %s (binary already exists).", TTm::GetCurLocTm().GetWebLogTimeStr().CStr(), ++SentFiles, FileName.CStr());
					TFile::Move(FileName, DestFNm, false);
				}
				else {
					TMem Buffer;
					// create a reading lock on the file
					hFile = CreateFile((FileName + ".lock").CStr(), GENERIC_READ, 0, NULL, CREATE_ALWAYS, 0, NULL);
					//DWORD LastError = GetLastError();

					// if we are able to create a valid handle it means that we can process a file that has not been processed yet
					if (hFile != INVALID_HANDLE_VALUE) {
						int Articles = ParseNewsFeedFile(FileName, Buffer);
						SentArticles += Articles;
						TNewsNotify::Notify->OnStatusFmt("[%s] Saving file nr. %d: %s (%d articles, total saved: %d).", TTm::GetCurLocTm().GetWebLogTimeStr().CStr(), ++SentFiles, FileName.CStr(), Articles, SentArticles);
						
						// it's possible that we kill the build process while it's generating the bin file. if it happens the bin file would be corrupted and next time the xml would be processed bin would not be recreated since it already exists
						// to prevent this problem we first create bin in the tmp file. if it's saved ok we only then rename it to bin file
						if (Articles > 0) {
							{ PSOut FOut = TFOut::New(BinFNm + ".tmp");
							Buffer.SaveMem(FOut); }
							TFile::Rename(BinFNm + ".tmp", BinFNm);
						}

						TFile::Move(FileName, DestFNm, false);
						
						CloseHandle(hFile);
						TFile::Del(FileName + ".lock");
					}
				}
			}
			catch (PExcept Except) {
				InfoLogFile->NotifyErr(TStr::Fmt("Exception while parsing file %s.", FileName.CStr()).CStr(), Except);
				printf("Exception while parsing file %s: %s\n", FileName.CStr(), Except->GetMsgStr().CStr());
			}
			catch (...) {
				InfoLogFile->NotifyErr("Unhandled exception.");
				printf("Unhandled exception.\n");
			}
		}
	}
}

void TNewsProc::StartOnlineBinary(const TStr& InBinFPath, const TStr& OutBinFPath)
{
	TStrV FileExV; FileExV.Add(".bin");
	TVec<TPair<TUInt, TStr> > FTmFNmV;
	int SentFiles = 0;
	
	while (true) {
		// (re-)read the folder for new files
		TNewsNotify::Notify->OnStatus("Enumerating files...");
		TTmStopWatch StopWatch(true);
		// prepare a list of crawled files
		EnumerateFiles(InBinFPath, FileExV, FTmFNmV);
		TNewsNotify::Notify->OnStatusFmt("Enumerating %d files took %.2f seconds.", FTmFNmV.Len(), StopWatch.GetSec());

		// sleep for few seconds, to make sure all is crawled and saved
		if (FTmFNmV.Len() < 10) {
			TNewsNotify::Notify->OnStatusFmt("Sleeping for %d seconds.", SleepSec);
			TSysProc::Sleep(SleepSec * 1000);
		}

		for (int N = 0; N < FTmFNmV.Len() - SkipLast; N++) {
			const TStr FileName = FTmFNmV[N].Val2;
			try {
				PSIn FIn = TFIn::New(FileName);
				TMem Buffer;
				TMem::LoadMem(FIn, Buffer);
				FIn = NULL;
														
				TNewsNotify::Notify->OnStatusFmt("[%s] Sending file nr. %d: %s.", TTm::GetCurLocTm().GetWebLogTimeStr().CStr(), ++SentFiles, FileName.CStr());
				if (Buffer.Len() > 0) {
					// store the data into the output buffer
					TMem EncodedBuffer;
					TNewsMemEncoding::EncodeMem(Buffer, EncodedBuffer);
					SendBuffer(EncodedBuffer);
				}

				// move the file
				TStr DstFNm = OutBinFPath + FileName.GetFBase();
				if (!TFile::Move(FileName, DstFNm, false))
					printf("WARNING: Unable to move file %s\n", FileName.CStr());
				if (SleepBinaryMSec > 0)
					TSysProc::Sleep(SleepBinaryMSec);
			}
			catch (PExcept Except) {
				InfoLogFile->NotifyErr(TStr::Fmt("Exception while parsing file %s.", FileName.CStr()).CStr(), Except);
				printf("Exception while parsing file %s: %s\n", FileName.CStr(), Except->GetMsgStr().CStr());
			}
			catch (...) {
				InfoLogFile->NotifyErr("Unhandled exception.");
				printf("Unhandled exception.\n");
			}
		}
	}
}

int TNewsProc::ParseNewsFeedFile(const TStr& InXmlFNm, TMem& OutBuffer) 
{
	PMem Buffer = TMem::New();
	PSOut BinSOut = TMemOut::New(Buffer);

	// check file not empty
	if (TFile::GetSize(InXmlFNm) == 0)
		return 0;

	// get file crawl time
	const TTm CrawlTm = TTm::GetTmFromMSecs(TFile::GetCreateTm(InXmlFNm));
	// open input XML (unzip if necessary)
	PSIn XmlSIn = InXmlFNm.IsSuffix(".gz") ? TZipIn::New(InXmlFNm) : TFIn::New(InXmlFNm);
	TXmlLx XmlLx(XmlSIn);
	// skip <xml>
	TXmlDoc::SkipTopTag(XmlSIn);
	// skip <article-set>
	TXmlDoc::SkipTopTag(XmlSIn);

	// load Items
	TTm StartTm, EndTm;
	int AllArticles = 0, GoodArticles = 0; uint64 AllMemUsed = 0;

	while (true) {
		// go to the next one
		PXmlDoc ItemXmlDoc;
		{
			ItemXmlDoc = TXmlDoc::LoadTxt(XmlLx);
			if (!ItemXmlDoc->IsOk())
				break;
		}
		// parse new article
		AllArticles++;

		try {
			TNewsArticle Article;
			PXmlTok ArticleTok = ItemXmlDoc->GetTok();
			//TStr ArticleXmlStr; TXmlDoc::New(ArticleTok)->SaveStr(ArticleXmlStr);
			// parse source type and check if to skip
			PXmlTok SourceTok = ArticleTok->GetTagTok("source");
			Article.SourceType = SourceTok->IsTagTok("type") ? SourceTok->GetTagTokStr("type") : "";
			// parse article language and check if to skip
			Article.ArticleLang = ArticleTok->GetTagTokStr("lang");
			// parse rest of source
			Article.SourceUri = SourceTok->IsTagTok("hostname") ? SourceTok->GetTagTokStr("hostname") : "";
			Article.SourceTitle = SourceTok->IsTagTok("name") ? SourceTok->GetTagTokStr("name") : "";
			// sometimes the title doesn't exist. in this case fallback to using hostname. don't use feed/title since it's not clean enough
			if (Article.SourceTitle == "")
				Article.SourceTitle = Article.SourceUri;
			if (SourceTok->IsSubTag("location")) {
				Article.SourceLocP = true;
				PXmlTok LocTok = SourceTok->GetTagTok("location");
				if (LocTok->IsSubTag("latitude") && LocTok->IsSubTag("longitude")) {
					Article.SourceLoc.Val1 = LocTok->GetTagTokStr("latitude").GetFlt();
					Article.SourceLoc.Val2 = LocTok->GetTagTokStr("longitude").GetFlt();
				}
				if (LocTok->IsSubTag("country")) { Article.SourceCountry = LocTok->GetTagTokStr("country"); }
				if (LocTok->IsSubTag("city")) { Article.SourceCity = LocTok->GetTagTokStr("city"); }
			}
			else {
				Article.SourceLocP = false;
			}
			// parse rest of article
			Article.ArticleUri = ArticleTok->GetTagTokStr("uri");
			
			/*if (InfoLogFile->GetVerbosityLevel() >= 1)
				InfoLogFile->NotifyVerboseFmt(1, "Processing article: %s", Article.ArticleUri.CStr());*/

			if (ArticleTok->IsSubTag("publish-date")) {
				const TStr ArticlePubTmStr = ArticleTok->GetTagTokStr("publish-date");
				Article.ArticlePubTm = TTm::GetTmFromWebLogDateTimeStr(ArticlePubTmStr, '-', ':', '.', 'T');
			}
			const TStr ArticleCrawlTmStr = ArticleTok->GetTagTokStr("retrieved-date");
			Article.ArticleCrawlTm = TTm::GetTmFromWebLogDateTimeStr(ArticleCrawlTmStr, '-', ':', '.', 'T');
			Article.ArticleTitle = ArticleTok->GetTagTokStr("title");

			// use the body from the plain text
			// we do this since non-english articles frequently don't have correctly parsed paragraphs
			Article.ArticleBody = GetClearTextBody(ArticleTok);

			// special treatment of enrycher-processed article's body
			switch (Source) {
				case NewsFeed:
					Article.ArticleBody = ArticleTok->GetTagTokStr("body-cleartext");
					break;
				case NewsFeedEnrycher:
					if (ArticleTok->IsTagTok("body-rych")) {
						// praser enrycher document
						PXmlTok DocTok = ArticleTok->GetTagTok("body-rych|item");
						PEnrycherDoc Doc = TEnrycherDoc::LoadXml(DocTok);
						ParseEnrycherDoc(Article, Doc);
					}
					else if (IgnoreIfNoAnnotations)
						continue;
					break;
				case NewsFeedXLike:
					if (ArticleTok->IsTagTok("body-xlike")) {
						// praser enrycher document
						const PXmlTok DocTok = ArticleTok->GetTagTok("body-xlike|item");
						ParseXLikeDoc(Article, DocTok);
					}
					else if (IgnoreIfNoAnnotations)
						continue;
					break;
			}

			// parse the category info from body-rich
			TXmlTokV CatXmlV; ArticleTok->GetTagTokV("body-rych|item|metadata|semantics|attribute", CatXmlV);
			for (int CatN = 0; CatN < CatXmlV.Len(); CatN++) {
				TStr Resource = CatXmlV[CatN]->GetStrArgVal("resource");
				Resource.ChangeStr("www.dmoz.org/Top/", "www.dmoz.org/");
				const double Wgt = CatXmlV[CatN]->GetFltArgVal("wgt", 1);
				if (Resource.IsPrefix("http://www.dmoz.org") && !IsIgnoredCategory(Resource)) {
					// find the category of the appropriate depth
					TStr ValidCat = GetValidCategoryDepth(Resource);
					Article.AddCategory(ValidCat, Wgt);
				}
			}

			// parse images
			if (ArticleTok->IsSubTag("img")) {
				ArticleTok->GetTagValV("img", false, Article.ImageUriV);
			}
			// parse similar articles
			TXmlTokV SimArticlesV; ArticleTok->GetTagTokV("similar-articles", SimArticlesV);
			for (int SimArticlesN = 0; SimArticlesN < SimArticlesV.Len(); SimArticlesN++) {
				PXmlTok SimArticles = SimArticlesV[SimArticlesN];
				const TStr Lang = SimArticles->GetStrArgVal("stream");
				TXmlTokV SimArticleV; SimArticles->GetTagTokV("a", SimArticleV);
				for (int SimArticleN = 0; SimArticleN < SimArticleV.Len(); SimArticleN++) {
					PXmlTok SimArticle = SimArticleV[SimArticleN];
					const TStr NewsfeedId = SimArticle->GetStrArgVal("id");
					const int SimWgt = TFlt::Round(100.0 * SimArticle->GetFltArgVal("sim"));
					Article.SimArticleV.Add(TStrStrIntTr(NewsfeedId, Lang, SimWgt));
				}
			}
			// get google cluster id
			if (ArticleTok->IsSubTag("google_story_id")) {
				Article.ArticleGoogleClustId = ArticleTok->GetTagTokStr("google_story_id");
			}
			// get bloombergness
			if (ArticleTok->IsSubTag("x-blb")) {
				TStr BlpStr = ArticleTok->GetTagTokStr("x-blb");
				if (BlpStr.IsFlt()) {
					const int BlpScore = TFlt::Round(10.0 * BlpStr.GetFlt());
					TStr Label = TInt::GetStr(BlpScore);
					Article.AuxKdV.Add(TStrKd("BlpScore", Label));
				}
			}
			// remember Newsfeed ID
			Article.AuxKdV.Add(TStrKd("NewsfeedId", ArticleTok->GetStrArgVal("id")));
			// check if filter passes the article
			if (Filter->Pass(Article) && !Filter->Ignore(Article)) {
				// save article
				if (!BinSOut.Empty()) { Article.Save(*BinSOut); }
				GoodArticles++;
				// remember claimed-times
				if (!StartTm.IsDef())
					EndTm = StartTm = Article.ArticleCrawlTm;
				else {
					if (StartTm > Article.ArticleCrawlTm) { StartTm = Article.ArticleCrawlTm; }
					if (EndTm < Article.ArticleCrawlTm) { EndTm = Article.ArticleCrawlTm; }
				}
			}
		}
		catch (PExcept Except) {
			TNewsNotify::Notify->OnStatusFmt("ERROR: Problem parsing article data for article %d in file %s: %s", AllArticles, InXmlFNm.CStr(), Except->GetMsgStr().CStr());
		}
		catch (...) {
			TNewsNotify::Notify->OnStatusFmt("ERROR: General error while parsing article data for article %d in file %s.", AllArticles, InXmlFNm.CStr());
		}
	}

	// report on what was parsed
	//TNewsNotify::Notify->OnStatusFmt("%s[%s]: %d / %d", InXmlFNm.CStr(), TUInt64::GetMegaStr(AllMemUsed).CStr(), AllArticles, GoodArticles);
	//TNewsNotify::Notify->OnStatusFmt(" crawled %s", CrawlTm.GetWebLogDateTimeStr().CStr());
	//TNewsNotify::Notify->OnStatusFmt(" from    %s to %s", StartTm.GetWebLogDateTimeStr().CStr(), EndTm.GetWebLogDateTimeStr().CStr());

	//// store the data into the output buffer
	//TNewsMemEncoding::EncodeMem(Buffer, OutBuffer);
	OutBuffer = *Buffer;

	// return size of parsed articles
	return GoodArticles;
}

void TNewsProc::ParseEnrycherDoc(TNewsArticle& Article, const PEnrycherDoc& Doc) 
{
	// extract body text
	//TChA Body;
	//for (int ParaN = 0; ParaN < Doc->GetParas(); ParaN++) {
	//	// we might get the title in the body - used so that enrycher also enryches the title
	//	if (ParaN == 0 && Doc->GetPara(ParaN).GetStr() == Article.ArticleTitle)
	//		continue;
	//	if (Body.Len() > 0) { Body += "\n\n"; }
	//	Body += Doc->GetPara(ParaN).GetStr();
	//}
	//Article.ArticleBody = Body;
	
	// extract article dates
	ExtractArticleDates(Article, Doc);

	const TStr FirstSentence = Doc->IsSentId(1) ? Doc->GetSent(1).GetStr() : "";
	bool PotentialLocation = false;
	TStr StartText = Article.ArticleBody.GetSubStr(0, 45);
	const StringPiece StartTextSP(StartText.CStr());
	bool PotentiallyStartsWithLoc = ReStartsWithLoc.Match(StartTextSP, 0, StartTextSP.size(), RE2::UNANCHORED, Matches, 1);
	
	// extract annotations
	TIntFltKdV AnnIdWgtV; TEnrycherUtil::GetAnnWgt(Doc, AnnIdWgtV);
	for (int AnnIdWgtN = 0; AnnIdWgtN < AnnIdWgtV.Len(); AnnIdWgtN++) {
		const int AnnId = AnnIdWgtV[AnnIdWgtN].Key;
		const double AnnWgt = AnnIdWgtV[AnnIdWgtN].Dat;
		const TEnrycherAnn& Ann = Doc->GetAnnById(AnnId);
		// remember the annotation
		TStr Type = Ann.GetType();
		TStr AnnLabel = Ann.GetStr(), AnnType;
		if (Type.IsStrIn("location")) { AnnType = NewsConst.TypeLocation; }
		else if (Type.IsStrIn("person")) { AnnType = NewsConst.TypePerson; }
		else if (Type.IsStrIn("organisation")) { AnnType = NewsConst.TypeOrganization; }
		else if (Type.IsStrIn("organization")) { AnnType = NewsConst.TypeOrganization; }
		else if (Type == "wiki") { AnnType = NewsConst.TypeWiki; }
		else { continue; }
		// get annotation uri
		TStrV AnnUriV; // = "enrycherann:" + TRdfUri::GetUriStr(AnnLabel);
		// check if there is wikipedia link
		for (int AttrN = 0; AttrN < Ann.GetSemAttrs(); AttrN++) {
			const TEnrycherSemAttr& Attr = Ann.GetSemAttr(AttrN);
			if (Attr.GetType() == "owl:sameAs") {
				if (Attr.GetResource().IsPrefix("dbr:")) {
					TStr AnnUri = Attr.GetResource();
					// resolve redirects
					AnnUri = GetWikiRedirect(AnnUri);
					// make wikipedia link
					AnnUri.ChangeStr("dbr:", "http://en.wikipedia.org/wiki/");
					// remember
					AnnUriV.AddUnique(AnnUri);
				} else if (Attr.GetResource().IsStrIn(".wikipedia.org/wiki/")) {
					TStr AnnUri = Attr.GetResource();
					AnnUriV.AddUnique(GetWikiRedirect(AnnUri));
				}
			}
		}
		// if english wiki url is present then 
		// 1. remember mappings from non-english to english uri (for future occurences of non-english-uri only)
		// 2. remove non-english mappings and use only the english one
		TStr EnglishWikiUrl = "";
		for (int AnnUriN = 0; AnnUriN < AnnUriV.Len(); AnnUriN++) {
			if (AnnUriV[AnnUriN].IsPrefix("http://en.wikipedia.org/wiki/")) {
				EnglishWikiUrl = AnnUriV[AnnUriN];
				break;
			}
		}
				
		if (EnglishWikiUrl != "") {
			// remember nonenglish_uri --> english_uri mappings
			// forget non-english urls if we have english one
			for (int AnnUriN = AnnUriV.Len()-1; AnnUriN >= 0; AnnUriN--) {
				if (!AnnUriV[AnnUriN].IsPrefix("http://en.wikipedia.org/wiki/")) {
					SetWikiRedirect(AnnUriV[AnnUriN], EnglishWikiUrl);
					AnnUriV.Del(AnnUriN);	
				}					
			}

			// forget all english urls except the first one (EnglishWikiUrl)
			// this solves the problem with enrycher where we get 
			//<attribute displayName = "dbr:California" resource = "dbr:California" type = "owl:sameAs" / >
			//<attribute displayName = "dbr:Kalifornien" resource = "dbr:Kalifornien" type = "owl:sameAs" / >
			// in which case only the first url is actually pointing to english wikipedia
			for (int AnnUriN = AnnUriV.Len() - 1; AnnUriN >= 0; AnnUriN--) {
				if (AnnUriV[AnnUriN].IsPrefix("http://en.wikipedia.org/wiki/") && AnnUriV[AnnUriN] != EnglishWikiUrl)
					AnnUriV.Del(AnnUriN);
			}
		}
		// if we didn't find an english uri then we have a list of non-english uris for the same concept
		// we need to remove all but one of this concepts otherwise we will be saving multiple annotations of the same concept
		// since we want to pick the same language no matter in which order we found the concepts we sort the vector first by name and then remove all but first item
		else {
			AnnUriV.Sort();
			while (AnnUriV.Len() > 1)
				AnnUriV.DelLast();
		}
				
		int TokIndexOfLocation = TInt::Mx;	// it's possible that we have multiple locations in the first sentence. we want to set as LocationConceptN the one that appears first
		// remember annotation
		for (int AnnUriN = 0; AnnUriN < AnnUriV.Len(); AnnUriN++) {
			TStr AnnUri = AnnUriV[AnnUriN];
			// get the redirect if existing
			AnnUri = GetWikiRedirect(AnnUri);
			// if based on the db we know that the uri is a location then make the type location
			if (IsLocationUri(AnnUri))
				AnnType = NewsConst.TypeLocation;
			if (IsEntityBlackList(AnnUri)) 
				continue; 
			
			// NOTE: i commented this out since we do this on the news server. there we can also reload the file, etc - here we need to restart building to load new file
			//// correct the url, type and label for the given concept - if we have info from the rules files
			//ConceptFix.CorrectConceptInfo(AnnUri, AnnType, AnnLabel, Article.ArticleLang);

			//// after we (possibly) corrected the url, we check if this url is maybe ignored
			//if (ConceptFix.HasConceptIgnore(AnnUri))
			//	continue;
						
			// if the article starts in a way that shows that potentially it mentions a location AND the currently tested concepts is a location
			if (PotentiallyStartsWithLoc && (AnnType == NewsConst.TypeLocation || WikiArticleLocationMappingH.IsKey(AnnUri))) {
				const int AnnInsts = Ann.GetAnnInsts();
				for (int InstN = 0; InstN < AnnInsts; InstN++) {
					const TEnrycherAnnInst& AnnInst = Ann.GetAnnInst(InstN);
					const TStr AnnDisplayName = AnnInst.GetStr();
					int Tokens = AnnInst.GetTokens();
					for (int TokN = 0; TokN < Tokens; TokN++) {
						int TokPos = FirstSentence.SearchStr(AnnDisplayName);
						if (AnnInst.GetSentId(TokN) == 1 && TokPos >= 0 && TokPos < 5 && TokPos < TokIndexOfLocation) {
							Article.LocationConceptN = Article.ConceptV.Len();
							TokIndexOfLocation = TokPos;
							//InfoLogFile->NotifyVerboseFmt(2, "\t\tArticle location %s inferred from: %s ....", Article.ConceptV[Article.ConceptV.Len() - 1].GetUri().CStr(), FirstSentence.GetSubStr(0, 50).CStr());

							// if we did find a location at the beginning of the article and this location is currently a URI that is in WikiArticleLocationMappingH then we have a case where the location
							// was not appropriately specified so a mistake was made. For example, having "Washingon (AP) " at the beginning will annotate Washington the state. Washington state will be in
							// WikiArticleLocationMappingH where the dat will have Washington D.C. . So, here we want to replace existing location uri with the one in WikiArticleLocationMappingH
							if (WikiArticleLocationMappingH.IsKey(AnnUri)) {
								AnnUri = WikiArticleLocationMappingH.GetDat(AnnUri);
								AnnType = NewsConst.TypeLocation;		// make sure we set the type to location
							}
							// NOTE: we don't change the AnnLabel here - this is not good but we hope that the concept has been seen before so the label was already set and won't be changed
						}
					}
				}
			}

			Article.ConceptV.Add(TNewsConcept(AnnUri, AnnLabel, AnnType, AnnWgt));
		}
	}

	if (InfoLogFile->GetVerbosityLevel() >= 4 && Article.LocationConceptN != TInt::Mx) {
		InfoLogFile->Notify(TNotifyType::ntInfo, TStr::Fmt("Found concept: %s in text:", Article.ConceptV[Article.LocationConceptN].GetUri().CStr()).CStr());
		InfoLogFile->Notify(TNotifyType::ntInfo, StartText.CStr());
		InfoLogFile->NotifyInfo("");
	}

	// extract topics and keywords
	//int TopicN = 1, TagN = 1; // topic and tag counter, for faking weights
	//for (int AttrN = 0; AttrN < Doc->GetSemAttrs(); AttrN++) {
	//	const TEnrycherSemAttr& Attr = Doc->GetSemAttr(AttrN);
	//	if (Attr.GetType() == "sioc:topic") {
	//		// check if dmoz topic
	//		if (!Attr.GetResource().IsPrefix("http://www.dmoz.org")) { continue; }
	//		// if yes, remember it
	//		TStr Label = Attr.GetDisplayName();
	//		// DMOZ category
	//		Article.ConceptV.Add(TNewsConcept(NewsConst.TypeCategory + ":" + Label, 
	//			Label, NewsConst.TypeCategory, 1.0 / double(TopicN++)));
	//	} else if (Attr.GetType() == "sioc:tag") {
	//		// dmoz keywords
	//		TStr Label = Attr.GetDisplayName();
	//		Article.ConceptV.Add(TNewsConcept(NewsConst.TypeKeyword + ":" + Label,
	//			Label, NewsConst.TypeKeyword, 1.0 / double(TagN++)));
	//	}
	//}
}

void TNewsProc::ParseXLikeDoc(TNewsArticle& Article, const PXmlTok& Doc) {
	PEnrycherDoc EnrycherDoc = TEnrycherDoc::LoadXml(Doc);
	ParseEnrycherDoc(Article, EnrycherDoc);
}

// send a buffer to the listeners
void TNewsProc::SendBuffer(const TMem& Buffer) {
	bool Ok; TStr MsgStr; PWebPg WebPg; int TryN = 1;
	while (true) {
		// generate http request
		PUrl Url = TUrl::New(AddArticleUrl);
		PHttpRq HttpRq = THttpRq::New(hrmPost, Url, THttp::TextFldVal, Buffer);
		// execute request
		TWebFetchBlocking::GetWebPg(HttpRq, Ok, MsgStr, WebPg, NULL);
		// break if request successfull
		if (Ok) { break; }
		// wait a bit otherwise
		Sleep(5000);
		TryN++;
		if (TryN % 10 == 0)
			TNewsNotify::Notify->OnNotifyFmt(ntErr, "There was a problem calling the article service: %s", MsgStr.CStr());
	}
}

void TNewsProc::ExtractArticleDates(TNewsArticle& Article, const PEnrycherDoc& Doc)
{
	if (DateExtractor.Empty())
		return;
	// proceed only if we have one of the known languages
	TExtractedDateColl ExtractedDateColl;
	TTm Date;
	if (Article.ArticleCrawlTm.IsDef())
		Date = Article.ArticleCrawlTm;
	DateExtractor->ExtractDates(Article.ArticleBody, Article.ArticleLang, Date, ExtractedDateColl);
	// if there are too many dates, then ignore them completely
	if (ExtractedDateColl.Len() > 0 && ExtractedDateColl.Len() <= 10) {
		Article.DatesJson = ExtractedDateColl.EncodeExtractedDates();
	}
}

void TNewsProc::LoadDbs() {
	TNotify::StdNotify->OnStatus("==== Loading databases... ====");
	// 
	// load wiki redirects
	if (TFile::Exists(DbsPath + "redirectsStatic.dat")) {
		TNotify::StdNotify->OnStatus("Loading dynamic wiki redirects from " + DbsPath + "redirectsStatic.dat");
		TFIn FIn(DbsPath + "redirectsStatic.dat");
		WikiStaticRedirectH.Load(FIn);
		TNotify::StdNotify->OnStatusFmt("Found %d dynamic wiki redirects.", WikiStaticRedirectH.Len());
	}
	else {
		TStrV StaticFNameV; StaticFNameV.Add("redirectsStaticGeneral"); StaticFNameV.Add("redirectsStaticLocation");
		for (int N = 0; N < StaticFNameV.Len(); N++) {
			const TStr FileNm = DbsPath + StaticFNameV[N] + ".txt";
			if (TFile::Exists(FileNm)) {
				TNotify::StdNotify->OnStatus("Loading static wiki redirects from " + FileNm);
				TSsParser Ss(FileNm, '\t');
				while (Ss.Next()) {
					if (Ss.GetFlds() == 2)
						WikiStaticRedirectH.AddDat(Ss.GetFld(0), Ss.GetFld(1));
				}
			}
		}
		TFOut FOut(DbsPath + "redirectsStatic.dat");
		WikiStaticRedirectH.Save(FOut);
		TNotify::StdNotify->OnStatusFmt("Found %d static wiki redirects.", WikiStaticRedirectH.Len());
	}
		
	bool SkipDynamicBin = false;
	// check if the txt file was edited manually after the bin was last saved. if yes, then ignore bin and load from txt
	if (TFile::Exists(DbsPath + RedirectDynamicFNm + ".dat") && TFile::Exists(DbsPath + RedirectDynamicFNm + ".txt"))
		SkipDynamicBin = TFile::GetLastWriteTm(DbsPath + RedirectDynamicFNm + ".txt") > TFile::GetLastWriteTm(DbsPath + RedirectDynamicFNm + ".dat");

	if (!SkipDynamicBin && TFile::Exists(DbsPath + RedirectDynamicFNm + ".dat")) {
		TNotify::StdNotify->OnStatus("Loading dynamic wiki redirects from " + DbsPath + RedirectDynamicFNm + ".dat");
		TFIn FIn(DbsPath + RedirectDynamicFNm + ".dat");
		WikiDynamicRedirectH.Load(FIn);
		TNotify::StdNotify->OnStatusFmt("Found %d dynamic wiki redirects.", WikiDynamicRedirectH.Len());
	}
	else if (TFile::Exists(DbsPath + RedirectDynamicFNm + ".txt")) {
		TNotify::StdNotify->OnStatus("Loading dynamic wiki redirects from " + DbsPath + RedirectDynamicFNm + ".txt");
		TSsParser Ss(DbsPath + RedirectDynamicFNm + ".txt", '\t');
		while (Ss.Next()) {
			if (Ss.GetFlds() == 2)
				WikiDynamicRedirectH.AddDat(Ss.GetFld(0), Ss.GetFld(1));
		}
		TFOut FOut(DbsPath + RedirectDynamicFNm + ".dat");
		WikiDynamicRedirectH.Save(FOut);
		TNotify::StdNotify->OnStatusFmt("Found %d dynamic wiki redirects.", WikiDynamicRedirectH.Len());
	}
	else
		TNotify::StdNotify->OnStatus("Warning: Wiki dynamic redirects file was not found at " + DbsPath + RedirectDynamicFNm + ".*");

	// 
	// load entity blacklist
	bool SkipBin = false;
	TStr FNameRoot = "entityBlacklist";
	// check if the txt file was edited manually after the bin was last saved. if yes, then ignore bin and load from txt
	if (TFile::Exists(DbsPath + FNameRoot + ".dat") && TFile::Exists(DbsPath + FNameRoot + ".txt"))
		SkipBin = TFile::GetLastWriteTm(DbsPath + FNameRoot + ".txt") > TFile::GetLastWriteTm(DbsPath + FNameRoot + ".dat");

	if (!SkipBin && TFile::Exists(DbsPath + FNameRoot + ".dat")) {
		TNotify::StdNotify->OnStatus("Loading entity blacklist from " + DbsPath + FNameRoot + ".dat");
		TFIn FIn(DbsPath + FNameRoot + ".dat");
		EntityBlackListH.Load(FIn);
		TNotify::StdNotify->OnStatusFmt("Found %d blacklisted entities.", EntityBlackListH.Len());
	}
	else if (TFile::Exists(DbsPath + FNameRoot + ".txt")) {
		TNotify::StdNotify->OnStatus("Loading entity blacklist from " + DbsPath + FNameRoot + ".txt");
		TFIn FIn(DbsPath + FNameRoot + ".txt"); TStr LnStr;
		while (FIn.GetNextLn(LnStr)) {
			if (!LnStr.Empty())
				EntityBlackListH.AddKey(LnStr);
		}
		TFOut FOut(DbsPath + FNameRoot + ".dat");
		EntityBlackListH.Save(FOut);
		TNotify::StdNotify->OnStatusFmt("Found %d blacklisted entities.", EntityBlackListH.Len());
	}
	else
		TNotify::StdNotify->OnStatus("Warning: Entity blacklist file was not found at " + DbsPath + FNameRoot + ".*");

	// database of locations
	SkipBin = false;
	FNameRoot = "entityLocation";
	// check if the txt file was edited manually after the bin was last saved. if yes, then ignore bin and load from txt
	if (TFile::Exists(DbsPath + FNameRoot + ".dat") && TFile::Exists(DbsPath + FNameRoot + ".txt"))
		SkipBin = TFile::GetLastWriteTm(DbsPath + FNameRoot + ".txt") > TFile::GetLastWriteTm(DbsPath + FNameRoot + ".dat");

	if (!SkipBin && TFile::Exists(DbsPath + FNameRoot + ".dat")) {
		TNotify::StdNotify->OnStatus("Loading entity location file from " + DbsPath + FNameRoot + ".dat");
		TFIn FIn(DbsPath + FNameRoot + ".dat");
		LocationH.Load(FIn);
		TNotify::StdNotify->OnStatusFmt("Found %d known locations.", LocationH.Len());
	}
	else if (TFile::Exists(DbsPath + FNameRoot + ".txt")) {
		TNotify::StdNotify->OnStatus("Loading entity location file from " + DbsPath + FNameRoot + ".txt");
		TFIn FIn(DbsPath + FNameRoot + ".txt"); TStr LnStr;
		while (FIn.GetNextLn(LnStr)) {
			if (LnStr.Empty())
				continue;
			// we trim the typical prefix so that we save space when serializing
			LnStr = LnStr.TrimLeft("http://en.wikipedia.org/wiki/");
			LocationH.AddKey(LnStr);
		}
		TFOut FOut(DbsPath + FNameRoot + ".dat");
		LocationH.Save(FOut);
		TNotify::StdNotify->OnStatusFmt("Found %d known locations.", LocationH.Len());
	}
	else
		TNotify::StdNotify->OnStatus("Warning: Entity location file was not found at " + DbsPath + FNameRoot + ".*");
		
	// a part of the data extraction from the article is also determining if the article has a specified location. that is the location that sometimes appears 
	// in the article at the beginning. To determine accurately we need a list of all known cities
	// additionally we also need to modify some frequently incorrectly specified places and replace them with correct annotations
	// example of this is Washington. if mentioned at the beginning of the article, the annotation becomes Washington state. We need to replace that annotation with Washington D.C.
	if (TFile::Exists(DbsPath + "articleLocationMapping.txt")) {
		TNotify::StdNotify->OnStatus("Loading dynamic wiki redirects from " + DbsPath + "articleLocationMapping.txt");
		TSsParser Ss(DbsPath + "articleLocationMapping.txt", '\t');
		while (Ss.Next()) {
			if (Ss.GetFlds() == 2)
				WikiArticleLocationMappingH.AddDat(Ss.GetFld(0), Ss.GetFld(1));
		}
		TNotify::StdNotify->OnStatusFmt("Found %d article location mappings.", WikiArticleLocationMappingH.Len());
	}
	
	TNotify::StdNotify->OnStatus("==== Loading finished. ====");
}

TStr TNewsProc::GetWikiRedirect(const TStr& InStr) {
	TStr Str = InStr;
	// check for redirect. support multiple redirects
	int TryCount = 0;
	uv_rwlock_rdlock(&HashLock);
	if (WikiStaticRedirectH.IsKey(InStr)) {
		while (WikiStaticRedirectH.IsKey(Str) && TryCount < 10) {
			TryCount++;
			Str = WikiStaticRedirectH.GetDat(Str);
		}
		if (TryCount >= 10)
			TNotify::StdNotify->OnStatus("Possibly circular static redirect for concept " + Str);
	}
	TryCount = 0;
	if (WikiDynamicRedirectH.IsKey(InStr)) {
		while (WikiDynamicRedirectH.IsKey(Str) && TryCount < 10) {
			TryCount++;
			Str = WikiDynamicRedirectH.GetDat(Str);
		}
		if (TryCount >= 10)
			TNotify::StdNotify->OnStatus("Possibly circular dynamic redirect for concept " + Str);
	}
	uv_rwlock_rdunlock(&HashLock);
	return Str;
}

bool TNewsProc::HasWikiRedirect(const TStr& InStr) {
	uv_rwlock_rdlock(&HashLock);
	bool HasRedirect = WikiStaticRedirectH.IsKey(InStr) || WikiDynamicRedirectH.IsKey(InStr);
	uv_rwlock_rdunlock(&HashLock);
	return HasRedirect;
}

void TNewsProc::SetWikiRedirect(const TStr& InStr, const TStr& OutStr)
{
	TStr Test = GetWikiRedirect(InStr);
	if (Test == OutStr)
		return;
	uv_rwlock_wrlock(&HashLock);
	WikiDynamicRedirectH.AddDat(InStr, OutStr);
	uv_rwlock_wrunlock(&HashLock);
}

void TNewsProc::SaveWikiRedirectFile()
{
	uv_rwlock_wrlock(&HashLock);
	TNotify::StdNotify->OnStatus("Saving dynamic wiki redirects...");
	TFOut FOut(DbsPath + RedirectDynamicFNm + ".txt");
	for (int KeyId = WikiDynamicRedirectH.FFirstKeyId(); WikiDynamicRedirectH.FNextKeyId(KeyId);)
		FOut.PutStrFmt("%s\t%s\n", WikiDynamicRedirectH.GetKey(KeyId), WikiDynamicRedirectH[KeyId].CStr());
	TNotify::StdNotify->OnStatus("Done");
	uv_rwlock_wrunlock(&HashLock);
}

bool TNewsProc::IsEntityBlackList(const TStr& EntityStr) {
	// check for redirect
	return EntityBlackListH.IsKey(EntityStr);
}

bool TNewsProc::IsLocationUri(const TStr& EntityStr) {
	// check for redirect
	TStr EntityStrShort = EntityStr.TrimLeft("http://en.wikipedia.org/wiki/");
	return LocationH.IsKey(EntityStrShort);
}

bool TNewsProc::IsIgnoredCategory(const TStr& CategoryNm)
{
	//TStr CategoryNm = _CategoryNm.GetLc();
	for (int N = 0; N < CatIgnoreV.Len(); N++) {
		if (CategoryNm.IsPrefix(CatIgnoreV[N]))
			return true;
	}
	return false;
}

TStr TNewsProc::GetValidCategoryDepth(const TStr& CategoryNm)
{
	for (int N = 0; N < CatPrefixV.Len(); N++) {
		if (CategoryNm.IsPrefix(CatPrefixV[N].Val1)) {
			const int MaxDepth = CatPrefixV[N].Val2;
			int CurrDepth = -3;	// we have 3 '/' in the http://dmoz.com/ prefix
			for (int i = 0; i < CategoryNm.Len(); i++) {
				if (CategoryNm.GetCh(i) == '/') CurrDepth++;
				if (CurrDepth >= MaxDepth)
					return CategoryNm.GetSubStr(0, i-1);
			}
		}
	}
	// we didn't find any suitable rule, so just return the full category name
	return CategoryNm;
}

// convert the <p>...</p> to double newlines
TStr TNewsProc::GetClearTextBody(const PXmlTok& ArticleTok)
{
	PXmlTok BodyTok = ArticleTok->GetTagTok("body-cleartext");
	int Tokens = BodyTok->GetSubToks();
	TChA Body;
	for (int N = 0; N < Tokens; N++) {
		TStr Text = BodyTok->GetSubTok(N)->GetTokStr(false);
		Text.ToTrunc();
		if (Text.Len() == 0)
			continue;
		Body += Text;
		if (N < Tokens-1)
			Body += "\n\n";
	}
	return Body;
}

///////////////////////////////
// News-Processing-Filter
TNewsProcFilter::TNewsProcFilter(const PNewsProcFilter& Filter)
{
	SourceTypeSet = TStrSet(Filter->SourceTypeSet);
	LangSet = TStrSet(Filter->LangSet);
	DomainSet = TStrSet(Filter->DomainSet);

	IgnoreDomainSet = TStrSet(IgnoreDomainSet);
	IgnoreSourceTypeSet = TStrSet(IgnoreSourceTypeSet);
	IgnoreLangSet = TStrSet(IgnoreLangSet);
}

TStr TNewsProcFilter::GetDomain(const TStr& UrlStr) const {
	// extract the basic url
	PUrl Url = TUrl::New(UrlStr);
	if (!Url->IsOk()) { return ""; }
	TStr Domain = Url->GetHostNm();
	// clean URLs only to keep only domain name
	if (Domain.CountCh('.') > 1) {
		// remove any subdomains
		int FirstDot = Domain.SearchChBack('.');
		int SecondDot = Domain.SearchChBack('.', FirstDot - 1);
		// handle cases like .co.uk
		if (Domain.Len() - SecondDot > 7) {
			Domain = Domain.GetSubStr(SecondDot + 1, Domain.Len() - 1);
		}
	}
	// done
	return Domain.GetLc();
}

TNewsProcFilter::TNewsProcFilter(const TStrSet& _SourceTypeSet, const TStrSet& _LangSet, const TStrSet& _DomainSet,
	const TStrSet& _IgnoreSourceTypeSet, const TStrSet& _IgnoreLangSet, const TStrSet& _IgnoreDomainSet) :
	SourceTypeSet(_SourceTypeSet), LangSet(_LangSet), IgnoreSourceTypeSet(_IgnoreSourceTypeSet), IgnoreLangSet(_IgnoreLangSet) {

	for (int KeyId = _DomainSet.FFirstKeyId(); _DomainSet.FNextKeyId(KeyId);)
		DomainSet.AddKey(GetDomain(_DomainSet.GetKey(KeyId)));
	for (int KeyId = _IgnoreDomainSet.FFirstKeyId(); _IgnoreDomainSet.FNextKeyId(KeyId);)
		IgnoreDomainSet.AddKey(GetDomain(_IgnoreDomainSet.GetKey(KeyId)));
	//if (!DomainFNm.Empty()) {
	//	TFIn FIn(DomainFNm); TStr LnStr;
	//	while (FIn.GetNextLn(LnStr)) {
	//		// shall we skip the line?
	//		if (LnStr.Empty() || LnStr.IsWs()) { continue; }
	//		if (LnStr[0] == '#') { continue; }
	//		// get domain
	//		TStr Domain = GetDomain(LnStr);
	//		if (!Domain.Empty()) { DomainSet.AddKey(Domain); }
	//	}
	//}
}

TStrSet TNewsProcFilter::ReadFile(const TStr& FileName)
{
	TStrSet Lines;
	if (TFile::Exists(FileName)) {
		TFIn FIn(FileName); TStr LnStr;
		while (FIn.GetNextLn(LnStr)) {
			// shall we skip the line?
			if (LnStr.Empty() || LnStr.IsWs()) { continue; }
			if (LnStr[0] == '#') { continue; }
			Lines.AddKey(LnStr);
		}
	}
	return Lines;
}

bool TNewsProcFilter::Pass(const TNewsArticle& Article) const {
	// if not filters, always pass
	if (LangSet.Empty() && SourceTypeSet.Empty() && DomainSet.Empty()) return true;
	// otherwise default is not pass
	if (LangSet.IsKey(Article.ArticleLang)) return true;
	if (SourceTypeSet.IsKey(Article.SourceType)) return true;
	if (DomainSet.IsKey(GetDomain(Article.ArticleUri))) return true;
	return false;
}

bool TNewsProcFilter::Ignore(const TNewsArticle& Article) const {
	if (IgnoreDomainSet.IsKey(GetDomain(Article.ArticleUri))) return true;
	if (IgnoreLangSet.IsKey(Article.ArticleLang)) return true;
	if (IgnoreSourceTypeSet.IsKey(Article.SourceType)) return true;
	if (Article.ArticleBody.Len() < MinBodyLen) return true;
	return false;
}