/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "NewsSearch.h"

// method for searching for events based on 
bool TNewsSearch::SearchEvents(const TEventQuery& Query, PRecSet& RecSet, PJsonVal& ErrorJson) const
{
	ErrorJson = NULL;
	RecSet = NULL;
	
	TStr Error;
	PRecSet EventRecSet;
	// search by concepts
	if (Query.ConceptUriV.Len() > 0) {
		
		bool Ok = FindEventsForConcepts(Query.ConceptUriV, EventRecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (EventRecSet->GetRecs() == 0) {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given topics/concepts"));
			return false;
		}
	}

	// search by keywords
	if (Query.Keywords != "") {
		PRecSet RecSet; bool Ok = FindEventsForKeywords(Query.Keywords, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given keywords"));
			return false;
		}
	}

	// search by location
	if (Query.LocationUriV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindEventsForLocations(Query.LocationUriV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given location"));
			return false;
		}
	}

	// search by publisher
	if (Query.PublisherUriV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindEventsForPublishers(Query.PublisherUriV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given publisher"));
			return false;
		}
	}

	// search by dates
	if (Query.DateStart != "" || Query.DateEnd != "") {
		PRecSet RecSet; bool Ok = FindEventsForDates(Query.DateStart, Query.DateEnd, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given time constraints"));
			return false;
		}
	}

	// search by languages
	if (Query.LangV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindEventsForLangs(Query.LangV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the specified language(s)"));
			return false;
		}
	}

	// search by categories
	if (Query.CategoryV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindEventsForCategories(Query.CategoryV, Query.CategoryIncludeSub, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No events match the specified categories"));
			return false;
		}
	}

	if (Query.DateMentionV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindEventsForDateMentions(Query.DateMentionV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No events match the specified date mentions"));
			return false;
		}
	}

	// search by article count
	if (Query.MinMaxArtsInEvent.Val1 > 0 || Query.MinMaxArtsInEvent.Val2 != TInt::Mx) {
		PRecSet RecSet; bool Ok = FindEventsForCoverage(Query.MinMaxArtsInEvent.Val1, Query.MinMaxArtsInEvent.Val2, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No events match the specified article counts"));
			return false;
		}
	}

	// if we already have no results then ignore the ignore conditions since they always just remove some elements
	if (EventRecSet.Empty() || EventRecSet->GetRecs() == 0) {
		EventRecSet = TRecSet::New(NewsBase->EventStore);
		return true;
	}

	// find the events to ignore and remove them from the results
	if (Query.IgnoreConceptUriV.Len() > 0) {
		// ignored concepts have to be processed one by one otherwise we would only remove events with ALL ignored concepts
		for (int N = 0; N < Query.IgnoreConceptUriV.Len(); N++) {
			PRecSet IgnoreSet; bool Ok = FindEventsForConcept(Query.IgnoreConceptUriV[N], IgnoreSet, Error);
			if (!Ok) {
				ErrorJson = TJsonVal::NewObj("error", Error);
				return false;
			}
			NewsBase->RemoveRecSets(EventRecSet, IgnoreSet);
		}
	}
	
	if (Query.IgnoreKeywords != "") {
		TStrV IgnoreKwV; Query.IgnoreKeywords.SplitOnAllCh(' ', IgnoreKwV);
		for (int N = 0; N < IgnoreKwV.Len(); N++) {
			PRecSet IgnoreSet; bool Ok = FindEventsForKeywords(IgnoreKwV[N], IgnoreSet, Error);
			if (Ok)
				NewsBase->RemoveRecSets(EventRecSet, IgnoreSet);
		}
	}

	if (Query.IgnorePublisherUriV .Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindEventsForPublishers(Query.IgnorePublisherUriV, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(EventRecSet, IgnoreSet);
	}

	if (Query.IgnoreLocationUriV.Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindEventsForLocations(Query.IgnoreLocationUriV, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(EventRecSet, IgnoreSet);
	}

	if (Query.IgnoreCategoryV.Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindEventsForCategories(Query.IgnoreCategoryV, Query.IgnoreCategoryIncludeSub, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(EventRecSet, IgnoreSet);
	}

	if (Query.IgnoreLangV.Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindEventsForLangs(Query.IgnoreLangV, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(EventRecSet, IgnoreSet);
	}

	RecSet = EventRecSet;
	return true;
}

bool TNewsSearch::SearchArticles(const TArticleQuery& Query, PRecSet& RecSet, PJsonVal& ErrorJson) const
{
	ErrorJson = NULL;
	RecSet = NULL;
	TStr Error;
	PRecSet ArticleRecSet;
	// search by concepts
	if (Query.ConceptUriV.Len() > 0) {
		// we call FindStoriesForConcepts instead of FindEvensForConcepts intentionally
		// this way we have for the stories Fqs that we can use for storing the stories
		bool Ok = FindArticlesForConcepts(Query.ConceptUriV, ArticleRecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (ArticleRecSet->GetRecs() == 0) {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given topics/concepts"));
			return false;
		}
	}

	// search by keywords
	if (Query.Keywords != "") {
		PRecSet RecSet; bool Ok = FindArticlesForKeywords(Query.Keywords, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(ArticleRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given keywords"));
			return false;
		}
	}

	// search by location
	if (Query.LocationUriV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindArticlesForLocations(Query.LocationUriV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(ArticleRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given location"));
			return false;
		}
	}

	// search by publisher
	if (Query.PublisherUriV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindArticlesForPublishers(Query.PublisherUriV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(ArticleRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given publisher"));
			return false;
		}
	}

	// search by languages
	if (Query.LangV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindArticlesForLangs(Query.LangV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(ArticleRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the specified language(s)"));
			return false;
		}
	}

	// search by categories
	if (Query.CategoryV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindArticlesForCategories(Query.CategoryV, Query.CategoryIncludeSub, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(ArticleRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the specified categories"));
			return false;
		}
	}

	// search by dates
	if (Query.DateStart != "" || Query.DateEnd != "") {
		PRecSet RecSet; bool Ok = FindArticlesForDates(Query.DateStart, Query.DateEnd, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(ArticleRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given time constraints"));
			return false;
		}
	}

	// search by date mentions
	if (Query.DateMentionV.Len() > 0) {
		PRecSet RecSet; bool Ok = FindArticlesForDateMentions(Query.DateMentionV, RecSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		if (RecSet->GetRecs() > 0)
			NewsBase->IntersectRecSets(ArticleRecSet, RecSet, false);
		else {
			ErrorJson = TJsonVal::NewObj("error", TStr("No results match the given date mentions"));
			return false;
		}
	}

	// if we already have no results then ignore the ignore conditions since they always just remove some elements
	if (ArticleRecSet.Empty() || ArticleRecSet->GetRecs() == 0) {
		ErrorJson = TJsonVal::NewObj("error", TStr("No results match the criteria"));
		return false;
	}

	// find the articles to ignore and remove them from the results
	if (Query.IgnoreConceptUriV.Len() > 0) {
		// ignored concepts have to be processed one by one otherwise we would only remove articles with ALL ignored concepts
		for (int N = 0; N < Query.IgnoreConceptUriV.Len(); N++) {
			PRecSet IgnoreSet; bool Ok = FindArticlesForConcept(Query.IgnoreConceptUriV[N], IgnoreSet, Error);
			if (!Ok) {
				ErrorJson = TJsonVal::NewObj("error", Error);
				return false;
			}
			NewsBase->RemoveRecSets(ArticleRecSet, IgnoreSet);
		}
	}

	if (Query.IgnoreKeywords != "") {
		TStrV IgnoreKwV; Query.IgnoreKeywords.SplitOnAllCh(' ', IgnoreKwV);
		for (int N = 0; N < IgnoreKwV.Len(); N++) {
			PRecSet IgnoreSet; bool Ok = FindArticlesForKeywords(IgnoreKwV[N], IgnoreSet, Error);
			if (Ok)
				NewsBase->RemoveRecSets(ArticleRecSet, IgnoreSet);
		}
	}

	if (Query.IgnorePublisherUriV.Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindArticlesForPublishers(Query.IgnorePublisherUriV, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(ArticleRecSet, IgnoreSet);
	}

	if (Query.IgnoreLocationUriV.Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindArticlesForLocations(Query.IgnoreLocationUriV, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(ArticleRecSet, IgnoreSet);
	}

	if (Query.IgnoreCategoryV.Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindArticlesForCategories(Query.IgnoreCategoryV, Query.IgnoreCategoryIncludeSub, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(ArticleRecSet, IgnoreSet);
	}

	if (Query.IgnoreLangV.Len() > 0) {
		PRecSet IgnoreSet; bool Ok = FindArticlesForLangs(Query.IgnoreLangV, IgnoreSet, Error);
		if (!Ok) {
			ErrorJson = TJsonVal::NewObj("error", Error);
			return false;
		}
		NewsBase->RemoveRecSets(ArticleRecSet, IgnoreSet);
	}

	RecSet = ArticleRecSet;
	return true;
}

// find articles that contain articles with ALL specified keywords
bool TNewsSearch::FindArticlesForKeywords(const TStr& Keywords, PRecSet& RecSet, TStr& Error) const
{
	// if we are not indexing text then return error
	int FieldKeyId = ArticleStore->GetFieldKeyId(ArticleStore->GetFieldId("Body"));
	if (!IndexVoc->IsKeyId(FieldKeyId)) {
		Error = "Article body does not have the body key. Unable to do keyword search";
		return false;
	}

	TQueryItem KwQuery(Base, FieldKeyId, Keywords);
	TPair<TBool, PRecSet> Results = Index->Search(Base, KwQuery, Index->GetDefMerger());

	RecSet = Results.Val2;
	return true;
}

bool TNewsSearch::FindArticlesForConcept(const TStr& ConceptUri, PRecSet& RecSet, TStr& Error) const
{
	if (!ConceptStore->IsRecNm(ConceptUri)) {
		Error = TStr::Fmt("Unknown concept '%s'", ConceptUri.CStr());
		return false;
	}
	// if the concept is a concept folder then potentially update the events assigned to the folder
	uint64 ConceptId = ConceptStore->GetRecId(ConceptUri);
	if (ConceptStore->GetFieldNmStr(ConceptId, "Type") == NewsConst.TypeConceptFolder) {
		if (NewsBase->ConceptFolderStore->IsRecNm(ConceptUri) == false) {
			Error = TStr::Fmt("Unknown concept folder '%s'", ConceptUri.CStr());
			return false;
		}
		ConceptFolderBase->UpdateConceptFolderArticles(ConceptUri);
		RecSet = NewsBase->ConceptFolderStore->GetRec(ConceptUri).DoJoin(Base, "hasArticle");
	}
	else
		RecSet = ConceptStore->GetRec(ConceptUri).DoJoin(Base, "hasArticle");
		
	return true;
}

bool TNewsSearch::FindArticlesForConcepts(const TStrV& ConceptUriV, PRecSet& RecSet, TStr& Error) const
{
	PRecSet ArticleRecSet;
	for (int ConceptN = 0; ConceptN < ConceptUriV.Len(); ConceptN++) {
		TStr ConceptUri = ConceptUriV[ConceptN];
		PRecSet ConceptRecSet;
		if (!FindArticlesForConcept(ConceptUri, ConceptRecSet, Error))		// if concept is not known then set the error and return
			return false;
		NewsBase->IntersectRecSets(ArticleRecSet, ConceptRecSet);
	}
	RecSet = ArticleRecSet;
	return true;
}

bool TNewsSearch::FindArticlesForLocations(const TStrV& LocationUriV, PRecSet& RecSet, TStr& Error) const
{
	TUInt64V CityRecIdV;
	for (int N = 0; N < LocationUriV.Len(); N++) {
		const TStr LocationUri = LocationUriV[N];
		if (!ConceptStore->IsRecNm(LocationUri)) {
			Error = TStr::Fmt("Location '%s' does not exist in the concept store", LocationUri.CStr());
			return false;
		}
		if (NewsBase->GeoBase->IsCountry(LocationUri)) {
			TStrV CityUriV; NewsBase->GeoBase->GetCitiesForCountry(LocationUri, CityUriV);
			for (int N = 0; N < CityUriV.Len(); N++) {
				if (ConceptStore->IsRecNm(CityUriV[N]))
					CityRecIdV.Add(ConceptStore->GetRecId(CityUriV[N]));
			}
		}
		else
			CityRecIdV.Add(ConceptStore->GetRecId(LocationUri));
	}
	
	PRecSet CityRecSet = TRecSet::New(ConceptStore, CityRecIdV);
	RecSet = CityRecSet->DoJoin(Base, "hasArticleForLocation");
	return true;
}

bool TNewsSearch::FindArticlesForPublisher(const TStr& PublisherUri, PRecSet& RecSet, TStr& Error) const
{
	if (!SourceStore->IsRecNm(PublisherUri)) {
		Error = TStr::Fmt("Unknown source '%s'", PublisherUri.CStr());
		return false;
	}
	// if the concept is a concept folder then potentially update the events assigned to the folder
	uint64 SourceId = SourceStore->GetRecId(PublisherUri);
	RecSet = SourceStore->GetRec(SourceId).DoJoin(Base, "hasArticle");
	return true;
}

bool TNewsSearch::FindArticlesForPublishers(const TStrV& PublisherUriV, PRecSet& RecSet, TStr& Error) const
{
	PRecSet PubRecSet;
	for (int N = 0; N < PublisherUriV.Len(); N++) {
		TStr PublisherUri = PublisherUriV[N];
		PRecSet PublisherRecSet;
		if (!FindArticlesForPublisher(PublisherUri, PublisherRecSet, Error))		// if publisher is not known then set the error and return
			return false;
		NewsBase->IntersectRecSets(PubRecSet, PublisherRecSet);
	}
	RecSet = PubRecSet;
	return true;
}

bool TNewsSearch::FindArticlesForDates(const TStr& StartDate, const TStr& EndDate, PRecSet& RecSet, TStr& Error) const
{
	if (!NewsBase->IndexVoc->IsKeyNm(ArticleStoreId, "Date")) {
		Error = "Unable to search by dates since date is not indexed using a key";
		return false;
	}
	const int DateKeyId = NewsBase->IndexVoc->GetKeyId(ArticleStoreId, "Date");
	uint64 Dates = NewsBase->IndexVoc->GetWords(DateKeyId);
	TQueryItemV QueryItemDateV;
	for (uint64 DateId = 0; DateId < Dates; DateId++) {
		TStr DateStr = NewsBase->IndexVoc->GetWordStr(DateKeyId, DateId);
		// ignore the empty date value
		if (DateStr == "")
			continue;
		if (StartDate != "" && DateStr < StartDate)
			continue;
		if (EndDate != "" && DateStr > EndDate)
			continue;
		// we search for stories that have extracted date equal to this value
		QueryItemDateV.Add(TQueryItem(NewsBase->Base, DateKeyId, DateId));
	}
	if (QueryItemDateV.Len() > 0)
		RecSet = NewsBase->Base->Search(TQueryItem(oqitOr, QueryItemDateV));
	else
		RecSet = TRecSet::New(ArticleStore);
	return true;
}

bool TNewsSearch::FindArticlesForLangs(const TStrV& LangV, PRecSet& RecSet, TStr& Error) const
{
	if (!NewsBase->IndexVoc->IsKeyNm(ArticleStoreId, "Language")) {
		Error = "Unable to search by language since it is not indexed using a key";
		return false;
	}
	const int LangKeyId = NewsBase->IndexVoc->GetKeyId(ArticleStoreId, "Language");
	TQueryItemV QueryItemLangV;
	for (int LangN = 0; LangN < LangV.Len(); LangN++) {
		if (!NewsBase->IndexVoc->IsWordStr(LangKeyId, LangV[LangN]))
			continue;
		QueryItemLangV.Add(TQueryItem(NewsBase->Base, LangKeyId, NewsBase->IndexVoc->GetWordId(LangKeyId, LangV[LangN])));
	}
	if (QueryItemLangV.Len() > 0)
		RecSet = NewsBase->Base->Search(TQueryItem(oqitOr, QueryItemLangV));
	else
		RecSet = TRecSet::New(ArticleStore);
	return true;
}

bool TNewsSearch::FindArticlesForCategories(const TStrV& CategoryUriV, const bool& IncludeSubCategories, PRecSet& RecSet, TStr& Error) const
{
	TUInt64V CatRecIdV;
	for (int N = 0; N < CategoryUriV.Len(); N++) {
		if (!NewsBase->CategoryStore->IsRecNm(CategoryUriV[N])) {
			Error = TStr::Fmt("Specified category '%s' is not known", CategoryUriV[N].CStr());
			return false;
		}
		CatRecIdV.Add(NewsBase->CategoryStore->GetRecId(CategoryUriV[N]));
	}

	if (IncludeSubCategories) {
		const TUInt64Set CatRecIdSet = NewsBase->CategoryBase->GetAllSubCategories(CatRecIdV);
		CatRecIdSet.GetKeyV(CatRecIdV);
	}
	// category rec set
	PRecSet CatRecSet = TRecSet::New(CategoryStore, CatRecIdV);
	// join articles for those categories
	RecSet = CatRecSet->DoJoin(Base, "hasArticle");
	return true;
}

bool TNewsSearch::FindArticlesForDateMentions(const TStrV& DateMentionV, PRecSet& RecSet, TStr& Error) const
{
	if (NewsBase->DateStore.Empty()) {
		Error = "Date store is not set up. There are no detected date mentions.";
		return false;
	}

	TUInt64V DateIdV;
	for (int N = 0; N < DateMentionV.Len(); N++) {
		if (NewsBase->DateStore->IsRecNm(DateMentionV[N]))
			DateIdV.Add(NewsBase->DateStore->GetRecId(DateMentionV[N]));
	}
	// rec set with searched dates
	PRecSet DateSet = TRecSet::New(NewsBase->DateStore, DateIdV);
	// articles that mention the specified dates
	RecSet = DateSet->DoJoin(Base, "hasArticle");
	return true;
}

// ///////////////
// SEARCH STORIES

// find stories that contain articles with the specified keywords
// method is a bit complicated since it computes a score for each story based on how often does the search term appear in the articles in the story
bool TNewsSearch::FindStoriesForKeywords(const TStr& Keywords, PRecSet& RecSet, TStr& Error) const
{
	PRecSet ArtRecSet; TStr ArtError;
	bool Ok = FindArticlesForKeywords(Keywords, ArtRecSet, ArtError);
	if (!Ok) {
		Error = ArtError;
		return false;
	}
	// count the number of times the matching articles point to the same story
	TRecDict StoryRecH;
	const uint64 Arts = ArtRecSet->GetRecs();
	for (int N = 0; N < Arts; N++) {
		const uint64 StoryId = ArticleStore->GetFieldNmUInt64(ArtRecSet->GetRecId(N), "hasStoryId");
		if (StoryId != TUInt64::Mx)
			StoryRecH.AddDat(StoryId)++;
	}
	
	TUInt64IntKdV StoryIdFqV;
	for (int KeyId = StoryRecH.FFirstKeyId(); StoryRecH.FNextKeyId(KeyId);) {
		const uint64 StoryId = StoryRecH.GetKey(KeyId);
		int Fq = StoryRecH[KeyId];
		int Arts = StoryStore->GetFieldNmInt(StoryId, "ArticleCount");
		if (Arts > 100) {
			Fq = int(Fq * 100.0 / double(Arts));
			Arts = 100;
		}
		// the story score should be between 0 - 100. 100 will be in the case that all the articles mention the search term
		/*double Score = 100 * Fq / (double) MAX(1, Arts);*/
		double Score = Fq;
		StoryIdFqV.Add(TUInt64IntKd(StoryId, (int) Score));
	}
	RecSet = TRecSet::New(StoryStore, StoryIdFqV, true);
	return true;
}

bool TNewsSearch::FindStoriesForConcept(const TStr& ConceptUri, PRecSet& RecSet, TStr& Error) const
{
	if (!ConceptStore->IsRecNm(ConceptUri)) {
		Error = TStr::Fmt("Unknown concept '%s'", ConceptUri.CStr());
		return false;
	}
	RecSet = ConceptStore->GetRec(ConceptUri).DoJoin(Base, "hasStory");
	return true;
}

bool TNewsSearch::FindStoriesForConcepts(const TStrV& ConceptUriV, PRecSet& RecSet, TStr& Error) const
{
	PRecSet StoryRecSet;
	for (int ConceptN = 0; ConceptN < ConceptUriV.Len(); ConceptN++) {
		const TStr ConceptUri = ConceptUriV[ConceptN];
		if (!ConceptStore->IsRecNm(ConceptUri)) {
			Error = TStr::Fmt("Unknown concept '%s'", ConceptUri.CStr());
			return false;
		}
		const PRecSet RecSet = ConceptStore->GetRec(ConceptUriV[ConceptN]).DoJoin(Base, "hasStory");
		NewsBase->IntersectRecSets(StoryRecSet, RecSet);
	}
	if (StoryRecSet.Empty())
		StoryRecSet = TRecSet::New(StoryStore);
	RecSet = StoryRecSet;
	return true;
}

bool TNewsSearch::FindStoriesForDates(const TStr& StartDate, const TStr& EndDate, PRecSet& RecSet, TStr& Error) const
{
	if (!IndexVoc->IsKeyNm(StoryStoreId, "StoryDate") || !IndexVoc->IsKeyNm(StoryStoreId, "StoryDateEnd")) {
		Error = "StoryDate and StoryDateEnd are not valid keys in the Story store. Unable to search stories by date.";
		return false;
	}
	const int StoryDateKeyId = IndexVoc->GetKeyId(StoryStoreId, "StoryDate");
	const int StoryDateEndKeyId = IndexVoc->GetKeyId(StoryStoreId, "StoryDateEnd");
	uint64 Dates = IndexVoc->GetWords(StoryDateKeyId);
	TQueryItemV QueryItemDateV;
	for (uint64 DateId = 0; DateId < Dates; DateId++) {
		TStr DateStr = IndexVoc->GetWordStr(StoryDateKeyId, DateId);
		// ignore the empty date value
		if (DateStr == "")
			continue;
		if (StartDate != "" && DateStr < StartDate)
			continue;
		if (EndDate != "" && DateStr > EndDate)
			continue;
		// we search for stories that have extracted date equal to this value
		QueryItemDateV.Add(TQueryItem(Base, StoryDateKeyId, DateId));
		// if extracted date was not set then we are satisfied also with the average date of the article in the Story
		QueryItemDateV.Add(TQueryItem(Base, StoryDateEndKeyId, DateId));
	}
	if (QueryItemDateV.Len() > 0) {
		RecSet = Base->Search(TQueryItem(oqitOr, QueryItemDateV));
		RecSet = SetStoryScoresBasedOnStoryQuality(RecSet);
	}
	else
		RecSet = TRecSet::New(StoryStore);
	return true;
}

// ///////////////
// SEARCH EVENTS

// find events that contain articles with the specified keywords
bool TNewsSearch::FindEventsForKeywords(const TStr& Keywords, PRecSet& RecSet, TStr& Error) const
{
	PRecSet StorySet; TStr StoryError;
	bool Ok = FindStoriesForKeywords(Keywords, StorySet, StoryError);
	if (!Ok) {
		Error = StoryError;
		return false;
	}
	THash<TUInt64, TInt> EventIdWgtH;
	THash<TUInt64, TInt> EventIdStoryCntH;
	for (int N = 0; N < StorySet->GetRecs(); N++) {
		uint64 StoryId = StorySet->GetRecId(N);
		TRec EventRec = StoryStore->GetRec(StoryId).DoSingleJoin(Base, "hasEvent");
		if (EventRec.IsDef()) {
			const uint64 EventId = EventRec.GetRecId();
			EventIdWgtH.AddDat(EventId) += StorySet->GetRecFq(N);
			EventIdStoryCntH.AddDat(EventId) += 1;
		}
	}
	// normalize event scores by story counts
	for (int KeyId = EventIdWgtH.FFirstKeyId(); EventIdWgtH.FNextKeyId(KeyId);)
		EventIdWgtH[KeyId] = int(EventIdWgtH[KeyId] / (double) EventIdStoryCntH[KeyId]);

	TUInt64IntKdV EventIdWgtKdV; EventIdWgtH.GetKeyDatKdV(EventIdWgtKdV);
	RecSet = TRecSet::New(EventStore, EventIdWgtKdV, true);
	return true;
}

bool TNewsSearch::FindEventsForConcept(const TStr& ConceptUri, PRecSet& RecSet, TStr& Error) const
{
	if (!ConceptStore->IsRecNm(ConceptUri)) {
		Error = TStr::Fmt("Unknown concept '%s'", ConceptUri.CStr());
		return false;
	}
	// if the concept is a concept folder then potentially update the events assigned to the folder
	const uint64 ConceptId = ConceptStore->GetRecId(ConceptUri);
	if (ConceptStore->GetFieldNmStr(ConceptId, "Type") == NewsConst.TypeConceptFolder) {
		if (NewsBase->ConceptFolderStore->IsRecNm(ConceptUri) == false) {
			Error = TStr::Fmt("Unknown concept folder '%s'", ConceptUri.CStr());
			return false;
		}
		ConceptFolderBase->UpdateConceptFolderEvents(ConceptUri);
		RecSet = NewsBase->ConceptFolderStore->GetRec(ConceptUri).DoJoin(Base, "hasEvent");
	}
	else 
		RecSet = ConceptStore->GetRec(ConceptUri).DoJoin(Base, "hasEvent");
	return true;
}

bool TNewsSearch::FindEventsForConcepts(const TStrV& ConceptUriV, PRecSet& RecSet, TStr& Error) const
{
	PRecSet EventRecSet;
	for (int ConceptN = 0; ConceptN < ConceptUriV.Len(); ConceptN++) {
		TStr ConceptUri = ConceptUriV[ConceptN];
		PRecSet ConceptRecSet;
		if (!FindEventsForConcept(ConceptUri, ConceptRecSet, Error))
			return false;
		NewsBase->IntersectRecSets(EventRecSet, ConceptRecSet);
	}
	if (EventRecSet.Empty())
		EventRecSet = TRecSet::New(EventStore);
	RecSet = EventRecSet;
	return true;
}

// for the given start and end date find the valid dates in between (including StartDate and EndDate)
// and find stories that either start or end between these dates
// TODO: if the event starts before StartDate and ends after EndDate we won't find it. this is a bug and should be fixed
bool TNewsSearch::FindEventsForDates(const TStr& StartDate, const TStr& EndDate, PRecSet& RecSet, TStr& Error) const
{
	if (!IndexVoc->IsKeyNm(EventStoreId, "EventDate") || !IndexVoc->IsKeyNm(EventStoreId, "EventDateEnd")) {
		Error = "EventDate and EventDateEnd are not valid keys in the Event store. Unable to search events by date.";
		return false;
	}
	const int EventDateKeyId = IndexVoc->GetKeyId(EventStoreId, "EventDate");
	const int EventDateEndKeyId = IndexVoc->GetKeyId(EventStoreId, "EventDateEnd");
	uint64 Dates = IndexVoc->GetWords(EventDateKeyId);
	TQueryItemV QueryItemDateV;
	for (uint64 DateId = 0; DateId < Dates; DateId++) {
		TStr DateStr = IndexVoc->GetWordStr(EventDateKeyId, DateId);
		// ignore the empty date value
		if (DateStr == "")
			continue;
		if (StartDate != "" && DateStr < StartDate)
			continue;
		if (EndDate != "" && DateStr > EndDate)
			continue;
		// we search for stories that have extracted date equal to this value
		QueryItemDateV.Add(TQueryItem(Base, EventDateKeyId, DateId));
		// if extracted date was not set then we are satisfied also with the average date of the article in the Event
		QueryItemDateV.Add(TQueryItem(Base, EventDateEndKeyId, DateId));
	}
	if (QueryItemDateV.Len() > 0) {
		RecSet = Base->Search(TQueryItem(oqitOr, QueryItemDateV));
		RecSet = SetEventScoresBasedOnStoryQuality(RecSet);
	}
	else
		RecSet = TRecSet::New(EventStore);
	return true;
}

// return events that have stories from at least one of the languages in LangV
bool TNewsSearch::FindEventsForLangs(const TStrV& LangV, PRecSet& RecSet, TStr& Error) const
{
	if (!IndexVoc->IsKeyNm(StoryStoreId, "Language")) {
		Error = "Language is not a valid key for the Story store. Unable to find events by language";
		return false;
	}
	const int LangKeyId = IndexVoc->GetKeyId(StoryStoreId, "Language");
	TQueryItemV QueryItemLangV;
	for (int LangN = 0; LangN < LangV.Len(); LangN++) {
		if (!IndexVoc->IsWordStr(LangKeyId, LangV[LangN]))
			continue;
		QueryItemLangV.Add(TQueryItem(Base, LangKeyId, IndexVoc->GetWordId(LangKeyId, LangV[LangN])));
	}
	if (QueryItemLangV.Len() > 0) {
		PRecSet StorySet = Base->Search(TQueryItem(oqitOr, QueryItemLangV));
		RecSet = StorySet->DoJoin(Base, "hasEvent");
		RecSet = SetEventScoresBasedOnStoryQuality(RecSet);
	}
	else
		RecSet = TRecSet::New(EventStore);
	return true;
}

bool TNewsSearch::FindEventsForLocations(const TStrV& LocationUriV, PRecSet& RecSet, TStr& Error) const
{
	TUInt64V CityRecIdV;
	for (int N = 0; N < LocationUriV.Len(); N++) {
		const TStr LocationUri = LocationUriV[N];
		if (!ConceptStore->IsRecNm(LocationUri)) {
			Error = TStr::Fmt("Location '%s' does not exist in the concept store", LocationUri.CStr());
			return false;
		}
		if (NewsBase->GeoBase->IsCountry(LocationUri)) {
			TStrV CityUriV; NewsBase->GeoBase->GetCitiesForCountry(LocationUri, CityUriV);
			for (int N = 0; N < CityUriV.Len(); N++) {
				if (ConceptStore->IsRecNm(CityUriV[N]))
					CityRecIdV.Add(ConceptStore->GetRecId(CityUriV[N]));
			}
		}
		else
			CityRecIdV.Add(ConceptStore->GetRecId(LocationUri));
	}

	PRecSet CityRecSet = TRecSet::New(ConceptStore, CityRecIdV);
	RecSet = CityRecSet->DoJoin(Base, "hasEventForLocation");
	RecSet = SetEventScoresBasedOnStoryQuality(RecSet);
	return true;
}

bool TNewsSearch::FindEventsForPublisher(const TStr& PublisherUri, PRecSet& RecSet, TStr& Error) const
{
	if (!SourceStore->IsRecNm(PublisherUri)) {
		Error = TStr::Fmt("Unknown source '%s'", PublisherUri.CStr());
		return false;
	}
	// if the concept is a concept folder then potentially update the events assigned to the folder
	const uint64 SourceId = SourceStore->GetRecId(PublisherUri);
	const PRecSet ArticlesFromPublisher = SourceStore->GetRec(SourceId).DoJoin(Base, "hasArticle");
	const PRecSet StoriesFromPublisher = ArticlesFromPublisher->DoJoin(Base, "hasStory");
	RecSet = StoriesFromPublisher->DoJoin(Base, "hasEvent");
	return true;
}

bool TNewsSearch::FindEventsForPublishers(const TStrV& PublisherUriV, PRecSet& RecSet, TStr& Error) const
{
	PRecSet PubRecSet;
	for (int N = 0; N < PublisherUriV.Len(); N++) {
		const TStr PublisherUri = PublisherUriV[N];
		PRecSet PublisherRecSet;
		if (!FindEventsForPublisher(PublisherUri, PublisherRecSet, Error))		// if publisher is not known then set the error and return
			return false;
		NewsBase->IntersectRecSets(PubRecSet, PublisherRecSet);
	}
	RecSet = PubRecSet;
	return true;
}

bool TNewsSearch::FindEventsForCategories(const TStrV& CategoryUriV, const bool& IncludeSubCategories, PRecSet& RecSet, TStr& Error) const
{
	TUInt64V CatRecIdV;
	for (int N = 0; N < CategoryUriV.Len(); N++) {
		if (!NewsBase->CategoryStore->IsRecNm(CategoryUriV[N])) {
			Error = TStr::Fmt("Specified category '%s' is not known", CategoryUriV[N].CStr());
			return false;
		}
		CatRecIdV.Add(NewsBase->CategoryStore->GetRecId(CategoryUriV[N]));
	}

	if (IncludeSubCategories) {
		const TUInt64Set CatRecIdSet = NewsBase->CategoryBase->GetAllSubCategories(CatRecIdV);
		CatRecIdSet.GetKeyV(CatRecIdV);
	}

	// category rec set
	PRecSet CatRecSet = TRecSet::New(CategoryStore, CatRecIdV);
	// join articles for those categories
	RecSet = CatRecSet->DoJoin(Base, "hasEvent");
	RecSet = SetEventScoresBasedOnStoryQuality(RecSet);
	return true;
}

bool TNewsSearch::FindEventsForDateMentions(const TStrV& DateMentionV, PRecSet& RecSet, TStr& Error) const
{
	if (NewsBase->DateStore.Empty()) {
		Error = "Date store is not set up. There are no detected date mentions.";
		return false;
	}

	TUInt64V DateIdV;
	for (int N = 0; N < DateMentionV.Len(); N++) {
		if (NewsBase->DateStore->IsRecNm(DateMentionV[N]))
			DateIdV.Add(NewsBase->DateStore->GetRecId(DateMentionV[N]));
	}
	// rec set with searched dates
	PRecSet DateSet = TRecSet::New(NewsBase->DateStore, DateIdV);
	PRecSet ArticleSet = DateSet->DoJoin(Base, "hasArticle");
	PRecSet StorySet = ArticleSet->DoJoin(Base, "hasStory");
	RecSet = StorySet->DoJoin(Base, "hasEvent");
	RecSet = SetEventScoresBasedOnStoryQuality(RecSet);
	return true;
}


// find events that is covered between MinArtCount and MaxArtCount
bool TNewsSearch::FindEventsForCoverage(const int& MinArtCount, const int& MaxArtCount, PRecSet& RecSet, TStr& Error) const
{
	const int ArticleCountKeyId = IndexVoc->GetKeyId(EventStoreId, "ArticleCountKey");
	const uint64 Counts = IndexVoc->GetWords(ArticleCountKeyId);
	TQueryItemV QueryItemDateV;
	for (uint64 CountId = 0; CountId < Counts; CountId++) {
		const TStr CountStr = IndexVoc->GetWordStr(ArticleCountKeyId, CountId);
		const int Count = CountStr.GetInt(0);
		if (Count >= MinArtCount && Count <= MaxArtCount)
			QueryItemDateV.Add(TQueryItem(Base, ArticleCountKeyId, CountId));
	}
	// find events for counts
	if (QueryItemDateV.Len() == 0)
		RecSet = TRecSet::New(StoryStore);
	else {
		RecSet = Base->Search(TQueryItem(oqitOr, QueryItemDateV));
		RecSet = SetEventScoresBasedOnStoryQuality(RecSet);
	}
	return true;
}

// utility method
// date mentions can be specificed as a series of dateMention=XXX&dateMention=YYY&...
// or as a range with the starting and ending date - usually as dateMentionStart=XXX&dateMentionEnd=YYY
TStrV TNewsSearch::ParseDateMentions(const TStrV& DateMentionV, const TStr& DateMentionStart, const TStr& DateMentionEnd) const
{
	TStrV DateV;
	if (NewsBase->DateStore.Empty())
		return DateV;
	if (DateMentionV.Len() > 0)
		DateV.AddV(DateMentionV);
	if (DateMentionStart != "" && DateMentionEnd != "" && !NewsBase->DateStore.Empty()) {
		PStoreIter Iter = NewsBase->DateStore->GetIter();
		while (Iter->Next()) {
			TStr Date = NewsBase->DateStore->GetRecNm(Iter->GetRecId());
			if (Date >= DateMentionStart && Date <= DateMentionEnd)
				DateV.Add(Date);
		}
	}
	return DateV;
}

// some event queries return list of events that don't have an inherent sorting (like when searching for events based on dates
// in such cases we would like to set Fq for each event based on the quality of the top story in the event
// the top story in the event should always be the hasPrimaryStory since multiple stories should always be merged to the event that has the primary story with the highest quality
PRecSet TNewsSearch::SetEventScoresBasedOnStoryQuality(const PRecSet& RecSet) const
{
	EAssertR(RecSet->GetStoreId() == NewsBase->EventStoreId, "Expected records to be for the event store.");
	TUInt64IntKdV RecIdFqV;
	for (int N = 0; N < RecSet->GetRecs(); N++) {
		const uint64 EventId = RecSet->GetRecId(N);
		const uint64 StoryId = EventStore->GetFieldNmUInt64(EventId, "hasPrimaryStoryId");
		if (StoryId == TUInt64::Mx)
			continue;
		const int Fq = NewsBase->GetStoryQuality(StoryId);
		RecIdFqV.Add(TUInt64IntKd(EventId, Fq));
	}
	const PRecSet NewEventSet = TRecSet::New(EventStore, RecIdFqV, true);
	return NewEventSet;
}

PRecSet TNewsSearch::SetStoryScoresBasedOnStoryQuality(const PRecSet& RecSet) const
{
	EAssertR(RecSet->GetStoreId() == NewsBase->StoryStoreId, "Expected records to be for the story store.");
	TUInt64IntKdV RecIdFqV;
	for (int N = 0; N < RecSet->GetRecs(); N++) {
		const uint64 StoryId = RecSet->GetRecId(N);
		const int Fq = NewsBase->GetStoryQuality(StoryId);
		RecIdFqV.Add(TUInt64IntKd(StoryId, Fq));
	}
	const PRecSet NewStorySet = TRecSet::New(StoryStore, RecIdFqV, true);
	return NewStorySet;
}