/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include <base.h>
#include "NewsSrvUtil.h"
#include <qminer_srv.h>

EnumArticleInfoFlags DefArticleFlags = EnumArticleInfoFlags(aifBasicInfo | aifIncludeBody | aifIncludeTitle | aifIncludeSourceInfo | aifIncludeEventUri );
EnumStoryInfoFlags DefStoryFlags = EnumStoryInfoFlags(sifIncludeBasicStats | sifIncludeCategory | sifIncludeLocation | sifIncludeStoryDate);
EnumEventInfoFlags DefEventFlags = EnumEventInfoFlags(eifIncludeArticleCounts | eifIncludeConcepts | eifIncludeMultiLingInfo | eifIncludeCategories | eifIncludeLocation);
EnumPublisherInfoFlags DefPublisherFlags = EnumPublisherInfoFlags(pifIncludeTitle | pifIncludeImportance);
EnumConceptInfoFlags DefConceptFlags = EnumConceptInfoFlags();
EnumConceptInfoFlags ConceptFlagsFull = EnumConceptInfoFlags(picIncludePrimImage | picIncludeAllImages | picIncludeDescription | picIncludeTrends);

///////////////////////////////////////////////
// PlusMachines-Server-Helper-Utilities
TNewsSrvUtil::TNewsSrvUtil(const PNewsBase& _NewsBase) : NewsBase(_NewsBase)
{ 
	GeoBase = NewsBase->GeoBase;

	ArticleStore = NewsBase->ArticleStore;
	ConceptStore = NewsBase->ConceptStore;
	ConceptClassStore = NewsBase->ConceptClassStore;
	ConceptFolderStore = NewsBase->ConceptFolderStore;
	UserStore = NewsBase->UserStore;
	CategoryStore = NewsBase->CategoryStore;
	StoryStore = NewsBase->StoryStore;
	SourceStore = NewsBase->SourceStore;
	EventStore = NewsBase->EventStore;
	ImageStore = NewsBase->ImageStore;
	TagStore = NewsBase->TagStore;
	
	ArticleStoreId = NewsBase->ArticleStoreId;
	ConceptStoreId = NewsBase->ConceptStoreId;;
	ConceptClassStoreId = NewsBase->ConceptClassStoreId;
	ConceptFolderStoreId = NewsBase->ConceptFolderStoreId;
	UserStoreId = NewsBase->UserStoreId;
	CategoryStoreId = NewsBase->CategoryStoreId;
	SourceStoreId = NewsBase->SourceStoreId;
	StoryStoreId = NewsBase->StoryStoreId;
	EventStoreId = NewsBase->EventStoreId;
	ImageStoreId = NewsBase->ImageStoreId;
	TagStoreId = NewsBase->TagStoreId;

	Base = NewsBase->Base;
}

EnumTimeWindow TNewsSrvUtil::GetTimeWindow(const TStr& Text)
{
	TStr TextLc = Text.GetLc();
	if (Text == "total")
		return twTotal;
	else if (Text == "month")
		return twMonth;
	else if (Text == "day")
		return twDay;
	else if (Text == "hour")
		return twHour;
	else
		return twUndef;
}

// for a given article, return the article's DateTime formatted based on the TimeWindow (e.g. "total", "2014-02" or "2014-02-01" or "2014-02-01 03:02:01"
TStr TNewsSrvUtil::GetKeyForTimeWindow(const uint64& ArticleId, const EnumTimeWindow& TimeWindow) const
{
	TStr Date;
	if (TimeWindow == twTotal)
		Date = "total";
	else if (TimeWindow == twDay)
		Date = NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "Date");
	else {
		TTm Tm; NewsBase->ArticleStore->GetFieldNmTm(ArticleId, "DateTime", Tm);
		if (TimeWindow == twHour)
			Date = Tm.GetWebLogDateTimeStr(true, " ", false);
		else if (TimeWindow == twMonth) 
			Date = TStr::Fmt("%04d-%02d", Tm.GetYear(), Tm.GetMonth());
	}
	return Date;
}

// return detailed info about the article
PJsonVal TNewsSrvUtil::Article(const uint64& ArticleId, const int& MaxBodyLen, const EnumConceptType& IncludeConceptTypes, const TStrV& LabelLanguageV, const EnumArticleInfoFlags& ArticleFlags) const {
	PJsonVal ResJson = TJsonVal::NewObj();
	if (!ArticleStore->IsRecId(ArticleId))
		return ResJson;
	TRec ArticleRec = ArticleStore->GetRec(ArticleId);
	
	// always add id and uri
	ResJson->AddToObj("id", TUInt64::GetStr(ArticleId));
	ResJson->AddToObj("uri", ArticleStore->GetRecNm(ArticleId));
	
	const bool IsDuplicate = ArticleStore->GetFieldNmUInt64(ArticleId, "isDuplicateOfId") != TUInt64::Mx;

	// basic info
	if (ArticleFlags & aifBasicInfo) {
		ResJson->AddToObj("lang", ArticleStore->GetFieldNmStr(ArticleId, "Language"));
		ResJson->AddToObj("isDuplicate", TJsonVal::NewBool(IsDuplicate));	// is this article a duplicate of another one?

		const TFlt CosSim = ArticleStore->GetFieldNmFlt(ArticleId, "CosToCentroid");
		TTm Tm; ArticleStore->GetFieldNmTm(ArticleId, "DateTime", Tm);

		ResJson->AddToObj("date", TJsonVal::NewStr(Tm.IsDef() ? Tm.GetWebLogDateStr() : "[unknown]"));
		ResJson->AddToObj("time", TJsonVal::NewStr(Tm.IsDef() ? Tm.GetHMSTColonDotStr(true, false) : "[unknown]"));
		ResJson->AddToObj("sim", TJsonVal::NewNum(CosSim));
	}

	// title
	if (ArticleFlags & aifIncludeTitle)
		ResJson->AddToObj("title", ArticleStore->GetFieldNmStr(ArticleId, "Title"));
	// body
	if (ArticleFlags & aifIncludeBody && MaxBodyLen != 0) {
		TStr Body = ArticleStore->GetFieldNmStr(ArticleId, "Body");
		if (MaxBodyLen != -1) {
			int LenToTake = Body.SearchChBack(' ', MaxBodyLen);
			Body = Body.GetSubStr(0, LenToTake) + "...";
		}
		ResJson->AddToObj("body", Body);
	}
	
	// add source info
	if (ArticleFlags & aifIncludeSourceInfo) {
		TRec SourceRec = ArticleRec.DoSingleJoin(Base, "hasSource");
		const uint64 SourceId = SourceRec.GetRecId();
		TStr SourceTitle = SourceStore->GetFieldNmStr(SourceId, "Title");
		TStr SourceUri = SourceStore->GetFieldNmStr(SourceId, "URI");
		if (SourceTitle == "")
			SourceTitle = SourceUri.TrimLeft("http://").TrimLeft("https://").TrimLeft("www.");
		ResJson->AddToObj("sourceTitle", SourceTitle.GetUc());
		ResJson->AddToObj("sourceUri", SourceUri);
		//ResJson->AddToObj("sourceId", TUInt64(SourceId).GetStr());
		
		// add also the source location
		if (!NewsBase->SourceStore->IsFieldNmNull(SourceRec.GetRecId(), "Location")) {
			const TFltPr LocPr = NewsBase->SourceStore->GetFieldNmFltPr(SourceRec.GetRecId(), "Location");
			ResJson->AddToObj("sourceLat", TJsonVal::NewNum(LocPr.Val1));
			ResJson->AddToObj("sourceLong", TJsonVal::NewNum(LocPr.Val2));
		}
	}

	// add concepts
	if (ArticleFlags & aifIncludeConcepts && IncludeConceptTypes != ctNone && LabelLanguageV.Len() > 0) {
		PRecSet ConceptSet = ArticleRec.DoJoin(Base, "hasConcept");
		ResJson->AddToObj("concepts", Concepts(ConceptSet, TUInt64Set(), TInt::Mx, IncludeConceptTypes, LabelLanguageV));
	}
	
	// list of article uris that are duplicate of this article
	if (ArticleFlags & aifIncludeDuplicateList) {
		PRecSet ArticleSet = ArticleRec.DoJoin(Base, "hasDuplicate");
		ResJson->AddToObj("duplicates", ArticleUris(ArticleSet));
	}

	if (ArticleFlags & aifIncludeOriginalArticleInfo && IsDuplicate) {
		const uint64 OrigArtId = ArticleStore->GetFieldNmUInt64(ArticleId, "isDuplicateOfId");
		ResJson->AddToObj("originalArticle", Article(OrigArtId, MaxBodyLen, IncludeConceptTypes, LabelLanguageV, ArticleFlags));
	}
	
	// add article image
	if (ArticleFlags & aifIncludeImage) {
		TRec ImageRec = ArticleRec.DoSingleJoin(Base, "hasImage");
		if (ImageRec.IsDef())
			ResJson->AddToObj("imageUri", ImageStore->GetFieldNmStr(ImageRec.GetRecId(), "URI"));
	}

	// story uri
	if (ArticleFlags & aifIncludeStoryUri || ArticleFlags & aifIncludeEventUri) {
		TRec StoryRec = ArticleRec.DoSingleJoin(Base, "hasStory");
		if (StoryRec.IsDef()) {
			const uint64 StoryId = StoryRec.GetRecId();
			if (ArticleFlags & aifIncludeStoryUri)
				ResJson->AddToObj("storyUri", StoryStore->GetRecNm(StoryId));
			if (ArticleFlags & aifIncludeEventUri) {
				const uint64 EventId = StoryStore->GetFieldNmUInt64(StoryId, "hasEventId");
				if (EventId != TUInt64::Mx)
					ResJson->AddToObj("eventUri", EventStore->GetRecNm(EventId));
			}
		}
	}
	
	// categories
	if (ArticleFlags & aifIncludeCategories) {
		TJsonValV CatV;
		PRecSet CategorySet = ArticleRec.DoJoin(Base, "hasCategory");
		for (int N = 0; N < CategorySet->GetRecs(); N++) {
			PJsonVal CatJson = Category(CategorySet->GetRecId(N));
			CatJson->AddToObj("wgt", CategorySet->GetRecFq(N));
			CatV.Add(CatJson);
		}
		ResJson->AddToObj("categories", CatV);
	}
	
	// location
	if (ArticleFlags & aifIncludeLocation) {
		const uint64 LocId = ArticleStore->GetFieldNmUInt64(ArticleId, "hasLocationId");
		if (LocId != TUInt64::Mx)
			ResJson->AddToObj("location", Location(LocId, LabelLanguageV));
	}
	
	// extracted dates
	if (ArticleFlags & aifIncludeExtractedDates) {
		TStr DateCollStr = ArticleStore->GetFieldNmStr(ArticleId, "ExtractedDates");
		if (DateCollStr != "")
			ResJson->AddToObj("extractedDates", TJsonVal::GetValFromStr(DateCollStr));
	}

	// newsfeed id 
	if (ArticleFlags & aifIncludeNewsfeedId && !ArticleStore->IsFieldNmNull(ArticleId, "NewsfeedId")) {
		ResJson->AddToObj("newsfeedId", ArticleStore->GetFieldNmStr(ArticleId, "NewsfeedId"));
	}

	return ResJson;
}

PJsonVal TNewsSrvUtil::Concept(const TUInt64& ConceptId, const int ConceptFreq, const TStrV& LabelLanguageV, const EnumConceptInfoFlags ConceptFlags) const 
{
	PJsonVal ConceptJson = TJsonVal::NewObj();
	if (!ConceptStore->IsRecId(ConceptId))
		return ConceptJson;
	const TStr ConceptUri = ConceptStore->GetFieldNmStr(ConceptId, "URI");
	const TStr ConceptType = ConceptStore->GetFieldNmStr(ConceptId, "Type");

	// uri for linking to entity pages
	ConceptJson->AddToObj("uri", ConceptUri);
	ConceptJson->AddToObj("id", ConceptId.GetStr());
	ConceptJson->AddToObj("type", ConceptType);
	if (ConceptFreq != -1)
		ConceptJson->AddToObj("score", TJsonVal::NewNum(ConceptFreq));

	// add also labels for the concept in other languages
	for (int N = 0; N < LabelLanguageV.Len(); N++)
		ConceptJson->AddToObj(NewsConst.GetJsonNmForLang(LabelLanguageV[N]), NewsBase->GetConceptLabel(ConceptUri, LabelLanguageV[N]));
	
	// include also all the images + concept description
	if (ConceptFlags & picIncludeDescription) {
		const TStr Desc = ConceptStore->GetFieldNmStr(ConceptId, "Description");
		ConceptJson->AddToObj("desc", Desc);
	}

	if (ConceptFlags & picIncludePrimImage) {
		TRec Concept = ConceptStore->GetRec(ConceptId);
		TRec PrimaryImage = Concept.DoSingleJoin(Base, "hasPrimaryImage");
		if (PrimaryImage.IsDef())
			ConceptJson->AddToObj("primaryImage", ImageStore->GetRecNm(PrimaryImage.GetRecId()));
	}

	if (ConceptFlags & picIncludeAllImages) {
		TRec Concept = ConceptStore->GetRec(ConceptId);
		PRecSet Images = Concept.DoJoin(Base, "hasImage");
		TStrV ImagesV;
		for (int N=0; N < Images->GetRecs(); N++)
			ImagesV.Add(ImageStore->GetRecNm(Images->GetRecId(N)));
		ConceptJson->AddToObj("images", TJsonVal::NewArr(ImagesV));
	}

	if (ConceptFlags & picIncludeTrends) {
		const uint64 LastArtId = ArticleStore->GetLastRecId();
		if (ArticleStore->IsRecId(LastArtId)) {
			TTm Tm; ArticleStore->GetFieldNmTm(LastArtId, "DateTime", Tm);
			TIntV FreqV(30);
			for (int N = 0; N < 30; N++) {
				TStr PopKey = Tm.GetYMDDashStr();
				PConceptPop Pop = NewsBase->TrendsManager->GetPopForKey(PopKey);
				if (!Pop.Empty())
					FreqV[30 - 1 - N] = Pop->GetConceptPop((TConceptId) ConceptId);
				Tm.AddDays(-1);
			}
			ConceptJson->AddToObj("trend", TJsonVal::NewArr(FreqV));
			// we additionally also add the date of the latest article. this could be relevant is one wants to extrapolate the likely trend of the concept for the latest day
			// when a new day starts the concept won't have a high trend and the latest value will therefore be very low
			ConceptJson->AddToObj("trendLatestDate", Tm.GetWebLogDateTimeStr(true, " ", false));	
		}
	}
		
	return ConceptJson;
}

PJsonVal TNewsSrvUtil::Event(const uint64& EventRecId, const TStrV& LabelLanguageV, const EnumConceptType& ConceptTypes, const EnumEventInfoFlags& EventFlags, const EnumStoryInfoFlags& StoryFlags, const int& EventImageCount, const int& StoryImageCount) const
{
	PJsonVal EventJson = TJsonVal::NewObj();
	if (!EventStore->IsRecId(EventRecId))
		return EventJson;
	// if the event has no more stories and the original story was redirected to another event, then return the info and the new event uri
	const uint64 RedirectedEventId = GetValidEventId(EventRecId);
	if (RedirectedEventId != EventRecId) {
		PJsonVal RetJson = TJsonVal::NewObj("warning", TStr::Fmt("The event has no stories assigned to it. It was merged with event %I64u", RedirectedEventId));
		RetJson->AddToObj("newEventUri", TJsonVal::NewNum(RedirectedEventId));
		return RetJson;
	}

	TStr Uri = EventStore->GetRecNm(EventRecId);
	//EventJson->AddToObj("id", (int) EventRecId);
	EventJson->AddToObj("uri", Uri);

	// add event concepts
	TRec EventRec = EventStore->GetRec(EventRecId);
	if (EventFlags & sifIncludeConcepts) {
		PRecSet ConceptRecSet = EventRec.DoJoin(Base, "hasConcept");
		ConceptRecSet->SortByFq(false);
		EventJson->AddToObj("concepts", Concepts(ConceptRecSet, TUInt64Set(), 100, ctAll, LabelLanguageV));
	}

	PRecSet StoryRecSet = EventRec.DoJoin(Base, "hasStory");
	// sort stories based on their relevance
	StoryRecSet->SortByFq(false);

	// add stories	
	if (EventFlags & eifIncludeStories)
		EventJson->AddToObj("stories", Stories(StoryRecSet, LabelLanguageV, ConceptTypes, StoryFlags, StoryImageCount));

	EventJson->AddToObj("eventDate", EventStore->GetFieldNmStr(EventRecId, "EventDate"));
	EventJson->AddToObj("eventDateEnd", EventStore->GetFieldNmStr(EventRecId, "EventDateEnd"));

	THash<TStr, TInt> LangToArtCountH;
	THash<TStr, TStr> LangToTitleH;
	THash<TStr, TStr> LangToSummaryH;
	THash<TUInt64, TInt> CatToWgtH;		// categories for the stories
	for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
		const uint64 StoryId = StoryRecSet->GetRecId(N);
		// get article counts per language
		const TStr Lang = StoryStore->GetFieldNmStr(StoryId, "Language");
		// ignore stories from non-specified languages
		/*if (!NewsConst.LangH.IsKey(Lang))
			continue;*/
		const bool CustomTitle = NewsBase->GetStoryFlag(StoryId, sfUserSetTitle);
		const bool CustomSummary = NewsBase->GetStoryFlag(StoryId, sfUserSetSummary);
			
		const int ArticleCount = StoryStore->GetFieldNmInt(StoryId, "ArticleCount");
		LangToArtCountH.AddDat(Lang) += ArticleCount;

		// update story info for particular language
		if (!LangToTitleH.IsKey(Lang) || CustomTitle)
			LangToTitleH.AddDat(Lang, StoryStore->GetFieldNmStr(StoryId, "Title"));
		if (!LangToSummaryH.IsKey(Lang) || CustomSummary)
			LangToSummaryH.AddDat(Lang, StoryStore->GetFieldNmStr(StoryId, "Summary"));
			
		// categories for stories
		PRecSet CatRecSet = StoryRecSet->GetRec(N).DoJoin(Base, "hasCategory");
		for (int CatN = 0; CatN < CatRecSet->GetRecs(); CatN++)
			CatToWgtH.AddDat(CatRecSet->GetRecId(CatN)) += CatRecSet->GetRecFq(CatN);
	}

	if (EventFlags & eifIncludeMultiLingInfo) {
		// add event info for each known language
		PJsonVal EventInfoPerLang = TJsonVal::NewObj();
		for (int KeyId = LangToTitleH.FFirstKeyId(); LangToTitleH.FNextKeyId(KeyId);) {
			const TStr Lang = LangToTitleH.GetKey(KeyId);
			const TStr Title = LangToTitleH.GetDat(Lang);
			const TStr Summary = LangToSummaryH.GetDat(Lang);
			PJsonVal EventInfo = TJsonVal::NewObj();
			EventInfo->AddToObj("title", Title);
			EventInfo->AddToObj("summary", Summary);
			EventInfoPerLang->AddToObj(Lang, EventInfo);
		}
		EventJson->AddToObj("multiLingInfo", EventInfoPerLang);
	}
		
	if (EventFlags & eifIncludeCategories) {
		// add most likely event categories
		CatToWgtH.SortByDat(false);
		TJsonValV CatV; int MaxWgt = -1;
		for (int Added = 0, KeyId = CatToWgtH.FFirstKeyId(); CatToWgtH.FNextKeyId(KeyId) && Added < 3; Added++) {
			if (MaxWgt == -1) MaxWgt = CatToWgtH[KeyId];
			PJsonVal CatJson = Category(CatToWgtH.GetKey(KeyId));
			CatJson->AddToObj("wgt", 100 * (CatToWgtH[KeyId] / (double) MaxWgt));
			CatV.Add(CatJson);
		}
		EventJson->AddToObj("categories", CatV);
	}

	if (EventFlags & eifIncludeLocation) {
		// add location
		TRec LocRec = EventRec.DoSingleJoin(Base, "hasLocation");
		if (LocRec.IsDef())
			EventJson->AddToObj("location", Location(LocRec.GetRecId(), LabelLanguageV));
	}

	if (EventFlags & eifIncludeImages) {
		// add 3 images from the top story
		if (StoryRecSet->GetRecs() > 0 && EventImageCount > 0) {
			PRecSet StoryArticles = StoryRecSet->DoJoin(Base, "hasArticle");
			StoryArticles->SortByFq();
			TStrV ImageUrlV;
			for (int N = 0; N < StoryArticles->GetRecs() && ImageUrlV.Len() < EventImageCount; N++) {
				PRecSet ImageSet = StoryArticles->GetRec(N).DoJoin(Base, "hasImage");
				if (ImageSet->GetRecs() > 0)
					ImageUrlV.Add(ImageStore->GetFieldNmStr(ImageSet->GetRecId(0), "URI"));
			}
			EventJson->AddToObj("images", TJsonVal::NewArr(ImageUrlV));
		}
	}

	if (EventFlags & eifIncludeArticleCounts) {
		// add article counts per language to event json
		int ArtTotal = 0;
		PJsonVal CountsJson = TJsonVal::NewObj();
		for (int N = LangToArtCountH.FFirstKeyId(); LangToArtCountH.FNextKeyId(N);) {
			CountsJson->AddToObj(LangToArtCountH.GetKey(N), LangToArtCountH[N]);
			ArtTotal += LangToArtCountH[N];
		}
		CountsJson->AddToObj("total", ArtTotal);
		EventJson->AddToObj("articleCounts", CountsJson);
	}

	return EventJson;
}

PJsonVal TNewsSrvUtil::Story(const uint64& StoryRecId, const int& Score, const TStrV& LabelLanguageV, const EnumConceptType& ConceptTypes, const EnumStoryInfoFlags& StoryFlags, const int& StoryImageCount) const
{
	PJsonVal StoryJson = TJsonVal::NewObj();
	if (!StoryStore->IsRecId(StoryRecId))
		return StoryJson;
	TRec StoryRec = StoryStore->GetRec(StoryRecId);
	StoryJson->AddToObj("uri", StoryStore->GetRecNm(StoryRecId));
	StoryJson->AddToObj("id", TUInt64(StoryRecId).GetStr());
	if (StoryFlags & sifIncludeBasicStats) {
		StoryJson->AddToObj("score", Score);
		StoryJson->AddToObj("lang", StoryStore->GetFieldNmStr(StoryRecId, "Language"));
		StoryJson->AddToObj("avgCos", TJsonVal::NewNum(StoryStore->GetFieldNmFlt(StoryRecId, "AvgCosToCentroid")));
		StoryJson->AddToObj("variance", TJsonVal::NewNum(StoryStore->GetFieldNmFlt(StoryRecId, "Variance")));
		StoryJson->AddToObj("articleCount", TJsonVal::NewNum(StoryStore->GetFieldNmInt(StoryRecId, "ArticleCount")));
		StoryJson->AddToObj("averageDate", StoryStore->GetFieldNmStr(StoryRecId, "AverageDate"));
	}	
	if (StoryFlags & sifIncludeTitle) {
		StoryJson->AddToObj("title", StoryStore->GetFieldNmStr(StoryRecId, "Title"));
		StoryJson->AddToObj("customTitle", TJsonVal::NewBool(NewsBase->GetStoryFlag(StoryRecId, sfUserSetTitle)));
	}
	if (StoryFlags & sifIncludeSummary) {
		StoryJson->AddToObj("summary", StoryStore->GetFieldNmStr(StoryRecId, "Summary"));
		StoryJson->AddToObj("customSummary", TJsonVal::NewBool(NewsBase->GetStoryFlag(StoryRecId, sfUserSetSummary)));
	}
	if (StoryFlags & sifIncludeConcepts) {
		PRecSet ConceptRecSet = StoryRec.DoJoin(Base, "hasConcept");
		ConceptRecSet->SortByFq(false);
		StoryJson->AddToObj("concepts", Concepts(ConceptRecSet, TUInt64Set(), 100, ConceptTypes, LabelLanguageV));
	}
	if (StoryFlags & sifIncludeLocation) {
		// location info - take city if we have it, otherwise try taking the country
		TRec LocRec = StoryRec.DoSingleJoin(Base, "hasLocation");
		if (LocRec.IsDef())
			StoryJson->AddToObj("location", Location(LocRec.GetRecId(), LabelLanguageV));
		StoryJson->AddToObj("customLocation", TJsonVal::NewBool(NewsBase->GetStoryFlag(StoryRecId, sfUserSetLocation)));
	}

	if (StoryFlags & sifIncludeCategory) {
		// categories for stories
		PRecSet CatRecSet = StoryRec.DoJoin(Base, "hasCategory");
		StoryJson->AddToObj("categories", Categories(CatRecSet));
	}
	
	// medoid article
	if (StoryFlags & sifIncludeMedoidArticle) {
		TRec MedoidArticleRec = StoryRec.DoSingleJoin(Base, "hasMedoid");
		if (MedoidArticleRec.IsDef() && ArticleStore->IsRecId(MedoidArticleRec.GetRecId()))
			StoryJson->AddToObj("medoidArticle", Article(MedoidArticleRec.GetRecId(), LongerArticleBodyLen));
		// if medoid was not yet set for the story then choose the most central article (highest fq)
		else {
			PRecSet StoryArticles = StoryRec.DoJoin(Base, "hasArticle");
			StoryArticles->SortByFq();
			if (StoryArticles->GetRecs() > 0)
				StoryJson->AddToObj("medoidArticle", Article(StoryArticles->GetRecId(0), LongerArticleBodyLen));
		}
	}
	// relevant story dates
	if (StoryFlags & sifIncludeStoryDate) {
		// this should be used when showing the story date
		StoryJson->AddToObj("storyDate", StoryStore->GetFieldNmStr(StoryRecId, "StoryDate"));
		StoryJson->AddToObj("storyDateEnd", StoryStore->GetFieldNmStr(StoryRecId, "StoryDateEnd"));
		StoryJson->AddToObj("customDate", TJsonVal::NewBool(NewsBase->GetStoryFlag(StoryRecId, sfUserSetDate)));

		StoryJson->AddToObj("extractedDate", StoryStore->GetFieldNmStr(StoryRecId, "ExtractedDate"));
		StoryJson->AddToObj("extractedDateEnd", StoryStore->GetFieldNmStr(StoryRecId, "ExtractedDateEnd"));
	}

	if (StoryFlags & sifIncludeExtractedDates) {
		TStr CommonDates = StoryStore->GetFieldNmStr(StoryRecId, "CommonDates");
		if (CommonDates != "")
			StoryJson->AddToObj("commonDates", TJsonVal::GetValFromStr(CommonDates));
	}
	
	return StoryJson;
}

PJsonVal TNewsSrvUtil::Source(const uint64& SourceRecId, const EnumPublisherInfoFlags& SourceFlags) const
{
	if (!SourceStore->IsRecId(SourceRecId))
		return TJsonVal::NewObj();
	PJsonVal ResJson = TJsonVal::NewObj();
	
	//ResJson->AddToObj("id", TJsonVal::NewStr(TUInt64(SourceRecId).GetStr()));
	ResJson->AddToObj("uri",  TJsonVal::NewStr(SourceStore->GetFieldNmStr(SourceRecId, "URI")));
	if (SourceFlags & pifIncludeTitle)
		ResJson->AddToObj("title",  TJsonVal::NewStr(SourceStore->GetFieldNmStr(SourceRecId, "Title")));
	if (SourceFlags & pifIncludeDescription) 
		ResJson->AddToObj("desc", TJsonVal::NewStr(SourceStore->GetFieldNmStr(SourceRecId, "Description")));
	if (SourceFlags & pifIncludeTags) {
		const PRecSet TagSet = NewsBase->SourceStore->GetRec(SourceRecId).DoJoin(NewsBase->Base, "hasTag");
		ResJson->AddToObj("tags", Tags(TagSet, false));
	}
	if (SourceFlags & pifIncludeDetails) {
		if (NewsBase->SourceStore->IsFieldNm("Details")) {
			const TStr DetailsStr = NewsBase->SourceStore->GetFieldNmStr(SourceRecId, "Details");
			bool Ok = false; TStr Msg;
			PJsonVal DetailsJson = TJsonVal::GetValFromStr(DetailsStr, Ok, Msg);
			if (Ok)
				ResJson->AddToObj("details", DetailsJson);				
		}
	}
	/*if (IncludeImages) {
		TRec ImageRec = NewsBase->SourceStore->GetRec(SourceRecId).DoSingleJoin(NewsBase->Base, "hasImage");
		TRec ImageThumbRec = NewsBase->SourceStore->GetRec(SourceRecId).DoSingleJoin(NewsBase->Base, "hasThumbImage");
		if (ImageRec.IsDef())
			ResJson->AddToObj("image", ImageRec.GetRecNm());
		if (ImageThumbRec.IsDef())
			ResJson->AddToObj("thumbImage", ImageThumbRec.GetRecNm());
	}*/
	
	if (SourceFlags & pifIncludeLocation) {
		TFltPr Location = SourceStore->GetFieldNmFltPr(SourceRecId, "Location");
		ResJson->AddToObj("lat", TJsonVal::NewNum(Location.Val1));
		ResJson->AddToObj("long", TJsonVal::NewNum(Location.Val2));
		ResJson->AddToObj("country", TJsonVal::NewStr(SourceStore->GetFieldNmStr(SourceRecId, "Country")));
		ResJson->AddToObj("city", TJsonVal::NewStr(SourceStore->GetFieldNmStr(SourceRecId, "City")));
	}
	if (SourceFlags & pifIncludeImportance)
		ResJson->AddToObj("importance",  TJsonVal::NewNum(SourceStore->GetFieldNmInt(SourceRecId, "Importance")));
	
	if (SourceFlags & pifIncludeArticleCount)
		ResJson->AddToObj("articles", TJsonVal::NewNum(NewsBase->NewsCounts->GetSourceFq(SourceRecId)));
	// in case the enabled == false then add the enabled flag to json
	if (NewsBase->IsSourceEnabled(SourceStore->GetRecNm(SourceRecId)) == false)
		ResJson->AddToObj("enabled", false);
	
	return ResJson;
}

PJsonVal TNewsSrvUtil::ConceptFolder(const uint64& ConceptFolderId, const bool& IncludeDetails, const bool& IncludeConceptInfo, const TStrV& LabelLanguageV) const
{
	if (!ConceptFolderStore->IsRecId(ConceptFolderId))
		return TJsonVal::NewObj("error", TStr::Fmt("Invalid concept folder id: %I64u", ConceptFolderId));
	
	PJsonVal ResJson;
	// include details or just uri
	if (IncludeDetails) {
		PConceptFolder Folder = TConceptFolder::New(ConceptFolderId, NewsBase->ConceptFolderStore);
		ResJson = Folder->GetJson();
	}
	else {
		ResJson = TJsonVal::NewObj("uri", ConceptFolderStore->GetRecNm(ConceptFolderId));
	}

	EnumConceptInfoFlags ConceptFlags = DefConceptFlags;
	if (IncludeConceptInfo)
		ConceptFlags = EnumConceptInfoFlags(picIncludePrimImage | picIncludeAllImages | picIncludeDescription | picIncludeTrends);

	// include concept info (label, ...)
	const TStr ConceptFolderUri = ConceptFolderStore->GetRecNm(ConceptFolderId);
	if (ConceptStore->IsRecNm(ConceptFolderUri))
		ResJson->AddToObj("concept", Concept(ConceptStore->GetRecId(ConceptFolderUri), -1, LabelLanguageV, ConceptFlags));
	return ResJson;
}

PJsonVal TNewsSrvUtil::User(const uint64& UserId) const
{
	if (UserStore.Empty())
		return TJsonVal::NewObj("error", TStr("User store is null"));
	if (!UserStore->IsRecId(UserId))
		return TJsonVal::NewObj("error", TStr::Fmt("Invalid user id: %I64u", UserId));

	PJsonVal ResJson = TJsonVal::NewObj("uri", UserStore->GetRecNm(UserId));
	ResJson->AddToObj("name", UserStore->GetFieldNmStr(UserId, "Name"));
	ResJson->AddToObj("lastAccessTime", UserStore->GetFieldNmFlt(UserId, "LastAccessTime"));
	return ResJson;
}

PJsonVal TNewsSrvUtil::Folder(const uint64 FolderRecId) const
{
	return TJsonVal::NewObj("error", "not implemented");
	/*PJsonVal ResJson = TJsonVal::NewObj();
	uint64 FolderId = RecentNewsBase->FolderStore->GetFieldUInt64(FolderRecId, RecentNewsBase->FolderStore->FolderIdFieldId);
	ResJson->AddToObj("folderId", TJsonVal::NewStr(TUInt64::GetStr(FolderId)));

	uint64 OwnerId = RecentNewsBase->FolderStore->GetFieldUInt64(FolderRecId, RecentNewsBase->FolderStore->FolderOwnerFieldId);
	ResJson->AddToObj("folderOwnerId", TJsonVal::NewStr(TUInt64::GetStr(OwnerId)));
		
	TStr FolderName = RecentNewsBase->FolderStore->GetFieldStr(FolderRecId, RecentNewsBase->FolderStore->FolderNameFieldId);
	ResJson->AddToObj("folderName", TJsonVal::NewStr(FolderName));
		
	TInt Visibility = RecentNewsBase->FolderStore->GetFolderVisibility(FolderRecId);
	ResJson->AddToObj("visibility", TJsonVal::NewStr(Visibility.GetStr()));
	return ResJson;*/
}


PJsonVal TNewsSrvUtil::Location(const uint64& ConceptRecId, const TStrV& LabelLanguageV, const bool& IncludeUri, const bool& IncludeLocation, const bool& IncludeDetails) const
{
	if (!ConceptStore->IsRecId(ConceptRecId))
		return TJsonVal::NewObj();
	const TStr LocationUri = ConceptStore->GetRecNm(ConceptRecId);
	return GeoBase->IsCity(LocationUri) ?
		GeoBase->GetInfoForCity(LocationUri, LabelLanguageV, IncludeUri, IncludeLocation, IncludeDetails, true) :
		GeoBase->GetInfoForCountry(LocationUri, LabelLanguageV, IncludeUri, IncludeLocation, IncludeDetails);
}

PJsonVal TNewsSrvUtil::Category(const uint64& CategoryRecId) const
{
	if (!CategoryStore->IsRecId(CategoryRecId))
		return TJsonVal::NewObj();
	PJsonVal Json = TJsonVal::NewObj("uri", TStr(CategoryStore->GetRecNm(CategoryRecId)));
	Json->AddToObj("label", TStr(CategoryStore->GetFieldNmStr(CategoryRecId, "Label")));
	return Json;
}

PJsonVal TNewsSrvUtil::Tag(const uint64& TagRecId) const
{
	if (TagStore.Empty() || !TagStore->IsRecId(TagRecId))
		return TJsonVal::NewObj();
	return TJsonVal::NewObj("uri", TagStore->GetRecNm(TagRecId));
}

PJsonVal TNewsSrvUtil::Articles(const PRecSet& ArticleSet, const int& MaxBodyLen, const TUInt64Set& IgnoreArticleSet, const bool& IncludeWgt, const EnumConceptType& IncludeConceptTypes, const TStrV& LabelLanguageV, const EnumArticleInfoFlags& ArticleFlags) const 
{
	TJsonValV ArticleJsonV;
	for (int ArticleN = 0; ArticleN < ArticleSet->GetRecs(); ArticleN++) {
		if (IgnoreArticleSet.IsKey(ArticleSet->GetRecId(ArticleN))) { continue; }
		const uint64 RecId = ArticleSet->GetRecId(ArticleN);
		// make sure we don't try to add a deleted article
		if (!ArticleStore->IsRecId(RecId))
			continue;
		PJsonVal ArticleJson = Article(RecId, MaxBodyLen, IncludeConceptTypes, LabelLanguageV, ArticleFlags);
		if (IncludeWgt)
			ArticleJson->AddToObj("wgt", ArticleSet->GetRecFq(ArticleN));
		ArticleJsonV.Add(ArticleJson);
	}
	return TJsonVal::NewArr(ArticleJsonV);
}

PJsonVal TNewsSrvUtil::Concepts(const PRecSet& ConceptSet, const TUInt64Set& IgnoreConceptSet, const int MaxCountPerType, const EnumConceptType& IncludeConceptTypes, const TStrV& LabelLanguageV, const EnumConceptInfoFlags ConceptFlags) const
{
	TJsonValV ConceptJsonV; 
	TStrH TypeToCount;
	for (int ConceptN = 0; ConceptN <  ConceptSet->GetRecs(); ConceptN++) {
		const uint64 ConceptId = ConceptSet->GetRecId(ConceptN);
		const int ConceptFreq = ConceptSet->GetRecFq(ConceptN);
		// skip ignore concepts
		if (IgnoreConceptSet.IsKey(ConceptId)) { continue; }
		// get concept parameters
		TStr ConceptType = ConceptStore->GetFieldNmStr(ConceptId, "Type");
		// type specific informations
		PJsonVal ConceptJson = NULL;
		const int CurrentCount = TypeToCount.IsKey(ConceptType) ? TypeToCount.GetDat(ConceptType).Val : 0;
		if (NewsConst.IsValidType(ConceptType, IncludeConceptTypes) && CurrentCount < MaxCountPerType) {
			ConceptJson = Concept(ConceptId, ConceptFreq, LabelLanguageV, ConceptFlags);
			TypeToCount.AddDat(ConceptType) += 1;
		} 
		else
			continue;
		ConceptJsonV.Add(ConceptJson);
	}	
	// add concepts to the json output
	return TJsonVal::NewArr(ConceptJsonV);
}

PJsonVal TNewsSrvUtil::Concepts(const TUInt64V& ConceptV, const TStrV& LabelLanguageV, const EnumConceptInfoFlags ConceptFlags) const
{
	TJsonValV ConceptJsonV;
	int Len = ConceptV.Len();
	for (int ConceptN = 0; ConceptN < Len; ConceptN++)
		ConceptJsonV.Add(Concept(ConceptV[ConceptN], -1, LabelLanguageV, ConceptFlags));
	// add concepts to the json output
	return TJsonVal::NewArr(ConceptJsonV);
}

// compute daily topics for articles in range ArticleSet[Page*Count] - ArticleSet[(Page+1)*Count
PJsonVal TNewsSrvUtil::DailyTopics(const PRecSet& ArticleSet, const TUInt64Set& IgnoreConceptSet, const TStrV& LabelLanguageV, const int& Page, const int& Count) const
{	
	PJsonVal ResJson = TJsonVal::NewObj();
	int StartInd = Page * Count;
	if (StartInd >= ArticleSet->GetRecs())
		return ResJson;

	// find the index of the first article that has the same day as article ArticleSet[Page * Count]
	TTm StartTm; ArticleStore->GetFieldNmTm(ArticleSet->GetRecId(StartInd), "DateTime", StartTm);
	TStr StartDate = StartTm.GetYMDDashStr();
	while (StartInd > 0) {
		int PrevStartInd = StartInd-1;
		TTm PrevStartTm; ArticleStore->GetFieldNmTm(ArticleSet->GetRecId(PrevStartInd), "DateTime", PrevStartTm);
		TStr PrevStartDate = PrevStartTm.GetYMDDashStr();
		if (PrevStartDate != StartDate)
			break;
		StartInd = PrevStartInd;
	}

	// find the index of the last article that has the same day as article ArticleSet[(Page+1) * Count]
	int EndInd = (Page+1) * Count;
	if (EndInd >= ArticleSet->GetRecs())
		EndInd = ArticleSet->GetRecs()-1;
	TTm EndTm; ArticleStore->GetFieldNmTm(ArticleSet->GetRecId(EndInd), "DateTime", EndTm);
	TStr EndDate = EndTm.GetYMDDashStr();
	while (EndInd < ArticleSet->GetRecs()-1) {
		int NextEndInd = EndInd+1;
		TTm NextStartTm; ArticleStore->GetFieldNmTm(ArticleSet->GetRecId(NextEndInd), "DateTime", NextStartTm);
		TStr NextStartDate = NextStartTm.GetYMDDashStr();
		if (NextStartDate != EndDate)
			break;
		EndInd = NextEndInd;
	}

	// find groups of articles for each day in the [StartInd, EndInd] range and from articles for one day extract most relevant concepts
	TUInt64V ArticleRecIdV; ArticleSet->GetRecIdV(ArticleRecIdV);
	TStr CurrDate = StartDate;
	int FirstIndexForDay = StartInd;
	for (int N = StartInd; N <= EndInd; N++) {
		TTm Tm; ArticleStore->GetFieldNmTm(ArticleSet->GetRecId(N), "DateTime", Tm);
		if (CurrDate != Tm.GetYMDDashStr() || N == EndInd) {
			TUInt64V ArticleRecIdForDayV; ArticleRecIdV.GetSubValV(FirstIndexForDay, N-1, ArticleRecIdForDayV);
			PRecSet ArticlesForDay = TRecSet::New(ArticleStore, ArticleRecIdForDayV);
			PRecSet RelatedConceptSet = ArticlesForDay->DoJoin(Base, "hasConcept", 1000, true);
			RelatedConceptSet->SortByFq(false);
			ResJson->AddToObj(CurrDate, Concepts(RelatedConceptSet, IgnoreConceptSet, 10, ctEntities, LabelLanguageV));
			FirstIndexForDay = N;
			CurrDate = Tm.GetYMDDashStr();
		}
	}
	return ResJson;
}

PJsonVal TNewsSrvUtil::Events(const PRecSet& EventRecSet, const TStrV& LabelLanguageV, const EnumConceptType& ConceptTypes, const EnumEventInfoFlags& EventFlags, const EnumStoryInfoFlags& StoryFlags, const int& EventImageCount, const int& StoryImageCount) const
{
	PJsonVal ArrayJson = TJsonVal::NewArr();
	int EventN = EventRecSet->GetRecs();
	for (int N = 0; N < EventN; N++) {
		PJsonVal EventJson = Event(EventRecSet->GetRecId(N), LabelLanguageV, ConceptTypes, EventFlags, StoryFlags, EventImageCount, StoryImageCount);
		EventJson->AddToObj("wgt", EventRecSet->GetRecFq(N));
		ArrayJson->AddToArr(EventJson);
	}
	
	return ArrayJson;
}

PJsonVal TNewsSrvUtil::Stories(const PRecSet& StoryRecSet, const TStrV& LabelLanguageV, const EnumConceptType& ConceptTypes, const EnumStoryInfoFlags& StoryFlags, const int& StoryImageCount) const
{
	PJsonVal ArrayJson = TJsonVal::NewArr();
	int StoryN = StoryRecSet->GetRecs();
	for (int N = 0; N < StoryN; N++)
		ArrayJson->AddToArr(Story(StoryRecSet->GetRecId(N), StoryRecSet->GetRecFq(N), LabelLanguageV, ConceptTypes, StoryFlags, StoryImageCount));
	return ArrayJson;
}

PJsonVal TNewsSrvUtil::Sources(const PRecSet& SourceSet, const EnumPublisherInfoFlags& SourceFlags) const
{
	TJsonValV JsonV;
	if (!SourceSet.Empty()) {
		for (int N = 0; N < SourceSet->GetRecs(); N++) 
			JsonV.Add(Source(SourceSet->GetRecId(N), SourceFlags));
	}
	return TJsonVal::NewArr(JsonV);
}

PJsonVal TNewsSrvUtil::Folders(const PRecSet& FolderSet) const 
{
	TJsonValV FolderJsonV;
	for (int FolderN = 0; FolderN < FolderSet->GetRecs(); FolderN++) {
		// prepare result JSon object
		uint64 FolderRecId = FolderSet->GetRecId(FolderN);
		PJsonVal ResJson = Folder(FolderRecId);
		FolderJsonV.Add(ResJson);
	}
	return TJsonVal::NewArr(FolderJsonV);
}

PJsonVal TNewsSrvUtil::Categories(const PRecSet& CategorySet, const int& MaxCount) const
{
	TJsonValV JsonV;
	for (int N = 0; N < CategorySet->GetRecs() && N < MaxCount; N++) {
		PJsonVal ResJson = Category(CategorySet->GetRecId(N));
		ResJson->AddToObj("wgt", CategorySet->GetRecFq(N));
		JsonV.Add(ResJson);
	}
	return TJsonVal::NewArr(JsonV);
}

PJsonVal TNewsSrvUtil::Tags(const PRecSet& TagSet, const bool& AddWgts) const
{
	TJsonValV JsonV;
	for (int N = 0; N < TagSet->GetRecs(); N++) {
		PJsonVal ResJson = Tag(TagSet->GetRecId(N));
		if (AddWgts)
			ResJson->AddToObj("wgt", TagSet->GetRecFq(N));
		JsonV.Add(ResJson);
	}
	return TJsonVal::NewArr(JsonV);
}

PJsonVal TNewsSrvUtil::ConceptFolders(const PRecSet& FolderSet, const bool& IncludeDetails, const bool& IncludeConceptInfo, const TStrV& LabelLanguageV) const
{
	TJsonValV FolderJsonV;
	for (int FolderN = 0; FolderN < FolderSet->GetRecs(); FolderN++)
		FolderJsonV.Add(ConceptFolder(FolderSet->GetRecId(FolderN), IncludeDetails, IncludeConceptInfo, LabelLanguageV));
	return TJsonVal::NewArr(FolderJsonV);
}

PJsonVal TNewsSrvUtil::Users(const PRecSet& UserSet) const
{
	TJsonValV JsonV;
	for (int N = 0; N < UserSet->GetRecs(); N++) {
		PJsonVal ResJson = User(UserSet->GetRecId(N));
		JsonV.Add(ResJson);
	}
	return TJsonVal::NewArr(JsonV);
}

// compute article source aggregate info
// return data of form: [ { title: string, loc: { lat: num, long: num }, counts: [{ count: num, date: string }] } ]
// if IgnoreDayTime == true, then we will put together all articles on the same date
PJsonVal TNewsSrvUtil::ArticleSourceAggr(const TStrKdV& FldNmValPrV, const PRecSet& ArticleSet, const TStr& FldNmPrefix) const
{
	const bool IncludeLocInfo = TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeLocInfo", true);
	const bool IncludeSourceUri = TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeSourceUri", false);
	const TStr TimeWindowStr = TSAppSrvFun::GetFldVal(FldNmValPrV, FldNmPrefix + "TimeWindow", "day");		// total, month, day, time
	const EnumTimeWindow TimeWindow = TNewsSrvUtil::GetTimeWindow(TimeWindowStr);
	
	THash<TFltPr, THash<TStr, TInt> > LocToDateCountH;
	THash<TUInt64, THash<TStr, TInt> > SourceIdToDateCountH;
	int Articles = ArticleSet->GetRecs();
	for (int N = 0; N < Articles; N++) {
		const TRec ArticleRec = ArticleSet->GetRec(N);
		const TRec SourceRec = ArticleSet->GetRec(N).DoSingleJoin(NewsBase->Base, "hasSource");
		if (NewsBase->SourceStore->IsFieldNmNull(SourceRec.GetRecId(), "Location"))
			continue;
		// find the appropriate key for the article's date and update the article count
		TStr Date = GetKeyForTimeWindow(ArticleRec.GetRecId(), TimeWindow);
		SourceIdToDateCountH.AddDat(SourceRec.GetRecId()).AddDat(Date) += 1;
	}

	TJsonValV ResArr;
	for (int SrcN = SourceIdToDateCountH.FFirstKeyId(); SourceIdToDateCountH.FNextKeyId(SrcN);) {
		const uint64 SourceId = SourceIdToDateCountH.GetKey(SrcN);
		const TFltPr LocPr = NewsBase->SourceStore->GetFieldNmFltPr(SourceId, "Location");

		PJsonVal LocJson = TJsonVal::NewObj();
		if (IncludeLocInfo) {
			PJsonVal LocObj = TJsonVal::NewObj("lat", TJsonVal::NewNum(LocPr.Val1));
			LocObj->AddToObj("long", TJsonVal::NewNum(LocPr.Val2));
			LocJson->AddToObj(LocObj);
		}
		if (IncludeSourceUri)
			LocJson->AddToObj("sourceUri", NewsBase->SourceStore->GetRecNm(SourceId));

		THash<TStr, TInt>& DateCountH = SourceIdToDateCountH[SrcN];
		TJsonValV DateArrV;
		for (int DateId = DateCountH.FFirstKeyId(); DateCountH.FNextKeyId(DateId);) {
			PJsonVal DateCountJson = TJsonVal::NewObj("date", DateCountH.GetKey(DateId));
			if (DateCountH[DateId] > 1)
				DateCountJson->AddToObj("count", TJsonVal::NewNum(DateCountH[DateId]));
			DateArrV.Add(DateCountJson);
		}
		LocJson->AddToObj("counts", TJsonVal::NewArr(DateArrV));
		LocJson->AddToObj("title", NewsBase->SourceStore->GetFieldNmStr(SourceId, "Title"));
		ResArr.Add(LocJson);
	}
	return TJsonVal::NewArr(ResArr);
}

// return date mentioned in articles about the event
// form: [{date: 2013-11-12, count: 123 }]
PJsonVal TNewsSrvUtil::ArticleDateMentionAggr(const PRecSet& ArticleSet) const
{
	if (NewsBase->DateStore.Empty())
		return TJsonVal::NewObj("error", "Date store is not set up. There are no detected date mentions.");

	PRecSet DateMentionSet = ArticleSet->DoJoin(NewsBase->Base, "hasDate");

	PJsonVal JsonArr = TJsonVal::NewArr();
	for (int D = 0; D < DateMentionSet->GetRecs(); D++) {
		const TStr MentionDateStr = NewsBase->DateStore->GetRecNm(DateMentionSet->GetRecId(D));
		PJsonVal DateCount = TJsonVal::NewObj("date", MentionDateStr);
		DateCount->AddToObj("count", DateMentionSet->GetRecFq(D));
		JsonArr->AddToArr(DateCount);
	}

	return JsonArr;
}


PJsonVal TNewsSrvUtil::EventsRecentActivity(const PRecSet& EventRecSet, const int& MaxCount, const TTm& OnlyUpdatesAfterTm, const uint64& LastActivityId, const TStr& Lang, const bool& EventsWithLocationOnly, const double& MinAvgCosSim, const bool& EventsWithLangOnly) const
{
	PJsonVal RetJson = TJsonVal::NewObj();
	PJsonVal EventsJson = TJsonVal::NewObj();
	TStrV EventUriV;
	TLstNd<TActivityItem>* Node = NewsBase->NewsActivity->StoryActivityL.Last();
	// remember the time of the latest activity - client can use this to get only updates after this one
	if (Node != NULL) {
		RetJson->AddToObj("lastUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("lastActivityId", Node->Val.GetActivityId().GetStr());
	}

	// skip events that are will not be in EventRecSet (unless EventRecSet is null - in that case don't remove any event)
	TUInt64Set EventRecIdSet;
	if (!EventRecSet.Empty())
		EventRecSet->GetRecIdSet(EventRecIdSet);

	int N = 0;
	while (Node != NULL && N < MaxCount) {
		const uint64 StoryId = Node->Val.GetItemId();
		const TTm& ActivityTm = Node->Val.GetTm();
		// go into the history only as far as the latest update or as specified
		if (ActivityTm <= OnlyUpdatesAfterTm)
			break;

		if (Node->Val.GetActivityId() <= LastActivityId)
			break;

		Node = Node->Prev();

		if (!NewsBase->StoryStore->IsRecId(StoryId))
			continue;
		// skip events that have too low quality
		if (MinAvgCosSim > 0 && NewsBase->StoryStore->GetFieldNmFlt(StoryId, "AvgCosToCentroid") < MinAvgCosSim)
			continue;
		const uint64 EventId = NewsBase->StoryStore->GetFieldNmUInt64(StoryId, "hasEventId");
		if (EventId == TUInt64::Mx)
			continue;
		if (!EventRecSet.Empty() && EventRecIdSet.IsKey(EventId) == false)
			continue;
		const TStr Uri = NewsBase->EventStore->GetRecNm(EventId);

		if (!EventsJson->IsObjKey(Uri)) {
			PJsonVal EventJson = TJsonVal::NewObj("uri", Uri);

			// /////
			// find the location of the event
			TRec LocRec = NewsBase->EventStore->GetRec(EventId).DoSingleJoin(NewsBase->Base, "hasLocation");
			if (!LocRec.IsDef() && EventsWithLocationOnly)
				continue;
			if (LocRec.IsDef())
				EventJson->AddToObj("location", Location(LocRec.GetRecId()));

			// /////
			// find the title of the event
			PRecSet StoryRecSet = NewsBase->EventStore->GetRec(EventId).DoJoin(NewsBase->Base, "hasStory");
			// sort stories based on their relevance
			StoryRecSet->SortByFq(false);
			TStr Title;
			for (int S = 0; S < StoryRecSet->GetRecs(); S++) {
				const uint64 StoryId = StoryRecSet->GetRecId(S);
				// get article counts per language
				const TStr StoryLang = NewsBase->StoryStore->GetFieldNmStr(StoryId, "Language");
				if (StoryLang == Lang) {
					Title = NewsBase->StoryStore->GetFieldNmStr(StoryId, "Title");
					break;
				}
			}
			// if we didn't find any story in the desired language then skip
			if (Title == "" && EventsWithLangOnly)
				continue;
			if (Title == "" && StoryRecSet->GetRecs() > 0)
				Title = NewsBase->StoryStore->GetFieldNmStr(StoryRecSet->GetRecId(0), "Title");

			EventJson->AddToObj("title", Title);

			EventJson->AddToObj("articleCount", GetEventArticleCount(EventId));

			EventsJson->AddToObj(Uri, EventJson);		// event uri -> event info
		}
		EventUriV.Add(Uri);
		N++;
	}
	// add info about the old update that we didn't include (can be used to compute the speed of changes)
	if (Node != NULL) {
		RetJson->AddToObj("oldUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("oldActivityId", Node->Val.GetActivityId().GetStr());
	}

	RetJson->AddToObj("activity", TJsonVal::NewArr(EventUriV));
	RetJson->AddToObj("eventInfo", EventsJson);
	return RetJson;
}

PJsonVal TNewsSrvUtil::ConceptFolderFull(const uint64& ConceptRecId, const bool& AddLabelsForConcepts, const TStrV& LabelLanguageV) const
{
	return TJsonVal::NewObj("error", "not implemented");
	//TConceptFolderRec& FolderRec = RecentNewsBase->FolderStore->GetConceptFolder(ConceptRecId);
	//PJsonVal ResJson = RecentNewsBase->GetJsonForFolder(FolderRec.FolderRecId);
	//// add also concept labels and source labels
	//if (AddLabelsForConcepts)
	//	RecentNewsBase->AddLabelsForFolderJson(ResJson);
	//	
	//TStr FolderUri = ConceptStore->GetRecNm(ConceptRecId);
	//ResJson->AddToObj("conceptFolderUri", TJsonVal::NewStr(FolderUri));
	//ResJson->AddToObj("autoAddArticles", TJsonVal::NewBool(FolderRec.AutoAddArticles));

	//return ResJson;
}

// return json with a list of all article uris from ArticleRecSet
PJsonVal TNewsSrvUtil::ArticleUris(const PRecSet& ArticleRecSet) const
{
	int Arts = ArticleRecSet->GetRecs();
	TStrV ArtV;
	for (int ArtN = 0; ArtN < Arts; ArtN++) {
		uint64 ArtId = ArticleRecSet->GetRecId(ArtN);
		if (ArticleStore->IsRecId(ArtId))
			ArtV.Add(ArticleStore->GetRecNm(ArtId));
	}
	return TJsonVal::NewArr(ArtV);
}

// return json with a list of all article ids from ArticleRecSet
PJsonVal TNewsSrvUtil::ArticleIds(const PRecSet& ArticleRecSet) const
{
	int Arts = ArticleRecSet->GetRecs();
	TStrV ArtV;
	for (int ArtN = 0; ArtN < Arts; ArtN++) {
		const TUInt64 ArtId = ArticleRecSet->GetRecId(ArtN);
		if (ArticleStore->IsRecId(ArtId))
			ArtV.Add(ArtId.GetStr());
	}
	return TJsonVal::NewArr(ArtV);
}

// return json with a list of all article uris from ArticleRecSet - use the FldNmValPrV params to sort the recset first
PJsonVal TNewsSrvUtil::ArticleUris(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet, const TStr& FldNmPrefix) const
{
	const TStr SortBy = TSAppSrvFun::GetFldVal(FldNmValPrV, FldNmPrefix + "SortBy", "none");	// none, id, date, cosSim, fq
	const bool SortByAsc = TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "SortByAsc", false);	// asc or desc order
	PStore ArticleStore = NewsBase->ArticleStore;

	// sort events by time
	if (SortBy.GetLc() == "date")
		ArticleRecSet->SortByField(SortByAsc, NewsBase->ArticleStore->GetFieldId("DateTime"));
	else if (SortBy.GetLc() == "cossim")
		// sort stories by relevance (frequency in rec set - should sort primarily by relevance of the searched concepts to the story)
		ArticleRecSet->SortByField(SortByAsc, NewsBase->ArticleStore->GetFieldId("CosToCentroid"));
	else if (SortBy.GetLc() == "id")
		ArticleRecSet->SortById(SortByAsc);
	else if (SortBy.GetLc() == "fq")
		ArticleRecSet->SortByFq(SortByAsc);
	
	return ArticleUris(ArticleRecSet);
}

// return json with a list of all article ids from ArticleRecSet - use the FldNmValPrV params to sort the recset first
PJsonVal TNewsSrvUtil::ArticleIds(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet, const TStr& FldNmPrefix) const
{
	const TStr SortBy = TSAppSrvFun::GetFldVal(FldNmValPrV, FldNmPrefix + "SortBy", "none");	// none, id, date, cosSim, fq
	const bool SortByAsc = TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "SortByAsc", false);	// asc or desc order
	PStore ArticleStore = NewsBase->ArticleStore;
	
	// sort events by time
	if (SortBy.GetLc() == "date")
		ArticleRecSet->SortByField(SortByAsc, NewsBase->ArticleStore->GetFieldId("DateTime"));
	else if (SortBy.GetLc() == "cossim")
		// sort stories by relevance (frequency in rec set - should sort primarily by relevance of the searched concepts to the story)
		ArticleRecSet->SortByField(SortByAsc, NewsBase->ArticleStore->GetFieldId("CosToCentroid"));
	else if (SortBy.GetLc() == "id")
		ArticleRecSet->SortById(SortByAsc);
	else if (SortBy.GetLc() == "fq")
		ArticleRecSet->SortByFq(SortByAsc);
	
	return ArticleIds(ArticleRecSet);
}

PJsonVal TNewsSrvUtil::Articles(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet, const TStr& FldNmPrefix) const
{
	const TStr SortBy = TSAppSrvFun::GetFldVal(FldNmValPrV, FldNmPrefix + "SortBy", "cosSim");	// date, id, cosSim, fq
	const bool SortByAsc = TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "SortByAsc", false);	// asc or desc order
	const int Page = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "Page", 0);
	const int Count = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "Count", DefaultNumberOfReturnedArticlesPerPage);
	const int BodyLen = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "BodyLen", DefaultArticleBodyLen);
	
	TStrV ConceptLangV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV, ctEntities);
	
	EnumArticleInfoFlags ArticleFlags = GetArticleFlags(FldNmValPrV, FldNmPrefix);

	// sort events by time
	if (SortBy.GetLc() == "date")
		ArticleRecSet->SortByField(SortByAsc, NewsBase->ArticleStore->GetFieldId("DateTime"));
	else if (SortBy.GetLc() == "cossim")
		// sort stories by relevance (frequency in rec set - should sort primarily by relevance of the searched concepts to the story)
		ArticleRecSet->SortByField(SortByAsc, NewsBase->ArticleStore->GetFieldId("CosToCentroid"));
	else if (SortBy.GetLc() == "id")
		ArticleRecSet->SortById(SortByAsc);
	else if (SortBy.GetLc() == "fq")
		ArticleRecSet->SortByFq(SortByAsc);

	// take Count events from page Page
	PRecSet SubArticleRecSet = ArticleRecSet->GetLimit(Count, Count*Page);

	PJsonVal ResJson = TJsonVal::NewObj();

	ResJson->AddToObj("results", Articles(SubArticleRecSet, BodyLen, TUInt64Set(), true, ConceptTypes, ConceptLangV, ArticleFlags));
	ResJson->AddToObj("resultCount", ArticleRecSet->GetRecs());

	/*PJsonVal SearchParamsJson = TJsonVal::NewObj();
	SearchParamsJson->AddToObj("page", Page);
	SearchParamsJson->AddToObj("count", Count);
	SearchParamsJson->AddToObj("sortBy", SortBy);
	ResJson->AddToObj("searchParams", SearchParamsJson);*/

	return ResJson;
}

// extract top concepts from the sample of articles and return top concepts
PJsonVal TNewsSrvUtil::RelatedConcepts(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet, const TUInt64Set& IgnoreConceptSet, const TStr& FldNmPrefix) const
{
	const int SampleSize = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "SampleSize", 1000);
	const int RelatedConceptCount = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "Count", 20);
	TStrV ConceptLangV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV, ctAll);

	if (RelatedConceptCount > 500)
		return TJsonVal::NewObj("error", TStr("The API does not allow obtaining more than 500 related concepts per call."));

	PRecSet RelatedConceptSet = ArticleRecSet->DoJoin(NewsBase->Base, "hasConcept", SampleSize, true);
	RelatedConceptSet->SortByFq(false);
	return Concepts(RelatedConceptSet, IgnoreConceptSet, RelatedConceptCount, ConceptTypes, ConceptLangV);
}

// return the list of articles as they were recently added - used to show activity on the map 
PJsonVal TNewsSrvUtil::ArticleRecentActivity(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet) const
{
	const int MaxCount = TSAppSrvFun::GetFldInt(FldNmValPrV, "articleRecentActivityMaxArticleCount", 60);
	const int MaxMinBack = TSAppSrvFun::GetFldInt(FldNmValPrV, "articleRecentActivityMaxMinsBack", 10 * 60);		// check maximum 10 hours into the history of activity
	TTm OnlyUpdatesAfterTm = TTm::GetCurLocTm(); OnlyUpdatesAfterTm.AddTime(0, -MaxMinBack);
	const uint64 LastArticleActivityId = TSAppSrvFun::GetFldUInt64(FldNmValPrV, "articleRecentActivityLastArticleActivityId", 0);
	const bool ArticlesWithLocationOnly = TSAppSrvFun::GetFldBool(FldNmValPrV, "articleRecentActivityArticlesWithLocationOnly", true);
	
	PJsonVal RetJson = TJsonVal::NewObj();
	PJsonVal ActivityArrJson = TJsonVal::NewArr();
	TLstNd<TActivityItem>* Node = NewsBase->NewsActivity->ArticleActivityL.Last();
	// remember the time of the latest activity - client can use this to get only updates after this one
	if (Node != NULL) {
		RetJson->AddToObj("lastUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("lastActivityId", Node->Val.GetActivityId().GetStr());
	}

	const TStr FldNmPrefix = "article";
	const int BodyLen = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "BodyLen", 0);
	TStrV ConceptLangV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV, ctEntities);

	EnumArticleInfoFlags DefFlags = EnumArticleInfoFlags(aifBasicInfo | aifIncludeSourceInfo | aifIncludeTitle);
	EnumArticleInfoFlags ArticleFlags = GetArticleFlags(FldNmValPrV, FldNmPrefix, DefFlags);

	// skip articles that are will not be in ArticleRecIdSet
	TUInt64Set ArticleRecIdSet;
	if (!ArticleRecSet.Empty())
		ArticleRecSet->GetRecIdSet(ArticleRecIdSet);

	int N = 0;
	while (Node != NULL && N < MaxCount) {
		const uint64 ArticleId = Node->Val.GetItemId();
		const TTm& ActivityTm = Node->Val.GetTm();
		// go into the history only as far as the latest update or as specified
		if (ActivityTm <= OnlyUpdatesAfterTm)
			break;
		if (Node->Val.GetActivityId() <= LastArticleActivityId)
			break;

		Node = Node->Prev();
		if (!ArticleRecIdSet.Empty() && ArticleRecIdSet.IsKey(ArticleId) == false)
			continue;
		if (!NewsBase->ArticleStore->IsRecId(ArticleId))
			continue;

		if (ArticlesWithLocationOnly && !NewsBase->HasLocationForArticleSource(ArticleId))
			continue;
		PJsonVal ArticleJson = Article(ArticleId, BodyLen, ConceptTypes, ConceptLangV, ArticleFlags);
		ActivityArrJson->AddToArr(ArticleJson);
		N++;
	}
	// add info about the old update that we didn't include (can be used to compute the speed of changes)
	if (Node != NULL) {
		RetJson->AddToObj("oldUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("oldActivityId", Node->Val.GetActivityId().GetStr());
	}
	// add current time - also can be used to compute amount of expected activity
	RetJson->AddToObj("currTime", TTm::GetCurLocTm().GetWebLogDateTimeStr());

	RetJson->AddToObj("activity", ActivityArrJson);
	return RetJson;
}

PJsonVal TNewsSrvUtil::DailyTopics(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet, const TUInt64Set& IgnoreConceptSet) const
{
	const TStr SortBy = TSAppSrvFun::GetFldVal(FldNmValPrV, "dailyTopicsSortBy", "date").GetLc();		// date, rel
	const int Page = TSAppSrvFun::GetFldInt(FldNmValPrV, "dailyTopicsPage", 0);
	const int Count = TSAppSrvFun::GetFldInt(FldNmValPrV, "dailyTopicsCount", DefaultNumberOfReturnedArticlesPerPage);
	TStrV ConceptLangV; TSAppSrvFun::GetFldValV(FldNmValPrV, "dailyTopicsConceptLang", ConceptLangV);
	if (ConceptLangV.Empty())
		ConceptLangV.Add("eng");
	return DailyTopics(ArticleRecSet, IgnoreConceptSet, ConceptLangV, Page, Count);
}

// get the event info for events in EventRecSet
PJsonVal TNewsSrvUtil::Events(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet, const TStr& FldNmPrefix)
{
	const TStr SortBy = TSAppSrvFun::GetFldVal(FldNmValPrV, FldNmPrefix + "SortBy", "");		// date, size, rel
	const bool SortByAsc = TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "SortByAsc", false);	// asc or desc order
	TStrV ConceptLangV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");

	const int  Page = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "Page", 0);
	const int  Count = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "Count", 25);
	const bool ReorderBySize = TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "ReorderBySize", true);
	TStrV ConceptTypeV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV, ctAll);
	const int EventImageCount = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "EventImageCount", 3);
	const int StoryImageCount = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "StoryImageCount", 3);

	EnumEventInfoFlags EventFlags = GetEventFlags(FldNmValPrV, FldNmPrefix);
	EnumStoryInfoFlags StoryFlags = GetStoryFlags(FldNmValPrV);

	if (Count > 100)
		return TJsonVal::NewObj("error", TStr("The API does not allow downloading more than 100 events per call."));

	PRecSet SubEventRecSet;
	// sort events by time
	if (SortBy.GetLc() == "date") {
		EventRecSet->SortByField(SortByAsc, EventStore->GetFieldId("EventDate"));
		SubEventRecSet = EventRecSet->GetLimit(Count, Count*Page);
	}
	else if (SortBy.GetLc() == "size") {
		EventRecSet->SortByField(SortByAsc, EventStore->GetFieldId("ArticleCount"));
		SubEventRecSet = EventRecSet->GetLimit(Count, Count*Page);
	}
	else if (SortBy.GetLc() == "rel") {
		// sort stories by relevance (frequency in rec set - should sort primarily by relevance of the searched concepts to the story)
		if (ReorderBySize) {
			// slightly adapt the order based on the event size
			TUInt64IntKdV RecIdKdV = EventRecSet->GetRecIdFqV();
			for (int N = 0; N < RecIdKdV.Len(); N++) {
				const uint64 EventId = RecIdKdV[N].Key;
				// we just care if the article count reached 500 - if yes, then don't add extra weight to the events
				const int ArtCount = EventStore->IsRecId(EventId) ? MIN(500, EventStore->GetFieldNmInt(EventId, "ArticleCount")) : 0;
				RecIdKdV[N].Dat += (int) (ArtCount / 25.);		// max added score will be 25
			}
			RecIdKdV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(SortByAsc));
			TUInt64IntKdV SubValRecIdKdV; RecIdKdV.GetSubValV(Count*Page, (Count*(Page + 1)) - 1, SubValRecIdKdV);
			SubEventRecSet = TRecSet::New(EventStore, SubValRecIdKdV, true);
		}
		else {
			EventRecSet->SortByFq(SortByAsc);
			SubEventRecSet = EventRecSet->GetLimit(Count, Count*Page);
		}
	}
	else {
		SubEventRecSet = EventRecSet->GetLimit(Count, Count*Page);
	}

	PJsonVal ResJson = TJsonVal::NewObj();

	ResJson->AddToObj("results", Events(SubEventRecSet, ConceptLangV, ConceptTypes, EventFlags, StoryFlags, EventImageCount, StoryImageCount));
	ResJson->AddToObj("resultCount", EventRecSet->GetRecs());

	/*PJsonVal SearchParamsJson = TJsonVal::NewObj();
	SearchParamsJson->AddToObj("page", Page);
	SearchParamsJson->AddToObj("count", Count);
	SearchParamsJson->AddToObj("sortBy", SortBy);
	ResJson->AddToObj("searchParams", SearchParamsJson);*/

	return ResJson;
}

EnumArticleInfoFlags TNewsSrvUtil::GetArticleFlags(const TStrKdV& FldNmValPrV, const TStr& FldNmPrefix, const EnumArticleInfoFlags& DefFlags) const
{
	// aifIncludeCategories = 1 << 7, aifIncludeLocation = 1 << 8
	int ArticleFlags = 0;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeBasicInfo", DefFlags & aifBasicInfo))
		ArticleFlags |= aifBasicInfo;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeBody", DefFlags & aifIncludeBody))
		ArticleFlags |= aifIncludeBody;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeTitle", DefFlags & aifIncludeTitle))
		ArticleFlags |= aifIncludeTitle;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeConcepts", DefFlags & aifIncludeConcepts))
		ArticleFlags |= aifIncludeConcepts;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeSourceInfo", DefFlags & aifIncludeSourceInfo))
		ArticleFlags |= aifIncludeSourceInfo;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeEventUri", DefFlags & aifIncludeEventUri))
		ArticleFlags |= aifIncludeEventUri;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeStoryUri", DefFlags & aifIncludeStoryUri))
		ArticleFlags |= aifIncludeStoryUri;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeDuplicateList", DefFlags & aifIncludeDuplicateList))
		ArticleFlags |= aifIncludeDuplicateList;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeOriginalArticleInfo", DefFlags & aifIncludeOriginalArticleInfo))
		ArticleFlags |= aifIncludeOriginalArticleInfo;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeCategories", DefFlags & aifIncludeCategories))
		ArticleFlags |= aifIncludeCategories;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeLocation", DefFlags & aifIncludeLocation))
		ArticleFlags |= aifIncludeLocation;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeImage", DefFlags & aifIncludeImage))
		ArticleFlags |= aifIncludeImage;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeExtractedDates", DefFlags & aifIncludeExtractedDates))
		ArticleFlags |= aifIncludeExtractedDates;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeNewsfeedId", DefFlags & aifIncludeNewsfeedId))
		ArticleFlags |= aifIncludeNewsfeedId;
	return (EnumArticleInfoFlags) ArticleFlags;
}

EnumStoryInfoFlags TNewsSrvUtil::GetStoryFlags(const TStrKdV& FldNmValPrV, const TStr& FldNmPrefix, const EnumStoryInfoFlags& DefFlags) const
{
	//{ IncludeBasicStats = 1 << 1, IncludeTitle = 1 << 2, IncludeSummary = 1 << 3, IncludeLocation = 1 << 4, IncludeCategory = 1 << 5, IncludeMedoidArticle = 1 << 6, IncludeStoryDate = 1 << 7, IncludeExtractedDates = 1 << 8 };
	int StoryFlags = 0;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeBasicStats", DefFlags & sifIncludeBasicStats))
		StoryFlags |= sifIncludeBasicStats;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeCategory", DefFlags & sifIncludeCategory))
		StoryFlags |= sifIncludeCategory;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeLocation", DefFlags & sifIncludeLocation))
		StoryFlags |= sifIncludeLocation;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeStoryDate", DefFlags & sifIncludeStoryDate))
		StoryFlags |= sifIncludeStoryDate;

	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeConcepts", DefFlags & sifIncludeConcepts))
		StoryFlags |= sifIncludeConcepts;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeTitle", DefFlags & sifIncludeTitle))
		StoryFlags |= sifIncludeTitle;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeSummary", DefFlags & sifIncludeSummary))
		StoryFlags |= sifIncludeSummary;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeMedoidArticle", DefFlags & sifIncludeMedoidArticle))
		StoryFlags |= sifIncludeMedoidArticle;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeExtractedDates", DefFlags & sifIncludeExtractedDates))
		StoryFlags |= sifIncludeExtractedDates;
	
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeAllDetails", false))
		StoryFlags |= (sifIncludeConcepts | sifIncludeTitle | sifIncludeSummary | sifIncludeExtractedDates);
	
	return (EnumStoryInfoFlags) StoryFlags;
}

EnumEventInfoFlags TNewsSrvUtil::GetEventFlags(const TStrKdV& FldNmValPrV, const TStr& FldNmPrefix, const EnumEventInfoFlags& DefFlags) const
{
	int EventFlags = 0;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeArticleCounts", DefFlags & eifIncludeArticleCounts))
		EventFlags |= eifIncludeArticleCounts;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeConcepts", DefFlags & eifIncludeConcepts))
		EventFlags |= eifIncludeConcepts;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeMultiLingInfo", DefFlags & eifIncludeMultiLingInfo))
		EventFlags |= eifIncludeMultiLingInfo;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeCategories", DefFlags & eifIncludeCategories))
		EventFlags |= eifIncludeCategories;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeLocation", DefFlags & eifIncludeLocation))
		EventFlags |= eifIncludeLocation;

	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeStories", DefFlags & eifIncludeStories))
		EventFlags |= eifIncludeStories;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeImages", DefFlags & eifIncludeImages))
		EventFlags |= eifIncludeImages;
	
	return (EnumEventInfoFlags) EventFlags;
}

EnumConceptInfoFlags TNewsSrvUtil::GetConceptFlags(const TStrKdV& FldNmValPrV, const TStr& FldNmPrefix, const EnumConceptInfoFlags& DefFlags) const
{
	int ConceptFlags = 0;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludePrimImage", DefFlags & picIncludePrimImage))
		ConceptFlags |= picIncludePrimImage;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeAllImages", DefFlags & picIncludeAllImages))
		ConceptFlags |= picIncludeAllImages;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeDescription", DefFlags & picIncludeDescription))
		ConceptFlags |= picIncludeDescription;
	if (TSAppSrvFun::GetFldBool(FldNmValPrV, FldNmPrefix + "IncludeTrends", DefFlags & picIncludeTrends))
		ConceptFlags |= picIncludeTrends;
	return (EnumConceptInfoFlags) ConceptFlags;
}

PJsonVal TNewsSrvUtil::EventsRecentActivity(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet) const
{
	const int MaxCount = TSAppSrvFun::GetFldInt(FldNmValPrV, "eventsRecentActivityMaxEventCount", 60);
	const int MaxMinBack = TSAppSrvFun::GetFldInt(FldNmValPrV, "eventsRecentActivityMaxMinsBack", 10 * 60);		// check maximum 10 hours into the history of activity
	TTm OnlyUpdatesAfterTm = TTm::GetCurLocTm(); OnlyUpdatesAfterTm.AddTime(0, -MaxMinBack);
	const uint64 LastEventActivityId = TSAppSrvFun::GetFldUInt64(FldNmValPrV, "eventsRecentActivityLastEventActivityId", 0);
	const TStr EventLang = TSAppSrvFun::GetFldVal(FldNmValPrV, "eventsRecentActivityEventLang", "eng");
	const bool EventsWithLocationOnly = TSAppSrvFun::GetFldBool(FldNmValPrV, "eventsRecentActivityEventsWithLocationOnly", true);
	const bool EventsWithLangOnly = TSAppSrvFun::GetFldBool(FldNmValPrV, "eventsRecentActivityEventsWithLangOnly", false);
	const TStr MinAvgCosSimStr = TSAppSrvFun::GetFldVal(FldNmValPrV, "eventsRecentActivityMinAvgCosSim", "0.0");
	const double MinAvgCosSim = MinAvgCosSimStr.GetFlt(0.0);

	return EventsRecentActivity(EventRecSet, MaxCount, OnlyUpdatesAfterTm, LastEventActivityId, EventLang, EventsWithLocationOnly, MinAvgCosSim, EventsWithLangOnly);
}

// return the number of articles assigned to the story (using the story rec)
int TNewsSrvUtil::GetStoryArticleCount(const uint64& StoryId) const
{
	if (!StoryStore->IsRecId(StoryId))
		return 0;
	return StoryStore->GetFieldNmInt(StoryId, "ArticleCount");
}

// return the number of articles assigned to the event (using the event rec)
int TNewsSrvUtil::GetEventArticleCount(const uint64& EventId) const
{
	if (!EventStore->IsRecId(EventId))
		return 0;
	return EventStore->GetFieldNmInt(EventId, "ArticleCount");
}

// the original story from the event can be reassigned to another event. 
// this method return the EventId if the event has at least one story in it
// if the event has no stories (the original story was reassigned to another event) then we return the id 
// of the event to which the primary story was assigned to
uint64 TNewsSrvUtil::GetValidEventId(const uint64& EventId) const
{
	WAssert(EventStore->IsRecId(EventId), "Invalid event id");
	if (!EventStore->IsRecId(EventId))
		return TUInt64::Mx;
	if (EventStore->GetRec(EventId).DoJoin(Base, "hasStory")->GetRecs() > 0)
		return EventId;
	TRec PrimaryStoryRec = EventStore->GetRec(EventId).DoSingleJoin(Base, "hasPrimaryStory");
	if (!PrimaryStoryRec.IsDef())
		return TUInt64::Mx;
	TRec EventRec = PrimaryStoryRec.DoSingleJoin(Base, "hasEvent");
	return EventRec.GetRecId();
}

void TNewsSrvUtil::GetConceptCountsByType(const PRecSet& ConceptRecSet, int& Persons, int& Locations, int& Organizations, int& Wikis) const
{
	int Concepts = ConceptRecSet->GetRecs();
	for (int N = 0; N < Concepts; N++) {
		const TStr Type = ConceptStore->GetFieldNmStr(ConceptRecSet->GetRecId(N), "Type");
		if (Type == NewsConst.TypePerson) Persons++;
		else if (Type == NewsConst.TypeLocation) Locations++;
		else if (Type == NewsConst.TypeOrganization) Organizations++;
		else if (Type == NewsConst.TypeWiki) Wikis++;
	}
}