/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "enrycherdoc.h"

///////////////////////////////
// Enrycher-Token
TEnrycherToken::TEnrycherToken(const PXmlTok& TokenTok, int SentId) {
	// parse token id
	TStr TokenIdStr = TokenTok->GetStrArgVal("id");
	TStr LeftId, RightId; TokenIdStr.SplitOnCh(LeftId, '.', RightId);
	EAssertR(LeftId.IsInt(), "Invalid Token ID " + TokenIdStr);
	EAssertR(LeftId.GetInt() == SentId, "Invalid Token ID " + TokenIdStr);
	EAssertR(RightId.IsInt(), "Invalid Token ID " + TokenIdStr);
	TokenId = RightId.GetInt();
	// pares token string
	TokenStr = TokenTok->GetTokStr(false);
	// parse token key-val pairs
	for (int ArgN = 0; ArgN < TokenTok->GetArgs(); ArgN++) {
		TStr ArgNm, ArgVal; TokenTok->GetArg(ArgN, ArgNm, ArgVal);
		if (ArgNm != "id") { KeyValH.AddDat(ArgNm, ArgVal); }
	}
}

PXmlTok TEnrycherToken::GetXmlTok(int SentId) const {
	// create token
    PXmlTok TokenTok = TXmlTok::New("token", GetStr());
	// output the token id (sentid.tokenid)
    TokenTok->AddArg("id", TStr::Fmt("%d.%d", SentId, GetId()));
	// output other key-value pairs
	int KeyId = KeyValH.FFirstKeyId();
	while (KeyValH.FNextKeyId(KeyId)) {
		TokenTok->AddArg(KeyValH.GetKey(KeyId), KeyValH[KeyId]);
	}
	// done
	return TokenTok;
}

///////////////////////////////
// Enrycher-Sentence
TEnrycherSent::TEnrycherSent(const PXmlTok& SentTok) {
    // this is always here
    SentId = SentTok->GetIntArgVal("id");
	if (SentTok->IsTagTok("plainText")) { 
		SentStr = SentTok->GetTagTok("plainText")->GetTokStr(false);
	} else if (SentTok->IsTagTok("text")) {
		SentStr = SentTok->GetTagTok("text")->GetTokStr(false);
	} else {
		TExcept::Throw("Missing sentence text in sentence " + SentId.GetStr());
	}
    // we parse tokens if tokenized
    TXmlTokV TokenTokV; SentTok->GetTagTokV("tokens|token", TokenTokV);
    TokenV.Gen(TokenTokV.Len(), 0);
    for (int TokenTokN = 0; TokenTokN < TokenTokV.Len(); TokenTokN++) {
        PXmlTok TokenTok = TokenTokV[TokenTokN];
        TokenV.Add(TEnrycherToken(TokenTok, SentId));
    }
}

PXmlTok TEnrycherSent::GetXmlTok() const {
    PXmlTok SentTok = TXmlTok::New("sentence");
    // always
    SentTok->AddArg("id", GetId());
    SentTok->AddSubTok(TXmlTok::New("plainText", GetStr()));
    // when tokenized
    if (IsTokenized()) {
        PXmlTok TokensTok = TXmlTok::New("tokens");
        for (int TokenN = 0; TokenN < GetTokens(); TokenN++) {
            TokensTok->AddSubTok(GetToken(TokenN).GetXmlTok(GetId())); }
        SentTok->AddSubTok(TokensTok);
    }
    return SentTok;
    // done
}

///////////////////////////////
// Enrycher-Paragraph
TEnrycherPara::TEnrycherPara(const PXmlTok& ParaTok) {
    // get tag name
    TagNm = ParaTok->GetTagNm();
    // get all sentences
    TXmlTokV SentTokV; ParaTok->GetTagTokV("sentence", SentTokV);
    for (int SentTokN = 0; SentTokN < SentTokV.Len(); SentTokN++) {
        PXmlTok SentTok = SentTokV[SentTokN];
        SentV.Add(TEnrycherSent(SentTok));
    }
}

PXmlTok TEnrycherPara::GetXmlTok() const {
    PXmlTok ParaTok = TXmlTok::New(TagNm);
    for (int SentN = 0; SentN < GetSents(); SentN++) {
        ParaTok->AddSubTok(GetSent(SentN).GetXmlTok());
    }
    return ParaTok;
}

TStr TEnrycherPara::GetStr() const { 
	TChA ChA;
	for (int SentN = 0; SentN < SentV.Len(); SentN++) {
		if (SentN > 0) { ChA += ' '; }
		ChA += SentV[SentN].GetStr();
	}
	return ChA;
}

///////////////////////////////
// Enrycher-Annotation-Instance
TEnrycherAnnInst::TEnrycherAnnInst(const PXmlTok& AnnInstTok) {
    // instance data
	TStr AnnInstIdStr = AnnInstTok->GetStrArgVal("id");
	if (AnnInstIdStr.IsInt()) {
		AnnInstId = AnnInstIdStr.GetInt();
	} else if (AnnInstIdStr.IsChIn('.')) {
		TStr LeftId, RightId; AnnInstIdStr.SplitOnCh(LeftId, '.', RightId);
		LeftId = LeftId.TrimLeft("E");	// new xlike format has ids of form "E3.2"
		EAssertR(LeftId.IsInt(), "Invalid Mention ID " + AnnInstIdStr);
		EAssertR(RightId.IsInt(), "Invalid Mention ID " + AnnInstIdStr);
		AnnInstId = LeftId.GetInt() * 1000 + RightId.GetInt();
	} else {
		// HACK:
		static int _AnnInstId = 2000;
		AnnInstId = _AnnInstId++;
	}
    AnnInstStr = AnnInstTok->GetStrArgVal("words");
    // instance tokens
    TXmlTokV TokenTokV; AnnInstTok->GetTagTokV("token", TokenTokV);
    for (int TokenTokN = 0; TokenTokN < TokenTokV.Len(); TokenTokN++) {
        // parse token id
        TStr TokenIdStr = TokenTokV[TokenTokN]->GetStrArgVal("id");
        TStr LeftId, RightId; TokenIdStr.SplitOnCh(LeftId, '.', RightId);
        EAssertR(LeftId.IsInt(), "Invalid Token ID " + TokenIdStr);
        EAssertR(RightId.IsInt(), "Invalid Token ID " + TokenIdStr);
        SentIdTokenIdV.Add(TIntPr(LeftId.GetInt(), RightId.GetInt()));
    }

	// xlike format contains mention tag: <mention sentenceId="3" words="motion"/>
	if (AnnInstTok->IsArg("sentenceId")) {
		const TStr SentId = AnnInstTok->GetStrArgVal("sentenceId");
		EAssertR(SentId != "", "XLike format contained annotation with empty sentenceId in mention tag");
		EAssertR(SentId.IsInt(), TStr::Fmt("XLike Sentence id ('%s') is not a valid integer", SentId.CStr()));
		SentIdTokenIdV.Add(TIntPr(SentId.GetInt(), TInt::Mx));
	}
}

PXmlTok TEnrycherAnnInst::GetXmlTok() const {
    PXmlTok AnnInstTok = TXmlTok::New("instance");
    AnnInstTok->AddArg("id", AnnInstId);
    AnnInstTok->AddArg("words", AnnInstStr);
    for (int SentIdTokenIdN = 0; SentIdTokenIdN < SentIdTokenIdV.Len(); SentIdTokenIdN++) {
        const TIntPr& SentIdTokenId = SentIdTokenIdV[SentIdTokenIdN];
        PXmlTok TokenTok = TXmlTok::New("token");
        TokenTok->AddArg("id", TStr::Fmt("%d.%d", SentIdTokenId.Val1.Val, SentIdTokenId.Val2.Val));
        AnnInstTok->AddSubTok(TokenTok);
    }
    return AnnInstTok;
}

///////////////////////////////
// Enrycher-Semantic-Attribute
TEnrycherSemAttr::TEnrycherSemAttr(const PXmlTok& SemAttrTok) {
	// attribute data
	TypeStr = SemAttrTok->GetStrArgVal("type", "owl:sameAs");
	ResourceStr = SemAttrTok->GetStrArgVal("resource");
	LiteralStr = SemAttrTok->GetTokStr(false);
	DisplayNameStr = SemAttrTok->GetStrArgVal("displayName");
	//HACK xlike
	if (ResourceStr.Empty()) { 
		// old xlike format has URL="..."
		if (SemAttrTok->IsArg("URL"))
			ResourceStr = SemAttrTok->GetStrArgVal("URL");
		else if (SemAttrTok->IsArg("URI"))
			ResourceStr = SemAttrTok->GetStrArgVal("URI");
		if (ResourceStr.Empty())
			throw PExcept(new TExcept("Unable to identify annotation uri"));
		const int LastSlash = ResourceStr.SearchChBack('/');
		if (LastSlash != -1) { DisplayNameStr = ResourceStr.Right(LastSlash+1); }
	}
	AboutStr = SemAttrTok->GetStrArgVal("about");
}

PXmlTok TEnrycherSemAttr::GetXmlTok() const {
	PXmlTok SemAttrTok = TXmlTok::New("attribute", LiteralStr);
    SemAttrTok->AddArg("type", TypeStr);
	if (IsResource()) { SemAttrTok->AddArg("resource", ResourceStr); }
	if (IsDisplayName()) { SemAttrTok->AddArg("displayName", DisplayNameStr); }
	if (IsAbout()) { SemAttrTok->AddArg("about", AboutStr); }
	return SemAttrTok;
}

///////////////////////////////
// Enrycher-Annotation
TEnrycherAnn::TEnrycherAnn(const PXmlTok& AnnTok) {
	//HACK xlike does not have annotation ids at the moment
	static int _AnnId = 1000;
    // annotation meta-data
	if (!AnnTok->IsArg("id"))
		AnnId = _AnnId++;
	else {
		TStr IdVal;
		if (AnnTok->IsArg("id"))
			IdVal = AnnTok->GetArgVal("id");
		IdVal = IdVal.TrimLeft("E");			// new format can start with "E"
		if (!IdVal.IsInt())
			throw PExcept(new TExcept("Unable to identify id number in the enrycher annotation for " + AnnTok->GetStrArgVal("displayName")));
		AnnId = IdVal.GetInt();
	}
    AnnStr = AnnTok->GetStrArgVal("displayName");
    AnnType = AnnTok->IsArg("type") ? AnnTok->GetStrArgVal("type") : "wiki";

	// HACK xlike does not have instances so we cannot compute weights of the concepts
	// instead, the <annotation tag contains an attribute weight that contains [0,1] value that we can
	// use as a weight. we store 1000*weight as an annotation instance that has sentence id = -1 and token id=1000*weight
	/*if (AnnTok->IsArg("weight")) {
		int Wgt = (int) (1000 * AnnTok->GetFltArgVal("weight", 0));
		TIntPr AnnInstPr(-1, Wgt);
		TIntPrV AnnInstPrV; AnnInstPrV.Add(AnnInstPr);
		TEnrycherAnnInst AnnInst("", AnnInstPrV);
		AnnInstV.Add(AnnInst);
	}*/

    // instances
	if (AnnTok->IsTagTok("instances")) {
		TXmlTokV AnnInstTokV; AnnTok->GetTagTokV("instances|instance", AnnInstTokV);
		for (int AnnInstTokN = 0; AnnInstTokN < AnnInstTokV.Len(); AnnInstTokN++) {
			TEnrycherAnnInst AnnInst(AnnInstTokV[AnnInstTokN]);
			if (!AnnInstV.IsIn(AnnInst)) { AnnInstV.Add(AnnInst); }
		}
	}
	// xlike instances
	else if (AnnTok->IsTagTok("mentions")) {		
		TXmlTokV AnnInstTokV; AnnTok->GetTagTokV("mentions|mention", AnnInstTokV);
		for (int AnnInstTokN = 0; AnnInstTokN < AnnInstTokV.Len(); AnnInstTokN++) {
			TEnrycherAnnInst AnnInst(AnnInstTokV[AnnInstTokN]);
			if (!AnnInstV.IsIn(AnnInst)) { AnnInstV.Add(AnnInst); }
		}
	}
    // semantics
	if (AnnTok->IsTagTok("semantics")) {
		TXmlTokV SemAttrTokV; AnnTok->GetTagTokV("semantics|attribute", SemAttrTokV);
		for (int SemAttrTokN = 0; SemAttrTokN < SemAttrTokV.Len(); SemAttrTokN++) {
			SemAttrV.Add(TEnrycherSemAttr(SemAttrTokV[SemAttrTokN]));
		}
	}
	// xlike attributes
	else if (AnnTok->IsTagTok("descriptions")) {
		TXmlTokV SemAttrTokV; AnnTok->GetTagTokV("descriptions|description", SemAttrTokV);
		for (int SemAttrTokN = 0; SemAttrTokN < SemAttrTokV.Len(); SemAttrTokN++) {
			SemAttrV.Add(TEnrycherSemAttr(SemAttrTokV[SemAttrTokN]));
		}
	}
}

PXmlTok TEnrycherAnn::GetXmlTok() const {
    PXmlTok AnnTok = TXmlTok::New("annotation");
    // annotation meta-data
    AnnTok->AddArg("id", GetId());
    AnnTok->AddArg("displayName", GetStr());
    AnnTok->AddArg("type", GetType());
    // instances
    PXmlTok AnnInstsTok = TXmlTok::New("instances");
    for (int AnnInstN = 0; AnnInstN < GetAnnInsts(); AnnInstN++) {
        AnnInstsTok->AddSubTok(GetAnnInst(AnnInstN).GetXmlTok());
    }
    AnnTok->AddSubTok(AnnInstsTok);
    // semantics
    PXmlTok SemTok = TXmlTok::New("semantics");
    for (int SemAttrN = 0; SemAttrN < GetSemAttrs(); SemAttrN++) {
        SemTok->AddSubTok(GetSemAttr(SemAttrN).GetXmlTok());
    }
    AnnTok->AddSubTok(SemTok);
	// done
	return AnnTok;
}

///////////////////////////////
// Enrycher-Assertion
TIntPr TEnrycherAssert::ParseAnnTok(const PXmlTok& Tok, const TStr& TagNm) const {
	EAssertR(Tok->IsTagTok(TagNm),
		TStr::Fmt("No %s in assertion %d!", TagNm.CStr(), AssertId.Val));
	PXmlTok SubTok = Tok->GetTagTok(TagNm);
	return TIntPr(
		SubTok->GetIntArgVal("annotationId"),
		SubTok->GetIntArgVal("instanceId"));
}

void TEnrycherAssert::ParseAnnTokV(const PXmlTok& Tok, const TStr& TagNm, TIntPrV& AnnIdV) const {
	TXmlTokV SubTokV; Tok->GetTagTokV(TagNm, SubTokV); AnnIdV.Gen(SubTokV.Len(), 0);
	for (int SubTokN = 0; SubTokN < SubTokV.Len(); SubTokN++) {
		const PXmlTok SubTok = SubTokV[SubTokN];
		AnnIdV.Add(TIntPr(
			SubTok->GetIntArgVal("annotationId"),
			SubTok->GetIntArgVal("instanceId")));
	}
}

PXmlTok TEnrycherAssert::GetAnnTok(const TIntPr& AnnId,
		const TStr& TagNm, const TEnrycherDoc* EnrycherDoc) const {

	PXmlTok AnnTok = TXmlTok::New(TagNm);
	AnnTok->AddArg("annotationId", AnnId.Val1);
	AnnTok->AddArg("instanceId", AnnId.Val2);
	AnnTok->AddArg("displayName", EnrycherDoc->GetAnnById(AnnId.Val1).GetStr());
	return AnnTok;
}

PXmlTok TEnrycherAssert::GetAnnTok(const TIntPr& AnnId, const TIntPrV& AnnIdV,
		const TStr& TagNm, const TStr& SubTagNm, const TEnrycherDoc* EnrycherDoc) const {

	PXmlTok AnnTok = GetAnnTok(AnnId, TagNm, EnrycherDoc);
	for (int AnnIdN = 0; AnnIdN < AnnIdV.Len(); AnnIdN++) {
		AnnTok->AddSubTok(GetAnnTok(AnnIdV[AnnIdN], SubTagNm, EnrycherDoc));
	}
	return AnnTok;
}


TEnrycherAssert::TEnrycherAssert(const PXmlTok& AssertTok) {
	AssertId = AssertTok->GetIntArgVal("id");
	// subject
	SubjAnnId = ParseAnnTok(AssertTok, "subject");
	ParseAnnTokV(AssertTok, "subject|modifier", SubjModAnnIdV);
	// verb
	VerbAnnId = ParseAnnTok(AssertTok, "verb");
	ParseAnnTokV(AssertTok, "verb|modifier", VerbModAnnIdV);
	// object
	ObjAnnId = ParseAnnTok(AssertTok, "object");
	ParseAnnTokV(AssertTok, "object|modifier", ObjModAnnIdV);
}

PXmlTok TEnrycherAssert::GetXmlTok(const TEnrycherDoc* EnrycherDoc) const {
	PXmlTok AssertTok = TXmlTok::New("assertion");
	AssertTok->AddArg("id", AssertId);
	AssertTok->AddSubTok(GetAnnTok(
		SubjAnnId, SubjModAnnIdV, "subject", "modifier", EnrycherDoc));
	AssertTok->AddSubTok(GetAnnTok(
		VerbAnnId, VerbModAnnIdV, "verb", "modifier", EnrycherDoc));
	AssertTok->AddSubTok(GetAnnTok(
		ObjAnnId, ObjModAnnIdV, "object", "modifier", EnrycherDoc));
	return AssertTok;
}

///////////////////////////////
// Enrycher-Pipeline-Trace
TEnrycherPipeline::TEnrycherPipeline(const PXmlTok& PipelineTok) {
    TXmlTokV ToolTokV; PipelineTok->GetTagTokV("tool", ToolTokV);
	for (int ToolTokN = 0; ToolTokN < ToolTokV.Len(); ToolTokN++) {
		const PXmlTok& ToolTok = ToolTokV[ToolTokN];
		// get tool name
		ToolNmV.Add(ToolTok->GetTokStr(false));
		// get execution time
		TStr ToolTmStr = ToolTok->GetStrArgVal("time");
		TTm ToolTm = TTm::GetTmFromWebLogDateTimeStr(
			ToolTmStr, '-', ':', '.', 'T');
		ToolTmV.Add(ToolTm);
		// get duration
		const int Duration = ToolTok->GetIntArgVal("durationTimeMillis", 0);
		ToolDurationV.Add(Duration);
	}
}

PXmlTok TEnrycherPipeline::GetXmlTok() const {
	PXmlTok PipelineTok = TXmlTok::New("pipeline");
	for (int ToolN = 0; ToolN < GetTools(); ToolN++) {
		PXmlTok ToolTok = TXmlTok::New("tool", GetToolNm(ToolN));
		ToolTok->AddArg("time", GetToolTm(ToolN).GetWebLogDateTimeStr(true, "T"));//, false));
		ToolTok->AddArg("durationTimeMillis", GetToolDuration(ToolN));
		PipelineTok->AddSubTok(ToolTok);
	}
	return PipelineTok;
}

///////////////////////////////
// Enrycher-Document
TEnrycherDoc::TEnrycherDoc(): DocId(TGuid::GenGuid()) {
	Pipeline = TEnrycherPipeline::New();
}

TEnrycherDoc::TEnrycherDoc(const PXmlTok& TopTok) {
	EAssertR(!TopTok.Empty(), "NULL document received");

	// get meta-data
    DocId = TopTok->GetStrArgVal("id");
    if (TopTok->IsSubTag("metadata")) {
		// document level semantic annotation
		if (TopTok->IsSubTag("metadata|semantics")) {
			TXmlTokV SemAttrTokV; TopTok->GetTagTokV("metadata|semantics|attribute", SemAttrTokV);
			for (int SemAttrTokN = 0; SemAttrTokN < SemAttrTokV.Len(); SemAttrTokN++) {
				SemAttrV.Add(TEnrycherSemAttr(SemAttrTokV[SemAttrTokN]));
			}
		}
		// pipeline execution trace
		if (TopTok->IsSubTag("metadata|pipeline")) {
			PXmlTok PipelineTok = TopTok->GetTagTok("metadata|pipeline");
			Pipeline = TEnrycherPipeline::LoadXml(PipelineTok); 
		} else {
			Pipeline = TEnrycherPipeline::New();
		}
    } else {
		Pipeline = TEnrycherPipeline::New();
	}
    // get paragraphs
    if (TopTok->IsSubTag("text")) {
        PXmlTok TextTok = TopTok->GetTagTok("text");
        const int ParaToks = TextTok->GetSubToks();
        for (int ParaTokN = 0; ParaTokN < ParaToks; ParaTokN++) {
            PXmlTok ParaTok = TextTok->GetSubTok(ParaTokN);
            if (ParaTok->IsTag()) {
				TEnrycherPara Para(ParaTok);
				if (!Para.Empty()) {
					const int ParaN = ParaV.Add(Para);
					// remember the sentence-id => (paragraph,sentenc)
					for (int SentN = 0; SentN < Para.GetSents(); SentN++) {
						TEnrycherSent& Sent = Para.GetSent(SentN);
						SentIdToParaNSentNH.AddDat(Sent.GetId()) = TIntPr(ParaN, SentN);
					}
				}
            }
        }
    }
	// Xlike format sentences
    if (TopTok->IsSubTag("sentences")) {
        PXmlTok TextTok = TopTok->GetTagTok("sentences");
		TEnrycherPara Para(TextTok);
		if (!Para.Empty()) {
			const int ParaN = ParaV.Add(Para);
			// remember the sentence-id => (paragraph,sentenc)
			for (int SentN = 0; SentN < Para.GetSents(); SentN++) {
				TEnrycherSent& Sent = Para.GetSent(SentN);
				SentIdToParaNSentNH.AddDat(Sent.GetId()) = TIntPr(ParaN, SentN);
			}
		}
    }
    // get annotations
    if (TopTok->IsSubTag("annotations")) {
        TXmlTokV AnnTokV; TopTok->GetTagTokV("annotations|annotation", AnnTokV);
        for (int AnnTokN = 0; AnnTokN < AnnTokV.Len(); AnnTokN++) {
			try {
				TEnrycherAnn Ann(AnnTokV[AnnTokN]);
				AnnH.AddDat(Ann.GetId(), Ann);
			}
			catch (PExcept Except) {
				TNotify::StdNotify->OnStatusFmt("ERROR: Problem parsing article annotation. Error: %s", Except->GetMsgStr().CStr());
			}
        }
    }
	// OLD xlike format annotations
	if (TopTok->IsSubTag("entities")) {
		TXmlTokV AnnTokV; TopTok->GetTagTokV("entities|entity", AnnTokV);
		for (int AnnTokN = 0; AnnTokN < AnnTokV.Len(); AnnTokN++) {
			try {
				TEnrycherAnn Ann(AnnTokV[AnnTokN]);
				AnnH.AddDat(Ann.GetId(), Ann);
			}
			catch (PExcept Except) {
				TNotify::StdNotify->OnStatusFmt("ERROR: Problem parsing article annotation. Error: %s", Except->GetMsgStr().CStr());
			}
		}
	}
	// xlike format annotations
    if (TopTok->IsSubTag("nodes")) {
        TXmlTokV AnnTokV; TopTok->GetTagTokV("nodes|node", AnnTokV);
        for (int AnnTokN = 0; AnnTokN < AnnTokV.Len(); AnnTokN++) {
			try {
				TEnrycherAnn Ann(AnnTokV[AnnTokN]);
				AnnH.AddDat(Ann.GetId(), Ann);
			}
			catch (PExcept Except) {
				TNotify::StdNotify->OnStatusFmt("ERROR: Problem parsing article annotation. Error: %s", Except->GetMsgStr().CStr());
			}
        }
    }
	// get assertions
	if (TopTok->IsSubTag("assertions")) {
        TXmlTokV AssertTokV; TopTok->GetTagTokV("assertions|assertion", AssertTokV);
        for (int AssertTokN = 0; AssertTokN < AssertTokV.Len(); AssertTokN++) {
			try {
				PXmlTok Assertion = AssertTokV[AssertTokN];
				if (Assertion->IsTagTok("subject") && Assertion->IsTagTok("verb") && Assertion->IsTagTok("object"))
					AssertV.Add(TEnrycherAssert(Assertion));
			}
			catch (PExcept Except) {
				TNotify::StdNotify->OnStatusFmt("ERROR: Problem parsing article annotation. Error: %s", Except->GetMsgStr().CStr());
			}
        }
	}
	// update the ids
	ResetNextIds();
}

void TEnrycherDoc::ResetNextIds() {
	// first sentence IDs
	NextSentId = 0;
	const int Paras = GetParas();
	for (int ParaN = 0; ParaN < Paras; ParaN++) {
		const TEnrycherPara& Para = GetPara(ParaN);
		const int Sents = Para.GetSents();
		for (int SentN = 0; SentN < Sents; SentN++) {
			const int SentId = Para.GetSent(SentN).GetId();
			UpdateNextId(SentId, NextSentId);
		}
	}
	// then the rest of IDs
	NextId = 0;
	int AnnKeyId = GetFirstAnn();
	while (GetNextAnn(AnnKeyId)) {
		const TEnrycherAnn& Ann = GetAnnByKeyId(AnnKeyId);
		UpdateNextId(Ann.GetId(), NextId);
		// check instances
		const int Insts = Ann.GetAnnInsts();
		for (int InstN = 0; InstN < Insts; InstN++) {
			UpdateNextId(Ann.GetAnnInst(InstN).GetId(), NextId);
		}
	}
	const int Asserts = GetAsserts();
	for (int AssertN = 0; AssertN < Asserts; AssertN++) {
		UpdateNextId(GetAssert(AssertN).GetAssertId(), NextId);
	}
}

PXmlDoc TEnrycherDoc::GetXmlDoc() const {
    PXmlTok TopTok = TXmlTok::New("item");

    // save meta-data
    TopTok->AddArg("id", DocId);
	PXmlTok MetaDataTok = TXmlTok::New("metadata");
	// document level semantic annotations
    PXmlTok SemTok = TXmlTok::New("semantics");
    for (int SemAttrN = 0; SemAttrN < GetSemAttrs(); SemAttrN++) {
        SemTok->AddSubTok(GetSemAttr(SemAttrN).GetXmlTok());
    }
    MetaDataTok->AddSubTok(SemTok);
	// pipeline execution trace
	if (!Pipeline.Empty()) {
		MetaDataTok->AddSubTok(Pipeline->GetXmlTok()); }
	TopTok->AddSubTok(MetaDataTok);

    // save paragraphs
    PXmlTok TextTok = TXmlTok::New("text");
    for (int ParaN = 0; ParaN < ParaV.Len(); ParaN++) {
		PXmlTok ParaTok = ParaV[ParaN].GetXmlTok();
        TextTok->AddSubTok(ParaTok);
    }
    TopTok->AddSubTok(TextTok);

    // save annotations
	if (!AnnH.Empty()) {
		PXmlTok AnnsTok = TXmlTok::New("annotations");
		int AnnKeyId = GetFirstAnn();
		while (GetNextAnn(AnnKeyId)) {
			AnnsTok->AddSubTok(GetAnnByKeyId(AnnKeyId).GetXmlTok()); }
		TopTok->AddSubTok(AnnsTok);
	}

	// save assertions
	if (!AssertV.Empty()) {
		PXmlTok AssertTok = TXmlTok::New("assertions");
		for (int AssertN = 0; AssertN < AssertV.Len(); AssertN++) {
			AssertTok->AddSubTok(AssertV[AssertN].GetXmlTok(this));
		}
		TopTok->AddSubTok(AssertTok);
	}

	// create and return XML doc
    return TXmlDoc::New(TopTok);
}

int TEnrycherDoc::AddSemAttr(const TEnrycherSemAttr& NewSemAttr) {
	// remember the new annotation semantic attribute
	int SemAttrN = SemAttrV.Add(NewSemAttr);
	return SemAttrN;
}

/* Iterate thorugh the <semantics> element to find the one describing the document-level sentiment. */
double TEnrycherDoc::GetSentiment() const { 
	// format sample: <attribute about="_:sentiment-of-eb3a6081-15e2-4852-9e2a-b2994b00ea59" type="kdo:hasScore">0.004708</attribute>
	for (int i=0; i<SemAttrV.Len(); i++) {
		if (SemAttrV[i].GetAbout().IsStrIn("sentiment-of") && SemAttrV[i].GetType()=="kdo:hasScore") {
			return SemAttrV[i].GetLiteral().GetFlt();
		}
	}
	return 0;
}

const TEnrycherSent& TEnrycherDoc::GetSent(int SentId) const {
	const TIntPr& ParaNSentN = SentIdToParaNSentNH.GetDat(SentId);
	return ParaV[ParaNSentN.Val1].GetSent(ParaNSentN.Val2);
}

bool TEnrycherDoc::IsSentId(const int SentId) const
{
	return SentIdToParaNSentNH.IsKey(SentId);
}

int TEnrycherDoc::AddPara(const TEnrycherPara& NewPara) { 
	// remember the new paragraph
	int ParaN = ParaV.Add(NewPara); 
	// assign proper ids to sentences and tokens
	TEnrycherPara& Para = ParaV[ParaN];
	for (int SentN = 0; SentN < Para.GetSents(); SentN++) {
		TEnrycherSent& Sent = Para.GetSent(SentN);
		Sent.SetId(NextSentId); NextSentId++;
		for (int TokenN = 0; TokenN < Sent.GetTokens(); TokenN++) {
			Sent.GetToken(TokenN).SetId(TokenN);
		}
		// remember the map
		SentIdToParaNSentNH.AddDat(Sent.GetId()) = TIntPr(ParaN, SentN);
	}
	// return sequence number of the new paragraph
	return ParaN;
}

int TEnrycherDoc::AddAnn(const TEnrycherAnn& NewAnn) {
	TInt AnnId = NextId; NextId++;
	// remember the new annotation
	AnnH.AddDat(AnnId, NewAnn);
	TEnrycherAnn& Ann = AnnH.GetDat(AnnId);
	// assign the annotation a proper id
	Ann.SetId(AnnId);
	
	return AnnId;
}

int TEnrycherDoc::AddAnnInst(int AnnId, const TEnrycherAnnInst& NewAnnInst) {
	// remember the new annotation instance
	TEnrycherAnn& Ann = AnnH.GetDat(AnnId);
	int AnnInstN = Ann.AddAnnInst(NewAnnInst);
	TEnrycherAnnInst& AnnInst = Ann.GetAnnInst(AnnInstN);
	// assign the annotation instance a proper id
	AnnInst.SetId(AnnInstN);
	return AnnInstN;
}

int TEnrycherDoc::AddAnnSemAttr(int AnnId, const TEnrycherSemAttr& NewSemAttr) {
	// remember the new annotation semantic attribute
	TEnrycherAnn& Ann = AnnH.GetDat(AnnId);
	int SemAttrN = Ann.AddSemAttr(NewSemAttr);
	return SemAttrN;
}

void TEnrycherDoc::GetAllText(TChA& DocChA) const {
	const int DocParas = ParaV.Len();
	for (int DocParaN = 0; DocParaN < DocParas; DocParaN++) {
		if (DocParaN > 0) { DocChA += '\n'; }
		const TEnrycherPara Para = ParaV[DocParaN];
		for (int SentN = 0; SentN < Para.GetSents(); SentN++) {
			if (DocParaN > 0) { DocChA += ' '; }
			DocChA += Para.GetSent(SentN).GetStr();
		}
	}
}

void TEnrycherDoc::AddAsciiCharsToUni(const char *Str, TIntV& XmlChA) {
	int Len = (int) strlen(Str);
	for (int i = 0; i < Len; i++) {
		XmlChA.Add(Str[i]);
	}
}

void TEnrycherDoc::AddStrToXmlUni(const TStr& PlainChA, TIntV& XmlChA) {
	TIntV UniChA(PlainChA.Len(), 0);
	AddStrToUni(PlainChA, UniChA);
	AddUniToXmlUni(UniChA, XmlChA);
}

void TEnrycherDoc::AddUniToXmlUni(const TIntV& PlainChA, TIntV& XmlChA){
	for (int ChN=0; ChN<PlainChA.Len(); ChN++) {
		int Ch=PlainChA[ChN];
		switch (Ch){
			case '"': AddAsciiCharsToUni("&quot;", XmlChA); break;
			case '&': AddAsciiCharsToUni("&amp;", XmlChA); break;
			case '\'':AddAsciiCharsToUni("&apos;", XmlChA); break;
			case '<': AddAsciiCharsToUni("&lt;", XmlChA); break;
			case '>': AddAsciiCharsToUni("&gt;", XmlChA); break;
			default: XmlChA.Add(Ch);
		}
	}
}

void TEnrycherDoc::AddStrToUni(const TStr& Str, TIntV& ChA) {
	TUnicodeDef::GetDef()->codec.DecodeUtf8(Str, ChA, false);
}


void TEnrycherDoc::AddTokToUni(const PXmlTok& Tok, const bool& XmlP, TIntV& ChA) {
	switch (Tok->GetSym()){
    case xsyWs:
		AddStrToUni(Tok->GetStr(), ChA); break;
    case xsyStr:
		if (XmlP){AddStrToXmlUni(Tok->GetStr(), ChA); } else {AddStrToUni(Tok->GetStr(), ChA);} break;
    case xsyQStr:
      if (XmlP){AddAsciiCharsToUni("<![CDATA[", ChA);}
      AddStrToUni(Tok->GetStr(), ChA);
      if (XmlP){AddAsciiCharsToUni("]]>", ChA);} break;
    case xsyTag:
      if (XmlP){
        ChA.Add('<'); AddStrToUni(Tok->GetStr(), ChA);
        for (int ArgN=0; ArgN<Tok->GetArgs(); ArgN++){
          TStr ArgNm; TStr ArgVal; Tok->GetArg(ArgN, ArgNm, ArgVal);
          if (XmlP){ArgVal=TXmlLx::GetXmlStrFromPlainStr(ArgVal);}
          char ArgValQCh=TXmlLx::GetArgValQCh(ArgVal);
          ChA.Add(' '); AddStrToUni(ArgNm, ChA); ChA.Add('=');
		  ChA.Add(ArgValQCh); AddStrToUni(ArgVal, ChA);; ChA.Add(ArgValQCh);
        }
      }
      if (Tok->GetSubToks()==0){
        if (XmlP){AddAsciiCharsToUni("/>", ChA);}
      } else {
		  if (XmlP){ChA.Add('>');}
        for (int SubTokN=0; SubTokN<Tok->GetSubToks(); SubTokN++){
          AddTokToUni(Tok->GetSubTok(SubTokN), XmlP, ChA);}
		if (XmlP){AddAsciiCharsToUni("</", ChA); AddStrToUni(Tok->GetStr(), ChA); ChA.Add('>');}
      }
      break;
    default: Fail;
  }
}

void TEnrycherDoc::SaveUni(const PXmlDoc& Doc, TIntV& Out) {
	AddTokToUni(Doc->GetTok(), true, Out);
}
void TEnrycherDoc::SaveTxt(const PXmlDoc& Doc, const PSOut& Out) {
	TIntV UnicodeDoc(1024,0);
	SaveUni(Doc, UnicodeDoc);
	TIntV Utf8(UnicodeDoc.Len(), 0);
	TUnicodeDef::GetDef()->EncodeUtf8(UnicodeDoc, Utf8);
	for (int i = 0; i < Utf8.Len(); i++) {
		int Char = Utf8[i];
		Assert(Char <= TUCh::Mx);
		Out->PutCh((char) Char);
	}
}
