/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ActivityBase.h"

//////////////////////////////////////////////////////////////////////
/// Load existing user base
TActivityBase::TActivityBase(const PNewsBase _NewsBase, const TStr& IndexFPath,  const TFAccess& _FAccess, 
		const int64& IndexCacheSize, const int64& StoreCacheSize) {

	FAccess = _FAccess;
	NewsBase = _NewsBase;

	EAssert(FAccess == faRdOnly || FAccess == faUpdate || FAccess == faRestore);

	GenericBase = TGenericBase::Load(IndexFPath, FAccess, IndexCacheSize, StoreCacheSize, TStrUInt64H(), false);

	CommonInit();
}

// create new user base
TActivityBase::TActivityBase(const TStr& StoreDefFNm, const PNewsBase _NewsBase, const TStr& IndexFPath, 
		const int64& IndexCacheSize, const int64& StoreCacheSize) 
{
	FAccess = faCreate;
	TQm::TEnv::Logger->OnStatus("Opening in create mode");

	NewsBase = _NewsBase;

	// prepare base
	GenericBase = TGenericBase::New(IndexFPath, IndexCacheSize, StoreCacheSize, false);
	// initialize store
	PJsonVal StoreDefVal = TJsonVal::GetValFromStr(TStr::LoadTxt(StoreDefFNm));
	GenericBase->CreateSchema(StoreDefVal);
	
	CommonInit();
}

void TActivityBase::CommonInit() {
	Base = GenericBase->Base;
	IndexVoc = Base->GetIndexVoc();
	Index = Base->GetIndex();
	
	UserStore = Base->GetStoreByStoreNm("User");
	FolderStore = Base->GetStoreByStoreNm("Folder");
	ArtActStore = Base->GetStoreByStoreNm("ArticleActivity");
	ActStore = Base->GetStoreByStoreNm("Activity");

	Base->AddStore(NewsBase->ArticleStore);
	Base->AddStore(NewsBase->SourceStore);
	Base->AddStore(NewsBase->StoryStore);
	Base->AddStore(NewsBase->ConceptStore);
				
	ArticleStoreId = NewsBase->ArticleStore->GetStoreId();
	ConceptStoreId = NewsBase->ConceptStore->GetStoreId();
	UserStoreId = UserStore->GetStoreId();
	FolderStoreId = FolderStore->GetStoreId();;
	ArtActStoreId = ArtActStore->GetStoreId();
	ActStoreId = ActStore->GetStoreId();

	// finish initalization
	Base->Init();

	// create folder joins if not existing
	
	// stored articles are the ones the user added to a folder
	if (!FolderStore->IsJoinNm("hasStoredArticle"))
		FolderStore->AddJoinDesc(TJoinDesc("hasStoredArticle", ArticleStoreId, FolderStoreId, IndexVoc));
	JoinFolderHasStoredArticlesId = FolderStore->GetJoinId("hasStoredArticle");

	// articles that the user liked
	if (!FolderStore->IsJoinNm("hasLikedArticle"))
		FolderStore->AddJoinDesc(TJoinDesc("hasLikedArticle", ArticleStoreId, FolderStoreId, IndexVoc));
	JoinFolderHasLikedArticlesId = FolderStore->GetJoinId("hasLikedArticle");

	// article that the user liked but then removed the like
	if (!FolderStore->IsJoinNm("hasLikedArticleInPast"))
		FolderStore->AddJoinDesc(TJoinDesc("hasLikedArticleInPast", ArticleStoreId, FolderStoreId, IndexVoc));
	JoinFolderHasLikedArticlesInPastId = FolderStore->GetJoinId("hasLikedArticleInPast");

	// activities for a folder
	if (!FolderStore->IsJoinNm("hasActivity"))
		FolderStore->AddJoinDesc(TJoinDesc("hasActivity", ActStoreId, FolderStoreId, IndexVoc));
	JoinFolderHasActivityId= FolderStore->GetJoinId("hasActivity");

	// ArticleStore --> Folder
	// an article has been liked by ... users
	if (!NewsBase->ArticleStore->IsJoinNm("hasBeenLikedBy"))
		NewsBase->ArticleStore->AddJoinDesc(TJoinDesc("hasBeenLikedBy", FolderStoreId, ArticleStoreId, IndexVoc));
	JoinArticleLikedById = NewsBase->ArticleStore->GetJoinId("hasBeenLikedBy");

	// an article has been stored by ... users
	if (!NewsBase->ArticleStore->IsJoinNm("hasBeenStoredBy"))
		NewsBase->ArticleStore->AddJoinDesc(TJoinDesc("hasBeenStoredBy", FolderStoreId, ArticleStoreId, IndexVoc));
	JoinArticleStoredBy = NewsBase->ArticleStore->GetJoinId("hasBeenStoredBy");

	// FolderStore --> ArticleActivity store
	if (!FolderStore->IsJoinNm("hasArtActStored")) {
		// stored articles are the ones the user added to a folder
		FolderStore->AddJoinDesc(TJoinDesc("hasArtActStored", ArtActStoreId, FolderStoreId, IndexVoc));
	}
	JoinFolderHasArtActStoredId = FolderStore->GetJoinId("hasArtActStored");

	// FolderStore --> ArticleActivity store
	if (!FolderStore->IsJoinNm("hasArtActLiked")) {
		FolderStore->AddJoinDesc(TJoinDesc("hasArtActLiked", ArtActStoreId, FolderStoreId, IndexVoc));
	}
	JoinFolderHasArtActLikedId = FolderStore->GetJoinId("hasArtActLiked");
}

TActivityBase::~TActivityBase() 
{ 
	// save if necessary
	if (FAccess != faRdOnly) {
		Base->RemoveStore(NewsBase->ArticleStore);
		Base->RemoveStore(NewsBase->SourceStore);
		Base->RemoveStore(NewsBase->StoryStore);
		Base->RemoveStore(NewsBase->ConceptStore);
	} else {
		TQm::TEnv::Logger->OnStatus("No saving neccessary!\n");
	}
}

PActivityBase TActivityBase::New(const TStr& CuratedNewsDef, const PNewsBase _NewsBase, const TStr& IndexFPath, const int64& IndexCacheSize, const int64& StoreCacheSize) {
			
	return TIndex::Exists(IndexFPath) ?
		new TActivityBase(_NewsBase, IndexFPath, faUpdate, IndexCacheSize, StoreCacheSize) :
		new TActivityBase(CuratedNewsDef, _NewsBase, IndexFPath, IndexCacheSize, StoreCacheSize);
}

//////////////////////////////////////
// folder methods
uint64 TActivityBase::AddFolder(const TStr& UserSqlId, const TStr& FolderId)
{
	const uint64 ExistingFolderRecId = GetFolderRecId(UserSqlId, FolderId);
	if (ExistingFolderRecId != TUInt64::Mx)
		return ExistingFolderRecId;

	PJsonVal UserJson = TJsonVal::NewObj();
	UserJson->AddToObj("UserSqlId", TJsonVal::NewStr(UserSqlId));
	//UserJson->AddToObj("hasFolder", FolderJson);
	const uint64 UserRecId = GenericBase->AddRec(UserStoreId, UserJson);

	TStr Key = CreateKey(UserSqlId, FolderId);
	PJsonVal FolderJson = TJsonVal::NewObj();
	FolderJson->AddToObj("URI", TJsonVal::NewStr(Key));
	const uint64 FolderRecId = GenericBase->AddRec(FolderStoreId, FolderJson);
	
	Index->IndexJoin(UserStore, "hasFolder", UserRecId, FolderRecId);

	return FolderRecId;
}

bool TActivityBase::IsFolder(const TStr& UserSqlId, const TStr& FolderId)
{
	return GetFolderRecId(UserSqlId, FolderId) != TUInt64::Mx;
}

//void TActivityBase::DeleteFolder(const TStr& UserSqlId, const TStr& FolderId)
//{
//
//}

uint64 TActivityBase::GetFolderRecId(const TStr& UserSqlId, const TStr& FolderId)
{
	if (!UserStore->IsRecNm(UserSqlId))
		return TUInt64::Mx;
	TStr Key = CreateKey(UserSqlId, FolderId);
	if (FolderStore->IsRecNm(Key))
		return FolderStore->GetRecId(Key);
	return TUInt64::Mx;
}

PRecSet TActivityBase::GetFoldersForUser(const TStr& UserSqlId)
{
	if (!UserStore->IsRecNm(UserSqlId))
		return NULL;
	uint64 UserRecId = UserStore->GetRecId(UserSqlId);
	TRec UserRec(UserStoreId, UserRecId);
	PRecSet FoldersRecSet = UserRec.DoJoin(Base, "hasFolder");
	return FoldersRecSet;
}

// add an article to a particular folder
void TActivityBase::AddStoredArticle(const uint64& FolderRecId, const uint64& ArticleRecId)
{
	// add article activity
	PJsonVal ActJson = TJsonVal::NewObj();
	TStr TimeNow = TTm::GetCurLocTm().GetWebLogDateTimeStr(true, "T", false);
	ActJson->AddToObj("DateTime", TimeNow);
	ActJson->AddToObj("ArticleRecId", (int) ArticleRecId);
	const uint64 ActRecId = GenericBase->AddRec(ArtActStoreId, ActJson);
	// join folder -> article activity store
	Index->IndexJoin(FolderStore, JoinFolderHasArtActStoredId, FolderRecId, ActRecId);
	// join folder -> article store
	Index->IndexJoin(FolderStore, JoinFolderHasStoredArticlesId, FolderRecId, ArticleRecId);
	// join article -> folder 
	Index->IndexJoin(NewsBase->ArticleStore, JoinArticleStoredBy, ArticleRecId, FolderRecId);
}

// remove the article that is already stored in a folder from the folder
void TActivityBase::RemoveStoredArticle(const uint64& FolderRecId, const uint64& ArticleRecId)
{
	// remove join folder -> article activity store
	PRecSet ArtSet = GetStoredArtAct(FolderRecId);
	for (int N=0; N < ArtSet->GetRecs(); N++) {
		uint64 ActRecId = ArtSet->GetRecId(N);
		if (ArtActStore->GetFieldNmUInt64(ActRecId, "ArticleRecId") == ArticleRecId) {
			Index->DeleteJoin(FolderStore, JoinFolderHasArtActStoredId, FolderRecId, ActRecId);
			break;
		}
	}
	// join folder -> article store
	Index->DeleteJoin(FolderStore, JoinFolderHasStoredArticlesId, FolderRecId, ArticleRecId);
	// remove article -> folder
	Index->DeleteJoin(NewsBase->ArticleStore, JoinArticleStoredBy, ArticleRecId, FolderRecId);
}

bool TActivityBase::IsLikedArticle(const uint64& FolderRecId, const uint64& ArticleRecId)
{
	PRecSet ArtSet = GetLikedArticles(FolderRecId);
	for (int N=0; N < ArtSet->GetRecs(); N++) {
		if (ArtSet->GetRecId(N) == ArticleRecId)
			return true;
	}
	return false;
}

bool TActivityBase::IsLikedArticleInPast(const uint64& FolderRecId, const uint64& ArticleRecId)
{
	PRecSet ArtSet = GetLikedArticlesInPast(FolderRecId);
	for (int N=0; N < ArtSet->GetRecs(); N++) {
		if (ArtSet->GetRecId(N) == ArticleRecId)
			return true;
	}
	return false;
}

void TActivityBase::AddLikedArticle(const uint64& FolderRecId, const uint64& ArticleRecId)
{
	// add article activity
	PJsonVal ActJson = TJsonVal::NewObj();
	TStr ArticleUri = NewsBase->ArticleStore->GetRecNm(ArticleRecId);
	TStr TimeNow = TTm::GetCurLocTm().GetWebLogDateTimeStr(true, "T", false);
	ActJson->AddToObj("DateTime", TimeNow);
	ActJson->AddToObj("ArticleURI", ArticleUri);
	const uint64 ActRecId = GenericBase->AddRec(ArtActStoreId, ActJson);
	// join folder -> article activity store
	Index->IndexJoin(FolderStore, JoinFolderHasArtActLikedId, FolderRecId, ActRecId);
	// join folder -> article store
	Index->IndexJoin(FolderStore, JoinFolderHasLikedArticlesId, FolderRecId, ArticleRecId);
	// join article -> folder 
	Index->IndexJoin(NewsBase->ArticleStore, JoinArticleLikedById, ArticleRecId, FolderRecId);
}

// remove the article that is already liked from the folder
void TActivityBase::RemoveLikedArticle(const uint64& FolderRecId, const uint64& ArticleRecId)
{
	// remove join folder -> article activity store
	PRecSet ArtSet = GetLikedArtAct(FolderRecId);
	for (int N=0; N < ArtSet->GetRecs(); N++) {
		uint64 ActRecId = ArtSet->GetRecId(N);
		if (ArtActStore->GetFieldNmUInt64(ActRecId, "ArticleRecId") == ArticleRecId) {
			// remove that the article is liked
			Index->DeleteJoin(FolderStore, JoinFolderHasArtActLikedId, FolderRecId, ActRecId);
			break;
		}
	}
	// remember that the article was liked in the past
	Index->IndexJoin(FolderStore, JoinFolderHasLikedArticlesInPastId, FolderRecId, ArticleRecId);

	// join folder -> article store
	Index->DeleteJoin(FolderStore, JoinFolderHasLikedArticlesId, FolderRecId, ArticleRecId);
	// remove article -> folder
	Index->DeleteJoin(NewsBase->ArticleStore, JoinArticleLikedById, ArticleRecId, FolderRecId);
}

// remove all curated articles for a folder and remove the folder
void TActivityBase::DeleteFolder(const uint64& FolderRecId)
{
	// delete liked article activity for the folder
	PRecSet ActSet = GetLikedArtAct(FolderRecId);
	for (int RecN=0; RecN < ActSet->GetRecs(); RecN++)
		Index->DeleteJoin(FolderStore, "hasArtActLiked", FolderRecId, ActSet->GetRecId(RecN));

	// delete stored article activity for the folder
	ActSet = GetStoredArtAct(FolderRecId);
	for (int RecN=0; RecN < ActSet->GetRecs(); RecN++)
		Index->DeleteJoin(FolderStore, "hasArtActStored", FolderRecId, ActSet->GetRecId(RecN));
	
	// delete stored articles for the folder
	ActSet = GetStoredArticles(FolderRecId);
	for (int RecN=0; RecN < ActSet->GetRecs(); RecN++)
		Index->DeleteJoin(FolderStore, "hasStoredArticle", FolderRecId, ActSet->GetRecId(RecN));
	
	// delete stored articles for the folder
	ActSet = GetLikedArticles(FolderRecId);
	for (int RecN=0; RecN < ActSet->GetRecs(); RecN++)
		Index->DeleteJoin(FolderStore, "hasLikedArticle", FolderRecId, ActSet->GetRecId(RecN));

	// delete stored articles for the folder
	ActSet = GetLikedArticlesInPast(FolderRecId);
	for (int RecN=0; RecN < ActSet->GetRecs(); RecN++)
		Index->DeleteJoin(FolderStore, "hasLikedArticleInPast", FolderRecId, ActSet->GetRecId(RecN));
	
	// remove the folder from the folder store
	//FolderStore->DeleteFolder(FolderRecId);
}


PRecSet TActivityBase::GetLikedArtAct(const uint64& FolderRecId)
{
	TRec FolderRec(FolderStoreId, FolderRecId);
	return FolderRec.DoJoin(Base, "hasArtActLiked");
}

PRecSet TActivityBase::GetStoredArtAct(const uint64& FolderRecId)
{
	TRec FolderRec(FolderStoreId, FolderRecId);
	return FolderRec.DoJoin(Base, "hasArtActStored");
}

// return the articles that the user added to a folder
PRecSet TActivityBase::GetStoredArticles(const uint64& FolderRecId)
{
	TRec FolderRec(FolderStoreId, FolderRecId);
	return FolderRec.DoJoin(Base, "hasStoredArticle");
}

PRecSet TActivityBase::GetLikedArticles(const uint64& FolderRecId)
{
	EAssertR(FolderStore->IsRecId(FolderRecId), "The folder id is not valid");
	TRec FolderRec(FolderStoreId, FolderRecId);
	return FolderRec.DoJoin(Base, "hasLikedArticle");
}

PRecSet TActivityBase::GetLikedArticlesInPast(const uint64& FolderRecId)
{
	EAssertR(FolderStore->IsRecId(FolderRecId), "The folder id is not valid");
	TRec FolderRec(FolderStoreId, FolderRecId);
	return FolderRec.DoJoin(Base, "hasLikedArticleInPast");
}

void TActivityBase::SaveData(const TStr& FName, const PStore Store)
{
	PSOut Output = TFOut::New(FName, false);
	PRecSet RecSet = Store->GetAllRecs();
	uint64 Count = RecSet->GetRecs();
	uchar StoreId = Store->GetStoreId();
	for (int N=0; N < Count; N++) {
		try
		{
			const TRec Rec = TRec(StoreId, RecSet->GetRecId(N));
			PJsonVal Json = Rec.GetJson(Base, true, true, true, true);
			Output->PutStr(TJsonVal::GetStrFromVal(Json) + "\n");
		}
		catch (PExcept E)
		{
			TQm::TEnv::Logger->OnStatus("Exception while saving data to file " + FName + ": " + E->GetMsgStr());
		}
	}
}


void TActivityBase::LoadData(const TStr& FName, const PStore Store)
{
	if (!TFile::Exists(FName)) {
		TQm::TEnv::Logger->OnStatus("Unable to find file: " + FName); 
		return;
	}
	uchar StoreId = Store->GetStoreId();
	TFIn Input(FName);
	while (true) {
		try
		{
			TChA Line;
			bool ok = Input.GetNextLn(Line);
			if (!ok)
				break;
			PJsonVal Json = TJsonVal::GetValFromStr(Line);
			GenericBase->AddRec(StoreId, Json);
		}
		catch (PExcept E)
		{
			TQm::TEnv::Logger->OnStatus("Exception while importing data from file " + FName + ": " + E->GetMsgStr());
		}
	}
}

void TActivityBase::SaveJoinsData()
{
	TTm Time = TTm::GetCurLocTm();
	TStr TimeStr = Time.GetYMDDashStr();

	PSOut LikedOutput = TFOut::New(TimeStr + "-hasLikedArticle.txt", false);
	PSOut LikedInPastOutput = TFOut::New(TimeStr + "-hasLikedArticleInPast.txt", false);
	PSOut StoredOutput = TFOut::New(TimeStr + "-hasStoredArticle.txt", false);

	PStoreIter Iter = FolderStore->GetIter();
	while (Iter->Next()) {
		const uint64 RecId = Iter->GetRecId();
		const TStr Key = FolderStore->GetRecNm(RecId);
		
		TRec FolderRec(FolderStoreId, RecId);
		PRecSet LikedSet = FolderRec.DoJoin(Base, "hasLikedArticle");
		if (LikedSet->GetRecs() > 0) {
			TStr LikedStr = GetListOfRecIds(LikedSet);
			LikedOutput->PutStr(Key + "," + LikedStr + "\n");
		}

		PRecSet LikedInPastSet = FolderRec.DoJoin(Base, "hasLikedArticleInPast");
		if (LikedInPastSet->GetRecs() > 0) {
			TStr LikedInPastStr = GetListOfRecIds(LikedInPastSet);
			LikedInPastOutput->PutStr(Key + "," + LikedInPastStr + "\n");
		}

		PRecSet StoredSet = FolderRec.DoJoin(Base, "hasStoredArticle");
		if (StoredSet->GetRecs() > 0) {
			TStr StoredStr = GetListOfRecIds(StoredSet);
			StoredOutput->PutStr(Key + "," + StoredStr + "\n");
		}
	}
}

void TActivityBase::LoadJoinsData(const TStr& FName, const int JoinId)
{
	if (!TFile::Exists(FName)) {
		TQm::TEnv::Logger->OnStatus("Unable to find file: " + FName); 
		return;
	}
	TFIn Input(FName);
	while (true) {
		try
		{
			TStr Line;
			bool ok = Input.GetNextLn(Line);
			if (!ok)
				break;
			TStrV ArticleIdStrV;
			Line.SplitOnAllCh(',', ArticleIdStrV);
			if (ArticleIdStrV.Len() == 0)
				continue;
			TStr Key = ArticleIdStrV[0];
			if (!FolderStore->IsRecNm(Key)) {
				TQm::TEnv::Logger->OnStatus("Folder " + Key + " does not exist. Not importing joins for it"); 
				continue;
			}
			const uint64 FolderRecId = FolderStore->GetRecId(Key);
			for (int N=1; N < ArticleIdStrV.Len(); N++) 
				Index->IndexJoin(FolderStore, JoinId, FolderRecId, ArticleIdStrV[N].GetUInt64(0));
		}
		catch (PExcept E)
		{
			TQm::TEnv::Logger->OnStatus("Exception while importing data from file " + FName + ": " + E->GetMsgStr());
		}
	}
}
// get a comma separated list of rec ids
TStr TActivityBase::GetListOfRecIds(const PRecSet& RecSet)
{
	TChA Output;
	int Count = RecSet->GetRecs();
	for (int N=0; N < Count; N++) {
		if (N > 0) Output += ',';
		Output += TUInt64(RecSet->GetRecId(N)).GetStr();
	}
	return Output;
}
