/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ActivityBase.h"
#include "ActivitySrv.h"
#include <base.h>
#include "../NewsMiner/NewsConst.h"

///////////////////////////////////////////////
// PlusMachines-Server-Helper-Utilities
TNewsSrvUtil::TNewsSrvUtil(const PNewsBase& _NewsBase, const PActivityBase& _ActivityBase): NewsBase(_NewsBase), ActivityBase(_ActivityBase) { 
}

PJsonVal TNewsSrvUtil::Article(const TRec& ArticleRec, const bool& FullBodyP) const {
	const uint64 ArticleId = ArticleRec.GetRecId();

	// get main fields	
	const TStr Uri = NewsBase->ArticleStore->GetRecNm(ArticleId);
	const TStr Title = NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "Title");
	const TStr Body = NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "Body");
	const TStr Lang = NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "Language");
	TTm Tm; NewsBase->ArticleStore->GetFieldNmTm(ArticleId, "DateTime", Tm);
	
	// fix paragraphs in body
	TStr ParaBody;
	if (!FullBodyP) { 
		// keep only first paragraph
		TStr Leftover; Body.SplitOnStr(ParaBody, "\n\n", Leftover); 
		const int MxChs = 300;
		if (ParaBody.Len() > MxChs) {
			int NextSpace = ParaBody.SearchCh(' ', MxChs);
			ParaBody = ParaBody.Left(NextSpace != -1 ? NextSpace : MxChs);
			ParaBody += "...";
		}
		ParaBody = "<p>" + ParaBody + "</p>";
	} else {
		// keep all paragraphs
		ParaBody = "<p>" + Body + "</p>";
		ParaBody.ChangeStrAll("\n\n", "</p><p>");
	}
	
	// get source info
	PRecSet SourceSet = ArticleRec.DoJoin(NewsBase->Base, "hasSource");
	TStr SourceTitle = "[unknown]", SourceUri = "#";
	uint64 SourceId = 0;
	if (!SourceSet->Empty()) {
		SourceId = SourceSet->GetRecId(0);
		SourceTitle = NewsBase->SourceStore->GetFieldNmStr(SourceId, "Title");
		SourceUri = NewsBase->SourceStore->GetFieldNmStr(SourceId, "URI");
		if (SourceTitle == "") {
			SourceTitle = SourceUri;
			SourceTitle.DelStr("http://");
			SourceTitle.DelStr("https://");
			SourceTitle.DelStr("www.");
		}
	}	

	// prepare result JSon object
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("id", TJsonVal::NewStr(TUInt64::GetStr(ArticleId)));
	ResJson->AddToObj("uri", TJsonVal::NewStr(Uri));
	ResJson->AddToObj("title", TJsonVal::NewStr(Title));
	ResJson->AddToObj("body", TJsonVal::NewStr(ParaBody));
	ResJson->AddToObj("date", TJsonVal::NewStr(Tm.IsDef() ? Tm.GetWebLogDateStr() : "[unknown]"));
	ResJson->AddToObj("time", TJsonVal::NewStr(Tm.IsDef() ? Tm.GetWebLogTimeStr() : "[unknown]"));
	ResJson->AddToObj("lang", TJsonVal::NewStr(Lang));
	ResJson->AddToObj("source_title", TJsonVal::NewStr(SourceTitle.GetUc()));
	ResJson->AddToObj("source_uri", TJsonVal::NewStr(SourceUri));
	ResJson->AddToObj("source_id", TJsonVal::NewStr(TInt::GetStr((int) SourceId)));

	// add article image 	
	PRecSet ImageSet = ArticleRec.DoJoin(NewsBase->Base, "hasImage");
	if (ImageSet->GetRecs() > 0) {
		TStr ImageUri = NewsBase->ImageStore->GetFieldNmStr(ImageSet->GetRecId(0), "URI");
		ResJson->AddToObj("image_uri", TJsonVal::NewStr(ImageUri));
	}

	// done
	return ResJson;
}

PJsonVal TNewsSrvUtil::Articles(const PRecSet& ArticleSet,
		const bool& FullBodyP, const TUInt64Set& IgnoreArticleSet) const {

	TJsonValV ArticleJsonV;
	for (int ArticleN = 0; ArticleN < ArticleSet->GetRecs(); ArticleN++) {
		if (IgnoreArticleSet.IsKey(ArticleSet->GetRecId(ArticleN))) { continue; }
		PJsonVal ArticleJson = Article(ArticleSet->GetRec(ArticleN), FullBodyP);
		ArticleJsonV.Add(ArticleJson);
	}
	return TJsonVal::NewArr(ArticleJsonV);
}

PJsonVal TNewsSrvUtil::FolderArticles(const PRecSet& ArticleSet, const int& Page, const int& Count) const 
{		
	// filter articles, by keeping 100 most recent
	//ArticleSet->SortByField(NewsBase->Base, false, NewsBase->ArticleStore->TimeFieldId);
	PJsonVal ResJson = TJsonVal::NewObj();

	ResJson->AddToObj("allArticlesCount",  TJsonVal::NewStr(TUInt64::GetStr(NewsBase->ArticleStore->GetRecs())));		// all article count
	ResJson->AddToObj("articleCount", TJsonVal::NewStr(TInt::GetStr(ArticleSet->GetRecs())));							// articles in the folder
	ResJson->AddToObj("goPrevious", TJsonVal::NewBool(Page > 0));								// can we go back in the list of articles
	ResJson->AddToObj("goNext", TJsonVal::NewBool(Page*Count + Count < ArticleSet->GetRecs()));		// do we have more articles to show
		
	//ArticleSet->SortById(false);
	PRecSet ArticleSetSmall = ArticleSet->GetLimit(Count, Page*Count);		// skip the first Offset articles and include only Count articles
		
	// store articles
	ResJson->AddToObj("articles", Articles(ArticleSetSmall, false));
	return ResJson;
}

///////////////////////////////////////////////
// PlusMachines-Server-Article
TNewsSrvArticle::TNewsSrvArticle(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil):
	  TSAppSrvFun("article", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) { 

}

TStr TNewsSrvArticle::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv) {
	// parse out the article
	uint64 ArticleId = TUInt64::Mx;
	if (IsFldNm(FldNmValPrV, "id")) {
		if (!GetFldVal(FldNmValPrV, "id").IsUInt64()) { return "article id not a number"; }
		ArticleId = GetFldVal(FldNmValPrV, "id").GetUInt64();
		if (!NewsBase->ArticleStore->IsRecId(ArticleId)) { return "bad article id"; }
	} else {
		return "no article id provided";
	}
	// get the article record
	TRec ArticleRec(NewsBase->ArticleStoreId, ArticleId);
	// get its concepts
	PRecSet ConceptSet = ArticleRec.DoJoin(NewsBase->Base, "hasConcept");
	ConceptSet->SortByFq(false);
	// get related articles
	PRecSet RelatedArticleSet = ConceptSet->DoJoin(NewsBase->Base, "hasArticle");
	RelatedArticleSet->SortByFq(false); RelatedArticleSet->Trunc(5);
	TUInt64Set IgnoreArticleSet = TUInt64Set::GetSet(ArticleId);

	// store article
	PJsonVal ArticleJson = NewsSrvUtil->Article(ArticleRec, true);
	
	// store related articles
	ArticleJson->AddToObj("related_articles", NewsSrvUtil->Articles(RelatedArticleSet, false, IgnoreArticleSet));
	
	// done
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("article", ArticleJson);
	return TJsonVal::GetStrFromVal(ResJson);
}

///////////////////////////////////////////////
// PlusMachines-Server-Folder
TStr TActivitySrv::GetStr(const TStrKdV& FldNmValPrV, const TStr& FldNm)
{
	EAssertR(IsFldNm(FldNmValPrV, FldNm), "no " + FldNm);
	return GetFldVal(FldNmValPrV, FldNm);
}

TStr TActivitySrv::GetStrIfIn(const TStrKdV& FldNmValPrV, const TStr& FldNm, const TStr& DefVal)
{
	if (!IsFldNm(FldNmValPrV, FldNm)) return DefVal;
	return GetFldVal(FldNmValPrV, FldNm);
}

int TActivitySrv::GetInt(const TStrKdV& FldNmValPrV, const TStr& FldNm)
{
	EAssertR(IsFldNm(FldNmValPrV, FldNm), "no " + FldNm);
	EAssertR(GetFldVal(FldNmValPrV, FldNm).IsInt(), FldNm + " not a number");
	return GetFldVal(FldNmValPrV, FldNm).GetInt();
}

bool TActivitySrv::GetBoolIfIn(const TStrKdV& FldNmValPrV, const TStr& FldNm, const bool& DefVal)
{
	if (!IsFldNm(FldNmValPrV, FldNm)) return DefVal;
	TStr Val = GetFldVal(FldNmValPrV, FldNm, "").GetLc();
	if (Val == "1" || Val == "true") return true;
	return false;
}

int TActivitySrv::GetIntIfIn(const TStrKdV& FldNmValPrV, const TStr& FldNm, const int& DefVal)
{
	if (!IsFldNm(FldNmValPrV, FldNm)) return DefVal;
	if (!GetFldVal(FldNmValPrV, FldNm).IsInt()) return DefVal;
	return GetFldVal(FldNmValPrV, FldNm).GetInt();
}

uint64 TActivitySrv::GetUInt64(const TStrKdV& FldNmValPrV, const TStr& FldNm)
{
	EAssertR(IsFldNm(FldNmValPrV, FldNm), "no " + FldNm);
	EAssertR(GetFldVal(FldNmValPrV, FldNm).IsUInt64(), FldNm + " not a number");
	return GetFldVal(FldNmValPrV, FldNm).GetUInt64();
}

uint64 TActivitySrv::GetUInt64IfIn(const TStrKdV& FldNmValPrV, const TStr& FldNm, const uint64& DefVal)
{
	if (!IsFldNm(FldNmValPrV, FldNm)) return DefVal;
	if (!GetFldVal(FldNmValPrV, FldNm).IsUInt64()) return DefVal;
	return GetFldVal(FldNmValPrV, FldNm).GetUInt64();
}

FolderVisibility TActivitySrv::GetVisibilityIfIn(const TStrKdV& FldNmValPrV, const FolderVisibility& DfVal)
{
	TStr VisibilityStr = GetStrIfIn(FldNmValPrV, "visibility", "");
	if (VisibilityStr == "")
		return DfVal;
	return (FolderVisibility) VisibilityStr.GetInt(DfVal);
}


TStr TActivitySrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try
	{
		// folder functions
		if (GetFunNm() == "articlesForFolder")
			ResJson = ArticlesForFolder(FldNmValPrV);
		else if (GetFunNm() == "storeArticleToFolder")
			ResJson = StoreArticleToFolder(FldNmValPrV);
		else if (GetFunNm() == "removeArticleFromFolder")
			ResJson = RemoveArticleFromFolder(FldNmValPrV);
		else if (GetFunNm() == "getStoredArticleUris")
			ResJson = GetStoredArticleUris(FldNmValPrV);
		else if (GetFunNm() == "getStoredArticles")
			ResJson = GetStoredArticles(FldNmValPrV);

		else if (GetFunNm() == "saveLikedArticle")
			ResJson = SaveLikedArticle(FldNmValPrV);
		else if (GetFunNm() == "unlikeArticle")
			ResJson = UnlikeArticle(FldNmValPrV);
		else if (GetFunNm() == "getLikedArticleUris")
			ResJson = GetLikedArticleUris(FldNmValPrV);

		else if (GetFunNm() == "addFolder")
			ResJson = AddFolder(FldNmValPrV);
		else if (GetFunNm() == "modifyFolder")
			ResJson = ModifyFolder(FldNmValPrV);
		else if (GetFunNm() == "removeFolder")
			ResJson = RemoveFolder(FldNmValPrV);

		else if (GetFunNm() == "removeUser")
			ResJson = RemoveUser(FldNmValPrV);
		
		else if (GetFunNm() == "saveActivityDataToFile")
			ResJson = SaveActivityDataToFile(FldNmValPrV);
		else if (GetFunNm() == "loadActivityDataFromFile")
			ResJson = LoadActivityDataFromFile(FldNmValPrV);
		else if (GetFunNm() == "getStatsActivity")
			ResJson = GetStats(FldNmValPrV);
		else if (GetFunNm() == "printDataActivity")
			ResJson = PrintData(FldNmValPrV);
		else
		{
			ResJson = TJsonVal::NewObj();
			ResJson->AddToObj("error", TJsonVal::NewStr("invalid function name: " + GetFunNm()));
			TQm::TEnv::Logger->OnStatus("invalid function name: " + GetFunNm()); 
		}

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E)
	{
		ResJson = TJsonVal::NewObj();
		ResJson->AddToObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TQm::TEnv::Logger->OnStatus("Error: " + E->GetMsgStr()); 
	}
	return TJsonVal::GetStrFromVal(ResJson);
}


//////////////////////////////////////
// folder functions

// return curated articles for folder
PJsonVal TActivitySrv::ArticlesForFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "");
	PJsonVal SearchParamsJson = TJsonVal::GetValFromStr(GetFldVal(FldNmValPrV, "searchParams", "{}"));
	const int Page = SearchParamsJson->GetObjNum("page", 0);
	const int Count = SearchParamsJson->GetObjNum("count", DefaultNumberOfReturnedArticlesPerPage);
	
	PRecSet ArticlesRecSet;
	// if no folder id is specified then return all stored articles for the user
	if (FolderId == "") {
		PRecSet FolderRecSet = ActivityBase->GetFoldersForUser(UserSqlId);
		TUInt64V ArticleIdV;
		for (int F=0; F < FolderRecSet->GetRecs(); F++) {
			const uint64 FolderRecId = FolderRecSet->GetRecId(F);
			PRecSet ArticlesInFolder = ActivityBase->GetStoredArticles(FolderRecId);
			TUInt64V IdV; ArticlesInFolder->GetRecIdV(IdV);
			ArticleIdV.AddV(IdV);
		}
		ArticleIdV.Sort(false);
		ArticlesRecSet = TRecSet::New(ActivityBase->ArticleStoreId, ArticleIdV);
	}
	// if id is specified return only articles in that folder
	else {
		if (!ActivityBase->IsFolder(UserSqlId, FolderId)) {
			TQm::TEnv::Logger->OnNotify(ntErr, TStr::Fmt("ArticlesForFolder: Folder id %s for user %s does not exist.", FolderId.CStr(), UserSqlId.CStr()));
			return NULL;
		}
		const uint64 FolderRecId = (uint64) ActivityBase->GetFolderRecId(UserSqlId, FolderId);
		ArticlesRecSet = ActivityBase->GetStoredArticles(FolderRecId);
	}

	ArticlesRecSet->SortById(false);

	PJsonVal ResJson = TJsonVal::NewObj();
	// add articles
	ResJson->AddToObj("results", NewsSrvUtil->FolderArticles(ArticlesRecSet, Page, Count));
	// add search params to the result
	ResJson->AddToObj("searchParams", SearchParamsJson);

	return ResJson;
}

// add article to folder
PJsonVal TActivitySrv::StoreArticleToFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "");
	
	if (FolderId == "") {
		TQm::TEnv::Logger->OnNotify(ntErr, "SaveArticleToFolder: Unable to save article to folder since no folder id was provided.");
		return NULL;
	}

	const uint64 FolderRecId = (uint64) ActivityBase->AddFolder(UserSqlId, FolderId);

	TStr ArticleJsonStr = GetFldVal(FldNmValPrV, "article", "{}");
	PJsonVal ArticleJson = TJsonVal::GetValFromStr(ArticleJsonStr);
	uint64 ArticleRecId = NewsBase->GenericBase->AddRec(ActivityBase->ArticleStoreId, ArticleJson);
	ActivityBase->AddStoredArticle(FolderRecId, ArticleRecId);
	return NULL;
}

PJsonVal TActivitySrv::GetStoredArticleUris(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "");
	const int ArticleCount = GetIntIfIn(FldNmValPrV, "count", 100);

	if (FolderId == "") {
		TQm::TEnv::Logger->OnNotify(ntErr, "SaveArticleToFolder: Unable to save article to folder since no folder id was provided.");
		return NULL;
	}
	
	const uint64 FolderRecId = (uint64) ActivityBase->GetFolderRecId(UserSqlId, FolderId);
	PJsonVal ResJson = TJsonVal::NewArr();
	if (FolderRecId == TUInt64::Mx) 
		return ResJson;
	
	PRecSet ArticleSet = ActivityBase->GetStoredArticles(FolderRecId);
	ArticleSet->SortById(false);
	int AllCount = ArticleSet->GetRecs();
	for (int N = 0; N < AllCount && N < ArticleCount; N++)
		ResJson->AddToArr(TJsonVal::NewStr(NewsBase->ArticleStore->GetRecNm(ArticleSet->GetRecId(N))));
	return ResJson;
}

// get more information about each stored article than using the GetStoredArticleUris
PJsonVal TActivitySrv::GetStoredArticles(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "");
	const int ArticleCount = GetIntIfIn(FldNmValPrV, "count", 100);

	if (FolderId == "") {
		TQm::TEnv::Logger->OnNotify(ntErr, "SaveArticleToFolder: Unable to save article to folder since no folder id was provided.");
		return NULL;
	}
	
	const uint64 FolderRecId = (uint64) ActivityBase->GetFolderRecId(UserSqlId, FolderId);
	PJsonVal ResJson = TJsonVal::NewArr();
	if (FolderRecId == TUInt64::Mx) 
		return ResJson;
	
	PRecSet ArticleSet = ActivityBase->GetStoredArticles(FolderRecId);
	ArticleSet->SortById(false);
	int AllCount = ArticleSet->GetRecs();
	ArticleSet->Trunc(ArticleCount);
	ResJson = NewsSrvUtil->Articles(ArticleSet, false);
	return ResJson;
}


// save the article as liked
PJsonVal TActivitySrv::SaveLikedArticle(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "");
	
	if (FolderId == "") {
		TQm::TEnv::Logger->OnNotify(ntErr, "SaveLikedArticle: Unable to save liked article since no folder id was provided.");
		return NULL;
	}
	const uint64 FolderRecId = (uint64) ActivityBase->AddFolder(UserSqlId, FolderId);

	const TStr ArticleJsonStr = GetFldVal(FldNmValPrV, "article", "{}");
	PJsonVal ArticleJson = TJsonVal::GetValFromStr(ArticleJsonStr);
	const uint64 ArticleRecId = ActivityBase->NewsBase->GenericBase->AddRec(ActivityBase->ArticleStoreId, ArticleJson);

	// is the article currently liked or was it liked in the past?
	const bool IsLikedInPast = ActivityBase->IsLikedArticle(FolderRecId, ArticleRecId) || ActivityBase->IsLikedArticleInPast(FolderRecId, ArticleRecId);
	ActivityBase->AddLikedArticle(FolderRecId, ArticleRecId);
	
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("isLikedInPast", IsLikedInPast);
	return ResJson;
}

// remove the article from the liked list for the user
PJsonVal TActivitySrv::UnlikeArticle(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "0");
	const TStr ArticleUri = GetStrIfIn(FldNmValPrV, "articleUri", "");
		
	if (!ActivityBase->IsFolder(UserSqlId, FolderId)) 
		return NULL;
	const uint64 FolderRecId = (uint64) ActivityBase->GetFolderRecId(UserSqlId, FolderId);

	if (!ActivityBase->NewsBase->ArticleStore->IsRecNm(ArticleUri))
		return NULL;
	uint64 ArticleRecId = ActivityBase->NewsBase->ArticleStore->GetRecId(ArticleUri);
	
	ActivityBase->RemoveLikedArticle(FolderRecId, ArticleRecId);
	return NULL;
}

// get last x liked articles
PJsonVal TActivitySrv::GetLikedArticleUris(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "0");
	const int ArticleCount = GetIntIfIn(FldNmValPrV, "count", 100);
	
	const uint64 FolderRecId = (uint64) ActivityBase->GetFolderRecId(UserSqlId, FolderId);
	PJsonVal ResJson = TJsonVal::NewArr();
	if (FolderRecId == TUInt64::Mx) 
		return ResJson;
	
	PRecSet ArticleSet = ActivityBase->GetLikedArticles(FolderRecId);
	ArticleSet->SortById(false);
	int AllCount = ArticleSet->GetRecs();
	for (int N = 0; N < AllCount && N < ArticleCount; N++)
		ResJson->AddToArr(TJsonVal::NewStr(NewsBase->ArticleStore->GetRecNm(ArticleSet->GetRecId(N))));
	return ResJson;
}

// remove article from the folder or remove the like
PJsonVal TActivitySrv::RemoveArticleFromFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "");
	const TStr ArticleUri = GetStrIfIn(FldNmValPrV, "articleUri", "");

	const uint64 FolderRecId = (uint64) ActivityBase->GetFolderRecId(UserSqlId, FolderId);
	if (FolderRecId == TUInt64::Mx) {
		TQm::TEnv::Logger->OnNotify(ntErr, TStr::Fmt("RemoveArticleFromFolder: Folder id %s for user %s does not exist.", FolderId.CStr(), UserSqlId.CStr()));
		return NULL;
	}
	
	if (!ActivityBase->NewsBase->ArticleStore->IsRecNm(ArticleUri))
		return NULL;

	const uint64 ArticleRecId = ActivityBase->NewsBase->ArticleStore->GetRecId(ArticleUri);
	// remove from folder (if it's in)
	ActivityBase->RemoveStoredArticle(FolderRecId, ArticleRecId);
	// should we remove the like also?
	//ActivityBase->RemoveLikedArticle(FolderRecId, ArticleRecId);
	
	return NULL;
}

// add a user's folder
PJsonVal TActivitySrv::AddFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	TStr FolderStr = GetFldVal(FldNmValPrV, "folderJson", "{}");
	PJsonVal FolderJson = TJsonVal::GetValFromStr(FolderStr);
	
	const TStr FolderId = FolderJson->GetObjStr("folderId", "");
	if (FolderId == "") {
		TQm::TEnv::Logger->OnNotify(ntInfo, TStr::Fmt("AddFolder. Invalid value for FolderId ('%s').", FolderId.CStr()));
		return NULL;
	}
	const uint64 FolderRecId = (uint64) ActivityBase->AddFolder(UserSqlId, FolderId);

	TStr Key = ActivityBase->CreateKey(UserSqlId, FolderId);
	PJsonVal ActivityFolderJson = TJsonVal::NewObj();
	ActivityFolderJson->AddToObj("URI", TJsonVal::NewStr(Key));
	ActivityFolderJson->AddToObj("Name", FolderJson->GetObjStr("folderName", ""));
	ActivityFolderJson->AddToObj("Visibility", (int) FolderJson->GetObjStr("visibility", "").GetInt(PrivateFolder));
	ActivityBase->GenericBase->AddRec(ActivityBase->FolderStoreId, ActivityFolderJson);
	
	return ActivityFolderJson;
}

// modify some properties of a folder
PJsonVal TActivitySrv::ModifyFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	TStr FolderStr = GetFldVal(FldNmValPrV, "folderJson", "{}");
	PJsonVal FolderJson = TJsonVal::GetValFromStr(FolderStr);

	const TStr FolderId = FolderJson->GetObjStr("folderId", "");
	if (FolderId == "") {
		TQm::TEnv::Logger->OnNotify(ntInfo, TStr::Fmt("ModifyFolder. Invalid value for FolderId ('%s').", FolderId.CStr()));
		return NULL;
	}
	const uint64 FolderRecId = (uint64) ActivityBase->AddFolder(UserSqlId, FolderId);

	TStr Key = ActivityBase->CreateKey(UserSqlId, FolderId);
	PJsonVal ActivityFolderJson = TJsonVal::GetValFromStr(FolderStr);
	ActivityFolderJson->AddToObj("URI", TJsonVal::NewStr(Key));
	ActivityFolderJson->AddToObj("Name", FolderJson->GetObjStr("folderName", ""));
	ActivityFolderJson->AddToObj("Visibility", (int) FolderJson->GetObjStr("visibility", "").GetInt(PrivateFolder));
	ActivityBase->GenericBase->AddRec(ActivityBase->FolderStoreId, ActivityFolderJson);
	
	return ActivityFolderJson;
}

// delete the folder. remove all the likes and articles from the folder
PJsonVal TActivitySrv::RemoveFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	const TStr FolderId = GetStrIfIn(FldNmValPrV, "folderId", "");

	const uint64 FolderRecId = (uint64) ActivityBase->GetFolderRecId(UserSqlId, FolderId);
	if (FolderRecId == TUInt64::Mx) {
		TQm::TEnv::Logger->OnNotify(ntErr, TStr::Fmt("RemoveFolder: Folder id %s for user %s does not exist.", FolderId.CStr(), UserSqlId.CStr()));
		return NULL;
	}
	
	ActivityBase->DeleteFolder(FolderRecId);
	return NULL;
}

// delete the folder. remove all the likes and articles from the folder
PJsonVal TActivitySrv::RemoveUser(const TStrKdV& FldNmValPrV)
{
	const TStr UserSqlId = GetStrIfIn(FldNmValPrV, "userSqlId", "");
	
	PRecSet FoldersRecSet = ActivityBase->GetFoldersForUser(UserSqlId);
	if (FoldersRecSet.Empty())
		return NULL;
	for (int N=0; N < FoldersRecSet->GetRecs(); N++) 
		ActivityBase->DeleteFolder(FoldersRecSet->GetRecId(N));

	return NULL;
}

PJsonVal TActivitySrv::SaveActivityDataToFile(const TStrKdV& FldNmValPrV)
{
	TTm Time = TTm::GetCurLocTm();
	TStr TimeStr = Time.GetYMDDashStr();
	ActivityBase->SaveData(TimeStr + "-activityUsers.txt", ActivityBase->UserStore);
	ActivityBase->SaveData(TimeStr + "-activityFolders.txt", ActivityBase->FolderStore);
	ActivityBase->SaveData(TimeStr + "-activityArticleActivity.txt", ActivityBase->ArtActStore);
	ActivityBase->SaveData(TimeStr + "-activityActivity.txt", ActivityBase->ActStore);
	//ActivityBase->SaveData(TimeStr + "-activityArticles.txt", NewsBase->ArticleStore);
	ActivityBase->SaveJoinsData();
	
	ActivityBase->UserStore->PrintAll(ActivityBase->Base, "printUsers.txt");
	ActivityBase->FolderStore->PrintAll(ActivityBase->Base, "printFolders.txt");
	ActivityBase->ArtActStore->PrintAll(ActivityBase->Base, "printArtAct.txt");
	ActivityBase->ActStore->PrintAll(ActivityBase->Base, "printAct.txt");

	return TJsonVal::NewObj("status", "Activity data was successfully saved into files");
}


PJsonVal TActivitySrv::LoadActivityDataFromFile(const TStrKdV& FldNmValPrV)
{
	ActivityBase->LoadData("activityUsers.txt", ActivityBase->UserStore);
	ActivityBase->LoadData("activityFolders.txt", ActivityBase->FolderStore);
	ActivityBase->LoadData("activityArticleActivity.txt", ActivityBase->ArtActStore);
	ActivityBase->LoadData("activityActivity.txt", ActivityBase->ActStore);

	ActivityBase->LoadJoinsData("hasLikedArticle.txt", ActivityBase->JoinFolderHasLikedArticlesId);
	ActivityBase->LoadJoinsData("hasLikedArticleInPast.txt", ActivityBase->JoinFolderHasLikedArticlesInPastId);
	ActivityBase->LoadJoinsData("hasStoredArticle.txt", ActivityBase->JoinFolderHasStoredArticlesId);

	uint64 cnt = NewsBase->ArticleStore->GetRecs();

	//ActivityBase->LoadData("activityArticles.txt", NewsBase->ArticleStore);
	return TJsonVal::NewObj("status", "Activity data was successfully loaded from files");
}

PJsonVal TActivitySrv::GetStats(const TStrKdV& FldNmValPrV)
{
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("users", TUInt64(ActivityBase->UserStore->GetRecs()).GetStr());
	ResJson->AddToObj("folders", TUInt64(ActivityBase->FolderStore->GetRecs()).GetStr());
	ResJson->AddToObj("articles", TUInt64(ActivityBase->NewsBase->ArticleStore->GetRecs()).GetStr());
	return ResJson;
}

PJsonVal TActivitySrv::PrintData(const TStrKdV& FldNmValPrV)
{
	PJsonVal ResJson = TJsonVal::NewObj();
	TStrV PrintV; GetFldValV(FldNmValPrV, "print", PrintV);
	if (PrintV.IsIn("userStore"))
		ActivityBase->UserStore->PrintAll(ActivityBase->Base, "PrintActivityUsers.txt");
	if (PrintV.IsIn("folderStore"))
		ActivityBase->FolderStore->PrintAll(ActivityBase->Base, "PrintActivityFolders.txt");
	if (PrintV.IsIn("actStore"))
		ActivityBase->ActStore->PrintAll(ActivityBase->Base, "PrintActivityActStore.txt");
	if (PrintV.IsIn("artActStore"))
		ActivityBase->ArtActStore->PrintAll(ActivityBase->Base, "PrintActivityArtActStore.txt");

	if (PrintV.IsIn("index"))
		ActivityBase->Base->PrintIndex("PrintActivityIndex.txt", true);
	if (PrintV.IsIn("indexVoc"))
		ActivityBase->Base->PrintIndexVoc("PrintActivityIndexVoc.txt");
	return ResJson;
}