/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ActivityBase.h"
#include "ActivitySrv.h"
#include "../NewsMiner/news.h"
#include "../Utils/FileUtils.h"

int main(int argc, char* argv[]){
    Try;
	TQm::TEnv::Init();
    // create environment
    Env = ::TEnv(argc, argv, TNotify::StdNotify);
    // command line parameters
    Env.PrepArgs("Plus-Machines-Server", 1);
	TStr NewsDefFNm = Env.GetIfArgPrefixStr("-newsDef:", "./settings/news.def", "News Schema Definition File");
	TStr ActivityNewsDefFNm = Env.GetIfArgPrefixStr("-activityDef:", "./settings/activity.def", "Activity Schema Definition File");
	const TStr IndexFPath = Env.GetIfArgPrefixStr("-folderIndex:", "./ActivityIndex/", "Index Folder");
	const TStr FolderIndexFPath = Env.GetIfArgPrefixStr("-folderActivity:", "./ActivityActivity/", "Folder for storing user activity");

	const int IndexCacheSizeMB = Env.GetIfArgPrefixInt("-indexcache:", 4000, "Cache Size (in MB)");
    const int ArticleCacheSizeMB = Env.GetIfArgPrefixInt("-articlecache:", 4000, "Cache Size (in MB)");
    const int ActivityCacheSizeMB = Env.GetIfArgPrefixInt("-activitycache:", 4000, "Cache Size (in MB)");
	TStr TempFPath = Env.GetIfArgPrefixStr("-temp:", "./temp/", "Temporary Folder");
	TStr StoryUrlStr = Env.GetIfArgPrefixStr("-addstory:", "", "Article-story web service");
    const int PortN = Env.GetIfArgPrefixInt("-port:", 8093, "server port");
	TStr SettingsFileStr = Env.GetIfArgPrefixStr("-settingsActivity:", "./settings/settingsActivity.json", "File with various settings");
	
	TStr SettingsStr = TFile::Exists(SettingsFileStr) ? TStr::LoadTxt(SettingsFileStr) : "{}";
	PJsonVal SettingsJson = TJsonVal::GetValFromStr(SettingsStr);

    if (Env.IsEndOfRun()){return 0;}

    // prepare unicode base
    if (!TUnicodeDef::IsDef()) { TUnicodeDef::Load("./UnicodeDef.Bin"); }
		
	// create directories if necessary
	TDir::GenDir(IndexFPath);
	TDir::GenDir(FolderIndexFPath);
	
    // calculate cache-size in bytes
    const int64 IndexCacheSize = int64(IndexCacheSizeMB) * int64(TInt::Mega);
    const int64 ArticleCacheSize = int64(ArticleCacheSizeMB) * int64(TInt::Mega);
	const int64 ActivityCacheSize = int64(ActivityCacheSizeMB) * int64(TInt::Mega);

	bool NewsBaseExists = TIndex::Exists(IndexFPath) && TFile::Exists(IndexFPath + "Source.BaseStore");
	PNewsBase NewsBase = NewsBaseExists ?
		TNewsBase::Load(IndexFPath, faUpdate, IndexCacheSize, ArticleCacheSize, TempFPath, false) :
		TNewsBase::New(NewsDefFNm, IndexFPath, IndexCacheSize, ArticleCacheSize, TempFPath, false);

	PActivityBase ActivityBase = TActivityBase::New(ActivityNewsDefFNm, NewsBase, FolderIndexFPath, IndexCacheSize, ActivityCacheSize);

	// load basic miner function
	TSAppSrvFunV SrvFunV;
	TSrvFun::RegDefFun(NewsBase->Base, SrvFunV);

	// load plus machines functions
	PNewsSrvUtil NewsSrvUtil = TNewsSrvUtil::New(NewsBase, ActivityBase);
	SrvFunV.Add(TNewsSrvArticle::New(NewsBase, NewsSrvUtil));
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "articlesForFolder"));			// get curated articles for a given folder
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "storeArticleToFolder"));		// add article to folder
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "removeArticleFromFolder"));		// remove the article from folder
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "getStoredArticleUris"));		// get uris for articles stored in the folder
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "getStoredArticles"));			// get full info about the stored article

	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "saveLikedArticle"));			// add article to folder
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "unlikeArticle"));				// unlike article
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "getLikedArticleUris"));			// get uris for liked articles

	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "addFolder"));					// add a folder
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "modifyFolder"));				// change properties of a folder
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "removeFolder"));				// remove the folder and links to stored articles
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "removeUser"));					// remove all users folders and links to stored articles
	
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "saveActivityDataToFile"));		// store all articles, users, folders, activities to files
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "loadActivityDataFromFile"));	// load all articles, users, folders, activities from files

	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "getStatsActivity"));					// get statistics
	SrvFunV.Add(TActivitySrv::New(NewsBase, ActivityBase, NewsSrvUtil, "printDataActivity"));					// print index

	// set the logging values
	for (int N=0; N< SrvFunV.Len(); N++) {
		SrvFunV[N]->LogRqToFile = SettingsJson->GetObjBool("LogReqRes", false);
		SrvFunV[N]->LogRqFolder = SettingsJson->GetObjStr("LogFolder", ".\\Logs\\Activity");
	}
	
	// prepare the web-server logger
	PNotify WebNotify = TNotify::StdNotify;
	// start server
	PWebSrv WebSrv = TSAppSrv::New(PortN, SrvFunV, WebNotify, true);
	TLoop::Run();
	
	return 0;
    Catch;
    return 1;
}
