/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "CategoryReq.h"
#include "EventRegBase.h"

TCategoryReq::TCategoryReq(const TStr& _CategorySrvUrl, const TStr& _ReqSaveFName) :
	CategorySrvUrl(_CategorySrvUrl), ReqSaveFName(_ReqSaveFName)
{
	CompletedCount = 0;
	ErrorCount = 0;
	Load();
}

TCategoryReq::~TCategoryReq()
{
	if (ReqSaveFName != "")
		Save();
}

void TCategoryReq::Load() 
{
	if (TFile::Exists(ReqSaveFName)) {
		TFIn FIn(ReqSaveFName);
		ReqStoryIdQ.Load(FIn);
		ReqStoryIdH.Load(FIn);
	}
}

void TCategoryReq::Save()
{
	TFOut FOut(ReqSaveFName);
	ReqStoryIdQ.Save(FOut);
	ReqStoryIdH.Defrag();
	ReqStoryIdH.Save(FOut);
}

void TCategoryReq::OnError(const int& FId, const TStr& MsgStr) {
	ErrorCount++;
	// put to file only first 1000 errors
	if (ErrorCount < 1000)
		EventRegBase->NewsBase->ErrorLogFile->NotifyErr("TCategoryReq::OnError. Error: %s", MsgStr.CStr());
	else
		TNotify::StdNotify->OnStatus("CategoryReq error: " + MsgStr);
	//Report();
	// repeat the last request
	//SendNextReq();
}

// when we receive the category info from the category server call the callback
void TCategoryReq::OnFetch(const int& FId, const PWebPg& WebPg) { 
	CompletedCount++;
	Report();

	TStr Data = WebPg->GetHttpBodyAsStr();
	PJsonVal JsonData = TJsonVal::GetValFromStr(Data);
	// call the callback
	if (EventRegBase != NULL) {
		EventRegBase->UpdateStoryCategoriesCallback(JsonData);
	}
	
	// since we succeeded we remove the latest req from the queue
	PopReq();
	// send the next request
	SendNextReq();
}

void TCategoryReq::PopReq()
{
	if (ReqStoryIdQ.Len() == 0)
		return;
	const uint64 Key = ReqStoryIdQ.Top();
	ReqStoryIdQ.Pop();
	ReqStoryIdH.DelIfKey(Key);
}

// send the next request if there are currently no elements already being fetched
void TCategoryReq::SendNextReq()
{
	// there is currently a request in progress. wait for it to finish, SendNextReq will be then called automatically
	if (ReqStoryIdQ.Len() == 0 || CategorySrvUrl == "" || GetConnUrls() > 3 || EventRegBase == NULL)
		return;

	const uint64 StoryId = ReqStoryIdQ.Top();
	if (!EventRegBase->StoryStore->IsRecId(StoryId))  {
		EventRegBase->NewsBase->ErrorLogFile->NotifyErr("TCategoryReq::SendNextReq. Queued story id was not valid story id");
		// we need to pop this request, otherwise we would be in a cycle
		PopReq();
		SendNextReq();
		return;
	}
	PRecSet ArticleRecSet = EventRegBase->StoryStore->GetRec(StoryId).DoJoin(EventRegBase->Base, "hasArticle");
	ArticleRecSet->SortByField(false, EventRegBase->ArticleStore->GetFieldId("CosToCentroid"));

	TStr MergedContent = "";
	for (int ArtN = 0; ArtN < 5 && ArtN < ArticleRecSet->GetRecs(); ArtN++) {
		MergedContent += EventRegBase->ArticleStore->GetFieldNmStr(ArticleRecSet->GetRecId(ArtN), "Title") + " ";
		MergedContent += EventRegBase->ArticleStore->GetFieldNmStr(ArticleRecSet->GetRecId(ArtN), "Body") + " ";
	}
	
	if (MergedContent == "")  {
		// empty content. pop this request, otherwise we get the problem that we didn't specify the content-length
		PopReq();
		SendNextReq();
		return;
	}

	TMem BodyMem; BodyMem += "action=getCategory&key=" + TUrl::EncodeUrlStr(EventRegBase->StoryStore->GetRecNm(StoryId)) + "&maxCat=3&text=" + TUrl::EncodeUrlStr(MergedContent);
	PHttpRq HttpRq = THttpRq::New(hrmPost, TUrl::New(CategorySrvUrl), THttp::AppW3FormFldVal, BodyMem);
	/*HttpRq->AddFldVal("action", "getCategory");
	HttpRq->AddFldVal("key", EventRegBase->StoryStore->GetRecNm(StoryId));
	HttpRq->AddFldVal("maxCat", "3");
	HttpRq->AddFldVal("text", MergedContent);*/
	FetchHttpRq(HttpRq);
	//TStr ReqUrl = CategorySrvUrl + "?action=getCategory&key=" + TUrl::EncodeUrlStr(EventRegBase->StoryStore->GetRecNm(StoryId)) + "&maxCat=3&text=" + TUrl::EncodeUrlStr(MergedContent);
	//FetchUrl(ReqUrl);
}

void TCategoryReq::RequestStoryCategory(const uint64& StoryId)
{
	// add story to the queue only if it's not already there
	if (!ReqStoryIdH.IsKey(StoryId)) {
		ReqStoryIdQ.Push(StoryId);
		ReqStoryIdH.AddKey(StoryId);
	}
	SendNextReq();
}