/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "EventRegBase.h"
#include "../Utils/JsonUtils.h"
#include "../Utils/FileUtils.h"

#define EPS 1e-4

TEventSettings::TEventSettings()
{
	Init(TJsonVal::NewObj());
}

TEventSettings::TEventSettings(const PJsonVal& Settings)
{
	Init(Settings);
}

void TEventSettings::Init(const PJsonVal& Settings)
{
	RecomputeStoryFeaturesEvery = (int) Settings->GetObjNum("RecomputeStoryFeaturesEvery", 10);
	
	StoryMergingMinMatchArts = (int) Settings->GetObjNum("StoryMergingMinMatchArts", 5);
	StoryMergingSlope = Settings->GetObjNum("StoryMergingSlope", 0.2);
	StoryMergingMinAvgSim = (int) Settings->GetObjNum("StoryMergingMinAvgSim", 50);
	StoryMergingMinBowSim = (int) Settings->GetObjNum("StoryMergingMinBowSim", 50);
	StoryMergingMinAnnArtsToUse = (int) Settings->GetObjNum("StoryMergingMinAnnArtsToUse", 0);
	StoryMergingMaxStoryDiffHrs = (int) Settings->GetObjNum("StoryMergingMaxStoryDiffHrs", 48);
	StoryMergingMinArticleCosSim = Settings->GetObjNum("StoryMergingMinArticleCosSim", 0.3);

	StoryPropTopConceptsToTake = (int) Settings->GetObjNum("StoryPropTopConceptsToTake", 30);
	StoryPropMinConceptWgtToTake = (int) Settings->GetObjNum("StoryPropMinConceptWgtToTake", 5);
	StoryPropMinCategoryWgtToTake = (int) Settings->GetObjNum("StoryPropMinCategoryWgtToTake", 30);
	StoryPropMinLocationWgtToTake = (int) Settings->GetObjNum("StoryPropMinLocationWgtToTake", 40);
	StoryPropMinLocationPrcToTake = (int) Settings->GetObjNum("StoryPropMinLocationPrcToTake", 30);
	StoryPropMinLocationPropToTake = (int) Settings->GetObjNum("StoryPropMinLocationPropToTake", 50);

	EventPropTopConceptsToTake = (int) Settings->GetObjNum("EventPropTopConceptsToTake", 30);
	EventPropMinConceptWgtToTake = (int) Settings->GetObjNum("EventPropMinConceptWgtToTake", 5);

	SimilarEventsNrTopConceptsUsed = (int) Settings->GetObjNum("SimilarEventsNrTopConceptsUsed", 10);
	SimilarEventsNrTopEventsPerConcept = (int) Settings->GetObjNum("SimilarEventsNrTopEventsPerConcept", 1000);
	SimilarEventsMaxEventsToEval = (int) Settings->GetObjNum("SimilarEventsMaxEventsToEval", 5000);

	SimilarStoriesNrTopConceptsUsed = (int) Settings->GetObjNum("SimilarStoriesNrTopConceptsUsed", 10);
	SimilarStoriesNrTopStoriesPerConcept = (int) Settings->GetObjNum("SimilarStoriesNrTopStoriesPerConcept", 1000);
	SimilarStoriesMaxStoriesToEval = (int) Settings->GetObjNum("SimilarStoriesMaxStoriesToEval", 5000);

	UseModelForStoryMerging = Settings->GetObjBool("UseModelForStoryMerging", false);		// use the classification model or the fixed decision rule when deciding to merge stories
	UseModelForStoryLocation = Settings->GetObjBool("UseModelForStoryLocation", false);		// use the classification model or the fixed decision rule when deciding the story location

	ComputeCategories = Settings->GetObjBool("ComputeCategories", true);

	DateMaxPosInTextToUse = (int) Settings->GetObjNum("DateMaxPosInTextToUse", 100);
	DateMinStoryDateFreqCnt = (int) Settings->GetObjNum("DateMinStoryDateFreqCnt", 2);
	DateMinStoryDateFreqSlope = Settings->GetObjNum("DateMinStoryDateFreqSlope", 0.2);

	DateImputationMinValidYearCnt = (int) Settings->GetObjNum("DateImputationMinValidYearCnt", 2);
	DateImputationMinValidYearSlope = Settings->GetObjNum("DateImputationMinValidYearSlope", 0.1);

	EventSampleSize = (int) Settings->GetObjNum("EventSampleSize", 2000);

	VerbosityLevel = (int) Settings->GetObjNum("VerbosityLevel", 3);
}

//////////////////////////////////////////////////////////////////////
/// create or load existing event base
TEventRegBase::TEventRegBase(const PNewsBase& _NewsBase, const TFAccess& _FAccess, const PNewsSearch& _NewsSearch, const PCategoryReq& _CategoryReq, const TStr& _DbsPath, const TStr& _SettingsFName) :
	NewsBase(_NewsBase), FAccess(_FAccess), NewsSearch(_NewsSearch), CategoryReq(_CategoryReq), DbsPath(_DbsPath), SettingsFName(_SettingsFName)
{
	InfoLogFile = NewsBase->GetInfoLogger();
	ActivityLogFile = NewsBase->GetActivityLogger();
	ErrorLogFile = NewsBase->GetErrorLogger();

	Base = NewsBase->Base;
	Index = NewsBase->Index;
	IndexVoc = NewsBase->IndexVoc;
			
	ArticleStore = NewsBase->ArticleStore;
	ConceptStore = NewsBase->ConceptStore;
	CategoryStore = NewsBase->CategoryStore;
	StoryStore = NewsBase->StoryStore;
	SourceStore = NewsBase->SourceStore;
	EventStore = NewsBase->EventStore;
	ImageStore = NewsBase->ImageStore;

	ArticleStoreId = NewsBase->ArticleStoreId;
	ConceptStoreId = NewsBase->ConceptStoreId;;
	CategoryStoreId = NewsBase->CategoryStoreId;
	SourceStoreId = NewsBase->SourceStoreId;
	StoryStoreId = NewsBase->StoryStoreId;
	EventStoreId = NewsBase->EventStoreId;
	ImageStoreId = NewsBase->ImageStoreId;

	Base = NewsBase->Base;
		
	StoryLinking = TLrnStoryLinking::New(NewsBase, DbsPath);
	StoryLocation = TLrnStoryLocation::New(NewsBase, DbsPath);

	ReloadSettings();

	Load();
}

TEventRegBase::~TEventRegBase() 
{ 
	if (!CategoryReq.Empty())
		CategoryReq->SetEventRegBase(NULL);
	// save if necessary
	if (FAccess != faRdOnly) {
		Save();
	} else {
		TNotify::StdNotify->OnStatus("No saving neccessary!\n");
	}
}

void TEventRegBase::Load()
{
	/*TStr FName = Base->GetFPath() + "EventRegBase.bin";
	if (TFile::Exists(FName)) {
		TFIn FIn(FName);
	}*/
}

void TEventRegBase::Save() 
{ 
	/*TStr FName = Base->GetFPath() + "EventRegBase.bin";
	TFOut FOut(FName);
	*/
}

// when an articles was moved from one cluster to another we have to clean some joins and 
// potentially remove empty event (event with no articles)
void TEventRegBase::OnRemoveArticleFromStory(const uint64& OldStoryId)
{
	EAssertR(StoryStore->IsRecId(OldStoryId), TStr::Fmt("Story id %I64u is invalid", OldStoryId));

	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::OnRemoveArticleFromStory. It seems we removed some articles from story '%I64u'.", OldStoryId);
	InfoLogFile->IncreaseSpaces();
	// joins from article <-> old story are deleted already in news miner
	// here we just need to update event info
	
	// decrease article count for the story
	int NewArticleCount = UpdateStoryArticleCount(OldStoryId);

	uint64 OldEventId = GetStoryEvent(OldStoryId);
	UpdateEventArticleCount(OldEventId);

	// if the event doesn't have any articles left mark it as invalid
	if (NewArticleCount == 0) {
		// clean the previously extracted information about the event
		UpdateStoryProperties(OldStoryId, false);		
		
		// put the story back to it's original event
		const uint64 EventRecId = GetStoryPrimaryEvent(OldStoryId);
		InfoLogFile->NotifyVerboseFmt(2, "Story '%I64u' has no more articles assigned to it. It is being assigned back to the primary event '%I64u'", OldStoryId, EventRecId);

		// assign story to original event and update props of the old and new event
		AssignStoryToEvent(OldStoryId, EventRecId);	
		
		// we need to update event information for old and new event
		UpdateEventProperties(OldEventId, false);
		if (OldEventId != EventRecId)
			UpdateEventProperties(EventRecId);
	}
	
	InfoLogFile->DecreaseSpaces();
}

void TEventRegBase::ConsiderUpdatingStoryInfo(const uint64& StoryId, const bool& ForceUpdate)
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	UpdateStoryArticleCount(StoryId);
	int ChangeCount = StoryStore->GetFieldNmInt(StoryId, "ChangeCount");
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::ConsiderUpdatingStoryInfo. Updating story info for story '%I64u', changeCount = %d, force update = %d", StoryId, ChangeCount, ForceUpdate ? 1 : 0);
	InfoLogFile->IncreaseSpaces();
	
	uint64 EventId = GetStoryEvent(StoryId);
	UpdateEventArticleCount(EventId);

	// extract from the articles in story detected dates
	// find common dates and associate them with the story
	if (ForceUpdate || (ChangeCount % Settings.RecomputeStoryFeaturesEvery == 0)) {
		ActivityLogFile->NotifyInfo("Updating properties of story %I64u ('%s')", StoryId, StoryStore->GetRecNm(StoryId).CStr());
		InfoLogFile->NotifyVerboseFmt(4, "Story title: '%s'", GetStoryTitle(StoryId).CStr());
		InfoLogFile->NotifyVerboseFmt(4, "Story body: '%s'", GetStoryBody(StoryId).CStr());
		InfoLogFile->NotifyVerboseFmt(4, "Article count: %d", GetStoryArticleCount(StoryId));
		
		UpdateStoryProperties(StoryId, false);	// no need to update article count here since we've already called UpdateStoryArticleCount
		
		// find stories that should be merged
		UpdateEventStories(StoryId);
		// update concept information for the event
		UpdateEventProperties(EventId);
	}
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerbose(3, "TEventRegBase::ConsiderUpdatingStoryInfo. Finished.");
}

void TEventRegBase::UpdateStoryProperties(const uint64& StoryId, const bool& ArticleCountP, const bool& TitleSummaryP, const bool& DatesP, const bool& ConceptsLocationP, const bool& CategoryP, const bool& QualityP, const bool& LastUpdateP)
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	if (ArticleCountP)
		UpdateStoryArticleCount(StoryId);
	if (TitleSummaryP)
		UpdateStoryTitleAndSummary(StoryId);
	if (DatesP)
		UpdateStoryDates(StoryId);
	if (ConceptsLocationP) {
		UpdateStoryConcepts(StoryId);
		UpdateStoryLocation(StoryId);
	}
	if (CategoryP)
		UpdateStoryCategories(StoryId);
	if (QualityP)
		UpdateStoryQuality(StoryId);
	if (LastUpdateP)
		UpdateStoryLastUpdate(StoryId);
}

void TEventRegBase::UpdateEventProperties(const uint64& EventId, const bool& ArticleCountP, const bool& ConceptsP, const bool& LocationP, const bool& DateP, const bool& CategoriesP)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	if (ArticleCountP)
		UpdateEventArticleCount(EventId);
	if (DateP)
		UpdateEventDate(EventId);
	if (ConceptsP)
		UpdateEventConcepts(EventId);
	if (LocationP)
		UpdateEventLocation(EventId);
	if (CategoriesP)
		UpdateEventCategories(EventId);
}

// ////////////////
// STORY UPDATES

// update story date
void TEventRegBase::UpdateStoryDates(const uint64& StoryId)
{
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryDates. Updating story dates for story '%I64u'", StoryId);
	InfoLogFile->IncreaseSpaces();
	
	// go over all articles in story and extract from them dates and compute average date
	PRecSet ArticleRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
	const int TotArticleCount = ArticleRecSet->GetRecs();
	
	TStr ExtractedDate, ExtractedDateEnd, AverageDate, CommonDates;
	
	TExtractedDateColl CombinedDateColl; TTm AvgTm;
	bool Ok = ComputeStoryDates(StoryId, CombinedDateColl, AvgTm);
	if (Ok)
		AverageDate = AvgTm.GetYMDDashStr();	// YYYY-MM-DD

	for (int i=0; i < 5 && i < CombinedDateColl.Len(); i++)
		InfoLogFile->NotifyVerboseFmt(2, "\tDate range: %s, Freq: %d", CombinedDateColl.GetExtractedDate(i).GetStr().CStr(), CombinedDateColl.GetExtractedDate(i).GetFreq());
		
	const int MinDateFreq = (int) (Settings.DateMinStoryDateFreqCnt + (Settings.DateMinStoryDateFreqSlope * TotArticleCount));
	TExtractedDate Date;
	bool ExtractedDateP = CombinedDateColl.GetMostFreqDate(Date, MinDateFreq, true);
	if (ExtractedDateP) {
		ExtractedDate = Date.GetDate();
		if (Date.IsDateRange())
			ExtractedDateEnd = Date.GetDateEnd();
	}
	CommonDates = CombinedDateColl.EncodeExtractedDates(5);
	
	if (InfoLogFile->GetVerbosityLevel() >= 2)
		InfoLogFile->NotifyVerboseFmt(2, "Updating story dates for story '%I64u'. ExtractedDate = ('%s', '%s'). AverageDate = '%s'. Articles = %d", StoryId, ExtractedDate.CStr(), ExtractedDateEnd.CStr(), AverageDate.CStr(), ArticleRecSet->GetRecs());

	// in DateV we now have all the dates detected in articles
	PJsonVal StoryUpdateRec = TJsonVal::NewObj();
	StoryStore->SetFieldNmStr(StoryId, "ExtractedDate", ExtractedDate);
	StoryStore->SetFieldNmStr(StoryId, "ExtractedDateEnd", ExtractedDateEnd);
	StoryStore->SetFieldNmStr(StoryId, "AverageDate", AverageDate);
	
	StoryStore->SetFieldNmStr(StoryId, "CommonDates", CommonDates);
	// update the StoryDate only if we don't have custom values set
	if (NewsBase->GetStoryFlag(StoryId, sfUserSetDate) == false) {
		// the story date should be set to the ExtractedDate if valid, otherwise to AverageDate
		StoryStore->SetFieldNmStr(StoryId, "StoryDate", ExtractedDate != "" ? ExtractedDate : AverageDate);
		StoryStore->SetFieldNmStr(StoryId, "StoryDateEnd", ExtractedDateEnd);
	}
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerbose(3, "TEventRegBase::UpdateStoryDates. Finished.");
}

// update story concepts
// take all the articles associated with the story
// find the concepts in those articles
// find the most popular X concepts (separately for entites and keywords)
// and create joins between these concepts and story
void TEventRegBase::UpdateStoryConcepts(const uint64& StoryId)
{
	TTmStopWatch Sw; Sw.Start();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryConcepts. Updating story concepts and locations for story '%I64u'", StoryId);
	InfoLogFile->IncreaseSpaces();
	PRecSet ArticleRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");

	THash<TUInt64, TFlt> ConceptIdToWgtH;
	THash<TUInt64, TInt> ConceptIdToCntH;
	THash<TUInt64, TFlt> LocationIdToWgtH;
	int Articles = ArticleRecSet->GetRecs();
	for (int N=0; N<Articles; N++) {
		const uint64 ArticleRecId = ArticleRecSet->GetRecId(N);
		const double CosSim = ArticleStore->GetFieldNmFlt(ArticleRecId, "CosToCentroid");
		const PRecSet ConceptRecSet = ArticleRecSet->GetRec(N).DoJoin(Base, "hasConcept");
		for (int ConceptN = 0; ConceptN < ConceptRecSet->GetRecs(); ConceptN++) {
			const uint64 ConceptId = ConceptRecSet->GetRecId(ConceptN);
			const int ConceptRecFq = ConceptRecSet->GetRecFq(ConceptN);
			// debug:
			//TStr Uri = ConceptStore->GetRecNm(ConceptId);
			// CosSim: [0,1]
			// ConceptRecFq: [0,5]
			// => ConceptWgt : [5]
			/*if (ConceptRecFq > 5)
				WarnNotify(TStr::Fmt("Warning: Concept %s has weight %d. Should be [0,5]", Uri.CStr(), ConceptRecFq));*/

			TFlt ConceptWgt = (double) ConceptRecFq * CosSim;
			ConceptIdToWgtH.AddDat(ConceptId) += ConceptWgt;
			ConceptIdToCntH.AddDat(ConceptId) += 1;
		}
	}

	// ///////////
	// get separate top lists for entities and keywords
	ConceptIdToWgtH.SortByDat(false);
	double TopWgt = 1;
	TUInt64V ConceptsToAddV;
	int AddedEnts = 0, AddedKwds = 0;
	for (int KeyId = ConceptIdToWgtH.FFirstKeyId(); ConceptIdToWgtH.FNextKeyId(KeyId);) {
		const double Wgt = ConceptIdToWgtH[KeyId];
		TopWgt = MAX(TopWgt, Wgt);
		const uint64 ConceptRecId = ConceptIdToWgtH.GetKey(KeyId);
		const TStr Type = ConceptStore->GetFieldNmStr(ConceptRecId, "Type");

		if (NewsConst.IsEntity(Type) && AddedEnts < Settings.StoryPropTopConceptsToTake) {
			AddedEnts++; ConceptsToAddV.Add(ConceptIdToWgtH.GetKey(KeyId));
		}
		else if (NewsConst.IsKeywordOrWiki(Type) && AddedKwds < Settings.StoryPropTopConceptsToTake) {
			AddedKwds++; ConceptsToAddV.Add(ConceptIdToWgtH.GetKey(KeyId));	
		}
		if (AddedKwds >= Settings.StoryPropTopConceptsToTake && AddedEnts >= Settings.StoryPropTopConceptsToTake)
			break;
	}
	
	// ///////
	// updates to concepts
	// delete old joins for concepts
	StoryStore->DelJoins("hasConcept", StoryId);
	
	// add joins for top entitys + keywords
	if (InfoLogFile->GetVerbosityLevel() >= 4)
		InfoLogFile->NotifyVerbose(4, "Relevant concepts are:");
	for (int N=0; N < ConceptsToAddV.Len(); N++) {
		uint64 ConceptId = ConceptsToAddV[N];
		int Wgt = TFlt::Round(100.0 * ConceptIdToWgtH.GetDat(ConceptId) / TopWgt);
		if (Wgt < Settings.StoryPropMinConceptWgtToTake)
			continue;
		StoryStore->AddJoin("hasConcept", StoryId, ConceptId, Wgt);	// story -> concept  &  concept -> story
		if (Settings.VerbosityLevel >= 4)
			InfoLogFile->NotifyVerboseFmt(4, "\t%3d (%3d arts) %s (id: %I64u)", Wgt, ConceptIdToCntH.GetDat(ConceptId).Val, ConceptStore->GetRecNm(ConceptId).CStr(), ConceptId);
	}

	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryConcepts. Finished. Time needed %.0f ms", Sw.GetMSec());
}

//
// TODO: debug this whole part of code - made significant changes
void TEventRegBase::UpdateStoryLocation(const uint64& StoryId)
{
	// updates related to locations are done only if a custom location was not set
	bool CustomLocationSet = NewsBase->GetStoryFlag(StoryId, sfUserSetLocation);
	if (CustomLocationSet)
		return;

	TTmStopWatch Sw; Sw.Start();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryLocation. Updating story locations for story '%I64u'", StoryId);
	InfoLogFile->IncreaseSpaces();
	
	// remove existing joins between the story and locations
	RemoveLocationFromStory(StoryId);

	// use the model for story location prediction
	if (Settings.UseModelForStoryMerging) {
		uint64 LocationConceptId = TUInt64::Mx;
		double LocationProb = 0;
		const TStr StoryUri = StoryStore->GetRecNm(StoryId);
		bool HasLocation = StoryLocation->PredictStoryLocation(this, StoryUri, LocationConceptId, LocationProb);
		if (HasLocation)
			SetLocationForStory(StoryId, LocationConceptId, 100 * LocationProb);
	}
	// use the static set of rules for predicting the story location
	else {
		PRecSet ArticleRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");

		THash<TUInt64, TFlt> ConceptIdToWgtH;
		THash<TUInt64, TInt> ConceptIdToCntH;
		//THash<TUInt64, TFlt> ExplicitLocIdToWgtH;
		int Articles = ArticleRecSet->GetRecs();
		for (int N = 0; N < Articles; N++) {
			const uint64 ArticleRecId = ArticleRecSet->GetRecId(N);
			const double CosSim = ArticleStore->GetFieldNmFlt(ArticleRecId, "CosToCentroid");
			const PRecSet ConceptRecSet = ArticleRecSet->GetRec(N).DoJoin(Base, "hasConcept");
			for (int ConceptN = 0; ConceptN < ConceptRecSet->GetRecs(); ConceptN++) {
				const uint64 ConceptId = ConceptRecSet->GetRecId(ConceptN);
				const int ConceptRecFq = ConceptRecSet->GetRecFq(ConceptN);
				// we only consider a city as a valid location of an event. a country is too general
				if (! NewsBase->GeoBase->IsCity(ConceptStore->GetRecNm(ConceptId)))
					continue;
				ConceptIdToWgtH.AddDat(ConceptId) += CosSim;
				ConceptIdToCntH.AddDat(ConceptId) += 1;
			}
			const uint64 ExplicitLoc = ArticleStore->GetFieldNmUInt64(ArticleRecId, "hasLocationId");
			// the weights of the explicitly defined locations are 10x more important
			if (ExplicitLoc != TUInt64::Mx && NewsBase->GeoBase->IsCity(ConceptStore->GetRecNm(ExplicitLoc)))
				ConceptIdToWgtH.AddDat(ExplicitLoc) += 10 * CosSim;
		}
	
		// re-sort since we've added weights
		ConceptIdToWgtH.SortByDat(false);

		double SumWgts = 0;
		for (int KeyId = ConceptIdToWgtH.FFirstKeyId(); ConceptIdToWgtH.FNextKeyId(KeyId);) {
			SumWgts += ConceptIdToWgtH[KeyId];
		}

		if (ConceptIdToWgtH.Len() > 0) {
			int KeyId = ConceptIdToWgtH.FFirstKeyId(); ConceptIdToWgtH.FNextKeyId(KeyId);
			const uint64 ConceptId = ConceptIdToWgtH.GetKey(KeyId);
			const TStr ConceptNm = ConceptStore->GetRecNm(ConceptId);

			// note: wgt can be > 100. wgt also has the article locations weights in them
			int Wgt = (int) (100 * ConceptIdToWgtH.GetDat(ConceptId)) / Articles;
			int Prc = (int) (100 * ConceptIdToCntH.GetDat(ConceptId)) / Articles;
			double Ratio = ConceptIdToWgtH.GetDat(ConceptId) / SumWgts;
			InfoLogFile->NotifyVerboseFmt(2, "Top location: Wgt: %d, Presence: %d/%d arts. Loc: %s. Story %I64u", Wgt, ConceptIdToCntH.GetDat(ConceptId), Articles, ConceptNm.CStr(), StoryId);
			if (100 * Ratio >= Settings.StoryPropMinLocationPropToTake && Wgt >= Settings.StoryPropMinLocationWgtToTake && Prc >= Settings.StoryPropMinLocationPrcToTake) {
				SetLocationForStory(StoryId, ConceptId, MIN(100, Wgt));
			}
		}
	}
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryLocation. Finished. Time needed %.0f ms", Sw.GetMSec());
}

// update story categories - assign the story to max 3 dmoz categories
void TEventRegBase::UpdateStoryCategories(const uint64& StoryId, const bool& ForceUpdate)
{
	if (ForceUpdate == false && Settings.ComputeCategories == false)
		return;
	// using dmoz we can do good categorization only for english clusters. other clusters mainly map to /regional/spain/... or /regional/south america/...
	if (GetStoryLang(StoryId) != "eng")
		return;
	CategoryReq->RequestStoryCategory(StoryId);
}

void TEventRegBase::UpdateStoryCategoriesCallback(const PJsonVal& CategoryData) 
{
	if (CategoryData->IsObj() && CategoryData->IsObjKey("error")) {
		InfoLogFile->NotifyVerboseFmt(0, "TEventRegBase::UpdateStoryCategoriesCallback. Error from category server: %s", CategoryData->GetObjStr("error").CStr());
		return;
	}
	if (! CategoryData->IsObj() || !CategoryData->IsObjKey("key") || !CategoryData->IsObjKey("categories") || !CategoryData->GetObjKey("categories")->IsArr()) {
		InfoLogFile->NotifyVerbose(0, "TEventRegBase::UpdateStoryCategoriesCallback. Error. Invalid JSON was provided back from the server. Unable to use the information in it.");
		return;
	}
	
	TStr StoryUri = CategoryData->GetObjStr("key");
	if (!StoryStore->IsRecNm(StoryUri)) {
		InfoLogFile->NotifyVerboseFmt(0, "TEventRegBase::UpdateStoryCategoriesCallback. Received an invalid story uri %s", StoryUri.CStr());
		return;
	}
	uint64 StoryId = StoryStore->GetRecId(StoryUri);
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryCategoriesCallback. Updating story categories for story '%I64u'", StoryId);
	InfoLogFile->IncreaseSpaces();

	// delete previous categories
	StoryStore->DelJoins("hasCategory", StoryId);
	
	PJsonVal CategoriesArr = CategoryData->GetObjKey("categories");
	for (int N=0; N < CategoriesArr->GetArrVals(); N++) {
		TStr Label = CategoriesArr->GetArrVal(N)->GetObjStr("label", "");
		double Prob = CategoriesArr->GetArrVal(N)->GetObjNum("prob", 0);
		int Wgt = (int) ceil(100 * Prob);
		if (Wgt >= Settings.StoryPropMinCategoryWgtToTake) {
			NewsBase->CategoryBase->CreateDMozCats(Label);
			const TStr CatUri = NewsBase->CategoryBase->GetDMozCatUriForLabel(Label);
			
			if (!CategoryStore->IsRecNm(CatUri))
				ErrorLogFile->NotifyErr("TEventRegBase::UpdateStoryCategoriesCallback. The dmoz category should already exist. CreateDMozCats call didn't do it's job");
			uint64 CatId = CategoryStore->GetRecId(CatUri);
		
			StoryStore->AddJoin("hasCategory", StoryId, CatId, Wgt);
			if (InfoLogFile->GetVerbosityLevel() >= 3)
				InfoLogFile->NotifyVerboseFmt(3, "Story '%s' was assigned to category '%s' with weight %d", GetStoryTitle(StoryId).CStr(), CatUri.CStr(), Wgt);
		}
	}
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerbose(3, "TEventRegBase::UpdateStoryCategoriesCallback. Finished.");

	// once we've updated category for a story, update categories also for the event
	uint64 EventId = GetStoryEvent(StoryId);
	UpdateEventCategories(EventId);
}

// update the story title and summary
void TEventRegBase::UpdateStoryTitleAndSummary(const uint64& StoryId)
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const bool CustomStoryTitle = NewsBase->GetStoryFlag(StoryId, sfUserSetTitle);
	const bool CustomStorySummary = NewsBase->GetStoryFlag(StoryId, sfUserSetSummary);

	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryTitleAndSummary. Updating story title and summary for story '%I64u'", StoryId);
	InfoLogFile->IncreaseSpaces();
	// currently we use a simple approach. 
	// use the title of most central article as the title
	// use the first paragraph of most central article as the summary
	// TODO: find a better method for finding a good title and body of the story
	PRecSet ArticleRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
	TStr Title, Summary;
	if (ArticleRecSet->GetRecs() > 0) {
		ArticleRecSet->SortByField(false, ArticleStore->GetFieldId("CosToCentroid"));
		Title = ArticleStore->GetFieldNmStr(ArticleRecSet->GetRecId(0), "Title");
		Summary = ArticleStore->GetFieldNmStr(ArticleRecSet->GetRecId(0), "Body").GetSubStr(0, 700);
	}
	// set new title and summary values only if we don't have custom values for these fields
	if (! CustomStoryTitle)
		StoryStore->SetFieldNmStr(StoryId, "Title", Title);
	if (! CustomStorySummary) 
		StoryStore->SetFieldNmStr(StoryId, "Summary", Summary);
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerbose(3, "TEventRegBase::UpdateStoryTitleAndSummary. Finished");
}

// update article count for the story
int TEventRegBase::UpdateStoryArticleCount(const uint64& StoryId)
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryArticleCount. Updating story article count for story '%I64u'", StoryId);
	InfoLogFile->IncreaseSpaces();

	const int ArticleCount = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle")->GetRecs();
	const int ChangeCount = StoryStore->GetFieldNmInt(StoryId, "ChangeCount");
	TInt ArticleCountKey = 10 * ((int)(ArticleCount / 10));
	
	StoryStore->SetFieldNmInt(StoryId, "ArticleCount", ArticleCount);
	StoryStore->SetFieldNmStr(StoryId, "ArticleCountKey", ArticleCountKey.GetStr());
	StoryStore->SetFieldNmInt(StoryId, "ChangeCount", ChangeCount + 1);
	
	const uint64 EventRecId = GetStoryEvent(StoryId);
	if (EventRecId != TUInt64::Mx)
		UpdateEventArticleCount(EventRecId);
	
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateStoryArticleCount. Finished. Article count = %d", ArticleCount);
	
	return ArticleCount;
}

// update the story quality. quality is reflected in the join fq between story <-> event
void TEventRegBase::UpdateStoryQuality(const uint64& StoryId)
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const uint64 EventId = GetStoryEvent(StoryId);
	if (EventId == TUInt64::Mx)
		return;
	int Score = ComputeStoryQuality(StoryId);
	// del previous score
	EventStore->DelJoin("hasStory", EventId, StoryId);
	// add new score
	EventStore->AddJoin("hasStory", EventId, StoryId, Score);
}

// update the LastUpdate flag for the story. needed for api call to get all events updated after X
void TEventRegBase::UpdateStoryLastUpdate(const uint64& StoryId)
{
	EAssertR(StoryStore->IsRecId(StoryId), "Invalid story id");

	TTm Tm = TTm::GetCurLocTm();
	StoryStore->SetFieldNmStr(StoryId, "LastUpdate", Tm.GetWebLogDateTimeStr(true, "T", false));
}

// ////////////////
// EVENT UPDATES

// go over all articles A in the primary story. for each article check all similar articles B. 
// for each similar article B see what is the similarity between A and B. If for multiple articles A we get
// the same article B then we only consider the occurence with the highest similarity between A and B. 
// this prevents the problem of over estimating similarity due to article duplicates (same articles in A will likely have 
// as most similar the same articles B.
// once we find the set of most similar articles B we find the stories that the articles in B belong to. We compute the score of each story SB
// by summing the similarities of articles in that story
void TEventRegBase::UpdateEventStories(const uint64& OrigStoryRecId)
{
	EAssertR(StoryStore->IsRecId(OrigStoryRecId), TStr::Fmt("Story id %I64u is invalid", OrigStoryRecId));
	
	TTmStopWatch Sw; Sw.Start();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventStories. Recomputing stories to merge for story '%I64u'", OrigStoryRecId);
	InfoLogFile->IncreaseSpaces();
	if (!StoryStore->IsRecId(OrigStoryRecId))
		return;
	
	// find group of stories that include OrigStoryRecId that should be merged into a single event
	// grouping is based on various factors, including CCA similarity, concept similarity between stories, etc.
	// procedure works as follows:
	// first start with the OrigStoryRecId and find what are other stories that should be merged with it. 
	// for the extended list of stories we update the SimStoryH and StoriesToMergeV
	// stop iterating this when we find no more stories that should be merged
	TUInt64Set StoriesToMergeH;
	TMergeEval MergeEval;
	// if the story is "too messy" then don't try to merge anything with it
	/*if (GetStoryVariance(OrigStoryRecId) > GetStoryAvgCosToCentroid(OrigStoryRecId)) {
		StoriesToMergeV.Add(OrigStoryRecId);
	}
	else {*/
		// NewlyAddedStoryV will contain stories that we would like to add in each iteration
		TUInt64Set DiffStoryIdH;	// here we will keep all the story ids that are know to belong to different events than any of the stories in StoriesToMergeV
		TQQueue<TUInt64> NewlyAddedStoryV;
		NewlyAddedStoryV.Push(OrigStoryRecId);
		int Tries = 0;		// remember how many things did we try -  we want to limit ourselves since it can become very slow
		while (NewlyAddedStoryV.Len() > 0) {
			if (Tries > 100) {
				ErrorLogFile->NotifyInfo("TEventRegBase::UpdateEventStories. Reached limit (100) in computing similar stories. Stopped checking additional stories to merge.");
				break;
			}
			const uint64 AddedStory = NewlyAddedStoryV.Top(); NewlyAddedStoryV.Pop();
			if (StoriesToMergeH.IsKey(AddedStory))
				continue;
			StoriesToMergeH.AddKey(AddedStory);
			
			// if we know (from user) that AddedStory describes a different event than another set of clusters, then we can keep those clusters away from StoriesToMerge
			DiffStoryIdH.AddKeyV(NewsBase->ValidatedStoryLinks->GetValidatedDiffEvents(AddedStory));
			// if we know (from user) that AddedStory describes the same event as another set of clusters then we automatically also add these clusters
			NewlyAddedStoryV.PushV(NewsBase->ValidatedStoryLinks->GetValidatedSameEvents(AddedStory));
							
			THash<TUInt64, TSimStoryInfo> SimStoryH;
			ComputeSimilarStoriesUsingCCA(AddedStory, SimStoryH, true, true, Settings.StoryMergingMinArticleCosSim);
			SimStoryH.SortByDat(false);
			InfoLogFile->NotifyVerboseFmt(4, "Obtained %d stories that are potentially same as story '%I64u' (%s). We will now test if they should be merged:", SimStoryH.Len(), AddedStory, GetStoryTitle(AddedStory).CStr());
			InfoLogFile->IncreaseSpaces();
			
			// take only the top 5 similar stories
			for (int KeyId = SimStoryH.FFirstKeyId(), N = 0; SimStoryH.FNextKeyId(KeyId) && N < 10; N++) {
				const uint64 TestStoryId = SimStoryH.GetKey(KeyId);
				// if the user provided feedback that these stories (AddedStory and TestStoryId) don't belong to the same event then don't even test
				if (DiffStoryIdH.IsKey(TestStoryId))
					continue;
				if (StoriesToMergeH.AddKey(TestStoryId))
					continue;
				Tries++;
				const TSimStoryInfo& TestSimStoryInfo = SimStoryH.GetDat(TestStoryId);
				if (ShouldStoriesMerge(AddedStory, TestStoryId, TestSimStoryInfo, MergeEval))
					NewlyAddedStoryV.Push(TestStoryId);
			}
			InfoLogFile->DecreaseSpaces();
		}
	//}
	
	// depending on the order in which the stories were added together it is still possible that some story was added although it was later found to be incorrect one for the event
	// for this purpose we check here again if we have some stories that don't belong togher and remove them
	TUInt64V StoriesToMergeV;
	for (int KeyId = StoriesToMergeH.FFirstKeyId(); StoriesToMergeH.FNextKeyId(KeyId);) {
		const uint64 StoryId = StoriesToMergeH.GetKey(KeyId);
		if (!DiffStoryIdH.IsKey(StoryId))
			StoriesToMergeV.Add(StoryId);
	}
	
	if (StoriesToMergeV.Len() == 0)
		ErrorLogFile->NotifyErr("TEventRegBase::UpdateEventStories. StoriesToMergeV should contain at least one element");
	if (StoriesToMergeV.Len() > 1 && InfoLogFile->GetVerbosityLevel() >= 4) {
		InfoLogFile->NotifyVerbose(4, "The following stories will be merged:");
		for (int N=0; N < StoriesToMergeV.Len(); N++)
			InfoLogFile->NotifyVerboseFmt(4, "\t%I64u - %s (avgCosSim = %.4f, variance = %.4f)", StoriesToMergeV[N], GetStoryTitle(StoriesToMergeV[N]).CStr(), GetStoryAvgCosToCentroid(StoriesToMergeV[N]), GetStoryVariance(StoriesToMergeV[N]));
	}
	MergeStories(StoriesToMergeV, true);
	
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::RecomputeEventStories. Finished. Time needed %.0f ms", Sw.GetMSec());
}

// merge stories into a single event
// input: one or more stories that should be merged into a single event
// first, we find the story with best quality (TopStoryId). all the stories will be merged to the primary event for the TopStory
// if any existing stories are already merged to TargetEventId, these stories (except TopStoryId) are then moved to another event
void TEventRegBase::MergeStories(const TUInt64V& StoryIdV, const bool& RemoveExistingTargetStories, const bool& UpdateEventInfo)
{
	if (StoryIdV.Len() == 0)
		return;
	// find the story with the highest quality - we will merge to it
	uint64 TopStoryId = TUInt64::Mx; int TopStoryQuality = -1;
	TUInt64V StoriesToMergeV;
	TUInt64Set ModifiedEventsSet;
	for (int N = 0; N < StoryIdV.Len(); N++) {
		EAssertR(StoryStore->IsRecId(StoryIdV[N]), "Story id is not a valid story");
		StoriesToMergeV.Add(StoryIdV[N]);
		const int StoryQuality = NewsBase->GetStoryQuality(StoryIdV[N]);
		if (StoryQuality > TopStoryQuality) {
			TopStoryQuality = StoryQuality;
			TopStoryId = StoryIdV[N];
		}
		ModifiedEventsSet.AddKey(GetStoryEvent(StoryIdV[N]));
	}
	if (StoriesToMergeV.Len() == 0 || TopStoryId == TUInt64::Mx)
		return;

	const uint64 TargetEventId = GetStoryPrimaryEvent(TopStoryId);

	// find the stories currently assigned to the target event and remove them
	if (RemoveExistingTargetStories) {
		TUInt64V ExistingTargetStoryV; GetEventStories(TargetEventId, ExistingTargetStoryV);
		for (int N = 0; N < StoriesToMergeV.Len(); N++)
			ExistingTargetStoryV.DelIfIn(StoriesToMergeV[N]);
		// assign those stories to their top story
		if (ExistingTargetStoryV.Len() > 0)
			MergeStories(ExistingTargetStoryV, true);
	}
	MergeStories(TargetEventId, StoriesToMergeV, UpdateEventInfo);
}

// merge the stories in StoryIdV into the TargetEventId
// update also event info for each of the old events as well as the new event
void TEventRegBase::MergeStories(const uint64& TargetEventId, const TUInt64V& StoryIdV, const bool& UpdateEventInfo)
{
	EAssertR(EventStore->IsRecId(TargetEventId), TStr::Fmt("Event id %I64u is invalid", TargetEventId));

	if (StoryIdV.Len() == 0)
		return;
	for (int StoryN=0; StoryN < StoryIdV.Len(); StoryN++) {
		const uint64 StoryId = StoryIdV[StoryN];
		if (! StoryStore->IsRecId(StoryId))
			continue;
		AssignStoryToEvent(StoryId, TargetEventId, true, false);
	}
	if (UpdateEventInfo) {
		UpdateEventProperties(TargetEventId);
	}
}

// split specified stories back into their primary event
void TEventRegBase::SplitStories(const TUInt64V& StoryIdV)
{
	for (int StoryN = 0; StoryN < StoryIdV.Len(); StoryN++) {
		const uint64 StoryId = StoryIdV[StoryN];
		if (!StoryStore->IsRecId(StoryId))
			continue;
		const uint64 PrimaryEventId = GetStoryPrimaryEvent(StoryId);
		AssignStoryToEvent(StoryId, PrimaryEventId, true, true);
	}
}

// assign a story to an event. Also update the old event
void TEventRegBase::AssignStoryToEvent(const uint64& StoryId, const uint64& EventId, const bool& UpdateOldEventInfo, const bool& UpdateNewEventInfo)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	// find old event for storyId. remove old joins and create new ones
	// update the old and new event
	const uint64 OldEventId = GetStoryEvent(StoryId);
	if (OldEventId == EventId)
		return;
	if (OldEventId == TUInt64::Mx) {
		ErrorLogFile->NotifyErr("TEventRegBase::AssignStoryToEvent. Story '%I64u' was not previously assigned to any event. This should never happen.", StoryId);
		return;
	}
	if (EventId == TUInt64::Mx) {
		ErrorLogFile->NotifyErr("TEventRegBase::AssignStoryToEvent. Event id '%I64u' is invalid. Unable to assign story '%I64u' to it.", EventId, StoryId);
		return;
	}

	int Score = ComputeStoryQuality(StoryId);
	EventStore->DelJoin("hasStory", OldEventId, StoryId);
	EventStore->AddJoin("hasStory", EventId, StoryId, Score);	// event -> story && story -> event
	
	// decrease the count of articles in old event
	if (UpdateOldEventInfo)
		UpdateEventProperties(OldEventId);
		
	// increase the count of articles in new event
	if (UpdateNewEventInfo)
		UpdateEventProperties(EventId);
}

// update article count for the whole event
int TEventRegBase::UpdateEventArticleCount(const uint64& EventId) 
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventArticleCount. Updating event article count for event '%I64u'", EventId);
	InfoLogFile->IncreaseSpaces();

	PRecSet StoryRecSet = EventStore->GetRec(EventId).DoJoin(Base, "hasStory");
	int Count = 0;
	for (int N=0; N < StoryRecSet->GetRecs(); N++) 
		Count += StoryStore->GetFieldNmInt(StoryRecSet->GetRecId(N), "ArticleCount");
	TInt CountKey = 10 * ((int)(Count / 10));
	
	EventStore->SetFieldNmInt(EventId, "ArticleCount", Count);
	EventStore->SetFieldNmStr(EventId, "ArticleCountKey", CountKey.GetStr());
		
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventArticleCount. Finished. Article count = %d", Count);
	
	return Count;
}

// take the concepts from all the stories assigned to the event and assign them to the event
void TEventRegBase::UpdateEventConcepts(const uint64& EventId)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventConcepts. Updating event concepts for event '%I64u'. Relevant concepts are:", EventId);
	InfoLogFile->IncreaseSpaces();

	// go over all event stories and make hash of concept id -> sum_i (storyWgt_i * concWgt_i)
	THash<TUInt64, TFlt> ConceptToWgtH;
	PRecSet StoryRecSet = EventStore->GetRec(EventId).DoJoin(Base, "hasStory");
	for (int StoryN = 0; StoryN < StoryRecSet->GetRecs(); StoryN++) {
		uint64 StoryId = StoryRecSet->GetRecId(StoryN);
		double StoryWgt = (double) StoryRecSet->GetRecFq(StoryN) / 100.0;
		PRecSet ConceptRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasConcept");
		for (int ConcN = 0; ConcN < ConceptRecSet->GetRecs(); ConcN++) {
			uint64 ConcId = ConceptRecSet->GetRecId(ConcN);
			int ConcFq = ConceptRecSet->GetRecFq(ConcN);
			ConceptToWgtH.AddDat(ConcId) += (StoryWgt * ConcFq);
		}
	}
	ConceptToWgtH.SortByDat(false);
		
	// remove all existing joins event <--> concept
	EventStore->DelJoins("hasConcept", EventId);

	// normalize all concepts to range [0, 100]
	if (ConceptToWgtH.Len() > 0) {
		double MaxFq = ConceptToWgtH[0];
		int AddedEnts = 0, AddedKwds = 0;
		for (int KeyId = ConceptToWgtH.FFirstKeyId(); ConceptToWgtH.FNextKeyId(KeyId);) {
			const uint64 ConcId = ConceptToWgtH.GetKey(KeyId);
			const double Fq = ConceptToWgtH[KeyId];
			const int Wgt = TFlt::Round(100 * Fq / (double) MaxFq);
			if (Wgt < Settings.EventPropMinConceptWgtToTake)
				break;
			const TStr Type = ConceptStore->GetFieldNmStr(ConcId, "Type");
			if (NewsConst.IsEntity(Type) && AddedEnts >= Settings.EventPropTopConceptsToTake)
				continue;
			if (NewsConst.IsKeywordOrWiki(Type) && AddedKwds >= Settings.EventPropTopConceptsToTake)
				continue;
			if (NewsConst.IsEntity(Type))
				AddedEnts++;
			else if (NewsConst.IsKeywordOrWiki(Type))
				AddedKwds++;
			// don't add other concept types to events
			else
				continue;
			EventStore->AddJoin("hasConcept", EventId, ConcId, Wgt);  // event -> concept  && concept -> event
			if (Settings.VerbosityLevel >= 4)
				InfoLogFile->NotifyVerboseFmt(4, "\t%3d %s (id: %I64u)", Wgt, ConceptStore->GetRecNm(ConcId).CStr(), ConcId);
		}
	}
	
	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerbose(3, "TEventRegBase::UpdateEventConcepts. Finished.");
}

// take the most likely location from the stories assigned to the event
void TEventRegBase::UpdateEventLocation(const uint64& EventId)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));
	
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventLocation. Updating event location for event '%I64u'", EventId);
	InfoLogFile->IncreaseSpaces();

	uint64 EventLocationRecId = TUInt64::Mx; int EventLocationFreq = 100;
	THash<TUInt64, TInt> LocIdToWgtH;
	TRec EventRec = EventStore->GetRec(EventId);
	PRecSet StoryRecSet = EventRec.DoJoin(Base, "hasStory");
	for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
		// if we've found one story in the event that has a custom location then use that location as the final event location
		if (NewsBase->GetStoryFlag(StoryRecSet->GetRecId(N), sfUserSetLocation)) {
			EventLocationRecId = StoryStore->GetRec(StoryRecSet->GetRecId(N)).DoSingleJoin(Base, "hasLocation").GetRecId();
			break;
		}
		PRecSet LocRecSet = StoryStore->GetRec(StoryRecSet->GetRecId(N)).DoJoin(Base, "hasLocation");
		if (LocRecSet->GetRecs() > 0)
			LocIdToWgtH.AddDat(LocRecSet->GetRecId(0)) += LocRecSet->GetRecFq(0);
	}
	LocIdToWgtH.SortByDat(false);
	if (EventLocationRecId == TUInt64::Mx && LocIdToWgtH.Len() > 0) {
		int FirstKey = LocIdToWgtH.FFirstKeyId(); LocIdToWgtH.FNextKeyId(FirstKey);
		EventLocationRecId = LocIdToWgtH.GetKey(FirstKey);
		EventLocationFreq = LocIdToWgtH[FirstKey];
	}

	// remove existing joins between the event and locations
	RemoveLocationFromEvent(EventId);
		
	// set the new location for the event
	if (EventLocationRecId != TUInt64::Mx)
		SetLocationForEvent(EventId, EventLocationRecId, EventLocationFreq);

	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerbose(3, "TEventRegBase::UpdateEventLocation. Finished.");
}

// take the most likely date from the stories assigned to the event
void TEventRegBase::UpdateEventDate(const uint64& EventId)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));
	
	InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventDate. Updating event date for event '%I64u'", EventId);
	InfoLogFile->IncreaseSpaces();

	int TotArticleCount = 0;
	TStr Date, DateEnd;
	THash<TStr, TInt> DateToWgt;
	TExtractedDateColl CombinedDateColl;
	PRecSet StoryRecSet = EventStore->GetRec(EventId).DoJoin(Base, "hasStory");
	for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
		uint64 StoryId = StoryRecSet->GetRecId(N);
		if (NewsBase->GetStoryFlag(StoryId, sfUserSetDate) == true) {
			Date = StoryStore->GetFieldNmStr(StoryId, "StoryDate");
			DateEnd = StoryStore->GetFieldNmStr(StoryId, "StoryDateEnd");
			break;
		}
		TStr CommonDates = StoryStore->GetFieldNmStr(StoryId, "CommonDates");
		TExtractedDateColl Coll = TExtractedDateColl::DecodeExtractedDates(CommonDates);
		CombinedDateColl.AddColl(Coll);
		TotArticleCount += GetStoryArticleCount(StoryId);	// we don't dare to use the GetEventArticleCount here since it might not be updated yet
		TStr AvgDate = StoryStore->GetFieldNmStr(StoryId, "AverageDate");
		DateToWgt.AddDat(AvgDate) += StoryRecSet->GetRecFq(N);
	}
	// if no custom date was set then try to determine one
	if (Date == "") {
		const int MinDateFreq = (int) (Settings.DateImputationMinValidYearCnt + Settings.DateImputationMinValidYearSlope * TotArticleCount);
		CombinedDateColl.ImputeMissingYears(MinDateFreq, (Settings.VerbosityLevel > 0 && InfoLogFile != NULL && InfoLogFile->NotifyInstV.Len() > 0) ? InfoLogFile->NotifyInstV[0] : NULL);

		TExtractedDate ExtractedDate;
		bool ExtractedDateP = CombinedDateColl.GetMostFreqDate(ExtractedDate, MinDateFreq, true);
		// we found a frequent enough complete extracted date
		if (ExtractedDateP) {
			Date = ExtractedDate.GetDate();
			if (ExtractedDate.IsDateRange())
				DateEnd = ExtractedDate.GetDateEnd();
			InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventDate. Found common extracted date %s.", ExtractedDate.GetStr().CStr());
		}
		// use the most common average date for the event
		else if (DateToWgt.Len() > 0) {
			DateToWgt.SortByDat(false);
			int KeyId = DateToWgt.FFirstKeyId(); DateToWgt.FNextKeyId(KeyId);
			Date = DateToWgt.GetKey(KeyId);
			InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventDate. Using the average date for the event: %s.", Date.CStr());
		}
	}
	else 
		InfoLogFile->NotifyVerboseFmt(3, "TEventRegBase::UpdateEventDate. Using custom date for the event: %s.", Date.CStr());
	
	EventStore->SetFieldNmStr(EventId, "EventDate", Date);
	EventStore->SetFieldNmStr(EventId, "EventDateEnd", DateEnd);

	InfoLogFile->DecreaseSpaces();
	InfoLogFile->NotifyVerbose(3, "TEventRegBase::UpdateEventDate. Finished.");
}

// take categories from stories assigned to event and find 3 with the highest weight
// go over all stories, for each story find all categories for it
// take sum (category_wgt * story_wgt) / sum(story_wgt)
void TEventRegBase::UpdateEventCategories(const uint64& EventId)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));
	
	TRec EventRec = EventStore->GetRec(EventId);
	PRecSet StoryRecSet = EventRec.DoJoin(Base, "hasStory");
	THash<TUInt64, TInt> CatToWgtH;		// categories for the stories
	THash<TUInt64, TInt> CatToStoryWgtH;
	for (int N = 0; N < StoryRecSet->GetRecs(); N++) {
		PRecSet CatRecSet = StoryRecSet->GetRec(N).DoJoin(Base, "hasCategory");
		for (int CatN = 0; CatN < CatRecSet->GetRecs(); CatN++) {
			CatToWgtH.AddDat(CatRecSet->GetRecId(CatN)) += (CatRecSet->GetRecFq(CatN) * StoryRecSet->GetRecFq(N));
			CatToStoryWgtH.AddDat(CatRecSet->GetRecId(CatN)) += StoryRecSet->GetRecFq(N);
		}
	}

	// normalize (average) the weights in CatToWgtH
	for (int KeyId = CatToWgtH.FFirstKeyId(); CatToStoryWgtH.FNextKeyId(KeyId);) {
		uint64 CatId = CatToStoryWgtH.GetKey(KeyId);
		CatToWgtH[KeyId] = (int) (CatToWgtH[KeyId] / (double) CatToStoryWgtH.GetDat(CatId));
	}

	// delete previous categories
	EventStore->DelJoins("hasCategory", EventId);
	
	// add most likely event categories
	CatToWgtH.SortByDat(false);
	// add 3 top rated categories. all categories should have weight > threshold, otherwise they would not be assigned to stories
	for (int Added = 0, KeyId = CatToWgtH.FFirstKeyId(); CatToWgtH.FNextKeyId(KeyId) && Added < 3; Added++) {
		uint64 CatId = CatToWgtH.GetKey(KeyId);
		int Wgt = CatToWgtH[KeyId];
		EventStore->AddJoin("hasCategory", EventId, CatId, Wgt);
		if (InfoLogFile->GetVerbosityLevel() >= 3)
			InfoLogFile->NotifyVerboseFmt(3, "Event '%I64u' was assigned to category '%s' with weight %d", EventId, CategoryStore->GetRecNm(CatId).CStr(), Wgt);
	}
}

// ///////////
// SEARCH METHODS
void TEventRegBase::GetConceptsForEvents(const PRecSet& EventSet, const EnumConceptType& ValidTypes, const bool& NormalizeByEventCount, THash<TUInt64, TFlt>& ConceptIdWgt)
{
	ConceptIdWgt.Clr();
	int Events = EventSet->GetRecs();
	for (int N = 0; N < Events; N++) {
		PRecSet ConceptSet = EventStore->GetRec(EventSet->GetRecId(N)).DoJoin(Base, "hasConcept");
		const int Concepts = ConceptSet->GetRecs();
		for (int ConcN = 0; ConcN < Concepts; ConcN++)
			ConceptIdWgt.AddDat(ConceptSet->GetRecId(ConcN)) += ConceptSet->GetRecFq(ConcN);
	}
	// if filter by concept type then remove the concepts not of appropriate type
	if (ValidTypes != ctAll) {
		for (int KeyId = ConceptIdWgt.FFirstKeyId(); ConceptIdWgt.FNextKeyId(KeyId);) {
			const uint64 ConcId = ConceptIdWgt.GetKey(KeyId);
			if (!NewsConst.IsValidType(ConceptStore->GetFieldNmStr(ConcId, "Type"), ValidTypes))
				ConceptIdWgt.DelKeyId(KeyId);
		}
		ConceptIdWgt.Defrag();
	}

	if (NormalizeByEventCount) {
		double Count = (double) EventSet->GetRecs();
		for (int KeyId = ConceptIdWgt.FFirstKeyId(); ConceptIdWgt.FNextKeyId(KeyId);)
			ConceptIdWgt[KeyId] /= Count;
	}
}



// /////
// SIMILARITY COMPUTATION

// find a set of stories that are most similar to StoryId. similarity is computed 
// by comparing concepts associated with stories
PRecSet TEventRegBase::GetSimilarStories(const uint64& StoryId, const int& Count, const TStrSet& ValidStoryLangH, const int& MaxDayDiff, const EnumConceptType& ConceptTypes) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	TTm StoryDate;  bool EventDateOk = GetStoryDate(StoryId, StoryDate);

	// if we have a time limit then first find the candidate set of stories that will be considered for the results
	PRecSet StoriesForDateRange = TRecSet::New(NewsBase->StoryStore);
	if (MaxDayDiff != TInt::Mx && EventDateOk) {
		TTm StartDate(StoryDate); StartDate.AddDays(-MaxDayDiff);
		TTm EndDate(StoryDate); EndDate.AddDays(MaxDayDiff);
		TStr StartDateStr = TEventQuery::GetQueryDate(StartDate);
		TStr EndDateStr = TEventQuery::GetQueryDate(EndDate);
		TStr Error;
		NewsSearch->FindStoriesForDates(StartDateStr, EndDateStr, StoriesForDateRange, Error);
	}

	// first find the candidate documents - if many events we don't want to test them all
	TUInt64IntKdV ConceptIdWgtV; GetStoryConcepts(StoryId, ConceptTypes, ConceptIdWgtV, true);
	THash<TUInt64, TFlt> CandidateStoryIdWgtH;
	for (int N = 0; N < Settings.SimilarStoriesNrTopConceptsUsed && N < ConceptIdWgtV.Len(); N++) {
		TFlt ConceptScore = ConceptIdWgtV[N].Dat / 100.0;		// take into account that some concepts are more relevant for the similarity than others
		PRecSet StorySet; TStr Error; bool Ok = NewsSearch->FindStoriesForConcept(ConceptStore->GetRecNm(ConceptIdWgtV[N].Key), StorySet, Error);
		if (!Ok)
			continue;
		// if we have a time limit then remove the stories that are outside the time span (even if they might have high score)
		if (StoriesForDateRange->GetRecs() > 0)
			NewsBase->IntersectRecSets(StorySet, StoriesForDateRange, false);

		// trunc if necessary
		if (StorySet->GetRecs() > Settings.SimilarStoriesNrTopStoriesPerConcept) {
			StorySet->SortByFq(false);
			StorySet->Trunc(Settings.SimilarStoriesNrTopStoriesPerConcept);
		}

		for (int E = 0; E < StorySet->GetRecs(); E++)
			CandidateStoryIdWgtH.AddDat(StorySet->GetRecId(E)) += ConceptScore * StorySet->GetRecFq(E);
	}

	CandidateStoryIdWgtH.DelIfKey(StoryId);
	// sort by relevance because we will only evaluate top X stories
	CandidateStoryIdWgtH.Defrag();
	CandidateStoryIdWgtH.SortByDat(false);

	THash<TUInt64, TFlt> StoryConcIdWgtH; GetStoryConceptTFIDF(StoryId, ConceptTypes, StoryConcIdWgtH);
	TUInt64IntKdV SimStoryKdV;
	int EvaluatedCount = 0;
	for (int KeyId = CandidateStoryIdWgtH.FFirstKeyId(); CandidateStoryIdWgtH.FNextKeyId(KeyId);) {
		uint64 StoryId2 = CandidateStoryIdWgtH.GetKey(KeyId);
		// filtering of the stories based on the language?
		if (ValidStoryLangH.Len() > 0) {
			TStr Lang = GetStoryLang(StoryId2);
			if (!ValidStoryLangH.IsKey(Lang))
				continue;
		}
		THash<TUInt64, TFlt> StoryConcIdWgtH2; GetStoryConceptTFIDF(StoryId2, ConceptTypes, StoryConcIdWgtH2);
		double Sim = GetConceptSim(StoryConcIdWgtH, StoryConcIdWgtH2, stCos);
		int SimPrc = (int) (100 * Sim);
		SimStoryKdV.Add(TUInt64IntKd(StoryId2, SimPrc));
		EvaluatedCount++;
		if (EvaluatedCount >= Settings.SimilarStoriesMaxStoriesToEval)
			break;
	}
	SimStoryKdV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
	SimStoryKdV.Trunc(Count);
	return TRecSet::New(StoryStore, SimStoryKdV, true);
}

// find other stories that are most similar to the stories associated with stories in event EventId
PRecSet TEventRegBase::GetSimilarStoriesToEvent(const uint64& EventId, const int& Count, const TStrSet& ValidStoryLangH, const int& MaxDayDiff, const EnumConceptType& ConceptTypes) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	TTm EventDate;  bool EventDateOk = GetEventDate(EventId, EventDate);

	TUInt64V EventStoriesV; GetEventStories(EventId, EventStoriesV);
	TUInt64Set EventStoriesH(EventStoriesV);

	// if we have a time limit then first find the candidate set of stories that will be considered for the results
	PRecSet StoriesForDateRange = TRecSet::New(NewsBase->StoryStore);
	if (MaxDayDiff != TInt::Mx && EventDateOk) {
		TTm StartDate(EventDate); StartDate.AddDays(-MaxDayDiff);
		TTm EndDate(EventDate); EndDate.AddDays(MaxDayDiff);
		TStr StartDateStr = TEventQuery::GetQueryDate(StartDate);
		TStr EndDateStr = TEventQuery::GetQueryDate(EndDate);
		TStr Error;
		NewsSearch->FindStoriesForDates(StartDateStr, EndDateStr, StoriesForDateRange, Error);
	}

	// first find the candidate documents - if many events we don't want to test them all
	TUInt64IntKdV ConceptIdWgtV; GetEventConcepts(EventId, ConceptTypes, ConceptIdWgtV, true);
	THash<TUInt64, TFlt> CandidateStoryIdWgtH;
	for (int N = 0; N < Settings.SimilarStoriesNrTopConceptsUsed && N < ConceptIdWgtV.Len(); N++) {
		TFlt ConceptScore = ConceptIdWgtV[N].Dat / 100.0;		// take into account that some concepts are more relevant for the similarity than others
		PRecSet StorySet; TStr Error; bool Ok = NewsSearch->FindStoriesForConcept(ConceptStore->GetRecNm(ConceptIdWgtV[N].Key), StorySet, Error);
		if (!Ok)
			continue;
		// if we have a time limit then remove the stories that are outside the time span (even if they might have high score)
		if (StoriesForDateRange->GetRecs() > 0)
			NewsBase->IntersectRecSets(StorySet, StoriesForDateRange, false);

		// trunc if necessary
		if (StorySet->GetRecs() > Settings.SimilarStoriesNrTopStoriesPerConcept) {
			StorySet->SortByFq(false);
			StorySet->Trunc(Settings.SimilarStoriesNrTopStoriesPerConcept);
		}

		for (int E = 0; E < StorySet->GetRecs(); E++) {
			if (!EventStoriesH.IsKey(StorySet->GetRecId(E)))
				CandidateStoryIdWgtH.AddDat(StorySet->GetRecId(E)) += ConceptScore * StorySet->GetRecFq(E);
		}
	}

	// sort by relevance because we will only evaluate top X stories
	CandidateStoryIdWgtH.SortByDat(false);

	THash<TUInt64, TFlt> EventConcIdWgtH; GetEventConceptTFIDF(EventId, ConceptTypes, EventConcIdWgtH);
	TUInt64IntKdV SimStoryKdV;
	int EvaluatedCount = 0;
	for (int KeyId = CandidateStoryIdWgtH.FFirstKeyId(); CandidateStoryIdWgtH.FNextKeyId(KeyId);) {
		uint64 StoryId2 = CandidateStoryIdWgtH.GetKey(KeyId);
		// filtering of the stories based on the language?
		if (ValidStoryLangH.Len() > 0) {
			TStr Lang = GetStoryLang(StoryId2);
			if (!ValidStoryLangH.IsKey(Lang))
				continue;
		}
		THash<TUInt64, TFlt> StoryConcIdWgtH2; GetStoryConceptTFIDF(StoryId2, ConceptTypes, StoryConcIdWgtH2);
		double Sim = GetConceptSim(EventConcIdWgtH, StoryConcIdWgtH2, stCos);
		int SimPrc = (int) (100 * Sim);
		SimStoryKdV.Add(TUInt64IntKd(StoryId2, SimPrc));
		EvaluatedCount++;
		if (EvaluatedCount >= Settings.SimilarStoriesMaxStoriesToEval)
			break;
	}
	SimStoryKdV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
	SimStoryKdV.Trunc(Count);
	return TRecSet::New(StoryStore, SimStoryKdV, true);
}


// find other stories that are most similar to the stories associated with stories in event EventId
PRecSet TEventRegBase::GetSimilarEvents(const uint64& EventId, const int& Count, const int& MaxDayDiff, const EnumConceptType& ConceptTypes) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	TTm EventDate; bool EventDateOk = GetEventDate(EventId, EventDate);
	// if we have a time limit then first find the candidate set of stories that will be considered for the results
	PRecSet EventsForDateRange = TRecSet::New(NewsBase->EventStore);
	if (MaxDayDiff != TInt::Mx && EventDateOk) {
		TTm StartDate(EventDate); StartDate.AddDays(-MaxDayDiff);
		TTm EndDate(EventDate); EndDate.AddDays(MaxDayDiff);
		TStr StartDateStr = TEventQuery::GetQueryDate(StartDate);
		TStr EndDateStr = TEventQuery::GetQueryDate(EndDate);
		TStr Error; NewsSearch->FindEventsForDates(StartDateStr, EndDateStr, EventsForDateRange, Error);
	}

	// first find the candidate documents - if many events we don't want to test them all
	TUInt64IntKdV ConceptIdWgtV; GetEventConcepts(EventId, ConceptTypes, ConceptIdWgtV, true);
	THash<TUInt64, TFlt> CandidateEventIdWgtH;
	for (int N = 0; N < Settings.SimilarEventsNrTopConceptsUsed && N < ConceptIdWgtV.Len(); N++) {
		const TFlt ConceptScore = ConceptIdWgtV[N].Dat / 100.0;		// take into account that some concepts are more relevant for the similarity than others
		PRecSet EventSet; TStr Error; bool Ok = NewsSearch->FindEventsForConcept(ConceptStore->GetRecNm(ConceptIdWgtV[N].Key), EventSet, Error);
		if (!Ok)
			continue;
		// if we have a time limit then remove the events that are outside the time span (even if they might have high score)
		if (EventsForDateRange->GetRecs() > 0)
			NewsBase->IntersectRecSets(EventSet, EventsForDateRange, false);

		// trunc if ncessary
		if (EventSet->GetRecs() > Settings.SimilarEventsNrTopEventsPerConcept) {
			EventSet->SortByFq(false);
			EventSet->Trunc(Settings.SimilarEventsNrTopEventsPerConcept);
		}

		for (int E = 0; E < EventSet->GetRecs(); E++)
			CandidateEventIdWgtH.AddDat(EventSet->GetRecId(E)) += ConceptScore * EventSet->GetRecFq(E);
	}

	CandidateEventIdWgtH.DelIfKey(EventId);
	// sort by relevance because we will only evaluate top X stories
	CandidateEventIdWgtH.Defrag();
	CandidateEventIdWgtH.SortByDat(false);

	// todo: maybe sort and take only a subset of candidates
	THash<TUInt64, TFlt> EventConcIdWgtH; GetEventConceptTFIDF(EventId, ConceptTypes, EventConcIdWgtH);
	TUInt64IntKdV SimEventKdV;
	int EvaluatedCount = 0;
	for (int KeyId = CandidateEventIdWgtH.FFirstKeyId(); CandidateEventIdWgtH.FNextKeyId(KeyId);) {
		uint64 EventId2 = CandidateEventIdWgtH.GetKey(KeyId);
		THash<TUInt64, TFlt> EventConcIdWgtH2; GetEventConceptTFIDF(EventId2, ConceptTypes, EventConcIdWgtH2);
		double Sim = GetConceptSim(EventConcIdWgtH, EventConcIdWgtH2, stCos);
		int SimPrc = (int) (100 * Sim);
		SimEventKdV.Add(TUInt64IntKd(EventId2, SimPrc));
		EvaluatedCount++;
		if (EvaluatedCount >= Settings.SimilarEventsMaxEventsToEval)
			break;
	}
	SimEventKdV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
	SimEventKdV.Trunc(Count);
	return TRecSet::New(EventStore, SimEventKdV, true);
}

PRecSet TEventRegBase::GetSimilarStoriesByCCA(const uint64& TestStoryId, const int& Count, const TStrSet& ValidStoryLangH) const
{
	THash<TUInt64, TSimStoryInfo> SimStoryH;
	ComputeSimilarStoriesUsingCCA(TestStoryId, SimStoryH, false, false);

	TUInt64IntKdV StoryIdWgtV;
	for (int KeyId = SimStoryH.FFirstKeyId(); SimStoryH.FNextKeyId(KeyId);) {
		const uint64 StoryId = SimStoryH.GetKey(KeyId);
		if (TestStoryId == StoryId)
			continue;
		const TStr Lang = GetStoryLang(StoryId);
		if (ValidStoryLangH.Len() > 0 && !ValidStoryLangH.IsKey(Lang))
			continue;
		const int Wgt = SimStoryH[KeyId].MaxArtCount > 0 ?
			MAX(1, (int) (100 * SimStoryH[KeyId].ArtMatchCount / SimStoryH[KeyId].MaxArtCount)) :
			1;
		StoryIdWgtV.Add(TUInt64IntKd(StoryId, Wgt));
	}
	StoryIdWgtV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
	StoryIdWgtV.Trunc(Count);
	return TRecSet::New(NewsBase->StoryStore, StoryIdWgtV, true);
}

PRecSet TEventRegBase::GetSimilarStoriesToEventByCCA(const uint64& TestEventId, const int& Count, const TStrSet& ValidStoryLangH) const
{
	THash<TUInt64, TSimStoryInfo> FinalSimStoryH;
	TUInt64V StoryIdV;
	GetEventStories(TestEventId, StoryIdV);
	for (int N = 0; N < StoryIdV.Len(); N++) {
		THash<TUInt64, TSimStoryInfo> SimStoryH;
		ComputeSimilarStoriesUsingCCA(StoryIdV[N], SimStoryH, false, false);
		for (int KeyId = SimStoryH.FFirstKeyId(); SimStoryH.FNextKeyId(KeyId);) {
			const TSimStoryInfo& Info = SimStoryH[KeyId];
			const uint64 StoryId = SimStoryH.GetKey(KeyId);
			// update existing sim story info
			if (FinalSimStoryH.IsKey(StoryId)) {
				FinalSimStoryH.GetDat(StoryId).ArtMatchCount += Info.ArtMatchCount;
				FinalSimStoryH.GetDat(StoryId).MaxArtCount += Info.MaxArtCount;
			}
			else
				FinalSimStoryH.AddDat(StoryId, Info);
		}
	}

	TUInt64IntKdV StoryIdWgtV;
	for (int KeyId = FinalSimStoryH.FFirstKeyId(); FinalSimStoryH.FNextKeyId(KeyId);) {
		const uint64 StoryId = FinalSimStoryH.GetKey(KeyId);
		if (StoryIdV.IsIn(StoryId))		// don't add stories in the EventId
			continue;
		const TStr Lang = GetStoryLang(StoryId);
		if (ValidStoryLangH.Len() > 0 && !ValidStoryLangH.IsKey(Lang))
			continue;
		const int Wgt = FinalSimStoryH[KeyId].MaxArtCount > 0 ?
			MAX(1, (int) (100 * FinalSimStoryH[KeyId].ArtMatchCount / FinalSimStoryH[KeyId].MaxArtCount)) :
			1;
		StoryIdWgtV.Add(TUInt64IntKd(StoryId, Wgt));
	}
	StoryIdWgtV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
	StoryIdWgtV.Trunc(Count);
	return TRecSet::New(NewsBase->StoryStore, StoryIdWgtV, true);
}

PRecSet TEventRegBase::GetSimilarEventsByCCA(const uint64& TestEventId, const int& Count) const
{
	PRecSet StorySet = GetSimilarStoriesToEventByCCA(TestEventId, TInt::Mx, TStrSet());
	THash<TUInt64, TInt> EventIdWgtH;
	for (int N = 0; N < StorySet->GetRecs(); N++) {
		const uint64 EventId = GetStoryEvent(StorySet->GetRecId(N));
		if (TestEventId == EventId)
			continue;
		const int Fq = StorySet->GetRecFq(N);
		if (EventIdWgtH.IsKey(EventId))
			EventIdWgtH[EventId] = MAX(EventIdWgtH.GetDat(EventId), Fq);
		else
			EventIdWgtH[EventId] = Fq;
	}
	TUInt64IntKdV RecIdFqV; EventIdWgtH.GetKeyDatKdV(RecIdFqV);
	RecIdFqV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
	RecIdFqV.Trunc(Count);
	return TRecSet::New(NewsBase->EventStore, RecIdFqV, true);
}

// find a set of stories that are most similar to StoryId. similarity is computed 
// by comparing concepts associated with stories
bool TEventRegBase::GetSimilarConcepts(PRecSet& RecSet, TStr& Error, const TUInt64& ConceptId, const int& Count, const EnumConceptType& ValidTypes)
{
	PRecSet EventSet; bool Ok = NewsSearch->FindEventsForConcept(ConceptStore->GetRecNm(ConceptId), EventSet, Error);
	if (!Ok)
		return false;

	const int MinConceptWgt = 20;
	// put events where ConceptId is very relevant on top
	EventSet->SortByFq(false);
	PRecSet SubEventSet = EventSet->GetLimit(500, 0);

	THash<TUInt64, TInt> ConceptToCntH;
	for (int EventN = 0; EventN < SubEventSet->GetRecs(); EventN++) {
		const TRec EventRec = SubEventSet->GetRec(EventN);
		PRecSet ConceptRecSet = EventRec.DoJoin(NewsBase->Base, "hasConcept");
		int Concepts = ConceptRecSet->GetRecs();
		for (int ConceptN = 0; ConceptN < Concepts; ConceptN++) {
			if (ConceptRecSet->GetRecFq(ConceptN) < MinConceptWgt)
				continue;
			const uint64 ConceptId = ConceptRecSet->GetRecId(ConceptN);
			const TStr ConceptType = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
			if (!NewsConst.IsValidType(ConceptType, ValidTypes))
				continue;
			ConceptToCntH.AddDat(ConceptId) += 1;
		}
	}

	ConceptToCntH.SortByDat(false);
	const int MaxWgt = SubEventSet->GetRecs();
	TUInt64IntKdV ConceptIdWgtV;
	for (int KeyId = ConceptToCntH.FFirstKeyId(); ConceptToCntH.FNextKeyId(KeyId) && ConceptIdWgtV.Len() < Count;) {
		ConceptIdWgtV.Add(TUInt64IntKd(ConceptToCntH.GetKey(KeyId), (int) (100 * ConceptToCntH[KeyId] / (double) MaxWgt)));
	}
	RecSet = TRecSet::New(ConceptStore, ConceptIdWgtV, true);
	return true;
}

// /////
// HELPER FUNCTIONS
double TEventRegBase::GetConceptSimEvents(const uint64& EventId1, const uint64& EventId2, const EnumSimType& SimType, const EnumConceptType& ConceptTypes) const
{
	THash<TUInt64, TFlt> ConcH1; GetEventConceptTFIDF(EventId1, ConceptTypes, ConcH1);
	THash<TUInt64, TFlt> ConcH2; GetEventConceptTFIDF(EventId2, ConceptTypes, ConcH2);
	return GetConceptSim(ConcH1, ConcH2, SimType);
}

double TEventRegBase::GetConceptSimStories(const uint64& StoryId1, const uint64& StoryId2, const EnumSimType& SimType, const EnumConceptType& ConceptTypes) const
{
	THash<TUInt64, TFlt> ConcH1; GetStoryConceptTFIDF(StoryId1, ConceptTypes, ConcH1);
	THash<TUInt64, TFlt> ConcH2; GetStoryConceptTFIDF(StoryId2, ConceptTypes, ConcH2);
	return GetConceptSim(ConcH1, ConcH2, SimType);
}

// compute the cosine/jaccard similarity
// if the hash sets don't contain any items return -1
double TEventRegBase::GetConceptSim(const THash<TUInt64, TFlt>& ConceptIdWgtH1, const THash<TUInt64, TFlt>& ConceptIdWgtH2, const EnumSimType& SimType) const
{
	const double Eps = 1e-6;
	if (ConceptIdWgtH1.Len() == 0 && ConceptIdWgtH2.Len() == 0)
		return -1;
	if (SimType == stCos) {
		double WordWgtSum = 0;
		for (int KeyId = ConceptIdWgtH1.FFirstKeyId(); ConceptIdWgtH1.FNextKeyId(KeyId);) {
			uint64 ConceptId = ConceptIdWgtH1.GetKey(KeyId);
			if (ConceptIdWgtH2.IsKey(ConceptId))
				WordWgtSum += ConceptIdWgtH1.GetDat(ConceptId) * ConceptIdWgtH2.GetDat(ConceptId);
		}
		if (WordWgtSum < Eps)
			return 0;
		double Norm1 = 0;
		for (int KeyId = ConceptIdWgtH1.FFirstKeyId(); ConceptIdWgtH1.FNextKeyId(KeyId);)
			Norm1 += TMath::Sqr(ConceptIdWgtH1[KeyId]);
		Norm1 = sqrt(Norm1);
		double Norm2 = 0;
		for (int KeyId = ConceptIdWgtH2.FFirstKeyId(); ConceptIdWgtH2.FNextKeyId(KeyId);)
			Norm2 += TMath::Sqr(ConceptIdWgtH2[KeyId]);
		Norm2 = sqrt(Norm2);
		if (Norm1*Norm2 < Eps)
			return 0;
		double CosSim = WordWgtSum / (Norm1*Norm2);
		return CosSim;
	}
	else if (SimType == stJaccard) {
		int Ints = 0, Union = ConceptIdWgtH1.Len();
		for (int KeyId = ConceptIdWgtH2.FFirstKeyId(); ConceptIdWgtH2.FNextKeyId(KeyId);) {
			uint64 ConceptId = ConceptIdWgtH2.GetKey(KeyId);
			if (ConceptIdWgtH1.IsKey(ConceptId))
				Ints++;
			else
				Union++;
		}
		double JacSim = (Union == 0) ? 0 : Ints / (double) Union;
		return JacSim;
	}
	else
		return 0;
}

void TEventRegBase::ReportStoryMerge(const uint64& PrimaryStoryId, const uint64& StoryId, const int& AvgSim, const uint64& EventRecId)
{
	if (InfoLogFile->GetVerbosityLevel() < 2)
		return;
	TRec Art1 =  StoryStore->GetRec(PrimaryStoryId).DoSingleJoin(Base, "hasMedoid");
	TRec Art2 = StoryStore->GetRec(StoryId).DoSingleJoin(Base, "hasMedoid");
	if (Art1.IsDef() && Art2.IsDef() &&
		ArticleStore->IsRecId(Art1.GetRecId()) && ArticleStore->IsRecId(Art2.GetRecId())) {
		TStr Title1 = GetStoryTitle(PrimaryStoryId);
		TStr Title2 = GetStoryTitle(StoryId);
		TStr Body1 = GetStoryBody(PrimaryStoryId);
		TStr Body2 = GetStoryBody(StoryId);
		Body1 = Body1.GetSubStr(0, 150);
		Body2 = Body2.GetSubStr(0, 150);
		InfoLogFile->NotifyVerboseFmt(2, "=== Mergining stories (similarity %.3f, event id %s )===\nTitle 1: %s\nTitle 2: %s\nBody 1: %s\nBody 2: %s\n======", AvgSim / 100.0, TUInt64(EventRecId).GetStr().CStr(), Title1.CStr(), Title2.CStr(), Body1.CStr(), Body2.CStr());
	}
}

TStr TEventRegBase::PrintEventInfo(const uint64& EventRecId)
{
	EAssertR(EventStore->IsRecId(EventRecId), TStr::Fmt("Event id %I64u is invalid", EventRecId));

	TChA Out;
	TRec EventRec(EventStore, EventRecId);
	TRec PrimStoryRec = EventRec.DoSingleJoin(Base, "hasPrimaryStory");
	TUInt64 PrimRecId = PrimStoryRec.IsDef() ? (TUInt64) PrimStoryRec.GetRecId() : TUInt64::Mx;
	PRecSet StoryRecSet = EventRec.DoJoin(Base, "hasStory");
	Out += TStr::Fmt("Event id: %I64u\nStory count: %d\nPrimary story id: %s\n", EventRecId, StoryRecSet->GetRecs(), PrimRecId.GetStr().CStr());
	for (int StoryN = 0; StoryN < StoryRecSet->GetRecs(); StoryN++) {
		TUInt64 StoryId = StoryRecSet->GetRecId(StoryN);
		TStr Lang = StoryStore->GetFieldNmStr(StoryId, "Language");
		TInt Changes = StoryStore->GetFieldNmInt(StoryId, "ChangeCount");
		Out += TStr::Fmt("============\nStory: %s (id: %I64u, lang: %s, change count: %d)\n", StoryStore->GetRecNm(StoryId).CStr(), StoryId.Val, Lang.CStr(), Changes.Val);
		PRecSet ConceptRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasConcept");
		ConceptRecSet->SortByFq(false);
		Out += TStr::Fmt("Concepts (%d):\n", ConceptRecSet->GetRecs());
		for (int ConcN=0; ConcN < ConceptRecSet->GetRecs(); ConcN++) {
			const uint64 ConcId = ConceptRecSet->GetRecId(ConcN);
			const TStr Uri = ConceptStore->GetRecNm(ConcId);
			const TStr Label = NewsBase->ConceptLabel->GetConceptLabel(Uri, "eng");
			Out += TStr::Fmt("   %03d  %s | %s\n", ConceptRecSet->GetRecFq(ConcN), Label.CStr(), Uri.CStr());
		}
		const TStr AvgDate = StoryStore->GetFieldNmStr(StoryId, "AverageDate");
		Out += TStr::Fmt("Average date: %s\n", AvgDate.CStr());
		const TStr DatesJson = StoryStore->GetFieldNmStr(StoryId, "CommonDates");
		TExtractedDateColl DateColl(DatesJson);
		Out += TStr::Fmt("Extracted dates (%d):\n", DateColl.Len());
		for (int DateN=0; DateN < DateColl.Len(); DateN++) {
			TExtractedDate Date = DateColl.GetExtractedDate(DateN);
			if (Date.IsDateRange())
				Out += TStr::Fmt("   %s - %s (freq: %d)\n", Date.GetDate().CStr(), Date.GetDateEnd().CStr(), Date.GetFreq());
			else
				Out += TStr::Fmt("   %s (freq: %d)\n", Date.GetDate().CStr(), Date.GetFreq());
		}
		PRecSet ArticleRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
		Out += TStr::Fmt("Articles (%d):\n", ArticleRecSet->GetRecs());
		for (int ArtN=0; ArtN < ArticleRecSet->GetRecs(); ArtN++) {
			TUInt64 ArtId = ArticleRecSet->GetRecId(ArtN);
			TStr Uri = ArticleStore->GetRecNm(ArtId);
			TStr Title = ArticleStore->GetFieldNmStr(ArtId, "Title");
			TStr Body = ArticleStore->GetFieldNmStr(ArtId, "Body");
			TFlt Cos = ArticleStore->GetFieldNmFlt(ArtId, "CosToCentroid");
			TTm Time; ArticleStore->GetFieldNmTm(ArtId, "DateTime", Time);
			Body = Body.GetSubStr(0, 150);
			Body.ChangeChAll('\n', ' ');
			Out += TStr::Fmt("  Art %s (id: %s)\n      Cos to centroid: %.4f, date: %s\n      Title: %s\n      Body: %s\n", Uri.CStr(), ArtId.GetStr().CStr(), Cos.Val, Time.GetWebLogDateTimeStr(true, " ", false).CStr(), Title.CStr(), Body.CStr());
		}
	}
	TStr OutStr = Out;
	OutStr.ChangeChAll('%', ' ');
	InfoLogFile->NotifyVerbose(0, OutStr.CStr());
	return Out;
}

TStr TEventRegBase::PrintArticleInfo(const uint64& ArticleRecId)
{
	if (!ArticleStore->IsRecId(ArticleRecId))
		return TStr::Fmt("Provided article id '%I64u' does not match any valid article", ArticleRecId);
	TChA Out;
	Out += "URI: " + ArticleStore->GetRecNm(ArticleRecId) + "\n";
	Out += "Title: " + ArticleStore->GetFieldNmStr(ArticleRecId, "Title") + "\n";
	Out += "Lang: " + ArticleStore->GetFieldNmStr(ArticleRecId, "Language") + "\n";
	TTm Time;
	ArticleStore->GetFieldNmTm(ArticleRecId, "DateTime", Time);
	Out += "Time: " + Time.GetWebLogDateTimeStr() + "\n";
	Out += "Body: " + ArticleStore->GetFieldNmStr(ArticleRecId, "Body") + "\n";
	TStr OutStr = Out;
	OutStr.ChangeChAll('%', ' ');
	InfoLogFile->NotifyVerbose(0, OutStr.CStr());
	return OutStr;
}

void TEventRegBase::ReloadSettings()
{
	const TStr SettingsStr = TFile::Exists(SettingsFName) ? TStr::LoadTxt(SettingsFName) : "{}";
	const PJsonVal AllSettings = TJsonVal::GetValFromStr(SettingsStr);
	const PJsonVal EventRegSettings = AllSettings->IsObjKey("eventRegistry") ? AllSettings->GetObjKey("eventRegistry") : TJsonVal::NewObj();

	SetSettings(EventRegSettings);
	InfoLogFile->SetVerbosityLevel(Settings.VerbosityLevel);
}

void TEventRegBase::SetSettings(const PJsonVal& SettingsJson)
{
	Settings = TEventSettings(SettingsJson);
	if (InfoLogFile != NULL)
		InfoLogFile->SetVerbosityLevel(Settings.VerbosityLevel);
}

// the original story from the event can be reassigned to another event. 
// this method return the EventId if the event has at least one story in it
// if the event has no stories (the original story was reassigned to another event) then we return the id 
// of the event to which the primary story was assigned to
uint64 TEventRegBase::GetValidEventId(const uint64& EventId) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));
	
	if (EventStore->GetRec(EventId).DoJoin(Base, "hasStory")->GetRecs() > 0)
		return EventId;
	TRec PrimaryStoryRec = EventStore->GetRec(EventId).DoSingleJoin(Base, "hasPrimaryStory");
	if (!PrimaryStoryRec.IsDef())
		return TUInt64::Mx;
	TRec EventRec = PrimaryStoryRec.DoSingleJoin(Base, "hasEvent");
	return EventRec.GetRecId();
}

// check the primary story and return the starting (and potentially ending dates) for the event
void TEventRegBase::GetEventDate(const uint64& EventId, TStr& DateStart, TStr& DateEnd) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	DateStart = ""; DateEnd = "";
	if (!EventStore->IsRecId(EventId))
		return;
	DateStart = EventStore->GetFieldNmStr(EventId, "EventDate");
	DateEnd = EventStore->GetFieldNmStr(EventId, "EventDateEnd");
}

// return the starting date of the event as TTm
bool TEventRegBase::GetEventDate(const uint64& EventId, TTm& EventDate) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	TStr DateStart = EventStore->GetFieldNmStr(EventId, "EventDate");
	return ParseDate(DateStart, EventDate);
}

// get the average cosine similarity to centroid of all articles belonging to event
double TEventRegBase::GetEventAvgCosToCentroid(const uint64& EventId) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	TUInt64V StoryIdV; GetEventStories(EventId, StoryIdV);
	int TotalSize = 0;
	double TotalAvg = 0;
	for (int N = 0; N < StoryIdV.Len(); N++) {
		const double Avg = GetStoryAvgCosToCentroid(StoryIdV[N]);
		const int StorySize = GetStoryArticleCount(StoryIdV[N]);
		TotalSize += StorySize;
		TotalAvg += Avg * StorySize;
	}
	if (TotalSize > 0)
		return TotalAvg / (double) TotalSize;
	else return 0;
}

// get the primary story id
uint64 TEventRegBase::GetEventPrimaryStoryId(const uint64& EventId) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	TRec Story = EventStore->GetRec(EventId).DoSingleJoin(Base, "hasPrimaryStory");
	return Story.GetRecId();
}

void TEventRegBase::GetEventStories(const uint64& EventId, TUInt64V& StoryIdV, const bool& SortByQualityP) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	StoryIdV.Clr();
	if (!EventStore->IsRecId(EventId))
		return;
	PRecSet StoryRecSet = EventStore->GetRec(EventId).DoJoin(Base, "hasStory");
	// sort by story freq join to event (freq represents the story quality)
	if (SortByQualityP)
		StoryRecSet->SortByFq(false);
	StoryRecSet->GetRecIdV(StoryIdV);
}

// get the central article and return its title
TStr TEventRegBase::GetStoryTitle(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const TStr Title = StoryStore->GetFieldNmStr(StoryId, "Title");
	if (Title != "")
		return Title;
	const TRec StoryRec = StoryStore->GetRec(StoryId);
	const TRec Art = StoryRec.DoSingleJoin(Base, "hasMedoid");
	if (Art.IsDef() && ArticleStore->IsRecId(Art.GetRecId()))
		return ArticleStore->GetFieldNmStr(Art.GetRecId(), "Title");
	const PRecSet ArtSet = StoryRec.DoJoin(Base, "hasArticle");
	if (ArtSet->GetRecs() > 0)
		return ArticleStore->GetFieldNmStr(ArtSet->GetRecId(0), "Title");
	return "";
}

// get the central article and return its content (maybe shortened)
TStr TEventRegBase::GetStoryBody(const uint64& StoryId, const int& MaxLen) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	TStr Summary = StoryStore->GetFieldNmStr(StoryId, "Summary");
	if (Summary != "") {
		Summary.ChangeChAll('\n', ' ');
		return Summary.GetSubStr(0, MaxLen);
	}
	uint64 ArtRecId = TUInt64::Mx;
	TRec StoryRec = StoryStore->GetRec(StoryId);
	TRec Art = StoryRec.DoSingleJoin(Base, "hasMedoid");
	if (Art.IsDef() && ArticleStore->IsRecId(Art.GetRecId()))
		ArtRecId = Art.GetRecId();
	else {
		PRecSet ArtSet = StoryRec.DoJoin(Base, "hasArticle");
		if (ArtSet->GetRecs() > 0)
			ArtRecId = ArtSet->GetRecId(0);
	}
	if (ArtRecId == TUInt64::Mx || !ArticleStore->IsRecId(ArtRecId))
		return "";
	TStr Body = ArticleStore->GetFieldNmStr(ArtRecId, "Body");
	Body.ChangeChAll('\n', ' ');
	return Body.GetSubStr(0, MaxLen);
}

// return starting (and potentially ending) dates of the story
void TEventRegBase::GetStoryDate(const uint64& StoryId, TStr& DateStart, TStr& DateEnd) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	DateStart = StoryStore->GetFieldNmStr(StoryId, "StoryDate");
	DateEnd = StoryStore->GetFieldNmStr(StoryId, "StoryDateEnd");
}

// return the starting date of the story as TTm
bool TEventRegBase::GetStoryDate(const uint64& StoryId, TTm& StoryDate) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	TStr DateStart = StoryStore->GetFieldNmStr(StoryId, "StoryDate");
	return ParseDate(DateStart, StoryDate);
}

TStr TEventRegBase::GetStoryLang(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	return StoryStore->GetFieldNmStr(StoryId, "Language");
}

void TEventRegBase::GetStoryAnnArticles(const uint64& StoryId, int& AnnotatedArts, int& TotalArts) const
{
	PRecSet Articles = StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasArticle");
	TotalArts = Articles->GetRecs();
	AnnotatedArts = 0;
	for (int N = 0; N < Articles->GetRecs(); N++) {
		if (ArticleStore->GetRec(Articles->GetRecId(N)).DoJoin(NewsBase->Base, "hasConcept")->GetRecs() > 0)
			AnnotatedArts++;
	}
}

// get the average time of the article in the story (in hours since 1.1.2001) and the standard deviation of the time
// to avoid the outliers also remove 10% of first and last articles and compute Avg and StDev on the subset
bool TEventRegBase::GetStoryTimes(const uint64& StoryId, double& AvgTime, double& StDevTime, double& CoreAvgTime, double& CoreStDevTime) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	PRecSet ArticleSet1 = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
	uint64 TimeStartMS = TTm::GetMSecsFromTm(TTm(2000, 1, 1));
	TFltV ArtHrV;
	for (int N = 0; N < ArticleSet1->GetRecs(); N++) {
		uint64 ArtId = ArticleSet1->GetRecId(N);
		if (!ArticleStore->IsRecId(ArtId))
			continue;
		TTm Tm; ArticleStore->GetFieldNmTm(ArtId, "DateTime", Tm);
		ArtHrV.Add((TSysTm::GetMSecsFromTm(Tm) - TimeStartMS) / (3600 * 1000));	// convert ms to hours
	}
	if (ArtHrV.Len() > 0) {
		TFlt Sum = 0;
		for (int N = 0; N < ArtHrV.Len(); N++)
			Sum += ArtHrV[N];
		AvgTime = Sum / (double) ArtHrV.Len();
		TFlt StDev = 0;
		for (int N = 0; N < ArtHrV.Len(); N++)
			StDev += (ArtHrV[N] - AvgTime) * (ArtHrV[N] - AvgTime);
		StDev /= (double) ArtHrV.Len();
		StDevTime = sqrt(StDev);

		int Off = (int) (ArtHrV.Len() / 10.0);
		Sum = 0;
		for (int N = Off; N < ArtHrV.Len() - Off; N++)
			Sum += ArtHrV[N];
		CoreAvgTime = Sum / (double) (MAX(1, ArtHrV.Len() - (2 * Off)));
		StDev = 0;
		for (int N = Off; N < ArtHrV.Len() - Off; N++)
			StDev += (ArtHrV[N] - CoreAvgTime) * (ArtHrV[N] - CoreAvgTime);
		StDev /= (double) (MAX(1, ArtHrV.Len() - (2 * Off)));
		CoreStDevTime = sqrt(StDev);
		return true;
	}
	else
		return false;
}

// return the event id to which the story StoryId is assigned to
uint64 TEventRegBase::GetStoryEvent(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const uint64 EventId = StoryStore->GetRec(StoryId).DoSingleJoin(Base, "hasEvent").GetRecId();
	AssertR(EventStore->IsRecId(EventId), "GetStoryEvent: Event id value invalid");
	return EventId;
}

// find the id of the event to which the StoryId is primarily assigned to
uint64 TEventRegBase::GetStoryPrimaryEvent(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const uint64 EventId = StoryStore->GetRec(StoryId).DoSingleJoin(Base, "hasPrimaryEvent").GetRecId();
	AssertR(EventStore->IsRecId(EventId), "GetStoryPrimaryEvent: Event id value invalid");
	return EventId;
}

int TEventRegBase::GetStoryArticleCount(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	return StoryStore->GetFieldNmInt(StoryId, "ArticleCount");
}

double TEventRegBase::GetStoryAvgCosToCentroid(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	return StoryStore->GetFieldNmFlt(StoryId, "AvgCosToCentroid");
}

double TEventRegBase::GetStoryVariance(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	return StoryStore->GetFieldNmFlt(StoryId, "Variance");
}

// return concepts assigned to the story. ignore concepts of type not specified in ValidConceptTypeH
void TEventRegBase::GetStoryConcepts(const uint64& StoryId, const EnumConceptType& ConceptTypes, TUInt64IntKdV& ConceptIdWgtV, const bool& SortP) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	PRecSet ConceptSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasConcept");
	for (int N = 0; N < ConceptSet->GetRecs(); N++) {
		const uint64 ConceptId = ConceptSet->GetRecId(N);
		if (NewsConst.IsValidType(ConceptStore->GetFieldNmStr(ConceptId, "Type"), ConceptTypes))
			ConceptIdWgtV.Add(TUInt64IntKd(ConceptId, ConceptSet->GetRecFq(N)));
	}
	if (SortP)
		ConceptIdWgtV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
}

void TEventRegBase::GetEventConcepts(const uint64& EventId, const EnumConceptType& ConceptTypes, TUInt64IntKdV& ConceptIdWgtV, const bool& SortP) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	PRecSet ConceptSet = EventStore->GetRec(EventId).DoJoin(Base, "hasConcept");
	for (int N = 0; N < ConceptSet->GetRecs(); N++) {
		const uint64 ConceptId = ConceptSet->GetRecId(N);
		if (NewsConst.IsValidType(ConceptStore->GetFieldNmStr(ConceptId, "Type"), ConceptTypes))
			ConceptIdWgtV.Add(TUInt64IntKd(ConceptId, ConceptSet->GetRecFq(N)));
	}
	if (SortP)
		ConceptIdWgtV.SortCmp(TCmpKeyDatByDat<TUInt64, TInt>(false));
}

// get the concepts for the given story. ignore concepts not specified in ConceptTypes. weight concepts based on the frequency of concept in all articles
void TEventRegBase::GetStoryConceptTFIDF(const uint64& StoryId, const EnumConceptType& ConceptTypes, THash<TUInt64, TFlt>& ConceptIdWgtH) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	ConceptIdWgtH.Clr();
	TUInt64IntKdV ConceptIdWgtV; GetStoryConcepts(StoryId, ConceptTypes, ConceptIdWgtV, false);
	for (int N = 0; N < ConceptIdWgtV.Len(); N++) {
		const uint64 ConceptId = ConceptIdWgtV[N].Key;
		const uint64 ConcFq = NewsBase->NewsCounts->GetConceptFq(ConceptId);	// DF
		const uint64 Arts = NewsBase->NewsCounts->GetAnnArticleCount();			// Docs
		const double IDF = double(MAX(1, Arts)) / MAX(1, ConcFq);
		ConceptIdWgtH.AddDat(ConceptId, ConceptIdWgtV[N].Dat * log(IDF));
	}
}

// compute weights and counts for cities mentioned in articles belonging to a story
void TEventRegBase::GetStoryLocationsInfo(const uint64& StoryId, TStoryLocationsInfo& StoryLocationsInfo) const
{
	PRecSet ArticleRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
	StoryLocationsInfo.Articles = ArticleRecSet->GetRecs();
	//double Sum
	for (int N = 0; N < StoryLocationsInfo.Articles; N++) {
		const uint64 ArticleRecId = ArticleRecSet->GetRecId(N);
		const double CosSim = ArticleStore->GetFieldNmFlt(ArticleRecId, "CosToCentroid");
		const PRecSet ConceptRecSet = ArticleRecSet->GetRec(N).DoJoin(Base, "hasConcept");
		StoryLocationsInfo.SumCosSim += CosSim;
		for (int ConceptN = 0; ConceptN < ConceptRecSet->GetRecs(); ConceptN++) {
			const uint64 ConceptId = ConceptRecSet->GetRecId(ConceptN);
			const int ConceptRecFq = ConceptRecSet->GetRecFq(ConceptN);
			// we only consider a city as a valid location of an event. a country is too general
			if (!NewsBase->GeoBase->IsCity(ConceptStore->GetRecNm(ConceptId)))
				continue;
			// debug:
			const TStr Uri = ConceptStore->GetRecNm(ConceptId);
			StoryLocationsInfo.CandidateLocationH.AddKey(ConceptId);
			StoryLocationsInfo.ConceptIdToWgtH.AddDat(ConceptId) += CosSim;
			StoryLocationsInfo.ConceptIdToCntH.AddDat(ConceptId) += 1;
			StoryLocationsInfo.SumConceptWgts += CosSim;
		}
		const uint64 ExplicitLoc = ArticleStore->GetFieldNmUInt64(ArticleRecId, "hasLocationId");
		// if we found the location of the article then increase its value based on the centrality of the article
		if (ExplicitLoc != TUInt64::Mx && NewsBase->GeoBase->IsCity(ConceptStore->GetRecNm(ExplicitLoc))) {
			StoryLocationsInfo.CandidateLocationH.AddKey(ExplicitLoc);
			StoryLocationsInfo.SumCosSimForExplicitArticles += CosSim;
			StoryLocationsInfo.ExplicitConceptIdToWgtH.AddDat(ExplicitLoc) += CosSim;
			StoryLocationsInfo.ExplicitConceptIdToCntH.AddDat(ExplicitLoc) += 1;
		}
	}
}

// get the concepts for the given event. ignore concepts not specified in ConceptTypes. weight concepts based on the frequency of concept in all articles
void TEventRegBase::GetEventConceptTFIDF(const uint64& EventId, const EnumConceptType& ConceptTypes, THash<TUInt64, TFlt>& ConceptIdWgtH) const
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	ConceptIdWgtH.Clr();
	TUInt64IntKdV ConceptIdWgtV; GetEventConcepts(EventId, ConceptTypes, ConceptIdWgtV, false);
	for (int N = 0; N < ConceptIdWgtV.Len(); N++) {
		const uint64 ConceptId = ConceptIdWgtV[N].Key;
		const uint64 ConcFq = NewsBase->NewsCounts->GetConceptFq(ConceptId);	// DF
		const uint64 Arts = NewsBase->NewsCounts->GetAnnArticleCount();			// Docs
		const double IDF = double(MAX(1, Arts)) / MAX(1, ConcFq);
		ConceptIdWgtH.AddDat(ConceptId, ConceptIdWgtV[N].Dat * log(IDF));
	}
}

// parse date in format YYYY-MM-DD into TTm
// return false if format is not correct
bool TEventRegBase::ParseDate(const TStr& DateStr, TTm& Date) const
{
	if (DateStr.Len() < 10 || DateStr[4] != '-' || DateStr[7] != '-')
		return false;
	TStr YearStr = DateStr.GetSubStr(0, 3);
	TStr MonthStr = DateStr.GetSubStr(5, 6);
	TStr DayStr = DateStr.GetSubStr(8, 9);
	if (!YearStr.IsInt() || !MonthStr.IsInt() || !DayStr.IsInt())
		return false;
	Date = TTm(YearStr.GetInt(), MonthStr.GetInt(), DayStr.GetInt());
	return true;
}

// compute an estimate of story quality. use features such as AvgCosToCentroid, Variance, ArticleCount, ...
// the score is used in situations such as when multiple stories in the same language exist for an event and we have to 
// choose which story info to use for event (see TEventRegSrvUtil::Event)
// output in range [1, 100]
// computation: 5 * sqrt(ArticleCount) * AvgCosToCentroid / Variance
// TODO: find a better function for estimating the story quality
int TEventRegBase::ComputeStoryQuality(const uint64& StoryId) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const double AvgCosToCentroid = StoryStore->GetFieldNmFlt(StoryId, "AvgCosToCentroid");    // range [0,1]
	const double Variance = StoryStore->GetFieldNmFlt(StoryId, "Variance");
	const int ArticleCount = StoryStore->GetFieldNmInt(StoryId, "ArticleCount");

	// a really poor-mans-score of the story
	/*int Score = MAX(1, (int) (ArticleCount * AvgCosToCentroid / MAX(0.1, Variance)));*/
	//double Score = sqrt(ArticleCount) * AvgCosToCentroid / MAX(0.1, Variance);
	if (Variance > 0.7)
		return 1;
	double VarianceInv = MAX(0, 1.0 - Variance);
	const double ArticleScore = MAX(1, MIN((double) ArticleCount, 100.0));
	double Score = ArticleScore * AvgCosToCentroid * VarianceInv;
	
	return MIN(100, MAX((int) Score, 1));
}

// go over all articles in the story, take the extracted dates and populate them in CombinedDateColl
bool TEventRegBase::ComputeStoryDates(const uint64& StoryId, TExtractedDateColl& CombinedDateColl, TTm& AverageTime) const
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	// go over all articles in story and extract from them dates and compute average date
	PRecSet ArticleRecSet = StoryStore->GetRec(StoryId).DoJoin(Base, "hasArticle");
	const int TotArticleCount = ArticleRecSet->GetRecs();
	uint64 TmSum = 0;
	for (int N = 0; N < TotArticleCount; N++) {
		uint64 ArticleRecId = ArticleRecSet->GetRecId(N);
		TTm Tm; ArticleStore->GetFieldNmTm(ArticleRecId, "DateTime", Tm);
		TStr Date = Tm.GetWebLogDateStr();
		TmSum += (uint) Tm.GetSecTm();
		if (ArticleStore->IsFieldNmNull(ArticleRecId, "ExtractedDates"))
			continue;
		TStr DateCollStr = ArticleStore->GetFieldNmStr(ArticleRecId, "ExtractedDates");
		TExtractedDateColl ArticleDateColl(DateCollStr);
		// remove the dates that are mentioned too far down in the text
		ArticleDateColl.FilterDatesByPosInText(0, Settings.DateMaxPosInTextToUse);
		CombinedDateColl.AddColl(ArticleDateColl);
	}
	// where we don't know the year of the event try to impute it from other date instances
	const int MinFreq = (int) (Settings.DateImputationMinValidYearCnt + Settings.DateImputationMinValidYearSlope * TotArticleCount);
	CombinedDateColl.ImputeMissingYears(MinFreq, (Settings.VerbosityLevel > 0 && InfoLogFile != NULL && InfoLogFile->NotifyInstV.Len() > 0) ? InfoLogFile->NotifyInstV[0] : NULL);
	// find the most popular date or range
	CombinedDateColl.SortByFreq(false);
	if (TotArticleCount > 0) {
		TSecTm AvgSecTm = TSecTm((uint) (TmSum / TotArticleCount));
		AverageTime = TTm(AvgSecTm);
	}
	return TotArticleCount > 0;
}

// return SimStoryInfo data about cca similarity between two stories
void TEventRegBase::ComputeCCASimilarityBetweenStories(const uint64& StoryRecId1, const uint64& StoryRecId2, TSimStoryInfo& SimStoryInfo, const double& MinStory1ArticleScore) const
{
	EAssertR(StoryStore->IsRecId(StoryRecId1), TStr::Fmt("Story id %I64u is invalid", StoryRecId1));
	EAssertR(StoryStore->IsRecId(StoryRecId2), TStr::Fmt("Story id %I64u is invalid", StoryRecId2));
		
	THash<TUInt64, TSimStoryInfo> SimStoryH;
	ComputeSimilarStoriesUsingCCA(StoryRecId1, SimStoryH, false, false, MinStory1ArticleScore);
	if (SimStoryH.IsKey(StoryRecId2))
		SimStoryInfo = SimStoryH.GetDat(StoryRecId2);
}

// compute similar stories to StoryRecId based on CCA. for each similar story compute 
void TEventRegBase::ComputeSimilarStoriesUsingCCA(const uint64& StoryRecId, THash<TUInt64, TSimStoryInfo>& SimStoryH, const bool& IgnoreStoriesBelowThreshold, const bool& IgnoreStoriesWithLargeVariance, const double& MinStoryArticleCosSim) const
{
	EAssertR(StoryStore->IsRecId(StoryRecId), TStr::Fmt("Story id %I64u is invalid", StoryRecId)); 

	// if the story has too high variance then don't do any merging
	if (IgnoreStoriesWithLargeVariance && GetStoryVariance(StoryRecId) > GetStoryAvgCosToCentroid(StoryRecId))
		return;
	
	// take all articles belonging to StoryRecId
	PRecSet PrimaryArticles = StoryStore->GetRec(StoryRecId).DoJoin(Base, "hasArticle");
	if (MinStoryArticleCosSim > EPS)
		PrimaryArticles->SortByField(false, NewsBase->ArticleStore->GetFieldId("CosToCentroid"));
	int PrimaryArticlesCnt = MIN(PrimaryArticles->GetRecs(), 300);		// some (bad) stories have several thousand articles. in such cases we want to stop soon enough
	TStrIntH LangToMaxCntH;
	for (int ArtN = 0; ArtN < PrimaryArticlesCnt; ArtN++) {
		const TRec ArtRec = PrimaryArticles->GetRec(ArtN);
		if (MinStoryArticleCosSim > EPS && ArticleStore->GetFieldNmFlt(ArtRec.GetRecId(), "CosToCentroid") < MinStoryArticleCosSim)
			break;

		// for all joins to similar articles independently for each language
		for (int LangN =0; LangN < NewsConst.LangSimilarV.Len(); LangN++) {
			const TStr Lang = NewsConst.LangSimilarV[LangN];
			const TStr JoinNm = "similar_" + Lang;
			// CCA often finds as similar articles a group of articles that are duplicates. when we resolve articles to their first article we would count the same article multiple times 
			// which would give greater evidence for similarity than it should be. therefore we remember each original article that we see and count it only once
			TUInt64Set SeenArticlesH;
			if (ArticleStore->IsJoinNm(JoinNm)) {
				PRecSet SimArtRecSet = ArtRec.DoJoin(Base, JoinNm);
				const int SimArtCnt = SimArtRecSet->GetRecs();
				for (int SimArtN = 0; SimArtN < SimArtCnt; SimArtN++) {
					uint64 SimArtId = SimArtRecSet->GetRecId(SimArtN);
					if (!ArticleStore->IsRecId(SimArtId))	// it's possible that we've already deleted the similar article
						continue;
					const int SimWgt = SimArtRecSet->GetRecFq(SimArtN);
					const uint64 OrigArtId = ArticleStore->GetFieldNmUInt64(SimArtId, "isDuplicateOfId");
					
					if (OrigArtId != TUInt64::Mx && !ArticleStore->IsRecId(OrigArtId))
						continue;
					// if article is a duplicate then use the original article id, otherwise use the sim. art. id
					if (OrigArtId != TUInt64::Mx)
						SimArtId = OrigArtId;
					if (SeenArticlesH.IsKey(SimArtId))
						continue;
					SeenArticlesH.AddKey(SimArtId);
					//
					// if we came to here then the SimArtId is a valid similar article of article ArtRec
					
					// increase the max count of matching articles 
					LangToMaxCntH.AddDat(Lang) += 1;
					// find to which story does the similar article belong to and add the story with the similarity weight to the SimStoryH
					const uint64 SimStoryRecId = ArticleStore->GetFieldNmUInt64(SimArtId, "hasStoryId");
					if (SimStoryRecId == TUInt64::Mx)
						continue;
					TSimStoryInfo& Info = SimStoryH.AddDat(SimStoryRecId);
					if (GetStoryLang(SimStoryRecId) != Lang) {
						TChA Report;
						Report += TStr::Fmt("Trying to find similar stories to story %I64u (%s). Story is in lang %s.\n", StoryRecId, StoryStore->GetRecNm(StoryRecId).CStr(), GetStoryLang(SimStoryRecId).CStr());
						Report += TStr::Fmt("We were checking the article %I64u that belongs to the story. Article is in lang %s\n", ArtRec.GetRecId(), ArticleStore->GetFieldNmStr(ArtRec.GetRecId(), "Language").CStr());
						Report += TStr::Fmt("For this article, there are %d similar articles in language %s\n", SimArtCnt, Lang.CStr());
						Report += TStr::Fmt("One of the similar articles is %I64u which is in language %s\n", SimArtId, ArticleStore->GetFieldNmStr(SimArtId, "Language").CStr());
						if (OrigArtId != TUInt64::Mx)
							Report += TStr::Fmt("The similar article is a duplicate. It's original article is %I64u, which is in lang %s\n", OrigArtId, ArticleStore->GetFieldNmStr(OrigArtId, "Language").CStr());
						Report += TStr::Fmt("There has to be a mistake. Since we were looking for similar articles in %s lang, we didn't expect to find the article to be in %s", Lang.CStr(), GetStoryLang(SimStoryRecId).CStr());
						ErrorLogFile->NotifyErr("TEventRegBase::ComputeSimilarStoriesUsingCCA. %s", Report.CStr());
					}
					Info.ArtMatchCount += 1;
					Info.SumWgts += SimWgt;
				}
			}
		}
	}
	for (int KeyId = SimStoryH.FFirstKeyId(); SimStoryH.FNextKeyId(KeyId); ) {
		const uint64 StoryId = SimStoryH.GetKey(KeyId);
		TSimStoryInfo& Info = SimStoryH.GetDat(StoryId);
		TStr StoryLang = GetStoryLang(StoryId);
		// immediately remove similar stories that have very few similar articles
		if (IgnoreStoriesBelowThreshold && Info.ArtMatchCount < Settings.StoryMergingMinMatchArts) {
			SimStoryH.DelKeyId(KeyId);
			continue;
		}
		if (IgnoreStoriesWithLargeVariance && GetStoryVariance(StoryId) > GetStoryAvgCosToCentroid(StoryId)) {
			SimStoryH.DelKeyId(KeyId);
			continue;
		}
		// store the max number of articles that we could expect as a match between stories
		// TODO: here we don't take into account that StoryId could be a huge story with many articles. to make this number correct we would have 
		// to compute also the similarity of StoryId->StoryRecId, count MaxArtCount in that case and add it to the bottom number
		Info.MaxArtCount = LangToMaxCntH.GetDat(StoryLang);
	}

	SimStoryH.Defrag();		// we might have removed some keys from the dict
}

// decide if the story TestStoryId should be merged with stories in MergedStoryV.
// use various criteria to decide. Among others use the SimStoryH data which contains CCA similarity of TestStoryId with other stories
bool TEventRegBase::ShouldStoriesMerge(const uint64& MergedStoryId, const uint64& TestStoryId, const TSimStoryInfo& TestSimStoryInfo, TMergeEval& MergeEval, const bool& FullEval)
{
	EAssertR(StoryStore->IsRecId(MergedStoryId), TStr::Fmt("Story id %I64u is invalid", MergedStoryId));
	EAssertR(StoryStore->IsRecId(TestStoryId), TStr::Fmt("Story id %I64u is invalid", TestStoryId));

	if (Settings.UseModelForStoryMerging) {
		double MergeProb;
		const TStr MergedStoryUri = StoryStore->GetRecNm(MergedStoryId);
		const TStr TestStoryUri = StoryStore->GetRecNm(TestStoryId);
		bool ModelOk = StoryLinking->ShouldStoriesMerge(this, MergedStoryUri, TestStoryUri, MergeProb);
		if (ModelOk)
			return MergeProb >= 0.5;
		// if model was not ok then use the default method for deciding to merge or not
		ErrorLogFile->NotifyErr("TEventRegBase::ShouldStoriesMerge. Warning: SVM was supposed to be used for deciding on story merging but model was not available");
	}

	MergeEval.ShouldMerge = true;

	// compare if TestStoryId and MergedStoryId are about the same event
	// comparison should be done so that it doesn't matter if we compare (Test,Merged) or (Merged,Test)
	// in other words, if we find that stories A,B and C should be merged, then it shouldn't matter in which order we try to merge them
	// A,B,C or B,A,C, ....
	MergeEval.CCAArtMatchCount = TestSimStoryInfo.ArtMatchCount;
	MergeEval.CCAAvgWgt = MergeEval.CCAArtMatchCount > 0 ? (int) (TestSimStoryInfo.SumWgts / float(MergeEval.CCAArtMatchCount)) : 0;
	MergeEval.CCAMaxMatchCount = TestSimStoryInfo.MaxArtCount;		// this is the maximum number of matched articles that we could get 

	MergeEval.CCAArticleTreshold = MAX(Settings.StoryMergingMinMatchArts, int(Settings.StoryMergingSlope * MergeEval.CCAMaxMatchCount));
	// if not enough articles are matched then don't merge stories
	if (MergeEval.CCAArtMatchCount < MergeEval.CCAArticleTreshold)
		MergeEval.ShouldMerge = false;
	if (MergeEval.ShouldMerge == false && !FullEval) return false;

	// we log only if at least the article threshold condition is met
	if (InfoLogFile->GetVerbosityLevel() >= 4)
		InfoLogFile->NotifyVerboseFmt(4, "Testing story '%I64u', fixed story '%I64u'. Article match count = %d (>= %d). Max match count = %d. Average wgt = %d (>= %d). Concept similarity = %.1f%% (>= %d%%).",
		TestStoryId, MergedStoryId, MergeEval.CCAArtMatchCount, MergeEval.CCAArticleTreshold, MergeEval.CCAMaxMatchCount, MergeEval.CCAAvgWgt, Settings.StoryMergingMinAvgSim, 100 * GetConceptSimStories(TestStoryId, MergedStoryId), Settings.StoryMergingMinBowSim);
	// average similarity of articles has to be above StoryMergingSim (in %)
	if (MergeEval.CCAAvgWgt < Settings.StoryMergingMinAvgSim)
		MergeEval.ShouldMerge = false;
	if (MergeEval.ShouldMerge == false && !FullEval) return false;

	// compare concept similarity only if enough articles are annotated
	int AnnotatedArts1, TotalArts1, AnnotatedArts2, TotalArts2;
	GetStoryAnnArticles(MergedStoryId, AnnotatedArts1, TotalArts1);
	GetStoryAnnArticles(TestStoryId, AnnotatedArts2, TotalArts2);
	MergeEval.AnnArticles1 = AnnotatedArts1; MergeEval.AnnArticles2 = AnnotatedArts2;
	MergeEval.TotalArts1 = TotalArts1; MergeEval.TotalArts2 = TotalArts2;
	MergeEval.ConceptSim = int(100 * GetConceptSimStories(TestStoryId, MergedStoryId));
	if (AnnotatedArts1 > Settings.StoryMergingMinAnnArtsToUse && AnnotatedArts2 > Settings.StoryMergingMinAnnArtsToUse) {
		// concept similarity on stories has to be above StoryMergingMinBowSim (100*cos sim)
		if (MergeEval.ConceptSim < Settings.StoryMergingMinBowSim)
			MergeEval.ShouldMerge = false;
		if (MergeEval.ShouldMerge == false && !FullEval) return false;
	}

	// check the time difference between the average dates in the stories
	double AvgTimeHrs1, AvgTimeHrs2, StDevTime1, StDevTime2;
	double CoreAvgTimeHrs1, CoreAvgTimeHrs2, CoreStDevTime1, CoreStDevTime2;
	bool ok1 = GetStoryTimes(MergedStoryId, AvgTimeHrs1, StDevTime1, CoreAvgTimeHrs1, CoreStDevTime1);
	bool ok2 = GetStoryTimes(TestStoryId, AvgTimeHrs2, StDevTime2, CoreAvgTimeHrs2, CoreStDevTime2);
	MergeEval.CoreAvgTimeHrs1 = CoreAvgTimeHrs1; MergeEval.CoreAvgTimeHrs2 = CoreAvgTimeHrs2;
	if (ok1 && ok2) {
		if (abs(CoreAvgTimeHrs1 - CoreAvgTimeHrs2) > Settings.StoryMergingMaxStoryDiffHrs)
			MergeEval.ShouldMerge = false;
	}
	else
		MergeEval.ShouldMerge = false;
	if (MergeEval.ShouldMerge == false && !FullEval) return false;

	// TODO: add other checks here. check the story category, ...

	// TODO: add a method for collecting learning data with story features and then use svm to find the most appropriate model
	if (MergeEval.ShouldMerge == true)
		InfoLogFile->NotifyVerboseFmt(4, "Test story '%I64u' (%s) is similar enough and will be merged with story '%I64u'.", TestStoryId, GetStoryTitle(TestStoryId).CStr(), MergedStoryId);
	return MergeEval.ShouldMerge;
}

void TEventRegBase::SetLocationForStory(const uint64& StoryId, const uint64& LocConceptId, const int& Wgt)
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const TStr LocUri = ConceptStore->GetRecNm(LocConceptId);
	InfoLogFile->NotifyVerboseFmt(2, "Assigning city for story '%I64u' to location %s with wgt %d", StoryId, LocUri.CStr(), Wgt);

	StoryStore->AddJoin("hasLocation", StoryId, LocConceptId, Wgt);
	// if location is a city then also add joins to the country where the city is
	if (NewsBase->GeoBase->IsCity(LocUri)) {
		const TStr CountryRecNm = NewsBase->GeoBase->GetCountryForCity(LocUri);
		if (ConceptStore->IsRecNm(CountryRecNm)) {
			const uint64 CountryRecId = ConceptStore->GetRecId(CountryRecNm);
			// only add join from concept to story so that we can support search by country
			Index->IndexJoin(ConceptStore, "hasStoryForLocation", CountryRecId, StoryId, Wgt);
		}
	}
}

void TEventRegBase::SetLocationForEvent(const uint64& EventId, const uint64& LocConceptId, const int& Wgt)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	const TStr LocUri = ConceptStore->GetRecNm(LocConceptId);
	InfoLogFile->NotifyVerboseFmt(2, "Assigning location for event '%I64u' to : %s", EventId, LocUri.CStr());

	EventStore->AddJoin("hasLocation", EventId, LocConceptId, Wgt);
	// if location is a city then also add joins to the country where the city is
	if (NewsBase->GeoBase->IsCity(LocUri)) {
		const TStr CountryRecNm = NewsBase->GeoBase->GetCountryForCity(LocUri);
		if (ConceptStore->IsRecNm(CountryRecNm)) {
			const uint64 CountryRecId = ConceptStore->GetRecId(CountryRecNm);
			// only add join from concept to event so that we can support search by country
			Index->IndexJoin(ConceptStore, "hasEventForLocation", CountryRecId, EventId, Wgt);
		}
	}
}

void TEventRegBase::RemoveLocationFromStory(const uint64& StoryId) 
{
	EAssertR(StoryStore->IsRecId(StoryId), TStr::Fmt("Story id %I64u is invalid", StoryId));

	const TRec StoryRec(StoryStore, StoryId);
	
	// remove city or country from location
	PRecSet ExistingLocations = StoryRec.DoJoin(Base, "hasLocation");
	for (int N = 0; N < ExistingLocations->GetRecs(); N++) {
		const uint64 LocationRecId = ExistingLocations->GetRecId(N);
		StoryStore->DelJoin("hasLocation", StoryId, LocationRecId);
		const TStr LocationUri = ConceptStore->GetRecNm(LocationRecId);
		// if location was a city then remove also the IndexJoin from country concept to story
		if (NewsBase->GeoBase->IsCity(LocationUri)) {
			const TStr CountryWikiUri = NewsBase->GeoBase->GetCountryForCity(LocationUri);
			if (ConceptStore->IsRecNm(CountryWikiUri))
				Index->DeleteJoin(ConceptStore, "hasStoryForLocation", ConceptStore->GetRecId(CountryWikiUri), StoryId);
		}
	}
}

void TEventRegBase::RemoveLocationFromEvent(const uint64& EventId)
{
	EAssertR(EventStore->IsRecId(EventId), TStr::Fmt("Event id %I64u is invalid", EventId));

	const TRec EventRec(EventStore, EventId);

	// remove city or country from location
	PRecSet ExistingLocations = EventRec.DoJoin(Base, "hasLocation");
	for (int N = 0; N < ExistingLocations->GetRecs(); N++) {
		const uint64 LocationRecId = ExistingLocations->GetRecId(N);
		EventStore->DelJoin("hasLocation", EventId, LocationRecId);
		const TStr LocationUri = ConceptStore->GetRecNm(LocationRecId);
		// if location was a city then remove also the IndexJoin from country concept to event
		if (NewsBase->GeoBase->IsCity(LocationUri)) {
			const TStr CountryWikiUri = NewsBase->GeoBase->GetCountryForCity(LocationUri);
			if (ConceptStore->IsRecNm(CountryWikiUri))
				Index->DeleteJoin(ConceptStore, "hasEventForLocation", ConceptStore->GetRecId(CountryWikiUri), EventId);
		}
	}
}

// ////////////////////////////
// TMergeEval
PJsonVal TMergeEval::GetJson(const uint64& StoryId1, const uint64& StoryId2, const PEventRegBase& EventRegBase) const 
{
	const TEventSettings Settings = EventRegBase->GetSettings();
	PJsonVal Json = TJsonVal::NewObj("story1Id", TJsonVal::NewNum(StoryId1));
	Json->AddToObj("story2Id", TJsonVal::NewNum(StoryId2));
	Json->AddToObj("story1Title", EventRegBase->GetStoryTitle(StoryId1));
	Json->AddToObj("story2Title", EventRegBase->GetStoryTitle(StoryId2));

	Json->AddToObj("artMatch", CCAArtMatchCount);
	Json->AddToObj("artMatchThresh", CCAArticleTreshold);

	Json->AddToObj("avgWeight", CCAAvgWgt);
	Json->AddToObj("avgWeightThresh", Settings.StoryMergingMinAvgSim);

	Json->AddToObj("maxArtCount", CCAMaxMatchCount);

	Json->AddToObj("conceptSim", ConceptSim);
	Json->AddToObj("conceptSimThresh", Settings.StoryMergingMinBowSim);

	Json->AddToObj("annArts1", AnnArticles1);
	Json->AddToObj("annArts2", AnnArticles2);
	Json->AddToObj("totalArts1", TotalArts1);
	Json->AddToObj("totalArts2", TotalArts2);
	Json->AddToObj("minAnnArtsToUse", Settings.StoryMergingMinAnnArtsToUse);

	Json->AddToObj("coreTimeDiff", int(abs(CoreAvgTimeHrs1 - CoreAvgTimeHrs2)));
	Json->AddToObj("coreTimeDiffThresh", Settings.StoryMergingMaxStoryDiffHrs);

	Json->AddToObj("shouldMerge", ShouldMerge);

	return Json;
}