#ifndef EVENTREGBASE_H
#define EVENTREGBASE_H

#include <qminer.h>
#include <logger.h>
#include "../NewsMiner/NewsBase.h"
#include "../NewsMiner/NewsSearch.h"
#include "../Utils/DateExtractor.h"
#include "../Srv/CategorySrv.h"
#include "../Learning/LrnStoryLinking.h"
#include "../Learning/LrnStoryLocation.h"
#include "CategoryReq.h"

using namespace TQm;

class TEventSettings
{
public:
	TEventSettings();
	TEventSettings(const PJsonVal& Settings);
	void Init(const PJsonVal& Settings);

	// the number of articles that need to change in a story before we trigger an update
	int RecomputeStoryFeaturesEvery;

	int StoryMergingMinMatchArts;
	double StoryMergingSlope;
	int StoryMergingMinAvgSim;			// minimum average similarity of matched articles as computed by CCA
	int StoryMergingMinBowSim;			// minimum BOW similarity on concepts of the stories
	int StoryMergingMinAnnArtsToUse;	// minimum number of articles that have to be annotated in a story to use the concept similarity as a criterion when merging the stories
	int StoryMergingMaxStoryDiffHrs;	// maximum number of hours that the average dates of the stories are apart
	double StoryMergingMinArticleCosSim;	// minimum cosine similarity of articles in the story to consider them when checking other stories to merge

	int StoryPropTopConceptsToTake;
	int StoryPropMinConceptWgtToTake;
	int StoryPropMinCategoryWgtToTake;  // min cos similarity (between 0-100) between a category and a story to create a join between them
	int StoryPropMinLocationWgtToTake;	// min weight (between 0-100) the location should have compared to other concepts to take it as story location
	int StoryPropMinLocationPrcToTake;	// min percent of articles that should have the location in order to use it for the story location
	int StoryPropMinLocationPropToTake;	// min percent of all weights for cities should be assigned to one city

	int EventPropTopConceptsToTake;		// number of concepts to take when assigning to concepts to events
	int EventPropMinConceptWgtToTake;	// concepts with total weight below this threshold won't be assigned to an event

	// params for computing similar events and stories
	int SimilarEventsNrTopConceptsUsed;
	int SimilarEventsNrTopEventsPerConcept;
	int SimilarEventsMaxEventsToEval;			// when having a set of event candidates, what number of events do we want to evaluate in details using GetConceptSim()
	int SimilarStoriesNrTopConceptsUsed;
	int SimilarStoriesNrTopStoriesPerConcept;
	int SimilarStoriesMaxStoriesToEval;			// when having a set of story candidates, what number of stories do we want to evaluate in details using GetConceptSim()

	bool UseModelForStoryMerging;	// use SVM model to merge stories. the model has to have learning data to be used
	bool UseModelForStoryLocation; // use SVM model to determine the story location. the model has to have learning data to be used
	bool ComputeCategories;			// do we need to compute categories for stories or not. used to make story updates faster

	int DateMaxPosInTextToUse;					// what is the maximum position of the date in the text in order for the date to be used when determining story date
	int DateMinStoryDateFreqCnt;				// min number of times the date has to be found in the story articles in order to use it as the ExtractedDate field value
	double DateMinStoryDateFreqSlope;			// if > 0 then use DateMinStoryDateFreqCnt +  DateMinStoryDateFreqSlope* article count as the formula to decide if an extracted date should be used as the ExtractedDate value
	int DateImputationMinValidYearCnt;			// min number of times a date with a valid year has to occur in order to consider it when imputing dates with missing year (to go from -02-12 to 2012-02-12)
	double DateImputationMinValidYearSlope;		// if > 0 then use DateImputationMinValidYearCnt +  DateImputationMinValidYearSlope* article count as the formula to decide if a date is a valid candidate for incomplete dates to be imputed

	int EventSampleSize;						// the event subset size to use when computing different things on sets of events; -1 for all (no sampling)
	int VerbosityLevel;
};

// helper class used to represent information about CCA similarity between two stories
class TSimStoryInfo
{
public:
	TInt ArtMatchCount;
	TFlt SumWgts;		// sum of weights from original story to this story. to find average weight compute SumSimWgts / MatchedArticles
	TInt MaxArtCount;		// max(articles in story 1, articles in story 2) - used to know what would be the maximum possible MatchedArticles

	// so that we can use the class in THash
	bool operator < (const TSimStoryInfo& Info) const { return ArtMatchCount < Info.ArtMatchCount; }
};

class TStoryLocationsInfo
{
public:
	THash<TUInt64, TFlt> ConceptIdToWgtH;
	THash<TUInt64, TInt> ConceptIdToCntH;
	THash<TUInt64, TFlt> ExplicitConceptIdToWgtH;
	THash<TUInt64, TInt> ExplicitConceptIdToCntH;
	
	TUInt64Set CandidateLocationH;		// set of all locations identified in the articles in the story
	TFlt SumCosSim;						// sum of cossim of all articles in story
	TFlt SumCosSimForExplicitArticles;	// sum of cossim of articles that have explicit location in the story
	TFlt SumConceptWgts;				// sum of all concept weights (=cossim) in all story articles - each article can have multiple locations so each article's cossim can be counted multiple times
	
	TInt Articles;
};

class TMergeEval {
public:
	TInt CCAArticleTreshold;
	TInt CCAArtMatchCount;
	TInt CCAMaxMatchCount;
	TInt CCAAvgWgt;
	TInt ConceptSim;
	TInt AnnArticles1, AnnArticles2, TotalArts1, TotalArts2;
	TFlt CoreAvgTimeHrs1, CoreAvgTimeHrs2;

	bool ShouldMerge;

	PJsonVal GetJson(const uint64& StoryId1, const uint64& StoryId2, const PEventRegBase& EventRegBase) const;
};


//typedef TTriple<TInt, TFlt, TInt> TSimStoryInfo;	// number of matched articles, sum of wgts, max possible count

//////////////////////////////////////

class TEventRegBase{
public:
  friend class TPt<TEventRegBase>;

private:
	TCRef CRef;
	TFAccess FAccess;
	TStr DbsPath;

	// index to base
	PIndex Index;
	PIndexVoc IndexVoc;

	TUInt64 NextEventId;
	PCategoryReq CategoryReq;

	TEventSettings Settings;
	PLogger InfoLogFile;
	PLogger ActivityLogFile;
	PLogger ErrorLogFile;
	
	// for loading existing index and stores for update or read-only
	TEventRegBase(const PNewsBase& _NewsBase, const TFAccess& FAccess, const PNewsSearch& _NewsSearch, const PCategoryReq& _CategoryReq, const TStr& _DbsPath, const TStr& SettingsFName);

public:
	enum EnumSimType { stCos, stJaccard };

	TStr SettingsFName;
	
	PNewsBase NewsBase;
	PNewsSearch NewsSearch;

	PStore ArticleStore;
	PStore ConceptStore;
	PStore CategoryStore;
	PStore StoryStore;
	PStore SourceStore;
	PStore EventStore;
	PStore ImageStore;
	PBase Base;

	uchar ArticleStoreId;
	uchar ConceptStoreId;
	uchar CategoryStoreId;
	uchar StoryStoreId;
	uchar SourceStoreId;
	uchar EventStoreId;
	uchar ImageStoreId;
	
	// cross-lingual story linking
	PStoryLinking StoryLinking;
	// determining story location
	PStoryLocation StoryLocation;

	~TEventRegBase();
	// for creating new index and stores or updating existing
	static PEventRegBase New(const PNewsBase& NewsBase, const PNewsSearch NewsSearch, const PCategoryReq& CategoryReq, const TStr& DbsPath, const TStr& SettingsFName) {
		return new TEventRegBase(NewsBase, TIndex::Exists(NewsBase->Base->GetFPath()) ? faUpdate : faCreate, NewsSearch, CategoryReq, DbsPath, SettingsFName);
	}

	/*void OnAddArticleToStory(const uint64& ArticleId, const uint64& StoryId);
	void OnAddArticlesToStory(const TUInt64V& ArticleIdV, const uint64& StoryId);*/
	void OnRemoveArticleFromStory(const uint64& OldStoryId);
	/*void OnMedoidUpdate(const uint64& StoryId);
	void OnStoryFinished(const uint64& StoryId);*/
	//void OnRemoveArticlesFromStory(const TUInt64V& ArticleIdV, const TUInt64& OldStoryId);
	
	//void RecomputeEventProps(const uint64& EventRecId);

	void ConsiderUpdatingStoryInfo(const uint64& StoryId, const bool& ForceUpdate = false);
	void UpdateStoryProperties(const uint64& StoryId, const bool& ArticleCountP = true, const bool& TitleSummaryP = true, const bool& DatesP = true, const bool& ConceptsLocationP = true, const bool& CategoryP = true, const bool& QualityP = true, const bool& LastUpdateP = true);
	void UpdateEventProperties(const uint64& EventId, const bool& ArticleCountP = true, const bool& ConceptsP = true, const bool& LocationP = true, const bool& DateP = true, const bool& CategoriesP = true);

	void UpdateStoryDates(const uint64& StoryId);
	void UpdateStoryConcepts(const uint64& StoryId);
	void UpdateStoryLocation(const uint64& StoryId);
	void UpdateStoryCategories(const uint64& StoryId, const bool& ForceUpdate = false);
	void UpdateStoryTitleAndSummary(const uint64& StoryId);
	int  UpdateStoryArticleCount(const uint64& StoryId);
	void UpdateStoryCategoriesCallback(const PJsonVal& CategoryData);
	void UpdateStoryQuality(const uint64& StoryId);
	void UpdateStoryLastUpdate(const uint64& StoryId);

	void UpdateEventStories(const uint64& StoryId);
	void UpdateEventConcepts(const uint64& EventId);
	void UpdateEventDate(const uint64& EventId);
	void UpdateEventCategories(const uint64& EventId);
	void UpdateEventLocation(const uint64& EventId);
	int  UpdateEventArticleCount(const uint64& EventId);

	void MergeStories(const TUInt64V& StoryIdV, const bool& RemoveExistingTargetStories = true, const bool& UpdateEventInfo = true);
	void MergeStories(const uint64& TargetEventId, const TUInt64V& StoryIdV, const bool& UpdateEventInfo = true);
	void AssignStoryToEvent(const uint64& StoryId, const uint64& EventId, const bool& UpdateOldEventInfo = true, const bool& UpdateNewEventInfo = true);
	void SplitStories(const TUInt64V& StoryIdV);
		
	// story merging
	bool ShouldStoriesMerge(const uint64& StoryId, const uint64& TestStoryId, const TSimStoryInfo& TestSimStoryInfo, TMergeEval& MergeEval, const bool& FullEval = false);

	void GetConceptsForEvents(const PRecSet& EventSet, const EnumConceptType& ValidTypes, const bool& NormalizeByEventCount, THash<TUInt64, TFlt>& ConceptIdWgt);
	
	/*PRecSet FindEventsForConcepts(const TStrV& ConceptUriV);
	PRecSet FindEventsForCoverage(const int& MinArtCount, const int& MaxArtCount = TInt::Mx);*/

	// helper methods
	uint64 GetValidEventId(const uint64& EventId) const;
	void GetEventDate(const uint64& EventId, TStr& DateStart, TStr& DateEnd) const;
	bool GetEventDate(const uint64& StoryId, TTm& StoryDate) const;
	void GetEventStories(const uint64& EventId, TUInt64V& StoryIdV, const bool& SortByQualityP = false) const;
	int GetEventArticleCount(const uint64& EventId) const { return EventStore->IsRecId(EventId) ? EventStore->GetFieldNmInt(EventId, "ArticleCount") : 0; }
	double GetEventAvgCosToCentroid(const uint64& EventId) const;
	uint64 GetEventPrimaryStoryId(const uint64& EventId) const;
	void GetStoryDate(const uint64& StoryId, TStr& DateStart, TStr& DateEnd) const;
	bool GetStoryDate(const uint64& StoryId, TTm& EventDate) const;
	TStr GetStoryTitle(const uint64& StoryId) const;
	TStr GetStoryBody(const uint64& StoryId, const int& MaxLen = 200) const;	// get the medoid article and return its content (maybe shortened)
	TStr GetStoryLang(const uint64& StoryId) const;
	void GetStoryAnnArticles(const uint64& StoryId, int& AnnotatedArts, int& TotalArts) const;
	bool GetStoryTimes(const uint64& StoryId, double& AvgTime, double& StDevTime, double& CoreAvgTime, double& CoreStDevTime) const;
	uint64 GetStoryEvent(const uint64& StoryId) const;
	uint64 GetStoryPrimaryEvent(const uint64& StoryId) const;
	int GetStoryArticleCount(const uint64& StoryId) const;
	double GetStoryAvgCosToCentroid(const uint64& StoryId) const;
	double GetStoryVariance(const uint64& StoryId) const;
	void GetStoryLocationsInfo(const uint64& StoryId, TStoryLocationsInfo& StoryLocationsInfo) const;
	bool ParseDate(const TStr& DateStr, TTm& Date) const;
	
	// get common dates in the story
	bool ComputeStoryDates(const uint64& StoryId, TExtractedDateColl& CombinedDateColl, TTm& AverageTime) const;
	int  ComputeStoryQuality(const uint64& StoryId) const;
	void ComputeCCASimilarityBetweenStories(const uint64& StoryRecId1, const uint64& StoryRecId2, TSimStoryInfo& SimStoryInfo, const double& MinStory1ArticleScore = 0) const;
	void ComputeSimilarStoriesUsingCCA(const uint64& StoryId, THash<TUInt64, TSimStoryInfo>& SimStoryH, const bool& IgnoreStoriesBelowThreshold = true, const bool& IgnoreStoriesWithLargeVariance = true, const double& MinStoryArticleCosSim = 0) const;
	
	void GetStoryConcepts(const uint64& StoryId, const EnumConceptType& ConceptTypes, TUInt64IntKdV& ConceptIdWgtV, const bool& SortP = true) const;
	void GetEventConcepts(const uint64& EventId, const EnumConceptType& ConceptTypes, TUInt64IntKdV& ConceptIdWgtV, const bool& SortP = true) const;
	void GetStoryConceptTFIDF(const uint64& StoryId, const EnumConceptType& ConceptTypes, THash<TUInt64, TFlt>& ConceptIdWgtH) const;
	void GetEventConceptTFIDF(const uint64& EventId, const EnumConceptType& ConceptTypes, THash<TUInt64, TFlt>& ConceptIdWgtH) const;
	double GetConceptSim(const THash<TUInt64, TFlt>& ConceptIdWgtH1, const THash<TUInt64, TFlt>& ConceptIdWgtH2, const EnumSimType& SimType = stCos) const;
	double GetConceptSimEvents(const uint64& EventId1, const uint64& EventId2, const EnumSimType& SimType = stCos, const EnumConceptType& ConceptTypes = ctConcepts) const;
	double GetConceptSimStories(const uint64& StoryId1, const uint64& StoryId2, const EnumSimType& SimType = stCos, const EnumConceptType& ConceptTypes = ctConcepts) const;

	void SetLocationForStory(const uint64& StoryId, const uint64& LocConceptId, const int& Wgt);
	void SetLocationForEvent(const uint64& EventId, const uint64& LocConceptId, const int& Wgt);
	void RemoveLocationFromStory(const uint64& StoryId);
	void RemoveLocationFromEvent(const uint64& EventId);

	// similarity methods
	PRecSet GetSimilarStories(const uint64& StoryId, const int& Count = 20, const TStrSet& ValidStoryLangH = TStrSet(), const int& MaxDayDiff = TInt::Mx, const EnumConceptType& ConceptTypes = ctConcepts) const;
	PRecSet GetSimilarStoriesToEvent(const uint64& EventId, const int& Count = 20, const TStrSet& ValidStoryLangH = TStrSet(), const int& MaxDayDiff = TInt::Mx, const EnumConceptType& ConceptTypes = ctConcepts) const;
	PRecSet GetSimilarEvents(const uint64& EventId, const int& Count = 20, const int& MaxDayDiff = TInt::Mx, const EnumConceptType& ConceptTypes = ctConcepts) const;
	PRecSet GetSimilarStoriesByCCA(const uint64& StoryId, const int& Count, const TStrSet& ValidStoryLangH) const;
	PRecSet GetSimilarStoriesToEventByCCA(const uint64& EventId, const int& Count, const TStrSet& ValidStoryLangH) const;
	PRecSet GetSimilarEventsByCCA(const uint64& EventId, const int& Count) const;

	bool GetSimilarConcepts(PRecSet& RecSet, TStr& Error, const TUInt64& StoryId, const int& Count = 20, const EnumConceptType& ValidTypes = ctEntities);

	PLogger GetInfoLogFile() const { return InfoLogFile; }
	PLogger GetErrorLogFile() const { return ErrorLogFile; }
	void SetInfoLogFile(PLogger Logger) { InfoLogFile = Logger; }
	void SetErrorLogFile(PLogger Logger) { ErrorLogFile = Logger; }

	TStr PrintEventInfo(const uint64& EventRecId);
	TStr PrintArticleInfo(const uint64& ArticleRecId);
	void ReloadSettings();
	void SetSettings(const PJsonVal& SettingsJson);
	const TEventSettings GetSettings() { return Settings; }

protected:
	void Load();
	void Save();

	void ReportStoryMerge(const uint64& PrimaryStoryId, const uint64& StoryId, const int& AvgSim, const uint64& EventRecId);	
};

#endif


