/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "EventRegSrvUtil.h"
#include <base.h>
#include <mine.h>
#include "../Utils/FileUtils.h"

// /////////////
// TEventRegSrvUtil

PJsonVal TEventRegSrvUtil::EventProps(const uint64& EventRecId) const
{
	PJsonVal EventJson = TJsonVal::NewObj();
	if (!EventStore->IsRecId(EventRecId))
		return EventJson;
		
	const TStr Uri = EventStore->GetRecNm(EventRecId);
	EventJson->AddToObj("id", (int) EventRecId);
	EventJson->AddToObj("uri", Uri);

	// add stories
	PJsonVal ArrayJson = TJsonVal::NewArr();
	PRecSet StoryRecSet = EventStore->GetRec(EventRecId).DoJoin(EventRegBase->Base, "hasStory");
	for (int EventN=0; EventN < StoryRecSet->GetRecs(); EventN++) {
		const uint64 StoryRecId = StoryRecSet->GetRecId(EventN);
		TRec MedoidArticleRec = StoryStore->GetRec(StoryRecId).DoSingleJoin(NewsBase->Base, "hasMedoid");
		if (MedoidArticleRec.IsDef() && NewsBase->ArticleStore->IsRecId(MedoidArticleRec.GetRecId()))
			ArrayJson->AddToArr(NewsSrvUtil->Article(MedoidArticleRec.GetRecId(), LongerArticleBodyLen));
		// if medoid was not yet set for the story then choose the most central article (highest fq)
		else {
			PRecSet StoryArticles = StoryStore->GetRec(StoryRecId).DoJoin(NewsBase->Base, "hasArticle");
			StoryArticles->SortByFq();
			if (StoryArticles->GetRecs() > 0)
				ArrayJson->AddToArr(NewsSrvUtil->Article(StoryArticles->GetRecId(0), LongerArticleBodyLen));
		}
	}
	EventJson->AddToObj("stories", ArrayJson);
	return EventJson;
}

PJsonVal TEventRegSrvUtil::EventsProps(const PRecSet& EventRecSet) const
{
	PJsonVal ArrayJson = TJsonVal::NewArr();
	const int Events = EventRecSet->GetRecs();
	for (int N = 0; N < Events; N++)
		ArrayJson->AddToArr(EventProps(EventRecSet->GetRecId(N)));
	return ArrayJson;
}

// return [{date: 2013-11-12, count: 123 }]
PJsonVal TEventRegSrvUtil::EventsTimeAggr(const PRecSet& EventRecSet)
{
	THash<TStr, TInt> DateToCountH;
	int Events = EventRecSet->GetRecs();
	for (int N = 0; N < Events; N++) {
		const TStr Date = EventStore->GetFieldNmStr(EventRecSet->GetRecId(N), "EventDate");
		if (Date != "")
			DateToCountH.AddDat(Date) += 1;
	}

	DateToCountH.SortByKey();

	PJsonVal JsonArr = TJsonVal::NewArr();
	for (int KeyId = DateToCountH.FFirstKeyId(); DateToCountH.FNextKeyId(KeyId);) {
		const TStr Date = DateToCountH.GetKey(KeyId);
		const int Count = DateToCountH[KeyId];
		PJsonVal DateCount = TJsonVal::NewObj("date", TJsonVal::NewStr(Date));
		DateCount->AddToObj("count", Count);
		JsonArr->AddToArr(DateCount);
	}
	return JsonArr;
}

// return location aggregates
// [{count:123, location: { cityUri: ..., country: { ... }} }, {...}]
PJsonVal TEventRegSrvUtil::EventsLocAggr(const PRecSet& EventRecSet, const TStrV& LabelLanguageV)
{
	THash<TUInt64, TInt> LocToCountH;
	int Events = EventRecSet->GetRecs();
	for (int N = 0; N < Events; N++) {
		TRec EventRec = EventRecSet->GetRec(N);
		TRec LocRec = EventRec.DoSingleJoin(NewsBase->Base, "hasLocation");
		if (LocRec.IsDef())
			LocToCountH.AddDat(LocRec.GetRecId()) += 1;
	}

	LocToCountH.SortByDat(false);

	PJsonVal JsonArr = TJsonVal::NewArr();
	for (int KeyId = LocToCountH.FFirstKeyId(); LocToCountH.FNextKeyId(KeyId);) {
		const uint64 LocRecId = LocToCountH.GetKey(KeyId);
		const int Count = LocToCountH[KeyId];

		PJsonVal LocJson = TJsonVal::NewObj("count", TJsonVal::NewNum(Count));
		LocJson->AddToObj("location", NewsSrvUtil->Location(LocRecId, LabelLanguageV));
		JsonArr->AddToArr(LocJson);
	}

	return JsonArr;
}

// return location aggregates where for each location we count the number of events per individual date
// [{ location: { cityUri: ...}, dateCountArr: [{date: "2014-02-01", count: 3},  {date: ..., count: ...}]}
PJsonVal TEventRegSrvUtil::EventsLocTimeAggr(const PRecSet& EventRecSet, const TStrV& LabelLanguageV)
{
	THash<TUInt64, THash<TStr, TInt> > LocToDateCountH;
	int Events = EventRecSet->GetRecs();
	for (int N = 0; N < Events; N++) {
		const TRec EventRec = EventRecSet->GetRec(N);
		const TRec LocRec = EventRec.DoSingleJoin(NewsBase->Base, "hasLocation");
		if (LocRec.IsDef()) {
			const TStr Date = EventStore->GetFieldNmStr(EventRec.GetRecId(), "EventDate");
			THash<TStr, TInt>& DateCountH = LocToDateCountH.AddDat(LocRec.GetRecId());
			if (Date == "")
				continue;
			DateCountH.AddDat(Date) += 1;
		}
	}

	PJsonVal JsonArr = TJsonVal::NewArr();
	for (int KeyId = LocToDateCountH.FFirstKeyId(); LocToDateCountH.FNextKeyId(KeyId);) {
		const uint64 LocRecId = LocToDateCountH.GetKey(KeyId);
		PJsonVal LocInfoJson = NewsSrvUtil->Location(LocRecId, LabelLanguageV, false, true, false);

		TStr Title;
		if (LocInfoJson->IsObjKey("label")) Title = LocInfoJson->GetObjStr("label");
		else if (LabelLanguageV.Len() > 0 && LocInfoJson->IsObjKey(NewsConst.GetJsonNmForLang(LabelLanguageV[0])))
			Title = LocInfoJson->GetObjStr(NewsConst.GetJsonNmForLang(LabelLanguageV[0]));
		PJsonVal LocCountsJson = TJsonVal::NewObj("title", Title);
		if (LocInfoJson->IsObjKey("lat"))
			LocCountsJson->AddToObj("lat", LocInfoJson->GetObjNum("lat"));
		if (LocInfoJson->IsObjKey("long"))
			LocCountsJson->AddToObj("long", LocInfoJson->GetObjNum("long"));
		
		THash<TStr, TInt>& DateCountH = LocToDateCountH[KeyId];
		TJsonValV DateArrV;
		for (int DateId = DateCountH.FFirstKeyId(); DateCountH.FNextKeyId(DateId);) {
			PJsonVal DateCountJson = TJsonVal::NewObj("date", DateCountH.GetKey(DateId));
			if (DateCountH[DateId] > 1)
				DateCountJson->AddToObj("count", TJsonVal::NewNum(DateCountH[DateId]));
			DateArrV.Add(DateCountJson);
		}
		LocCountsJson->AddToObj("counts", TJsonVal::NewArr(DateArrV));
		JsonArr->AddToArr(LocCountsJson);
	}

	return JsonArr;
}

// compute who are the publishers that report most about the events in the EventRecSet
PJsonVal TEventRegSrvUtil::EventsTopPublisherAggr(const PRecSet& EventRecSet, const int& Count, const bool& IncludePublisherDetails)
{
	THash<TUInt64, TInt> SourceToCntH;
	int Events = EventRecSet->GetRecs();
	for (int E = 0; E < Events; E++) {
		THashSet<TUInt64> SourceIdH;
		const PRecSet ArticleSet = EventRecSet->GetRec(E).DoJoin(NewsBase->Base, "hasStory")->DoJoin(NewsBase->Base, "hasArticle");
		for (int N = 0; N < ArticleSet->GetRecs(); N++) {
			const uint64 SourceId = NewsBase->ArticleStore->GetFieldNmUInt64(ArticleSet->GetRecId(N), "hasSourceId");
			SourceIdH.AddKey(SourceId);
		}
		for (int KeyId = SourceIdH.FFirstKeyId(); SourceIdH.FNextKeyId(KeyId);) {
			SourceToCntH.AddDat(SourceIdH.GetKey(KeyId))++;
		}
	}
	SourceToCntH.SortByDat(false);
	int Flags = 0;
	if (IncludePublisherDetails)
		Flags = pifIncludeTitle | pifIncludeDescription | pifIncludeTags | pifIncludeLocation | pifIncludeImportance | pifIncludeDetails;
	TJsonValV SourceAggrV;
	for (int KeyId = SourceToCntH.FFirstKeyId(), N = 0; SourceToCntH.FNextKeyId(KeyId) && N < Count; N++) {
		const uint64 SourceId = SourceToCntH.GetKey(KeyId);
		const int Frequency = SourceToCntH[KeyId];
		PJsonVal SourceJson = NewsSrvUtil->Source(SourceId, (EnumPublisherInfoFlags) Flags);
		SourceJson->AddToObj("frequency", Frequency);
		SourceJson->AddToObj("ratio", TJsonVal::NewNum(Frequency / (double)Events));
		SourceAggrV.Add(SourceJson);
	}
	return TJsonVal::NewArr(SourceAggrV);
}

// events -> stories -> articles. For each article find mentioned dates
PJsonVal TEventRegSrvUtil::EventsDateMentionAggr(const PRecSet& EventRecSet, const int& MinDaysApart, const int& MinDateMentionCount)
{
	if (NewsBase->DateStore.Empty())
		return TJsonVal::NewObj("error", "Date store is not set up. There are no detected date mentions.");

	// number of events on a particular date
	THash<TStr, TInt> DateToEventCntH;
	// (event date, date mentioned in article) -> count
	THash<TPair<TStr, TStr>, TInt> DatePairToCntH;
	for (int E = 0; E < EventRecSet->GetRecs(); E++) {
		const TStr EventDateStr = NewsBase->EventStore->GetFieldNmStr(EventRecSet->GetRecId(E), "EventDate");
		if (EventDateStr == "")
			continue;
		TStrV EventPartV; EventDateStr.SplitOnAllCh('-', EventPartV);
		if (EventPartV.Len() != 3)	// comparing dates outside this range will throw exception
			continue;
		//NewsBase->InfoLogFile->NotifyInfo("Event date: %s", EventDateStr.CStr());
		DateToEventCntH.AddDat(EventDateStr) += 1;
		const TTm EventDate(EventPartV[0].GetInt(), EventPartV[1].GetInt(), EventPartV[2].GetInt());
		if (!ValidDateH.IsKey(EventDateStr))
			ValidDateH.AddDat(EventDateStr, TSysTm::IsValidDate(EventDate));
		if (ValidDateH.GetDat(EventDateStr).Val == false)
			continue;	// invalid date values (such as 2013-09-31)
		PRecSet StorySet = EventRecSet->GetRec(E).DoJoin(NewsBase->Base, "hasStory");
		PRecSet ArticleRecSet = StorySet->DoJoin(NewsBase->Base, "hasArticle");
		PRecSet DateMentionSet = ArticleRecSet->DoJoin(NewsBase->Base, "hasDate");
		for (int D = 0; D < DateMentionSet->GetRecs(); D++) {
			const TStr MentionDateStr = NewsBase->DateStore->GetRecNm(DateMentionSet->GetRecId(D));
			if (MinDaysApart > 0) {
				TStrV MentionPartV; MentionDateStr.SplitOnAllCh('-', MentionPartV);
				if (MentionPartV.Len() < 3)
					continue;
				//NewsBase->InfoLogFile->NotifyInfo("Date mention: %s", MentionDateStr.CStr());
				const TTm MentionDate(MentionPartV[0].GetInt(), MentionPartV[1].GetInt(), MentionPartV[2].GetInt());
				if (!ValidDateH.IsKey(MentionDateStr))
					ValidDateH.AddDat(MentionDateStr, TSysTm::IsValidDate(MentionDate));
				if (ValidDateH.GetDat(MentionDateStr).Val == false)
					continue;	// invalid date values (such as 2013-09-31)

				const uint64 DiffDays = TTm::GetDiffDays(EventDate, MentionDate);
				if (!(DiffDays >= MinDaysApart))
					continue;
			}
			DatePairToCntH.AddDat(TPair<TStr, TStr>(EventDateStr, MentionDateStr)) += DateMentionSet->GetRecFq(D);
		}
	}

	// prepare json to return
	// add first event counts for each date
	TJsonValV EventCountV;
	for (int KeyId = DateToEventCntH.FFirstKeyId(); DateToEventCntH.FNextKeyId(KeyId);) {
		PJsonVal EventCountJson = TJsonVal::NewObj("e", DateToEventCntH.GetKey(KeyId));		// event date
		EventCountJson->AddToObj("c", DateToEventCntH[KeyId]);								// count
		EventCountV.Add(EventCountJson);
	}

	TJsonValV DatePairV;
	for (int KeyId = DatePairToCntH.FFirstKeyId(); DatePairToCntH.FNextKeyId(KeyId);) {
		const int Count = DatePairToCntH[KeyId];
		if (Count < MinDateMentionCount)
			continue;
		const TPair<TStr, TStr> DatePair = DatePairToCntH.GetKey(KeyId);
		PJsonVal DatePairJson = TJsonVal::NewObj("e", DatePair.Val1);		// event date
		DatePairJson->AddToObj("d", DatePair.Val2);							// date mention
		DatePairJson->AddToObj("c", Count);									// count
		DatePairV.Add(DatePairJson);
	}

	PJsonVal RetJson = TJsonVal::NewObj("eventCounts", TJsonVal::NewArr(EventCountV));
	RetJson->AddToObj("datePairs", TJsonVal::NewArr(DatePairV));
	return RetJson;
}

PJsonVal TEventRegSrvUtil::EventsCategoryAggr(const PRecSet& EventRecSet)
{
	THash<TUInt64, TInt> CatToCountH;
	int Events = EventRecSet->GetRecs();
	for (int N = 0; N < Events; N++) {
		TRec EventRec = EventRecSet->GetRec(N);
		PRecSet CatRecSet = EventRec.DoJoin(NewsBase->Base, "hasCategory");
		for (int CatN = 0; CatN < CatRecSet->GetRecs(); CatN++) {
			CatToCountH.AddDat(CatRecSet->GetRecId(CatN)) += 1;
		}
	}

	CatToCountH.SortByDat(false);

	PJsonVal JsonArr = TJsonVal::NewArr();
	for (int KeyId = CatToCountH.FFirstKeyId(); CatToCountH.FNextKeyId(KeyId);) {
		const uint64 LocRecId = CatToCountH.GetKey(KeyId);
		const int Count = CatToCountH[KeyId];

		PJsonVal CatJson = TJsonVal::NewObj("count", TJsonVal::NewNum(Count));
		CatJson->AddToObj("title", NewsBase->CategoryStore->GetRecNm(LocRecId));
		JsonArr->AddToArr(CatJson);
	}

	return JsonArr;
}

// return concept aggregates
// [{freq:123, label:..., type:..., }, ...]
PJsonVal TEventRegSrvUtil::EventsConceptAggr(const PRecSet& EventRecSet, const EnumConceptType& ValidTypes, const int& ConceptCount, const TStrV& LabelLanguageV)
{
	THash<TUInt64, TFlt> Concepts;
	EventRegBase->GetConceptsForEvents(EventRecSet, ValidTypes, true, Concepts);
	Concepts.SortByDat(false);
	
	// add keys for each of the valid concept types
	TStrH TypeToCountH;
	for (int N = 0; N < NewsConst.ConceptTypeV.Len(); N++) {
		if (NewsConst.IsValidType(NewsConst.ConceptTypeV[N], ValidTypes))
			TypeToCountH.AddDat(NewsConst.ConceptTypeV[N]);
	}

	int AddedConcepts = 0;
	const int ExpectedConcepts = TypeToCountH.Len() * ConceptCount;
	TJsonValV ArrJson;
	for (int N = 0, KeyId = Concepts.FFirstKeyId(); Concepts.FNextKeyId(KeyId); N++) {
		const TStr ConceptType = NewsBase->ConceptStore->GetFieldNmStr(Concepts.GetKey(KeyId), "Type");
		if (TypeToCountH.GetDat(ConceptType) < ConceptCount) {
			PJsonVal ConceptJson = NewsSrvUtil->Concept(Concepts.GetKey(KeyId), (int) Concepts[KeyId], LabelLanguageV);
			ArrJson.Add(ConceptJson);
			TypeToCountH.AddDat(ConceptType) += 1;
			AddedConcepts++;
		}
		// did we already get ConceptCount concepts for each desired type?
		if (AddedConcepts == ExpectedConcepts)
			break;
	}
	return TJsonVal::NewArr(ArrJson);
}

PJsonVal TEventRegSrvUtil::EventsKeywordAggr(const PRecSet& EventRecSet, const TStr& Lang, const int& SampleSize)
{
	PRecSet StorySet = EventRecSet->DoJoin(NewsBase->Base, "hasStory");
	// keep only stories that are in the appropriate language
	PRecSet StoryLangSet;
	PRecSet Clone = StorySet->Clone();
	Clone->FilterByFieldStr(NewsBase->StoryStore->GetFieldId("Language"), Lang);
	if (Clone->GetRecs() == 0)
		StoryLangSet = StorySet;
	else
		StoryLangSet = Clone;
	// select only a reasonable set of stories
	const int Stories = MIN(StoryLangSet->GetRecs(), SampleSize);
	PRecSet SampleStorySetSet = StoryLangSet->GetSampleRecSet(Stories, false);
	PRecSet ArticleSet = SampleStorySetSet->DoJoin(NewsBase->Base, "hasMedoid");
	// if some articles are invalid simply return
	for (int N = 0; N < ArticleSet->GetRecs(); N++) {
		if (! NewsBase->ArticleStore->IsRecId(ArticleSet->GetRecId(N)))
			return TJsonVal::NewObj("error", TStr("Medoid contained an invalid article id"));
	}

	PSwSet SwSet;
	if (Lang == "eng")
		SwSet = TSwSet::New(swstEn523);
	else if (Lang == "spa")
		SwSet = TSwSet::New(swstEs);
	else if (Lang == "deu")
		SwSet = TSwSet::New(swstGe);
	else if (Lang == "slv")
		SwSet = TSwSet::New(swstSi);
	else
		SwSet = TSwSet::New();
	PAggr KwAggr = TAggrs::TKeywords::New(NewsBase->Base, "keywords", ArticleSet, TFtrExts::TBagOfWords::New(NewsBase->Base, NewsBase->ArticleStore, NewsBase->ArticleStore->GetFieldId("Body"), TFtrExts::bowmConcat, TTokenizers::THtmlUnicode::New(SwSet, TStemmer::New(stmtPorter, true))), Stories);
	PJsonVal KwData = KwAggr->SaveJson();
	return KwData->GetObjKey("keywords");
}

// find common entities in the EventRecSet
// don't just take the most frequently occuring entities. discount the entities as TF*IDF
PJsonVal TEventRegSrvUtil::EventsConceptMatrix(const PRecSet& EventRecSet, const EnumConceptType& ValidTypes, const TStrV& LabelLanguageV, const int& ConceptCount, const MatrixScoringMethod& ScoringMethod, const int& SampleSize)
{
	// go through all the concepts and count how frequently each pair occurs together in the events
	THash<TPair<TUInt64, TUInt64>, TInt> ConceptPairToCntH;
	THash<TUInt64, TInt> ConceptToCntH;
	
	const int Events = MIN(EventRecSet->GetRecs(), SampleSize);
	GetConceptPairs(EventRecSet, ValidTypes, ConceptPairToCntH, ConceptToCntH, Events, 10);

	// remove concepts that don't appear in at lest 5% of all events
	const double Thresh = Events * 0.05;
	for (int KeyId = ConceptToCntH.FFirstKeyId(); ConceptToCntH.FNextKeyId(KeyId);) {
		if ((double) ConceptToCntH[KeyId] < Thresh)
			ConceptToCntH.DelKeyId(KeyId);
	}

	// find "interestingness" of pairs of concepts (ci,cj)
	THash<TPair<TUInt64, TUInt64>, TFlt> ConceptPairToScoreH;
	for (int KeyId = ConceptPairToCntH.FFirstKeyId(); ConceptPairToCntH.FNextKeyId(KeyId);) {
		TPair<TUInt64, TUInt64> Pair = ConceptPairToCntH.GetKey(KeyId);
		double PairCount = ConceptPairToCntH[KeyId];
		TStr Concept1 = NewsBase->ConceptStore->GetRecNm(Pair.Val1);
		TStr Concept2 = NewsBase->ConceptStore->GetRecNm(Pair.Val2);
		// we removed the less frequent concepts so skip pairs that include any of them
		if (!ConceptToCntH.IsKey(Pair.Val1) || !ConceptToCntH.IsKey(Pair.Val2))
			continue;
		const double C1 = (double) ConceptToCntH.GetDat(Pair.Val1);
		const double C2 = (double) ConceptToCntH.GetDat(Pair.Val2);
		double Score = 0;
		// find pairs of concepts (ci,cj) that have high COF*IDF = conceptPairToCntH(ci,cj) * log ( |EventsRecSet| / (ConceptToCntH(ci)) * log( |EventRecSet| / ConceptToCntH(cj) ) )
		if (ScoringMethod == pairTfIdf)
			Score = PairCount * log(Events / C1) * log(Events / C2);
		else if (ScoringMethod == chiSquare) {
			const double E12 = C1 * C2 / (double) Events;
			const double O12 = PairCount;
			if (O12 > E12)
				Score = Sqr(O12-E12) / E12 * log(PairCount);	// take chi square but also prefer pairs with high coocurrence score
		}
		else if (ScoringMethod == pmi) {
			const double P12 = PairCount / (double) Events;
			const double P1 = C1 / (double) Events;
			const double P2 = C2 / (double) Events;
			Score = log(P12 / (P1*P2));
		}
		ConceptPairToScoreH.AddDat(Pair, Score);
	}

	// for each concept compute its score as score_ci = sum(ConceptPairToScoreH( (ci, X) ) ) + sum(ConceptPairToScoreH( (X, ci) ) )
	THash<TUInt64, TFlt> ConceptToScoreH;
	for (int KeyId = ConceptPairToScoreH.FFirstKeyId(); ConceptPairToScoreH.FNextKeyId(KeyId);) {
		TPair<TUInt64, TUInt64> Pair = ConceptPairToScoreH.GetKey(KeyId);
		const double PairScore = ConceptPairToScoreH[KeyId];
		ConceptToScoreH.AddDat(Pair.Val1) += abs(PairScore);
		ConceptToScoreH.AddDat(Pair.Val2) += abs(PairScore);
	}

	// sort concepts by score and produce the subset of concepts that we will send
	ConceptToScoreH.SortByDat(false);
	TUInt64Set ChosenConceptsH;
	for (int Count = 0, KeyId = ConceptToScoreH.FFirstKeyId(); Count < ConceptCount && ConceptToScoreH.FNextKeyId(KeyId); Count++)
		ChosenConceptsH.AddKey(ConceptToScoreH.GetKey(KeyId));
	
	PJsonVal FreqMatrixJson = TJsonVal::NewArr();
	PJsonVal ScoreMatrixJson = TJsonVal::NewArr();
	PJsonVal ConceptListJson = TJsonVal::NewArr();
	// return info about co-occurence of best ranked concepts
	for (int KeyId = ChosenConceptsH.FFirstKeyId(); ChosenConceptsH.FNextKeyId(KeyId); ) {
		uint64 ConceptId = ChosenConceptsH.GetKey(KeyId);
		ConceptListJson->AddToArr(NewsSrvUtil->Concept(ConceptId, -1, LabelLanguageV));

		TFltV FreqArrV;
		TFltV ScoreArrV;
		for (int KeyId2 = ChosenConceptsH.FFirstKeyId(); ChosenConceptsH.FNextKeyId(KeyId2);) {
			uint64 ConceptId2 = ChosenConceptsH.GetKey(KeyId2);
			TPair<TUInt64, TUInt64> Key(MIN(ConceptId, ConceptId2), MAX(ConceptId, ConceptId2));
			double Count;
			if (ConceptId == ConceptId2)
				Count = ConceptToCntH.GetDat(ConceptId);
			else
				Count = ConceptPairToCntH.IsKey(Key) ? (double) ConceptPairToCntH.GetDat(Key) : 0.0;
			FreqArrV.Add(Count);
			double Score = 0;
			if (ConceptPairToScoreH.IsKey(TPair<TUInt64, TUInt64>(ConceptId, ConceptId2))) Score = ConceptPairToScoreH.GetDat(TPair<TUInt64, TUInt64>(ConceptId, ConceptId2));
			else if (ConceptPairToScoreH.IsKey(TPair<TUInt64, TUInt64>(ConceptId2, ConceptId))) Score = ConceptPairToScoreH.GetDat(TPair<TUInt64, TUInt64>(ConceptId2, ConceptId));
			ScoreArrV.Add(Score);
		}
		FreqMatrixJson->AddToArr(TJsonVal::NewArr(FreqArrV));
		ScoreMatrixJson->AddToArr(TJsonVal::NewArr(ScoreArrV));
	}
	PJsonVal ResJson = TJsonVal::NewObj("concepts", ConceptListJson);
	ResJson->AddToObj("sampleSize", Events);
	ResJson->AddToObj("freqMatrix", FreqMatrixJson);
	ResJson->AddToObj("scoreMatrix", ScoreMatrixJson);
	
	return ResJson;
}

PJsonVal TEventRegSrvUtil::EventsConceptGraph(const PRecSet& EventRecSet, const EnumConceptType& ValidTypes, const TStrV& LabelLanguageV, const int& ConceptCount, const int& LinkCount, const int& SampleSize)
{
	//
	// go through all the concepts and count how frequently each pair occurs together in the events
	//
	THash<TPair<TUInt64, TUInt64>, TInt> ConceptPairToCntH;
	THash<TUInt64, TInt> ConceptToCntH;
	GetConceptPairs(EventRecSet, ValidTypes, ConceptPairToCntH, ConceptToCntH, SampleSize, 10);

	// 
	// normalize top concepts by idf
	//
	THash<TUInt64, TFlt> ConceptToWgtH;
	// get total number of articles
	const uint64 TotArts = NewsBase->NewsCounts->GetArticleCount();
	// normalize concepts by IDF(c)
	for (int KeyId = ConceptToCntH.FFirstKeyId(); ConceptToCntH.FNextKeyId(KeyId);) {
		const uint64 ConceptId = ConceptToCntH.GetKey(KeyId);
		const int ConceptCnt = ConceptToCntH[KeyId];
		// get number of articles annotated with Concept
		const uint64 ConceptArts = NewsBase->NewsCounts->GetConceptFq(ConceptId);
		if (ConceptArts == 0 || TotArts == 0 || ConceptArts > TotArts) {
			NewsBase->InfoLogFile->NotifyErr("Warning: TEventRegSrvUtil::EventsConceptGraph. Concept count values are not what we expected. Concept count = %I64u, article count = %I64u.", ConceptArts, TotArts);
			continue;
		}
		const double IDF = log(TotArts / (double) ConceptArts);
		const double ConceptWgt = ConceptCnt * IDF;
		ConceptToWgtH.AddDat(ConceptId, ConceptWgt);
	}

	PJsonVal RetJson = TJsonVal::NewObj();
	// remember what would be the max possible score for a concept pair
	RetJson->AddToObj("maxPossibleScore", MIN(EventRecSet->GetRecs(), SampleSize));
	//
	// add top concepts
	ConceptToWgtH.SortByDat(false);
	THash<TUInt64, TInt> TopConcepts;
	int Added = 0;
	for (int KeyId = ConceptToWgtH.FFirstKeyId(); Added < ConceptCount && ConceptToWgtH.FNextKeyId(KeyId); Added++)
		TopConcepts.AddDat(ConceptToWgtH.GetKey(KeyId), (int) ConceptToWgtH[KeyId]);
	
	TUInt64IntKdV ConceptRecKdV; TopConcepts.GetKeyDatKdV(ConceptRecKdV);
	PRecSet ConceptRecSet = TRecSet::New(NewsBase->ConceptStore, ConceptRecKdV, true);
	RetJson->AddToObj("concepts", NewsSrvUtil->Concepts(ConceptRecSet, TUInt64Set(), TInt::Mx, ctAll, LabelLanguageV));

	//
	// add frequencies between top concepts
	ConceptPairToCntH.SortByDat(false);
	TJsonValV PairsJsonV;
	int AddedLinks = 0;
	for (int KeyId = ConceptPairToCntH.FFirstKeyId(); AddedLinks < LinkCount && ConceptPairToCntH.FNextKeyId(KeyId); AddedLinks++) {
		TPair<TUInt64, TUInt64> Pair = ConceptPairToCntH.GetKey(KeyId);
		if (! TopConcepts.IsKey(Pair.Val1) || ! TopConcepts.IsKey(Pair.Val2)) 
			continue;
		PJsonVal PairJson = TJsonVal::NewObj("id1", TJsonVal::NewNum((int) Pair.Val1));
		PairJson->AddToObj("id2", TJsonVal::NewNum((int) Pair.Val2));
		PairJson->AddToObj("score", TJsonVal::NewNum(ConceptPairToCntH[KeyId]));
		PairsJsonV.Add(PairJson);
	}
	RetJson->AddToObj("links", PairsJsonV);
	
	return RetJson;
}

// compute how the top concepts from the stories trend over time - used to show the flow through time
PJsonVal TEventRegSrvUtil::EventsTrendingConcepts(const PRecSet& EventRecSet, const TStrV& TrendingConceptV, const int& ConceptCount, const TStrV& LabelLanguageV)
{
	// find all articles from all the stories of the matching events
	PRecSet AllStoryRecSet = EventRecSet->DoJoin(NewsBase->Base, "hasStory");
	PRecSet ArticleRecSet = AllStoryRecSet->DoJoin(NewsBase->Base, "hasArticle");

	TUInt64V ConceptIdV;
	// go over each concept. find articles for the concept and make intersection with ArticleRecSet
	// save for each of the concepts the number of times the concept occurs in events distributed over time
	THash<TStr, THash<TUInt64, TInt> > DateToConceptIdCountH;
	for (int N = 0; N < TrendingConceptV.Len(); N++) {
		// ignore invalid concept uris
		if (!NewsBase->ConceptStore->IsRecNm(TrendingConceptV[N]))
			continue;
		const uint64 ConceptId = NewsBase->ConceptStore->GetRecId(TrendingConceptV[N]);
		ConceptIdV.Add(ConceptId);
		PRecSet ArticlesForConcept = NewsBase->ConceptStore->GetRec(ConceptId).DoJoin(NewsBase->Base, "hasArticle");
		NewsBase->IntersectRecSets(ArticlesForConcept, ArticleRecSet);
		for (int A = 0; A < ArticlesForConcept->GetRecs(); A++) {
			TStr Date = NewsBase->ArticleStore->GetFieldNmStr(ArticlesForConcept->GetRecId(A), "Date");
			DateToConceptIdCountH.AddDat(Date).AddDat(ConceptId) += 1;
		}
	}

	// create a trends array where each element is an object having "date" and an array of (concept, freq) for that date
	TJsonValV DateArrJsonV;
	PJsonVal ArtsPerDayJson = TJsonVal::NewObj();
	DateToConceptIdCountH.SortByKey(true);
	for (int KeyId = DateToConceptIdCountH.FFirstKeyId(); DateToConceptIdCountH.FNextKeyId(KeyId);) {
		const TStr Date = DateToConceptIdCountH.GetKey(KeyId);
		PJsonVal DateJson = TJsonVal::NewObj("date", Date);
		PJsonVal ConceptFqs = TJsonVal::NewObj();
		THash<TUInt64, TInt>& ConceptIdCountH = DateToConceptIdCountH[KeyId];
		for (int CKeyId = ConceptIdCountH.FFirstKeyId(); ConceptIdCountH.FNextKeyId(CKeyId);) {
			ConceptFqs->AddToObj(ConceptIdCountH.GetKey(CKeyId).GetStr(), TJsonVal::NewNum(ConceptIdCountH[CKeyId]));
		}
		DateJson->AddToObj("conceptFreq", ConceptFqs);
		const uint64 ArtsForDate = NewsBase->NewsCounts->GetAnnArticleCount(Date);
		DateJson->AddToObj("totArts", TJsonVal::NewNum((int) ArtsForDate));
		DateArrJsonV.Add(DateJson);
	}
	PJsonVal RetJson = TJsonVal::NewObj("trends", TJsonVal::NewArr(DateArrJsonV));

	PRecSet ConceptRecSet = TRecSet::New(NewsBase->ConceptStore, ConceptIdV);
	RetJson->AddToObj("conceptInfo", NewsSrvUtil->Concepts(ConceptRecSet, TUInt64Set(), TInt::Mx, ctAll, LabelLanguageV));
	return RetJson;
}

PJsonVal TEventRegSrvUtil::EventsGetEventClusters(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet)
{
	const int KeywordCount = TSAppSrvFun::GetFldInt(FldNmValPrV, "eventClustersKeywordCount", 30);
	const TStr LabelLang = TSAppSrvFun::GetFldVal(FldNmValPrV, "eventClustersConceptLang", "eng");
	const int MaxEventsToCluster = TSAppSrvFun::GetFldInt(FldNmValPrV, "eventClustersMaxEventsToCluster", 10000);
	const int EventsUsedForKeywords = TSAppSrvFun::GetFldInt(FldNmValPrV, "eventClustersEventsUsedForKeywords", 3000);
	const bool UseLog = TSAppSrvFun::GetFldBool(FldNmValPrV, "eventClustersUseLog", true);

	PStore ConceptStore = NewsBase->ConceptStore;

	THash<TUInt64, TInt> ConceptIdToDFH;	// document frequency of each concept in the EventRecSet
	int Events = EventRecSet->GetRecs();
	TVec<TUInt64IntKdV> ConceptKdVV;		// concepts for each event
	PBowDocBs BowDocBs = TBowDocBs::New();
	for (int N = 0; N < Events; N++) {
		const uint64 EventId = EventRecSet->GetRecId(N);
		// get event concepts with weights
		TUInt64IntKdV ConceptKdV; EventRegBase->GetEventConcepts(EventId, ctAll, ConceptKdV, false);
		ConceptKdVV.Add(ConceptKdV);
		for (int C = 0; C < ConceptKdV.Len(); C++)
			BowDocBs->AddWordStr(ConceptStore->GetRecNm(ConceptKdV[C].Key));
	}

	// compute bow. each doc is one event. words are the concepts and their weights is log(word score)
	for (int N = 0; N < Events; N++) {
		const TUInt64IntKdV ConceptKdV = ConceptKdVV[N];
		TIntFltPrV WIdWgtPrV;
		for (int C = 0; C < ConceptKdV.Len(); C++) {
			const int Wgt = ConceptKdV[C].Dat;
			const int WId = BowDocBs->AddWordStr(ConceptStore->GetRecNm(ConceptKdV[C].Key));
			WIdWgtPrV.Add(TIntFltPr(WId, UseLog ? log(Wgt) : Wgt));
		}
		BowDocBs->AddDoc(TUInt64(EventRecSet->GetRecId(N)).GetStr(), TStrV(), WIdWgtPrV);
	}

	// compute TFIDF weights
	PBowDocWgtBs BowDocWgtBs = TBowDocWgtBs::New(BowDocBs, bwwtNrmTFIDF);
	// do iterative k-means
	IterativeKMeans IKMeans(BowDocBs, BowDocWgtBs, MaxEventsToCluster);
	// compute keywords for each of the subclusters
	IKMeans.ComputeKeywords(KeywordCount, EventsUsedForKeywords);

	const TClusterInfo& ClusterInfo = IKMeans.GetClusterInfo();
	// find all concepts that appear in any of the clusters as keywords. add them as json object {concept id: concept info, ....}
	TStrSet ConceptUriSet = ClusterInfo.GetAllKeywords();
	PJsonVal ConceptJson = TJsonVal::NewObj();
	for (int KeyId = ConceptUriSet.FFirstKeyId(); ConceptUriSet.FNextKeyId(KeyId); ) {
		const TUInt64 ConceptId = ConceptStore->GetRecId(ConceptUriSet.GetKey(KeyId));
		TStr Label = NewsBase->ConceptLabel->GetConceptLabel(ConceptUriSet.GetKey(KeyId), LabelLang);
		ConceptJson->AddToObj(ConceptId.GetStr(), Label);
	}
	PJsonVal ResJson = TJsonVal::NewObj("concepts", ConceptJson);
	// add cluster info. for each cluster just store an array of { conceptid: x, wgt: y}
	ResJson->AddToObj("clusters", GetClusterDetails(ClusterInfo, BowDocBs));
	return ResJson;
}

// /////////////
// INFO FROM INDIVIDUAL EVENTS

//
// get event keywords - computed from the articles assigned in the event
PJsonVal TEventRegSrvUtil::EventKeywordAggr(const uint64& EventId, const TStr& Lang)
{
	TStrV LangV; LangV.Add(Lang);
	const PRecSet ArticleRecSet = GetEventArticles(EventId, LangV);
	if (ArticleRecSet->GetRecs() == 0)
		return TJsonVal::NewObj("error", TStr("There are no articles for the specified language so keywords can not be extracted."));

	// select only a reasonable set of stories
	PSwSet SwSet;
	if (Lang == "eng")
		SwSet = TSwSet::New(swstEn523);
	else if (Lang == "spa")
		SwSet = TSwSet::New(swstEs);
	else if (Lang == "deu")
		SwSet = TSwSet::New(swstGe);
	else if (Lang == "slv")
		SwSet = TSwSet::New(swstSi);
	else
		SwSet = TSwSet::New();
	int Articles = MIN(ArticleRecSet->GetRecs(), 500);
	PAggr KwAggr = TAggrs::TKeywords::New(NewsBase->Base, "keywords", ArticleRecSet, TFtrExts::TBagOfWords::New(NewsBase->Base, NewsBase->ArticleStore, NewsBase->ArticleStore->GetFieldId("Body"), TFtrExts::bowmConcat, TTokenizers::THtmlUnicode::New(SwSet, TStemmer::New(stmtPorter, true))), Articles);
	PJsonVal KwData = KwAggr->SaveJson();
	return KwData->GetObjKey("keywords");
}


// get information about the sources that publish articles about the specific event
// return the location and title of the sources, date of the article, title of the article, ...
PJsonVal TEventRegSrvUtil::EventSourceAggr(const TStrKdV& FldNmValPrV, const uint64& EventId)
{
	PRecSet StorySet = NewsBase->EventStore->GetRec(EventId).DoJoin(NewsBase->Base, "hasStory");
	PRecSet ArticleSet = StorySet->DoJoin(NewsBase->Base, "hasArticle");
	ArticleSet->SortByField(true, NewsBase->ArticleStore->GetFieldId("DateTime"));
		
	return NewsSrvUtil->ArticleSourceAggr(FldNmValPrV, ArticleSet);
}

// trending info about the event
PJsonVal TEventRegSrvUtil::EventArticleTrend(const TStrKdV& FldNmValPrV, const uint64& EventRecId, const TStr& FldNmPrefix) const
{
	const double MinCosSim = TSAppSrvFun::GetFldFlt(FldNmValPrV, FldNmPrefix + "MinArticleCosSim", -1);
	TStrV ConceptLangV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV);
	TStrV LangV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "Lang", LangV);
	if (LangV.Empty()) 
		LangV.AddV(NewsConst.LangCoreV);

	PRecSet ArticleSet = GetEventArticles(EventRecId, LangV);
	ArticleSet->SortByField(true, NewsBase->ArticleStore->GetFieldId("DateTime"));
	if (MinCosSim > 0) {
		ArticleSet->FilterByFieldFlt(NewsBase->ArticleStore->GetFieldId("CosToCentroid"), MinCosSim, 2);	// 1 should be enough (instead of 2) but due to rounding errors we might remove some articles using 1
	}
	PJsonVal ResJson = TJsonVal::NewObj();
	
	TStrKdV ParamV(FldNmValPrV);
	TSAppSrvFun::SetFldNmInt(ParamV, "articleTrendCount", ArticleSet->GetRecs());										// we have to return all articles
	TSAppSrvFun::SetFldNmVal(ParamV, "articleTrendSortBy", "date");														// order by date
	TSAppSrvFun::SetFldNmBool(ParamV, "articleTrendSortByAsc", true);													// increasing order
	if (!TSAppSrvFun::IsFldNm(ParamV, "articleTrendBodyLen")) TSAppSrvFun::SetFldNmInt(ParamV, "articleTrendBodyLen", -1);	// don't return body (if not specified otherwise)

	return NewsSrvUtil->Articles(ParamV, ArticleSet, FldNmPrefix);
}

// return date mentioned in articles about the event
// form: [{date: 2013-11-12, count: 123 }]
PJsonVal TEventRegSrvUtil::EventDateMentionAggr(const uint64& EventId)
{
	PRecSet StorySet = EventStore->GetRec(EventId).DoJoin(NewsBase->Base, "hasStory");
	PRecSet ArticleRecSet = StorySet->DoJoin(NewsBase->Base, "hasArticle");
	return NewsSrvUtil->ArticleDateMentionAggr(ArticleRecSet);
}
// ////////////

// get frequencies of concept pairs for concepts in EventRecSet
void TEventRegSrvUtil::GetConceptPairs(const PRecSet& EventRecSet, const EnumConceptType& ValidTypes, THash<TPair<TUInt64, TUInt64>, TInt>& ConceptPairToCntH, THash<TUInt64, TInt>& ConceptToCntH, const int& SampleSize, const int& MinConceptWgt)
{
	ConceptPairToCntH.Clr();
	ConceptToCntH.Clr();
	PRecSet SampleSet = EventRecSet->GetRecs() > SampleSize ? EventRecSet->GetSampleRecSet(SampleSize, false) : EventRecSet;
	for (int EventN = 0; EventN < SampleSet->GetRecs(); EventN++) {
		TRec EventRec = SampleSet->GetRec(EventN);
		PRecSet ConceptRecSet = EventRec.DoJoin(NewsBase->Base, "hasConcept");
		int Concepts = ConceptRecSet->GetRecs();
		for (int ConceptN = 0; ConceptN < Concepts; ConceptN++) {
			if (ConceptRecSet->GetRecFq(ConceptN) < MinConceptWgt)
				continue; 
			uint64 ConceptId = ConceptRecSet->GetRecId(ConceptN);
			TStr ConceptType = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
			if (!NewsConst.IsValidType(ConceptType, ValidTypes))
				continue;
			ConceptToCntH.AddDat(ConceptId) += 1;
			for (int ConceptN2 = ConceptN + 1; ConceptN2 < Concepts; ConceptN2++) {
				if (ConceptRecSet->GetRecFq(ConceptN2) < MinConceptWgt)
					continue;
				uint64 ConceptId2 = ConceptRecSet->GetRecId(ConceptN2);
				TStr ConceptType2 = NewsBase->ConceptStore->GetFieldNmStr(ConceptId2, "Type");
				if (!NewsConst.IsValidType(ConceptType2, ValidTypes))
					continue;
				TPair<TUInt64, TUInt64> Key = TPair<TUInt64, TUInt64>(MIN(ConceptId, ConceptId2), MAX(ConceptId, ConceptId2));
				ConceptPairToCntH.AddDat(Key) += 1;
			}
		}
	}
}

// return a RecSet containing articles assigned to EventId belonging to languages in LangV
PRecSet TEventRegSrvUtil::GetEventArticles(const uint64& EventId, const TStrV& LangV) const
{
	const PRecSet StoryRecSet = EventStore->GetRec(EventId).DoJoin(NewsBase->Base, "hasStory");
	if (LangV.Len() > 0)
		StoryRecSet->FilterByFieldStrSet(NewsBase->StoryStore->GetFieldId("Language"), TStrSet(LangV));

	PRecSet ArticleRecSet = StoryRecSet->DoJoin(NewsBase->Base, "hasArticle");
	return ArticleRecSet;
}

// print keywords for ClusterInfo and keywords for each of the subclusters
PJsonVal TEventRegSrvUtil::GetClusterDetails(const TClusterInfo& ClusterInfo, const PBowDocBs& BowDocBs)
{
	// add event count
	PJsonVal RetJson = TJsonVal::NewObj("eventCount", TJsonVal::NewNum(ClusterInfo.DIdV.Len()));
	
	// add list of ids of the events in the current cluster
	// TODO: sort the ids by their distance to centroid. closest first
	TStrV EventIdV;
	for (int N = 0; N < ClusterInfo.DIdSimV.Len(); N++) {
		const int DId = ClusterInfo.DIdSimV[N].Key;
		// debug: should be in decreasing order
		const double Sim = ClusterInfo.DIdSimV[N].Dat;
		TStr EventId = BowDocBs->GetDocNm(DId);
		EventIdV.Add(EventId);
	}
	RetJson->AddToObj("events", TJsonVal::NewArr(EventIdV));

	// add keywords for the current cluster
	int KWs = ClusterInfo.KWordSet->GetKWords();
	TJsonValV ConceptJsonV;
	for (int N = 0; N < KWs; N++) {
		const TStr ConceptNm = ClusterInfo.KWordSet->GetKWordStr(N);
		PJsonVal ConceptJson = TJsonVal::NewObj("id", (int) NewsBase->ConceptStore->GetRecId(ConceptNm));
		ConceptJson->AddToObj("wgt", ClusterInfo.KWordSet->GetKWordWgt(N));
		ConceptJsonV.Add(ConceptJson);
	}
	RetJson->AddToObj("topConcepts", ConceptJsonV);

	// recursively add keywords for sub clusters
	TJsonValV SubClusterJsonV;
	for (int N = 0; N < ClusterInfo.SubClusterV.Len(); N++) {
		SubClusterJsonV.Add(GetClusterDetails(ClusterInfo.SubClusterV[N], BowDocBs));
	}
	RetJson->AddToObj("children", SubClusterJsonV);

	// return
	return RetJson;
}
