/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include <base.h>
#include "../NewsMiner/NewsBase.h"
#include "../NewsMiner/GeoBase.h"
#include "../NewsMiner/NewsSearch.h"

#include "EventRegBase.h"
#include "EventRegSrvUtil.h"
#include "CategoryReq.h"

#include "../Srv/AddArticleSrv.h"
#include "../Srv/AddStorySrv.h"
#include "../Srv/ArticleSrv.h"
#include "../Srv/NewsSrv.h"
#include "../Srv/StoryLinkingSrv.h"
#include "../Srv/ConceptFolderSrv.h"
#include "../Srv/CategorySrv.h"
#include "../Srv/ConceptSrv.h"
#include "../Srv/ConceptClassSrv.h"
#include "../Srv/ConceptTrendsSrv.h"
#include "../Srv/EventRegSrv.h"
#include "../Srv/EventCompareSrv.h"
#include "../Srv/ServerAdminSrv.h"
#include "../Srv/SourceSrv.h"
#include "../Srv/StorySrv.h"
#include "../Srv/ConceptDailyTrendsSrv.h"

#include "../Utils/GroupNotify.h"
#include "../Utils/ValidatedStoryLinks.h"
#include "../Utils/ValidatedArticleLinks.h"

#ifdef GLib_WIN

BOOL CtrlHandler(DWORD fdwCtrlType)
{
	switch (fdwCtrlType) {
		// Handle the CTRL-C signal.
		case CTRL_C_EVENT:
			printf("Ctrl-C event. Stoping the message queue.\n");
			// stop the server loop
			//TLoop::AsyncExit();
			TQm::TEnv::ReturnCode = 0;
			return TRUE;

			// CTRL-CLOSE: confirm that the user wants to exit.
		case CTRL_CLOSE_EVENT:
			printf("Ctrl-Close event\n\n");
			return(TRUE);

			// Pass other signals to the next handler.
		case CTRL_BREAK_EVENT:
			printf("Ctrl-Break event\n\n");
			return FALSE;

		case CTRL_LOGOFF_EVENT:
			// stop the server loop
			TLoop::Stop();
			TQm::TEnv::ReturnCode = 0;
			printf("Ctrl-Logoff event\n\n");
			return FALSE;

		case CTRL_SHUTDOWN_EVENT:
			TLoop::Stop();
			TQm::TEnv::ReturnCode = 0;
			printf("Ctrl-Shutdown event\n\n");
			return FALSE;

		default:
			return FALSE;
	}
}
#endif

PNotify BigExceptionLogger;

bool BigExceptionHandler(char* MsgCStr)
{
	if (!BigExceptionLogger.Empty()) {
		TStr Output(MsgCStr);
#ifdef GLib_WIN
		Output += "Stack trace:\n";
		Output += TBufferStackWalker::GetStackTrace();
#endif
		BigExceptionLogger->OnStatus(Output);
	}
	TExcept::ThrowFull(MsgCStr, "");
	return false;
}

int StartService(int argc, char* argv[])
{
	// create environment
	Env = ::TEnv(argc, argv, TNotify::StdNotify);
	// command line parameters
	Env.PrepArgs("News miner server", 1);

	// store filename
	TStr NewsStoreDefFNm = Env.GetIfArgPrefixStr("-newsDef:", "./settings/news.def", "News schema Definition File");
	TStr GeoStoreDefFNm = Env.GetIfArgPrefixStr("-geoDef:", "./settings/geo.def", "Geo info schema Definition File");

	// port
	const int PortN = Env.GetIfArgPrefixInt("-port:", 8092, "server port");
	// paths
	const TStr IndexFPath = Env.GetIfArgPrefixStr("-folderIndex:", "./IndexArticles/", "Index Folder");
	//const TStr EventFPath = Env.GetIfArgPrefixStr("-folderEvents:", "./EventEvents/", "Folder for event store");
	const TStr TempFPath = Env.GetIfArgPrefixStr("-temp:", "./temp/", "Temporary Folder");
	const TStr GeoFPath = Env.GetIfArgPrefixStr("-geoIndex:", "./IndexGeo/", "GeoInfo Index Folder");
	const TStr DbsPath = Env.GetIfArgPrefixStr("-dbsPath:", "./dbs/Articles/", "Folder with databases and models");

	const TStr GeoCountryFNm = Env.GetIfArgPrefixStr("-geoCountryFNm:", "geoCountryInfo.json", "GeoInfo country data file");
	const TStr GeoCityFNm = Env.GetIfArgPrefixStr("-geoCityFNm:", "geoCityInfo.json", "GeoInfo city data file");

	// adding articles
	bool AddArticleP = Env.GetIfArgPrefixBool("-add:", true, "Expose-Add-Functions");

	// is it ok if geo data files are missing?
	const bool OkIfGeoDataMissingP = Env.GetIfArgPrefixBool("-okIfGeoDataMissing:", false, "Is it ok if geo data is missing");

	// clustering
	const TStrV StoryUrlV = Env.GetIfArgPrefixStrV("-addstory:", "Article-story service [lang1,lang2,...langN@url]");

	// categories for stories
	const TStr CategorySrvUrl = Env.GetIfArgPrefixStr("-addcategory:", "", "Story category service [e.g. http://localhost:8099/categorySrv]");

	// index sizes
	const int IndexCacheSizeMB = Env.GetIfArgPrefixInt("-indexcache:", 4000, "Cache Size (in MB)");
	const int ArticleCacheSizeMB = Env.GetIfArgPrefixInt("-articlecache:", 4000, "Cache Size (in MB)");

	// settings
	const TStr SettingsFNm = Env.GetIfArgPrefixStr("-settingsArticles:", "./settings/settingsArticles.json", "File with settings related to event registry");
	const TStr LangInfoFNm = Env.GetIfArgPrefixStr("-settingsLangInfo:", "./settings/langInfo.json", "File with settings related to languages");

	// do we want to log requests and responses to files
	bool LogRqToFileP = Env.GetIfArgPrefixBool("-logRq", false, "Log web requests to ./Logs/Articles folder");

	bool CreateIndexFromDumpP = Env.GetIfArgPrefixBool("-createIndexFromDump:", false, "Set to true if you with to create a new index from data in ./Dump folder");
	bool CreateDumpP = Env.GetIfArgPrefixBool("-createDump:", false, "Set to true if you just wish to generate json files with all content. Use -createIndexFromDump to create new database from the json files.");
	AddArticleP = AddArticleP && !CreateDumpP;		// if we wish to create a dump then open only in readonly mode

	if (Env.IsEndOfRun()) { return -1; }

	// prepare unicode base
	if (!TUnicodeDef::IsDef()) { TUnicodeDef::Load("./UnicodeDef.Bin"); }

#ifdef GLib_WIN
	//SetConsoleCtrlHandler((PHANDLER_ROUTINE) CtrlHandler, TRUE);
#endif

	BigExceptionLogger = TFileNotify::New("./Logs/Log-EventRegistry-ExtremeExceptions.txt", true, true);

	// set the languages we wish to consider (in autosuggest, and other operations)
	PJsonVal LangInfo = TJsonVal::NewObj();
	if (TFile::Exists(LangInfoFNm))
		LangInfo = TJsonVal::GetValFromStr(TStr::LoadTxt(LangInfoFNm));
	NewsConst.SetLangInfo(LangInfo);

	// create directories if necessary
	TDir::GenDir(IndexFPath);
	TDir::GenDir(IndexFPath + "CustomData/");
	TDir::GenDir(GeoFPath);
	TDir::GenDir("./Logs");
	TDir::GenDir("./Logs/Articles");
	//TDir::GenDir(EventFPath);

	// calculate cache-size in bytes
	const int64 IndexCacheSize = int64(IndexCacheSizeMB) * int64(TInt::Mega);
	const int64 ArticleCacheSize = int64(ArticleCacheSizeMB) * int64(TInt::Mega);

	PNewsFixDb NewsFixDb = TNewsFixDb::New(DbsPath, true);

	// load concept label
	PConceptLabel ConceptLabel = TConceptLabel::New(DbsPath);

	TNotify::StdNotify->OnStatus("Creating/loading geo store...");
	bool GeoBaseExists = TIndex::Exists(GeoFPath) && TFile::Exists(GeoFPath + "Country.BaseStore");
	PGeoBase GeoBase = GeoBaseExists ?
		TGeoBase::Load(GeoFPath, faRdOnly, ConceptLabel) :
		TGeoBase::New(GeoFPath, GeoStoreDefFNm, ConceptLabel);
	if (!GeoBaseExists) {
		if ((!TFile::Exists(DbsPath + GeoCountryFNm) || !TFile::Exists(DbsPath + GeoCityFNm)) && !OkIfGeoDataMissingP) {
			TNotify::StdNotify->OnStatusFmt("Failed to locate the file with country or city info. Check if %s%s and %s%s exist.", DbsPath.CStr(), GeoCityFNm.CStr(), DbsPath.CStr(), GeoCountryFNm.CStr());
			return -1;
		}
		GeoBase->ImportData(NewsFixDb, DbsPath + GeoCountryFNm, DbsPath + GeoCityFNm);
	}

	TNotify::StdNotify->OnStatus("Creating/loading news base...");
	bool NewsBaseExists = TIndex::Exists(IndexFPath) && TFile::Exists(IndexFPath + "Source.BaseStore");
	PNewsBase NewsBase = NewsBaseExists ?
		TNewsBase::Load(GeoBase, NewsFixDb, ConceptLabel, IndexFPath, AddArticleP ? faUpdate : faRdOnly, IndexCacheSize, ArticleCacheSize, SettingsFNm, DbsPath, TempFPath) :
		TNewsBase::New(GeoBase, NewsFixDb, ConceptLabel, NewsStoreDefFNm, IndexFPath, IndexCacheSize, ArticleCacheSize, SettingsFNm, DbsPath, TempFPath);
	TNotify::StdNotify->OnStatus("Done.");


	PConceptFolderBase ConceptFolderBase = TConceptFolderBase::New(NewsBase, SettingsFNm);

	PNewsSearch NewsSearch = TNewsSearch::New(NewsBase, ConceptFolderBase);
	/*
	if ((!NewsBaseExists || !GeoBaseExists) && AddArticleP) {
		// make sure we have all the location uris in the ConceptStore
		TNotify::StdNotify->OnStatus("Importing geo concepts into concept store...");
		GeoBase->ImportGeoConcepts(NewsBase());
		//RecentNewsBase->SetFixesForGeoConcepts(true);	// enable this only if you start with fresh fix rules
		TNotify::StdNotify->OnStatus("Done.");
	}
	*/
	PNewsSrvUtil NewsSrvUtil = TNewsSrvUtil::New(NewsBase);

	PCategoryReq CategoryReq = TCategoryReq::New(CategorySrvUrl, IndexFPath + "CustomData/Req-StoriesToCategorySrv.bin");
	if (CategorySrvUrl == "")
		TNotify::StdNotify->OnStatus("WARNING: Url for category server was not provided. Category service will not be called.");

	// load event store
	PEventRegBase EventRegBase = TEventRegBase::New(NewsBase, NewsSearch, CategoryReq, DbsPath, SettingsFNm);

	// add the daily trends service - get popularity of different concepts through last X days
	PConceptDailyTrends DailyTrends = TConceptDailyTrends::New(NewsBase, NewsSearch, NewsSrvUtil, IndexFPath + "CustomData/TrendsDaily.bin");
	DailyTrends->SetCustomParams("D0,D1", "D2,D3,D4,D5,D6,D7,D8,D9,D10,D11,D12,D13,D14,D15", 50, 0, 100);
	DailyTrends->UpdateTrends();

	// event registry service functions
	PEventRegSrvUtil EventRegSrvUtil = TEventRegSrvUtil::New(NewsBase, EventRegBase, NewsSrvUtil);

	// array of server functions
	TSAppSrvFunV SrvFunV;


	//// load basic miner function
	TSrvFun::RegDefFun(NewsBase->Base, SrvFunV);

	// create category manager and server
	PSAppSrvFun CategorySrv = TCategorySrv::New(NewsBase, NewsSrvUtil);
	SrvFunV.Add(CategorySrv);

	// auto suggest for categories
	TCategoryAutoSuggestSrv* CategoryAutoSuggestSrv = TCategoryAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
	SrvFunV.Add(CategoryAutoSuggestSrv);

	// "print" server
	SrvFunV.Add(TPrintSrv::New(NewsBase, NewsSrvUtil));
	// "overview" server
	SrvFunV.Add(TOverviewSrv::New(NewsBase, NewsSrvUtil));

	// "article" server
	PSAppSrvFun ArticleSrv = TArticleSrv::New(NewsBase, NewsSearch, NewsSrvUtil);
	ArticleSrv->SetLogRqToFile(LogRqToFileP);
	SrvFunV.Add(ArticleSrv);
	// "concept" server
	PSAppSrvFun ConceptSrv = TConceptSrv::New(NewsBase, NewsSearch, NewsSrvUtil);
	ConceptSrv->SetLogRqToFile(LogRqToFileP);
	SrvFunV.Add(ConceptSrv);
	// "source" server
	PSAppSrvFun SourceSrv = TSourceSrv::New(NewsBase, NewsSrvUtil);
	SourceSrv->SetLogRqToFile(LogRqToFileP);
	SrvFunV.Add(SourceSrv);
	// "story" server
	PSAppSrvFun StorySrv = TStorySrv::New(NewsBase, EventRegBase, NewsSrvUtil);
	StorySrv->SetLogRqToFile(LogRqToFileP);
	SrvFunV.Add(StorySrv);

	// "suggestConcepts" server
	TConceptAutoSuggestSrv* ConceptAutoSuggestSrv = TConceptAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
	SrvFunV.Add(ConceptAutoSuggestSrv);

	// "suggestSources" server
	TSourceAutoSuggestSrv* SourceAutoSuggestSrv = TSourceAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
	SrvFunV.Add(SourceAutoSuggestSrv);

	// "suggestSourceLocations" server
	SrvFunV.Add(TSourceLocationAutoSuggestSrv::New(NewsBase, NewsSrvUtil));

	// "suggestConceptClasses" server
	if (NewsBase->Base->IsStoreNm("ConceptClass")) {
		TConceptClassAutoSuggestSrv* ConceptClassAutoSuggestSrv = TConceptClassAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
		SrvFunV.Add(ConceptClassAutoSuggestSrv);
		ConceptClassAutoSuggestSrv->UpdatePrefixes();
	}

	// "conceptTrends" server - add the trends management functions (getTrends, save, build)
	TConceptTrendsSrv* ConceptTrendsSrv = TConceptTrendsSrv::New(NewsBase);
	ConceptTrendsSrv->SetLogRqToFile(LogRqToFileP);
	SrvFunV.Add(ConceptTrendsSrv);

	// "conceptDailyTrends" server - daily trends methods (getTopTrends, ...)
	/*TConceptDailyTrendsExSrv* DailyTrendsSrv = TConceptDailyTrendsExSrv::New(NewsBase, DailyTrends);
	DailyTrendsSrv->SetLogRqToFile(LogRqToFileP);
	SrvFunV.Add(DailyTrendsSrv);*/
	TConceptDailyTrendsSrv* DailyTrendsSrv = TConceptDailyTrendsSrv::New(NewsBase, NewsSrvUtil);
	DailyTrendsSrv->SetLogRqToFile(LogRqToFileP);
	SrvFunV.Add(DailyTrendsSrv);

	// "conceptFolder" server
	SrvFunV.Add(TConceptFolderSrv::New(NewsBase, ConceptFolderBase, NewsSrvUtil, LogRqToFileP));

	PConceptClassBase ConceptClassBase = TConceptClassBase::New(NewsBase);
	// "conceptClass" server
	SrvFunV.Add(TConceptClassSrv::New(NewsBase, ConceptClassBase, NewsSrvUtil, LogRqToFileP));

	// add location search functionas
	SrvFunV.Add(TGeoAutoSuggestSrv::New(GeoBase));

	PNotify ClusteringCallNotify = TFileNotify::New("./Logs/Log-ArticleToClusteringCalls.txt", true, true);
	PSendArticleToClustering SendArticleToClustering = TSendArticleToClustering::New(NewsBase, SettingsFNm, ClusteringCallNotify);

	TClustUtil* ClustUtil = new	TClustUtil(NewsBase(), EventRegBase());

	// support for /add-story call - for setting article stories, medoids, duplicates, etc
	SrvFunV.Add(TAddStorySrv::New(ClustUtil));

	// main query options for events:
	// "event" server - getEventInfo, getEventArticles, getEventStats, getSimilarEvents, ...
	SrvFunV.Add(TEventRegSrv::New(NewsBase, EventRegBase, NewsSearch, NewsSrvUtil, EventRegSrvUtil, LogRqToFileP));

	// "eventCompare" server - events comparison methods
	SrvFunV.Add(TEventCompareSrv::New(NewsBase, EventRegBase, NewsSrvUtil, EventRegSrvUtil, LogRqToFileP));

	// "storylinking" server
	SrvFunV.Add(TStoryLinkingSrv::New(NewsBase, EventRegBase, NewsSrvUtil, LogRqToFileP));

	// set the logging values
	for (int N=0; N< SrvFunV.Len(); N++)
		SrvFunV[N]->SetLogRqFolder("./Logs/Articles");

	// /////////
	// the "add" server - function to add the articles
	if (AddArticleP) {
		SrvFunV.Add(TAddArticleSrv::New(NewsBase, SendArticleToClustering));
		SrvFunV.Add(TAddArticleJsonSrv::New(NewsBase, SendArticleToClustering));
	}

	// once we've imported the data we can also build geo suggestions
	GeoBase->LoadOrUpdatePrefixes();
	// update the prefixes before we start
	TNotify::StdNotify->OnStatus("Updating concept and source prefixes...");
	ConceptAutoSuggestSrv->LoadOrUpdatePrefixes();
	CategoryAutoSuggestSrv->LoadOrUpdatePrefixes();
	SourceAutoSuggestSrv->UpdatePrefixes();
	TNotify::StdNotify->OnStatus("Done.");

	// prepare the web-server logger
	//PNotify Notify = new TFileNotify(IndexFPath + "Log-ArticlesWebRequests.txt");
	//PNotify Notify = TNotify::StdNotify;
	TVec<PNotify> NotifierV;
	NotifierV.Add(TNotify::StdNotify());
	NotifierV.Add(TFileNotify::New("./Logs/Log-ArticlesWebRequests.txt", true, true));
	PNotify Notify = TGroupNotify::New(NotifierV);

	// set the event reg base and start sending waiting category requests
	CategoryReq->SetEventRegBase(EventRegBase());

	// start server
	//WebSrv = TReplaySrv::New(PortN, SrvFunV, Notify, true);
	TReplaySrv* WebSrv = TReplaySrv::New(PortN, SrvFunV, Notify, true);

	const TStr ReqLogFNm = IndexFPath + "CustomData/ArticlesWebRqLog.bin";
	// first replay any data that was not properly serialized by exiting the program normally
	WebSrv->ReplayLog(ReqLogFNm);
	WebSrv->StartLogging(ReqLogFNm, true, true);

	// set the urls of the clustering services
	// we do this here because we want first to process the transactional log so that we don't resend to clustering the same articles that we've already sent
	SendArticleToClustering->SetStoryUrlV(StoryUrlV, IndexFPath + "CustomData/Req-ArticlesToClustering");

	// start the loop
	TLoop::Run();

	WebSrv->StopLogging();
	delete WebSrv;

	return 0;
}

int main(int argc, char* argv[])
{
	TQm::TEnv::Init();

	// set a hanlder for all exceptions that would otherwise throw a ExeStop
	TOnExeStop::PutOnExeStopF(BigExceptionHandler);
	Env = ::TEnv(argc, argv, TNotify::StdNotify);
	const TStr IndexFPath = Env.GetIfArgPrefixStr("-folderIndex:", "./IndexArticles/", "Index Folder");
	try
	{
		//
		// start the ER service
		int Ret = StartService(argc, argv);
		// negative return values mean that we did not come to the normal loop so we don't want to remove the log file but instead just exit
		if (Ret < 0)
			return 0;

		//
		// web service was shut down and data was successfully saved to disk
		// we can now remove the transaction log file
		const TStr ReqLogFNm = IndexFPath + "CustomData/ArticlesWebRqLog.bin";
		TNotify::StdNotify->OnStatus("Data sucessfully closed. Removing the transaction logging file.");
		TReplaySrv::RemoveLogData(ReqLogFNm);
		return 0;
	}
	catch (PExcept E) {
		ErrNotify(E->GetMsgStr());
		return 1;
	}
	catch (...) {
		ErrNotify("Unrecognized exception occured.");
		return 1;
	}
}
