/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include <base.h>
#include "../NewsMiner/news.h"
#include "../NewsMiner/GeoBase.h"
#include "../NewsMiner/NewsSearch.h"
#include "../NewsMiner/NewsCounts.h"

#include "../ServerArticles/EventRegBase.h"
#include "../ServerArticles/EventRegSrvUtil.h"
#include "../ServerArticles/CategoryReq.h"

#include "../Srv/AddArticleSrv.h"
#include "../Srv/AddStorySrv.h"
#include "../Srv/ArticleSrv.h"
#include "../Srv/NewsSrv.h"
#include "../Srv/ConceptFolderSrv.h"
#include "../Srv/CategorySrv.h"
#include "../Srv/ConceptSrv.h"
#include "../Srv/ConceptClassSrv.h"
#include "../Srv/ConceptTrendsSrv.h"
#include "../Srv/EventRegSrv.h"
#include "../Srv/EventCompareSrv.h"
#include "../Srv/SourceSrv.h"
#include "../Srv/ServerAdminSrv.h"
#include "../Srv/ConceptDailyTrendsSrv.h"

#include "../Learning/CrossLingStoryLink.h"

#include "../Utils/GroupNotify.h"

#include <qminer_js.h>

#ifdef GLib_WIN
BOOL CtrlHandler(DWORD fdwCtrlType)
{
	switch (fdwCtrlType) {
		// Handle the CTRL-C signal.
		case CTRL_C_EVENT:
			printf("Ctrl-C event. Stoping the message queue.\n");
			// stop the server loop
			//TLoop::AsyncExit();
			TQm::TEnv::ReturnCode = 0;
			return TRUE;

			// CTRL-CLOSE: confirm that the user wants to exit.
		case CTRL_CLOSE_EVENT:
			printf("Ctrl-Close event\n\n");
			return(TRUE);

			// Pass other signals to the next handler.
		case CTRL_BREAK_EVENT:
			printf("Ctrl-Break event\n\n");
			return FALSE;

		case CTRL_LOGOFF_EVENT:
			// stop the server loop
			TLoop::Stop();
			TQm::TEnv::ReturnCode = 0;
			printf("Ctrl-Logoff event\n\n");
			return FALSE;

		case CTRL_SHUTDOWN_EVENT:
			TLoop::Stop();
			TQm::TEnv::ReturnCode = 0;
			printf("Ctrl-Shutdown event\n\n");
			return FALSE;

		default:
			return FALSE;
	}
}
#endif

int main(int argc, char* argv[]){

	try {
		TQm::TEnv::Init();
		// create environment
		Env = ::TEnv(argc, argv, TNotify::StdNotify);
		// command line parameters
		Env.PrepArgs("News miner server", 1);

		// store filename
		TStr NewsStoreDefFNm = Env.GetIfArgPrefixStr("-newsDef:", "./settings/news.def", "News schema Definition File");
		TStr GeoStoreDefFNm = Env.GetIfArgPrefixStr("-geoDef:", "./settings/geo.def", "Geo info schema Definition File");

		// port
		const int PortN = Env.GetIfArgPrefixInt("-port:", 8092, "server port");
		// paths
		const TStr IndexFPath = Env.GetIfArgPrefixStr("-folderIndex:", "./IndexArticles/", "Index Folder");
			//const TStr EventFPath = Env.GetIfArgPrefixStr("-folderEvents:", "./EventEvents/", "Folder for event store");
		const TStr TempFPath = Env.GetIfArgPrefixStr("-temp:", "./temp/", "Temporary Folder");
		const TStr GeoFPath = Env.GetIfArgPrefixStr("-geoIndex:", "./IndexGeo/", "GeoInfo Index Folder");
		const TStr DbsPath = Env.GetIfArgPrefixStr("-dbsPath:", "./dbs/Articles/", "Folder with databases and models");

		const TStr GeoCountryFNm = Env.GetIfArgPrefixStr("-geoCountryFNm:", "countryInfo.json", "GeoInfo country data file");
		const TStr GeoCityFNm = Env.GetIfArgPrefixStr("-geoCityFNm:", "cityInfo.json", "GeoInfo city data file");

		// adding articles
		const bool AddArticleP = Env.GetIfArgPrefixBool("-add:", true, "Expose-Add-Functions");

		// javascript
		const TStrV JsFNmV = Env.GetIfArgPrefixStrV("-js:", TStrV(), "User scripts");
		const TStrV InitJsonStrV = Env.GetIfArgPrefixStrV("-jsinit:", TStrV(), "Script initialization parameters");

        // is it ok if geo data files are missing?
		const bool OkIfGeoDataMissingP = Env.GetIfArgPrefixBool("-okIfGeoDataMissing:", false, "Is it ok if geo data is missing");

		// clustering
		const TStrV StoryUrlV = Env.GetIfArgPrefixStrV("-addstory:", "Article-story service [lang1,lang2,...langN@url]");
		const bool ClusterArticlesP = Env.GetIfArgPrefixBool("-clusterArticles:", true, "Perform article clustering (true|false)");

		// categories for stories
		const TStr CategorySrvUrl = Env.GetIfArgPrefixStr("-addcategory:", "", "Story category service [e.g. http://localhost:8099/categorySrv]");

		// index sizes
		const int IndexCacheSizeMB = Env.GetIfArgPrefixInt("-indexcache:", 4000, "Cache Size (in MB)");
		const int ArticleCacheSizeMB = Env.GetIfArgPrefixInt("-articlecache:", 4000, "Cache Size (in MB)");

		// languages we want to consider
		const TStr Langs = Env.GetIfArgPrefixStr("-langs:", "eng,deu,spa,zho,slv", "Comma separated list of languages to consider (eng,deu,spa,zho,slv)");

		// settings
		const TStr SettingsFNm = Env.GetIfArgPrefixStr("-settingsArticles:", "./settings/settingsArticles.json", "File with settings related to event registry");

		// do we want to log requests and responses to files
		bool LogRqToFileP = Env.GetIfArgPrefixBool("-logRq", false, "Log web requests to ./Logs/Articles folder");

		bool CreateIndexFromDumpP = Env.GetIfArgPrefixBool("-createIndexFromDump:", false, "Set to true if you with to create a new index from data in ./Dump folder");
		bool CreateDumpP = Env.GetIfArgPrefixBool("-createDump:", false, "Set to true if you just wish to generate json files with all content. Use -createIndexFromDump to create new database from the json files.");

		if (Env.IsEndOfRun()) { return 0; }

		// prepare unicode base
		if (!TUnicodeDef::IsDef()) { TUnicodeDef::Load("./UnicodeDef.Bin"); }

#ifdef GLib_WIN
		//SetConsoleCtrlHandler((PHANDLER_ROUTINE) CtrlHandler, TRUE);
#endif

		// set the languages we wish to consider (in autosuggest, and other operations)
		TStrV LangV; Langs.SplitOnAllCh(',', LangV);
		NewsConst.SetLangs(LangV);

		// create directories if necessary
		TDir::GenDir(IndexFPath);
		TDir::GenDir(GeoFPath);
		TDir::GenDir("./Logs");
		TDir::GenDir("./Logs/Articles");
		//TDir::GenDir(EventFPath);

		// calculate cache-size in bytes
		const int64 IndexCacheSize = int64(IndexCacheSizeMB) * int64(TInt::Mega);
		const int64 ArticleCacheSize = int64(ArticleCacheSizeMB) * int64(TInt::Mega);

		bool NewsBaseExists = TIndex::Exists(IndexFPath) && TFile::Exists(IndexFPath + "Source.BaseStore");
		PNewsBase NewsBase = NewsBaseExists ?
			TNewsBase::Load(IndexFPath, AddArticleP ? faUpdate : faRdOnly, IndexCacheSize, ArticleCacheSize, SettingsFNm, DbsPath, TempFPath) :
			TNewsBase::New(NewsStoreDefFNm, IndexFPath, IndexCacheSize, ArticleCacheSize, SettingsFNm, DbsPath, TempFPath);

		bool GeoBaseExists = TIndex::Exists(GeoFPath) && TFile::Exists(GeoFPath + "Country.BaseStore");
		PGeoBase GeoBase = GeoBaseExists ?
			TGeoBase::Load(GeoFPath, faRdOnly, NewsConst.LangV) :
			TGeoBase::New(GeoFPath, GeoStoreDefFNm, NewsConst.LangV);
		if (!GeoBaseExists) {
			if ((!TFile::Exists(DbsPath + GeoCountryFNm) || !TFile::Exists(DbsPath + GeoCityFNm)) && !OkIfGeoDataMissingP) {
				TQm::TEnv::Logger->OnStatusFmt("Failed to locate the file with country or city info. Check if %s%s and %s%s exist.", DbsPath.CStr(), GeoCityFNm.CStr(), DbsPath.CStr(), GeoCountryFNm.CStr());
				return 0;
			}
			GeoBase->ImportData(NewsBase->NewsFixDb, DbsPath + GeoCountryFNm, DbsPath + GeoCityFNm, "");
		}
		// once we've imported the data we can also build geo suggestions
		GeoBase->BuildSuggestions();


		PLogger ConceptFolderLogger = TLogger::New(3);
		ConceptFolderLogger->AddLogger(new TFileNotify("./Logs/Log-ConceptFolders.txt", true, true));
		PConceptFolderBase ConceptFolderBase = TConceptFolderBase::New(NewsBase, SettingsFNm, ConceptFolderLogger);

		PNewsSearch NewsSearch = TNewsSearch::New(NewsBase, GeoBase, ConceptFolderBase);

		if ((!NewsBaseExists || !GeoBaseExists) && AddArticleP) {
			// make sure we have all the location uris in the ConceptStore
			TQm::TEnv::Logger->OnStatus("Importing geo concepts into concept store...");
			GeoBase->ImportGeoConcepts(NewsBase);
			//RecentNewsBase->SetFixesForGeoConcepts(true);	// enable this only if you start with fresh fix rules
			TQm::TEnv::Logger->OnStatus("Done.");
		}

		// array of server functions
		TSAppSrvFunV SrvFunV;


		//// load basic miner function
		TSrvFun::RegDefFun(NewsBase->Base, SrvFunV);

		PNewsSrvUtil NewsSrvUtil = TNewsSrvUtil::New(NewsBase, GeoBase);

		// create category manager and server
		PSAppSrvFun CategorySrv = TCategorySrv::New(NewsBase, NewsSrvUtil);
		SrvFunV.Add(CategorySrv);

		// auto suggest for categories
		TCategoryAutoSuggestSrv* CategoryAutoSuggestSrv = TCategoryAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
		SrvFunV.Add(CategoryAutoSuggestSrv);

		// "print" server
		SrvFunV.Add(TPrintSrv::New(NewsBase, GeoBase, NewsSrvUtil));
		// "overview" server
		SrvFunV.Add(TOverviewSrv::New(NewsBase, GeoBase, NewsSrvUtil));

		// cross-lingual story linking
		PCrossLingStoryLink CrossLingStoryLink = TCrossLingStoryLink::New(NewsBase, DbsPath);

		PCategoryReq CategoryReq = TCategoryReq::New(CategorySrvUrl, IndexFPath + "Req-StoriesToCategorySrv.dat");
		if (CategorySrvUrl == "")
			TQm::TEnv::Logger->OnStatus("WARNING: Url for category server was not provided. Category service will not be called.");

		// load event store
		PEventRegBase EventRegBase = TEventRegBase::New(NewsBase, GeoBase, NewsSearch, CategoryReq, CrossLingStoryLink, NewsConst.LangV, SettingsFNm);

		// "article" server
		PSAppSrvFun ArticleSrv = TArticleSrv::New(NewsBase, NewsSearch, NewsSrvUtil);
		ArticleSrv->SetLogRqToFile(LogRqToFileP);
		SrvFunV.Add(ArticleSrv);
		// "concept" server
		PSAppSrvFun ConceptSrv = TConceptSrv::New(NewsBase, NewsSearch, NewsSrvUtil);
		ConceptSrv->SetLogRqToFile(LogRqToFileP);
		SrvFunV.Add(ConceptSrv);
		// "source" server
		PSAppSrvFun SourceSrv = TSourceSrv::New(NewsBase, NewsSrvUtil);
		SourceSrv->SetLogRqToFile(LogRqToFileP);
		SrvFunV.Add(SourceSrv);
		// "story" server
		PSAppSrvFun StorySrv = TStorySrv::New(NewsBase, NewsSrvUtil);
		StorySrv->SetLogRqToFile(LogRqToFileP);
		SrvFunV.Add(StorySrv);

		// "suggestConcepts" server
		TConceptAutoSuggestSrv* ConceptAutoSuggestSrv = TConceptAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
		SrvFunV.Add(ConceptAutoSuggestSrv);

		// "suggestSources" server
		TSourceAutoSuggestSrv* SourceAutoSuggestSrv = TSourceAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
		SrvFunV.Add(SourceAutoSuggestSrv);

		// "suggestSourceLocations" server
		SrvFunV.Add(TSourceLocationAutoSuggestSrv::New(NewsBase, NewsSrvUtil));

		// "suggestConceptClasses" server
		if (NewsBase->Base->IsStoreNm("ConceptClass")) {
			TConceptClassAutoSuggestSrv* ConceptClassAutoSuggestSrv = TConceptClassAutoSuggestSrv::New(NewsBase, NewsSrvUtil);
			SrvFunV.Add(ConceptClassAutoSuggestSrv);
			ConceptClassAutoSuggestSrv->UpdatePrefixes();
		}

		// "conceptTrends" server - add the trends management functions (getTrends, save, build)
		TConceptTrendsSrv* ConceptTrendsSrv = TConceptTrendsSrv::New(NewsBase);
		ConceptTrendsSrv->SetLogRqToFile(LogRqToFileP);
		SrvFunV.Add(ConceptTrendsSrv);

		// add the daily trends web service - get popularity of different concepts through last X days
		PConceptDailyTrends DailyTrends = TConceptDailyTrends::New(NewsBase, NewsSearch, NewsSrvUtil, IndexFPath + "TrendsDaily.bin");
		DailyTrends->SetCustomParams("D0,D1", "D2,D3,D4,D5,D6,D7,D8,D9,D10,D11,D12,D13,D14,D15", 50, 100, 100);
		DailyTrends->UpdateTrends();

		// "conceptDailyTrends" server - daily trends methods (getTopTrends, ...)
		TConceptDailyTrendsSrv* DailyTrendsSrv = TConceptDailyTrendsSrv::New(NewsBase, DailyTrends);
		DailyTrendsSrv->SetLogRqToFile(LogRqToFileP);
		SrvFunV.Add(DailyTrendsSrv);

		// "conceptFolder" server
		SrvFunV.Add(TConceptFolderSrv::New(NewsBase, ConceptFolderBase, NewsSrvUtil, LogRqToFileP));

		if (NewsBase->Base->IsStoreNm("ConceptClass")) {
			PConceptClassBase ConceptClassBase = TConceptClassBase::New(NewsBase);
			// "conceptClass" server
			SrvFunV.Add(TConceptClassSrv::New(NewsBase, ConceptClassBase, NewsSrvUtil, LogRqToFileP));
		}

		// add location search functionas
		SrvFunV.Add(TGeoAutoSuggestSrv::New(GeoBase));

		PNotify ClusteringCallNotify = TFileNotify::New("./Logs/Log-ArticleToClusteringCalls.txt", true, true);
		PSendArticleToClustering SendArticleToClustering = TSendArticleToClustering::New(NewsBase, ClusteringCallNotify);

		// if we want to cluster articles then set up appropriate call/response methods
		if (ClusterArticlesP) {
			TClustUtil* ClustUtil = new	TClustUtil(NewsBase(), EventRegBase());

			// if we want to use external clustering service then we need to be able to handle the add-story calls from it
			if (StoryUrlV.Len() > 0) {
				// support for /add-story call - for setting article stories, medoids, duplicates, etc
				SrvFunV.Add(TAddStorySrv::New(ClustUtil));
			}

				// event registry functions
			PEventRegSrvUtil EventRegSrvUtil = TEventRegSrvUtil::New(NewsBase, EventRegBase, NewsSrvUtil);

				// main query options for events:
			// "event" server - getEventInfo, getEventArticles, getEventStats, getSimilarEvents, ...
			SrvFunV.Add(TEventRegSrv::New(NewsBase, EventRegBase, NewsSearch, NewsSrvUtil, EventRegSrvUtil, LogRqToFileP));

			// "eventCompare" server - events comparison methods
			SrvFunV.Add(TEventCompareSrv::New(NewsBase, EventRegBase, NewsSrvUtil, EventRegSrvUtil, LogRqToFileP));

			// "crossLingStoryLink" server
			SrvFunV.Add(TCrossLingStoryLinkSrv::New(NewsBase, EventRegBase, CrossLingStoryLink, NewsSrvUtil, LogRqToFileP));

			SendArticleToClustering->SetStoryUrlV(StoryUrlV, IndexFPath + "Req-ArticlesToClustering");
		}

		// set the logging values
		for (int N=0; N< SrvFunV.Len(); N++)
			SrvFunV[N]->SetLogRqFolder(".\\Logs\\Articles");

		// /////////
		// the "add" server - function to add the articles
		if (AddArticleP) {
			SrvFunV.Add(TAddArticleSrv::New(NewsBase, SendArticleToClustering));
			SrvFunV.Add(TAddArticleJsonSrv::New(NewsBase, SendArticleToClustering));
		}

		// update the prefixes before we start
		TQm::TEnv::Logger->OnStatus("Updating concept and source prefixes...");
		ConceptAutoSuggestSrv->LoadOrUpdatePrefixes();
		CategoryAutoSuggestSrv->LoadOrUpdatePrefixes();
		SourceAutoSuggestSrv->UpdatePrefixes();
		TQm::TEnv::Logger->OnStatus("Done.");

		// prepare the web-server logger
		//PNotify Notify = new TFileNotify(IndexFPath + "Log-ArticlesWebRequests.txt");
		//PNotify Notify = TNotify::StdNotify;
		TVec<PNotify> NotifierV;
		NotifierV.Add(TNotify::StdNotify());
		NotifierV.Add(TFileNotify::New("./Logs/Log-ArticlesWebRequests.txt", true, true));
		PNotify Notify = TGroupNotify::New(NotifierV);

		// set the event reg base and start sending waiting category requests
		CategoryReq->SetEventRegBase(EventRegBase());
		CategoryReq->SendNextReq();

		TVec<TQm::PScript> ScriptV;
		// javascript
		for (int JsFNmN = 0; JsFNmN < JsFNmV.Len(); JsFNmN++) {
			// javascript file
			TStr ScriptFNm = JsFNmV[JsFNmN];
			// script name
			TStr ScriptNm = ScriptFNm.GetFMid();
			// include folder
			TStrV IncludeFPath = TStrV::GetV(ScriptFNm.GetFPath());
			// initialize javascrip engine
			PScript Script = TScript::New(NewsBase->Base, ScriptNm, ScriptFNm, IncludeFPath, TVec<TJsFPath>());
			ScriptV.Add(Script);		// we have to remember the script otherwise the destructor for it gets called and calling the method crashes the exe
			// prepare server functions
			Script->RegSrvFun(SrvFunV);
			//SrvFunV.Add(TJsSrvFun::New(NewsBase->GenericBase, ScriptNm, Script));
		}

		// start server
		PWebSrv WebSrv = TSAppSrv::New(PortN, SrvFunV, Notify, true);
		TLoop::Run();
		return 0;
	}
	catch (PExcept E)
	{
		ErrNotify(E->GetMsgStr());
		return 1;
	}
	catch (...){
		ErrNotify("Unrecognized exception occured.");
		return 1;
	}
}
