/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "CategorySrv.h"

//////////////////////////////////////
/// TCrossLingStoryLinkSrv

TCategorySrv::TCategorySrv(const TStr& CategoryDbPath) : TSAppSrvFun("categorySrv", saotJSon)
{
	// load DMoz-classifier
	TNotify::StdNotify->OnStatus("Loading DMoz classifier...");
	DMozCfy = TDMozCfy::LoadFPath(CategoryDbPath);
	if (DMozCfy.Empty())
		TNotify::StdNotify->OnStatus("WARNING: Necessary files for loading the DMOZ classifier were not found. Categorization will not be available.");
	else
		TNotify::StdNotify->OnStatus("Classifier loaded successfully.");
}

TStr TCategorySrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "getCategory")
			ResJson = GetCategory(FldNmValPrV, RqEnv);
		else
			ResJson = TJsonVal::NewObj("error", TStr("No known action was provided in the request."));

		// if one of the calls returned null we need to init to empty object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (const PExcept& E) {
		ResJson = TJsonVal::NewObj();
		ResJson->AddToObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus(TStr::Fmt("Error: %s", E->GetMsgStr().CStr()));
	}
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback"))
		return GetFldVal(FldNmValPrV, "callback", "") + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

PJsonVal TCategorySrv::GetCategory(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	try {
		// get text to categorize from the body of the request
		/*PHttpRq HttpRq = RqEnv->GetHttpRq();
		TMem InMem; HttpRq->GetBodyAsMem(InMem);
		TStr Text = InMem.GetAsStr();*/
		
		TTmStopWatch StopWatch(true);
		TStr RequestUri = GetFldVal(FldNmValPrV, "key");
		TInt MaxCat = GetFldInt(FldNmValPrV, "maxCat", 3);
		TStr Text = GetFldVal(FldNmValPrV, "text", "");

		// find top 3 categories for the story
		TStrFltKdV CatNmWgtV; TStrFltKdV KeyWdWgtV;
		if (! DMozCfy.Empty())
			DMozCfy->Classify(Text, CatNmWgtV, KeyWdWgtV, MaxCat);

		
		TJsonValV CatArrV;
		for (int CatN = 0; CatN < CatNmWgtV.Len(); CatN++) {
			TStr Label = "http://www.dmoz.org/" + CatNmWgtV[CatN].Key.TrimLeft("Top/");
			double Prob = CatNmWgtV[CatN].Dat;
			PJsonVal CatJson = TJsonVal::NewObj("label", Label);
			CatJson->AddToObj("prob", TJsonVal::NewNum(Prob));
			CatArrV.Add(CatJson);
		}
		PJsonVal CatArrJson = TJsonVal::NewArr(CatArrV);
		TNotify::StdNotify->OnStatusFmt("Returning %d categories.", CatArrV.Len());

		PJsonVal RetJson = TJsonVal::NewObj("categories", CatArrJson);
		RetJson->AddToObj("key", RequestUri);
		RetJson->AddToObj("timeNeeded", TStr::Fmt("%d miliseconds", StopWatch.GetMSecInt()));
		return RetJson;
	}
	catch (const PExcept& Except) { 
		TNotify::StdNotify->OnStatus(TStr::Fmt("Error while computing a category: %s", Except->GetMsgStr().CStr()));
		return TJsonVal::NewObj("error", TStr::Fmt("Error while computing a category: %s", Except->GetMsgStr().CStr()));
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("Unrecognized exception occured while computing a category");
		return TJsonVal::NewObj("error", TStr("Unrecognized exception occured while computing a category"));
	}
}