/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include <qminer_srv.h>
#include "UserBase.h"
#include "UserSrv.h"

using namespace TQm;

int main(int argc, char* argv[]){
	try {
		TQm::TEnv::Init();
		// create environment
		Env = ::TEnv(argc, argv, TNotify::StdNotify);
		// command line parameters
		Env.PrepArgs("Plus-Machines-Server", 1);

		TStr NewsStoreDefFNm = Env.GetIfArgPrefixStr("-newsDef:", "./settings/news.def", "News schema Definition File");
		const TStr IndexFPath = Env.GetIfArgPrefixStr("-userindex:", "./IndexUsers/", "User Index Folder");

		TStr TempFPath = Env.GetIfArgPrefixStr("-temp:", "./temp/", "Temporary Folder");

		const int IndexCacheSizeMB = Env.GetIfArgPrefixInt("-indexcache:", 4000, "Cache Size (in MB)");
		const int DataCacheSizeMB = Env.GetIfArgPrefixInt("-datacache:", 4000, "Cache Size (in MB)");

		const bool UpdateP = Env.GetIfArgPrefixBool("-update:", true, "update mode");

		const int Port = Env.GetIfArgPrefixInt("-port:", 8091, "server port");
		TStr SettingsFileStr = Env.GetIfArgPrefixStr("-settingsUsers:", "./settings/settingsUsers.json", "File with various settings");

		TStr SettingsStr = TFile::Exists(SettingsFileStr) ? TStr::LoadTxt(SettingsFileStr) : "{}";
		PJsonVal SettingsJson = TJsonVal::GetValFromStr(SettingsStr);

		if (Env.IsEndOfRun()) { return 0; }

		// prepare unicode base
		if (!TUnicodeDef::IsDef()) { TUnicodeDef::Load("./UnicodeDef.Bin"); }

		// create directories if necessary
		TDir::GenDir(IndexFPath);

		// calculate cache-size in bytes
		const int64 IndexCacheSize = int64(IndexCacheSizeMB) * int64(TInt::Mega);
		const int64 DataCacheSize = int64(DataCacheSizeMB) * int64(TInt::Mega);

		bool ExistsP = TIndex::Exists(IndexFPath) && TFile::Exists(IndexFPath + "User.BaseStore");
		PUserBase UserBase = ExistsP ?
			TUserBase::Load(IndexFPath, UpdateP ? faUpdate : faRdOnly, IndexCacheSize, DataCacheSize, TempFPath) :
			TUserBase::New(NewsStoreDefFNm, IndexFPath, IndexCacheSize, DataCacheSize, TempFPath);

		// load basic miner function
		TSAppSrvFunV SrvFunV;
		TSrvFun::RegDefFun(UserBase->Base, SrvFunV);

		// load plus machines functions
		PUserSrvUtil UserSrvUtil = TUserSrvUtil::New(UserBase);

		SrvFunV.Add(TUsersSrv::New(UserBase, UserSrvUtil));

		// set the logging values
		for (int N = 0; N < SrvFunV.Len(); N++) {
			SrvFunV[N]->SetLogRqToFile(SettingsJson->GetObjBool("LogReqRes", false));
			SrvFunV[N]->SetLogRqFolder(SettingsJson->GetObjStr("LogFolder", "./Logs/Users"));
		}

		// prepare the web-server logger
		PNotify WebNotify = TNotify::StdNotify;
		// start server
		PWebSrv WebSrv = TSAppSrv::New(Port, SrvFunV, WebNotify, true);
		TLoop::Run();

		return 0;
	}
	catch (PExcept E) {
		TQm::TEnv::Logger->OnStatus("Error: " + E->GetMsgStr());
		return 1;
	}
	catch (...) {
		return 1;
	}    
}
