/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "UserBase.h"
#include "../Utils/JsonUtils.h"

///////////////////////////////
// News-Base
TUserBase::TUserBase(const TStr& StoreDefFNm, const TStr& FPath, const int64& IndexCacheSize, const int64& DataCacheSize, const TStr& TempFPath) {

	PJsonVal StoreDefVal = TJsonVal::GetValFromStr(TStr::LoadTxt(StoreDefFNm));
	// delete some stores from the news.def
	TStrSet StoreH; StoreH.AddKey("Concept"); StoreH.AddKey("ConceptFolder"); StoreH.AddKey("User");
	RemoveJsonStores(StoreDefVal, StoreH);

	for (int StoreN = 0; StoreN < StoreDefVal->GetArrVals(); StoreN++) {
		PJsonVal Store = StoreDefVal->GetArrVal(StoreN);
		PJsonVal Joins = Store->GetObjKey("joins");
		for (int N = Joins->GetArrVals() - 1; N >= 0; N--) {
			TStr TargetStore = Joins->GetArrVal(N)->GetObjStr("store");
			if (!StoreH.IsKey(TargetStore))
				Joins->DelArrVal(N);
		}
	}
	
	// prepare base
	Base = NewBase(FPath, StoreDefVal, IndexCacheSize, DataCacheSize, TStrUInt64H(), false);
		
	// initialize news base
	Init();
	// finish initalization
	Base->Init(TempFPath);

	SaveBase(Base);
}

void TUserBase::RemoveJsonStores(const PJsonVal& JsonScheme, const TStrSet& KeepStores)
{
	for (int N = JsonScheme->GetArrVals() - 1; N >= 0; N--) {
		if (! KeepStores.IsKey(JsonScheme->GetArrVal(N)->GetObjStr("name")))
			JsonScheme->DelArrVal(N);
	}
}

TUserBase::TUserBase(const TStr& FPath, const TFAccess& FAccess, const int64& IndexCacheSize, const int64& DataCacheSize, const TStr& TempFPath) {

	// prepare base
	Base = LoadBase(FPath, FAccess, IndexCacheSize, DataCacheSize, TStrUInt64H(), false);
	// initialize news base
	Init();
	// finish initalization
	Base->Init(TempFPath);

	// load the source importance data if the file exists
	TStr UserBaseFile = Index->GetIndexFPath() + "UserBase.bin";
	if (TFile::Exists(UserBaseFile)) {
		try {
			TFIn FIn(UserBaseFile);
			ConceptsH.Load(FIn);
			SourcesH.Load(FIn);
			LocationsH.Load(FIn);
			AvailableFolderUriV.Load(FIn);
		}
		catch (PExcept E) {
			TQm::TEnv::Logger->OnStatus("Warning: UserBase.dat file was found but we were not able to load it.");
			// todo: maybe try to identify an appropriate value for LastFolderId
		}
	}
}

void TUserBase::Init() {
	// prepare shortcuts
	IndexVoc = Base->GetIndexVoc();
	Index = Base->GetIndex();
	// prepare concept store
	EAssertR(Base->IsStoreNm("Concept"), "Concept store does not exist");
	ConceptStore = Base->GetStoreByStoreNm("Concept");
	ConceptStoreId = ConceptStore->GetStoreId();
	// prepare concept folder store
	EAssertR(Base->IsStoreNm("ConceptFolder"), "ConceptFolder store does not exist");
	ConceptFolderStore = Base->GetStoreByStoreNm("ConceptFolder");
	ConceptFolderStoreId = ConceptFolderStore->GetStoreId();
	// prepare user store
	EAssertR(Base->IsStoreNm("User"), "User store does not exist");
	UserStore = Base->GetStoreByStoreNm("User");
	UserStoreId = UserStore->GetStoreId();
}

PUserBase TUserBase::New(const TStr& StoreDefFNm, const TStr& FPath,
	const int64& IndexCacheSize, const int64& DataCacheSize, const TStr& TempFPath) {

	return new TUserBase(StoreDefFNm, FPath, IndexCacheSize, DataCacheSize, TempFPath);
}

PUserBase TUserBase::Load(const TStr& FPath, const TFAccess& FAccess,
	const int64& IndexCacheSize, const int64& DataCacheSize, const TStr& TempFPath) {

	return new TUserBase(FPath, FAccess, IndexCacheSize, DataCacheSize, TempFPath);
}

void TUserBase::Save() { 
	TStr UserBaseFile = Index->GetIndexFPath() + "UserBase.bin";
	TFOut FOut(UserBaseFile);
	ConceptsH.Save(FOut);
	SourcesH.Save(FOut);
	LocationsH.Save(FOut);
	AvailableFolderUriV.Save(FOut);
}

//////////////////////////////////////
// user management and methods

// create a new user
uint64 TUserBase::AddUser(const TStr& UserUri, const TStr& Name) 
{ 
	PUser User = TUser::New();
	User->Name = Name;
	User->Uri = UserUri;
	User->LastAccessTime = TUser::GetTime();

	uint64 UserId = UserStore->AddRec(User->GetJsonForStore());
	return UserId;	
}

PUser TUserBase::GetUser(const uint64& UserId)
{
	if (!UserStore->IsRecId(UserId))
		return NULL;
	return TUser::New(UserId, UserStore);
}

PUser TUserBase::GetUser(const TStr& UserUri)
{
	if (!UserStore->IsRecNm(UserUri))
		return NULL;
	return TUser::New(UserUri, UserStore);
}

// change the name of the user
void TUserBase::SetUserName(const TStr& UserUri, const TStr& Name)
{
	if (!UserStore->IsRecNm(UserUri))
		return;
	PUser User = GetUser(UserUri);
	User->Name = Name;
	UserStore->AddRec(User->GetJsonForStore());
}

// return the id of the home folder for the user
uint64 TUserBase::GetHomeFolderId(const uint64& UserId)
{
	if (!UserStore->IsRecId(UserId))
		return TUInt64::Mx;
	
	TRec HomeFolder = UserStore->GetRec(UserId).DoSingleJoin(Base, "hasHomeFolder");
	return HomeFolder.GetRecId();
}

void TUserBase::SetAsHomeFolder(const uint64& UserId, const uint64& FolderId)
{
	if (!UserStore->IsRecId(UserId))
		return;
	// validate concept folder id
	if (!ConceptFolderStore->IsRecId(FolderId))
		return;
	// make as home folder
	UserStore->AddJoin("hasHomeFolder", UserId, FolderId);
}

// return the id of the folder named FolderName. The folder has to be owned by the UserId
//uint64 TUserBase::GetFolderIdByName(const uint64& UserId, const TStr& FolderName)
//{
//	EAssertR(UserStore->IsRecId(UserId), "The user id is not valid");
//
//	TRec UserRec(UserStore->GetStoreId(), UserId);
//	PRecSet FolderSet = UserRec.DoJoin(Base, UserStore->JoinHasFoldersId);
//	for (int N=0; N < FolderSet->GetRecs(); N++)
//	{
//		if (FolderStore->GetFieldStr(FolderSet->GetRecId(N), FolderStore->FolderNameFieldId) == FolderName)
//			return FolderSet->GetRecId(N);
//	}
//	return TUInt64::Mx;
//}
//
//// given the owner's id and the folder id return the folder rec id
//uint64 TUserBase::GetFolderId(const uint64& UserId, const TUInt64& FolderId)
//{
//	EAssertR(UserStore->IsRecId(UserId), "The user id is not valid");
//
//	TRec UserRec(UserStore->GetStoreId(), UserId);
//	PRecSet FolderSet = UserRec.DoJoin(Base, UserStore->JoinHasFoldersId);
//	for (int N=0; N < FolderSet->GetRecs(); N++)
//	{
//		if (FolderStore->GetFieldUInt64(FolderSet->GetRecId(N), FolderStore->FolderIdFieldId) == FolderId)
//			return FolderSet->GetRecId(N);
//	}
//	return TUInt64::Mx;
//}

// return the list of folder ids created by the user (without the home folder)
PRecSet TUserBase::GetFoldersForUser(const uint64& UserId, const bool& IncludeHomeFolder)
{
	if (!UserStore->IsRecId(UserId))
		return TRecSet::New(ConceptFolderStore);

	PRecSet FolderSet = UserStore->GetRec(UserId).DoJoin(Base, "hasConceptFolder");
	if (!IncludeHomeFolder)
		FolderSet->RemoveRecId(GetHomeFolderId(UserId));
	return FolderSet;
}

// add/replace uri vectors with id vectors. add uris to ConceptH, ...
// so that we can later convert ids back to uris
void TUserBase::AddIdsForFolderJson(const PJsonVal& JsonData, const bool& RemoveUris)
{
	if (JsonData->IsObjKey("conceptUriV")) {
		PJsonVal ConceptUriV = JsonData->GetObjKey("conceptUriV");
		PJsonVal ConceptIdV = TJsonVal::NewArr();
		for (int N = 0; N < ConceptUriV->GetArrVals(); N++) {
			TStr ConceptUri = ConceptUriV->GetArrVal(N)->GetStr();
			int ConceptId = AddConcept(ConceptUri);
			ConceptIdV->AddToArr(TJsonVal::NewNum(ConceptId));
		}
		JsonData->AddToObj("conceptIdV", ConceptIdV);
	}
	if (JsonData->IsObjKey("fixedConceptUriV")) {
		PJsonVal ConceptUriV = JsonData->GetObjKey("fixedConceptUriV");
		PJsonVal ConceptIdV = TJsonVal::NewArr();
		for (int N = 0; N < ConceptUriV->GetArrVals(); N++) {
			const TStr ConceptUri = ConceptUriV->GetArrVal(N)->GetStr();
			int ConceptId = AddConcept(ConceptUri);
			ConceptIdV->AddToArr(TJsonVal::NewNum(ConceptId));
		}
		JsonData->AddToObj("fixedConceptIdV", ConceptIdV);
	}
	if (JsonData->IsObjKey("sourceUriV")) {
		PJsonVal SourceUriV = JsonData->GetObjKey("sourceUriV");
		PJsonVal SourceIdV = TJsonVal::NewArr();
		for (int N = 0; N < SourceUriV->GetArrVals(); N++) {
			const TStr SourceUri = SourceUriV->GetArrVal(N)->GetStr();
			int SourceId = AddSource(SourceUri);
			SourceIdV->AddToArr(TJsonVal::NewNum(SourceId));
		}
		JsonData->AddToObj("sourceIdV", SourceIdV);
	}
	if (JsonData->IsObjKey("locationUriV")) {
		PJsonVal LocationUriV = JsonData->GetObjKey("locationUriV");
		PJsonVal LocationIdV = TJsonVal::NewArr();
		for (int N = 0; N < LocationUriV->GetArrVals(); N++) {
			TStr Loc = LocationUriV->GetArrVal(N)->GetStr();
			int LocationId = AddLocation(Loc);
			LocationIdV->AddToArr(TJsonVal::NewNum(LocationId));
		}
		JsonData->AddToObj("locationIdV", LocationIdV);
	}

	if (RemoveUris) {
		JsonData->DelObjKey("conceptUriV");
		JsonData->DelObjKey("fixedConceptUriV");
		JsonData->DelObjKey("sourceUriV");
		JsonData->DelObjKey("locationUriV");
	}
}

// add/replace id vectors with uri vectors. 
void TUserBase::AddUrisForFolderJson(const PJsonVal& JsonData, const bool& RemoveIds)
{
	if (JsonData->IsObjKey("conceptIdV")) {
		PJsonVal ConceptIdV = JsonData->GetObjKey("conceptIdV");
		PJsonVal ConceptUriV = TJsonVal::NewArr();
		for (int N = 0; N < ConceptIdV->GetArrVals(); N++) {
			int ConceptId = (int) ConceptIdV->GetArrVal(N)->GetNum();
			const TStr ConceptUri = GetConcept(ConceptId);
			ConceptUriV->AddToArr(TJsonVal::NewStr(ConceptUri));
		}
		JsonData->AddToObj("conceptUriV", ConceptUriV);
	}
	if (JsonData->IsObjKey("fixedConceptIdV")) {
		PJsonVal ConceptIdV = JsonData->GetObjKey("fixedConceptIdV");
		PJsonVal ConceptUriV = TJsonVal::NewArr();
		for (int N = 0; N < ConceptIdV->GetArrVals(); N++) {
			uint64 ConceptId = (int) ConceptIdV->GetArrVal(N)->GetNum();
			const TStr ConceptUri = GetConcept(ConceptId);
			ConceptUriV->AddToArr(TJsonVal::NewStr(ConceptUri));
		}
		JsonData->AddToObj("fixedConceptUriV", ConceptUriV);
	}
	if (JsonData->IsObjKey("sourceIdV")) {
		PJsonVal SourceIdV = JsonData->GetObjKey("sourceIdV");
		PJsonVal SourceUriV = TJsonVal::NewArr();
		for (int N = 0; N < SourceIdV->GetArrVals(); N++) {
			uint64 SourceId = (int) SourceIdV->GetArrVal(N)->GetNum();
			const TStr SourceUri = GetSource(SourceId);
			SourceUriV->AddToArr(TJsonVal::NewStr(SourceUri));
		}
		JsonData->AddToObj("sourceUriV", SourceUriV);
	}
	if (JsonData->IsObjKey("locationIdV")) {
		PJsonVal LocationIdV = JsonData->GetObjKey("locationIdV");
		PJsonVal LocationUriV = TJsonVal::NewArr();
		for (int N = 0; N < LocationIdV->GetArrVals(); N++) {
			uint64 LocationId = (int) LocationIdV->GetArrVal(N)->GetNum();
			TStr Loc = GetLocation(LocationId);
			LocationUriV->AddToArr(TJsonVal::NewStr(Loc));
		}
		JsonData->AddToObj("locationUriV", LocationUriV);
	}

	if (RemoveIds) {
		JsonData->DelObjKey("conceptIdV");
		JsonData->DelObjKey("fixedConceptIdV");
		JsonData->DelObjKey("sourceIdV");
		JsonData->DelObjKey("locationIdV");
	}
}

//////////////////////////////////////
// folder methods
// return the owner of the concept folder
uint64 TUserBase::GetFolderOwner(const TStr& FolderUri)
{
	if (!ConceptFolderStore->IsRecNm(FolderUri))
		return TUInt64::Mx;

	TRec UserRec = ConceptFolderStore->GetRec(FolderUri).DoSingleJoin(Base, "hasUser");
	return UserRec.GetRecId();
}

// after a folder is deleted add the uri to the list of available uris
void TUserBase::AddAvailableFolderUri(const TStr& FolderUri)
{
	AvailableFolderUriV.Add(FolderUri);
}

// if there an available uri, get it, otherwise return empty
TStr TUserBase::GetAvailableFolderUri()
{
	if (AvailableFolderUriV.Len() == 0)
		return "";
	TStr Uri = AvailableFolderUriV[AvailableFolderUriV.Len() - 1];
	AvailableFolderUriV.DelLast();
	return Uri;
}

TStr TUserBase::GetNewFolderUri()
{
	TStr Uri = GetAvailableFolderUri();
	
	if (Uri == "") {
		uint64 NewRecId = 0;
		if (ConceptFolderStore->GetRecs() > 0)
			NewRecId = ConceptFolderStore->GetLastRecId() + 1;
		Uri = TStr::Fmt("%s%I64u", NewsConst.TypeConceptFolder.CStr(), NewRecId);
	}
	return Uri;
}

// create a new folder for the given user
uint64 TUserBase::AddFolder(const uint64& UserId, const TStr& FolderName, const FolderVisibility& Visibility) 
{ 
	if (!UserStore->IsRecId(UserId))
		return TUInt64::Mx;
	
	TStr Uri = GetNewFolderUri();
	// we are re-using a concept folder that has existed in the past. concept <--> concept folder already exists
	bool ExistingUri = ConceptFolderStore->IsRecNm(Uri);

	PJsonVal Json = TJsonVal::NewObj("URI", Uri);
	Json->AddToObj("Label", FolderName);
	Json->AddToObj("Type", NewsConst.TypeConceptFolder);
	
	const uint64 ConceptRecId = ConceptStore->AddRec(Json);

	PConceptFolder Folder = TConceptFolder::New();
	Folder->Uri = Uri;
	Folder->Visibility = (int) Visibility;
	const uint64 CFRecId = ConceptFolderStore->AddRec(Folder->GetJsonForStore());

	if (!ExistingUri) {
		ConceptStore->AddJoin("hasConceptFolder", ConceptRecId, CFRecId);
	}

	UserStore->AddJoin("hasConceptFolder", UserId, CFRecId);
	return CFRecId;
}

PJsonVal TUserBase::GetFolderJson(const uint64& FolderId)
{
	PConceptFolder Folder = GetFolder(FolderId);
	if (Folder.Empty())
		return TJsonVal::NewObj();
	return Folder->GetJson();
}

PConceptFolder TUserBase::GetFolder(const uint64& FolderId)
{
	if (!ConceptFolderStore->IsRecId(FolderId))
		return NULL;
	const TStr ConceptFolderUri = ConceptFolderStore->GetRecNm(FolderId);
	PConceptFolder Folder = TConceptFolder::New(ConceptFolderUri, ConceptFolderStore);
	return Folder;
}

// delete an existing folder
void TUserBase::DeleteFolder(const uint64& FolderId)
{
	if (!ConceptFolderStore->IsRecId(FolderId)) {
		// todo: log error
		return;
	}
	TStr FolderUri = ConceptFolderStore->GetRecNm(FolderId);
	uint64 UserId = GetFolderOwner(FolderUri);
	
	// put an empty folder in place of existing ConceptFolderUri
	PConceptFolder Folder = TConceptFolder::New();
	Folder->Uri = ConceptFolderStore->GetRecNm(FolderId);
	UpdateFolder(Folder);
		
	if (ConceptStore->IsRecNm(FolderUri)) {
		uint64 ConceptId = ConceptStore->GetRecId(FolderUri);
		// keep the concept folder <--> concept in the index. the next use of this concept folder uri will reuse it
		// just clear the label for concept
		if (ConceptStore->IsRecId(ConceptId)) {
			PJsonVal ConcJson = TJsonVal::NewObj("URI", ConceptStore->GetRecNm(ConceptId));
			ConcJson->AddToObj("Label", TStr(""));
			ConcJson->AddToObj("Description", TStr(""));
			ConceptStore->AddRec(ConcJson);
		}
	}
		
	// delete concept folder <--> user
	UserStore->DelJoin("hasConceptFolder", UserId, FolderId);
	
	// mark the concept folder as 
	AddAvailableFolderUri(FolderUri);
}

// update folder meta information
void TUserBase::UpdateFolder(const PConceptFolder& Folder)
{
	if (!ConceptFolderStore->IsRecNm(Folder->Uri)) {
		// todo: log error
		return;
	}
	// call AddRec instead of UpdateRec since we have URI inside and not $id
	ConceptFolderStore->AddRec(Folder->GetJsonForStore());
}

void TUserBase::UpdateConcept(const TStr& ConceptUri, const TStr& Label)
{
	PJsonVal ConcJson = TJsonVal::NewObj("URI", ConceptUri);
	ConcJson->AddToObj("Label", Label);
	ConceptStore->AddRec(ConcJson);
}

// increase article count by one for FolderId
void TUserBase::IncreaseArticleCount(const uint64& FolderId)
{
	if (!ConceptFolderStore->IsRecId(FolderId))
		return;
	const TStr FolderUri = ConceptFolderStore->GetRecNm(FolderId);
	const int ArticlesInFolder = ConceptFolderStore->GetFieldNmInt(FolderId, "ArticlesInFolder");
	PJsonVal Json = TJsonVal::NewObj("URI", FolderUri);
	Json->AddToObj("ArticlesInFolder", TJsonVal::NewNum(ArticlesInFolder + 1));
	ConceptFolderStore->AddRec(Json);
}

// update the +concepts of the folder using the concepts in the article
void TUserBase::UpdateFolderConceptsWithArticleConcepts(const uint64 FolderId, const TUInt64FltKdV& ImportantArticleConceptsV)
{
	NotifyVerbose(3, "-- Folder before update -- ");
	PrintFolderDefinition(FolderId);

	PConceptFolder Folder = TConceptFolder::New(FolderId, ConceptFolderStore);
	
	//int FolderArticleCount = FolderArticles->GetRecs();	
	const double DecayRate = 0.97;
	const double MaxArticleWgt = 20;
	double ArticleWgt = min(20, max(3, Folder->ArticlesInFolder));	// Wgt should be between 3 and 20 based on the number of articles currently in the folder
	
	// make the old weights in the folder SpV a bit smaller using the DecayRate
	TIntFltKdV SpVData;
	
	for (int N=0; N < Folder->ConceptIdV.Len(); N++) {
		const int ConceptId = Folder->ConceptIdV[N];
		if (Folder->IsConceptFixed(ConceptId)) continue;	// don't changed fixed concepts
		if (Folder->ConceptWgtV[N] < 0) continue;			// don't change weights for concepts that we wish to ignore

		Folder->ConceptWgtV[N] = Folder->ConceptWgtV[N] * DecayRate;
	}
	
	// add wgts for concepts in the article
	for (int N=0; N < ImportantArticleConceptsV.Len(); N++)
	{
		uint64 ConceptId = ImportantArticleConceptsV[N].Key;
		if (Folder->IsConceptFixed(ConceptId)) continue;
		int Ind = Folder->ConceptIdV.SearchForw(ConceptId, 0);
		if (Ind != -1 && Folder->ConceptWgtV[Ind] < 0) continue;	// don't change weights for concepts that we wish to ignore
		// put the concept to the [0, 10] range
		double Val = ImportantArticleConceptsV[N].Dat * 2 * MaxArticleWgt / ArticleWgt;
		// update the vector with 5% of the values from the annotations - 95% of the values are kept from the old values in the SpV
		double Diff = Val * (1.0-DecayRate);
		if (Ind == -1)
			Ind = Folder->AddConcept(ConceptId, 0);
		double  OldVal = Folder->ConceptWgtV[Ind];
		Folder->ConceptWgtV[Ind] += Diff;
		NotifyVerboseFmt(3, "Increasing concept %s wgt from %.2f for %.3f", GetConcept(ConceptId).CStr(), OldVal, Diff);
	}
	
	// every once in a while remove the concepts that are noise. this also enables to filter out topics that died out
	// TODO!! fix this. watch that we don't remove items with negative weights
	if (Folder->ArticlesInFolder % 40 == 0) {
		for (int N = Folder->ConceptIdV.Len() - 1; N >= 0; N--) {
			if (Folder->ConceptWgtV[N] > 0 && Folder->ConceptWgtV[N] < 0.05) {
				Folder->ConceptIdV.Del(N);
				Folder->ConceptWgtV.Del(N);
			}
		}
	}
	if (VerbosityLevel >= 3) {
		NotifyVerbose(3, "-- Folder after update -- ");
		PrintFolderDefinition(FolderId);
		NotifyVerbose(3, "--------------");
	}
}

//////////////////////////////////////
// following people

// get the list of friends that are followed by user with UserId
PRecSet TUserBase::GetFollowingFriends(const uint64& UserId)
{
	PRecSet FollowingSet = UserStore->GetRec(UserId).DoJoin(Base, "isFollowing");
	return FollowingSet;
}

PRecSet TUserBase::GetFollowedByFriends(const uint64& UserId)
{
	PRecSet FollowedBySet = UserStore->GetRec(UserId).DoJoin(Base, "isFollowedBy");
	return FollowedBySet;
}

// follow another friend
void TUserBase::FollowAdd(const uint64& UserId, const uint64& FriendUserId)
{
	UserStore->AddJoin("isFollowing", UserId, FriendUserId);
}

// unfollow a friend
void TUserBase::FollowRemove(const uint64& UserId, const uint64& FriendUserId)
{
	UserStore->DelJoin("isFollowing", UserId, FriendUserId);
}

// for a given folder find the set of folders that have most similar ConceptSpV
//PRecSet TUserBase::FindSimilarFolders(const uint64& FolderId, int MaxFolderCount)
//{
//	PBowSim BowSim = TBowSim::New(bstCos);
//	PConceptFolder Folder = GetFolder(FolderId);
//	
//	const double MinSim = 0.3;
//	THashSet<TFltUInt64Pr> SimilarityFolderIdH;
//	for (int FolderN = 0; FolderN < FolderStore->GetRecs(); FolderN++) {
//		if (FolderN == FolderId)
//			continue;
//		TFolderRec Folder = FolderStore->GetFolder(FolderN);
//		double Sim = BowSim->GetSim(FolderTest.ConceptSpV, Folder.ConceptSpV);
//		if (Sim < MinSim)
//			continue;
//		SimilarityFolderIdH.AddKey(TFltUInt64Pr(Sim, FolderN));
//	}
//
//	// take from the hash only MaxFolderCount most similar folders
//	TUInt64IntKdV MostSimilarFoldersV;
//	for (int Cnt = 0; Cnt < MaxFolderCount && SimilarityFolderIdH.Len() > 0; Cnt++) {
//		TFltUInt64Pr MaxVal = TFltUInt64Pr(-1, 0);
//		int KeyId = SimilarityFolderIdH.FFirstKeyId();
//		while (SimilarityFolderIdH.FNextKeyId(KeyId))  {
//			TFltUInt64Pr Val = SimilarityFolderIdH.GetKey(KeyId);
//			if (Val.Val1 > MaxVal.Val1)
//				MaxVal = Val;
//		}
//		MostSimilarFoldersV.Add(TUInt64IntKd(MaxVal.Val2, (int)100*MaxVal.Val1));
//		SimilarityFolderIdH.DelKey(MaxVal);
//	}
//
//	PRecSet RecSet = TRecSet::New(FolderStoreId, MostSimilarFoldersV, true);
//	return RecSet;
//}

// find users who seem to have similar interests as user with UserId
//PRecSet TUserBase::FindSimilarUsers(const uint64 UserId, int MaxUserCount)
//{
//	PRecSet Folders = GetFoldersForUser(UserId);
//
//	THash<TUInt64, TInt> UsersToScoresH;
//	for (int FolderN=0; FolderN < Folders->GetRecs(); FolderN++) {
//		PRecSet SimilarFolders = FindSimilarFolders(Folders->GetRecId(FolderN), 50);
//		for (int SimFN=0; SimFN < SimilarFolders->GetRecs(); SimFN++) {
//			uint64 SimFId = SimilarFolders->GetRecId(SimFN);
//			uint64 UserId = FolderStore->GetFolderOwner(SimFId);
//			int SimFq = SimilarFolders->GetRecFq(SimFN);
//			UsersToScoresH.AddDat(UserId, UsersToScoresH.IsKey(UserId) ? UsersToScoresH.GetDat(UserId) + SimFq : SimFq);
//		}
//	}
//	UsersToScoresH.SortByDat(false);
//
//	TUInt64IntKdV MostSimilarUsersV;
//	int KeyId = UsersToScoresH.FFirstKeyId();
//	while (UsersToScoresH.FNextKeyId(KeyId) && MostSimilarUsersV.Len() < MaxUserCount) {
//		TUInt64 UserId = UsersToScoresH.GetKey(KeyId);
//		MostSimilarUsersV.Add(TUInt64IntKd(UserId, UsersToScoresH.GetDat(UserId)));
//	}
//	PRecSet RecSet = TRecSet::New(UserStoreId, MostSimilarUsersV, true);
//	return RecSet;
//}

//////////////////////////////////////
// helper functions

void TUserBase::GetFolderConcepts(const uint64& FolderId, const bool& SortByWgts, TUInt64FltKdV& ConceptV)
{
	PConceptFolder Folder = GetFolder(FolderId);
	if (Folder.Empty())
		return;
	
	for (int N=0; N < Folder->ConceptIdV.Len(); N++) {
		uint64 ConceptId = Folder->ConceptIdV[N];
		TFlt Wgt = Folder->ConceptWgtV[N];
		ConceptV.Add(TUInt64FltKd(ConceptId, Wgt));
	}
	if (SortByWgts)
		ConceptV.SortCmp(TCmpKeyDatByDat<TUInt64, TFlt>(false));
}

void TUserBase::PrintRecSet(const PRecSet& RecSet, const int& PrintCount)
{
	NotifyVerboseFmt(3, "Printing RecSet. Recs= %d", PrintCount);
	for (int RecN=0; RecN < PrintCount && RecN < RecSet->GetRecs(); RecN++)
		NotifyVerboseFmt(3, "RecId = %I64u, RecWgt = %d", RecSet->GetRecId(RecN), RecSet->GetRecFq(RecN));
}

void TUserBase::PrintFolderDefinition(const uint64& FolderId, const int& NumOfConcepts)
{
	if (!ConceptFolderStore->IsRecId(FolderId))
		return;
	const TStr FolderUri = ConceptFolderStore->GetRecNm(FolderId);
	if (!ConceptStore->IsRecNm(FolderUri))
		return;
	const uint64 ConceptId = ConceptStore->GetRecId(FolderUri);
	PConceptFolder Folder = GetFolder(FolderId);
	TUInt64FltKdV ConceptV;
	GetFolderConcepts(FolderId, true, ConceptV);
	
	NotifyVerbose(3, "-------------------");
	//NotifyVerboseFmt(3, "Folder name: %s", ConceptStore->GetFieldNmStr(ConceptId, "Label").CStr());
	NotifyVerboseFmt(3, "Number of concepts: %d", Folder->ConceptIdV.Len());
	NotifyVerbose(3, "Top folder concepts:");
	for (int N=0; N < ConceptV.Len() && N < 10; N++)
		NotifyVerboseFmt(3, "%.2f %s (fixed=%s)", ConceptV[N].Dat, GetConcept((int) ConceptV[N].Key).CStr(), TBool::GetStr(Folder->IsConceptFixed((int)ConceptV[N].Key)).CStr());
}

//PRecSet TUserBase::FindUsers(const TStr& Prefix, int MaxCount)
//{
//	TStr PrefixStr = Prefix.GetLc();
//	PIndexVoc IndexVoc = Base->GetIndexVoc();
//	TVec<TStr> WordStrV;
//	// find all the matching words
//	const uint64 Words = IndexVoc->GetWords(UserNameKWdsKeyId);
//	for (uint64 WordId = 0; WordId < Words; WordId++) {
//		TStr WordStr = IndexVoc->GetWordStr(UserNameKWdsKeyId, WordId);
//		WordStr = WordStr.GetLc();
//		if (PrefixStr.Empty() || WordStr.IsPrefix(PrefixStr) || WordStr.SearchStr(" " + PrefixStr, 0) != -1) {
//			//const uint64 WordFq = IndexVoc->GetWordFq(UserNameKWdsKeyId, WordId);
//			WordStrV.Add(WordStr);
//			if (WordStrV.Len() > MaxCount)
//				break;
//		}
//	}
//	// sort and limit
//	WordStrV.Sort(true);
//	WordStrV.Trunc(10);
//	
//	TUInt64Set UserIdSet;
//	for (int N=0; N < WordStrV.Len(); N++)
//	{
//		TStr WordStr = WordStrV[N];
//		TUInt64V WordIdV; IndexVoc->AddWordIdV(UserNameKWdsKeyId, WordStr, WordIdV);
//		PQuery Query = TQuery::New(Base, TQueryItem(Base, UserNameKWdsKeyId, WordStr));
//		// get result set
//		PRecSet Set = Base->Search(Query);
//		for (int N=0; N < Set->GetRecs(); N++)
//			UserIdSet.AddKey(Set->GetRecId(N));
//	}
//	TUInt64V UserIdV;
//	UserIdSet.GetKeyV(UserIdV);
//	UserIdV.Trunc(10);
//	PRecSet Users = TRecSet::New(UserStoreId, UserIdV);
//	return Users;
//}
//
//PRecSet TUserBase::FindFolders(const TStr& Prefix, int MaxCount)
//{
//	TStr PrefixStr = Prefix.GetLc();
//	PIndexVoc IndexVoc = Base->GetIndexVoc();
//	TVec<TStr> WordStrV;
//	// find all the matching words
//	const uint64 Words = IndexVoc->GetWords(FolderNameKWdsKeyId);
//	for (uint64 WordId = 0; WordId < Words; WordId++) {
//		TStr WordStr = IndexVoc->GetWordStr(FolderNameKWdsKeyId, WordId);
//		WordStr = WordStr.GetLc();
//		if (PrefixStr.Empty() || WordStr.IsPrefix(PrefixStr) || WordStr.SearchStr(" " + PrefixStr, 0) != -1) {
//			//const uint64 WordFq = IndexVoc->GetWordFq(FolderNameKWdsKeyId, WordId);
//			WordStrV.Add(WordStr);
//			if (WordStrV.Len() > MaxCount)
//				break;
//		}
//	}
//	// sort and limit
//	WordStrV.Sort(true);
//	WordStrV.Trunc(10);
//	
//	TUInt64Set FolderIdSet;
//	for (int N=0; N < WordStrV.Len(); N++)
//	{
//		TStr WordStr = WordStrV[N];
//		TUInt64V WordIdV; IndexVoc->AddWordIdV(FolderNameKWdsKeyId, WordStr, WordIdV);
//		PQuery Query = TQuery::New(Base, TQueryItem(Base, FolderNameKWdsKeyId, WordStr));
//		// get result set
//		PRecSet Set = Base->Search(Query);
//		for (int N=0; N < Set->GetRecs(); N++)
//			FolderIdSet.AddKey(Set->GetRecId(N));
//	}
//	TUInt64V FolderIdV;
//	FolderIdSet.GetKeyV(FolderIdV);
//	FolderIdV.Trunc(10);
//	PRecSet Folders = TRecSet::New(FolderStoreId, FolderIdV);
//	return Folders;
//}


// notifications
void TUserBase::NotifyVerbose(const int VerbosityLevel, const char *Str) 
{
	if (Logger != NULL && VerbosityLevel >= VerbosityLevel)
		Logger->NotifyVerbose(VerbosityLevel, Str); 
}

void TUserBase::NotifyVerboseFmt(const int VerbosityLevel, const char *FmtStr, ...) 
{
	if (Logger != NULL && VerbosityLevel >= VerbosityLevel) {
		va_list valist; va_start(valist, FmtStr);
		const int RetVal = vsnprintf(NotifyBuf, 100*1024-2, FmtStr, valist);
		if (RetVal < 0) return;
		Logger->NotifyVerbose(VerbosityLevel, NotifyBuf);
		va_end(valist); 
	}
}