/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "UserSrv.h"
#include "UserBase.h"

///////////////////////////////////////////////
// PlusMachines-Server-Helper-Utilities
TUserSrvUtil::TUserSrvUtil(const PUserBase& _UserBase): UserBase(_UserBase) { 
	UserStore = UserBase->UserStore;
	ConceptFolderStore = UserBase->ConceptFolderStore;
}

PJsonVal TUserSrvUtil::Folders(const PRecSet& FolderSet, const bool& IncludeFullInfo) const
{
	PJsonVal ArrayJson = TJsonVal::NewArr();
	int Recs = FolderSet->GetRecs();
	for (int N = 0; N < Recs; N++)
		ArrayJson->AddToArr(Folder(FolderSet->GetRecId(N), IncludeFullInfo));
	return ArrayJson;
}

PJsonVal TUserSrvUtil::User(const TStr& UserUri, const bool& IncludeFolderList, const bool& IncludeFullFolderInfo) const
{
	return User(UserStore->GetRecId(UserUri), IncludeFolderList, IncludeFullFolderInfo);
}

PJsonVal TUserSrvUtil::User(const uint64& UserId, const bool& IncludeFolderList, const bool& IncludeFullFolderInfo) const
{
	if (!UserStore->IsRecId(UserId))
		return TJsonVal::NewObj();
	PJsonVal ResJson = TJsonVal::NewObj("userUri", UserStore->GetRecNm(UserId));
	ResJson->AddToObj("name", UserStore->GetFieldNmStr(UserId, "Name"));
	if (IncludeFolderList) {
		PRecSet FolderSet = UserBase->GetFoldersForUser(UserId, false);
		TJsonValV FolderJsonV;
		for (int N = 0; N < FolderSet->GetRecs(); N++)
			FolderJsonV.Add(Folder(FolderSet->GetRecId(N), IncludeFullFolderInfo));
		ResJson->AddToObj("folders", FolderJsonV);
	}
	return ResJson;
}

PJsonVal TUserSrvUtil::Folder(const uint64& ConceptFolderId, const bool& IncludeFullInfo) const
{
	if (!ConceptFolderStore->IsRecId(ConceptFolderId))
		return TJsonVal::NewObj();
	PJsonVal ResJson;
	if (IncludeFullInfo) {
		PConceptFolder Folder = TConceptFolder::New(ConceptFolderId, UserBase->ConceptFolderStore);
		ResJson = Folder->GetJson();
		// replace ids with uris
		UserBase->AddUrisForFolderJson(ResJson, true);
	}
	else {
		ResJson = TJsonVal::NewObj("uri", UserBase->ConceptFolderStore->GetRecNm(ConceptFolderId));
	}
	TStr ConceptUri = UserBase->ConceptFolderStore->GetRecNm(ConceptFolderId);
	if (UserBase->ConceptStore->IsRecNm(ConceptUri)) {
		uint64 ConceptId = UserBase->ConceptStore->GetRecId(ConceptUri);
		ResJson->AddToObj("label", UserBase->ConceptStore->GetFieldNmStr(ConceptId, "Label"));
	}
	return ResJson;
}

PJsonVal TUserSrvUtil::Users(const PRecSet& UserSet, const bool& IncludeFolderList, const bool& IncludeFullFolderInfo) const
{
	TJsonValV UserJsonV;
	for (int UserN = 0; UserN < UserSet->GetRecs(); UserN++) 
		UserJsonV.Add(User(UserSet->GetRecId(UserN), IncludeFolderList, IncludeFullFolderInfo));
	return TJsonVal::NewArr(UserJsonV);
}

///////////////////////////////////////////////
// PlusMachines-Server-Folder

TStr TUsersSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		// creating/modifying user
		if (Action == "addUser")
			ResJson = AddUser(FldNmValPrV);
		else if (Action == "setUserName")
			ResJson = SetUserName(FldNmValPrV);
		else if (Action == "getUserProfile")
			ResJson = GetUserProfile(FldNmValPrV);
		
		// folder functions
		else if (Action == "addFolder")
			ResJson = AddFolder(FldNmValPrV);
		else if (Action == "deleteFolder")
			ResJson = DeleteFolder(FldNmValPrV);
		else if (Action == "getFolderJson")
			ResJson = GetFolderJson(FldNmValPrV);
		else if (Action == "updateFolder")
			ResJson = UpdateFolder(FldNmValPrV);
		else if (Action == "addConceptToFolder")
			ResJson = AddConceptToFolder(FldNmValPrV);
		else if (Action == "removeConceptFromFolder")
			ResJson = RemoveConceptFromFolder(FldNmValPrV);
		
		// editing interests, relevant sources, relevant locations
		else if (Action == "updateFolderWithArticleConcepts")
			ResJson = UpdateFolderWithArticleConcepts(FldNmValPrV);

		// search functions
		/*else if (Action == "findUsers")
			ResJson = FindUsers(FldNmValPrV);
		else if (Action == "findFolders")
			ResJson = FindFolders(FldNmValPrV);
		else if (Action == "findSimilarFolders")
			ResJson = FindSimilarFolders(FldNmValPrV);*/

		// following friends
		else if (Action == "followAdd")
			ResJson = FollowAdd(FldNmValPrV);
		else if (Action == "followRemove")
			ResJson = FollowRemove(FldNmValPrV);

		// admin functionality (concept folders, ...)
		//else if (Action == "adminOptions")
		//	ResJson = AdminOptions(FldNmValPrV);
		else if (Action == "users")
			ResJson = Users(FldNmValPrV);
		else if (Action == "saveUserDataToFile")
			ResJson = SaveUserDataToFile(FldNmValPrV);
		else if (Action == "loadUserDataFromFile")
			ResJson = LoadUserDataFromFile(FldNmValPrV);
		else if (Action == "clearUserData")
			ResJson = ClearUserData(FldNmValPrV);
		else if (Action == "getStats")
			ResJson = GetStats(FldNmValPrV);
		else if (Action == "printData")
			ResJson = PrintData(FldNmValPrV);
		else {
			ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("No valid actions were provided"));
			TQm::TEnv::Logger->OnStatus("invalid action name for user: " + Action);
		}

		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj();
		ResJson->AddToObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TQm::TEnv::Logger->OnStatus("Error: " + E->GetMsgStr());
	}

	// done
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}


//////////////////////////////////////
// creating/modifying user
PJsonVal TUsersSrv::AddUser(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr UserName = GetFldVal(FldNmValPrV, "userName", "");
	const bool CreateHome = GetFldBool(FldNmValPrV, "createHome", true);

	uint64 UserId;
	// new user
	if (!UserBase->UserStore->IsRecNm(UserUri)) {
		UserId = UserBase->AddUser(UserUri, UserName);
		if (CreateHome) {
			uint64 ConceptFolderId = UserBase->AddFolder(UserId, HomeFolderName, PrivateFolder);
			UserBase->SetAsHomeFolder(UserId, ConceptFolderId);
		}
	}
	// existing user
	else
		UserId = UserBase->UserStore->GetRecId(UserUri);
	return UserSrvUtil->User(UserId, true);
}

PJsonVal TUsersSrv::SetUserName(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr UserName = GetFldVal(FldNmValPrV, "userName", "");
	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));

	UserBase->SetUserName(UserUri, UserName);
	return UserSrvUtil->User(UserUri);
}

PJsonVal TUsersSrv::GetUserProfile(const TStrKdV& FldNmValPrV)
{
	TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	uint64 UserId = UserBase->UserStore->GetRecId(UserUri);
	
	// user info + list of all his folders
	PJsonVal ResJson = UserSrvUtil->User(UserUri, true);

	// add the definition of the home folder (likes, interests, ...)
	uint64 HomeFolderId = UserBase->GetHomeFolderId(UserId);
	if (HomeFolderId != TUInt64::Mx)
		ResJson->AddToObj("homeFolder", UserBase->GetFolderJson(HomeFolderId));
		
	//// add the last X liked articles. used to show the correct icon on the articles
	//PRecSet LikedArticlesRecSet = UserBase->Base->Search(TQuery::New(UserBase->Base, TQueryItem(UserBase->Base, UserBase->LikedArticlesKeyId, HomeFolderId, oqctEqual)));
	//LikedArticlesRecSet->SortById(false);
	//LikedArticlesRecSet->Trunc(100);	// choose the number of liked articles to return
	//TIntV ArticleIdV;
	//for (int N=0; N < LikedArticlesRecSet->GetRecs(); N++)
	//	ArticleIdV.Add((int) LikedArticlesRecSet->GetRecId(N));
	//PJsonVal LikedArticleIdArr = TJsonVal::NewArr(ArticleIdV);
	//ResJson->AddToObj("likedArticles", LikedArticleIdArr);
	return ResJson;
}

//////////////////////////////////////
// folder functions
PJsonVal TUsersSrv::AddFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FolderName = GetFldVal(FldNmValPrV, "folderName", "");
	const TStr FolderStr = GetFldVal(FldNmValPrV, "folder", "{}");
	const FolderVisibility Visibility = (FolderVisibility) GetFldInt(FldNmValPrV, "visibility", (int) PrivateFolder);

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	uint64 UserId = UserBase->UserStore->GetRecId(UserUri);

	if (FolderName == "")
		return TJsonVal::NewObj("error", TStr::Fmt("Invalid folder name: '%s'", FolderName.CStr()));

	uint64 FolderId = UserBase->AddFolder(UserId, FolderName, Visibility);
	if (FolderId == TUInt64::Mx)
		return TJsonVal::NewObj("error", TStr("Failed to create a folder"));

	// if we got the "folder" then update the new folder with the content from it
	if (IsFldNm(FldNmValPrV, "folder")) {
		PJsonVal FolderJson = TJsonVal::GetValFromStr(FolderStr);
		UserBase->AddIdsForFolderJson(FolderJson);
		PConceptFolder Folder = TConceptFolder::New(FolderJson);
		Folder->Uri = UserBase->ConceptFolderStore->GetRecNm(FolderId);
		UserBase->UpdateFolder(Folder);
	}
	
	return UserSrvUtil->Folder(FolderId, true);
}

PJsonVal TUsersSrv::UpdateFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FolderStr = GetFldVal(FldNmValPrV, "folder", "{}");

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	uint64 UserId = UserBase->UserStore->GetRecId(UserUri);

	PJsonVal FolderJson = TJsonVal::GetValFromStr(FolderStr);
	// change uris to ids
	UserBase->AddIdsForFolderJson(FolderJson);
	PConceptFolder Folder = TConceptFolder::New(FolderJson);
	if (! UserBase->ConceptFolderStore->IsRecNm(Folder->Uri))
		return TJsonVal::NewObj("error", TStr::Fmt("Concept folder '%s' is not known", Folder->Uri.CStr()));
	
	UserBase->UpdateFolder(Folder);

	// update the label in case it changed
	TStr Label = FolderJson->GetObjStr("label", "");
	if (Label != "" && UserBase->ConceptStore->IsRecNm(Folder->Uri))
		UserBase->UpdateConcept(Folder->Uri, Label);
	return TJsonVal::NewObj("info", TStr("Folder updated"));
}

// add a particular concept to the folder
PJsonVal TUsersSrv::AddConceptToFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");
	const TStr ConceptUri = GetFldVal(FldNmValPrV, "conceptUri", "");

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	const uint64 UserId = UserBase->UserStore->GetRecId(UserUri);
	uint64 FolderId = TUInt64::Mx;
	if (UserBase->ConceptFolderStore->IsRecNm(FolderUri))
		FolderId = UserBase->ConceptFolderStore->GetRecId(FolderUri);
	else
		FolderId = UserBase->GetHomeFolderId(UserId);
	if (UserId != UserBase->GetFolderOwner(UserBase->ConceptFolderStore->GetRecNm(FolderId)))
		return TJsonVal::NewObj("error", TStr::Fmt("User '%s' is not the owner of the folder '%s'. Unable to modify the folder", UserUri.CStr(), FolderUri.CStr()));

	int ConceptId = UserBase->AddConcept(ConceptUri);

	PConceptFolder Folder = UserBase->GetFolder(FolderId);
	Folder->AddConcept(ConceptId);
	UserBase->UpdateFolder(Folder);

	return TJsonVal::NewObj("info", "concept was added to the folder");
}

// remove a particular concept from the folder
PJsonVal TUsersSrv::RemoveConceptFromFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");
	const TStr ConceptUri = GetFldVal(FldNmValPrV, "conceptUri", "");

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	const uint64 UserId = UserBase->UserStore->GetRecId(UserUri);
	uint64 FolderId = TUInt64::Mx;
	if (UserBase->ConceptFolderStore->IsRecNm(FolderUri))
		FolderId = UserBase->ConceptFolderStore->GetRecId(FolderUri);
	else
		FolderId = UserBase->GetHomeFolderId(UserId);
	if (UserId != UserBase->GetFolderOwner(UserBase->ConceptFolderStore->GetRecNm(FolderId)))
		return TJsonVal::NewObj("error", TStr::Fmt("User '%s' is not the owner of the folder '%s'. Unable to modify the folder", UserUri.CStr(), FolderUri.CStr()));

	int ConceptId = UserBase->AddConcept(ConceptUri);

	PConceptFolder Folder = UserBase->GetFolder(FolderId);
	Folder->RemoveConcept(ConceptId);
	UserBase->UpdateFolder(Folder);

	return TJsonVal::NewObj("info", "concept was remved from the folder");
}

PJsonVal TUsersSrv::DeleteFolder(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");
	if (UserBase->VerbosityLevel >= 2)
		//UserBase->NotifyVerboseFmt(2, "Remove folder %I64u for user %I64u", FolderId, UserRecId);

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	uint64 UserId = UserBase->UserStore->GetRecId(UserUri);

	if (!UserBase->ConceptFolderStore->IsRecNm(FolderUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Concept folder with uri '%s' does not exist", FolderUri.CStr()));
	if (UserId != UserBase->GetFolderOwner(FolderUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User '%s' is not the owner of the folder '%s'. Unable to delete the folder", UserUri.CStr(), FolderUri.CStr()));
		
	UserBase->DeleteFolder(UserBase->ConceptFolderStore->GetRecId(FolderUri));
	
	return TJsonVal::NewObj("info", TStr("Folder was successfully removed"));
}

PJsonVal TUsersSrv::GetFolderJson(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	const uint64 UserId = UserBase->UserStore->GetRecId(UserUri);

	uint64 FolderId = TUInt64::Mx;
	if (UserBase->ConceptFolderStore->IsRecNm(FolderUri)) {
		if (UserId != UserBase->GetFolderOwner(FolderUri))
			return TJsonVal::NewObj("error", TStr::Fmt("User '%s' is not the owner of the folder '%s'. Unable to get the folder", UserUri.CStr(), FolderUri.CStr()));
		FolderId = UserBase->ConceptFolderStore->GetRecId(FolderUri);
	}
	else
		FolderId = UserBase->GetHomeFolderId(UserId);

	if (!UserBase->ConceptFolderStore->IsRecId(FolderId))
		return TJsonVal::NewObj("error", TStr("Unable to get the concept folder for provided folderUri or the home folder."));

	return UserSrvUtil->Folder(FolderId, true);
}

//////////////////////////////////////
// editing interests, relevant sources, relevant locations
PJsonVal TUsersSrv::UpdateFolderWithArticleConcepts(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");
	PJsonVal ArticleJson = TJsonVal::GetValFromStr(GetFldVal(FldNmValPrV, "article", "{}"));

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	uint64 UserId = UserBase->UserStore->GetRecId(UserUri);

	if (!UserBase->ConceptFolderStore->IsRecNm(FolderUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Concept folder with uri '%s' does not exist", FolderUri.CStr()));
	if (UserId != UserBase->GetFolderOwner(FolderUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User '%s' is not the owner of the folder '%s'. Unable to delete the folder", UserUri.CStr(), FolderUri.CStr()));
		
	const uint64 FolderId = UserBase->ConceptFolderStore->GetRecId(FolderUri);
	TUInt64FltKdV EntityV, TopicV, FolderConceptV;
	
	// extract here the list of artice concept ids and add them to the folder
	GetArticleConceptsByType(ArticleJson, EntityV, TopicV, FolderConceptV, true);

	// take from article 3 most important annotations for each type
	TUInt64FltKdV ConceptV;
	const int NumOfConcepts = 3;
	EntityV.Trunc(NumOfConcepts); TopicV.Trunc(NumOfConcepts); 
	ConceptV.AddV(EntityV);
	ConceptV.AddV(TopicV);
	
	UserBase->UpdateFolderConceptsWithArticleConcepts(FolderId, ConceptV);
	UserBase->IncreaseArticleCount(FolderId);	// increase the number of articles in folder (used for computing importance of new concepts)
	
	return NULL;
}

//////////////////////////////////////
// search functions
//PJsonVal TUsersSrv::FindUsers(const TStrKdV& FldNmValPrV)
//{
//	EAssertR(IsFldNm(FldNmValPrV, "prefix"), "no prefix");
//	const TStr Prefix = GetFldVal(FldNmValPrV, "prefix");
//	if (UserBase->VerbosityLevel >= 2)
//		UserBase->NotifyVerboseFmt(2, "Finding users with prefix %s", Prefix.CStr());
//
//	PRecSet UserSet = UserBase->FindUsers(Prefix, 10);
//	PJsonVal ResJson = UserSrvUtil->Users(UserSet);
//	return ResJson;
//}
//
//PJsonVal TUsersSrv::FindFolders(const TStrKdV& FldNmValPrV)
//{
//	EAssertR(IsFldNm(FldNmValPrV, "prefix"), "no prefix");
//	const TStr Prefix = GetFldVal(FldNmValPrV, "prefix");
//	if (UserBase->VerbosityLevel >= 2)
//		UserBase->NotifyVerboseFmt(2, "Finding folders with prefix %s", Prefix.CStr());
//
//	PRecSet FolderSet = UserBase->FindFolders(Prefix, 10);
//	PJsonVal ResJson = UserSrvUtil->Folders(FolderSet);
//	return ResJson;
//}
//
//PJsonVal TUsersSrv::FindSimilarFolders(const TStrKdV& FldNmValPrV)
//{
//	const uint64 UserRecId = GetFolderOwner(FldNmValPrV);
//	const uint64 FolderId = GetUInt64IfIn(FldNmValPrV, "folderId", TUInt64::Mx);
//
//	const uint64 FolderRecId = UserBase->GetFolderRecId(UserRecId, FolderId);
//	if (FolderRecId == TUInt64::Mx) {
//		TQm::TEnv::Logger->OnNotify(ntErr, "Invalid folderId value");
//		return NULL;
//	}
//	if (UserBase->VerbosityLevel >= 2)
//		UserBase->NotifyVerboseFmt(2, "Finding similar folders to folder %I64u for user %I64u", FolderId, UserRecId);
//
//	PRecSet FolderSet = UserBase->FindSimilarFolders(FolderId, 20);
//	PJsonVal ResJson = UserSrvUtil->Folders(FolderSet);
//	return ResJson;
//}

//////////////////////////////////////
// following friends
PJsonVal TUsersSrv::FollowAdd(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FriendUri = GetFldVal(FldNmValPrV, "friendUri", "");
	
	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	if (!UserBase->UserStore->IsRecNm(FriendUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Friend with uri '%s' does not exist", FriendUri.CStr()));

	const uint64 UserId = UserBase->UserStore->GetRecId(UserUri);
	const uint64 FriendId = UserBase->UserStore->GetRecId(FriendUri);
	UserBase->FollowAdd(UserId, FriendId);
	return NULL;
}

PJsonVal TUsersSrv::FollowRemove(const TStrKdV& FldNmValPrV)
{
	const TStr UserUri = GetFldVal(FldNmValPrV, "userUri", "");
	const TStr FriendUri = GetFldVal(FldNmValPrV, "friendUri", "");

	if (!UserBase->UserStore->IsRecNm(UserUri))
		return TJsonVal::NewObj("error", TStr::Fmt("User with uri '%s' does not exist", UserUri.CStr()));
	if (!UserBase->UserStore->IsRecNm(FriendUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Friend with uri '%s' does not exist", FriendUri.CStr()));

	const uint64 UserId = UserBase->UserStore->GetRecId(UserUri);
	const uint64 FriendId = UserBase->UserStore->GetRecId(FriendUri);
	UserBase->FollowRemove(UserId, FriendId);
	return NULL;
}

//////////////////////////////////////
// admin functionality (concept folders, ...)

PJsonVal TUsersSrv::Users(const TStrKdV& FldNmValPrV)
{
	PRecSet UserSet = UserBase->UserStore->GetAllRecs();
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("users", UserSrvUtil->Users(UserSet));
	return ResJson;
}

// return a list of all concept folders
PJsonVal TUsersSrv::SaveUserDataToFile(const TStrKdV& FldNmValPrV)
{
	TStr Time = TTm::GetCurLocTm().GetYMDDashStr();
	
	TFOut FOut1(Time + "-users.txt", false);
	PStoreIter Iter = UserBase->UserStore->GetIter();
	while (Iter->Next()) {
		const uint64 UserId = Iter->GetRecId();
		const TStr UserUri = UserBase->UserStore->GetRecNm(UserId);
		PUser User = UserBase->GetUser(UserUri);
		PJsonVal UserJson = User->GetJsonForStore();
		FOut1.PutStrLn(TJsonVal::GetStrFromVal(UserJson));
	}

	TFOut FOut2(Time + "-folders.txt", false);
	Iter = UserBase->ConceptFolderStore->GetIter();
	while (Iter->Next()) {
		const uint64 FolderId = Iter->GetRecId();
		PConceptFolder Folder = UserBase->GetFolder(FolderId);
		PJsonVal FolderJson = Folder->GetJsonForStore();
		// add also the owner info
		TRec UserRec = UserBase->ConceptFolderStore->GetRec(FolderId).DoSingleJoin(UserBase->Base, "hasUser");
		FolderJson->AddToObj("hasUser", TStr::Fmt("%I64u", UserRec.GetRecId()));
		// add the folder name
		if (!UserBase->ConceptStore->IsRecNm(Folder->Uri))
			continue;
		uint64 ConceptId = UserBase->ConceptStore->GetRecId(Folder->Uri);
		TStr FolderName = UserBase->ConceptStore->GetFieldNmStr(ConceptId, "Label");
		FolderJson->AddToObj("FolderName", FolderName);
		FOut2.PutStrLn(TJsonVal::GetStrFromVal(FolderJson));
	}
	
	return TJsonVal::NewObj("info", TStr::Fmt("Saved user data to %s-users.txt and %s-folders.txt file.", Time.CStr(), Time.CStr()));
}

// restore concept folders from a file
PJsonVal TUsersSrv::LoadUserDataFromFile(const TStrKdV& FldNmValPrV)
{
	TStr UsersFileName = GetFldVal(FldNmValPrV, "fileName", "users.txt");
	TStr FoldersFileName = GetFldVal(FldNmValPrV, "fileName", "folders.txt");
	TChA Line;
	int UserCnt = 0, FolderCnt = 0;
	if (!TFile::Exists(UsersFileName)) {
		TQm::TEnv::Logger->OnStatus("Unable to find file: " + UsersFileName);
		return TJsonVal::NewObj("info", TStr::Fmt("Unable to locate %s.txt file", UsersFileName.CStr()));
	}
	else {
		TFIn Input(UsersFileName);
		while (Input.GetNextLn(Line)) {
			try {
				PJsonVal JsonData = TJsonVal::GetValFromStr(Line);
				UserBase->UserStore->AddRec(JsonData);
				UserCnt++;
			}
			catch (PExcept E) {
				TQm::TEnv::Logger->OnStatus("Exception while loading user data file: " + E->GetMsgStr());
			}
		}
	}
	
	if (!TFile::Exists(FoldersFileName)) {
		TQm::TEnv::Logger->OnStatus("Unable to find file: " + FoldersFileName);
		return TJsonVal::NewObj("info", TStr::Fmt("Unable to locate %s.txt file", FoldersFileName.CStr()));
	}
	else {
		TFIn Input(FoldersFileName);
		while (Input.GetNextLn(Line)) {
			try {
				PJsonVal JsonData = TJsonVal::GetValFromStr(Line);
				TUInt64 UserId = JsonData->GetObjStr("hasUser", "").GetUInt64(TUInt64::Mx);
				if (UserId == TUInt64::Mx)
					continue;
				TStr FolderName = JsonData->GetObjStr("FolderName", "");
				PConceptFolder Folder = TConceptFolder::New(JsonData);
				UserBase->AddFolder(UserId, FolderName, (FolderVisibility) (int) Folder->Visibility);
				UserBase->ConceptFolderStore->AddRec(JsonData);
				FolderCnt++;
			}
			catch (PExcept E) {
				TQm::TEnv::Logger->OnStatus("Exception while loading user data file: " + E->GetMsgStr());
			}
		}
	}
	return TJsonVal::NewObj("info", TStr::Fmt("Read %d users and %d folders from users.txt.", UserCnt, FolderCnt));
}

PJsonVal TUsersSrv::ClearUserData(const TStrKdV& FldNmValPrV)
{
	PStoreIter Iter = UserBase->ConceptFolderStore->GetIter();
	while (Iter->Next()) {
		const uint64 FolderId = Iter->GetRecId();
		UserBase->DeleteFolder(FolderId);
	}
	
	return TJsonVal::NewObj("info", "Cleared user data.");
}

PJsonVal TUsersSrv::GetStats(const TStrKdV& FldNmValPrV)
{
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("users", TUInt64(UserBase->UserStore->GetRecs()).GetStr());
	ResJson->AddToObj("folders", TUInt64(UserBase->ConceptFolderStore->GetRecs()).GetStr());
	return ResJson;
}

PJsonVal TUsersSrv::PrintData(const TStrKdV& FldNmValPrV)
{
	TDir::GenDir("./Print");
	TStr Prefix = "./Print/" + TTm::GetCurLocTm().GetYMDDashStr() + " ";

	PJsonVal ResJson = TJsonVal::NewObj();
	
	TStrV PrintV; GetFldValV(FldNmValPrV, "data", PrintV);
	if (PrintV.IsIn("userStore"))
		UserBase->UserStore->PrintAll(UserBase->Base, Prefix + "UsersUserStore.txt");
	if (PrintV.IsIn("conceptFolderStore"))
		UserBase->ConceptFolderStore->PrintAll(UserBase->Base, Prefix + "UsersConceptFolderStore.txt");
	if (PrintV.IsIn("conceptStore"))
		UserBase->ConceptStore->PrintAll(UserBase->Base, Prefix + "UsersConceptStore.txt");
	if (PrintV.IsIn("index"))
		UserBase->Base->PrintIndex(Prefix + "UsersIndex.txt", true);
	if (PrintV.IsIn("indexVoc"))
		UserBase->Base->PrintIndexVoc(Prefix + "UsersIndexVoc.txt");

	return ResJson;
}

// use the hasConcept array in ArticleJson and extract from it the list of mentioned entities, topics, categories and conceptFolders
void TUsersSrv::GetArticleConceptsByType(const PJsonVal& ArticleJson, TUInt64FltKdV& EntityV, TUInt64FltKdV& TopicV, TUInt64FltKdV& ConceptFolderV, const bool& SortByWgts)
{
	PJsonVal ConceptJson = ArticleJson->GetObjKey("hasConcept");
	if (!ConceptJson.Empty()) {
		const int ConceptCnt = ConceptJson->GetArrVals();
		for (int N=0; N < ConceptCnt; N++) {
			PJsonVal Concept = ConceptJson->GetArrVal(N);
			//const int ConceptRecId = (int) Concept->GetObjNum("$id", -1);		// this id is not valid in this store. we have to use the uri and create our own id
			const TStr ConceptUri = Concept->GetObjStr("URI", "");
			const TStr ConceptType = Concept->GetObjStr("Type", "");
			const int ConceptId = UserBase->AddConcept(ConceptUri);
			const TFlt Fq = Concept->GetObjNum("$fq", 0);

			if (NewsConst.EntitySet.IsKey(ConceptType))
				EntityV.Add(TUInt64FltKd(ConceptId, Fq));
			else if (NewsConst.WikiSet.IsKey(ConceptType))
				TopicV.Add(TUInt64FltKd(ConceptId, Fq));
			else if (NewsConst.ConceptFolderSet.IsKey(ConceptType))
				ConceptFolderV.Add(TUInt64FltKd(ConceptId, Fq));
		}

		if (SortByWgts) {
			EntityV.SortCmp(TCmpKeyDatByDat<TUInt64, TFlt>(false));
			TopicV.SortCmp(TCmpKeyDatByDat<TUInt64, TFlt>(false));
			ConceptFolderV.SortCmp(TCmpKeyDatByDat<TUInt64, TFlt>(false));
		}
	}
}
