/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "AddArticleSrv.h"

//////////////////////////////////////
// TSendArticleToClustering - service for sending articles to the clustering service

TSendArticleToClustering::TSendArticleToClustering(const PNewsBase& _NewsBase, const TStr& _SettingsFName, const PNotify& _Notify) :
	NewsBase(_NewsBase), SettingsFName(_SettingsFName), Notify(_Notify)
{
	ArticleStore = NewsBase->ArticleStore;
	ConceptStore = NewsBase->ConceptStore;
	Base = NewsBase->Base;

	const TStr SettingsStr = TFile::Exists(SettingsFName) ? TStr::LoadTxt(SettingsFName) : "{}";
	const PJsonVal AllSettings = TJsonVal::GetValFromStr(SettingsStr);
	const PJsonVal Settings = AllSettings->IsObjKey("sendArticleToClustering") ? AllSettings->GetObjKey("sendArticleToClustering") : TJsonVal::NewObj();

	EntityWgt = Settings->GetObjNum("EntityWgt", 1.0);
	NonEntityWgt = Settings->GetObjNum("NonEntityWgt", 0.3);
	MultiplyWithConceptWgts = Settings->GetObjBool("MultiplyWithConceptWgts", false);

	Notify->OnStatus("TSendArticleToClustering started.");
}

void TSendArticleToClustering::SetStoryUrlV(const TStrV& _StoryUrlV, const TStr& WebReqFName) { 
	StoryUrlV = TStrV(_StoryUrlV);
	Notify->OnStatus("Setting story urls:");
	// initialize story url
	for (int StoryUrlN = 0; StoryUrlN < StoryUrlV.Len(); StoryUrlN++) {
		TStr LangStoryUrl = StoryUrlV[StoryUrlN];
		Assert(LangStoryUrl.IsChIn('@'));
		TStr Langs, StoryUrl; LangStoryUrl.SplitOnCh(Langs, '@', StoryUrl);
		TStrV LangV; Langs.SplitOnAllCh(',', LangV);
		for (int LangN=0; LangN < LangV.Len(); LangN++)
			LangStoryUrlH.AddDat(LangV[LangN], StoryUrl);
		Notify->OnStatusFmt("Send languages %s to url: %s", Langs.CStr(), StoryUrl.CStr());
		// create a fetcher for each clustering service. this way if one falls down we can still send things to others
		StoryUrlToWebPgFetch.AddDat(StoryUrl, TClustReqPersist::New(Langs + " ", WebReqFName + "-" + Langs + ".bin", true, true, Notify));
	}
}

// send a new article to the clustering service
void TSendArticleToClustering::SendArticle(const uint64& ArticleId)
{
	const TStr Lang = ArticleStore->GetFieldNmStr(ArticleId, "Language");
	const TStr NewsfeedId = ArticleStore->GetFieldNmStr(ArticleId, "NewsfeedId");
	
	const TStr LangKey = LangStoryUrlH.IsKey(Lang) ? Lang : "default";
	TStr ClusteringUrl = LangStoryUrlH.IsKey(LangKey) ? LangStoryUrlH.GetDat(LangKey) : "";
	PClustReqPersist ClustReq = (LangStoryUrlH.IsKey(LangKey) && StoryUrlToWebPgFetch.IsKey(LangStoryUrlH.GetDat(LangKey))) ?
		StoryUrlToWebPgFetch.GetDat(LangStoryUrlH.GetDat(LangKey)) : NULL;
	
	if (ClusteringUrl != "" && !ClustReq.Empty()) {
		try {	
			const TStr Title = ArticleStore->GetFieldNmStr(ArticleId, "Title");
			TStr Body = ArticleStore->GetFieldNmStr(ArticleId, "Body");
			// we shorten the text if needed - otherwise the clustering could take too much time and we would always get a timeout
			if (Body.Len() > 50000)
				Body = Body.GetSubStr(0, 50000);
			TTm Tm; ArticleStore->GetFieldNmTm(ArticleId, "DateTime", Tm);

			// forward for clustering into stories
			// assemble URL
			TChA ArticleUrl = ClusteringUrl;
			ArticleUrl += "?id=" + NewsfeedId;
			ArticleUrl += "&lang=" + TUrl::EncodeUrlStr(Lang);
			ArticleUrl += "&text=" + TUrl::EncodeUrlStr(Body);
			ArticleUrl += "&title=" + TUrl::EncodeUrlStr(Title);

			uint64 TimeMS = TSysTm::GetMSecsFromTm(Tm);
			// get seconds since 1970/1/1
			uint64 TimeStartMS = TTm::GetMSecsFromTm(TTm(1970,1,1));
			double DiffS = (TimeMS - TimeStartMS) / 1000;
			TFlt DayTimeStamp = DiffS / 86400.0;
			ArticleUrl += "&date="; ArticleUrl += TUrl::EncodeUrlStr(TFlt::GetStr(DayTimeStamp, "%.3f"));

			// if we also want to send concepts then add them
			if (EntityWgt > NewsConst.Eps || NonEntityWgt > NewsConst.Eps) {
				PRecSet ConceptRecSet = ArticleStore->GetRec(ArticleId).DoJoin(Base, "hasConcept");
				const int ConcCount = ConceptRecSet->GetRecs();
				TStrV ConcIdStrV, ConcWgtStrV; 
				for (int N=0; N < ConcCount; N++) {
					TStr Type = ConceptStore->GetFieldNmStr(ConceptRecSet->GetRecId(N), "Type");
					if (NewsConst.IsEntity(Type)) {
						if (EntityWgt > NewsConst.Eps) {
							ConcIdStrV.Add(TUInt64(ConceptRecSet->GetRecId(N)).GetStr());
							const double Wgt = MultiplyWithConceptWgts ? EntityWgt * ConceptRecSet->GetRecFq(N) : EntityWgt;
							ConcWgtStrV.Add(TFlt::GetStr(Wgt, "%.2f"));
						}
					}
					else {
						if (NonEntityWgt > NewsConst.Eps) {
							ConcIdStrV.Add(TUInt64(ConceptRecSet->GetRecId(N)).GetStr());
							const double Wgt = MultiplyWithConceptWgts ? EntityWgt * ConceptRecSet->GetRecFq(N) : NonEntityWgt;
							ConcWgtStrV.Add(TFlt::GetStr(Wgt, "%.2f"));
						}
					}
				}
			
				const TStr ConceptIdV = TStr::GetStr(ConcIdStrV, ",");
				const TStr ConceptWgtV = TStr::GetStr(ConcWgtStrV, ",");
				ArticleUrl += "&conceptIds="; ArticleUrl += TUrl::EncodeUrlStr(ConceptIdV);
				ArticleUrl += "&conceptWgts="; ArticleUrl += TUrl::EncodeUrlStr(ConceptWgtV);
			}

			// log the activity
			NewsBase->ActivityLogFile->NotifyInfo("Sending article %s ('%s') to clustering service", NewsfeedId.CStr(), ArticleStore->GetRecNm(ArticleId).CStr());
			
			ClustReq->FetchUrl(ArticleUrl);
		}
		catch (PExcept E) { 
			NewsBase->ErrorLogFile->NotifyErr(TStr::Fmt("TSendArticleToClustering::SendArticle. Exception occured while sending article %s", NewsfeedId.CStr()).CStr(), E);
		}
		catch (...) {
			NewsBase->ErrorLogFile->NotifyErr("TSendArticleToClustering::SendArticle. General exception.");
		}
	}
	else {
		Notify->OnNotify(ntWarn, "Unable to find the clustering service for article in language " + Lang);
	}
}

//////////////////////////////////////
// TAddArticleSrv - service that accepts new articles in the binary format

TAddArticleSrv::TAddArticleSrv(const PNewsBase& _NewsBase, const PSendArticleToClustering& _SendArticleToClustering) :
	TSrvFun(_NewsBase->Base, "add", saotXml), NewsBase(_NewsBase), SendArticleToClustering(_SendArticleToClustering)
{
}

PXmlDoc TAddArticleSrv::ExecXml(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	TUInt64V NewArticleIdV;
	try {
		TTmStopWatch Timer; Timer.Start();
		// get new articles from the HTTP request body
		PHttpRq HttpRq = RqEnv->GetHttpRq();
		TMem InMem; HttpRq->GetBodyAsMem(InMem);
		TMem BodyMem; TNewsMemEncoding::DecodeMem(InMem, BodyMem);
		PSIn NewsSIn = TMemIn::New(BodyMem);
		// parse articles from stream
		//TNewsProc::IndexArticleSIn(NewsBase, NewsSIn, NewArticleIdV);
		while (!NewsSIn->Eof()) {
			TNewsArticle Article(*NewsSIn);
			const uint64 ArticleId = NewsBase->AddArticle(Article);
			if (ArticleId != TUInt64::Mx) {
				NewArticleIdV.Add(ArticleId);
				const TStr NewsfeedId = NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "NewsfeedId");
				NewsBase->ActivityLogFile->NotifyInfo("New article %s ('%s') was added", NewsfeedId.CStr(), Article.ArticleUri.CStr());
			}
			else {
				//NewsBase->ActivityLogFile->NotifyErr("Adding a new article failed. New article id is invalid. Article url: '%s'", Article.ArticleUri.CStr());
				//NewsBase->ErrorLogFile->NotifyErr("Adding a new article failed. New article id is invalid. Article url: '%s'", Article.ArticleUri.CStr());
				
				// if something is wrong with the article we throw an exception. if nothing is wrong but we just don't want to add the article then we return Mx
				NewsBase->InfoLogFile->NotifyErr("Adding a new article failed. The new article already exists or the source is disabled. Url: '%s'", Article.ArticleUri.CStr());
			}
		}
		Timer.Stop();
		// notify about new load
		TNotify::StdNotify->OnStatusFmt("[%s] %d new records in %.4fs (last article id: %I64u)",
			GetFunNm().CStr(), NewArticleIdV.Len(), Timer.GetSec(), NewArticleIdV.Len() > 0 ? (uint64) NewArticleIdV[NewArticleIdV.Len() - 1] : 0);
	}
	catch (PExcept E) {
		NewsBase->ErrorLogFile->NotifyErr("TAddArticleSrv::ExecXml. Error loading records.", E);
		TNotify::StdNotify->OnStatusFmt("Error loading records: %s", E->GetMsgStr().CStr());
	}
	catch (...) {
		NewsBase->ErrorLogFile->NotifyErr("TAddArticleSrv::ExecXml. General error while loading records.");
		TNotify::StdNotify->OnStatus(TStr::Fmt("Error loading records."));
	}

	for (int NewArticleIdN = 0; NewArticleIdN < NewArticleIdV.Len(); NewArticleIdN++) {
		try {
			uint64 ArticleId = NewArticleIdV[NewArticleIdN];
			SendArticleToClustering->SendArticle(ArticleId);
		}
		catch (PExcept E) {
			NewsBase->ErrorLogFile->NotifyErr("TAddArticleSrv::ExecXml. Error contacting clustering service.", E);
			TNotify::StdNotify->OnStatusFmt("Error contacting clustering service: %s", E->GetMsgStr().CStr());
		}
	}
	// print current status regarding waiting queue of requests, successes and fails
	SendArticleToClustering->Report();
		
	// done
	return TXmlDoc::New(TXmlTok::New(GetFunNm(), "Ok"));
}


//////////////////////////////////////
// TAddArticleJsonSrv - service that accepts new articles as an arraj of json objects compatible with the news.def format

TAddArticleJsonSrv::TAddArticleJsonSrv(const PNewsBase& _NewsBase, const PSendArticleToClustering& _SendArticleToClustering) :
	TSrvFun(_NewsBase->Base, "addJson", saotXml), NewsBase(_NewsBase), SendArticleToClustering(_SendArticleToClustering)
{
}


// ////////////////////////////////
// WARNING WARNING WARNING WARNING
//
// adding articles as json object is not yet fully supported
// what is missing is creating parent categories for each category and importing the similar_X articles
// 
PXmlDoc TAddArticleJsonSrv::ExecXml(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	TUInt64V NewArticleIdV;
	try {
		TTmStopWatch Timer; Timer.Start();
		// get new articles from the HTTP request body
		PHttpRq HttpRq = RqEnv->GetHttpRq();
		TMem Body; HttpRq->GetBodyAsMem(Body);
		PSIn BodySIn = TMemIn::New(Body);
		
		PJsonVal ArticleArr = TJsonVal::GetValFromSIn(BodySIn);
		if (!ArticleArr->IsArr()) {
			TNotify::StdNotify->OnStatus("AddJson request did not contain an array of articles to be added.");
			NewsBase->ErrorLogFile->NotifyErr("TAddArticleJsonSrv::ExecXml. AddJson request did not contain an array of articles to be added.");
			return TXmlDoc::New(TXmlTok::New(GetFunNm(), "AddJson request did not contain an array of articles to be added."));
		}
		int Arts = ArticleArr->GetArrVals();
		for (int N = 0; N < Arts; N++) {
			PJsonVal ArticleJson = ArticleArr->GetArrVal(N);
			const uint64 ArticleId = NewsBase->AddArticle(ArticleJson);
			if (ArticleId != TUInt64::Mx) {
				const TStr NewsfeedId = NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "NewsfeedId");
				NewsBase->ActivityLogFile->NotifyInfo("New article %s ('%s') was added", NewsfeedId.CStr(), ArticleJson->GetObjStr("URI", "").CStr());
				NewArticleIdV.Add(ArticleId);
			}
			else {
				NewsBase->ActivityLogFile->NotifyErr("Adding a new article failed. New article id is invalid. Article url: '%s'", ArticleJson->GetObjStr("URI", "").CStr());
				NewsBase->ErrorLogFile->NotifyErr("Adding a new article failed. New article id is invalid. Article url: '%s'", ArticleJson->GetObjStr("URI", "").CStr());
			}
		}
		TNotify::StdNotify->OnStatusFmt("%d new articles (last id: %I64u)", NewArticleIdV.Len(), NewArticleIdV.Len() > 0 ? (uint64) NewArticleIdV[NewArticleIdV.Len() - 1] : 0);
		Timer.Stop();
		// notify about new load
		/*TNotify::StdNotify->OnStatusFmt("[%s] %d new records in %.4fs (last article id: %I64u)",
			GetFunNm().CStr(), NewArticleIdV.Len(), Timer.GetSec(), NewArticleIdV.Len() > 0 ? NewArticleIdV[NewArticleIdV.Len() - 1] : 0);*/
	}
	catch (PExcept E) {
		NewsBase->ErrorLogFile->NotifyErr("TAddArticleJsonSrv::ExecXml. Error loading records.", E);
		TNotify::StdNotify->OnStatusFmt("Error loading records: %s", E->GetMsgStr().CStr());
	}
	catch (...) {
		NewsBase->ErrorLogFile->NotifyErr("TAddArticleJsonSrv::ExecXml. General error while loading records.");
		TNotify::StdNotify->OnStatus(TStr::Fmt("Error loading records."));
	}

	for (int NewArticleIdN = 0; NewArticleIdN < NewArticleIdV.Len(); NewArticleIdN++) {
		try {
			uint64 ArticleId = NewArticleIdV[NewArticleIdN];
			SendArticleToClustering->SendArticle(ArticleId);
		}
		catch (PExcept Except) {
			TNotify::StdNotify->OnStatusFmt("Error contacting clustering service: %s", Except->GetMsgStr().CStr());
		}
	}
	// print current status regarding waiting queue of requests, successes and fails
	SendArticleToClustering->Report();

	// done
	return TXmlDoc::New(TXmlTok::New(GetFunNm(), "Ok"));
}