/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "AddStorySrv.h"
#include "../Utils/FileUtils.h"

TClustUtil::TClustUtil(TNewsBase* _NewsBase, TEventRegBase* _EventRegBase) : NewsBase(_NewsBase), EventRegBase(_EventRegBase)
{
}

//called by clustering when an article is assigned to a cluster
void TClustUtil::SetArticle(const TStr& ArticleName, const TStr& StoryName, const double& CosToCentroid)
{
	TTmStopWatch Sw; Sw.Start();
	const uint64 ArticleId = NewsBase->GetArticleIdForNewsfeedId(ArticleName);
	
	if (!NewsBase->ArticleStore->IsRecId(ArticleId)) {
		NewsBase->ErrorLogFile->NotifyErr("TClustUtil::SetArticle. Article id %s is not a known article name.", ArticleName.CStr());
		NewsBase->ActivityLogFile->NotifyErr("We received a cluster info for article %s that does not exist in the database.", ArticleName.CStr());
		return;
	}

	// if the user assigned the article to a story, then don't modify it's cluster membership
	if (NewsBase->ValidatedArticleLinks->HasStoryForArticle(ArticleId)) {
		const uint64 StoryId = NewsBase->StoryStore->IsRecNm(StoryName) ? NewsBase->StoryStore->GetRecId(StoryName) : TUInt64::Mx;
		NewsBase->InfoLogFile->NotifyVerboseFmt(3, "TClustUtil::SetArticle. Article %s was not assigned to story %I64u ('%s') since we have a user-set story for the article.", ArticleName.CStr(), StoryId, StoryName.CStr());
		return;
	}

	// NOTE: this is commented out since we never reassing a single article to a new story. when reassignment happens is when the cluster is split into two
	//if (EventRegBase != NULL) {
	//	// get the old storyid so that we can check if the old story needs to be removed...
	//	TRec OldStoryRec = NewsBase->ArticleStore->GetRec(ArticleId).DoSingleJoin(NewsBase->Base, "hasStory");
	//	if (OldStoryRec.IsDef()) {
	//		EventRegBase->OnRemoveArticleFromStory(OldStoryRec.GetRecId());
	//		EventRegBase->ConsiderUpdatingStoryInfo(OldStoryRec.GetRecId());
	//		if (NewsBase->StoryStore->GetRecNm(OldStoryRec.GetRecId()) == StoryName)
	//			TNotify::StdNotify->OnStatusFmt("TClustUtil:SetArticle. WARNING: The article %I64u is reassigned to the same story", ArticleId);
	//	}
	//}

	// remember the old story id - if set
	uint64 OldStoryId = TUInt64::Mx;
	if (NewsBase->ActivityLogFile->GetVerbosityLevel() >= 5)
		OldStoryId = NewsBase->ArticleStore->GetFieldNmUInt64(ArticleId, "hasStoryId");

	// create the story and create the joins to the article
	uint64 StoryId = NewsBase->AddStory(ArticleId, StoryName);
	
	NewsBase->InfoLogFile->NotifyVerboseFmt(3, "TClustUtil::SetArticle. Assigning article %s to story %I64u ('%s')", ArticleName.CStr(), StoryId, StoryName.CStr());
	NewsBase->InfoLogFile->IncreaseSpaces();
	NewsBase->ActivityLogFile->NotifyInfo("Article %s ('%s') was assigned to story %I64u ('%s')", ArticleName.CStr(), NewsBase->ArticleStore->GetRecNm(ArticleId).CStr(), StoryId, StoryName.CStr());

	if (OldStoryId != TUInt64::Mx && NewsBase->ActivityLogFile->GetVerbosityLevel() >= 5) {
		PRecSet ArtSet = NewsBase->StoryStore->GetRec(OldStoryId).DoJoin(NewsBase->Base, "hasArticle");
		NewsBase->ActivityLogFile->NotifyInfo("\tArticle %s was previously assigned to story %I64u which now contains %d articles.", ArticleName.CStr(), StoryId, ArtSet->GetRecs());
	}
	
	if (CosToCentroid >= 0)
		NewsBase->ArticleStore->SetFieldNmFlt(ArticleId, "CosToCentroid", CosToCentroid);
	EventRegBase->ConsiderUpdatingStoryInfo(StoryId);

	// update story activity
	NewsBase->NewsActivity->AddStoryActivity(StoryId);

	Sw.Stop();
	NewsBase->InfoLogFile->DecreaseSpaces();
	NewsBase->InfoLogFile->NotifyVerboseFmt(3, "TClustUtil:SetArticle. Finished. Time needed %.0f ms", Sw.GetMSec());
}

// clustering can notify us that a new article _ArticleId is a duplicate of some previously indexed article _OriginalArticleId
// the article _ArticleId is then ignored in the clustering
void TClustUtil::SetDuplicate(const TStr& ArticleName, const TStr& OriginalArticleName)
{
	const uint64 ArticleId = NewsBase->GetArticleIdForNewsfeedId(ArticleName);
	const uint64 OriginalArticleId = NewsBase->GetArticleIdForNewsfeedId(OriginalArticleName);
	
	if (!NewsBase->ArticleStore->IsRecId(ArticleId) || !NewsBase->ArticleStore->IsRecId(OriginalArticleId)) {
		NewsBase->ErrorLogFile->NotifyErr("TClustUtil::SetDuplicate. Article %s or original article %s are not valid names.", ArticleName.CStr(), OriginalArticleName.CStr());
		NewsBase->ActivityLogFile->NotifyErr("We received info about article duplicates with invalid parameters. Article %s or original article %s are not valid numbers.", ArticleName.CStr(), OriginalArticleName.CStr());
		return;
	}

	NewsBase->ActivityLogFile->NotifyInfo("Article %s ('%s') was found to be a duplicate of article %s ('%s')", ArticleName.CStr(), NewsBase->ArticleStore->GetRecNm(ArticleId).CStr(), OriginalArticleName.CStr(), NewsBase->ArticleStore->GetRecNm(OriginalArticleId).CStr());

	if (NewsBase->InfoLogFile->GetVerbosityLevel() >= 5) {
		NewsBase->InfoLogFile->NotifyVerboseFmt(5, "TClustUtil:SetDuplicate. -- Article %s was found to be a duplicate of article %s --", ArticleName.CStr(), OriginalArticleName.CStr());
		NewsBase->InfoLogFile->NotifyVerboseFmt(5, "TClustUtil:SetDuplicate. Content of article %s: %s", ArticleName.CStr(), NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "Body").GetSubStr(0, 100).CStr());
		NewsBase->InfoLogFile->NotifyVerboseFmt(5, "TClustUtil:SetDuplicate. Content of article %s: %s", OriginalArticleName.CStr(), NewsBase->ArticleStore->GetFieldNmStr(OriginalArticleId, "Body").GetSubStr(0, 100).CStr());
	}

	TStr Date = NewsBase->ArticleStore->GetFieldNmStr(ArticleId, "Date");

	// 
	// add join that will tell that the article is a duplicate
	NewsBase->ArticleStore->AddJoin("isDuplicateOf", ArticleId, OriginalArticleId);

	// 
	// if we've found the article to be a duplicate then we also want to decrease trending values for the concepts in the article
	PRecSet ConceptSet = NewsBase->ArticleStore->GetRec(ArticleId).DoJoin(NewsBase->Base, "hasConcept");
	int MinWgt = NewsBase->TrendsManager->MinConceptWgt;
	TConceptIdV ConceptV;
	for (int N = 0; N < ConceptSet->GetRecs(); N++) {
		if (ConceptSet->GetRecFq(N) >= MinWgt)
			ConceptV.Add((int) ConceptSet->GetRecId(N));
	}
	if (ConceptV.Len() > 0) {
		TTm Tm; NewsBase->ArticleStore->GetFieldNmTm(ArticleId, "DateTime", Tm);
		NewsBase->TrendsManager->RemoveConceptV(Tm, ConceptV);
	}
}

// we have new info about the cluster medoid, variance and average distance of articles to the centroid
void TClustUtil::SetClusterInfo(const TStr& MedoidArticleName, const TStr& StoryName, const TStrV& ArticleNameV, const TFltV& CosToCentroidV, const TFlt& Variance, const TFlt& AvgCosToCentroid)
{
	TTmStopWatch Sw; Sw.Start();
	
	const uint64 MedoidId = NewsBase->GetArticleIdForNewsfeedId(MedoidArticleName);
	// report current story status
	if (NewsBase->ActivityLogFile->GetVerbosityLevel() >= 5) {
		if (NewsBase->StoryStore->IsRecNm(StoryName)) {
			const uint64 StoryId = NewsBase->StoryStore->GetRecId(StoryName);
			PRecSet ArtSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasArticle");
			TUInt64V ArtIdV; ArtSet->GetRecIdV(ArtIdV);
			NewsBase->ActivityLogFile->NotifyInfo("====== Changing the articles for story %I64u ('%s') ======", StoryId, StoryName.CStr());
			NewsBase->ActivityLogFile->NotifyInfo("Story currently contains %d articles: %s", ArtIdV.Len(), TStr::GetStr(ArtIdV, ", ").CStr());
		}
		else
			NewsBase->ActivityLogFile->NotifyInfo("====== Creating a new story ======");
	}
	
	NewsBase->InfoLogFile->NotifyVerboseFmt(3, "TClustUtil:SetClusterInfo. Setting medoid %s, story %s, articleV length %d, cosToCentroidV length %d, variance %.3f, avgCosToCentroid %.3f", MedoidArticleName.CStr(), StoryName.CStr(), ArticleNameV.Len(), CosToCentroidV.Len(), Variance.Val, AvgCosToCentroid.Val);
	NewsBase->InfoLogFile->IncreaseSpaces();
	TUInt64Set OldStoryIdH;
	// assign all provided articles to StoryName
	// it's possible that the story is new and we will create a new record for it
	for (int ArtN = 0; ArtN < ArticleNameV.Len(); ArtN++) {
		const uint64 ArticleId = NewsBase->GetArticleIdForNewsfeedId(ArticleNameV[ArtN]);
		if (! NewsBase->ArticleStore->IsRecId(ArticleId)) 
			continue;
		// if the user assigned the article to a story, then don't modify it's cluster membership
		if (NewsBase->ValidatedArticleLinks->HasStoryForArticle(ArticleId)) {
			const uint64 StoryId = NewsBase->StoryStore->IsRecNm(StoryName) ? NewsBase->StoryStore->GetRecId(StoryName) : TUInt64::Mx;
			NewsBase->InfoLogFile->NotifyVerboseFmt(3, "TClustUtil::SetClusterInfo. Article %s was not assigned to story %I64u ('%s') since we have a user-set story for the article.", ArticleNameV[ArtN].CStr(), StoryId, StoryName.CStr());
			continue;
		}
		//debug. get old story id and then call event reg with the id. this should be redundant
		const TRec OldStoryRec = NewsBase->ArticleStore->GetRec(ArticleId).DoSingleJoin(NewsBase->Base, "hasStory");
		if (OldStoryRec.IsDef())
			OldStoryIdH.AddKey(OldStoryRec.GetRecId());

		// update story for article
		NewsBase->AddStory(ArticleId, StoryName);
	}

	uint64 StoryId = TUInt64::Mx;
	// report new story status
	if (NewsBase->StoryStore->IsRecNm(StoryName)) {
		StoryId = NewsBase->StoryStore->GetRecId(StoryName);
		PRecSet ArtSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasArticle");
		TUInt64V ArtIdV; ArtSet->GetRecIdV(ArtIdV);
		NewsBase->ActivityLogFile->NotifyInfo("Updated story %I64u ('%s') so that now contains %d articles: %s", StoryId, StoryName.CStr(), ArtIdV.Len(), TStr::GetStr(ArtIdV, ", ").CStr());
	}
	
	// story should exist by now. if it doesn't then we obviously don't have any articles for it and we can exit
	if (! NewsBase->StoryStore->IsRecNm(StoryName)) {
		NewsBase->ErrorLogFile->NotifyErr("TClustUtil::SetClusterInfo. Setting cluster info for unknown story '%s'", StoryName.CStr());
		TNotify::StdNotify->OnStatusFmt("TClustUtil::SetClusterInfo. ERROR. Setting cluster info for unknown story '%s'", StoryName.CStr());
		NewsBase->InfoLogFile->DecreaseSpaces();
		return;
	}
		
	// call event reg to notify it that we've removed some articles from some stories
	// if we find that the new articles are assigned to the same story as before then don't call the EventRegBase->OnRemoveArticleFromStory
	OldStoryIdH.DelIfKey(StoryId);
	OldStoryIdH.Defrag();
	for (int KeyId=OldStoryIdH.FFirstKeyId(); OldStoryIdH.FNextKeyId(KeyId); )
		EventRegBase->OnRemoveArticleFromStory(OldStoryIdH.GetKey(KeyId));
			
	// set new distances from centroid for each article in the story
	if (ArticleNameV.Len() == CosToCentroidV.Len()) {
		for (int N = 0; N < ArticleNameV.Len(); N++) {
			const double CosSim = CosToCentroidV[N];
			const uint64 ArticleId = NewsBase->GetArticleIdForNewsfeedId(ArticleNameV[N]);
			if (!NewsBase->ArticleStore->IsRecId(ArticleId)) {
				NewsBase->ErrorLogFile->NotifyErr("TClustUtil::SetClusterInfo. Article %s is not a known article.", ArticleNameV[N].CStr());
				continue;
			}
			// NOTE: we should not set the cosine similarity of an article that was set by user to another story. 
			// if the user assigned the article to a story, then don't change it's cosine similarity to cluster
			/*if (NewsBase->ValidatedArticleLinks->HasStoryForArticle(ArticleId))
				continue;*/
			NewsBase->ArticleStore->SetFieldNmFlt(ArticleId, "CosToCentroid", CosSim);
		}
	}
	else {
		NewsBase->ErrorLogFile->NotifyErr("TClustUtil:SetClusterInfo. The lists of articles and list of cos to centroid were of different length. Ignoring their information.");
		TNotify::StdNotify->OnStatus("TClustUtil:SetClusterInfo. ERROR. The lists of articles and list of cos to centroid were of different length. Ignoring their information.");
	}

	// update the cluster variance and avg dist from centroid	
	if (AvgCosToCentroid >= 0)
		NewsBase->StoryStore->SetFieldNmFlt(StoryId, "AvgCosToCentroid", AvgCosToCentroid);
	if (Variance >= 0)
		NewsBase->StoryStore->SetFieldNmFlt(StoryId, "Variance", Variance);
	
	// set new medoid (if present and valid)
	if (NewsBase->ArticleStore->IsRecId(MedoidId))
		NewsBase->PutMedoid(MedoidId, StoryName);

	// after we assing all articles to the story we want to update the story dates for it
	EventRegBase->ConsiderUpdatingStoryInfo(StoryId, true);

	// report info about changed stories
	if (NewsBase->ActivityLogFile->GetVerbosityLevel() >= 5) {
		if (OldStoryIdH.Len() > 0) {
			NewsBase->ActivityLogFile->NotifyInfo("Update of this story also changed the following stories:");
			for (int KeyId = OldStoryIdH.FFirstKeyId(); OldStoryIdH.FNextKeyId(KeyId);) {
				const uint64 ChangedStoryId = OldStoryIdH.GetKey(KeyId);
				PRecSet ArtSet = NewsBase->StoryStore->GetRec(ChangedStoryId).DoJoin(NewsBase->Base, "hasArticle");
				TUInt64V ArtIdV; ArtSet->GetRecIdV(ArtIdV);
				NewsBase->ActivityLogFile->NotifyInfo("\tStory %I64u ('%s') now contains %d articles: %s", ChangedStoryId, NewsBase->StoryStore->GetRecNm(ChangedStoryId).CStr(), ArtIdV.Len(), TStr::GetStr(ArtIdV, ", ").CStr());
			}
		}
		NewsBase->ActivityLogFile->NotifyInfo("====== Finished changing the articles for story %I64u ('%s') ======", StoryId, StoryName.CStr());
	}

	Sw.Stop();
	NewsBase->InfoLogFile->DecreaseSpaces();
	NewsBase->InfoLogFile->NotifyVerboseFmt(3, "TClustUtil::SetClusterInfo took %.0f ms", Sw.GetMSec());
}

void TClustUtil::MarkArticlesStoryless(const TStr& StoryName, const TStrV& ArticleNameV)
{
	if (! NewsBase->StoryStore->IsRecNm(StoryName))
		return;
	const uint64 StoryId = NewsBase->StoryStore->GetRecId(StoryName);
	for (int N = 0; N < ArticleNameV.Len(); N++) {
		const uint64 ArticleId = NewsBase->GetArticleIdForNewsfeedId(ArticleNameV[N]);
		if (!NewsBase->ArticleStore->IsRecId(ArticleId)) {
			NewsBase->ErrorLogFile->NotifyErr("TClustUtil::MarkArticlesStoryless. Article %s is not a known article.", ArticleNameV[N].CStr());
			continue;
		}
		NewsBase->ActivityLogFile->NotifyInfo("Article %s ('%s') was marked as storyless", ArticleNameV[N].CStr(), NewsBase->ArticleStore->GetRecNm(ArticleId).CStr());
		// here we just forget the story -> article link. the article -> story will remain so that we will know that which is the closest story for any article
		NewsBase->Index->DeleteJoin(NewsBase->StoryStore, "hasArticle", StoryId, ArticleId);
	}

	// NOTE: we expect that this call is always made before the cluster is deleted
	// if this is true, then we don't need to call EventRegBase->ConsiderUpdatingStoryInfo(StoryId, true);
	// since we always receive the SetClusterInfo() call immediately after this one
	// if this is not true, then call ConsiderUpdatingStoryInfo(); here
}

// a cluster was deleted from the clustering service (because the articles in it were too old and represented a past event)
void TClustUtil::SetClusterDeleted(const TStr& StoryName)
{
	// we don't need to do anything here since this call always comes after the SetClusterInfo() which updates all event properties
	// if some custom cleanup has to be done, then add it here

	if (!NewsBase->StoryStore->IsRecNm(StoryName))
		return;
	const uint64 StoryId = NewsBase->StoryStore->GetRecId(StoryName);
	NewsBase->ActivityLogFile->NotifyInfo("Story %I64u ('%s') is now forgotten.", StoryId, StoryName.CStr());

	// for the articles by distance from the centroid. set CanDeleteWhenOld to false only to 3 articles closest to centroid
	//PRecSet ArticleSet = NewsBase->StoryStore->GetRec(StoryName).DoJoin(NewsBase->Base, "hasArticle");
	//ArticleSet->SortByFq(false);
	//for (int N = 0; N < ArticleSet->GetRecs() && N < 3; N++) {
	//	uint64 ArticleId = ArticleSet->GetRecId(N);
	//	NewsBase->ArticleStore->SetFieldNmBool(ArticleId, "CanDeleteWhenOld", false);	// mark the article as un-deletable
	//}

	//if (! NewsBase->StoryStore->IsRecNm(StoryName)) {
	//	Logger->NotifyVerboseFmt(0, "TClustUtil:SetClusterDeleted. ERROR. Setting cluster info for unknown story %s", StoryName.CStr());
	//	TNotify::StdNotify->OnStatusFmt("TClustUtil:SetClusterDeleted. ERROR. Setting cluster info for unknown story %s", StoryName.CStr());
	//	return;
	//}
	//Logger->NotifyVerboseFmt(3, "TClustUtil:SetClusterDeleted. A cluster %s was removed from clustering service.", StoryName.CStr());
	//Logger->IncreaseSpaces();
	//uint64 StoryId = NewsBase->StoryStore->GetRecId(StoryName);
	//// force the last update of the story
	//EventRegBase->ConsiderUpdatingStoryInfo(StoryId, true);
	//Logger->DecreaseSpaces();
}

///////////////////////////////
// News-Stories-Function
TStr TAddStorySrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	try {
		if (! IsFldNm(FldNmValPrV, "action")) {
			NewsBase->ErrorLogFile->NotifyErr("TAddStorySrv::ExecJSon. Received a clustering request without a specified action. Maybe you are using an obsolete version of clustering service.");
			return "{}";
		}
		const TStr Action = GetFldVal(FldNmValPrV, "action", "");
		const TStr StoryId = GetFldVal(FldNmValPrV, "cid");
		const TStr ArticleId = GetFldVal(FldNmValPrV, "aid");

		// article is assigned to a story
		if (Action == "assign") {
			const double CosToCentroid = GetFldVal(FldNmValPrV, "cosToCentroid").GetFlt(0);
			ClustUtil->SetArticle(ArticleId, StoryId, CosToCentroid);
		} 
		// the article is a duplicate of another article
		else if (Action == "duplicate") {
			const TStr OriginalArticleId = GetFldVal(FldNmValPrV, "origAid");
			ClustUtil->SetDuplicate(ArticleId, OriginalArticleId);
		}
		// set medoid for the story
		else if (Action == "clusterInfo") {
			const TStr MedoidId = GetFldVal(FldNmValPrV, "medoidAid");
			const TStr ArtIdStr = GetFldVal(FldNmValPrV, "aidV", "");
			TStrV ArtIdStrV; ArtIdStr.SplitOnAllCh(',', ArtIdStrV);
		
			const TStr CosToCentroidStr = GetFldVal(FldNmValPrV, "cosToCentroidV", "");
			TStrV CosToCentroidStrV; CosToCentroidStr.SplitOnAllCh(',', CosToCentroidStrV);
			TFltV CosToCentroidV;
			for (int N=0; N < CosToCentroidStrV.Len(); N++)
				CosToCentroidV.Add(CosToCentroidStrV[N].GetFlt(0));

			const TStr AvgCosToCentroidStr = GetFldVal(FldNmValPrV, "avgCosToCentroid", "-1");
			const TFlt AvgCosToCentroid = AvgCosToCentroidStr.GetFlt(-1);
			const TStr VarianceStr = GetFldVal(FldNmValPrV, "variance", "-1");
			const TFlt Variance = VarianceStr.GetFlt(-1);

			ClustUtil->SetClusterInfo(MedoidId, StoryId, ArtIdStrV, CosToCentroidV, Variance, AvgCosToCentroid);
		}
		else if (Action == "clusterDeleted") {
			ClustUtil->SetClusterDeleted(StoryId);
		}
		else if (Action == "markArticlesStoryless") {
			// for now we don't yet get data for this action, but it's a valid action
		}
		else {
			TNotify::StdNotify->OnStatus("TAddStorySrv::ExecJSon. Received a call with unrecognized or incomplete arguments.");
			NewsBase->ErrorLogFile->NotifyErr("TAddStorySrv::ExecJSon. Received a call with unrecognized or incomplete arguments: %s", RqEnv->GetHttpRq()->GetUrl()->GetUrlStr().CStr());
		}
	}
	catch (PExcept E) {
		TNotify::StdNotify->OnNotifyFmt(ntErr, "TAddStorySrv::ExecJSon. %s", E->GetMsgStr().CStr());
		TNotify::StdNotify->OnNotifyFmt(ntErr, E->GetLocStr().CStr());
		NewsBase->ErrorLogFile->NotifyErr("TAddStorySrv::ExecJSon.", E);
	}
	catch (...) {
		TNotify::StdNotify->OnNotify(ntErr, "TAddStorySrv::ExecJSon. Unrecognized exception in TAddStorySrv");
		NewsBase->ErrorLogFile->NotifyErr("TAddStorySrv::ExecJSon. Unrecognized exception in TAddStorySrv");
	}
	// return something
	return "{}";
}
