/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ArticleSrv.h"

#include "../Utils/FileUtils.h"
#include "../NewsMiner/NewsConst.h"


///////////////////////////////////////////////
// "article" requests
TArticleSrv::TArticleSrv(const PNewsBase& _NewsBase, const PNewsSearch& _NewsSearch, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("article", saotJSon), NewsBase(_NewsBase), NewsSearch(_NewsSearch), NewsSrvUtil(_NewsSrvUtil) 
{
	ReportResponseSize = true;
	ArticleStore = NewsBase->ArticleStore;
}

TStr TArticleSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv) 
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "getArticle")
			ResJson = GetArticle(FldNmValPrV);
		else if (Action == "getArticles")
			ResJson = GetArticles(FldNmValPrV);
		else if (Action == "getLatestArticles")
			ResJson = GetLatestArticles(FldNmValPrV);
		else if (Action == "getOriginalArticle")
			ResJson = GetOriginalArticle(FldNmValPrV);
		else {
			ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("No valid actions were provided. Valid values are: getArticle, getArticles"));
			TNotify::StdNotify->OnStatus("invalid action name for concept: " + Action);
		}

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		NewsBase->ErrorLogFile->NotifyErr("TArticleSrv::ExecJSon.", E);
		TNotify::StdNotify->OnStatus("TArticleSrv::ExecJSon. Error: " + E->GetMsgStr());
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception."));
		NewsBase->ErrorLogFile->NotifyErr("TArticleSrv::ExecJSon. General exception.");
	}
	
	// done
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

PJsonVal TArticleSrv::GetSingleArticle(const TStr& Uri, const TStrKdV& FldNmValPrV)
{
	if (!ArticleStore->IsRecNm(Uri))
		return TJsonVal::NewObj("error", TStr("bad article uri"));
	
	const uint64 ArticleId = ArticleStore->GetRecId(Uri);
	
	// get one or more result types that you would like
	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("info");
		
	// store article
	PJsonVal ArticleJson = TJsonVal::NewObj();
	if (ResultTypeV.IsIn("info")) {
		const int BodyLen = GetFldInt(FldNmValPrV, "infoBodyLen", -1);
		TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "infoConceptType", ConceptTypeV);
		EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV, ctAll);
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "infoConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		EnumArticleInfoFlags ArticleFlags = NewsSrvUtil->GetArticleFlags(FldNmValPrV, "info");
		ArticleJson->AddToObj("info", NewsSrvUtil->Article(ArticleId, BodyLen, ConceptTypes, ConceptLangV, ArticleFlags));
	}
	// the list of similar articles according to CCA
	if (ResultTypeV.IsIn("similarArticles")) {
		TStrV SimilarLangV; GetFldValV(FldNmValPrV, "similarArticlesLang", SimilarLangV);

		if (!SimilarLangV.Empty()) {
			PJsonVal SimilarDict = TJsonVal::NewObj();
			for (int N = 0; N < SimilarLangV.Len(); N++) {
				PRecSet ArticleSet = NewsBase->GetSimilarArticles(ArticleId, SimilarLangV[N]);
				SimilarDict->AddToObj(SimilarLangV[N], NewsSrvUtil->Articles(FldNmValPrV, ArticleSet, "similarArticles"));
			}
			ArticleJson->AddToObj("similarArticles", SimilarDict);
		}
	}
	// the list of articles that are duplicates of this article
	if (ResultTypeV.IsIn("duplicatedArticles")) {
		PRecSet DuplArticleSet = ArticleStore->GetRec(ArticleId).DoJoin(NewsBase->Base, "hasDuplicate");
		ArticleJson->AddToObj("duplicatedArticles", NewsSrvUtil->Articles(FldNmValPrV, DuplArticleSet, "duplicatedArticles"));
	}
	// if this article is a duplicate, then this returns the info about the original article
	if (ResultTypeV.IsIn("originalArticle")) {
		const int BodyLen = GetFldInt(FldNmValPrV, "originalArticleBodyLen", -1);
		TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "originalArticleConceptType", ConceptTypeV);
		EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV, ctAll);
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "originalArticleConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		EnumArticleInfoFlags ArticleFlags = NewsSrvUtil->GetArticleFlags(FldNmValPrV, "originalArticle");

		const uint64 OrigArtId = ArticleStore->GetFieldNmUInt64(ArticleId, "isDuplicateOfId");
		if (OrigArtId != TUInt64::Mx)
			ArticleJson->AddToObj("originalArticle", NewsSrvUtil->Article(OrigArtId, BodyLen));
	}

	return ArticleJson;
}

// get information about a specific article
PJsonVal TArticleSrv::GetArticle(const TStrKdV& FldNmValPrV) 
{
	if (!IsFldNm(FldNmValPrV, "articleUri") && !IsFldNm(FldNmValPrV, "articleId"))
		return TJsonVal::NewObj("error", TStr("no article uri/id provided. Set articleUri or articleId parameter."));
	TStrV UriV; GetFldValV(FldNmValPrV, "articleUri", UriV);
	TStrV IdV; GetFldValV(FldNmValPrV, "articleId", IdV);
	for (int N = 0; N < IdV.Len(); N++) {
		const uint64 Id = IdV[N].GetUInt64(TUInt64::Mx);
		if (ArticleStore->IsRecId(Id))
			UriV.Add(ArticleStore->GetRecNm(Id));
	}
	
	if (UriV.Len() == 1) {
		return GetSingleArticle(UriV[0], FldNmValPrV);
	}
	else {
		PJsonVal RetObj = TJsonVal::NewObj();
		for (int N = 0; N < UriV.Len(); N++)
			RetObj->AddToObj(UriV[N], GetSingleArticle(UriV[N], FldNmValPrV));
		return RetObj;
	}
}

// return latest count of articles
PJsonVal TArticleSrv::GetLatestArticles(const TStrKdV& FldNmValPrV)
{
	int Count = GetFldInt(FldNmValPrV, "count", 200);
	Count = min(Count, 200);
	int RelatedConceptCount = GetFldInt(FldNmValPrV, "relatedConceptCount", 20);
	bool IncludeSourceLocation = GetFldBool(FldNmValPrV, "includeSourceLocation", false);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");

	const PStore ArtStore = NewsBase->Base->GetStoreByStoreNm("Article");
	const uint64 ArticleId = ArtStore->GetLastRecId();
	const uint64 StartRecs = ArticleId > Count ? ArticleId - Count : 0;
	TUInt64V ArticleV;
	for (uint64 N = StartRecs; N <= ArticleId; N++)
		ArticleV.Add(N);
	PRecSet ArticleSet = TRecSet::New(ArticleStore, ArticleV);

	PJsonVal ResJson = TJsonVal::NewObj();

	// send articles
	const PJsonVal Articles = NewsSrvUtil->Articles(ArticleSet, -1);
	ResJson->AddToObj("results", Articles);
	if (IncludeSourceLocation) {
		int Vals = Articles->GetArrVals();
		for (int N = 0; N < Vals; N++) {
			PJsonVal Article = Articles->GetArrVal(N);
			TStr Uri = Article->GetObjStr("source_uri", "");
			uint64 SourceRecId = NewsBase->SourceStore->GetRecId(Uri);
			TFltPr Location = NewsBase->SourceStore->GetFieldNmFltPr(SourceRecId, "Location");
			Article->AddToObj("source_lat", TJsonVal::NewStr(Location.Val1.GetStr()));
			Article->AddToObj("source_long", TJsonVal::NewStr(Location.Val2.GetStr()));
		}
	}
	if (RelatedConceptCount > 0) {
		PRecSet RelatedConceptSet = ArticleSet->DoJoin(NewsBase->Base, "hasConcept", 1000, true);
		RelatedConceptSet->SortByFq(false);
		ResJson->AddToObj("relatedConcepts", NewsSrvUtil->Concepts(RelatedConceptSet, TUInt64Set(), Count, ctEntities, ConceptLangV));
	}

	return ResJson;
}

PJsonVal TArticleSrv::GetOriginalArticle(const TStrKdV& FldNmValPrV)
{
	if (!IsFldNm(FldNmValPrV, "uri"))
		return TJsonVal::NewObj("error", TStr("no article uri provided"));
	TStr Uri = GetFldVal(FldNmValPrV, "uri");
	if (!ArticleStore->IsRecNm(Uri))
		return TJsonVal::NewObj("error", TStr("bad article uri"));
	
	const uint64 ArticleId = ArticleStore->GetRecId(Uri);
	const uint64 OrigArticleId = ArticleStore->GetFieldNmUInt64(ArticleId, "isDuplicateOfId");
	if (OrigArticleId == TUInt64::Mx)
		return TJsonVal::NewObj("error", TStr("Provided article is not a duplicate"));

	if (!ArticleStore->IsRecId(OrigArticleId))
		return TJsonVal::NewObj("error", TStr("Original article id is not in the database anymore"));

	const int BodyLen = GetFldInt(FldNmValPrV, "originalArticleBodyLen", -1);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "originalArticleConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	EnumArticleInfoFlags ArticleFlags = NewsSrvUtil->GetArticleFlags(FldNmValPrV, "originalArticle");
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "originalArticleConceptType", ConceptTypeV);
	const EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);

	// store article
	PJsonVal ArticleJson = NewsSrvUtil->Article(OrigArticleId, BodyLen, ValidTypes, ConceptLangV, ArticleFlags);
	return ArticleJson;
}

PJsonVal TArticleSrv::GetArticlesForUrls(const TStrKdV& FldNmValPrV)
{
	TStrV ArticleUriV; GetFldValV(FldNmValPrV, "articleUri", ArticleUriV);

	TJsonValV ArticleV;
	for (int N = 0; N < ArticleUriV.Len(); N++) {
		if (!ArticleStore->IsRecNm(ArticleUriV[N]))
			continue;
		uint64 ArticleId = ArticleStore->GetRecId(ArticleUriV[N]);
		PJsonVal ArticleJson = NewsSrvUtil->Article(ArticleId, 500);
		ArticleV.Add(ArticleJson);
	}

	return TJsonVal::NewObj("articles", TJsonVal::NewArr(ArticleV));
}

// search for articles that match different criteria
PJsonVal TArticleSrv::GetArticles(const TStrKdV& FldNmValPrV)
{
	TArticleQuery Query;
	Query.Keywords = GetFldVal(FldNmValPrV, "keywords", "");
	GetFldValV(FldNmValPrV, "conceptUri", Query.ConceptUriV);
	GetFldValV(FldNmValPrV, "lang", Query.LangV);
	GetFldValV(FldNmValPrV, "locationUri", Query.LocationUriV);
	GetFldValV(FldNmValPrV, "publisherUri", Query.PublisherUriV);
	GetFldValV(FldNmValPrV, "categoryUri", Query.CategoryV);
	Query.CategoryIncludeSub = GetFldBool(FldNmValPrV, "categoryIncludeSub", true);

	Query.DateStart = GetFldVal(FldNmValPrV, "dateStart", "");
	Query.DateEnd = GetFldVal(FldNmValPrV, "dateEnd", "");

	TStrV DateMentionV; GetFldValV(FldNmValPrV, "dateMention", DateMentionV);
	TStr DateMentionStart = GetFldVal(FldNmValPrV, "dateMentionStart", "");
	TStr DateMentionEnd = GetFldVal(FldNmValPrV, "dateMentionEnd", "");
	Query.DateMentionV = NewsSearch->ParseDateMentions(DateMentionV, DateMentionStart, DateMentionEnd);

	Query.IgnoreKeywords = GetFldVal(FldNmValPrV, "ignoreKeywords", "");
	GetFldValV(FldNmValPrV, "ignoreConceptUri", Query.IgnoreConceptUriV);
	GetFldValV(FldNmValPrV, "ignoreLang", Query.IgnoreLangV);
	GetFldValV(FldNmValPrV, "ignoreConceptUri", Query.IgnoreConceptUriV);
	GetFldValV(FldNmValPrV, "ignoreLocationUri", Query.IgnoreLocationUriV);
	GetFldValV(FldNmValPrV, "ignorePublisherUri", Query.IgnorePublisherUriV);
	GetFldValV(FldNmValPrV, "ignoreCategoryUri", Query.IgnoreCategoryV);
	Query.IgnoreCategoryIncludeSub = GetFldBool(FldNmValPrV, "ignoreCategoryIncludeSub", true);

	// user can also provide a comma separated list of uris if he wishes to compute som aggegrates on them
	// if eventUriList is provided then other query arguments (if existing) will be ignored
	TStr ArticleIdList = GetFldVal(FldNmValPrV, "articleIdList", "");
	TStrV ArticleUriV; GetFldValV(FldNmValPrV, "articleUri", ArticleUriV);

	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("articles");
	
	PRecSet ArticleRecSet;
	PJsonVal ErrorJson;
	if (ArticleIdList.Len() > 0) {
		TStrV ArticleStrIdV;
		ArticleIdList.SplitOnAllCh(',', ArticleStrIdV);
		TUInt64V ArticleIdV;
		for (int N = 0; N < ArticleStrIdV.Len(); N++) {
			const uint64 ArticleId = ArticleStrIdV[N].GetUInt64(TUInt64::Mx);
			if (ArticleStore->IsRecId(ArticleId))
				ArticleIdV.Add(ArticleId);
		}
		ArticleRecSet = TRecSet::New(ArticleStore, ArticleIdV);
	}
	else if (ArticleUriV.Len() > 0) {
		TUInt64V ArticleIdV;
		for (int N = ArticleUriV.Len() - 1; N >= 0; N--) {
			const TStr ArticleUri = ArticleUriV[N];	
			if (ArticleStore->IsRecNm(ArticleUri))
				ArticleIdV.Add(ArticleStore->GetRecId(ArticleUri));
		}
		ArticleRecSet = TRecSet::New(ArticleStore, ArticleIdV);
	}
	else {
		const bool Ok = NewsSearch->SearchArticles(Query, ArticleRecSet, ErrorJson);
		if (!Ok && !ErrorJson.Empty())
			return ErrorJson;
	}
	if (ArticleRecSet.Empty() || ArticleRecSet->GetRecs() == 0)
		return TJsonVal::NewObj("error", TStr("No results match the search conditions."));

	PJsonVal ResJson = TJsonVal::NewObj();
	// list of events
	if (ResultTypeV.IsIn("articles")) {
		const int  Count = TSAppSrvFun::GetFldInt(FldNmValPrV, "articlesCount", 25);
		if (Count > 200) {
			PJsonVal ErrJson = TJsonVal::NewObj("error", TStr("The API does not allow downloading more than 200 articles per call."));
			ResJson->AddToObj("articles", ErrJson);
		}
		else
			ResJson->AddToObj("articles", NewsSrvUtil->Articles(FldNmValPrV, ArticleRecSet));
	}
	// list of uris
	if (ResultTypeV.IsIn("uriList"))
		ResJson->AddToObj("uriList", NewsSrvUtil->ArticleUris(FldNmValPrV, ArticleRecSet));
	
	// list of article ids
	if (ResultTypeV.IsIn("articleIds"))
		ResJson->AddToObj("articleIds", NewsSrvUtil->ArticleIds(FldNmValPrV, ArticleRecSet));

	// time aggregates
	if (ResultTypeV.IsIn("timeAggr"))
		ResJson->AddToObj("timeAggr", ArticleTimeAggr(FldNmValPrV, ArticleRecSet));
	
	// category aggregates
	if (ResultTypeV.IsIn("categoryAggr"))
		ResJson->AddToObj("categoryAggr", ArticleCategoryAggr(FldNmValPrV, ArticleRecSet));
	
	// concept aggregates
	if (ResultTypeV.IsIn("conceptAggr"))
		ResJson->AddToObj("conceptAggr", ArticleConceptAggr(FldNmValPrV, ArticleRecSet));
	
	// source location aggregates
	if (ResultTypeV.IsIn("sourceAggr")) {
		PJsonVal SourceAggrArr = NewsSrvUtil->ArticleSourceAggr(FldNmValPrV, ArticleRecSet);
		ResJson->AddToObj("sourceAggr", SourceAggrArr);
	}
	// keyword aggregates
	if (ResultTypeV.IsIn("keywordAggr")) {
		const TStr KeywordLang = GetFldVal(FldNmValPrV, "keywordAggrLang", "eng");
		const int KeywordCount = GetFldInt(FldNmValPrV, "keywordAggrCount", 50);
		ResJson->AddToObj("keywordAggr", ArticleKeywordAggr(FldNmValPrV, ArticleRecSet, KeywordLang, KeywordCount));
	}
	// concept matrix
	if (ResultTypeV.IsIn("conceptMatrix"))
		ResJson->AddToObj("conceptMatrix", ArticleConceptMatrix(FldNmValPrV, ArticleRecSet));
	
	// concept graph
	if (ResultTypeV.IsIn("conceptGraph"))
		ResJson->AddToObj("conceptGraph", ArticleConceptGraph(FldNmValPrV, ArticleRecSet));
	
	// trending concepts
	if (ResultTypeV.IsIn("trendingConcepts"))
		ResJson->AddToObj("trendingConcepts", ArticleTrendingConcepts(FldNmValPrV, ArticleRecSet));
	
	// date mention aggregates
	if (ResultTypeV.IsIn("dateMentionAggr"))
		ResJson->AddToObj("dateMentionAggr", NewsSrvUtil->ArticleDateMentionAggr(ArticleRecSet));
	
	// return the recently added articles and their timing information
	if (ResultTypeV.IsIn("recentActivity")) {
		PJsonVal ArticlesJson = TJsonVal::NewObj("articles", NewsSrvUtil->ArticleRecentActivity(FldNmValPrV, ArticleRecSet));
		ResJson->AddToObj("recentActivity", ArticlesJson);
	}
	if (ResJson->GetObjKeys() == 0)
		ResJson->AddToObj("error", TStr("unrecognized resultType was specified. Recognized values are list, timeAggr, categoryAggr, conceptAggr, conceptGraph, trendingConcepts, dateMentionAggr, recentActivity."));

	return ResJson;

	// if user specified some text he wishes to ignore we have to go through the articles in the rec set and
	// cut out articles that contain the text
	/*if (ParamsJson->IsObjKey("ignoreText")) {
		TStr IgnoreText = ParamsJson->GetObjStr("ignoreText", "");
		int Count = ArticleRecSet->GetRecs();

		RE2::Options ReOptions;
		ReOptions.set_case_sensitive(false);
		ReOptions.set_utf8(false);
		RE2 Re(IgnoreText.CStr(), ReOptions);
		StringPiece Matches[10];

		TUInt64IntKdV ValidArtV;
		for (int N = 0; N < Count; N++) {
			uint64 RecId = ArticleRecSet->GetRecId(N);
			TStr Content = ArticleStore->GetFieldNmStr(RecId, "Title") + " " + ArticleStore->GetFieldNmStr(RecId, "Body");
			const StringPiece SPContent(Content.CStr());
			if (Re.Match(SPContent, 0, SPContent.size(), RE2::UNANCHORED, Matches, 10)) {
				TNotify::StdNotify->OnStatusFmt("Skipping article %I64u since it contains '%s'", RecId, Matches[0].ToString().c_str());
			}
			else
				ValidArtV.Add(TUInt64IntKd(RecId, ArticleRecSet->GetRecFq(N)));
		}
		ArticleRecSet = TRecSet::New(NewsBase->ArticleStoreId, ValidArtV, true);
	}
*/
}

// return the number of articles per day
PJsonVal TArticleSrv::ArticleTimeAggr(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet, const TStr& FldNmPrefix)
{
	const TStr TimeWindowStr = TSAppSrvFun::GetFldVal(FldNmValPrV, FldNmPrefix + "TimeWindow", "day");		// total, month, day, time
	const EnumTimeWindow TimeWindow = TNewsSrvUtil::GetTimeWindow(TimeWindowStr);

	THash<TStr, TInt> DateToCountH;
	int Articles = ArticleRecSet->GetRecs();
	for (int N = 0; N < Articles; N++) {
		// find the appropriate key for the article's date and update the article count
		TStr Date = NewsSrvUtil->GetKeyForTimeWindow(ArticleRecSet->GetRecId(N), TimeWindow);
		DateToCountH.AddDat(Date) += 1;
	}
	// order by date
	DateToCountH.SortByKey();

	PJsonVal JsonArr = TJsonVal::NewArr();
	for (int KeyId = DateToCountH.FFirstKeyId(); DateToCountH.FNextKeyId(KeyId);) {
		TStr Date = DateToCountH.GetKey(KeyId);
		int Count = DateToCountH[KeyId];
		PJsonVal DateCount = TJsonVal::NewObj("date", TJsonVal::NewStr(Date));
		DateCount->AddToObj("count", Count);
		JsonArr->AddToArr(DateCount);
	}
	return JsonArr;
}

// category aggregate
PJsonVal TArticleSrv::ArticleCategoryAggr(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet)
{
	const int SampleSize = GetFldInt(FldNmValPrV, "categoryAggrSampleSize", 20000);
	PRecSet SampleArticleSet = ArticleRecSet->GetSampleRecSet(SampleSize, false);
	THash<TUInt64, TInt> CatToCountH;
	int Articles = SampleArticleSet->GetRecs();
	for (int N = 0; N < Articles; N++) {
		TRec ArtRec = SampleArticleSet->GetRec(N);
		TRec CatRec = ArtRec.DoSingleJoin(NewsBase->Base, "hasCategory");
		if (CatRec.IsDef())
			CatToCountH.AddDat(CatRec.GetRecId()) += 1;
	}

	CatToCountH.SortByDat(false);

	PJsonVal JsonArr = TJsonVal::NewArr();
	for (int KeyId = CatToCountH.FFirstKeyId(); CatToCountH.FNextKeyId(KeyId);) {
		TUInt64 LocRecId = CatToCountH.GetKey(KeyId);
		TStr CatNm = NewsBase->CategoryStore->GetRecNm(LocRecId);
		int Count = CatToCountH[KeyId];

		PJsonVal CatJson = TJsonVal::NewObj("count", TJsonVal::NewNum(Count));
		CatJson->AddToObj("title", CatNm);
		JsonArr->AddToArr(CatJson);
	}

	return JsonArr;
}

PJsonVal TArticleSrv::ArticleConceptAggr(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet)
{
	// take a subset of articles to compute the concept aggregates
	const int SampleSize = GetFldInt(FldNmValPrV, "conceptAggrSampleSize", 1000);
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptAggrConceptType", ConceptTypeV);
	const int ConceptCount = GetFldInt(FldNmValPrV, "conceptAggrConceptCount", 25);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptAggrConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	const EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);

	PRecSet ConceptSet = ArticleRecSet->DoJoin(NewsBase->Base, "hasConcept", SampleSize);
	THash<TUInt64, TInt> Concepts; ConceptSet->GetRecIdFqH(Concepts);
	Concepts.SortByDat(false);
	int MaxCount = Concepts.Len() > 0 ? (int) Concepts[0] : 1;

	TJsonValV ArrJson;
	
	TStrH TypeToCountH;
	for (int N = 0; N < NewsConst.ConceptTypeV.Len(); N++) {
		if (NewsConst.IsValidType(NewsConst.ConceptTypeV[N], ValidTypes))
			TypeToCountH.AddDat(NewsConst.ConceptTypeV[N]);
	}
	
	for (int N = 0, KeyId = Concepts.FFirstKeyId(); Concepts.FNextKeyId(KeyId); ) {
		const TStr ConceptType = NewsBase->ConceptStore->GetFieldNmStr(Concepts.GetKey(KeyId), "Type");
		if (!NewsConst.IsValidType(ConceptType, ValidTypes))
			continue;
		
		if (TypeToCountH.GetDat(ConceptType) < ConceptCount) {
			PJsonVal ConceptJson = NewsSrvUtil->Concept(Concepts.GetKey(KeyId), int (100 * Concepts[KeyId] / (double) MaxCount), ConceptLangV);
			ArrJson.Add(ConceptJson);
			TypeToCountH.AddDat(ConceptType) += 1;
		}
		// did we already get ConceptCount concepts for each desired type?
		int Sum = 0;
		for (int KeyId = TypeToCountH.FFirstKeyId(); TypeToCountH.FNextKeyId(KeyId);)
			Sum += TypeToCountH[KeyId];
		if (Sum == TypeToCountH.Len() * ConceptCount)
			break;
	}
	return TJsonVal::NewArr(ArrJson);
}

PJsonVal TArticleSrv::ArticleKeywordAggr(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet, const TStr& Lang, const int& KeywordCount)
{
	const int SampleSize = GetFldInt(FldNmValPrV, "keywordAggrSampleSize", 500);

	// select articles in appropriate language
	PRecSet ArticleLangSet;
	ArticleLangSet = ArticleRecSet->Clone();
	ArticleLangSet->FilterByFieldStr(ArticleStore->GetFieldId("Language"), Lang);
	if (ArticleLangSet->GetRecs() == 0)		// if no articles for given language then use all articles
		ArticleLangSet = ArticleRecSet;
	
	PSwSet SwSet;
	if (Lang == "eng")
		SwSet = TSwSet::New(swstEn523);
	else if (Lang == "spa")
		SwSet = TSwSet::New(swstEs);
	else if (Lang == "deu")
		SwSet = TSwSet::New(swstGe);
	else if (Lang == "slv")
		SwSet = TSwSet::New(swstSi);
	else
		SwSet = TSwSet::New();
	int Articles = min(ArticleLangSet->GetRecs(), SampleSize);
	PAggr KwAggr = TAggrs::TKeywords::New(NewsBase->Base, "keywords", ArticleLangSet, TFtrExts::TBagOfWords::New(NewsBase->Base, ArticleStore, ArticleStore->GetFieldId("Body"), TFtrExts::bowmConcat, TTokenizers::THtml::New(SwSet, TStemmer::New(stmtPorter, true))), Articles);
	PJsonVal KwData = KwAggr->SaveJson();
	return KwData->GetObjKey("keywords");
}

PJsonVal TArticleSrv::ArticleConceptMatrix(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet)
{
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptMatrixConceptType", ConceptTypeV);
	EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
	const int ConceptCount = GetFldInt(FldNmValPrV, "conceptMatrixConceptCount", 25);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptMatrixConceptLang", ConceptLangV);
	const int SampleSize = GetFldInt(FldNmValPrV, "conceptMatrixSampleSize", 500);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	const TStr MeasureStr = GetFldVal(FldNmValPrV, "conceptMatrixMeasure", "pmi");		// pmi (pointwise mutual information), pairTfIdf (pair frequence * IDF of individual concepts), chiSquare
	MatrixScoringMethod Measure = pmi;
	if (MeasureStr == "pmi") Measure = pmi;
	else if (MeasureStr == "pairTfIdf") Measure = pairTfIdf;
	else if (MeasureStr == "chiSquare") Measure = chiSquare;
	
	// go through all the concepts and count how frequently each pair occurs together in the events
	THash<TPair<TUInt64, TUInt64>, TInt> ConceptPairToCntH;
	THash<TUInt64, TInt> ConceptToCntH;
	const int Articles = min(ArticleRecSet->GetRecs(), SampleSize);
	GetConceptPairs(ArticleRecSet, ValidTypes, ConceptPairToCntH, ConceptToCntH, SampleSize);

	// find "interestingness" of pairs of concepts (ci,cj)
	THash<TPair<TUInt64, TUInt64>, TFlt> ConceptPairToScoreH;
	for (int KeyId = ConceptPairToCntH.FFirstKeyId(); ConceptPairToCntH.FNextKeyId(KeyId);) {
		TPair<TUInt64, TUInt64> Pair = ConceptPairToCntH.GetKey(KeyId);
		double Score;
		const int PairCount = ConceptPairToCntH[KeyId];
		const double C1 = (double) ConceptToCntH.GetDat(Pair.Val1);
		const double C2 = (double) ConceptToCntH.GetDat(Pair.Val2);
		// find pairs of concepts (ci,cj) that have high COF*IDF = conceptPairToCntH(ci,cj) * log ( |EventsRecSet| / (ConceptToCntH(ci)) * log( |EventRecSet| / ConceptToCntH(cj) ) )
		if (Measure == pairTfIdf)
			Score = PairCount * log(Articles / (double) ConceptToCntH.GetDat(Pair.Val1)) * log(Articles / (double) ConceptToCntH.GetDat(Pair.Val2));
		else if (Measure == chiSquare) {
			const double E12 = C1 * C2 / (double) Articles;
			const double O12 = PairCount;
			if (O12 > E12)
				Score = Sqr(O12 - E12) / E12 * log(PairCount);	// take chi square but also prefer pairs with high coocurrence score
		}
		else if (Measure == pmi) {
			const double P12 = PairCount / (double) Articles;
			const double P1 = C1 / (double) Articles;
			const double P2 = C2 / (double) Articles;
			Score = log(P12 / (P1*P2));
		}
		ConceptPairToScoreH.AddDat(Pair, Score);
	}

	// for each concept compute its score as score_ci = sum(ConceptPairToScoreH( (ci, X) ) ) + sum(ConceptPairToScoreH( (X, ci) ) )
	THash<TUInt64, TFlt> ConceptToScoreH;
	for (int KeyId = ConceptPairToScoreH.FFirstKeyId(); ConceptPairToScoreH.FNextKeyId(KeyId);) {
		TPair<TUInt64, TUInt64> Pair = ConceptPairToScoreH.GetKey(KeyId);
		int PairScore = (int) ConceptPairToScoreH[KeyId];
		ConceptToScoreH.AddDat(Pair.Val1) += abs(PairScore);
		ConceptToScoreH.AddDat(Pair.Val2) += abs(PairScore);
	}

	// sort concepts by score and produce the subset of concepts that we will send
	ConceptToScoreH.SortByDat(false);
	TUInt64Set ChosenConceptsH;
	for (int Count = 0, KeyId = ConceptToScoreH.FFirstKeyId(); Count < ConceptCount && ConceptToScoreH.FNextKeyId(KeyId); Count++)
		ChosenConceptsH.AddKey(ConceptToScoreH.GetKey(KeyId));

	PJsonVal FreqMatrixJson = TJsonVal::NewArr();
	PJsonVal ScoreMatrixJson = TJsonVal::NewArr();
	PJsonVal ConceptListJson = TJsonVal::NewArr();
	// return info about co-occurence of best ranked concepts
	for (int KeyId = ChosenConceptsH.FFirstKeyId(); ChosenConceptsH.FNextKeyId(KeyId);) {
		uint64 ConceptId = ChosenConceptsH.GetKey(KeyId);
		ConceptListJson->AddToArr(NewsSrvUtil->Concept(ConceptId, -1, ConceptLangV));
		
		TFltV FreqArrV;
		TFltV ScoreArrV;
		for (int KeyId2 = ChosenConceptsH.FFirstKeyId(); ChosenConceptsH.FNextKeyId(KeyId2);) {
			uint64 ConceptId2 = ChosenConceptsH.GetKey(KeyId2);
			TPair<TUInt64, TUInt64> Key(min(ConceptId, ConceptId2), max(ConceptId, ConceptId2));
			double Count;
			if (ConceptId == ConceptId2)
				Count = ConceptToCntH.GetDat(ConceptId);
			else
				Count = ConceptPairToCntH.IsKey(Key) ? (int) ConceptPairToCntH.GetDat(Key) : 0;
			FreqArrV.Add(Count);
			double Score = 0;
			if (ConceptPairToScoreH.IsKey(TPair<TUInt64, TUInt64>(ConceptId, ConceptId2))) Score = ConceptPairToScoreH.GetDat(TPair<TUInt64, TUInt64>(ConceptId, ConceptId2));
			else if (ConceptPairToScoreH.IsKey(TPair<TUInt64, TUInt64>(ConceptId2, ConceptId))) Score = ConceptPairToScoreH.GetDat(TPair<TUInt64, TUInt64>(ConceptId2, ConceptId));
			ScoreArrV.Add(Score);
		}
		FreqMatrixJson->AddToArr(TJsonVal::NewArr(FreqArrV));
		ScoreMatrixJson->AddToArr(TJsonVal::NewArr(ScoreArrV));
	}
	PJsonVal ResJson = TJsonVal::NewObj("concepts", ConceptListJson);
	ResJson->AddToObj("sampleSize", Articles);
	ResJson->AddToObj("freqMatrix", FreqMatrixJson);
	ResJson->AddToObj("scoreMatrix", ScoreMatrixJson);
	return ResJson;
}

PJsonVal TArticleSrv::ArticleConceptGraph(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet)
{
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptGraphConceptType", ConceptTypeV);
	EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
	const int ConceptCount = GetFldInt(FldNmValPrV, "conceptGraphConceptCount", 25);
	const int LinkCount = GetFldInt(FldNmValPrV, "conceptGraphLinkCount", 50);
	const int SampleSize = GetFldInt(FldNmValPrV, "conceptGraphSampleSize", 500);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptGraphConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");

	// if not types specified take all entities
	TStrSet ConceptTypeH(ConceptTypeV);
	if (ConceptTypeH.Len() == 0)
		ConceptTypeH = NewsConst.EntitySet;

	// go through all the concepts and count how frequently each pair occurs together in the events
	THash<TPair<TUInt64, TUInt64>, TInt> ConceptPairToCntH;
	THash<TUInt64, TInt> ConceptToCntH;
	GetConceptPairs(ArticleRecSet, ValidTypes, ConceptPairToCntH, ConceptToCntH, SampleSize);

	PJsonVal RetJson = TJsonVal::NewObj();
	// remember what would be the max possible score for a concept pair
	RetJson->AddToObj("maxPossibleScore", MIN(ArticleRecSet->GetRecs(), SampleSize));
	
	//
	// add top concepts
	ConceptToCntH.SortByDat(false);
	THash<TUInt64, TInt> TopConcepts;
	int Added = 0, KeyId;
	for (KeyId = ConceptToCntH.FFirstKeyId(); Added < ConceptCount && ConceptToCntH.FNextKeyId(KeyId); Added++)
		TopConcepts.AddDat(ConceptToCntH.GetKey(KeyId), ConceptToCntH[KeyId]);

	TUInt64IntKdV ConceptRecKdV; TopConcepts.GetKeyDatKdV(ConceptRecKdV);
	PRecSet ConceptRecSet = TRecSet::New(NewsBase->ConceptStore, ConceptRecKdV, true);
	RetJson->AddToObj("concepts", NewsSrvUtil->Concepts(ConceptRecSet, TUInt64Set(), TInt::Mx, ctAll, ConceptLangV));

	//
	// add frequencies between top concepts
	ConceptPairToCntH.SortByDat(false);
	TJsonValV PairsJsonV;
	int AddedLinks = 0;
	for (int KeyId = ConceptPairToCntH.FFirstKeyId(); AddedLinks < LinkCount && ConceptPairToCntH.FNextKeyId(KeyId); AddedLinks++) {
		TPair<TUInt64, TUInt64> Pair = ConceptPairToCntH.GetKey(KeyId);
		if (!TopConcepts.IsKey(Pair.Val1) || !TopConcepts.IsKey(Pair.Val2))
			continue;
		PJsonVal PairJson = TJsonVal::NewObj("id1", TJsonVal::NewNum(Pair.Val1));
		PairJson->AddToObj("id2", TJsonVal::NewNum(Pair.Val2));
		PairJson->AddToObj("score", TJsonVal::NewNum(ConceptPairToCntH[KeyId]));
		PairsJsonV.Add(PairJson);
	}
	RetJson->AddToObj("links", PairsJsonV);

	return RetJson;
}

PJsonVal TArticleSrv::ArticleTrendingConcepts(const TStrKdV& FldNmValPrV, const PRecSet& ArticleRecSet)
{
	const int ConceptCount = GetFldInt(FldNmValPrV, "trendingConceptsConceptCount", 10);
	// do we have custom concept urls provided?
	TStrV TrendingConceptV; GetFldValV(FldNmValPrV, "trendingConceptsConceptUri", TrendingConceptV);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "trendingConceptsConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");

	// find which concepts do we want to show trends for
	if (TrendingConceptV.Len() == 0) {
		PJsonVal TopConceptsJson = ArticleConceptAggr(FldNmValPrV, ArticleRecSet);
		for (int N = 0; TopConceptsJson->IsArr() && N < ConceptCount && N < TopConceptsJson->GetArrVals(); N++)
			TrendingConceptV.Add(TopConceptsJson->GetArrVal(N)->GetObjStr("uri"));
	}

	TUInt64V ConceptIdV;
	// go over each concept. find articles for the concept and make intersection with ArticleRecSet
	// save for each of the concepts the number of times the concept occurs in events distributed over time
	THash<TStr, THash<TUInt64, TInt> > DateToConceptIdCountH;
	for (int N = 0; N < TrendingConceptV.Len(); N++) {
		// ignore invalid concept uris
		if (!NewsBase->ConceptStore->IsRecNm(TrendingConceptV[N]))
			continue;
		const uint64 ConceptId = NewsBase->ConceptStore->GetRecId(TrendingConceptV[N]);
		ConceptIdV.Add(ConceptId);
		PRecSet ArticlesForConcept = NewsBase->ConceptStore->GetRec(ConceptId).DoJoin(NewsBase->Base, "hasArticle");
		NewsBase->IntersectRecSets(ArticlesForConcept, ArticleRecSet);
		for (int ArtN = 0; ArtN < ArticlesForConcept->GetRecs(); ArtN++) {
			const TStr Date = ArticleStore->GetFieldNmStr(ArticlesForConcept->GetRecId(ArtN), "Date");
			DateToConceptIdCountH.AddDat(Date).AddDat(ConceptId) += 1;
		}
	}

	// create a trends array where each element is an object having "date" and an array of (concept, freq) for that date
	TJsonValV DateArrJsonV;
	DateToConceptIdCountH.SortByKey(true);
	for (int KeyId = DateToConceptIdCountH.FFirstKeyId(); DateToConceptIdCountH.FNextKeyId(KeyId);) {
		TStr Date = DateToConceptIdCountH.GetKey(KeyId);
		PJsonVal DateJson = TJsonVal::NewObj("date", Date);
		PJsonVal ConceptFqs = TJsonVal::NewObj();
		THash<TUInt64, TInt>& ConceptIdCountH = DateToConceptIdCountH[KeyId];
		for (int CKeyId = ConceptIdCountH.FFirstKeyId(); ConceptIdCountH.FNextKeyId(CKeyId);) {
			ConceptFqs->AddToObj(ConceptIdCountH.GetKey(CKeyId).GetStr(), TJsonVal::NewNum(ConceptIdCountH[CKeyId]));
		}
		DateJson->AddToObj("conceptFreq", ConceptFqs);
		const uint64 ArtsForDate = NewsBase->NewsCounts->GetAnnArticleCount(Date);
		DateJson->AddToObj("totArts", TJsonVal::NewNum(ArtsForDate));
		DateArrJsonV.Add(DateJson);
	}
	PJsonVal RetJson = TJsonVal::NewObj("trends", TJsonVal::NewArr(DateArrJsonV));

	PRecSet ConceptRecSet = TRecSet::New(NewsBase->ConceptStore, ConceptIdV);
	RetJson->AddToObj("conceptInfo", NewsSrvUtil->Concepts(ConceptRecSet, TUInt64Set(), TInt::Mx, ctAll, ConceptLangV));
	return RetJson;
}

void TArticleSrv::GetConceptPairs(const PRecSet& ArticleRecSet, const EnumConceptType& ValidTypes, THash<TPair<TUInt64, TUInt64>, TInt>& ConceptPairToCntH, THash<TUInt64, TInt>& ConceptToCntH, const int SampleSize)
{
	ConceptPairToCntH.Clr();
	ConceptToCntH.Clr();
	PRecSet SampleSet = ArticleRecSet->GetRecs() > SampleSize ? ArticleRecSet->GetSampleRecSet(SampleSize, false) : ArticleRecSet;
	for (int ArtN = 0; ArtN < SampleSet->GetRecs(); ArtN++) {
		PRecSet ConceptRecSet = SampleSet->GetRec(ArtN).DoJoin(NewsBase->Base, "hasConcept");
		int Concepts = ConceptRecSet->GetRecs();
		for (int ConceptN = 0; ConceptN < Concepts; ConceptN++) {
			uint64 ConceptId = ConceptRecSet->GetRecId(ConceptN);
			TStr ConceptType = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
			if (!NewsConst.IsValidType(ConceptType, ValidTypes))
				continue;
			ConceptToCntH.AddDat(ConceptId) += 1;
			for (int ConceptN2 = ConceptN + 1; ConceptN2 < Concepts; ConceptN2++) {
				uint64 ConceptId2 = ConceptRecSet->GetRecId(ConceptN2);
				TStr ConceptType2 = NewsBase->ConceptStore->GetFieldNmStr(ConceptId2, "Type");
				if (!NewsConst.IsValidType(ConceptType, ValidTypes))
					continue;
				TPair<TUInt64, TUInt64> Key = TPair<TUInt64, TUInt64>(min(ConceptId, ConceptId2), max(ConceptId, ConceptId2));
				ConceptPairToCntH.AddDat(Key) += 1;
			}
		}
	}
}