/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include <base.h>
#include "CategorySrv.h"
#include "../NewsMiner/NewsSrvUtil.h"

///////////////////////////////////////////////
// Category server - providing names of categories, subcategories, ...
TCategorySrv::TCategorySrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil):
	  TSAppSrvFun("category", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) { 

	CategoryBase = NewsBase->CategoryBase;
}

void TCategorySrv::AddSubChildren(PJsonVal& CatArray)
{
	AssertR(CatArray->IsArr(), "Subchildren can only be added to an array of categories");
	int CatLen = CatArray->GetArrVals();
	for (int N=0; N < CatLen; N++) {
		PJsonVal Cat = CatArray->GetArrVal(N);
		const TStr CatUri = Cat->GetObjStr("uri", "");
		PRecSet ChildCatRecSet = CategoryBase->GetChildCategories(CatUri);
		if (ChildCatRecSet->GetRecs() > 0) {
			// build a json for each category in ChildCatRecSet
			PJsonVal ChildCatArrJson = CategoryBase->Categories(ChildCatRecSet);
			// for each of the children recursively call AddSubChildren to add subchildren
			AddSubChildren(ChildCatArrJson);
			// add subchildren as "children" to json
			Cat->AddToObj("children", ChildCatArrJson);
		}
	}
}

TStr TCategorySrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv) {
	PJsonVal ResJson;
	try {
		ResJson = PerformAction(FldNmValPrV, RqEnv);
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		NewsBase->ErrorLogFile->NotifyErr("TCategorySrv::ExecJSon.", E);
		TNotify::StdNotify->OnStatus("Error: " + E->GetMsgStr());
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception."));
	}
	
	TStr RetVal = TJsonVal::GetStrFromVal(ResJson);
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		const TStr& Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + RetVal + ")";
	}
	else
		return RetVal;
}

PJsonVal TCategorySrv::PerformAction(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	const TStr& Action = GetFldVal(FldNmValPrV, "action", "");
	if (Action == "") {
		return TJsonVal::NewObj("error", TStr("You need to specify action parameter. Possible values for it are:"));
	}
	else if (Action == "getCategories") {
		const TStr ParentCategory = GetFldVal(FldNmValPrV, "parentUri", CategoryBase->GetDMozUriPrefix());
		const bool RecursivelyP = GetFldInt(FldNmValPrV, "recursively", 1) == 1;
		PRecSet Children = CategoryBase->GetChildCategories(ParentCategory);
		if (Children->GetRecs() == 0)
			return TJsonVal::NewObj("results", TJsonVal::NewArr());
		PJsonVal ChildrenJson = CategoryBase->Categories(Children);
		if (RecursivelyP)
			AddSubChildren(ChildrenJson);
		return TJsonVal::NewObj("results", ChildrenJson);
	}
	else
		return TJsonVal::NewObj("error", TStr("Unknown action '" + Action + "'.'}"));
}

TCategoryAutoSuggestSrv::TCategoryAutoSuggestSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) : 
	TSAppSrvFun("suggestCategories", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil)
{
	CategoryBase = NewsBase->CategoryBase;
}

TCategoryAutoSuggestSrv::~TCategoryAutoSuggestSrv()
{
	TFOut FOut(NewsBase->Base->GetFPath() + "CategoryAutoSuggestCache.dat");
	CachedCategoryCount.Save(FOut);
	CategoryPrefixToIdWgtH.Save(FOut);
}

TStr TCategoryAutoSuggestSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	TStr Action = GetFldVal(FldNmValPrV, "action", "");
	if (Action == "updatePrefixes") {
		UpdatePrefixes();
		return "{\"info\": \"Categories updated.\"}";
	}

	PJsonVal ResJson = NULL;
	try {
		ResJson = FindCategories(FldNmValPrV);

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("Error: " + E->GetMsgStr());
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TStr("General exception"));
		TNotify::StdNotify->OnStatus("Error: General exception");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

void TCategoryAutoSuggestSrv::LoadOrUpdatePrefixes()
{
	try {
		if (TFile::Exists(NewsBase->Base->GetFPath() + "CategoryAutoSuggestCache.dat")) {
			TFIn FIn(NewsBase->Base->GetFPath() + "CategoryAutoSuggestCache.dat");
			TInt CategoryCount;
			CategoryCount.Load(FIn);
			if (CategoryCount + 10 > NewsBase->CategoryStore->GetRecs()) {
				TNotify::StdNotify->OnStatus("Category suggestions are being loaded from the cache...");
				CachedCategoryCount = CategoryCount;
				CategoryPrefixToIdWgtH.Load(FIn);
				return;
			}
			else
				TNotify::StdNotify->OnStatus("File with category suggestions is out of date.");
		}
		else
			TNotify::StdNotify->OnStatus("File with category suggestions does not exist.");
		UpdatePrefixes();
	}
	catch (PExcept E) {
		TNotify::StdNotify->OnStatusFmt("TCategoryAutoSuggestSrv::LoadOrUpdatePrefixes: %s", E->GetMsgStr().CStr());
		NewsBase->ErrorLogFile->NotifyErr("TCategoryAutoSuggestSrv::LoadOrUpdatePrefixes.", E);
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("TCategoryAutoSuggestSrv::LoadOrUpdatePrefixes: General exception");
		NewsBase->ErrorLogFile->NotifyErr("TCategoryAutoSuggestSrv::LoadOrUpdatePrefixes: General exception");
	}
}

void TCategoryAutoSuggestSrv::UpdatePrefixes()
{
	TTmStopWatch StopWatch(true);
	THash<TUInt64, TUInt64> CategoryIdToFqH;

	PNewsCounts NewsCounts = NewsBase->NewsCounts;
	CachedCategoryCount = NewsBase->CategoryStore->GetRecs();
	PStoreIter Iter = NewsBase->CategoryStore->GetIter();
	while (Iter->Next()) {
		uint64 CategoryId = Iter->GetRecId();
		const TStr& CategoryUri = NewsBase->CategoryStore->GetRecNm(CategoryId);
		const TStr& CategoryNm = NewsBase->CategoryStore->GetFieldNmStr(CategoryId, "Label");
		int CategoryScore = CategoryNm.CountCh('/');
		CategoryIdToFqH.AddDat(CategoryId, CategoryScore);
	}

	// sort from most to least frequently used concepts
	CategoryIdToFqH.SortByDat(true);

	CategoryPrefixToIdWgtH.Clr();
	for (int KeyId = CategoryIdToFqH.FFirstKeyId(); CategoryIdToFqH.FNextKeyId(KeyId);) {
		TUInt64 CategoryRecId = CategoryIdToFqH.GetKey(KeyId);
		int Fq = (int) CategoryIdToFqH[KeyId];
		
		TStr NameLc = NewsBase->CategoryStore->GetFieldNmStr(CategoryRecId, "Label").GetLc();
		NameLc.ChangeChAll('/', ' ');
		TStrV WordParts; NameLc.SplitOnAllCh(' ', WordParts);
		for (int Parts = 0; Parts < WordParts.Len(); Parts++) {
			TStr Sub1 = WordParts[Parts].GetSubStr(0, 0);
			TStr Sub2 = WordParts[Parts].GetSubStr(0, 1);
			CategoryPrefixToIdWgtH.AddDat(Sub1).Add(TUInt64IntKd(CategoryRecId, Fq));
			if (Sub2.Len() == 2)
				CategoryPrefixToIdWgtH.AddDat(Sub2, TVec<TUInt64IntKd>()).Add(TUInt64IntKd(CategoryRecId, Fq));
		}
	}
	
	double BuildingTm = StopWatch.GetSec();
	StopWatch.Reset(true);

	TNotify::StdNotify->OnStatusFmt("Update category prefixes took: %.2f sec", BuildingTm);
}

PJsonVal TCategoryAutoSuggestSrv::FindCategories(const TStrKdV& FldNmValPrV)
{
	TTmStopWatch StopWatch(true);

	// construct query
	const TStr PrefixStr = GetFldVal(FldNmValPrV, "prefix");
	const TStr PrefixStrLc = PrefixStr.GetLc();

	const TInt Page = GetFldInt(FldNmValPrV, "page", 0);
	const TInt Count = GetFldInt(FldNmValPrV, "count", 20);

	PJsonVal SuggArr = TJsonVal::NewArr();

	TUInt64Set AddedIdH;	// see which concepts we already added so that we don't add them twice
	TUInt64IntKdV SuggestionsV;

	int AddedCount = 0;
	const int ToSkip = Page * Count;
	int Skipped = 0;
	
	GetCategorySuggestions(PrefixStr, SuggestionsV);

	int SuggestionsCount = SuggestionsV.Len();
	for (int N = 0; N < SuggestionsCount; N++) {
		TUInt64 CategoryRecId = SuggestionsV[N].Key;
		// for the same prefix, the same concept can match multiple times so we need to check to add it only once
		if (AddedIdH.IsKey(CategoryRecId))
			continue;
		AddedIdH.AddKey(CategoryRecId);
		if (!NewsBase->CategoryStore->IsRecId(CategoryRecId))
			continue;
		// does the label match the given prefix
		TStr Label = NewsBase->CategoryStore->GetFieldNmStr(CategoryRecId, "Label");
		TStr LabelLc = Label.GetLc();
		LabelLc.ChangeChAll('/', ' ');
		if (!LabelLc.IsPrefix(PrefixStrLc) && LabelLc.SearchStr(" " + PrefixStrLc, 0) == -1)
			continue;

		// if we are here we have a valid concept that should be added to the suggestions list
		// now only check if we want some Page > 0. if yes, then skip the necessary number of suggestions
		if (Skipped < ToSkip) { Skipped++; continue; }

		PJsonVal Sugg = CategoryBase->Category(CategoryRecId);
		SuggArr->AddToArr(Sugg);

		AddedCount++;
		if (AddedCount >= Count)
			break;
	}
	TNotify::StdNotify->OnStatusFmt("Finding suggestions took: %.2fms", StopWatch.GetMSec());
	return SuggArr;
}

void TCategoryAutoSuggestSrv::GetCategorySuggestions(const TStr& Prefix, TVec<TUInt64IntKd>& SuggV) const
{
	TStr PrefixLc = Prefix.GetLc();
	TStr PrefixKey = PrefixLc.GetSubStr(0, 1);
	SuggV.Clr();
	if (!CategoryPrefixToIdWgtH.IsKey(PrefixKey))
		return;
	SuggV = CategoryPrefixToIdWgtH.GetDat(PrefixKey);
}