/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ConceptClassSrv.h"
#include "../NewsMiner/NewsSrvUtil.h"
#include <base.h>

///////////////////////////////////////////////
// base class for managing the category store - creating missing categories, finding subcategories, ...
TConceptClassBase::TConceptClassBase(const PNewsBase& _NewsBase) : NewsBase(_NewsBase)
{
	Base = NewsBase->Base;
	Index = NewsBase->Index;
	ConceptStore = NewsBase->ConceptStore;
	ConceptClassStore = NewsBase->ConceptClassStore;
	ConceptStoreId = NewsBase->ConceptStoreId;
	ConceptClassStoreId = NewsBase->ConceptClassStoreId;

	CustomClassPrefix = "concept-class";

	Load();
}

void TConceptClassBase::Load()
{
	try {
		const TStr FName = NewsBase->Base->GetFPath() + "CustomData/ConceptClassSrv.bin";
		if (TFile::Exists(FName)) {
			TFIn FIn(FName);
			NextUriId.Load(FIn);
			AvailableClassUriV.Load(FIn);
		}
		else
			NextUriId = (int) ConceptClassStore->GetRecs();
	}
	catch (PExcept E) {
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassBase::Load.", E);
	}
	catch (...) {
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassBase::Load. Error: General exception");
	}
}

void TConceptClassBase::Save()
{
	try {
		TStr FName = NewsBase->Base->GetFPath() + "CustomData/ConceptClassSrv.bin";
		TFOut FOut(FName);
		NextUriId.Save(FOut);
		AvailableClassUriV.Save(FOut);
	}
	catch (PExcept E) {
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassBase::Save.", E);
	}
	catch (...) {
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassBase::Save. Error: General exception");
	}
}

TStr TConceptClassBase::GetNextUri()
{
	TStr Uri;
	if (GetAvailableClassUri(Uri))
		return Uri;
	Uri = TStr::Fmt("%s-%d", CustomClassPrefix.CStr(), NextUriId);
	NextUriId++;
	return Uri;
}

// import classes from the specified file. the file should contain lines of json objects, where each json
// object represents one class (or a modification of a class, such as join to another class)
bool TConceptClassBase::ImportClasses(const TStr& FName, TStr& RetMsg)
{
	RetMsg = "";
	if (!TFile::Exists(FName)) {
		RetMsg = TStr::Fmt("File %s not found. Failed to import classes.", FName.CStr());
		return false;
	}
	TFIn FIn(FName);
	TStr Ln; int OkCount = 0, FailedCount = 0;
	while (FIn.GetNextLn(Ln)) {
		try {
			PJsonVal ClassJson = TJsonVal::GetValFromStr(Ln);
			if (!ClassJson->IsObj())
				continue;
			// joins between the classes - add to the index
			if (ClassJson->IsObjKey("hasSuperClass")) {
				const TStr ClassUri = ClassJson->GetObjStr("URI");
				const TStr SuperClassUri = ClassJson->GetObjKey("hasSuperClass")->GetObjStr("URI");
				if (!ConceptClassStore->IsRecNm(ClassUri) || !ConceptClassStore->IsRecNm(SuperClassUri))
					continue;
				const uint64 ClassId = ConceptClassStore->GetRecId(ClassUri);
				const uint64 SuperClassId = ConceptClassStore->GetRecId(SuperClassUri);
				ConceptClassStore->AddJoin("hasSubClass", SuperClassId, ClassId);
			}
			// individual classes - add them to the store
			else
				ConceptClassStore->AddRec(ClassJson);
			OkCount++;
		}
		catch (PExcept Ex) {
			FailedCount++;
		}
	}

	RetMsg = TStr::Fmt("Successfully imported %d classes, failed at %d classes", OkCount, FailedCount);
	return true;
}

bool TConceptClassBase::ImportInstances(const TStr& FName, TStr& RetMsg)
{
	RetMsg = "";
	if (!TFile::Exists(FName)) {
		RetMsg = TStr::Fmt("File %s not found. Failed to import instances.", FName.CStr());
		return false;
	}
	TFIn FIn(FName);
	TStr Ln; int OkCount = 0, FailedCount = 0, Total = 0;
	while (FIn.GetNextLn(Ln)) {
		try {
			Total++;
			PJsonVal Json = TJsonVal::GetValFromStr(Ln);
			if (!Json->IsObj() || !Json->IsObjKey("ConceptUri") || !Json->IsObjKey("ClassUri"))
				continue;
			const TStr ConceptUri = Json->GetObjStr("ConceptUri");
			const TStr ClassUri = Json->GetObjStr("ClassUri");
			if (!NewsBase->ConceptStore->IsRecNm(ConceptUri)) {
				TNotify::StdNotify->OnStatus("Unknown concept uri: " + ConceptUri);
				continue;
			}
			if (!NewsBase->ConceptClassStore->IsRecNm(ClassUri))
				continue;
			const uint64 ConceptId = ConceptStore->GetRecId(ConceptUri);
			const uint64 ClassId = ConceptClassStore->GetRecId(ClassUri);

			ConceptStore->AddJoin("hasConceptClass", ConceptId, ClassId);

			OkCount++;
		}
		catch (PExcept Ex) {
			FailedCount++;
		}
	}

	RetMsg = TStr::Fmt("Successfully imported %d instances, %d ignored, %d failed", OkCount, Total - OkCount - FailedCount, FailedCount);
	return true;
}

uint64 TConceptClassBase::AddClass(const TStr& Label, const TStr& Description, const TStr& Type, const TStrV& ConceptUriV, const TStr& SuperClassUri)
{
	TStr Uri = GetNextUri();
	PJsonVal Json = TJsonVal::NewObj("URI", Uri);
	Json->AddToObj("Description", Description);
	Json->AddToObj("Type", Type);
	uint64 Id = ConceptClassStore->AddRec(Json);

	SetLabel(Uri, Label);

	if (ConceptClassStore->IsRecNm(SuperClassUri)) {
		const uint64 SuperId = ConceptClassStore->GetRecId(SuperClassUri);
		ConceptClassStore->AddJoin("hasSuperClass", Id, SuperId);
	}

	for (int N = 0; N < ConceptUriV.Len(); N++) {
		if (!ConceptStore->IsRecNm(ConceptUriV[N])) {
			TNotify::StdNotify->OnStatus("Unknown concept uri: " + ConceptUriV[N]);
			continue;
		}
		const uint64 ConceptId = ConceptStore->GetRecId(ConceptUriV[N]);
		ConceptClassStore->AddJoin("hasConcept", Id, ConceptId);
	}
	return Id;
}

void TConceptClassBase::SetLabel(const TStr& Uri, const TStr& Label, const TStr& Lang)
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return;
	const TStr Labels = ConceptClassStore->GetFieldNmStr(ConceptClassStore->GetRecId(Uri), "LabelsJson");
	bool Ok; TStr Msg;
	PJsonVal LabelsJson = TJsonVal::GetValFromStr(Labels, Ok, Msg);
	if (!Ok)
		LabelsJson = TJsonVal::NewObj();
	LabelsJson->AddToObj(Lang, Label);
	const TStr UpdatedLabels = TJsonVal::GetStrFromVal(LabelsJson);
	ConceptClassStore->SetFieldNmStr(ConceptClassStore->GetRecId(Uri), "LabelsJson", UpdatedLabels);
}

TStr TConceptClassBase::GetLabel(const TStr& Uri, const TStr& Lang) const
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return "";
	const TStr Labels = ConceptClassStore->GetFieldNmStr(ConceptClassStore->GetRecId(Uri), "LabelsJson");
	bool Ok; TStr Msg;
	PJsonVal LabelsJson = TJsonVal::GetValFromStr(Labels, Ok, Msg);
	if (!Ok)
		return "";
	if (LabelsJson->IsObjKey(Lang))
		return LabelsJson->GetObjStr(Lang);
	if (LabelsJson->IsObjKey(DefLang))
		return LabelsJson->GetObjStr(DefLang);
	return "";
}

bool TConceptClassBase::HasLabel(const TStr& Uri, const TStr& Lang) const
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return false;
	const TStr Labels = ConceptClassStore->GetFieldNmStr(ConceptClassStore->GetRecId(Uri), "LabelsJson");
	bool Ok; TStr Msg;
	PJsonVal LabelsJson = TJsonVal::GetValFromStr(Labels, Ok, Msg);
	if (!Ok)
		return false;
	return LabelsJson->IsObjKey(Lang);
}

PJsonVal TConceptClassBase::GetLabels(const TStr& Uri) const
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return TJsonVal::New();
	const TStr Labels = ConceptClassStore->GetFieldNmStr(ConceptClassStore->GetRecId(Uri), "LabelsJson");
	bool Ok; TStr Msg;
	PJsonVal LabelsJson = TJsonVal::GetValFromStr(Labels, Ok, Msg);
	if (!Ok)
		return TJsonVal::New();
	return LabelsJson;
}

// set the Uri to have ParentUri as the parent
// if ParentUri == "" (or any invalid uri) then we make Uri as the root concept
void TConceptClassBase::SetSuperClass(const TStr& Uri, const TStr& SuperUri)
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return;
	uint64 ClassId = ConceptClassStore->GetRecId(Uri);
	TRec CurrentParentRec = ConceptClassStore->GetRec(ClassId).DoSingleJoin(Base, "hasSuperClass");
	if (CurrentParentRec.IsDef()) {
		uint64 CurrentParentId = CurrentParentRec.GetRecId();
		ConceptClassStore->DelJoin("hasSuperClass", ClassId, CurrentParentId);
	}
	if (ConceptClassStore->IsRecNm(SuperUri)) {
		uint64 SuperClassId = ConceptClassStore->GetRecId(SuperUri);
		ConceptClassStore->AddJoin("hasSuperClass", ClassId, SuperClassId);
	}
}

int TConceptClassBase::AddInstancesToClass(const TStr& Uri, const TUInt64V& ConceptIdV)
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return 0;
	const uint64 ClassId = ConceptClassStore->GetRecId(Uri);
	int AddedCount = 0;
	for (int N = 0; N < ConceptIdV.Len(); N++) {
		if (!NewsBase->ConceptStore->IsRecId(ConceptIdV[N]))
			continue;
		const uint64 ConceptId = ConceptIdV[N];
		ConceptClassStore->AddJoin("hasConcept", ClassId, ConceptId);
		AddedCount++;
	}
	return AddedCount;
}

void TConceptClassBase::RemoveInstancesFromClass(const TStr& Uri, const TUInt64V& ConceptIdV)
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return;
	const uint64 ClassId = ConceptClassStore->GetRecId(Uri);
	for (int N = 0; N < ConceptIdV.Len(); N++) {
		if (!NewsBase->ConceptStore->IsRecId(ConceptIdV[N]))
			continue;
		const uint64 ConceptId = ConceptIdV[N];
		ConceptClassStore->DelJoin("hasConcept", ClassId, ConceptId);
	}
}

void TConceptClassBase::DeleteClass(const TStr& Uri)
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return;
	const uint64 ClassId = ConceptClassStore->GetRecId(Uri);
	TRec SuperClassRec = ConceptClassStore->GetRec(Uri).DoSingleJoin(Base, "hasSuperClass");
	const uint64 SuperClassId = SuperClassRec.GetRecId();

	// delete join from Uri to parent
	if (SuperClassId != TUInt64::Mx)
		ConceptClassStore->DelJoin("hasSuperClass", ClassId, SuperClassId);
	// delete joins from children to Uri
	TUInt64Set SubClassRecs = GetSubClasses(Uri);
	for (int KeyId = SubClassRecs.FFirstKeyId(); SubClassRecs.FNextKeyId(KeyId);) {
		const uint64 SubClassId = SubClassRecs.GetKey(KeyId);
		ConceptClassStore->DelJoin("hasSuperClass", SubClassId, ClassId);
		// redirect children to the parent of the deleted class
		if (SuperClassId != TUInt64::Mx)
			ConceptClassStore->AddJoin("hasSuperClass", SubClassId, SuperClassId);
	}
	PRecSet ConceptSet = ConceptClassStore->GetRec(ClassId).DoJoin(Base, "hasConcept");
	for (int N = 0; N < ConceptSet->GetRecs(); N++) {
		const uint64 ConceptId = ConceptSet->GetRecId(N);
		ConceptClassStore->DelJoin("hasConcept", ClassId, ConceptId);
	}
	PJsonVal Json = TJsonVal::NewObj("URI", Uri);
	Json->AddToObj("Label", "");
	Json->AddToObj("Description", "");
	Json->AddToObj("Type", "");
	ConceptClassStore->AddRec(Json);
	NewsBase->ConceptLabel->RemoveConcept(Uri);

	// reuse the uri (only of the custom classes, not dbpedia ones) for the future classes
	if (Uri.SearchStr(CustomClassPrefix) == 0)
		AvailableClassUriV.AddUnique(Uri);
}

PRecSet TConceptClassBase::GetInstances(const TStr& Uri) const
{
	if (!ConceptClassStore->IsRecNm(Uri))
		return TRecSet::New(ConceptClassStore);
	return ConceptClassStore->GetRec(Uri).DoJoin(Base, "hasConcept");
}

void TConceptClassBase::ConceptUriToV(const TStrV& ConceptUriV, TUInt64V& ConceptIdV) const
{
	int Len = ConceptUriV.Len();
	ConceptIdV.Clr();
	for (int N = 0; N < Len; N++) {
		if (ConceptStore->IsRecNm(ConceptUriV[N]))
			ConceptIdV.Add(ConceptStore->GetRecId(ConceptUriV[N]));
	}
}

// if there an available uri, get it, otherwise return empty
bool TConceptClassBase::GetAvailableClassUri(TStr& Uri)
{
	if (AvailableClassUriV.Len() == 0)
		return false;
	Uri = AvailableClassUriV[AvailableClassUriV.Len() - 1];
	AvailableClassUriV.DelLast();
	return true;
}

TUInt64Set TConceptClassBase::GetAllSubClasses(const TStr& ClassUri) const
{
	return GetAllSubClasses(ConceptClassStore->GetRecId(ClassUri));
}

TUInt64Set TConceptClassBase::GetAllSubClasses(const uint64& ClassId) const
{
	TUInt64Set FullRetSet; FullRetSet.AddKey(ClassId);
	PRecSet ClassSet = TRecSet::New(ConceptClassStore, ClassId);
	while (ClassSet->GetRecs() > 0) {
		ClassSet = ClassSet->DoJoin(NewsBase->Base, "hasSubClass");
		TUInt64V ClassV; ClassSet->GetRecIdV(ClassV);
		FullRetSet.AddKeyV(ClassV);
	}
	return FullRetSet;
}

TUInt64Set TConceptClassBase::GetSubClasses(const TStr& ClassUri) const
{
	return GetSubClasses(ConceptClassStore->GetRecId(ClassUri));
}

TUInt64Set TConceptClassBase::GetSubClasses(const uint64& ClassId) const
{
	TUInt64Set RecIdSet;
	if (!ConceptClassStore->IsRecId(ClassId))
		return RecIdSet;
	PRecSet SubClasses = ConceptClassStore->GetRec(ClassId).DoJoin(NewsBase->Base, "hasSubClass");
	SubClasses->GetRecIdSet(RecIdSet);
	return RecIdSet;
}

// get the uris of all the custom classes that
void TConceptClassBase::GetValidConceptClasses(const TStrSet& TypeH, TStrV& ConceptClassUriV)
{
	PStoreIter Iter = NewsBase->ConceptClassStore->GetIter();
	TStrSet ClassUriH;
	while (Iter->Next()) {
		const uint64 ClassId = Iter->GetRecId();
		TStr Type = NewsBase->ConceptClassStore->GetFieldNmStr(ClassId, "Type");
		if (! TypeH.IsKey(Type))
			continue;
		const TStr Uri = NewsBase->ConceptClassStore->GetRecNm(ClassId);
		ClassUriH.AddKey(Uri);
	}
	for (int N = 0; N < AvailableClassUriV.Len(); N++) {
		ClassUriH.DelIfKey(AvailableClassUriV[N]);
	}
	ConceptClassUriV.Clr();
	ClassUriH.GetKeyV(ConceptClassUriV);
}

///////////////////////////////////////////////
// Concept Class server - providing administration for concept classes
TConceptClassSrvBase::TConceptClassSrvBase(const TStr& FunNm, const PNewsBase& _NewsBase, const PConceptClassBase& _ConceptClassBase, const PNewsSrvUtil& _NewsSrvUtil, const bool& _LogRqToFile) :
TSAppSrvFun(FunNm, saotJSon), NewsBase(_NewsBase), ConceptClassBase(_ConceptClassBase), NewsSrvUtil(_NewsSrvUtil)
{
	LogRqToFile = _LogRqToFile;
}

PJsonVal TConceptClassSrvBase::ConceptClass(const uint64& ClassId, const bool& IncludeConceptsAndDescription, const TStrV& LabelLanguageV) const
{
	if (!NewsBase->ConceptClassStore->IsRecId(ClassId))
		return TJsonVal::NewObj();
	const TStr Uri = NewsBase->ConceptClassStore->GetRecNm(ClassId);
	PJsonVal ResJson = TJsonVal::NewObj("uri", Uri);
	ResJson->AddToObj("type", NewsBase->ConceptClassStore->GetFieldNmStr(ClassId, "Type"));
	PJsonVal Labels = ConceptClassBase->GetLabels(Uri);
	// add also labels for the concept in other languages
	for (int N = 0; N < LabelLanguageV.Len(); N++) {
		const TStr JsonNm = NewsConst.GetJsonNmForLang(LabelLanguageV[N]);
		const TStr Label = Labels->GetObjStr(LabelLanguageV[N], "");
		ResJson->AddToObj(JsonNm, Label);
	}
	if (IncludeConceptsAndDescription) {
		PRecSet ConceptSet = NewsBase->ConceptClassStore->GetRec(ClassId).DoJoin(NewsBase->Base, "hasConcept");
		ResJson->AddToObj("concepts", NewsSrvUtil->Concepts(ConceptSet));
		ResJson->AddToObj("description", NewsBase->ConceptClassStore->GetFieldNmStr(ClassId, "Description"));
	}
	return ResJson;
}

///////////////////////////////////////////////
// Concept Class server - providing administration for concept classes
TStr TConceptClassSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");

		if (Action == "getClass")
			ResJson = GetClass(FldNmValPrV);

		else {
			ResJson = TJsonVal::NewObj("error", TStr::Fmt("Unknown action '%s' was provided in the request.", Action.CStr()));
			TNotify::StdNotify->OnStatus("invalid function name: " + GetFunNm());
		}

	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TConceptClassSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TConceptClassSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

// get current class definition
PJsonVal TConceptClassSrv::GetClass(const TStrKdV& FldNmValPrV)
{
	const TStr Uri = GetFldVal(FldNmValPrV, "uri", "");
	const bool IncludeAllLabels = GetFldBool(FldNmValPrV, "includeAllLabels", false);

	if (!NewsBase->ConceptClassStore->IsRecNm(Uri))
		return TJsonVal::NewObj("error", TStr::Fmt("Unknown concept class uri '%s'", Uri.CStr()));
	uint64 ClassId = NewsBase->ConceptClassStore->GetRecId(Uri);
	return ConceptClass(ClassId, true);
}
