/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ConceptFolderSrv.h"


// /////////////////////
// "conceptFolder" server
TStr TConceptFolderSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		// articles for a given folderUri
		if (Action == "getArticles")
			ResJson = GetArticles(FldNmValPrV);
		// get events for folder
		else if (Action == "getEvents")
			ResJson = GetEvents(FldNmValPrV);
		// suggested articles for a given folder definition
		else if (Action == "getArticlesForFolder")
			ResJson = GetArticlesForFolder(FldNmValPrV);
		// suggested events for a given folder definition
		else if (Action == "getEventsForFolder")
			ResJson = GetEventsForFolder(FldNmValPrV);
		else if (Action == "getFolderJson")
			ResJson = GetFolderJson(FldNmValPrV);
		// given a folderJson from user add concept and source labels
		else if (Action == "addLabelsForFolderJson")
			ResJson = AddLabelsForFolderJson(FldNmValPrV);
		else if (Action == "addConceptDetailsFolderJson")
			ResJson = AddConceptDetailsFolderJson(FldNmValPrV);
		// return the list of concept folders
		else if (Action == "listFolders")
			ResJson = ListFolders(FldNmValPrV);
		else {
			ResJson = TJsonVal::NewObj("error", TStr::Fmt("Unknown action '%s' was provided in the request.", Action.CStr()));
			TNotify::StdNotify->OnStatus("invalid function name: " + GetFunNm());
		}

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TConceptFolderSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TConceptFolderSrv::ExecJSon. ", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TConceptFolderSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TConceptFolderSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

// get the articles for the given concept folder folderUri
// update the folder first (unless updateIfNecessary == false)
PJsonVal TConceptFolderSrv::GetArticles(const TStrKdV& FldNmValPrV)
{
	TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");
	if (!NewsBase->ConceptStore->IsRecNm(FolderUri) || !NewsBase->ConceptFolderStore->IsRecNm(FolderUri))
		return TJsonVal::NewObj("error", TStr("Invalid folderUri value"));
	bool UpdateIfNecessary = GetFldBool(FldNmValPrV, "updateIfNecessary", true);

	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("articles");

	// update the concept folder if necessary
	if (UpdateIfNecessary)
		ConceptFolderBase->UpdateConceptFolderArticles(FolderUri);

	const uint64 ConceptFolderId = NewsBase->ConceptFolderStore->GetRecId(FolderUri);
	PRecSet ArticleRecSet = NewsBase->ConceptFolderStore->GetRec(ConceptFolderId).DoJoin(NewsBase->Base, "hasArticle");

	PJsonVal ResJson = TJsonVal::NewObj();
	// list of events
	if (ResultTypeV.IsIn("articles"))
		ResJson->AddToObj("articles", NewsSrvUtil->Articles(FldNmValPrV, ArticleRecSet));
	// recent activity
	if (ResultTypeV.IsIn("recentActivity")) {
		PJsonVal ArticlesJson = TJsonVal::NewObj("articles", NewsSrvUtil->ArticleRecentActivity(FldNmValPrV, ArticleRecSet));
		ResJson->AddToObj("recentActivity", ArticlesJson);
	}
	return ResJson;
}

PJsonVal TConceptFolderSrv::GetEvents(const TStrKdV& FldNmValPrV)
{
	TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");
	if (!NewsBase->ConceptStore->IsRecNm(FolderUri) || !NewsBase->ConceptFolderStore->IsRecNm(FolderUri))
		return TJsonVal::NewObj("error", TStr("Invalid folderUri value"));
	bool UpdateIfNecessary = GetFldBool(FldNmValPrV, "updateIfNecessary", true);

	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("events");

	// update the concept folder if necessary
	if (UpdateIfNecessary)
		ConceptFolderBase->UpdateConceptFolderArticles(FolderUri);

	const uint64 ConceptFolderId = NewsBase->ConceptFolderStore->GetRecId(FolderUri);
	PRecSet EventRecSet = NewsBase->ConceptFolderStore->GetRec(ConceptFolderId).DoJoin(NewsBase->Base, "hasEvent");

	PJsonVal ResJson = TJsonVal::NewObj();
	// list of events
	if (ResultTypeV.IsIn("events"))
		ResJson->AddToObj("events", NewsSrvUtil->Events(FldNmValPrV, EventRecSet));
	// recent activity
	if (ResultTypeV.IsIn("recentActivity")) {
		PJsonVal EventsJson = TJsonVal::NewObj("events", NewsSrvUtil->EventsRecentActivity(FldNmValPrV, EventRecSet));
		ResJson->AddToObj("recentActivity", EventsJson);
	}
	return ResJson;
}

// return recent folder articles based on the interests specified in the folder json
PJsonVal TConceptFolderSrv::GetArticlesForFolder(const TStrKdV& FldNmValPrV)
{
	TTmStopWatch StopWatchX(true);
	const int MaxDaysBack = GetFldInt(FldNmValPrV, "maxDaysBack", 15);
	PJsonVal FolderJson = TJsonVal::GetValFromStr(GetFldVal(FldNmValPrV, "folder", "{}"));
	// uris -> ids
	PConceptFolder Folder = TConceptFolder::New(FolderJson, NewsBase);

	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("articles");			// list, relatedConcepts, relatedSources, dailyTopics

	const uint64 MinArticleId = (MaxDaysBack != 0) ? (uint64) ConceptFolderBase->GetArticleIdForTime(MaxDaysBack) : 0;
	// compute suggested articles for the folder
	PRecSet ArticleRecSet = ConceptFolderBase->GetSuggestedArticlesForFolder(Folder, MinArticleId, false);

	PJsonVal ResJson = TJsonVal::NewObj();
	// list of articles
	if (ResultTypeV.IsIn("articles"))
		ResJson->AddToObj("articles", NewsSrvUtil->Articles(FldNmValPrV, ArticleRecSet));
	// list of related concepts
	if (ResultTypeV.IsIn("relatedConcepts")) {
		// ignore current concepts
		TUInt64Set IgnoreConceptSet(Folder->ConceptIdV);
		ResJson->AddToObj("relatedConcepts", NewsSrvUtil->RelatedConcepts(FldNmValPrV, ArticleRecSet, IgnoreConceptSet));
	}
	// list of related sources
	if (ResultTypeV.IsIn("relatedSources")) {
		const int RelatedSourcesCount = GetFldInt(FldNmValPrV, "relatedSourcesCount", 40);
		TTmStopWatch StopWatch(true);
		PRecSet RelatedSourceSet = ArticleRecSet->DoJoin(NewsBase->Base, "hasSource", 1000, true);
		RelatedSourceSet->SortByFq(false);
		RelatedSourceSet->Trunc(RelatedSourcesCount);
		ResJson->AddToObj("relatedSources", NewsSrvUtil->Sources(RelatedSourceSet));
	}
	// daily topics for the articles inside the Page and Page + Count
	if (ResultTypeV.IsIn("dailyTopics")) {
		TUInt64Set ConceptSet(Folder->ConceptIdV);
		ResJson->AddToObj("dailyTopics", NewsSrvUtil->DailyTopics(FldNmValPrV, ArticleRecSet, ConceptSet));
	}
	if (ResultTypeV.IsIn("recentActivity")) {
		PJsonVal ArticlesJson = TJsonVal::NewObj("articles", NewsSrvUtil->ArticleRecentActivity(FldNmValPrV, ArticleRecSet));
		ResJson->AddToObj("recentActivity", ArticlesJson);
	}
	// add also the folder definition to the result
	ResJson->AddToObj("folder", FolderJson);

	ConceptFolderBase->NotifyVerboseFmt(2, "ArticlesForFolder. Total time to get final articles: %d ms", StopWatchX.GetMSecInt());
	return ResJson;
}

// return recent folder articles based on the interests specified in the folder json
PJsonVal TConceptFolderSrv::GetEventsForFolder(const TStrKdV& FldNmValPrV)
{
	TTmStopWatch StopWatchX(true);
	const int MaxDaysBack = GetFldInt(FldNmValPrV, "maxDaysBack", 15);
	PJsonVal FolderJson = TJsonVal::GetValFromStr(GetFldVal(FldNmValPrV, "folder", "{}"));
	// uris -> ids
	PConceptFolder Folder = TConceptFolder::New(FolderJson, NewsBase);

	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("events");			// list, relatedConcepts, relatedSources, dailyTopics

	const uint64 MinEventId = (MaxDaysBack != 0) ? (uint64) ConceptFolderBase->GetEventIdForTime(MaxDaysBack) : 0;
	// compute suggested articles for the folder
	PRecSet EventRecSet = ConceptFolderBase->GetSuggestedEventsForFolder(Folder, MinEventId);

	PJsonVal ResJson = TJsonVal::NewObj();
	// list of articles
	if (ResultTypeV.IsIn("events"))
		ResJson->AddToObj("events", NewsSrvUtil->Events(FldNmValPrV, EventRecSet));
	// recent activity
	if (ResultTypeV.IsIn("recentActivity")) {
		PJsonVal EventsJson = TJsonVal::NewObj("events", NewsSrvUtil->EventsRecentActivity(FldNmValPrV, EventRecSet));
		ResJson->AddToObj("recentActivity", EventsJson);
	}
	// add also the folder definition to the result
	ResJson->AddToObj("folder", FolderJson);

	ConceptFolderBase->NotifyVerboseFmt(2, "ArticlesForFolder. Total time to get final articles: %d ms", StopWatchX.GetMSecInt());
	return ResJson;
}

PJsonVal TConceptFolderSrv::GetFolderJson(const TStrKdV& FldNmValPrV)
{
	const TStr FolderUri = GetFldVal(FldNmValPrV, "folderUri", "");
	const bool IncludeLabels = GetFldBool(FldNmValPrV, "includeLabels", true);
	TStrV LangV; GetFldValV(FldNmValPrV, "lang", LangV);
	if (LangV.Empty())
		LangV.Add("eng");

	if (!NewsBase->ConceptFolderStore->IsRecNm(FolderUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Concept folder with uri '%s' does not exist", FolderUri.CStr()));

	uint64 FolderId = NewsBase->ConceptFolderStore->GetRecId(FolderUri);
	PJsonVal Folderson = NewsSrvUtil->ConceptFolder(FolderId, true, true, LangV);
	ConceptFolderBase->AddUrisForFolderJson(Folderson);
	if (IncludeLabels)
		ConceptFolderBase->AddLabelsForFolderJson(Folderson);
	return Folderson;
}

// add into the json for a folder the list of labels for used concepts and sources
PJsonVal TConceptFolderSrv::AddLabelsForFolderJson(const TStrKdV& FldNmValPrV)
{
	PJsonVal JsonData = TJsonVal::GetValFromStr(GetFldVal(FldNmValPrV, "folder", "{}"));
	ConceptFolderBase->AddLabelsForFolderJson(JsonData);
	return JsonData;
}

PJsonVal TConceptFolderSrv::AddConceptDetailsFolderJson(const TStrKdV& FldNmValPrV)
{
	PJsonVal JsonData = TJsonVal::GetValFromStr(GetFldVal(FldNmValPrV, "folder", "{}"));
	EnumConceptInfoFlags ConceptFlags = (EnumConceptInfoFlags) (picIncludePrimImage);

	if (JsonData->IsObjKey("conceptUriV")) {
		PJsonVal ConceptsJson = JsonData->GetObjKey("conceptUriV");
		TJsonValV ConceptJsonV;
		for (int N = 0; N < ConceptsJson->GetArrVals(); N++) {
			const TStr ConceptUri = ConceptsJson->GetArrVal(N)->GetStr();
			if (NewsBase->ConceptStore->IsRecNm(ConceptUri)) {
				const uint64 ConceptId = NewsBase->ConceptStore->GetRecId(ConceptUri);
				ConceptJsonV.Add(NewsSrvUtil->Concept(ConceptId, -1, NewsConst.LangCoreV, ConceptFlags));
			}
		}
		JsonData->AddToObj("conceptV", TJsonVal::NewArr(ConceptJsonV));
	}

	return JsonData;
}

// return a list of all concept folders
PJsonVal TConceptFolderSrv::ListFolders(const TStrKdV& FldNmValPrV)
{
	const int Page = GetFldInt(FldNmValPrV, "page", 0);
	const int Count = GetFldInt(FldNmValPrV, "count", 100);

	TStrV LangV; GetFldValV(FldNmValPrV, "lang", LangV);
	if (LangV.Empty())
		LangV.Add("eng");

	TUInt64V RecIdV;
	const uint64 Recs = NewsBase->ConceptFolderStore->GetRecs();
	const int Start = Page * Count;
	const int End = MIN((Page + 1) * Count, Recs);
	for (int N = Start; N < End; N++)
		RecIdV.Add(N);

	const PRecSet RecSet = TRecSet::New(NewsBase->ConceptFolderStore, RecIdV);
	// don't include sources, concepts, ... but do include the label of the concept folder etc...
	return NewsSrvUtil->ConceptFolders(RecSet, false, true, LangV);
}
