/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ConceptSrv.h"


///////////////////////////////////////////////
// "concept" requests
TConceptSrv::TConceptSrv(const PNewsBase& _NewsBase, const PNewsSearch& _NewsSearch, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("concept", saotJSon), NewsBase(_NewsBase), NewsSearch(_NewsSearch), NewsSrvUtil(_NewsSrvUtil)
{
}

TStr TConceptSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "getArticles")
			ResJson = GetArticles(FldNmValPrV);
		else if (Action == "getInfo")
			ResJson = GetInfo(FldNmValPrV);
		else if (Action == "getRelatedConcepts")
			ResJson = GetRelatedConcepts(FldNmValPrV);
		else {
			ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("No valid actions were provided. Valid values are: getArticles, getInfo"));
			TNotify::StdNotify->OnStatus("invalid action name for concept: " + Action);
		}

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TConceptSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TConceptSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TConceptSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TConceptSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

// provide information about one or more concepts
PJsonVal TConceptSrv::GetInfo(const TStrKdV& FldNmValPrV)
{
	if (!IsFldNm(FldNmValPrV, "uri"))
		return TJsonVal::NewObj("error", TJsonVal::NewStr("Concept uri was not provided"));
	TStrV UriV; GetFldValV(FldNmValPrV, "uri", UriV);
	EnumConceptInfoFlags ConceptFlags = NewsSrvUtil->GetConceptFlags(FldNmValPrV);

	PJsonVal JsonObj = TJsonVal::NewObj();
	for (int N = 0; N < UriV.Len(); N++) {
		PJsonVal ResJson;
		if (!NewsBase->ConceptStore->IsRecNm(UriV[N]))
			ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("Provided uri is not valid concept uri."));
		else {
			const uint64 ConceptId = NewsBase->ConceptStore->GetRecId(UriV[N]);

			ResJson = NewsSrvUtil->Concept(ConceptId, -1, NewsConst.LangCoreV, ConceptFlags);
			TStr ConceptType = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
			if (ConceptType == NewsConst.TypeLocation)
				ResJson->AddToObj("location", NewsSrvUtil->Location(ConceptId, NewsConst.LangCoreV, true, true, true));
		}
		if (UriV.Len() == 1)
			return ResJson;
		JsonObj->AddToObj(UriV[N], ResJson);
	}

	return JsonObj;
}

// suggest related concepts to the one(s) provided in the uri parameter
// ignore the ones provided in ignoreUri
PJsonVal TConceptSrv::GetRelatedConcepts(const TStrKdV& FldNmValPrV)
{
	// analyze most recent articles (recent=true) or randomly selected ones (recent = false)
	const bool RecentP = GetFldBool(FldNmValPrV, "recent", false);
	// don't count articles where concept weight for the article is smaller than minConceptWgt
	const int MinConceptWgt = GetFldInt(FldNmValPrV, "minConceptWgt", 0);
	// don't count concept as a related in an article if it's weight is smaller than minRelatedConceptWgt
	const int MinRelatedConceptWgt = GetFldInt(FldNmValPrV, "minRelatedConceptWgt", 0);
	// number of returned suggested concepts
	const int Count = GetFldInt(FldNmValPrV, "count", 20);
	// sample size of articles to consider
	const int ArticleSampleSize = GetFldInt(FldNmValPrV, "articleSampleSize", 2000);
	// the concept types of interest
	TStrV TypeV; GetFldValV(FldNmValPrV, "conceptType", TypeV);
	EnumConceptType ValidTypes = NewsConst.GetValidTypes(TypeV);

	// one or more concepts to which we want to get related concepts to
	TStrV ConceptUriV; GetFldValV(FldNmValPrV, "uri", ConceptUriV);
	for (int N = ConceptUriV.Len() - 1; N >= 0; N--) {
		if (!NewsBase->ConceptStore->IsRecNm(ConceptUriV[N]))
			ConceptUriV.Del(N);
	}
	if (ConceptUriV.Len() == 0)
		return TJsonVal::NewObj("error", TJsonVal::NewStr("No valid concept uris were provided"));

	TStrV IgnoreUriV; GetFldValV(FldNmValPrV, "ignoreUri", IgnoreUriV);
	TUInt64V IgnoreIdV;
	for (int N = 0; N < IgnoreUriV.Len(); N++) {
		if (NewsBase->ConceptStore->IsRecNm(IgnoreUriV[N]))
			IgnoreIdV.Add(NewsBase->ConceptStore->GetRecId(IgnoreUriV[N]));
	}

	TUInt64H ConceptFqH;
	for (int N = 0; N < ConceptUriV.Len(); N++) {
		PRecSet ArticleSet = NewsBase->ConceptStore->GetRec(ConceptUriV[N]).DoJoin(NewsBase->Base, "hasArticle");
		if (MinConceptWgt > 0)
			ArticleSet->FilterByFq(MinConceptWgt, TInt::Mx);
		if (RecentP)
			ArticleSet = ArticleSet->GetLimit(ArticleSampleSize, MAX(0, ArticleSet->GetRecs() - ArticleSampleSize));
		else
			ArticleSet = ArticleSet->GetSampleRecSet(ArticleSampleSize, false);
		PRecSet ConceptSet = ArticleSet->DoJoin(NewsBase->Base, "hasConcept");
		for (int C = 0; C < ConceptSet->GetRecs(); C++) {
			if (ConceptSet->GetRecFq(C) >= MinRelatedConceptWgt)
				ConceptFqH.AddDat(ConceptSet->GetRecId(C)) += 1;
		}
	}

	PJsonVal SuggArr = TJsonVal::NewArr();
	ConceptFqH.SortByDat(false);
	TUInt64Set IgnoreSet(IgnoreIdV);
	TUInt64IntKdV ConceptIdFqV;
	int MaxVal = -1;
	for (int KeyId = ConceptFqH.FFirstKeyId(); ConceptFqH.FNextKeyId(KeyId) && ConceptIdFqV.Len() < Count;) {
		uint64 ConceptId = ConceptFqH.GetKey(KeyId);
		if (IgnoreSet.IsKey(ConceptId))
			continue;
		const TStr& Type = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
		if (!NewsConst.IsValidType(Type, ValidTypes))
			continue;
		if (MaxVal < 0)
			MaxVal = ConceptFqH[KeyId];
		const int Fq = 100 * ConceptFqH[KeyId] / MaxVal;
		ConceptIdFqV.Add(TUInt64IntKd(ConceptId, Fq));
	}

	PRecSet SuggConceptSet = TRecSet::New(NewsBase->ConceptStore, ConceptIdFqV, true);
	return NewsSrvUtil->Concepts(SuggConceptSet);
}

PJsonVal TConceptSrv::GetArticles(const TStrKdV& FldNmValPrV)
{
	if (!IsFldNm(FldNmValPrV, "uri"))
		return TJsonVal::NewObj("error", TJsonVal::NewStr("Concept uri was not provided"));
	TStr Uri = GetFldVal(FldNmValPrV, "uri", "");
	if (!NewsBase->ConceptStore->IsRecNm(Uri))
		return TJsonVal::NewObj("error", TJsonVal::NewStr("Provided uri is not valid concept uri."));

	// take the list of last 500 articles and find the common concepts in them
	const int RelatedConceptCount = GetFldInt(FldNmValPrV, "relatedConceptCount", 0);
	// include info about the top concepts in the articles
	const bool IncludeConceptInfo = GetFldBool(FldNmValPrV, "includeConceptInfo", false);
	// split articles by days and extract top topics (concepts) for each day
	const bool ExtractDailyTopics = GetFldBool(FldNmValPrV, "extractDailyTopics", false);
	// the length of the article body to return
	const int ArticleBodyLen = GetFldInt(FldNmValPrV, "articleBodyLen", DefaultArticleBodyLen);

	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptLang", ConceptLangV);
	if (ConceptLangV.Empty())
		ConceptLangV.Add("eng");

	const int Page = GetFldInt(FldNmValPrV, "page", 0);
	const int Count = GetFldInt(FldNmValPrV, "count", DefaultNumberOfReturnedArticlesPerPage);
	const TStr SortBy = GetFldVal(FldNmValPrV, "sortBy", "");		// date or rel
	const bool SortByRel = (SortBy.GetLc() == "rel");

	PRecSet ArticleSet; TStr Error;
	bool Ok = NewsSearch->FindArticlesForConcept(Uri, ArticleSet, Error);
	if (!Ok)
		return TJsonVal::NewObj("error", Error);

	if (SortByRel)
		ArticleSet->SortByFq(false);
	else
		ArticleSet->Reverse();

	const int ArticleCount = ArticleSet->GetRecs();
	PRecSet ArticleSubSet = ArticleSet->GetLimit(Count, Page*Count);		// skip the first Offset articles and include only Count articles

	PJsonVal ResJson = TJsonVal::NewObj();
	// add articles
	ResJson->AddToObj("results", NewsSrvUtil->Articles(ArticleSubSet, ArticleBodyLen));
	// include daily topics
	if (ExtractDailyTopics && SortByRel == false) {
		TUInt64Set IgnoreSet; IgnoreSet.AddKey(NewsBase->ConceptStore->GetRecId(Uri));
		ResJson->AddToObj("dailyTopics", NewsSrvUtil->DailyTopics(ArticleSet, IgnoreSet, ConceptLangV, Page, Count));
	}
	// store concept details
	if (IncludeConceptInfo)
		ResJson->AddToObj("conceptInfo", NewsSrvUtil->Concept(NewsBase->ConceptStore->GetRecId(Uri), -1, ConceptLangV));

	if (RelatedConceptCount > 0) {
		// get concepts from the last X articles
		PRecSet ArticleSubSet2 = ArticleSet->GetLimit(500, 0);
		PRecSet RelatedConceptSet = ArticleSubSet2->DoJoin(NewsBase->Base, "hasConcept", 1000, true);
		RelatedConceptSet->SortByFq(false);
		RelatedConceptSet->Trunc(RelatedConceptCount);
		ResJson->AddToObj("relatedConcepts", NewsSrvUtil->Concepts(RelatedConceptSet, TUInt64Set(), TInt::Mx, ctAll, ConceptLangV));
	}
	ResJson->AddToObj("resultCount", TJsonVal::NewStr(TInt::GetStr(ArticleCount)));
	return ResJson;
}

