/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "EventCompareSrv.h"


TStr TEventCompareSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		ResJson = ProcessRequest(FldNmValPrV);
		// if one of the calls returned null we need to init to empty object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TEventCompareSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TEventCompareSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TEventCompareSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TEventCompareSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

PJsonVal TEventCompareSrv::ProcessRequest(const TStrKdV& FldNmValPrV)
{
	PRecSet EventRecSet0, EventRecSet1;
	PJsonVal ErrorJson0, ErrorJson1;
	GetEvents(FldNmValPrV, 0, EventRecSet0, ErrorJson0);
	GetEvents(FldNmValPrV, 1, EventRecSet1, ErrorJson1);
	if (!ErrorJson0.Empty())
		return ErrorJson0;
	if (!ErrorJson1.Empty())
		return ErrorJson1;
	if (EventRecSet0.Empty() || EventRecSet1.Empty())
		return TJsonVal::NewObj("error", TStr("No results match the criteria"));

	PJsonVal ResJson = NULL;
	TStr Action = GetFldVal(FldNmValPrV, "action", "");
	if (Action == "compareConcepts")
		ResJson = CompareConcepts(FldNmValPrV, EventRecSet0, EventRecSet1);
	else if (Action == "compareLocations")
		ResJson = CompareLocations(FldNmValPrV, EventRecSet0, EventRecSet1);
	else if (Action == "compareLocationsWithTime")
		ResJson = CompareLocationsWithTime(FldNmValPrV, EventRecSet0, EventRecSet1);
	else if (Action == "compareTimeline")
		ResJson = CompareTimeline(FldNmValPrV, EventRecSet0, EventRecSet1);
	else if (Action == "compareCategories")
		ResJson = CompareCategories(FldNmValPrV, EventRecSet0, EventRecSet1);
	else
		ResJson = TJsonVal::NewObj("error", TStr("No known action was provided in the request. Valid actions are compareConcepts, compareLocations, compareTimeline, compareCategories."));
	// add the count of events in both groups
	if (!ResJson.Empty()) {
		PJsonVal CountJson = TJsonVal::NewObj();
		CountJson->AddToObj("group0", EventRecSet0.Empty() ? 0 : EventRecSet0->GetRecs());
		CountJson->AddToObj("group1", EventRecSet1.Empty() ? 0 : EventRecSet1->GetRecs());
		ResJson->AddToObj("counts", CountJson);
	}
	return ResJson;
}

// parse the query conditions from the FldNmValPrV
// GroupIndex should be 0 or 1 - this will parse conceptUri0, locationUri0, ... or conceptUri1, locationUri1...
bool TEventCompareSrv::GetEvents(const TStrKdV& FldNmValPrV, const int& GroupIndex, PRecSet& EventRecSet, PJsonVal& ErrorJson)
{
	TStr Group = TStr::Fmt("%d", GroupIndex);

	TEventQuery Query;
	GetFldValV(FldNmValPrV, "conceptUri" + Group, Query.ConceptUriV);
	Query.Keywords = GetFldVal(FldNmValPrV, "keywords" + Group, "");
	Query.DateStart = GetFldVal(FldNmValPrV, "timeStart" + Group, "");
	Query.DateEnd = GetFldVal(FldNmValPrV, "timeEnd" + Group, "");
	GetFldValV(FldNmValPrV, "lang" + Group, Query.LangV);
	GetFldValV(FldNmValPrV, "locationUri" + Group, Query.LocationUriV);
	GetFldValV(FldNmValPrV, "publisherUri" + Group, Query.PublisherUriV);
	GetFldValV(FldNmValPrV, "categoryUri" + Group, Query.CategoryV);
	Query.CategoryIncludeSub = GetFldBool(FldNmValPrV, "categoryIncludeSub" + Group, true);

	TInt MinArts = GetFldInt(FldNmValPrV, "minArticlesInEvent" + Group, 0);
	TInt MaxArts = GetFldInt(FldNmValPrV, "maxArticlesInEvent" + Group, TInt::Mx);
	Query.MinMaxArtsInEvent = TIntPr(MinArts, MaxArts);
	
	TStrV DateMentionV; GetFldValV(FldNmValPrV, "dateMention" + Group, DateMentionV);
	TStr DateMentionStart = GetFldVal(FldNmValPrV, "dateMentionStart" + Group, "");
	TStr DateMentionEnd = GetFldVal(FldNmValPrV, "dateMentionEnd" + Group, "");
	Query.DateMentionV = EventRegBase->NewsSearch->ParseDateMentions(DateMentionV, DateMentionStart, DateMentionEnd);
	
	Query.IgnoreKeywords = GetFldVal(FldNmValPrV, "ignoreKeywords" + Group, "");
	GetFldValV(FldNmValPrV, "ignoreLang" + Group, Query.IgnoreLangV);
	GetFldValV(FldNmValPrV, "ignoreConceptUri" + Group, Query.IgnoreConceptUriV);
	GetFldValV(FldNmValPrV, "ignoreLocationUri" + Group, Query.IgnoreLocationUriV);
	GetFldValV(FldNmValPrV, "ignorePublisherUri" + Group, Query.IgnorePublisherUriV);
	GetFldValV(FldNmValPrV, "ignoreCategoryUri" + Group, Query.IgnoreCategoryV);
	Query.IgnoreCategoryIncludeSub = GetFldBool(FldNmValPrV, "ignoreCategoryIncludeSub" + Group, true);

	return EventRegBase->NewsSearch->SearchEvents(Query, EventRecSet, ErrorJson);
}

// use the two rec sets to identify concepts that separate first group from the other and vice-versa
PJsonVal TEventCompareSrv::CompareConcepts(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet0, const PRecSet& EventRecSet1)
{
	// did we want to return just concepts of specific type?
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptType", ConceptTypeV);
	EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("");

	THash<TUInt64, TFlt> Concepts0, Concepts1;
	EventRegBase->GetConceptsForEvents(EventRecSet0, ValidTypes, true, Concepts0);
	EventRegBase->GetConceptsForEvents(EventRecSet1, ValidTypes, true, Concepts1);
	// hide the searched concepts from Concepts0 and Concepts1 - they would have too high weight
	TStrV ConceptUriV, ConceptUri2V;
	GetFldValV(FldNmValPrV, "conceptUri0", ConceptUriV);
	GetFldValV(FldNmValPrV, "conceptUri1", ConceptUri2V);
	ConceptUriV.AddV(ConceptUri2V);
	for (int N = 0; N < ConceptUriV.Len(); N++) {
		if (NewsBase->ConceptStore->IsRecNm(ConceptUriV[N])) {
			Concepts0.DelIfKey(NewsBase->ConceptStore->GetRecId(ConceptUriV[N]));
			Concepts1.DelIfKey(NewsBase->ConceptStore->GetRecId(ConceptUriV[N]));
		}
	}
	
	// find concepts that most differ between the two groups
	THash<TUInt64, TFlt> TopGroup0, TopGroup1, SumGroups;
	for (int KeyId = Concepts0.FFirstKeyId(); Concepts0.FNextKeyId(KeyId);) {
		uint64 ConcId = Concepts0.GetKey(KeyId);
		TFlt C0, C1; Concepts0.IsKeyGetDat(ConcId, C0); Concepts1.IsKeyGetDat(ConcId, C1);
		TopGroup0.AddDat(ConcId, C0 - C1);
		SumGroups.AddDat(ConcId) += C0;
	}
	for (int KeyId = Concepts1.FFirstKeyId(); Concepts1.FNextKeyId(KeyId);) {
		uint64 ConcId = Concepts1.GetKey(KeyId);
		TFlt C0, C1; Concepts0.IsKeyGetDat(ConcId, C0); Concepts1.IsKeyGetDat(ConcId, C1);
		TopGroup1.AddDat(ConcId, C1 - C0);
		SumGroups.AddDat(ConcId) += C1;
	}
	// TODO: optimize the sorting - only find top count elements
	Concepts0.Defrag(); Concepts1.Defrag();
	Concepts0.SortByDat(false); Concepts1.SortByDat(false);
	TopGroup0.SortByDat(false); TopGroup1.SortByDat(false); 
	SumGroups.SortByDat(false);
	
	int Count = GetFldInt(FldNmValPrV, "count", 30);
	TJsonValV TopArr0, TopArr1, DiffArr0, DiffArr1, CommonArr;
	for (int N = 0, KeyId = Concepts0.FFirstKeyId(); N < Count && Concepts0.FNextKeyId(KeyId); N++) {
		PJsonVal ConceptJson = NewsSrvUtil->Concept(Concepts0.GetKey(KeyId), (int) Concepts0[KeyId], ConceptLangV);
		TopArr0.Add(ConceptJson);
	}
	for (int N = 0, KeyId = Concepts1.FFirstKeyId(); N < Count && Concepts1.FNextKeyId(KeyId); N++) {
		PJsonVal ConceptJson = NewsSrvUtil->Concept(Concepts1.GetKey(KeyId), (int) Concepts1[KeyId], ConceptLangV);
		TopArr1.Add(ConceptJson);
	}
	for (int N = 0, KeyId = TopGroup0.FFirstKeyId(); N < Count && TopGroup0.FNextKeyId(KeyId); N++) {
		PJsonVal ConceptJson = NewsSrvUtil->Concept(TopGroup0.GetKey(KeyId), -1, ConceptLangV);
		uint64 ConcId = TopGroup0.GetKey(KeyId);
		ConceptJson->AddToObj("groupFreq0", TJsonVal::NewNum(Concepts0.GetDat(ConcId)));
		ConceptJson->AddToObj("groupFreq1", TJsonVal::NewNum(Concepts1.IsKey(ConcId) ? Concepts1.GetDat(ConcId).Val : 0.0));
		DiffArr0.Add(ConceptJson);
	}
	for (int N = 0, KeyId = TopGroup1.FFirstKeyId(); N < Count && TopGroup1.FNextKeyId(KeyId); N++) {
		PJsonVal ConceptJson = NewsSrvUtil->Concept(TopGroup1.GetKey(KeyId), -1, ConceptLangV);
		uint64 ConcId = TopGroup1.GetKey(KeyId);
		ConceptJson->AddToObj("groupFreq0", TJsonVal::NewNum(Concepts0.IsKey(ConcId) ? Concepts0.GetDat(ConcId).Val : 0.0));
		ConceptJson->AddToObj("groupFreq1", TJsonVal::NewNum(Concepts1.GetDat(ConcId)));
		DiffArr1.Add(ConceptJson);
	}
	for (int N = 0, KeyId = SumGroups.FFirstKeyId(); N < Count && SumGroups.FNextKeyId(KeyId); N++) {
		PJsonVal ConceptJson = NewsSrvUtil->Concept(SumGroups.GetKey(KeyId), int (SumGroups[KeyId]), ConceptLangV);
		CommonArr.Add(ConceptJson);
	}
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("topConc0", TopArr0);		// top concepts from group0
	ResJson->AddToObj("topConc1", TopArr1);		// top concepts from group1
	ResJson->AddToObj("diffConc0", DiffArr0);	// concepts from group0 most different to group1
	ResJson->AddToObj("diffConc1", DiffArr1);	// concepts from group1 most different to group0
	ResJson->AddToObj("commonConc", CommonArr);	// top shared concepts
	return ResJson;
}


PJsonVal TEventCompareSrv::CompareLocations(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet0, const PRecSet& EventRecSet1)
{
	PJsonVal RetJson = TJsonVal::NewObj();
	RetJson->AddToObj("group0", EventRegSrvUtil->EventsLocAggr(EventRecSet0));
	RetJson->AddToObj("group1", EventRegSrvUtil->EventsLocAggr(EventRecSet1));
	return RetJson;
}

PJsonVal TEventCompareSrv::CompareLocationsWithTime(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet0, const PRecSet& EventRecSet1)
{
	PJsonVal RetJson = TJsonVal::NewObj();
	RetJson->AddToObj("group0", EventRegSrvUtil->EventsLocTimeAggr(EventRecSet0));
	RetJson->AddToObj("group1", EventRegSrvUtil->EventsLocTimeAggr(EventRecSet1));
	return RetJson;
}

PJsonVal TEventCompareSrv::CompareTimeline(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet0, const PRecSet& EventRecSet1)
{
	PJsonVal RetJson = TJsonVal::NewObj();
	RetJson->AddToObj("group0", EventRegSrvUtil->EventsTimeAggr(EventRecSet0));
	RetJson->AddToObj("group1", EventRegSrvUtil->EventsTimeAggr(EventRecSet1));
	return RetJson;
}

PJsonVal TEventCompareSrv::CompareCategories(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet0, const PRecSet& EventRecSet1)
{
	PJsonVal RetJson = TJsonVal::NewObj();
	RetJson->AddToObj("group0", EventRegSrvUtil->EventsCategoryAggr(EventRecSet0));
	RetJson->AddToObj("group1", EventRegSrvUtil->EventsCategoryAggr(EventRecSet1));
	return RetJson;
}