/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "EventRegSrv.h"

#include <base.h>
#include <mine.h>
//#include <gks/gkswf.h>
//#include <gks/vizmapgks.h>


// //////////////
// Event registry methods

uint64 TEventRegSrv::GetEventId(const TStrKdV& FldNmValPrV)
{
	TStr EventUri = GetFldVal(FldNmValPrV, "eventUri");
	if (EventRegBase->EventStore->IsRecNm(EventUri))
		return EventRegBase->EventStore->GetRecId(EventUri);
	return TUInt64::Mx;
}

TStr TEventRegSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try
	{
		// event functions
		TStr Action = GetFldVal(FldNmValPrV, "action", "");

		if (Action == "getEvents")
			ResJson = GetEvents(FldNmValPrV);

		else if (Action == "getEvent")
			ResJson = GetEvent(FldNmValPrV);

		else if (Action == "getConceptGraph")
			ResJson = GetConceptGraph(FldNmValPrV);
		else if (Action == "getConceptsDocAtlas")
			ResJson = GetConceptsDocAtlas(FldNmValPrV);

		else if (Action == "getUpdatedEvents")
			ResJson = GetUpdatedEvents(FldNmValPrV);
		else if (Action == "getTrendingEvents")
			ResJson = GetTrendingEvents(FldNmValPrV);

		else if (Action == "getSimilarConcepts")
			ResJson = GetSimilarConcepts(FldNmValPrV);

		else if (Action == "getEventIdForStory")
			ResJson = GetEventIdForStory(FldNmValPrV);

		else
		{
			ResJson = TJsonVal::NewObj();
			ResJson->AddToObj("error", TStr("Unrecognized action specified. Valid values are: getEvents, getEventInfo, getEventTrend, getEventArticles, getTrendingEvents, getSimilarEvents, getSimilarStories, getSimilarConcepts, getConceptGraph."));
			TNotify::StdNotify->OnStatus("TEventRegSrv: Invalid action name: " + Action);
		}


		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TEventRegSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TEventRegSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TEventRegSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TEventRegSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

PJsonVal TEventRegSrv::GetEvents(const TStrKdV& FldNmValPrV)
{
	TEventQuery Query;
	Query.Keywords = GetFldVal(FldNmValPrV, "keywords", "");
	GetFldValV(FldNmValPrV, "conceptUri", Query.ConceptUriV);
	GetFldValV(FldNmValPrV, "lang", Query.LangV);
	GetFldValV(FldNmValPrV, "locationUri", Query.LocationUriV);
	GetFldValV(FldNmValPrV, "publisherUri", Query.PublisherUriV);
	GetFldValV(FldNmValPrV, "categoryUri", Query.CategoryV);
	Query.CategoryIncludeSub = GetFldBool(FldNmValPrV, "categoryIncludeSub", true);

	Query.DateStart = GetFldVal(FldNmValPrV, "dateStart", "");
	Query.DateEnd = GetFldVal(FldNmValPrV, "dateEnd", "");
	TInt MinArts = GetFldInt(FldNmValPrV, "minArticlesInEvent", 0);
	TInt MaxArts = GetFldInt(FldNmValPrV, "maxArticlesInEvent", TInt::Mx);
	Query.MinMaxArtsInEvent = TIntPr(MinArts, MaxArts);

	TStrV DateMentionV; GetFldValV(FldNmValPrV, "dateMention", DateMentionV);
	TStr DateMentionStart = GetFldVal(FldNmValPrV, "dateMentionStart", "");
	TStr DateMentionEnd = GetFldVal(FldNmValPrV, "dateMentionEnd", "");
	Query.DateMentionV = NewsSearch->ParseDateMentions(DateMentionV, DateMentionStart, DateMentionEnd);

	Query.IgnoreKeywords = GetFldVal(FldNmValPrV, "ignoreKeywords", "");
	GetFldValV(FldNmValPrV, "ignoreConceptUri", Query.IgnoreConceptUriV);
	GetFldValV(FldNmValPrV, "ignoreLang" , Query.IgnoreLangV);
	GetFldValV(FldNmValPrV, "ignoreLocationUri", Query.IgnoreLocationUriV);
	GetFldValV(FldNmValPrV, "ignorePublisherUri", Query.IgnorePublisherUriV);
	GetFldValV(FldNmValPrV, "ignoreCategoryUri", Query.IgnoreCategoryV);
	Query.IgnoreCategoryIncludeSub = GetFldBool(FldNmValPrV, "ignoreCategoryIncludeSub", true);

	// user can also provide a comma separated list of uris if he wishes to compute som aggegrates on them
	// if eventUriList is provided then other query arguments (if existing) will be ignored
	TStr EventUriList = GetFldVal(FldNmValPrV, "eventUriList", "");

	// get one or more result types that you would like
	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("events");

	PRecSet EventRecSet;
	if (EventUriList.Len() > 0) {
		TStrV EventUriV;
		EventUriList.SplitOnAllCh(',', EventUriV);
		TUInt64V EventIdV;
		for (int N = 0; N < EventUriV.Len(); N++) {
			if (EventStore->IsRecNm(EventUriV[N]))
				EventIdV.Add(EventStore->GetRecId(EventUriV[N]));
		}
		EventRecSet = TRecSet::New(EventStore, EventIdV);
	}
	else
	{
		PJsonVal ErrorJson;
		const bool Ok = NewsSearch->SearchEvents(Query, EventRecSet, ErrorJson);
		if (!Ok && !ErrorJson.Empty())
			return ErrorJson;
	}
	if (EventRecSet.Empty() || EventRecSet->GetRecs() == 0)
		return TJsonVal::NewObj("error", TStr("No results match the search conditions."));

	PJsonVal ResJson = TJsonVal::NewObj();
	// list of events
	if (ResultTypeV.IsIn("events")) {
		if (TSAppSrvFun::GetFldInt(FldNmValPrV, "eventsCount", 25) > 200) {
			PJsonVal ErrJson = TJsonVal::NewObj("error", TStr("The API does not allow downloading more than 200 events per call."));
			ResJson->AddToObj("events", ErrJson);
		}
		else
			ResJson->AddToObj("events", NewsSrvUtil->Events(FldNmValPrV, EventRecSet));
	}
	// only the list of event uris
	if (ResultTypeV.IsIn("uriList")) {
		TStrV UriV;
		for (int N = 0; N < EventRecSet->GetRecs(); N++)
			UriV.Add(TUInt64(EventRecSet->GetRecId(N)).GetStr());
		ResJson->AddToObj("uriList", TJsonVal::NewArr(UriV));
	}
	// time aggregates
	if (ResultTypeV.IsIn("timeAggr"))
		ResJson->AddToObj("timeAggr", EventRegSrvUtil->EventsTimeAggr(EventRecSet));
	// location aggregates
	if (ResultTypeV.IsIn("locAggr")) {
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "locAggrConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		ResJson->AddToObj("locAggr", EventRegSrvUtil->EventsLocAggr(EventRecSet, ConceptLangV));
	}
	// location at time aggregates
	if (ResultTypeV.IsIn("locTimeAggr")) {
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "locTimeAggrConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		ResJson->AddToObj("locTimeAggr", EventRegSrvUtil->EventsLocTimeAggr(EventRecSet, ConceptLangV));
	}
	// top publishers for the events
	if (ResultTypeV.IsIn("topPublisherAggr")) {
		bool IncludePublisherDetails = GetFldBool(FldNmValPrV, "topPublisherAggrIncludePublisherDetails", true);
		int TopPublishers = GetFldInt(FldNmValPrV, "topPublisherAggrTopPublisherCount", 30);
		ResJson->AddToObj("topPublisherAggr", EventRegSrvUtil->EventsTopPublisherAggr(EventRecSet, TopPublishers, IncludePublisherDetails));
	}
	// concept aggregates
	if (ResultTypeV.IsIn("conceptAggr")) {
		TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptAggrConceptType", ConceptTypeV);
		EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
		const int ConceptCount = GetFldInt(FldNmValPrV, "conceptAggrConceptCount", 25);
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptAggrConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		ResJson->AddToObj("conceptAggr", EventRegSrvUtil->EventsConceptAggr(EventRecSet, ValidTypes, ConceptCount, ConceptLangV));
	}
	// keyword aggregates
	if (ResultTypeV.IsIn("keywordAggr")) {
		const TStr Lang = GetFldVal(FldNmValPrV, "keywordAggrLang", "eng");
		//const int KeywordCount = GetFldInt(FldNmValPrV, "keywordAggrKeywordCount", 50);
		const int SampleSize = GetFldInt(FldNmValPrV, "keywordAggrSampleSize", 500);
		ResJson->AddToObj("keywordAggr", EventRegSrvUtil->EventsKeywordAggr(EventRecSet, Lang, SampleSize));
	}
	// concept graph
	if (ResultTypeV.IsIn("conceptGraph")) {
		TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptGraphConceptType", ConceptTypeV);
		EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptGraphConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		const int ConceptCount = GetFldInt(FldNmValPrV, "conceptGraphConceptCount", 25);
		const int LinkCount = GetFldInt(FldNmValPrV, "conceptGraphLinkCount", 50);
		const int SampleSize = GetFldInt(FldNmValPrV, "conceptGraphSampleSize", 500);

		ResJson->AddToObj("conceptGraph", EventRegSrvUtil->EventsConceptGraph(EventRecSet, ValidTypes, ConceptLangV, ConceptCount, LinkCount, SampleSize));
	}
	// concept co-ocurrence matrix
	if (ResultTypeV.IsIn("conceptMatrix")) {
		TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptMatrixConceptType", ConceptTypeV);
		EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptMatrixConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		const int ConceptCount = GetFldInt(FldNmValPrV, "conceptMatrixConceptCount", 25);
		const TStr MeasureStr = GetFldVal(FldNmValPrV, "conceptMatrixMeasure", "pmi");		// pmi (pointwise mutual information), pairTfIdf (pair frequence * IDF of individual concepts), chiSquare
		const int SampleSize = GetFldInt(FldNmValPrV, "conceptMatrixSampleSize", 500);
		TEventRegSrvUtil::MatrixScoringMethod Measure = TEventRegSrvUtil::pmi;
		if (MeasureStr == "pmi") Measure = TEventRegSrvUtil::pmi;
		else if (MeasureStr == "pairTfIdf") Measure = TEventRegSrvUtil::pairTfIdf;
		else if (MeasureStr == "chiSquare") Measure = TEventRegSrvUtil::chiSquare;

		ResJson->AddToObj("conceptMatrix", EventRegSrvUtil->EventsConceptMatrix(EventRecSet, ValidTypes, ConceptLangV, ConceptCount, Measure, SampleSize));
	}
	// trending concepts
	if (ResultTypeV.IsIn("trendingConcepts")) {
		const int ConceptCount = GetFldInt(FldNmValPrV, "trendingConceptsConceptCount", 10);
		// do we have custom concept urls provided?
		TStrV TrendingConceptsStrV; GetFldValV(FldNmValPrV, "trendingConceptsConceptUri", TrendingConceptsStrV);
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "trendingConceptsConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");

		// find which concepts do we want to show trends for
		if (TrendingConceptsStrV.Len() == 0) {
			TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "trendingConceptsConceptType", ConceptTypeV);
			EnumConceptType ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
			THash<TUInt64, TFlt> Concepts;
			EventRegBase->GetConceptsForEvents(EventRecSet, ValidTypes, true, Concepts);
			Concepts.SortByDat(false);
			for (int Added = 0, KeyId = Concepts.FFirstKeyId(); Concepts.FNextKeyId(KeyId) && Added < ConceptCount; Added++)
				TrendingConceptsStrV.Add(NewsBase->ConceptStore->GetRecNm(Concepts.GetKey(KeyId)));
		}

		ResJson->AddToObj("trendingConcepts", EventRegSrvUtil->EventsTrendingConcepts(EventRecSet, TrendingConceptsStrV, ConceptCount, ConceptLangV));
	}
	// date mention aggregates
	if (ResultTypeV.IsIn("dateMentionAggr")) {
		const int MinDaysApart = GetFldInt(FldNmValPrV, "dateMentionAggrMinDaysApart", 0);
		const int MinDateMentionCount = GetFldInt(FldNmValPrV, "dateMentionAggrMinDateMentionCount", 0);

		ResJson->AddToObj("dateMentionAggr", EventRegSrvUtil->EventsDateMentionAggr(EventRecSet, MinDaysApart, MinDateMentionCount));
	}
	// clusters aggregates
	if (ResultTypeV.IsIn("eventClusters"))
		ResJson->AddToObj("eventClusters", EventRegSrvUtil->EventsGetEventClusters(FldNmValPrV, EventRecSet));

	// category aggregates
	if (ResultTypeV.IsIn("categoryAggr"))
		ResJson->AddToObj("categoryAggr", EventRegSrvUtil->EventsCategoryAggr(EventRecSet));

	// return the recently edited events and their timing information
	if (ResultTypeV.IsIn("recentActivity")) {
		PJsonVal EventsJson = TJsonVal::NewObj("events", NewsSrvUtil->EventsRecentActivity(FldNmValPrV, EventRecSet));
		ResJson->AddToObj("recentActivity", EventsJson);
	}

	if (ResultTypeV.IsIn("docAtlasOnConcepts"))
		ResJson->AddToObj("docAtlasOnConcepts", GetDocAtlasOnConcepts(FldNmValPrV, EventRecSet));

	if (ResJson->GetObjKeys() == 0)
		ResJson->AddToObj("error", TStr("unrecognized resultType was specified. Recognized values are list, timeAggr, locAggr, locTimeAggr, dateMentionAggr, categoryAggr, conceptAggr, keywordAggr, eventClusters, conceptMatrix, conceptGraph, trendingConcepts, recentActivity."));

	return ResJson;
}

// return information about one or more events
// the information returned depends on the value of the resultType parameter
PJsonVal TEventRegSrv::GetEvent(const TStrKdV& FldNmValPrV)
{
	TStrV EventUriV; GetFldValV(FldNmValPrV, "eventUri", EventUriV);

	if (EventUriV.Len() == 0)
		return TJsonVal::NewObj("error", TStr("There is no event uri provided in the request"));
	else if (EventUriV.Len() == 1)
		return GetSingleEvent(FldNmValPrV, EventUriV[0]);
	else {
		if (EventUriV.Len() > 200)
			return TJsonVal::NewObj("error", TStr("The API does not allow downloading more than 200 events per call."));

		PJsonVal ResJson = TJsonVal::NewObj();
		for (int N = 0; N < EventUriV.Len(); N++) {
			const TStr EventUri = EventUriV[N];
			ResJson->AddToObj(EventUri, GetSingleEvent(FldNmValPrV, EventUri));
		}
		return ResJson;
	}
}

// return all necessary info about a single event (called by GetEvent)
PJsonVal TEventRegSrv::GetSingleEvent(const TStrKdV& FldNmValPrV, const TStr& EventUri)
{
	if (!NewsBase->EventStore->IsRecNm(EventUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Provided event uri (%s) is not a valid event uri", EventUri.CStr()));

	const uint64 _EventId = NewsBase->EventStore->GetRecId(EventUri);
	// check if the provided event has some stories assigned to it. if not, find the event where the story was merged to and return the info
	const uint64 EventId = EventRegBase->GetValidEventId(_EventId);
	if (EventId == TUInt64::Mx)
		return TJsonVal::NewObj("error", TStr::Fmt("Provided event uri (%s) is not a valid event uri", EventUri.CStr()));
	if (EventId != _EventId) {
		PJsonVal RedirectJson = TJsonVal::NewObj("newEventUri", NewsBase->EventStore->GetRecNm(EventId));
		RedirectJson->AddToObj("info", TStr::Fmt("Event '%s' was merged to event '%s'.", EventUri.CStr(), NewsBase->EventStore->GetRecNm(EventId).CStr()));
		return RedirectJson;
	}

	// get one or more result types that you would like
	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("info");

	PJsonVal ResJson = TJsonVal::NewObj();

	// return event info
	if (ResultTypeV.IsIn("info")) {
		TUInt64V EventV; EventV.Add(EventId);
		PRecSet EventSet = TRecSet::New(EventStore, EventV);
		PJsonVal EventsResJson = NewsSrvUtil->Events(FldNmValPrV, EventSet, "info");
		PJsonVal EventJson = TJsonVal::NewObj();
		if (EventsResJson->IsObjKey("results") && EventsResJson->GetObjKey("results")->IsArr() && EventsResJson->GetObjKey("results")->GetArrVals() > 0)
			EventJson = EventsResJson->GetObjKey("results")->GetArrVal(0);
		EventJson->DelObjKey("wgt");
		ResJson->AddToObj("info", EventJson);
	}

	// list of articles
	if (ResultTypeV.IsIn("articles")) {
		TStrV LangV; GetFldValV(FldNmValPrV, "articlesLang", LangV);
		if (LangV.Len() == 0)
			LangV = NewsConst.LangCoreV;
		PRecSet ArticleSet = EventRegSrvUtil->GetEventArticles(EventId, LangV);
		ResJson->AddToObj("articles", NewsSrvUtil->Articles(FldNmValPrV, ArticleSet));
	}

	// list of similar articles - these are articles that are not currently in the event but have a has similar join to it
	// for each language we only consider similar articles in the same language
	if (ResultTypeV.IsIn("relatedArticles")) {
		// get the valid join names
		TStrV LangV; GetFldValV(FldNmValPrV, "articlesLang", LangV);
		if (LangV.Len() == 0)
			LangV = NewsConst.LangCoreV;

		THash<TUInt64, TInt> SimilarArtH;
		for (int LangN = 0; LangN < LangV.Len(); LangN++) {
			// get event articles
			TStrV SingleLangV; SingleLangV.Add(LangV[LangN]);
			PRecSet ArticleSet = EventRegSrvUtil->GetEventArticles(EventId, SingleLangV);
			TUInt64V ArticleV; ArticleSet->GetRecIdV(ArticleV);
			THashSet<TUInt64> IgnoredArtH(ArticleV);

			const TStr JoinNm = "similar_" + LangV[LangN];
			if (!ArticleStore->IsJoinNm(JoinNm))
				continue;

			for (int N = 0; N < ArticleSet->GetRecs(); N++) {
				const uint64 ArtId = ArticleSet->GetRecId(N);
				const TInt ArtWgt = (int) (ArticleStore->GetFieldNmFlt(ArtId, "CosToCentroid") * 100);
				PRecSet SimilarArts = ArticleStore->GetRec(ArtId).DoJoin(NewsBase->Base, JoinNm);
				for (int ArtN = 0; ArtN < SimilarArts->GetRecs(); ArtN++) {
					const uint64 SimArtId = SimilarArts->GetRecId(ArtN);
					// if the article is already in the event or if it is a duplicate then ignore it
					if (IgnoredArtH.IsKey(SimArtId) || ArticleStore->GetFieldNmUInt64(SimArtId, "isDuplicateOfId") != TUInt64::Mx)
						continue;
					SimilarArtH.AddDat(SimArtId) += ArtWgt * SimilarArts->GetRecFq(ArtN);
				}
			}
		}
		SimilarArtH.SortByDat(false);
		TUInt64IntKdV SimilarKdV; SimilarArtH.GetKeyDatKdV(SimilarKdV);
		PRecSet SimilarArtSet = TRecSet::New(ArticleStore, SimilarKdV, true);
		ResJson->AddToObj("relatedArticles", NewsSrvUtil->Articles(FldNmValPrV, SimilarArtSet));
	}

	// return list of article uris for one or more events
	// ret json : {"articleUris" : ["article uri1", "article uri2", ...] }
	if (ResultTypeV.IsIn("articleUris")) {
		TStrV LangV; GetFldValV(FldNmValPrV, "articleUrisLang", LangV);
		if (LangV.Len() == 0)
			LangV = NewsConst.LangCoreV;
		PRecSet ArticleSet = EventRegSrvUtil->GetEventArticles(EventId, LangV);
		ResJson->AddToObj("articleUris", NewsSrvUtil->ArticleUris(FldNmValPrV, ArticleSet));
	}

	if (ResultTypeV.IsIn("articleIds")) {
		TStrV LangV; GetFldValV(FldNmValPrV, "articlesIdsLang", LangV);
		if (LangV.Len() == 0)
			LangV = NewsConst.LangCoreV;
		PRecSet ArticleSet = EventRegSrvUtil->GetEventArticles(EventId, LangV);
		ResJson->AddToObj("articleIds", NewsSrvUtil->ArticleIds(FldNmValPrV, ArticleSet));
	}

	// tag cloud data for an event
	if (ResultTypeV.IsIn("keywordAggr")) {
		const TStr KeywordLang = GetFldVal(FldNmValPrV, "keywordAggrLang", "eng");
		//const int KeywordCount = GetFldInt(FldNmValPrV, "keywordCount", 50);
		ResJson->AddToObj("keywordAggr", EventRegSrvUtil->EventKeywordAggr(EventId, KeywordLang));
	}

	// source info and location aggregate
	if (ResultTypeV.IsIn("sourceAggr"))
		ResJson->AddToObj("sourceAggr", EventRegSrvUtil->EventSourceAggr(FldNmValPrV, EventId));

	// date mention aggregates
	if (ResultTypeV.IsIn("dateMentionAggr"))
		ResJson->AddToObj("dateMentionAggr", EventRegSrvUtil->EventDateMentionAggr(EventId));

	// article trends
	if (ResultTypeV.IsIn("articleTrend"))
		ResJson->AddToObj("articleTrend", EventRegSrvUtil->EventArticleTrend(FldNmValPrV, EventId));

	// similar events
	if (ResultTypeV.IsIn("similarEvents"))
		ResJson->AddToObj("similarEvents", SimilarEvents(FldNmValPrV, EventId));

	// similar stories
	if (ResultTypeV.IsIn("similarStories"))
		ResJson->AddToObj("similarStories", SimilarStories(FldNmValPrV, EventId));

	if (ResJson->GetObjKeys() == 0)
		ResJson->AddToObj("error", TStr("unrecognized resultType was specified. Recognized values are articleList, articleUris, info, keywordAggr, sourceAggr, dateMentionAggr, articleTrend."));

	return ResJson;
}

// get a list of events that are similar to the specified event
PJsonVal TEventRegSrv::SimilarEvents(const TStrKdV& FldNmValPrV, const uint64& EventId)
{
	const int Count = GetFldInt(FldNmValPrV, "similarEventsCount", 20);
	const bool IncludeThisEvent = GetFldBool(FldNmValPrV, "similarEventsIncludeSelf", false);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "similarEventsConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	const TStr Source = GetFldVal(FldNmValPrV, "similarEventsSource", "concept");	// concept, cca - how to compute similarity
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "similarEventsConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV);

	// settings for source == concept
	const int MaxDayDiff = GetFldInt(FldNmValPrV, "similarEventsMaxDayDiff", TInt::Mx);

	// add article trend info?
	const bool AddArticleTrendInfo = GetFldBool(FldNmValPrV, "similarEventsAddArticleTrendInfo", false);
	const int AggrHours = GetFldInt(FldNmValPrV, "similarEventsAggrHours", 6);	// group articles withing 6 hours into the same key

	EnumEventInfoFlags EventFlags = NewsSrvUtil->GetEventFlags(FldNmValPrV, "similarEvents");

	if (!EventStore->IsRecId(EventId))
		return TJsonVal::NewObj("error", TStr("There is no event for the provided eventId or storyId."));

	PRecSet SimEventRecSet;
	if (Source == "concept")
		SimEventRecSet = EventRegBase->GetSimilarEvents(EventId, Count - (IncludeThisEvent ? 1 : 0), MaxDayDiff, ConceptTypes);
	else if (Source == "cca")
		SimEventRecSet = EventRegBase->GetSimilarEventsByCCA(EventId, Count);
	else
		SimEventRecSet = TRecSet::New(NewsBase->EventStore);

	PJsonVal RetJson = TJsonVal::NewObj();
	TJsonValV SimEventV;
	for (int N=0; N < SimEventRecSet->GetRecs(); N++) {
		const uint64 SimEventId = SimEventRecSet->GetRecId(N);
		PJsonVal EventJson = NewsSrvUtil->Event(SimEventId, ConceptLangV, ConceptTypes, EventFlags);
		const double Sim = SimEventRecSet->GetRecFq(N) / 100.0;
		EventJson->AddToObj("sim", TJsonVal::NewNum(Sim));
		SimEventV.Add(EventJson);
	}
	if (IncludeThisEvent) {
		PJsonVal EventJson = NewsSrvUtil->Event(EventId, ConceptLangV, ConceptTypes, EventFlags);
		EventJson->AddToObj("sim", TJsonVal::NewNum(1.0));
		SimEventV.Add(EventJson);
	}
	RetJson->AddToObj("similarEvents", TJsonVal::NewArr(SimEventV));

	// add info that can be used to show riverflow of articles in each of the events
	if (AddArticleTrendInfo) {
		TUInt64V EventIdV;  SimEventRecSet->GetRecIdV(EventIdV);
		EventIdV.Add(EventId);
		// store dates up to an hour scale - e.g. 2014-01-15:02
		THash<TStr, THash<TUInt64, TInt> > DateToEventIdCountH;
		for (int E = 0; E < EventIdV.Len(); E++) {
			PRecSet ArticleSet = TRec(EventStore, EventIdV[E]).DoJoin(NewsBase->Base, "hasStory")->DoJoin(NewsBase->Base, "hasArticle");
			for (int A = 0; A < ArticleSet->GetRecs(); A++) {
				TTm Date; ArticleStore->GetFieldNmTm(ArticleSet->GetRecId(A), "DateTime", Date);
				const int HourKey = AggrHours * (int) (Date.GetHour() / AggrHours);	// make keys only for hours 00, 06, 12, 18
				TStr DateStr = TStr::Fmt("%4d-%02d-%02d:%02d", Date.GetYear(), Date.GetMonth(), Date.GetDay(), HourKey);
				DateToEventIdCountH.AddDat(DateStr).AddDat(EventIdV[E]) += 1;
			}
		}
		// create a trends array where each element is an object having "date" and an array of (concept, freq) for that date
		TJsonValV DateArrJsonV;
		DateToEventIdCountH.SortByKey(true);
		for (int KeyId = DateToEventIdCountH.FFirstKeyId(); DateToEventIdCountH.FNextKeyId(KeyId);) {
			const TStr Date = DateToEventIdCountH.GetKey(KeyId);
			PJsonVal DateJson = TJsonVal::NewObj("date", Date);
			PJsonVal EventFqs = TJsonVal::NewObj();
			THash<TUInt64, TInt>& EventIdCountH = DateToEventIdCountH[KeyId];
			for (int CKeyId = EventIdCountH.FFirstKeyId(); EventIdCountH.FNextKeyId(CKeyId);)
				EventFqs->AddToObj(EventIdCountH.GetKey(CKeyId).GetStr(), TJsonVal::NewNum(EventIdCountH[CKeyId]));
			DateJson->AddToObj("eventFreq", EventFqs);
			DateArrJsonV.Add(DateJson);
		}
		RetJson->AddToObj("trends", TJsonVal::NewArr(DateArrJsonV));
	}

	return RetJson;
}

// get stories that are similar to the given event
PJsonVal TEventRegSrv::SimilarStories(const TStrKdV& FldNmValPrV, const uint64& EventId)
{
	const int Count = GetFldInt(FldNmValPrV, "similarStoriesCount", 20);
	TStrV LangV; GetFldValV(FldNmValPrV, "similarStoriesLang", LangV);		// in which languages should be the stories
	TStrSet LangH(LangV);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "similarStoriesConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "similarStoriesConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV);
	const TStr Source = GetFldVal(FldNmValPrV, "similarStoriesSource", "concept");	// concept, cca - how to compute similarity

	// settings for source == concept
	const int MaxDayDiff = GetFldInt(FldNmValPrV, "similarStoriesMaxDayDiff", TInt::Mx);

	EnumStoryInfoFlags StoryFlags = NewsSrvUtil->GetStoryFlags(FldNmValPrV, "similarStories");

	PRecSet SimStoryRecSet;
	if (Source == "concept")
		SimStoryRecSet = EventRegBase->GetSimilarStoriesToEvent(EventId, Count, LangH, MaxDayDiff, ConceptTypes);
	else if (Source == "cca")
		SimStoryRecSet = EventRegBase->GetSimilarStoriesToEventByCCA(EventId, Count, LangH);
	else
		SimStoryRecSet = TRecSet::New(NewsBase->StoryStore);

	PJsonVal RetJson = TJsonVal::NewObj();
	TJsonValV SimStoryV;
	for (int N = 0; N < SimStoryRecSet->GetRecs(); N++) {
		uint64 SimStoryId = SimStoryRecSet->GetRecId(N);
		int Score = SimStoryRecSet->GetRecFq(N);
		PJsonVal StoryJson = NewsSrvUtil->Story(SimStoryId, Score, ConceptLangV, ConceptTypes, StoryFlags);
		StoryJson->AddToObj("sim", SimStoryRecSet->GetRecFq(N));
		SimStoryV.Add(StoryJson);
	}
	return TJsonVal::NewArr(SimStoryV);
}

// get concepts that appear in similar events
PJsonVal TEventRegSrv::GetSimilarConcepts(const TStrKdV& FldNmValPrV)
{
	const TStr ConceptUri = GetFldVal(FldNmValPrV, "conceptUri", "");
	if (! NewsBase->ConceptStore->IsRecNm(ConceptUri))
		return TJsonVal::NewObj("error", TStr("There is no valid conceptUri provided."));
	int Count = GetFldInt(FldNmValPrV, "count", 20);
	bool AddDetails = GetFldBool(FldNmValPrV, "addDetails", false);
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	EnumConceptInfoFlags ConceptFlags = NewsSrvUtil->GetConceptFlags(FldNmValPrV);

	uint64 ConceptId = NewsBase->ConceptStore->GetRecId(ConceptUri);
	PRecSet SimConceptRecSet; TStr Error;
	bool Ok = EventRegBase->GetSimilarConcepts(SimConceptRecSet, Error, ConceptId, Count);
	if (!Ok)
		return TJsonVal::NewObj("error", Error);

	PJsonVal RetJson = TJsonVal::NewObj("conceptInfo", EventRegSrvUtil->NewsSrvUtil->Concept(ConceptId, -1, ConceptLangV, ConceptFlags));
	TJsonValV SimConceptV;
	for (int N=0; N < SimConceptRecSet->GetRecs(); N++) {
		uint64 SimConceptId = SimConceptRecSet->GetRecId(N);
		int Fq = SimConceptRecSet->GetRecFq(N);
		PJsonVal ConceptJson = EventRegSrvUtil->NewsSrvUtil->Concept(SimConceptId, Fq, ConceptLangV, ConceptFlags);
		ConceptJson->AddToObj("similarity", SimConceptRecSet->GetRecFq(N));
		SimConceptV.Add(ConceptJson);
	}
	RetJson->AddToObj("similarConcepts", TJsonVal::NewArr(SimConceptV));
	return RetJson;
}

// return the graph of entities related to concept conceptUri
// compute by checking the latest 200 events and find co-occuring concepts
PJsonVal TEventRegSrv::GetConceptGraph(const TStrKdV& FldNmValPrV)
{
	const TStr ConceptUri = GetFldVal(FldNmValPrV, "conceptUri");
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "conceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "conceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV);

	const int ConceptCount = GetFldInt(FldNmValPrV, "conceptCount", 25);
	const int LinkCount = GetFldInt(FldNmValPrV, "linkCount", 50);

	if (!NewsBase->ConceptStore->IsRecNm(ConceptUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Unknown concept uri '' specified", ConceptUri.CStr()));

	PRecSet EventSet = NewsBase->ConceptStore->GetRec(ConceptUri).DoJoin(NewsBase->Base, "hasEvent");
	if (EventRegBase->GetSettings().EventSampleSize > 0)
		EventSet = EventSet->GetSampleRecSet(EventRegBase->GetSettings().EventSampleSize, false);
	EventSet->SortById(false);

	PJsonVal RetJson = EventRegSrvUtil->EventsConceptGraph(EventSet, ConceptTypes, ConceptLangV, ConceptCount, LinkCount);
	return RetJson;
}

// return a json with event uris that have changed since "lastUpdate"
PJsonVal TEventRegSrv::GetUpdatedEvents(const TStrKdV& FldNmValPrV)
{
	TStr DateStr = GetFldVal(FldNmValPrV, "lastUpdate", "");	// format: 1970-01-01 00:00:00
	int Count = GetFldInt(FldNmValPrV, "count", 1000);
	int Page = GetFldInt(FldNmValPrV, "page", 0);

	PJsonVal RetJson = TJsonVal::NewObj();
	TStrV SubEventUriV;
	if (DateStr != "") {
		TTm LastUpdateDate = TTm::GetTmFromWebLogDateTimeStr(DateStr, '-', ':', '.', ' ');

		TStrSet EventUriH;

		TTm Tm;
		PStoreIter Iter = StoryStore->GetIter();
		while (Iter->Next()) {
			StoryStore->GetFieldNmTm(Iter->GetRecId(), "LastUpdate", Tm);
			if (Tm >= LastUpdateDate) {
				uint64 EventRecId = StoryStore->GetFieldNmUInt64(Iter->GetRecId(), "hasEventId");
				if (EventRecId != TUInt64::Mx)
					EventUriH.AddKey(EventStore->GetRecNm(EventRecId));
			}
		}
		TStrV EventUriV; EventUriH.GetKeyV(EventUriV);
		EventUriV.Sort();
		EventUriV.GetSubValV(Page*Count, max(0, ((Page + 1)*Count) - 1), SubEventUriV);
		RetJson->AddToObj("totalCount", EventUriV.Len());
	}
	else {
		PStoreIter Iter = EventStore->GetIter();
		uint64 N = 0;
		int Start = Page * Count;
		int End = (Page + 1)*Count;
		while (Iter->Next()) {
			if (N >= End) break;
			if (N >= Start)
				SubEventUriV.Add(EventStore->GetRecNm(Iter->GetRecId()));
			N++;
		}
		RetJson->AddToObj("totalCount", TJsonVal::NewNum(EventStore->GetRecs()));
	}

	RetJson->AddToObj("changedEventUris", TJsonVal::NewArr(SubEventUriV));
	RetJson->AddToObj("page", Page);
	RetJson->AddToObj("count", Count);
	return RetJson;
}

PJsonVal TEventRegSrv::GetEventIdForStory(const TStrKdV& FldNmValPrV)
{
	TStr StoryUri = GetFldVal(FldNmValPrV, "storyUri", "");
	TStr StoryIdStr = GetFldVal(FldNmValPrV, "storyId", "");
	if (StoryUri == "" && StoryIdStr == "")
		return TJsonVal::NewObj("error", TStr("No storyUri or storyId were provided in the request"));

	uint64 StoryId = TUInt64::Mx;
	if (StoryUri != "" && NewsBase->StoryStore->IsRecNm(StoryUri))
		StoryId = NewsBase->StoryStore->GetRecId(StoryUri);
	else if (StoryIdStr.IsUInt64() && NewsBase->StoryStore->IsRecId(StoryIdStr.GetUInt64()))
		StoryId = StoryIdStr.GetUInt64();

	if (NewsBase->StoryStore->IsRecId(StoryId)) {
		uint64 EventId = EventRegBase->GetStoryEvent(StoryId);
		return TJsonVal::NewObj("eventUri", NewsBase->EventStore->GetRecNm(EventId));
	}
	else
		return TJsonVal::NewObj("error", TStr("No events were found for the provided storyUri/storyId value"));
}

// return a list of "count" events that are "trending"
// consider event as trending if it is from the last day and contains a lot of articles
PJsonVal TEventRegSrv::GetTrendingEvents(const TStrKdV& FldNmValPrV)
{
	TStrV LangV; GetFldValV(FldNmValPrV, "trendingEventsLang", LangV);
	const int Count = GetFldInt(FldNmValPrV, "trendingEventsCount", 5);
	const bool IncludeDetails = GetFldBool(FldNmValPrV, "trendingEventsIncludeDetails", true);
	const double MinCosSim = GetFldFlt(FldNmValPrV, "trendingEventsMinSim", 0.4);

	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "trendingEventsConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "similarStoriesConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV);
	EnumEventInfoFlags EventFlags = NewsSrvUtil->GetEventFlags(FldNmValPrV, "trendingEvents");

	const uint64 LastArtId = ArticleStore->GetLastRecId();

	// take last two days. use them to find stories from those days
	TTm Date; ArticleStore->GetFieldNmTm(LastArtId, "DateTime", Date);
	TStr EndDate = Date.GetWebLogDateStr();
	Date.AddDays(-1);
	PRecSet EventSet; TStr Error; bool Ok = NewsSearch->FindEventsForDates(Date.GetWebLogDateStr(), EndDate, EventSet, Error);
	if (!Ok)
		return TJsonVal::NewObj("error", Error);

	EventSet->SortByField(false, EventStore->GetFieldId("ArticleCount"));

	TStrSet LangH(LangV);
	PJsonVal RetJson = TJsonVal::NewObj();
	TJsonValV TrendingEventsV;
	int Added = 0;
	for (int N = 0; N < EventSet->GetRecs() && Added < Count; N++) {
		uint64 EventId = EventSet->GetRecId(N);
		// only return events that have articles consistently about the same event
		double Sim = EventRegBase->GetEventAvgCosToCentroid(EventId);
		if (Sim < MinCosSim)
			continue;
		// if lang filter then ignore events without appropriate language reporting
		if (LangH.Len() > 0) {
			TUInt64V StoryIdV; EventRegBase->GetEventStories(EventId, StoryIdV);
			bool FoundLang = false;
			for (int S = 0; S < StoryIdV.Len(); S++)
				FoundLang |= LangH.IsKey(EventRegBase->GetStoryLang(StoryIdV[S]));
			if (!FoundLang)
				continue;
		}
		PJsonVal TrendJson = EventRegSrvUtil->EventArticleTrend(FldNmValPrV, EventId, "trendingEvents");
		if (IncludeDetails) {
			PJsonVal EventDetails = NewsSrvUtil->Event(EventSet->GetRecId(N), ConceptLangV, ConceptTypes, EventFlags);
			TrendJson->AddToObj("eventDetails", EventDetails);
		}
		TrendingEventsV.Add(TrendJson);
		Added++;
	}
	PJsonVal TrendingEventsJson = TJsonVal::NewArr(TrendingEventsV);
	RetJson->AddToObj("trendingEvents", TrendingEventsJson);
	return RetJson;
}

// get document atlas built on the concepts
PJsonVal TEventRegSrv::GetConceptsDocAtlas(const TStrKdV& FldNmValPrV)
{
	// create a new sub bow - could be simpler but we need to replace story uris with titles
	//PBowDocBs Bow = TBowDocBs::New();
	//THash<TUInt64, TStr> StoryIdToTitleH;
	//int Docs = EventRegBase->ConceptBowBs->GetDocs();
	//for (int DId = 0; DId < Docs; DId++) {
	//	TIntFltPrV StoryIdWgtV;
	//	PBowSpV SpV = EventRegBase->ConceptBowBs->GetDocSpV(DId);
	//	// ignore concepts if they occur in less than X stories
	//	int Len = SpV->Len();
	//	if (SpV->Len() < 100)
	//		continue;
	//	for (int StoryN = 0; StoryN < SpV->GetWIds(); StoryN++) {
	//		int StoryWId = SpV->GetWId(StoryN);
	//		uint64 StoryId = EventRegBase->ConceptBowBs->GetWordStr(StoryWId).GetUInt64(TUInt64::Mx);
	//		AssertR(StoryStore->IsRecId(StoryId), "Story id is not a valid story");
	//		TStr StoryTitle;
	//		if (StoryIdToTitleH.IsKey(StoryId))
	//			StoryTitle = StoryIdToTitleH.GetDat(StoryId);
	//		else {
	//			StoryTitle = EventRegBase->GetStoryTitle(StoryId).GetSubStr(0, 100) + "...";
	//			StoryIdToTitleH.AddDat(StoryId, StoryTitle);
	//		}
	//		int WId = Bow->AddWordStr(StoryTitle);
	//		StoryIdWgtV.Add(TIntFltPr(WId, (float) SpV->GetWgt(StoryN)));
	//	}
	//	Bow->AddDoc(EventRegBase->ConceptBowBs->GetDocNm(DId), TStrV(), StoryIdWgtV);
	//}
	//TNotify::StdNotify->OnStatusFmt("Computed bow contains %d docs (of %d total)", Bow->GetDocs(), Docs);
	//// parameters
	//const int FontSize = 15;

	/*PVizMap Map = TVizMapFactory::NewVizMap(Bow);
	PVizMapFrame Frame = Map->GetVizMapFrame(0);
	PVizMapContext VizMapContext = TVizMapContext::New(Map);
	TWfBmpGks SmallGks(2000, 1400);
	VizMapContext->Paint(SmallGks.GetGks(), false, FontSize, 0, 1.0, -1, false, true, FontSize, false, -1, -1);
	TStr SmallFNm = TFile::GetUniqueFNm("./large-concepts.png");
	SmallGks.SavePng(SmallFNm);
*/
	return NULL;
}

PJsonVal TEventRegSrv::GetDocAtlasOnConcepts(const TStrKdV& FldNmValPrV, const PRecSet& EventRecSet)
{
	//return NULL;
	//PRecSet SampleRecSet = EventRecSet;
	//if (EventRegBase->GetSettings().EventSampleSize > 0)
	//	SampleRecSet = EventRecSet->GetSampleRecSet(EventRegBase->GetSettings().EventSampleSize, false);
	//// create a new sub bow - could be simpler but we need to replace concept uris with labels
	//PBowDocBs Bow = TBowDocBs::New();
	////TIntV DIdV;
	//for (int N = 0; N < SampleRecSet->GetRecs(); N++) {
	//	TUInt64 RecId = SampleRecSet->GetRecId(N);
	//	int DId = EventRegBase->EventBowBs->GetDId(RecId.GetStr());
	//	//DIdV.Add(DId);
	//	TIntFltPrV ConcIdWgtV;
	//	PBowSpV SpV = EventRegBase->EventBowBs->GetDocSpV(DId);
	//	for (int ConceptN = 0; ConceptN < SpV->GetWIds(); ConceptN++) {
	//		int ConceptWId = SpV->GetWId(ConceptN);
	//		TStr ConceptUri = EventRegBase->EventBowBs->GetWordStr(ConceptWId);
	//		TStr ConceptLabel = NewsBase->ConceptStore->GetFieldNmStr(NewsBase->ConceptStore->GetRecId(ConceptUri), "Label");
	//		int WId = Bow->AddWordStr(ConceptLabel);
	//		ConcIdWgtV.Add(TIntFltPr(WId, (float) SpV->GetWgt(ConceptN)));
	//	}
	//	Bow->AddDoc(RecId.GetStr(), TStrV(), ConcIdWgtV);
	//}
	//// parameters
	//const int FontSize = 15;

	/*PVizMap Map = TVizMapFactory::NewVizMap(Bow);
	PVizMapFrame Frame = Map->GetVizMapFrame(0);
	PVizMapContext VizMapContext = TVizMapContext::New(Map);
	TWfBmpGks SmallGks(1000, 800);
	VizMapContext->Paint(SmallGks.GetGks(), false, FontSize, 0, 1.0, -1, false, true, FontSize, false, -1, -1);
	TStr SmallFNm = TFile::GetUniqueFNm("./large-stories.png");
	SmallGks.SavePng(SmallFNm);
*/
	return TJsonVal::NewObj();
}
