/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "NewsSrv.h"
#include "../Utils/FileUtils.h"

///////////////////////////////
// TConceptAutoSuggestSrv - auto suggest for concepts
TConceptAutoSuggestSrv::TConceptAutoSuggestSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("suggestConcepts", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) 
{ 
	NewsFixDb = NewsBase->NewsFixDb;
}

TConceptAutoSuggestSrv::~TConceptAutoSuggestSrv()
{
	try {
		TFOut FOut(NewsBase->Base->GetFPath() + "ConceptAutoSuggestCache.dat");
		CachedConceptCount.Save(FOut);
		NextLangInd.Save(FOut);
		ConceptIdToFqH.Save(FOut);
		ConceptPrefixToIdWgtHH.Save(FOut);
	}
	catch (PExcept E) {
		NewsBase->ErrorLogFile->NotifyErr("TConceptAutoSuggestSrv::~TConceptAutoSuggestSrv. Error saving ConceptAutoSuggestCache.dat.", E);
	}
	catch (...) {
		NewsBase->ErrorLogFile->NotifyErr("TConceptAutoSuggestSrv::~TConceptAutoSuggestSrv. General error while saving ConceptAutoSuggestCache.dat.");
	}
}

// find the concepts that match the user provided prefix
TStr TConceptAutoSuggestSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv) 
{
	PJsonVal ResJson = NULL;
	try
	{
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "updatePrefixes") {
			UpdatePrefixes();
			return "{\"info\": \"Concepts updated.\"}";
		}

		TStrV SourceV; GetFldValV(FldNmValPrV, "source", SourceV);
		EnumConceptType ValidTypes = NewsConst.GetValidTypes(SourceV);

		ResJson = FindConcepts(FldNmValPrV, ValidTypes);

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TConceptAutoSuggestSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TConceptAutoSuggestSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TConceptAutoSuggestSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TConceptAutoSuggestSrv::ExecJSon. General exception.");
	}
	
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

void TConceptAutoSuggestSrv::GetConceptSuggestions(const TStr& Prefix, const TStr& Lang, TVec<TUIntIntKd>& SuggV) const
{
	TStr PrefixLc = Prefix.GetLc();
	TStr PrefixKey = PrefixLc.GetSubStr(0, 1);
	SuggV.Clr();
	if (! ConceptPrefixToIdWgtHH.IsKey(Lang))
		return;
	if (! ConceptPrefixToIdWgtHH.GetDat(Lang).IsKey(PrefixKey))
		return;
	SuggV = ConceptPrefixToIdWgtHH.GetDat(Lang).GetDat(PrefixKey);
}

void TConceptAutoSuggestSrv::LoadOrUpdatePrefixes()
{
	if (TFile::Exists(NewsBase->Base->GetFPath() + "ConceptAutoSuggestCache.dat")) {
		try {
			TFIn FIn(NewsBase->Base->GetFPath() + "ConceptAutoSuggestCache.dat");
			TInt ConceptCount;
			ConceptCount.Load(FIn);
			if (ConceptCount + 1000 > NewsBase->ConceptStore->GetRecs()) {
				TNotify::StdNotify->OnStatus("Concept suggestions are being loaded from the cache...");
				CachedConceptCount = ConceptCount;
				NextLangInd.Load(FIn);
				ConceptIdToFqH.Load(FIn);
				ConceptPrefixToIdWgtHH.Load(FIn);
				ConceptIdToFqH.GetKeyDatKdV(ConceptIdFqV);
				return;
			}
			else
				TNotify::StdNotify->OnStatus("File with concept suggestions is out of date.");
		}
		catch (PExcept E) {
			NewsBase->ErrorLogFile->NotifyErr("TConceptAutoSuggestSrv::LoadOrUpdatePrefixes. Error loading ConceptAutoSuggestCache.dat.", E);
		}
		catch (...) {
			NewsBase->ErrorLogFile->NotifyErr("TConceptAutoSuggestSrv::LoadOrUpdatePrefixes. General error while loading ConceptAutoSuggestCache.dat.");
		}
	}
	else
		TNotify::StdNotify->OnStatus("File with concept suggestions does not exist.");
		
	UpdateDataCounts();
	TTmStopWatch StopWatch(true);
	TNotify::StdNotify->OnStatus("Updating concept suggestions for all languages...");
	for (int N = 0; N < NewsConst.LangAllV.Len(); N++)
		UpdatePrefixes(false);
	const double Tm = StopWatch.GetSec();
	TNotify::StdNotify->OnStatusFmt("Update took: %.2fsec", Tm);
}

void TConceptAutoSuggestSrv::UpdateDataCounts()
{
	TTmStopWatch StopWatch(true);
	PNewsCounts NewsCounts = NewsBase->NewsCounts;
	CachedConceptCount = (int) NewsBase->ConceptStore->GetRecs();
	ConceptIdToFqH.Clr();
	PStoreIter Iter = NewsBase->ConceptStore->GetIter();
	while (Iter->Next()) {
		uint64 ConceptId = Iter->GetRecId();

		const TStr ConceptType = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
		const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(ConceptId);

		// ignored concepts should not be suggested (this will also include the rewritten concepts)
		if (NewsFixDb->HasConceptIgnore(ConceptUri))
			continue;

		int ConceptFq = 0;
		// concept folder popularity we have to check differently
		if (ConceptType == NewsConst.TypeConceptFolder) {
			if (NewsBase->ConceptFolderStore->IsRecNm(ConceptUri)) {
				uint64 ConceptFolderId = NewsBase->ConceptFolderStore->GetRecId(ConceptUri);
				ConceptFq = NewsBase->ConceptFolderStore->GetFieldNmInt(ConceptFolderId, "ArticlesInFolder");
			}
		}
		else
			ConceptFq = (int) NewsCounts->GetConceptFq(ConceptId);
		ConceptIdToFqH.AddDat((uint) ConceptId, ConceptFq);
	}
	const double ObtainingFreqsTm = StopWatch.GetSec();
	StopWatch.Reset(true);

	TNotify::StdNotify->OnStatusFmt("Concept update data counts took: %.2fsec", ObtainingFreqsTm);

	// sort from most to least frequently used concepts
	ConceptIdToFqH.SortByDat(false);
	ConceptIdToFqH.GetKeyDatKdV(ConceptIdFqV);
}

// update prefixes for the next language
void TConceptAutoSuggestSrv::UpdatePrefixes(const bool& Report)
{
	if (NextLangInd >= NewsConst.LangAllV.Len()) {
		UpdateDataCounts();
		NextLangInd = (NextLangInd + 1) % (NewsConst.LangAllV.Len() + 1);
		return;
	}

	TTmStopWatch StopWatch(true);

	const TStr Lang = NewsConst.LangAllV[NextLangInd];
	NextLangInd = (NextLangInd + 1) % (NewsConst.LangAllV.Len() + 1);
	ConceptPrefixToIdWgtHH.AddDat(Lang, PrefToSugg());		// add empty set - needed for next updates for the same language (otherwise we would just increase counts)
		
	for (int KeyId = ConceptIdToFqH.FFirstKeyId(); ConceptIdToFqH.FNextKeyId(KeyId);) {
		const uint ConceptRecId = ConceptIdToFqH.GetKey(KeyId);
		const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(ConceptRecId);
		const int Fq = (int) ConceptIdToFqH[KeyId];

		TStr Name = NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, Lang);
		Name = Name.GetLc();
		Name.ChangeCh('.', ' ');
		Name.ChangeCh(',', ' ');

		PrefToSugg& ConceptPrefixToIdWgtH = ConceptPrefixToIdWgtHH.AddDat(Lang);
				
		TStrV WordPartV; Name.SplitOnAllCh(' ', WordPartV);
		for (int Parts = 0; Parts < WordPartV.Len(); Parts++) {
			const TStr Sub1 = WordPartV[Parts].GetSubStr(0, 0);
			const TStr Sub2 = WordPartV[Parts].GetSubStr(0, 1);
			ConceptPrefixToIdWgtH.AddDat(Sub1).Add(TUIntIntKd(ConceptRecId, Fq));
			if (Sub2.Len() == 2)
				ConceptPrefixToIdWgtH.AddDat(Sub2).Add(TUIntIntKd(ConceptRecId, Fq));
		}
	}
				
	const double PrefixBuildingTm = StopWatch.GetSec();

	if (Report)
		TNotify::StdNotify->OnStatusFmt("Update entity prefixes for lang %s took: %.2fsec", Lang.CStr(), PrefixBuildingTm);
}

PJsonVal TConceptAutoSuggestSrv::FindConcepts(const TStrKdV& FldNmValPrV, EnumConceptType ValidTypes) 
{
	TTmStopWatch StopWatch(true);

	// construct query
	const TStr PrefixStr = GetFldVal(FldNmValPrV, "prefix");
	const TStr PrefixStrLc = PrefixStr.GetLc();
	
	const TStr Lang = GetFldVal(FldNmValPrV, "lang", "").GetLc();
	const TInt Page = GetFldInt(FldNmValPrV, "page", 0);
	const TInt Count = GetFldInt(FldNmValPrV, "count", 20);
	TStrV LabelLangV; GetFldValV(FldNmValPrV, "labelLang", LabelLangV);
		
	PJsonVal SuggArr = TJsonVal::NewArr();
	
	TUInt64Set SeenIdH;	// see which concepts we already added so that we don't add them twice
	TUIntIntKdV SuggestionsV;
	
	int AddedCount = 0;
	const int ToSkip = Page * Count;
	int Skipped = 0;
	// if the user didn't specify any prefix then return a list of most frequently used concepts
	// we don't just request (Page+1) * Count elements since we might want to ignore entities/keywords and then we would get too few results
	if (PrefixStr.Len() > 0)
		GetConceptSuggestions(PrefixStr, Lang, SuggestionsV);
	else
		SuggestionsV = ConceptIdFqV;

	bool NonEmptyPrefix = PrefixStr.Len() > 0;
	int SuggestionsCount = SuggestionsV.Len();
	for (int N = 0; N < SuggestionsCount; N++) {
		const uint64 ConceptRecId = (uint64) SuggestionsV[N].Key;
		// for the same prefix, the same concept can match multiple times so we need to check to add it only once
		if (SeenIdH.IsKey(ConceptRecId))
			continue;
		SeenIdH.AddKey(ConceptRecId);
		if (!NewsBase->ConceptStore->IsRecId(ConceptRecId))
			continue;
		const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(ConceptRecId);
		TStr Label = NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, Lang);
		if (NonEmptyPrefix) {
			// does the label match the given prefix
			TStr LabelLc = Label.GetLc();
			if (!LabelLc.IsPrefix(PrefixStrLc) && LabelLc.SearchStr(" " + PrefixStrLc, 0) == -1)
				continue;
		}
		// if we want only keywords and the word is not a keyword then ignore it
		TStr Type = NewsBase->ConceptStore->GetFieldNmStr(ConceptRecId, "Type");
	
		if (NewsConst.IsValidType(Type, ValidTypes)) {

			// if the concept is a concept folder that is private (invisible) then don't include it
			if (NewsConst.ConceptFolderSet.IsKey(Type)) {
				TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(ConceptRecId);
				if (NewsBase->ConceptFolderStore->IsRecNm(ConceptUri)) {
					const uint64 ConceptFolderId = NewsBase->ConceptFolderStore->GetRecId(ConceptUri);
					if (NewsBase->ConceptFolderStore->GetFieldNmInt(ConceptFolderId, "Visibility") == (int) PrivateFolder)
						continue;
				}
			}
				
			// if we are here we have a valid concept that should be added to the suggestions list
			// now only check if we want some Page > 0. if yes, then skip the necessary number of suggestions
			if (Skipped < ToSkip) { Skipped++; continue; }

			const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(ConceptRecId);
			PJsonVal Sugg = TJsonVal::NewObj("uri", ConceptUri);
			Sugg->AddToObj("fq", SuggestionsV[N].Dat);
			Sugg->AddToObj("id", TJsonVal::NewStr(TUInt64::GetStr(ConceptRecId)));
			Sugg->AddToObj("type", Type);
			Sugg->AddToObj("label", NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, DefLang));
			// add also extra labels for languages specified using labelLang (used for things like concept admin
			for (int N = 0; N < LabelLangV.Len(); N++) {
				Sugg->AddToObj(NewsConst.GetJsonNmForLang(LabelLangV[N]), NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, LabelLangV[N]));
			}
			SuggArr->AddToArr(Sugg);

			AddedCount++;
			if (AddedCount >= Count)
				break;
		}
	}
	TNotify::StdNotify->OnStatusFmt("Finding suggestions took: %.2fms", StopWatch.GetMSec());
	return SuggArr;
}


///////////////////////////////
// TConceptAutoSuggestSrv - auto suggest for concepts
TConceptClassAutoSuggestSrv::TConceptClassAutoSuggestSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("suggestConceptClasses", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) 
{ 
	UpdatePrefixes();
}

TConceptClassAutoSuggestSrv::~TConceptClassAutoSuggestSrv()
{
	
}

// find the concepts that match the user provided prefix
TStr TConceptClassAutoSuggestSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "updatePrefixes") {
			UpdatePrefixes();
			return "{\"info\": \"Concept classes updated.\"}";
		}

		TStrV SourceV; GetFldValV(FldNmValPrV, "source", SourceV);
		EnumConceptClassType Type = cctNone;
		if (SourceV.Len() == 0)
			Type = cctAll;
		else {
			if (SourceV.IsIn("dbpedia")) Type = EnumConceptClassType(Type | cctDbpedia);
			if (SourceV.IsIn("custom")) Type = EnumConceptClassType(Type | cctCustom);
		}
		
		ResJson = FindClasses(FldNmValPrV, Type);

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TConceptClassAutoSuggestSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassAutoSuggestSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TConceptClassAutoSuggestSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TConceptClassAutoSuggestSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

void TConceptClassAutoSuggestSrv::GetClassSuggestions(const TStr& Prefix, const TStr& Lang, TVec<TUInt64IntKd>& SuggV) const
{
	TStr PrefixLc = Prefix.GetLc();
	TStr PrefixKey = PrefixLc.GetSubStr(0, 1);
	SuggV.Clr();
	if (!ClassPrefixToIdWgtHH.IsKey(Lang))
		return;
	if (!ClassPrefixToIdWgtHH.GetDat(Lang).IsKey(PrefixKey))
		return;
	SuggV = ClassPrefixToIdWgtHH.GetDat(Lang).GetDat(PrefixKey);
}

void TConceptClassAutoSuggestSrv::UpdatePrefixes()
{
	THash<TUInt64, TInt> ClassIdToFqH;

	PStoreIter Iter = NewsBase->ConceptClassStore->GetIter();
	while (Iter->Next()) {
		uint64 ClassId = Iter->GetRecId();

		int ClassScore = NewsBase->ConceptClassStore->GetRec(ClassId).DoJoin(NewsBase->Base, "hasConcept")->GetRecs();
		ClassIdToFqH.AddDat(ClassId, ClassScore);
	}

	// sort from most to least frequently used concepts
	ClassIdToFqH.SortByDat(false);

	for (int KeyId = ClassIdToFqH.FFirstKeyId(); ClassIdToFqH.FNextKeyId(KeyId);) {
		const TUInt64 ConceptRecId = ClassIdToFqH.GetKey(KeyId);
		const TStr ConceptUri = NewsBase->ConceptClassStore->GetRecNm(ConceptRecId);
		const int Fq = (int) ClassIdToFqH[KeyId];

		for (int N = 0; N < NewsConst.LangAllV.Len(); N++) {
			const TStr Lang = NewsConst.LangAllV[N];
			TStr Name = NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, Lang);
			Name = Name.GetLc();
			Name.ChangeCh('.', ' ');
			Name.ChangeCh(',', ' ');

			PrefToSugg ConceptPrefixToIdWgtH = ClassPrefixToIdWgtHH.AddDat(Lang);
			TStrV WordParts; Name.SplitOnAllCh(' ', WordParts);
			for (int Parts = 0; Parts < WordParts.Len(); Parts++) {
				const TStr Sub1 = WordParts[Parts].GetSubStr(0, 0);
				const TStr Sub2 = WordParts[Parts].GetSubStr(0, 1);
				if (!ConceptPrefixToIdWgtH.IsKey(Sub1))
					ConceptPrefixToIdWgtH.AddDat(Sub1, TVec<TUInt64IntKd>());
				ConceptPrefixToIdWgtH.GetDat(Sub1).Add(TUInt64IntKd(ConceptRecId, Fq));
				if (Sub2.Len() == 2) {
					if (!ConceptPrefixToIdWgtH.IsKey(Sub2))
						ConceptPrefixToIdWgtH.AddDat(Sub2, TVec<TUInt64IntKd>());
					ConceptPrefixToIdWgtH.GetDat(Sub2).Add(TUInt64IntKd(ConceptRecId, Fq));
				}
			}
		}
	}
}

PJsonVal TConceptClassAutoSuggestSrv::FindClasses(const TStrKdV& FldNmValPrV, const EnumConceptClassType& ValidTypes)
{
	// construct query
	const TStr PrefixStr = GetFldVal(FldNmValPrV, "prefix");
	const TStr PrefixStrLc = PrefixStr.GetLc();

	const TStr Lang = GetFldVal(FldNmValPrV, "lang", "").GetLc();
	const TInt Page = GetFldInt(FldNmValPrV, "page", 0);
	const TInt Count = GetFldInt(FldNmValPrV, "count", 20);
	TStrV ConceptLabelV; GetFldValV(FldNmValPrV, "labelLang", ConceptLabelV);
	for (int N = ConceptLabelV.Len() - 1; N >= 0; N--) {
		if (!NewsBase->ConceptClassStore->IsFieldNm(ConceptLabelV[N]))
			ConceptLabelV.Del(N);
	}

	PJsonVal SuggArr = TJsonVal::NewArr();

	const TStr FieldName = "Label" + Lang.GetCap();
	if (!NewsBase->ConceptClassStore->IsFieldNm(FieldName))
		return SuggArr;
	if (PrefixStr == "")
		return SuggArr;

	TUInt64Set AddedIdH;	// see which concepts we already added so that we don't add them twice
	TUInt64IntKdV SuggestionsV;

	int AddedCount = 0;
	const int ToSkip = Page * Count;
	int Skipped = 0;
	// if the user didn't specify any prefix then return a list of most frequently used concepts
	// we don't just request (Page+1) * Count elements since we might want to ignore entities/keywords and then we would get too few results
	GetClassSuggestions(PrefixStr, Lang, SuggestionsV);

	int SuggestionsCount = SuggestionsV.Len();
	for (int N = 0; N < SuggestionsCount; N++) {
		TUInt64 ClassId = SuggestionsV[N].Key;
		// for the same prefix, the same concept can match multiple times so we need to check to add it only once
		if (AddedIdH.IsKey(ClassId))
			continue;
		AddedIdH.AddKey(ClassId);
		if (!NewsBase->ConceptClassStore->IsRecId(ClassId))
			continue;
		// does the label match the given prefix
		TStr Label = NewsBase->ConceptClassStore->GetFieldNmStr(ClassId, FieldName);
		TStr LabelLc = Label.GetLc();
		if (!LabelLc.IsPrefix(PrefixStrLc) && LabelLc.SearchStr(" " + PrefixStrLc, 0) == -1)
			continue;

		// if we want only keywords and the word is not a keyword then ignore it
		TStr Type = NewsBase->ConceptClassStore->GetFieldNmStr(ClassId, "Type");
				
		if ((ValidTypes & cctDbpedia && NewsConst.IsClassDbpedia(Type)) ||
			(ValidTypes & cctCustom && NewsConst.IsClassCustom(Type))) {

			// if we are here we have a valid concept that should be added to the suggestions list
			// now only check if we want some Page > 0. if yes, then skip the necessary number of suggestions
			if (Skipped < ToSkip) { Skipped++; continue; }

			PJsonVal Sugg = TJsonVal::NewObj("uri", NewsBase->ConceptClassStore->GetRecNm(ClassId));
			Sugg->AddToObj("fq", SuggestionsV[N].Dat);
			Sugg->AddToObj("id", TJsonVal::NewStr(TUInt64::GetStr(ClassId)));
			Sugg->AddToObj("type", Type);
			Sugg->AddToObj("label", Label);
			// add also extra labels for languages specified using labelLang (used for things like concept admin
			if (ConceptLabelV.Len() > 0) {
				for (int N = 0; N < ConceptLabelV.Len(); N++)
					Sugg->AddToObj(ConceptLabelV[N], NewsBase->ConceptClassStore->GetFieldNmStr(ClassId, ConceptLabelV[N]));
			}
			SuggArr->AddToArr(Sugg);

			AddedCount++;
			if (AddedCount >= Count)
				break;
		}
	}
	return SuggArr;
}

///////////////////////////////
// TSourceAutoSuggestSrv - auto suggest for news source
TSourceAutoSuggestSrv::TSourceAutoSuggestSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("suggestSources", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) {  }

// provide news sources that match the prefix
TStr TSourceAutoSuggestSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv) 
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "updatePrefixes") {
			UpdatePrefixes();
			return "{\"info\": \"Sources updated.\"}";
		}

		const TStr PrefixStr = GetFldVal(FldNmValPrV, "prefix");
		const TStr PrefixStrLc = PrefixStr.GetLc();
		
		const TInt Page = GetFldInt(FldNmValPrV, "page", 0);
		const TInt Count = GetFldInt(FldNmValPrV, "count", 20); 
		
		TVec<TUInt64> PrefixMatchesV;
		if (PrefixStrLc.Len() > 0)
			GetSourceSuggestions(PrefixStr, PrefixMatchesV);
		// on empty prefix just return the most popular sources
		else
			PrefixMatchesV = SourceIdV;
		
		// if we just want top sources then don't even compare the prefix
		bool NonEmptyPrefix = PrefixStr.Len() > 0;
		const int ToSkip = Page * Count;
		
		TUInt64Set SeenSourcesH;
		TUInt64V SourceIdV;
		int AddedCount = 0;
		int Skipped = 0;
		for (int N = 0; N < PrefixMatchesV.Len() && AddedCount < Count; N++) {
			TUInt64 SourceRecId = PrefixMatchesV[N];
			if (NonEmptyPrefix) {
				TStr SourceLabel = NewsBase->SourceStore->GetFieldNmStr(SourceRecId, "Title");
				// some sources have empty title. in this case we have to improvise using the url
				if (SourceLabel == "") {
					SourceLabel = NewsBase->SourceStore->GetRecNm(SourceRecId);
					SourceLabel = GetCleanSourceNameFromUri(SourceLabel);
				}
				TStr SourceLabelLc = SourceLabel.ToLc();
				if (SourceLabelLc.SearchStr(PrefixStrLc) == -1)
					continue;
			}
			
			if (SeenSourcesH.IsKey(SourceRecId))
				continue;
			SeenSourcesH.AddKey(SourceRecId);
			
			// if we are here we have a valid source that should be added to the suggestions list
			// now only check if we want some Page > 0. if yes, then skip the necessary number of suggestions
			if (Skipped < ToSkip) { Skipped++; continue; }

			SourceIdV.Add(SourceRecId);
			AddedCount++;
		}

		PRecSet SourcesRecSet = TRecSet::New(NewsBase->SourceStore, SourceIdV);
		const int Flags = pifIncludeTitle | pifIncludeDescription | pifIncludeTags | pifIncludeLocation | pifIncludeImportance | pifIncludeDetails | pifIncludeArticleCount;
		ResJson = NewsSrvUtil->Sources(SourcesRecSet, (EnumPublisherInfoFlags) Flags);
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TSourceAutoSuggestSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TSourceAutoSuggestSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TSourceAutoSuggestSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TSourceAutoSuggestSrv::ExecJSon. General exception.");
	}
	
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

void TSourceAutoSuggestSrv::GetSourceSuggestions(TStr Prefix, TVec<TUInt64>& SuggV)
{
	const TStr PrefixLc = Prefix.GetLc();
	const TStr PrefixKey = PrefixLc.GetSubStr(0, 1);
	if (SourcePrefixToFqWIdV.IsKey(PrefixKey))
		SuggV = SourcePrefixToFqWIdV.GetDat(PrefixKey);
	else
		SuggV.Clr();
}

void TSourceAutoSuggestSrv::UpdatePrefixes()
{
	try {
		TTmStopWatch StopWatch(true);
		SourcePrefixToFqWIdV.Clr();

		THash<TUInt64, TUInt64> SourceIdToFqH;

		PStoreIter Iter = NewsBase->SourceStore->GetIter();
		PNewsCounts NewsCounts = NewsBase->NewsCounts;
		while (Iter->Next()) {
			uint64 SourceId = Iter->GetRecId();
			uint64 SourceFq = NewsCounts->GetSourceFq(SourceId);
			SourceIdToFqH.AddDat(SourceId, SourceFq);
		}
		// sort from most to least frequent source
		SourceIdToFqH.SortByDat(false);
		
		// ignore common uri parts
		TStrV IgnoreV; TStr IgnoreText("www|http|biz|com|org|si|info|net|edu|biz|cat|gov|uk|co"); IgnoreText.SplitOnAllCh('|', IgnoreV);
		TStrSet IgnoreH(IgnoreV);
		
		SourceIdV.Clr();
		SourceIdToFqH.GetKeyV(SourceIdV);

		for (int KeyId = SourceIdToFqH.FFirstKeyId(); SourceIdToFqH.FNextKeyId(KeyId);) {
			const uint64 SourceRecId = SourceIdToFqH.GetKey(KeyId);
			// use the title and the url as the searchable text
			const TStr Title = NewsBase->SourceStore->GetFieldNmStr(SourceRecId, "Title");
			const TStr Uri = NewsBase->SourceStore->GetRecNm(SourceRecId);
			TStr Text = Title + " " + Uri;
			Text.ChangeChAll('.', ' ');
			Text.ToLc();
						
			TStrV WordParts; Text.SplitOnAllCh(' ', WordParts);
			for (int PartN = 0; PartN < WordParts.Len(); PartN++) {
				if (IgnoreH.IsKey(WordParts[PartN]))
					continue;
				const TStr Sub1 = WordParts[PartN].GetSubStr(0, 0);
				const TStr Sub2 = WordParts[PartN].GetSubStr(0, 1);
				SourcePrefixToFqWIdV.AddDat(Sub1).Add(SourceRecId);
				if (Sub2.Len() == 2)
					SourcePrefixToFqWIdV.AddDat(Sub2).Add(SourceRecId);
			}
		}
		TNotify::StdNotify->OnStatusFmt("Update source prefixes took: %.2fms", StopWatch.GetMSec());
	}
	catch (PExcept E) {
		TNotify::StdNotify->OnStatus("TSourceAutoSuggestSrv::UpdatePrefixes. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TSourceAutoSuggestSrv::UpdatePrefixes.", E);
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("TSourceAutoSuggestSrv::UpdatePrefixes. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TSourceAutoSuggestSrv::UpdatePrefixes. General exception.");
	}
}

TStr TSourceAutoSuggestSrv::GetCleanSourceNameFromUri(const TStr& Source)
{
	TStr SourceLabel = TStr(Source);
	int Start = SourceLabel.SearchStr("//");
	int End = SourceLabel.SearchStr("/", Start+2);
	if (End == -1)
		End = SourceLabel.Len();
	SourceLabel = SourceLabel.GetSubStr(max(0, Start), End);
	if (SourceLabel.IsSuffix(".com"))
		SourceLabel.DelSubStr(SourceLabel.Len()-4, SourceLabel.Len());
	SourceLabel.ChangeChAll('.', ' ');
	SourceLabel.ChangeChAll('-', ' ');
	SourceLabel.DelStr("www");
	return SourceLabel;
}

///////////////////////////////
// TSourceLocationAutoSuggestSrv - auto suggest for news source locations

TSourceLocationAutoSuggestSrv::TSourceLocationAutoSuggestSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("suggestSourceLocations", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil)
{
	IndexVoc = NewsBase->Base->GetIndexVoc();
}

// find locations of news sources that match the given prefix
// locations can be countries and cities
TStr TSourceLocationAutoSuggestSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr PrefixStr = GetFldVal(FldNmValPrV, "prefix");
		TStr PrefixStrLc = PrefixStr.GetLc();
		if (PrefixStrLc.Empty()) {
			return TJsonVal::GetStrFromVal(TJsonVal::NewArr());
		}
		TInt Count = GetFldInt(FldNmValPrV, "count", 10);

		// CityCountry and Country keys use the same vocabulary therefore they have the same keyId
		// by iterating one we get all countries and cities
		const int CityCountryKeyId = IndexVoc->GetKeyId(NewsBase->SourceStoreId, "CityCountry");
		THash<TStr, TUInt64> CitiesCountriesFqH;
		uint64 Words = IndexVoc->GetWords(CityCountryKeyId);
		for (uint64 WordId = 0; WordId < Words; WordId++) {
			TStr WordStr = IndexVoc->GetWordStr(CityCountryKeyId, WordId);
			TStr WordStrLc = WordStr.GetLc();
			if (WordStrLc.IsPrefix(PrefixStrLc))
				CitiesCountriesFqH.AddDat(WordStr, IndexVoc->GetWordFq(CityCountryKeyId, WordId));  // use this frequency to select the most frequent concepts and then only on the most frequent ones do the queries	
		}
		CitiesCountriesFqH.SortByDat(false);

		// find the most frequently used words and add only a subset of them	
		PJsonVal ResJson = TJsonVal::NewArr();
		int CityCountryId = CitiesCountriesFqH.FFirstKeyId();
		int AddedCount = 0;
		while (CitiesCountriesFqH.FNextKeyId(CityCountryId) && AddedCount < 20) {
			TStr Word; TUInt64 Fq;
			CitiesCountriesFqH.GetKeyDat(CityCountryId, Word, Fq);
			ResJson->AddToArr(TJsonVal::NewStr(Word));
		}
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TSourceLocationAutoSuggestSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TSourceLocationAutoSuggestSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TSourceLocationAutoSuggestSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TSourceLocationAutoSuggestSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

///////////////////////////////////////////////
// "print" requests
TPrintSrv::TPrintSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("print", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) { }


TStr TPrintSrv::GetPathName(const TStr& FName)
{
	TTm Tm = TTm::GetCurLocTm();
	return "./Print/" + Tm.GetYMDDashStr() + " " + FName;
}

TStr TPrintSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TDir::GenDir("./Print");
		bool PrintJoins = GetFldBool(FldNmValPrV, "printJoins", false);
		TStrV DataV; GetFldValV(FldNmValPrV, "data", DataV);
		// news data
		if (DataV.IsIn("articleStore") && !NewsBase->ArticleStore.Empty())
			NewsBase->ArticleStore->PrintAll(NewsBase->Base, GetPathName("ArticlesArticleStore.txt"), PrintJoins);
		if (DataV.IsIn("sourceStore") && !NewsBase->SourceStore.Empty())
			NewsBase->SourceStore->PrintAll(NewsBase->Base, GetPathName("ArticlesSourceStore.txt"), PrintJoins);
		if (DataV.IsIn("conceptStore") && !NewsBase->ConceptStore.Empty())
			NewsBase->ConceptStore->PrintAll(NewsBase->Base, GetPathName("ArticlesConceptStore.txt"), PrintJoins);
		if (DataV.IsIn("conceptClassStore") && !NewsBase->ConceptClassStore.Empty())
			NewsBase->ConceptClassStore->PrintAll(NewsBase->Base, GetPathName("ArticlesConceptClassStore.txt"), PrintJoins);
		if (DataV.IsIn("conceptFolderStore") && !NewsBase->ConceptFolderStore.Empty())
			NewsBase->ConceptFolderStore->PrintAll(NewsBase->Base, GetPathName("ArticlesConceptFolderStore.txt"), PrintJoins);
		if (DataV.IsIn("userStore") && !NewsBase->UserStore.Empty())
			NewsBase->UserStore->PrintAll(NewsBase->Base, GetPathName("ArticlesUserStore.txt"), PrintJoins);
		if (DataV.IsIn("imageStore") && !NewsBase->ImageStore.Empty())
			NewsBase->ImageStore->PrintAll(NewsBase->Base, GetPathName("ArticlesImageStore.txt"), PrintJoins);
		if (DataV.IsIn("dateStore") && !NewsBase->DateStore.Empty())
			NewsBase->DateStore->PrintAll(NewsBase->Base, GetPathName("ArticlesDateStore.txt"), PrintJoins);
		if (DataV.IsIn("categoryStore") && !NewsBase->CategoryStore.Empty())
			NewsBase->CategoryStore->PrintAll(NewsBase->Base, GetPathName("ArticlesCategoryStore.txt"), PrintJoins);
		if (DataV.IsIn("storyStore") && !NewsBase->StoryStore.Empty())
			NewsBase->StoryStore->PrintAll(NewsBase->Base, GetPathName("ArticlesStoryStore.txt"), PrintJoins);
		if (DataV.IsIn("eventStore") && !NewsBase->EventStore.Empty())
			NewsBase->EventStore->PrintAll(NewsBase->Base, GetPathName("ArticlesEventStore.txt"), PrintJoins);

		// geo data
		if (DataV.IsIn("countryStore") && !NewsBase->GeoBase->CountryStore.Empty())
			NewsBase->GeoBase->CountryStore->PrintAll(NewsBase->Base, GetPathName("ArticlesCountryStore.txt"), PrintJoins);
		if (DataV.IsIn("cityStore") && !NewsBase->GeoBase->CityStore.Empty())
			NewsBase->GeoBase->CityStore->PrintAll(NewsBase->Base, GetPathName("ArticlesCityStore.txt"), PrintJoins);
		if (DataV.IsIn("languageStore") && !NewsBase->GeoBase->LanguageStore.Empty())
			NewsBase->GeoBase->LanguageStore->PrintAll(NewsBase->Base, GetPathName("ArticlesLanguageStore.txt"), PrintJoins);

		// index
		if (DataV.IsIn("index"))
			NewsBase->Base->PrintIndex(GetPathName("ArticlesIndex.txt"), true);
		if (DataV.IsIn("indexVoc"))
			NewsBase->Base->PrintIndexVoc(GetPathName("ArticlesIndexVoc.txt"));

		// top concepts
		if (DataV.IsIn("topConcepts"))
			PrintTopConcepts(FldNmValPrV);

		if (DataV.Len() == 0)
			ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("No valid print requests were received. Use data=...Store to specify which stores to print."));
		else
			ResJson = TJsonVal::NewObj("info", TJsonVal::NewStr("Necessary data was saved to disk."));
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TPrintSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TPrintSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TPrintSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TPrintSrv::ExecJSon. General exception.");
	}

	// done
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

void TPrintSrv::PrintTopConcepts(const TStrKdV& FldNmValPrV)
{
	TStrV TypeV; GetFldValV(FldNmValPrV, "conceptType", TypeV);
	EnumConceptType ValidTypes = NewsConst.GetValidTypes(TypeV);
	const int Count = GetFldInt(FldNmValPrV, "count", 10000);

	THash<TUInt64, TInt> ConceptToFqH;
	PStoreIter Iter = NewsBase->ConceptStore->GetIter();
	while (Iter->Next()) {
		uint64 ConceptId = Iter->GetRecId();
		ConceptToFqH.AddDat(ConceptId, (int) NewsBase->NewsCounts->GetConceptFq(ConceptId));
	}
	ConceptToFqH.SortByDat(false);
	int Added = 0;
	TStr FName = GetPathName("TopConcepts.txt");
	TFOut FOut(FName);
	for (int KeyId = ConceptToFqH.FFirstKeyId(); ConceptToFqH.FNextKeyId(KeyId) && Added <  Count;) {
		const uint64 ConceptId = ConceptToFqH.GetKey(KeyId);
		TStr Type = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
		if (NewsConst.IsValidType(Type, ValidTypes)) {
			FOut.PutStrFmtLn("%s\t%s\t%d", NewsBase->ConceptStore->GetRecNm(ConceptId).CStr(), Type.CStr(), ConceptToFqH[KeyId]);
			Added++;
		}
	}
}

///////////////////////////////////////////////
// "print" requests
TOverviewSrv::TOverviewSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("overview", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) { }


TStr TOverviewSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "getRecentStats")
			ResJson = GetRecentStats(FldNmValPrV);
		else if (Action == "getArticleActivity")
			ResJson = GetArticleActivity(FldNmValPrV);
		else if (Action == "getEventActivity")
			ResJson = GetEventActivity(FldNmValPrV);
		else if (Action == "getAllActivity")
			ResJson = GetAllActivity(FldNmValPrV);
		else {
			ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("No valid actions were provided. Valid values are: getRecentStats, getArticleActivity, getEventActivity."));
		}

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TOverviewSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TOverviewSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TOverviewSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TOverviewSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

PJsonVal TOverviewSrv::GetRecentStats(const TStrKdV& FldNmValPrV)
{
	PJsonVal RetJson = TJsonVal::NewObj();
	RetJson->AddToObj("totalArticleCount", TJsonVal::NewNum((int) NewsBase->ArticleStore->GetRecs()));
	RetJson->AddToObj("totalConceptCount", TJsonVal::NewNum((int) NewsBase->ConceptStore->GetRecs()));
	RetJson->AddToObj("totalSourceCount", TJsonVal::NewNum((int) NewsBase->SourceStore->GetRecs()));
	RetJson->AddToObj("totalEventCount", TJsonVal::NewNum((int) NewsBase->EventStore->GetRecs()));
	RetJson->AddToObj("totalCategoryCount", TJsonVal::NewNum((int) NewsBase->CategoryStore->GetRecs()));
		
	const uint64 LastArtId = NewsBase->ArticleStore->GetLastRecId();
	if (NewsBase->ArticleStore->IsRecId(LastArtId)) {
		TJsonValV DailyArticleV;
		TJsonValV DailyStoryV;
		TTm Tm; NewsBase->ArticleStore->GetFieldNmTm(LastArtId, "DateTime", Tm);
		for (int N = 0; N < 30; N++) {
			const TStr Date = Tm.GetWebLogDateStr();
			const uint64 ArticleCount = NewsBase->NewsCounts->GetArticleCount(Date);
			const uint64 StoryCount = NewsBase->NewsCounts->GetStoryCount(Date);
			PJsonVal ArticleJson = TJsonVal::NewObj("date", Date);
			ArticleJson->AddToObj("count", TJsonVal::NewNum((int) ArticleCount));
			DailyArticleV.Add(ArticleJson);
			PJsonVal StoryJson = TJsonVal::NewObj("date", Date);
			StoryJson->AddToObj("count", TJsonVal::NewNum((int) StoryCount));
			DailyStoryV.Add(StoryJson);

			Tm.AddDays(-1);
		}
		RetJson->AddToObj("recentArticles", TJsonVal::NewArr(DailyArticleV));
		RetJson->AddToObj("recentStories", TJsonVal::NewArr(DailyStoryV));
	}
	
	return RetJson;
}

// return array containing info about the last X added articles 
PJsonVal TOverviewSrv::GetArticleActivity(const TStrKdV& FldNmValPrV)
{
	const int ArticleCount = GetFldInt(FldNmValPrV, "articleCount", 100);
	const TStr ArticleUpdatesAfter = GetFldVal(FldNmValPrV, "articleUpdatesAfter", "");
	const uint64 LastArticleActivityId = GetFldUInt64(FldNmValPrV, "lastArticleActivityId", 0);
	const bool ArticlesWithLocationOnly = GetFldBool(FldNmValPrV, "articlesWithLocationOnly", true);

	const TStr FldNmPrefix = "article";
	const int BodyLen = TSAppSrvFun::GetFldInt(FldNmValPrV, FldNmPrefix + "BodyLen", 0);
	TStrV ConceptLangV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; TSAppSrvFun::GetFldValV(FldNmValPrV, FldNmPrefix + "ConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV, ctEntities);

	EnumArticleInfoFlags DefFlags = EnumArticleInfoFlags(aifBasicInfo | aifIncludeSourceInfo | aifIncludeTitle);
	EnumArticleInfoFlags ArticleFlags = NewsSrvUtil->GetArticleFlags(FldNmValPrV, FldNmPrefix, DefFlags);

	PJsonVal RetJson = TJsonVal::NewObj();
	PJsonVal ActivityArrJson = TJsonVal::NewArr();
	TLstNd<TActivityItem>* Node = NewsBase->NewsActivity->ArticleActivityL.Last();
	// remember the time of the latest activity - client can use this to get only updates after this one
	if (Node != NULL) {
		RetJson->AddToObj("lastUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("lastActivityId", Node->Val.GetActivityId().GetStr());
	}
	int N = 0;
	while (Node != NULL && N < ArticleCount) {
		const uint64 ArticleId = Node->Val.GetItemId();
		const TTm& ActivityTm = Node->Val.GetTm();
		// go into the history only as far as the latest update
		if (ArticleUpdatesAfter != "" && ActivityTm.GetWebLogDateTimeStr() <= ArticleUpdatesAfter)
			break;
		if (Node->Val.GetActivityId() <= LastArticleActivityId)
			break;

		Node = Node->Prev();
		if (! NewsBase->ArticleStore->IsRecId(ArticleId))
			continue;
		if (ArticlesWithLocationOnly && !NewsBase->HasLocationForArticleSource(ArticleId))
			continue;
		PJsonVal ArticleJson = NewsSrvUtil->Article(ArticleId, BodyLen, ConceptTypes, ConceptLangV, ArticleFlags);
		ActivityArrJson->AddToArr(ArticleJson);
		N++;
	}
	// add info about the old update that we didn't include (can be used to compute the speed of changes)
	if (Node != NULL) {
		RetJson->AddToObj("oldUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("oldActivityId", Node->Val.GetActivityId().GetStr());
	}

	RetJson->AddToObj("activity", ActivityArrJson);
	return RetJson;
}

// return { "activity": [12,2,5,71,2,2], "eventInfo": { 12: { event info }, 2: {}, ... } }
// the same event can appear multiple times in the activity array (for each time it was changed)
PJsonVal TOverviewSrv::GetEventActivity(const TStrKdV& FldNmValPrV)
{
	// number of returned events
	const int EventCount = GetFldInt(FldNmValPrV, "eventCount", 50);
	// report only events that have a known location
	const bool EventsWithLocationOnly = GetFldBool(FldNmValPrV, "eventsWithLocationOnly", true);
	// the language of the event title to show
	const TStr EventLang = GetFldVal(FldNmValPrV, "eventLang", "eng");
	// return only events that are at least in eventLang language
	const bool EventsWithLangOnly = GetFldBool(FldNmValPrV, "eventsWithLangOnly", false);
	// consider only events that were updated after the specified date
	const TStr EventUpdatesAfter = GetFldVal(FldNmValPrV, "eventUpdatesAfter", "");
	// show only events that have story activity id greater than lastStoryActivityId
	const uint64 LastStoryActivityId = GetFldUInt64(FldNmValPrV, "lastStoryActivityId", 0);

	PJsonVal RetJson = TJsonVal::NewObj();
	
	PJsonVal EventsJson = TJsonVal::NewObj();
	TStrV EventUriV;
	TUInt64Set SeenEventUriSet;		// add each event only once to the eventInfo object
		
	TLstNd<TActivityItem>* Node = NewsBase->NewsActivity->StoryActivityL.Last();
	// remember the time of the latest activity - client can use this to get only updates after this one
	if (Node != NULL) {
		RetJson->AddToObj("lastUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("lastActivityId", Node->Val.GetActivityId().GetStr());
	}
	int N = 0;
	while (Node != NULL && N < EventCount) {
		const uint64 ActivityStoryId = Node->Val.GetItemId();
		const TTm& ActivityTm = Node->Val.GetTm();
		// go into the history only as far as the latest update
		if (EventUpdatesAfter != "" && ActivityTm.GetWebLogDateTimeStr() <= EventUpdatesAfter)
			break;
		if (Node->Val.GetActivityId() <= LastStoryActivityId)
			break;

		Node = Node->Prev();

		if (! NewsBase->StoryStore->IsRecId(ActivityStoryId))
			continue;

		const TRec EventRec = NewsBase->StoryStore->GetRec(ActivityStoryId).DoSingleJoin(NewsBase->Base, "hasEvent");
		if (!EventRec.IsDef())
			continue;
		const uint64 EventId = EventRec.GetRecId();
		const TStr Uri = NewsBase->EventStore->GetRecNm(EventId);

		if (!EventsJson->IsObjKey(Uri)) {
			PJsonVal EventJson = TJsonVal::NewObj("uri", Uri);

			// /////
			// find the location of the event
			TRec LocRec = EventRec.DoSingleJoin(NewsBase->Base, "hasLocation");
			if (!LocRec.IsDef() && EventsWithLocationOnly)
				continue;
			if (LocRec.IsDef())
				EventJson->AddToObj("location", NewsSrvUtil->Location(LocRec.GetRecId()));

			// /////
			// find the title of the event
			PRecSet StoryRecSet = EventRec.DoJoin(NewsBase->Base, "hasStory");
			// sort stories based on their relevance
			StoryRecSet->SortByFq(false);
			TStr Title;
			for (int S = 0; S < StoryRecSet->GetRecs(); S++) {
				const uint64 StoryId = StoryRecSet->GetRecId(S);
				// get article counts per language
				const TStr StoryLang = NewsBase->StoryStore->GetFieldNmStr(StoryId, "Language");
				if (StoryLang == EventLang) {
					Title = NewsBase->StoryStore->GetFieldNmStr(StoryId, "Title");
					break;
				}
			}
			if (Title == "" && StoryRecSet->GetRecs() > 0)
				Title = NewsBase->StoryStore->GetFieldNmStr(StoryRecSet->GetRecId(0), "Title");

			EventJson->AddToObj("title", Title);

			EventJson->AddToObj("articleCount", NewsSrvUtil->GetEventArticleCount(EventId));

			EventsJson->AddToObj(Uri, EventJson);		// event uri -> event info
		}
		EventUriV.Add(Uri);
		N++;
	}
	// add info about the old update that we didn't include (can be used to compute the speed of changes)
	if (Node != NULL) {
		RetJson->AddToObj("oldUpdate", Node->Val.GetTm().GetWebLogDateTimeStr());
		RetJson->AddToObj("oldActivityId", Node->Val.GetActivityId().GetStr());
	}

	RetJson->AddToObj("activity", TJsonVal::NewArr(EventUriV));
	RetJson->AddToObj("eventInfo", EventsJson);
	return RetJson;
}

PJsonVal TOverviewSrv::GetAllActivity(const TStrKdV& FldNmValPrV)
{
	const bool AddArticles = GetFldBool(FldNmValPrV, "addArticles", true);
	const bool AddEvents = GetFldBool(FldNmValPrV, "addEvents", true);
	PJsonVal RetJson = TJsonVal::NewObj();
	PJsonVal RecentActivityJson = TJsonVal::NewObj();
	if (AddArticles)
		RecentActivityJson->AddToObj("articles", NewsSrvUtil->ArticleRecentActivity(FldNmValPrV, NULL));
		//RecentActivityJson->AddToObj("articles", GetArticleActivity(FldNmValPrV));
	if (AddEvents)
		RecentActivityJson->AddToObj("events", NewsSrvUtil->EventsRecentActivity(FldNmValPrV, NULL));
		//RecentActivityJson->AddToObj("events", GetEventActivity(FldNmValPrV));

	RetJson->AddToObj("recentActivity", RecentActivityJson);
	return RetJson;
}