/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */

#include "SourceSrv.h"


///////////////////////////////////////////////
// "source" requests
TSourceSrv::TSourceSrv(const PNewsBase& _NewsBase, const PNewsSrvUtil& _NewsSrvUtil) :
TSAppSrvFun("source", saotJSon), NewsBase(_NewsBase), NewsSrvUtil(_NewsSrvUtil) { }

TStr TSourceSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "getArticles")
			ResJson = GetArticles(FldNmValPrV);

		else if (Action == "getInfo")
			ResJson = GetInfo(FldNmValPrV);

		else {
			ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("No valid actions were provided. Valid values are: getArticles, getInfo"));
			TNotify::StdNotify->OnStatus("invalid action name for source: " + Action);
		}

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TSourceSrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TSourceSrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TSourceSrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TSourceSrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

// provide information about the news source
PJsonVal TSourceSrv::GetInfo(const TStrKdV& FldNmValPrV)
{
	const TStr SourceUri = GetFldVal(FldNmValPrV, "uri", "");
	if (!NewsBase->SourceStore->IsRecNm(SourceUri))
		return TJsonVal::NewObj();
	const uint64 SourceId = NewsBase->SourceStore->GetRecId(SourceUri);
	const int Flags = pifIncludeTitle | pifIncludeDescription | pifIncludeTags | pifIncludeLocation | pifIncludeImportance | pifIncludeDetails;
	return NewsSrvUtil->Source(SourceId, (EnumPublisherInfoFlags) Flags);
}

PJsonVal TSourceSrv::GetArticles(const TStrKdV& FldNmValPrV)
{
	const TStr SourceUri = GetFldVal(FldNmValPrV, "uri", "");
	if (!NewsBase->SourceStore->IsRecNm(SourceUri))
		return TJsonVal::NewObj("error", TStr::Fmt("Invalid source uri specified: %s", SourceUri.CStr()));
	const uint64 SourceRecId = NewsBase->SourceStore->GetRecId(SourceUri);

	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("articles");

	const PRecSet ArticleSet = NewsBase->SourceStore->GetRec(SourceRecId).DoJoin(NewsBase->Base, "hasArticle");

	const int ArticleCount = ArticleSet->GetRecs();

	PJsonVal ResJson = TJsonVal::NewObj();
	// list of articles
	if (ResultTypeV.IsIn("articles"))
		ResJson->AddToObj("articles", NewsSrvUtil->Articles(FldNmValPrV, ArticleSet));

	// list of related concepts
	if (ResultTypeV.IsIn("relatedConcepts"))
		ResJson->AddToObj("relatedConcepts", NewsSrvUtil->RelatedConcepts(FldNmValPrV, ArticleSet));

	// daily topics for the articles inside the Page and Page + Count
	if (ResultTypeV.IsIn("dailyTopics"))
		ResJson->AddToObj("dailyTopics", NewsSrvUtil->DailyTopics(FldNmValPrV, ArticleSet, TUInt64Set()));

	// store source details
	ResJson->AddToObj("sourceInfo", NewsSrvUtil->Source(SourceRecId));
	return ResJson;
}
