/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "StoryLinkingSrv.h"
#include "../ServerArticles/EventRegBase.h"

//////////////////////////////////////
/// TStoryLinkingSrv
TStoryLinkingSrv::TStoryLinkingSrv(const PNewsBase& _NewsBase, const PEventRegBase& _EventRegBase, const PNewsSrvUtil& _NewsUtil, const bool _LogRqToFile) :
	TSAppSrvFun("storyLinking", saotJSon), NewsBase(_NewsBase), EventRegBase(_EventRegBase), NewsUtil(_NewsUtil)
{
	StoryLinking = EventRegBase->StoryLinking;
	LogRqToFile = _LogRqToFile;
	Rnd.Randomize();
}

TStr TStoryLinkingSrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "suggestStoryPair")
			ResJson = SuggestStoryPair(FldNmValPrV);
		else if (Action == "getStoryPair")
			ResJson = GetStoryPair(FldNmValPrV);
		else if (Action == "setClassForStoryPair")
			ResJson = SetClassForStoryPair(FldNmValPrV);
		else if (Action == "buildModel")
			ResJson = BuildModel(FldNmValPrV);
		else if (Action == "saveData")
			ResJson = SaveData(FldNmValPrV);
		else if (Action == "rebuildData")
			ResJson = RebuildData(FldNmValPrV);
		else if (Action == "predictForStoryPair")
			ResJson = PredictForStoryPair(FldNmValPrV);
		else
			ResJson = TJsonVal::NewObj("error", TStr("No known action was provided in the request. Valid actions are suggestStoryPair, getStoryPair, setClassForStoryPair, buildModel, saveData, predictForStoryPair"));

		// if one of the calls returned null we need to init to empty object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj();
		ResJson->AddToObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TStoryLinkingSrv::ExecJSon. Error: %s", E->GetMsgStr().CStr());
	}
	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

PJsonVal TStoryLinkingSrv::GetStoryPair(const TStrKdV& FldNmValPrV)
{
	TStr StoryUri1 = GetFldVal(FldNmValPrV, "storyUri1", "");
	TStr StoryUri2 = GetFldVal(FldNmValPrV, "storyUri2", "");

	if (!NewsBase->StoryStore->IsRecNm(StoryUri1))
		return TJsonVal::NewObj("error", TStr::Fmt("Invalid storyUri1 value was provided (%s)", StoryUri1.CStr()));
	if (!NewsBase->StoryStore->IsRecNm(StoryUri2))
		return TJsonVal::NewObj("error", TStr::Fmt("Invalid storyUri2 value was provided (%s)", StoryUri2.CStr()));

	const uint64 StoryId1 = NewsBase->StoryStore->GetRecId(StoryUri1);
	const uint64 StoryId2 = NewsBase->StoryStore->GetRecId(StoryUri2);
	return StoryPair(FldNmValPrV, StoryId1, StoryId2);
}

// provide the user with info about two stories that are potentially describing the same event
// input contains languages known by the user
PJsonVal TStoryLinkingSrv::SuggestStoryPair(const TStrKdV& FldNmValPrV)
{
	TStrV LangV; GetFldValV(FldNmValPrV, "lang", LangV);
	if (LangV.Len() == 0)
		LangV = NewsConst.LangCoreV;
	PBase Base = NewsBase->Base;
	PStore StoryStore = NewsBase->StoryStore;

	// randomly choose a language for story1
	TStr Lang = LangV[Rnd.GetUniDevInt(LangV.Len())];
	PRecSet StoryRecSet = Base->Search(TQuery::New(Base, TQueryItem(Base, "Story", "Language", Lang)));
	uint64 Stories = StoryRecSet->GetRecs();
	int TryN = 0;
	// try finding a positive example
	if (Rnd.GetUniDevInt(100) % 2 == 0) {
		TNotify::StdNotify->OnStatus("Generating a likely positive example");
		while (TryN < 1000 && Stories > 0) {
			TryN++;
			const int StoryN = Rnd.GetUniDevInt(StoryRecSet->GetRecs());
			const uint64 StoryId1 = StoryRecSet->GetRecId(StoryN);
			const TStr StoryUri1 = StoryStore->GetRecNm(StoryId1);
			TRec EventRec = StoryStore->GetRec(StoryId1).DoSingleJoin(Base, "hasEvent");
			PRecSet StorySet = EventRec.DoJoin(Base, "hasStory");
			for (int N = 0; N < StorySet->GetRecs(); N++) {
				uint64 StoryId2 = StorySet->GetRecId(N);
				const TStr StoryUri2 = StoryStore->GetRecNm(StoryId2);
				const TStr Lang2 = EventRegBase->GetStoryLang(StoryId2);
				if (StoryId2 != StoryId1 && LangV.IsIn(Lang2) && !StoryLinking->IsKnownPair(StoryUri1, StoryUri2))
					return StoryPair(FldNmValPrV, StoryId1, StoryId2);
			}
		}
	}
	// try finding negative example
	else {
		TNotify::StdNotify->OnStatus("Generating a likely negative example");
		while (TryN < 1000 && Stories > 0) {
			TryN++;
			const int StoryN = Rnd.GetUniDevInt(StoryRecSet->GetRecs());
			const uint64 StoryId1 = StoryRecSet->GetRecId(StoryN);
			const TStr StoryUri1 = StoryStore->GetRecNm(StoryId1);
			THash<TUInt64, TSimStoryInfo> SimStoryH;
			EventRegBase->ComputeSimilarStoriesUsingCCA(StoryId1, SimStoryH);
			for (int KeyId = SimStoryH.FFirstKeyId(); SimStoryH.FNextKeyId(KeyId);) {
				const uint64 StoryId2 = SimStoryH.GetKey(KeyId);
				const TStr StoryUri2 = StoryStore->GetRecNm(StoryId2);
				const TStr Lang2 = EventRegBase->GetStoryLang(StoryId2);
				if (StoryId2 != StoryId1 && LangV.IsIn(Lang2) && !StoryLinking->IsKnownPair(StoryUri1, StoryUri2))
					return StoryPair(FldNmValPrV, StoryId1, StoryId2);
			}
		}
	}
	return TJsonVal::NewObj("error", TStr("No valid story pairs were found for the specified languages."));
}

// user provided information whether the two stories are a match or not
PJsonVal TStoryLinkingSrv::SetClassForStoryPair(const TStrKdV& FldNmValPrV)
{
	TStr StoryUri1 = GetFldVal(FldNmValPrV, "storyUri1");
	TStr StoryUri2 = GetFldVal(FldNmValPrV, "storyUri2");
	TStr Username = GetFldVal(FldNmValPrV, "userName");

	if (!NewsBase->StoryStore->IsRecNm(StoryUri1) || !NewsBase->StoryStore->IsRecNm(StoryUri2))
		return TJsonVal::NewObj("error", TStr("Invalid story uri was provided"));
	if (!IsFldNm(FldNmValPrV, "shouldMerge"))
		return TJsonVal::NewObj("error", TStr("shouldMerge value was not provided for the given stories"));
	bool ShouldMerge = GetFldBool(FldNmValPrV, "shouldMerge", false);
	// store the learning example
	StoryLinking->AddLearningExample(EventRegBase, StoryUri1, StoryUri2, ShouldMerge, Username);
	StoryLinking->SaveLearningData();
	return TJsonVal::NewObj("info", TStr("Learning example was stored."));
}

// rebuild model with current learning examples
PJsonVal TStoryLinkingSrv::BuildModel(const TStrKdV& FldNmValPrV)
{
	PJsonVal RetJson = StoryLinking->BuildModel(true);
	return RetJson;
}

// save current example table
PJsonVal TStoryLinkingSrv::SaveData(const TStrKdV& FldNmValPrV)
{
	StoryLinking->SaveLearningData();
	return TJsonVal::NewObj("info", TStr("Data was saved."));
}

PJsonVal TStoryLinkingSrv::RebuildData(const TStrKdV& FldNmValPrV)
{
	StoryLinking->RebuildLearningTable(EventRegBase);
	StoryLinking->SaveLearningData();
	return TJsonVal::NewObj("info", TStr("Data was rebuilt and saved."));
}

// make prediction for provided story pair
PJsonVal TStoryLinkingSrv::PredictForStoryPair(const TStrKdV& FldNmValPrV)
{
	TStr StoryUri1 = GetFldVal(FldNmValPrV, "storyUri1");
	TStr StoryUri2 = GetFldVal(FldNmValPrV, "storyUri2");
	if (!NewsBase->StoryStore->IsRecNm(StoryUri1) || !NewsBase->StoryStore->IsRecNm(StoryUri2))
		return TJsonVal::NewObj("error", "Invalid story uri was provided");
	double Probability = 0;
	bool Ok = StoryLinking->ShouldStoriesMerge(EventRegBase, StoryUri1, StoryUri2, Probability);
	if (!Ok)
		return TJsonVal::NewObj("error", "Unable to predict if stories should merge. There is currently no model built.");
	PJsonVal ResJson = TJsonVal::New();
	ResJson->AddToObj("info", TStr::Fmt("Stories should merge with probability %.3f", Probability));
	ResJson->AddToObj("probability", Probability);
	ResJson->AddToObj("classVal", TJsonVal::NewBool(Probability >= 0.5));
	return ResJson;
}

PJsonVal TStoryLinkingSrv::StoryPair(const TStrKdV& FldNmValPrV, const uint64& StoryId1, const uint64& StoryId2)
{
	PRecSet ArticleSet1 = NewsBase->StoryStore->GetRec(StoryId1).DoJoin(NewsBase->Base, "hasArticle");
	PRecSet ArticleSet2 = NewsBase->StoryStore->GetRec(StoryId2).DoJoin(NewsBase->Base, "hasArticle");
	
	EnumStoryInfoFlags StoryFlags = NewsUtil->GetStoryFlags(FldNmValPrV);
	PJsonVal ResJson = TJsonVal::NewObj();
	PJsonVal Story1 = NewsUtil->Story(StoryId1, 0, NewsConst.LangCoreV, ctAll, StoryFlags);
	PJsonVal Story2 = NewsUtil->Story(StoryId2, 0, NewsConst.LangCoreV, ctAll, StoryFlags);
	Story1->AddToObj("eventId", TJsonVal::NewNum(EventRegBase->GetStoryEvent(StoryId1)));
	Story2->AddToObj("eventId", TJsonVal::NewNum(EventRegBase->GetStoryEvent(StoryId2)));
	ResJson->AddToObj("story1", Story1);
	ResJson->AddToObj("story2", Story2);
	ResJson->AddToObj("articles1", NewsUtil->Articles(FldNmValPrV, ArticleSet1));
	ResJson->AddToObj("articles2", NewsUtil->Articles(FldNmValPrV, ArticleSet2));

	TSimStoryInfo SimInfo; EventRegBase->ComputeCCASimilarityBetweenStories(StoryId1, StoryId2, SimInfo, EventRegBase->GetSettings().StoryMergingMinArticleCosSim);
	TMergeEval MergeEval;
	EventRegBase->ShouldStoriesMerge(StoryId1, StoryId2, SimInfo, MergeEval, true);		// get full info about whether the stories should be merged or not
	PJsonVal SimInfoJson = MergeEval.GetJson(StoryId1, StoryId2, EventRegBase);
	ResJson->AddToObj("simInfo", SimInfoJson);

	return ResJson;
}
