/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "StorySrv.h"

///////////////////////////////////////////////
// "story" requests
TStorySrv::TStorySrv(const PNewsBase& _NewsBase, const PEventRegBase& _EventRegBase, const PNewsSrvUtil& _NewsSrvUtil) :
	TSAppSrvFun("story", saotJSon), NewsBase(_NewsBase), EventRegBase(_EventRegBase), NewsSrvUtil(_NewsSrvUtil) { 
}

TStr TStorySrv::ExecJSon(const TStrKdV& FldNmValPrV, const PSAppSrvRqEnv& RqEnv)
{
	PJsonVal ResJson = NULL;
	try {
		TStr Action = GetFldVal(FldNmValPrV, "action", "");
		if (Action == "getStory")
			ResJson = GetStory(FldNmValPrV);
		else if (Action == "getArticlesRelatedToStory")
			ResJson = GetArticlesRelatedToStory(FldNmValPrV);

		// some method calls return null so we have to create an empty json object
		if (ResJson.Empty())
			ResJson = TJsonVal::NewObj();
	}
	catch (PExcept E) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr(E->GetMsgStr()));
		TNotify::StdNotify->OnStatus("TStorySrv::ExecJson. Error: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TStorySrv::ExecJSon.", E);
	}
	catch (...) {
		ResJson = TJsonVal::NewObj("error", TJsonVal::NewStr("General exception"));
		TNotify::StdNotify->OnStatus("TStorySrv::ExecJson. General exception.");
		NewsBase->ErrorLogFile->NotifyErr("TStorySrv::ExecJSon. General exception.");
	}

	// handle jsonP requests
	if (IsFldNm(FldNmValPrV, "callback")) {
		TStr Callback = GetFldVal(FldNmValPrV, "callback", "");
		return Callback + "(" + TJsonVal::GetStrFromVal(ResJson) + ")";
	}
	else
		return TJsonVal::GetStrFromVal(ResJson);
}

PJsonVal TStorySrv::GetStory(const TStrKdV& FldNmValPrV)
{
	TStrV StoryUriV; GetFldValV(FldNmValPrV, "storyUri", StoryUriV);

	if (StoryUriV.Len() == 0)
		return TJsonVal::NewObj("error", TStr("There is no story uri provided in the request"));
	else if (StoryUriV.Len() == 1)
		return GetSingleStory(FldNmValPrV, StoryUriV[0]);
	else {
		PJsonVal ResJson = TJsonVal::NewObj();
		for (int N = 0; N < StoryUriV.Len(); N++)
			ResJson->AddToObj(StoryUriV[N], GetSingleStory(FldNmValPrV, StoryUriV[N]));
		return ResJson;
	}
}

PJsonVal TStorySrv::GetSingleStory(const TStrKdV& FldNmValPrV, const TStr& StoryUri)
{
	if (!NewsBase->StoryStore->IsRecNm(StoryUri))
		return TJsonVal::NewObj("error", TJsonVal::NewStr("Provided uri is not valid story uri."));

	const uint64 StoryId = NewsBase->StoryStore->GetRecId(StoryUri);

	TStrV ResultTypeV; GetFldValV(FldNmValPrV, "resultType", ResultTypeV);
	if (ResultTypeV.Len() == 0)
		ResultTypeV.Add("articles");

	const PRecSet ArticleSet = NewsBase->StoryStore->GetRec(StoryUri).DoJoin(NewsBase->Base, "hasArticle");

	PJsonVal ResJson = TJsonVal::NewObj();
	// store story details
	if (ResultTypeV.IsIn("info")) {
		TStrV ConceptLangV; GetFldValV(FldNmValPrV, "infoConceptLang", ConceptLangV);
		if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
		TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "infoConceptType", ConceptTypeV);
		EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV);
		const int StoryImageCount = GetFldInt(FldNmValPrV, "infoStoryImageCount", 3);
		EnumStoryInfoFlags StoryFlags = NewsSrvUtil->GetStoryFlags(FldNmValPrV, "info");
		ResJson->AddToObj("info", NewsSrvUtil->Story(StoryId, 0, ConceptLangV, ConceptTypes, StoryFlags, StoryImageCount));
	}
	// list of articles
	if (ResultTypeV.IsIn("articles"))
		ResJson->AddToObj("articles", NewsSrvUtil->Articles(FldNmValPrV, ArticleSet));
	// article uris
	if (ResultTypeV.IsIn("articleUris"))
		ResJson->AddToObj("articleUris", NewsSrvUtil->ArticleUris(FldNmValPrV, ArticleSet));
	// article ids
	if (ResultTypeV.IsIn("articleIds"))
		ResJson->AddToObj("articleIds", NewsSrvUtil->ArticleIds(FldNmValPrV, ArticleSet));
	// list of related concepts
	if (ResultTypeV.IsIn("relatedConcepts"))
		ResJson->AddToObj("relatedConcepts", NewsSrvUtil->RelatedConcepts(FldNmValPrV, ArticleSet));
	// article trends - information that can be used to show how the articles in the story are trending over time
	if (ResultTypeV.IsIn("articleTrend")) {
		const PRecSet ArticleRecSet = NewsBase->StoryStore->GetRec(StoryId).DoJoin(NewsBase->Base, "hasArticle");
		ArticleRecSet->SortByField(true, NewsBase->ArticleStore->GetFieldId("DateTime"));
		PJsonVal TrendJson = TJsonVal::NewObj();
		TrendJson->AddToObj("articles", NewsSrvUtil->Articles(ArticleRecSet, ShortArticleBodyLen));
		TrendJson->AddToObj("totalCount", ArticleRecSet->GetRecs());
		ResJson->AddToObj("articleTrend", TrendJson);
	}
	// related articles - computed by checking the CCA linked articles
	if (ResultTypeV.IsIn("relatedArticles")) {
		TStr Lang = GetFldVal(FldNmValPrV, "relatedArticlesLang", "eng");
		if (!NewsBase->ArticleStore->IsJoinNm("similar_" + Lang))
			return TJsonVal::NewObj("error", TStr::Fmt("similar_%s is not a valid join name", Lang.CStr()));
		PRecSet SimilarArtSet = ArticleSet->DoJoin(NewsBase->Base, "similar_" + Lang);
		ResJson->AddToObj("relatedArticles", NewsSrvUtil->Articles(FldNmValPrV, ArticleSet, "relatedArticles"));
	}
	// similar stories - computed by concepts or CCA
	if (ResultTypeV.IsIn("similarStories")) {
		ResJson->AddToObj("similarStories", SimilarStories(FldNmValPrV, StoryUri));
	}
	return ResJson;
}

// get stories that are similar to the given event
PJsonVal TStorySrv::SimilarStories(const TStrKdV& FldNmValPrV, const TStr& StoryUri)
{
	int Count = GetFldInt(FldNmValPrV, "similarStoriesCount", 20);
	TStrV LangV; GetFldValV(FldNmValPrV, "similarStoriesLang", LangV);
	TStrSet LangH(LangV);
	int MaxDayDiff = GetFldInt(FldNmValPrV, "similarStoriesMaxDayDiff", TInt::Mx);
	const TStr Source = GetFldVal(FldNmValPrV, "similarStoriesSource", "concept");	// concept, cca - how to compute similarity

	// info about returning stories
	TStrV ConceptLangV; GetFldValV(FldNmValPrV, "similarStoriesConceptLang", ConceptLangV);
	if (ConceptLangV.Empty()) ConceptLangV.Add("eng");
	TStrV ConceptTypeV; GetFldValV(FldNmValPrV, "similarStoriesConceptType", ConceptTypeV);
	EnumConceptType ConceptTypes = NewsConst.GetValidTypes(ConceptTypeV);
	EnumStoryInfoFlags StoryFlags = NewsSrvUtil->GetStoryFlags(FldNmValPrV, "similarStories");

	if (!NewsBase->StoryStore->IsRecNm(StoryUri))
		return TJsonVal::NewObj("error", TStr("There is no story for the provided storyUri."));

	const uint64 StoryId = NewsBase->StoryStore->GetRecId(StoryUri);
	PRecSet SimStoryRecSet;
	if (Source == "concept")
		SimStoryRecSet = EventRegBase->GetSimilarStories(StoryId, Count, LangH, MaxDayDiff, ConceptTypes);
	else if (Source == "cca")
		SimStoryRecSet = EventRegBase->GetSimilarStoriesByCCA(StoryId, Count, LangH);
	else
		SimStoryRecSet = TRecSet::New(NewsBase->StoryStore);

	PJsonVal RetJson = TJsonVal::NewObj();
	TJsonValV SimStoryV;
	for (int N = 0; N < SimStoryRecSet->GetRecs(); N++) {
		uint64 SimStoryId = SimStoryRecSet->GetRecId(N);
		int Score = SimStoryRecSet->GetRecFq(N);
		PJsonVal StoryJson = NewsSrvUtil->Story(SimStoryId, Score, ConceptLangV, ConceptTypes, StoryFlags);
		StoryJson->AddToObj("sim", SimStoryRecSet->GetRecFq(N));
		SimStoryV.Add(StoryJson);
	}
	return TJsonVal::NewArr(SimStoryV);
}

PJsonVal TStorySrv::GetArticlesRelatedToStory(const TStrKdV& FldNmValPrV)
{
	TStr StoryUri = GetFldVal(FldNmValPrV, "uri");
	if (!NewsBase->StoryStore->IsRecNm(StoryUri))
		return TJsonVal::NewObj("error", TStr("Invalid storyUri value provided."));

	const TStr Lang = GetFldVal(FldNmValPrV, "lang");
	const int ArticleBodyLen = GetFldInt(FldNmValPrV, "articleBodyLen", DefaultArticleBodyLen);

	THash<TUInt64, TInt> ArticleIdWgtH;
	PRecSet PrimaryArticles = NewsBase->StoryStore->GetRec(StoryUri).DoJoin(NewsBase->Base, "hasArticle");
	for (int ArtN = 0; ArtN < PrimaryArticles->GetRecs(); ArtN++) {
		TRec ArtRec = PrimaryArticles->GetRec(ArtN);

		TStr JoinNm = "similar_" + Lang;
		if (NewsBase->ArticleStore->IsJoinNm(JoinNm)) {
			PRecSet SimArtRecSet = ArtRec.DoJoin(NewsBase->Base, JoinNm);
			const int SimArtCnt = SimArtRecSet->GetRecs();
			for (int SimArtN = 0; SimArtN < SimArtCnt; SimArtN++) {
				TRec SimArt = SimArtRecSet->GetRec(SimArtN);
				TRec OrigArt = SimArt.DoSingleJoin(NewsBase->Base, "isDuplicateOf");
				// if article is a duplicate then use the original article id, otherwise use the sim. art. id
				uint64 SimArtId = (OrigArt.IsDef()) ? OrigArt.GetRecId() : SimArt.GetRecId();
				ArticleIdWgtH.AddDat(SimArtId) += 1;
			}
		}
	}
	ArticleIdWgtH.SortByDat(false);
	TUInt64IntKdV ArticleIdWgtV; ArticleIdWgtH.GetKeyDatKdV(ArticleIdWgtV);
	PRecSet ArticleSet = TRecSet::New(NewsBase->ArticleStore, ArticleIdWgtV, true);
	PJsonVal RetVal = TJsonVal::NewObj("results", NewsSrvUtil->Articles(ArticleSet, ArticleBodyLen, TUInt64Set(), true));
	return RetVal;
}