/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ConceptDailyTrends.h"

// ///////////
// TConceptDailyTrends
TConceptDailyTrends::TConceptDailyTrends(const PNewsBase& _NewsBase, const PNewsSearch& _NewsSearch, const PNewsSrvUtil& _NewsSrvUtil, const TStr& _FName) :
		TSAppSrvFun("foo"), NewsBase(_NewsBase), NewsSearch(_NewsSearch), NewsSrvUtil(_NewsSrvUtil), FName(_FName) {
	TrendsManager = NewsBase->TrendsManager;
	NewsBaseFix = NewsBase->NewsBaseFix;
	InfoLogger = TLogger::New(3);
	InfoLogger->AddLogger(new TFileNotify("./Logs/Log-ConceptDailyTrends.txt", true, true));
	LabelLang = "eng";
	Load();
}


void TConceptDailyTrends::Load()
{
	if (TFile::Exists(FName)) {
		InfoLogger->NotifyInfo("Loading concept daily trends info.");
		TNotify::StdNotify->OnStatus("Loading concept daily trends info...");
		try {
			TFIn FIn(FName);
			HiddenConcepts.Load(FIn);
			EventIdConceptIdPrV.Load(FIn);
			TNotify::StdNotify->OnStatus("Loaded.");
		}
		catch (...) {
			TNotify::StdNotify->OnStatus("Exception while loading daily trends.");
		}
	}
	else {
		InfoLogger->NotifyInfo("No concept daily trends data found.");
		TNotify::StdNotify->OnStatus("No concept daily trends data found.");
	}
}

void TConceptDailyTrends::Save() const
{
	try
	{
		TFOut FOut(FName);
		HiddenConcepts.Save(FOut);
		EventIdConceptIdPrV.Save(FOut);
	}
	catch (PExcept E)  { 
		TNotify::StdNotify->OnStatus("TConceptDailyTrends::Save. Error while saving daily trends: " + E->GetMsgStr());
		NewsBase->ErrorLogFile->NotifyErr("TConceptDailyTrends::Save.", E);
	}
}

void TConceptDailyTrends::SetCustomParams(const TStr& TestPopOffs, const TStr& NullPopOffs, const int& IgnoreTestConceptsBelowFq, const int& IgnoreNullConceptsBelowFq, const int& RetCount, const TStr& ConceptTypes)
{
	TStrKdV FldNmValPrV;
	FldNmValPrV.Add(TStrKd("testPop", TestPopOffs));
	FldNmValPrV.Add(TStrKd("nullPop", NullPopOffs));
	FldNmValPrV.Add(TStrKd("ignoreTestConceptsBelowFq", TInt::GetStr(IgnoreTestConceptsBelowFq)));
	FldNmValPrV.Add(TStrKd("ignoreNullConceptsBelowFq", TInt::GetStr(IgnoreNullConceptsBelowFq)));
	FldNmValPrV.Add(TStrKd("conceptCount", TInt::GetStr(RetCount)));
	TStrV ConceptTypeV; ConceptTypes.SplitOnAllCh(',', ConceptTypeV);
	for (int N = 0; N < ConceptTypeV.Len(); N++)
		FldNmValPrV.Add(TStrKd("conceptType", ConceptTypeV[N]));
	SetParams(FldNmValPrV);
}

// update trends with the specified values
PJsonVal TConceptDailyTrends::UpdateTrends()
{
	TNotify::StdNotify->OnStatus("Updating daily trends...");
	// clear cached TopTrendsJson - it should be recomputed since we have new concept trends
	ClearCache();
	ValidTypes = ctEntities;
	Trends = NULL;

	TStr Lang = GetFldVal(ParamsPrV, "labelLang", "eng");
	LabelLang = Lang;
	
	TStrV TestPopKeyV, NullPopKeyV;
	if (IsFldNm(ParamsPrV, "testPop")) {
		TStr TestPop = GetFldVal(ParamsPrV, "testPop", "D0");
		TestPopKeyV = TrendsManager->ParsePopKeys(TestPop);
	}
	if (IsFldNm(ParamsPrV, "nullPop")) {
		TStr NullPop = GetFldVal(ParamsPrV, "nullPop", "Total");
		NullPopKeyV = TrendsManager->ParsePopKeys(NullPop);
	}
		
	InfoLogger->NotifyInfo("Updating trends...");
	InfoLogger->NotifyInfo("Test population offsets: %s", TStr::GetStr(TestPopKeyV, ", ").CStr());
	InfoLogger->NotifyInfo("Null population offsets: %s", TStr::GetStr(NullPopKeyV, ", ").CStr());

	// validate the correctness of necessary params
	if (TestPopKeyV.Len() == 0) {
		TNotify::StdNotify->OnStatus("testPopOffs and testPopKeys did not contain any valid values for a population. Values should be of type 'Total', 'Y0', 'M1' or 'D2' or a comma separated list of them, like 'M1,M2,M3'.");
		return TJsonVal::NewObj("error", TStr("testPopOffs and testPopKeys did not contain any valid values for a population. Values should be of type 'Total', 'Y0', 'M1' or 'D2' or a comma separated list of them, like 'M1,M2,M3'."));
	}
	if (NullPopKeyV.Len() == 0) {
		TNotify::StdNotify->OnStatus("nullPopOffs and nullPopKeys did not contain any valid values for a population. Values should be of type 'Total', 'Y0', 'M1' or 'D2' or a comma separated list of them, like 'M1,M2,M3'.");
		return TJsonVal::NewObj("error", TStr("nullPopOffs and nullPopKeys did not contain any valid values for a population. Values should be of type 'Total', 'Y0', 'M1' or 'D2' or a comma separated list of them, like 'M1,M2,M3'."));
	}
		
	const int IgnoreTestConceptsBelowFq = GetFldInt(ParamsPrV, "ignoreTestConceptsBelowFq", 10);
	const int IgnoreNullConceptsBelowFq = GetFldInt(ParamsPrV, "ignoreNullConceptsBelowFq", 10);
	const int ConceptCount = GetFldInt(ParamsPrV, "conceptCount", 30);
	TStrV ConceptTypeV; GetFldValV(ParamsPrV, "conceptType", ConceptTypeV);
	ValidTypes = NewsConst.GetValidTypes(ConceptTypeV);
	InfoLogger->NotifyInfo("Ignore test concepts below fq = %d, Ignore null concepts below fq = %d, desired concept count = %d, concept types = %s", IgnoreTestConceptsBelowFq, IgnoreNullConceptsBelowFq, ConceptCount, TStr::GetStr(ConceptTypeV, ", ").CStr());
	
	// most popular trends computed using test and null distributions
	TestPop = TrendsManager->GetPopForKeys(TestPopKeyV);
	NullPop = TrendsManager->GetPopForKeys(NullPopKeyV);
	Trends = TrendsManager->ComparePops(TestPop, NullPop, IgnoreTestConceptsBelowFq, IgnoreNullConceptsBelowFq, true, tmPearsonResidual);
	
	TStrV OffDescV;
	for (int N=0; N < 30; N++) {
		TStr Off = "D" + TInt(N).GetStr();
		OffDescV.Add(Off);
		PConceptPop Pop = TrendsManager->GetPopForOff(Off);
		PConceptTrends Trends = TrendsManager->ComparePops(Pop, NullPop, 0, IgnoreNullConceptsBelowFq, false, tmPearsonResidual);
		MonthlyTrendV.Add(Trends);
	}
		
	PJsonVal ArrJson = TJsonVal::NewArr();
	int Added = 0;
	const int ConcCount = Trends->Len();
	for (int C=0; C < ConcCount; C++) {
		TConceptId Id = Trends->GetTopConceptId(C);
		// was the concept hidden by the admin?
		if (IsHidden(Id) || ! NewsBase->ConceptStore->IsRecId((uint64) Id)) 
			continue;
		// do we want concepts of this type?
		TStr Type = NewsBase->ConceptStore->GetFieldNmStr((uint64) Id, "Type");
		if (! NewsConst.IsValidType(Type, ValidTypes))
			continue;

		const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm((uint64) Id);
		PJsonVal ConceptTrendJson = ConceptTrend(Id, MonthlyTrendV);
		// send also info about test and null population frequencies so that we can see what kind of concepts we get as most popular
		ConceptTrendJson->AddToObj("testPopFq", TestPop->GetConceptPop(Id));
		ConceptTrendJson->AddToObj("nullPopFq", NullPop->GetConceptPop(Id));
		if (C < 10) 
			InfoLogger->NotifyInfo("Top concept %d: testFq: %d, nullFq: %d, chi: %.3f, uri: %s", C, TestPop->GetConceptPop(Id), NullPop->GetConceptPop(Id), Trends->GetChiForConcept(Id, 0), NewsBase->ConceptStore->GetRecNm((uint64) Id).CStr());
		ArrJson->AddToArr(ConceptTrendJson);
		Added++;
		if (Added >= ConceptCount)
			break;
	}
	InfoLogger->NotifyInfo("Computed a set of %d top trending concepts", ArrJson->GetArrVals());
	TopTrendsJson = TJsonVal::NewObj("trends", ArrJson);
	// todo: maybe add some other info, if needed
	TNotify::StdNotify->OnStatus("Daily trends updated.");
	return TJsonVal::NewObj("info", TStr::Fmt("Trends were updated. Trends contain %d concepts (desired number: %d).", ArrJson->GetArrVals(), ConceptCount));
}

// return the list of top concepts that are of the desired type
void TConceptDailyTrends::GetTopConceptsByType(const EnumConceptType& ConceptType, const int& _ConceptCount, TUInt64V& TopConceptV)
{
	// if the user didn't specify the number of desired concepts then use the number specified by default conceptCount parameters used for returning the getTopTrends data
	const int ConceptCount = _ConceptCount > 0 ? _ConceptCount : GetFldInt(ParamsPrV, "conceptCount", 10);
	
	TopConceptV.Clr();
	if (MonthlyTrendV.Len() == 0) UpdateTrends();
	if (MonthlyTrendV.Len() == 0 || Trends.Empty())
		return;
	const int ConcCount = Trends->Len();
	for (int C = 0; C < ConcCount && TopConceptV.Len() < ConceptCount; C++) {
		TConceptId Id = Trends->GetTopConceptId(C);
		// was the concept hidden by the admin?
		if (IsHidden(Id) || !NewsBase->ConceptStore->IsRecId((uint64) Id))
			continue;
		// do we want to ignore concepts of this type?
		TStr Type = NewsBase->ConceptStore->GetFieldNmStr((uint64) Id, "Type");
		if (!NewsConst.IsValidType(Type, ConceptType))
			continue;
		TopConceptV.Add((uint64) Id);
	}
}

// find the relevance of individual concepts in conceptV and the choose ConceptCount top concepts and return them in TopConceptV
void TConceptDailyTrends::GetTopConceptsFromList(const TUInt64V& ConceptV, const int& ConceptCount, TUInt64V& TopConceptV)
{
	TopConceptV.Clr();
	if (MonthlyTrendV.Len() == 0) UpdateTrends();
	if (MonthlyTrendV.Len() == 0 || Trends.Empty())
		return;
	// find trend values for specified concepts
	THash<TUInt64, TFlt> ConceptToTrendH;
	for (int N = 0; N < ConceptV.Len(); N++) {
		// debug:
		const uint64 ConceptId = (uint64) ConceptV[N];
		const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(ConceptId);
		const float Chi = Trends->GetChiForConcept( (TConceptId) ConceptV[N]);
		ConceptToTrendH.AddDat(ConceptV[N], Chi);
	}
		
	// sort by trend
	ConceptToTrendH.SortByDat(false);
	// return top ConceptCount concepts
	for (int KeyId = ConceptToTrendH.FFirstKeyId(); ConceptToTrendH.FNextKeyId(KeyId) && TopConceptV.Len() < ConceptCount;) {
		TConceptId Id = TConceptId (ConceptToTrendH.GetKey(KeyId));
		if (IsHidden(Id))
			continue;
		TopConceptV.Add((uint64) Id);
	}
}

PJsonVal TConceptDailyTrends::ConceptTrends(const TUInt64V& ConceptIdV, const int& NrOfDays)
{
	if (MonthlyTrendV.Len() == 0) UpdateTrends();
	if (MonthlyTrendV.Len() == 0 || Trends.Empty() || TestPop.Empty() || NullPop.Empty())
		return TJsonVal::NewObj("error", TStr("A necessary data structure is still invalid. Unable to compute results."));

	// choose a subset of days in the last month (could use all)
	TConceptTrendsV PopV; MonthlyTrendV.GetSubValV(0, NrOfDays, PopV);

	const int ConcCount = Trends->Len();
	PJsonVal ArrJson = TJsonVal::NewArr();
	for (int N = 0; N < ConceptIdV.Len(); N++) {
		const uint64 ConceptId = ConceptIdV[N];
		const float Chi = Trends->GetChiForConcept((TConceptId) ConceptId, 0);
		if (Chi == 0)
			continue;
		PJsonVal ConceptTrendJson = ConceptTrend(ConceptId, PopV);
		// send also info about test and null population frequencies so that we can see what kind of concepts we get as most popular
		ConceptTrendJson->AddToObj("testPopFq", TJsonVal::NewNum(TestPop->GetConceptPop(ConceptId)));
		ConceptTrendJson->AddToObj("nullPopFq", TJsonVal::NewNum(NullPop->GetConceptPop(ConceptId)));
		ArrJson->AddToArr(ConceptTrendJson);
	}
	//PJsonVal TopTrends = TJsonVal::NewObj("trends", ArrJson);
	return ArrJson;
}


// get the monthly trend for a given concept
PJsonVal TConceptDailyTrends::GetConceptTrendArray(const uint64& ConceptRecId, const TConceptTrendsV& TrendV) const
{
	if (!NewsBase->ConceptStore->IsRecId(ConceptRecId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown concept id %I64u", ConceptRecId)));
	// add concept's monthly trend
	TFltV ChiArr;
	for (int FqN = 0; FqN < TrendV.Len(); FqN++)
		ChiArr.Add(TrendV[FqN]->GetChiForConcept((TConceptId) ConceptRecId, 0));
	return TJsonVal::NewArr(ChiArr);
}

// return trend info for the specified concept
// the trend includes concept name, uri, chi and trend for all days in ConceptTrendsV
PJsonVal TConceptDailyTrends::ConceptTrend(const uint64& ConceptId, const TConceptTrendsV& ConceptTrendsV) const
{
	PJsonVal ConcJson = TJsonVal::NewObj();
	if (!NewsBase->ConceptStore->IsRecId(ConceptId))
		return ConcJson;
	if (Trends.Empty())
		return ConcJson;
	const float Chi = Trends->GetChiForConcept((TConceptId) ConceptId, 0);
	const TStr ConceptUri = NewsBase->ConceptStore->GetRecNm(ConceptId);
	const TStr ConceptLabel = NewsBase->ConceptLabel->GetConceptLabel(ConceptUri, LabelLang);
	ConcJson->AddToObj("uri", ConceptUri);
	ConcJson->AddToObj("type", NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type"));
	ConcJson->AddToObj("chi", Chi);
	ConcJson->AddToObj("label", ConceptLabel);
	TRec ImgRec = NewsBase->ConceptStore->GetRec(ConceptId).DoSingleJoin(NewsBase->Base, "hasPrimaryImage");
	if (ImgRec.IsDef())
		ConcJson->AddToObj("img", NewsBase->ImageStore->GetFieldNmStr(ImgRec.GetRecId(), "URI"));
	TFltV ChiV;
	int Len = ConceptTrendsV.Len();
	for (int N=0; N < Len; N++)
		ChiV.Add(ConceptTrendsV[N]->GetChiForConcept((TConceptId) ConceptId, 0));
	PJsonVal FreqArr = TJsonVal::NewArr(ChiV);
	ConcJson->AddToObj("trend", FreqArr);
	return ConcJson;
}

void TConceptDailyTrends::GetTopTrends(const EnumConceptType& ValidTypes, const int& ConceptCount, TUInt64V& TopConceptV)
{
	TopConceptV.Clr();
	if (Trends.Empty() || TestPop.Empty() || NullPop.Empty())
		return;

	const int ConcCount = Trends->Len();
	for (int C = 0; C < ConcCount && TopConceptV.Len() < ConceptCount; C++) {
		TConceptId Id = Trends->GetTopConceptId(C);
		// was the concept hidden by the admin?
		if (IsHidden(Id) || !NewsBase->ConceptStore->IsRecId((uint64) Id))
			continue;
		// do we want to ignore concepts of this type?
		TStr Type = NewsBase->ConceptStore->GetFieldNmStr((uint64) Id, "Type");
		if (!NewsConst.IsValidType(Type, ValidTypes))
			continue;
		TopConceptV.Add((uint64) Id);
	}
}

PRecSet TConceptDailyTrends::GetArticlesForConcepts(const TUInt64V& ConceptIdV)
{
	PRecSet ConceptRecSet = TRecSet::New(NewsBase->ConceptStore, ConceptIdV);
	ArticleRecSet = ConceptRecSet->DoJoin(NewsBase->Base, "hasArticle");
	ArticleRecSet->SortById(false);
	//ArticleRecSetTimeStamp = TTm::GetCurLocTm();
	return ArticleRecSet;
}

PJsonVal TConceptDailyTrends::GetArticles(const PRecSet& ArticleSet, const TStrV& LabelLanguageV, const int& Page, const int& Count)
{
	/*TUInt64V ConceptRecIdV;
	for (int N = 0; N < ConceptV.Len(); N++)
		if (NewsBase->ConceptStore->IsRecId(ConceptV[N]))
			ConceptRecIdV.Add((uint64) ConceptV[N]);
	
	if (ArticleRecSet.Empty() || TTm::GetDiffSecs(TTm::GetCurLocTm(), ArticleRecSetTimeStamp) > 120) {
		
		TNotify::StdNotify->OnStatus("Trends: Recomputing top articles for top trends..."); 
	}*/
	PJsonVal ResJson = TJsonVal::NewObj();
	PRecSet SubArticleRecSet = GetNonDuplArticles(ArticleRecSet, Page, Count);
	PJsonVal ArticleJsonArr = NewsSrvUtil->Articles(SubArticleRecSet, DefaultArticleBodyLen);
	// add to the array of articles info about the source location and top concept - used to show the map and top concepts
	int Arts = ArticleJsonArr->GetArrVals();
	for (int N = 0; N < Arts; N++) {
		PJsonVal ArtJson = ArticleJsonArr->GetArrVal(N);
		TStr Uri = ArtJson->GetObjStr("uri");
		TRec SourceRec = NewsBase->ArticleStore->GetRec(Uri).DoSingleJoin(NewsBase->Base, "hasSource");
		TFltPr Location = NewsBase->SourceStore->GetFieldNmFltPr(SourceRec.GetRecId(), "Location");
		ArtJson->AddToObj("source_lat", TJsonVal::NewStr(Location.Val1.GetStr()));
		ArtJson->AddToObj("source_long", TJsonVal::NewStr(Location.Val2.GetStr()));
		PRecSet ConceptSet = NewsBase->ArticleStore->GetRec(Uri).DoJoin(NewsBase->Base, "hasConcept");
		ConceptSet->SortByFq(false);
		for (int C = 0; C < ConceptSet->GetRecs(); C++) {
			if (NewsConst.IsEntity(NewsBase->ConceptStore->GetFieldNmStr(ConceptSet->GetRecId(C), "Type")))  {
				ArtJson->AddToObj("topConcept", NewsSrvUtil->Concept(ConceptSet->GetRecId(C), -1, LabelLanguageV));
				break;
			}
		}
	}
	ResJson->AddToObj("articles", ArticleJsonArr);
	return ResJson;
}

// get concept name, type, uri, description and image
PJsonVal TConceptDailyTrends::GetConceptInfo(const uint64& ConceptId, const TStrV& LabelLanguageV, const bool& IncludeDesc, const bool& IncludeImg) const
{
	if (!NewsBase->ConceptStore->IsRecId(ConceptId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown concept id %I64u", ConceptId)));
	
	PJsonVal ConceptInfoJson = NewsSrvUtil->Concept(ConceptId, -1, LabelLanguageV);
	// add the concept description
	if (IncludeDesc && !NewsBase->ConceptStore->IsFieldNmNull(ConceptId, "Description"))
		ConceptInfoJson->AddToObj("desc", NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Description"));
	// add the concept image
	TRec ImageRec = NewsBase->ConceptStore->GetRec(ConceptId).DoSingleJoin(NewsBase->Base, "hasPrimaryImage");
	if (IncludeImg && ImageRec.IsDef())
		ConceptInfoJson->AddToObj("img", NewsBase->ImageStore->GetRecNm(ImageRec.GetRecId()));
	return ConceptInfoJson;
}

// get concepts most related to the ConceptRecId. Relatedness is computed by looking at latest 500 articles
// containing ConceptRecId and finding the most frequently coocurring concepts
PJsonVal TConceptDailyTrends::GetRelatedConceptsFromLatestArticles(const uint64& ConceptId, const int& RelatedConceptCount, const int& LatestArticles) const
{
	if (!NewsBase->ConceptStore->IsRecId(ConceptId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown concept id %I64u", ConceptId)));
	PRecSet ArticleSet = NewsBase->ConceptStore->GetRec(ConceptId).DoJoin(NewsBase->Base, "hasArticle");
	int Start = max(0, ArticleSet->GetRecs() - LatestArticles);
	PRecSet ArticleSubSet = ArticleSet->GetLimit(LatestArticles, Start);
	THashSet<TConceptId> IgnoreConceptH; IgnoreConceptH.AddKey((TConceptId) ConceptId);
	return GetRelatedConcepts(ArticleSubSet, RelatedConceptCount, IgnoreConceptH, ctEntities);
}

PJsonVal TConceptDailyTrends::GetRelatedConcepts(const PRecSet& ArticleRecSet, const int& RelatedConceptCount, const THashSet<TConceptId> IgnoreConceptH, const EnumConceptType& ValidTypes) const
{
	// add most related concepts and their trends
	PRecSet RelatedConceptSet = ArticleRecSet->DoJoin(NewsBase->Base, "hasConcept", 1000, true);
	RelatedConceptSet->SortByFq(false);

	// add related concepts
	PJsonVal RelatedConcArr = TJsonVal::NewArr();
	int RelatedTotal = RelatedConceptSet->GetRecs();
	TJsonValV ConceptArrV;
	// take the daily trends computed for each day
	for (int ConcN = 0; ConcN < RelatedTotal && ConceptArrV.Len() < RelatedConceptCount; ConcN++) {
		TConceptId ConceptId = (TConceptId) RelatedConceptSet->GetRecId(ConcN);
		// don't add the same concept or a hidden one
		if (IgnoreConceptH.IsKey(ConceptId) || IsHidden(ConceptId))
			continue;
		TStr Type = NewsBase->ConceptStore->GetFieldNmStr((uint64) ConceptId, "Type");
		if (! NewsConst.IsValidType(Type, ValidTypes))
			continue;
		ConceptArrV.Add(ConceptTrend(ConceptId, MonthlyTrendV));
	}
	return TJsonVal::NewArr(ConceptArrV);
}

PJsonVal TConceptDailyTrends::GetConceptTrends(const PRecSet& ConceptSet, const int& RelatedConceptCount, const THashSet<TConceptId> IgnoreConceptH, const EnumConceptType& ValidTypes) const
{
	// add most related concepts and their trends
	ConceptSet->SortByFq(false);

	// add related concepts
	PJsonVal RelatedConcArr = TJsonVal::NewArr();
	const int RelatedTotal = ConceptSet->GetRecs();
	TJsonValV ConceptArrV;
	// take the daily trends computed for each day
	for (int ConcN = 0; ConcN < RelatedTotal && ConceptArrV.Len() < RelatedConceptCount; ConcN++) {
		TConceptId ConceptId = (TConceptId) ConceptSet->GetRecId(ConcN);
		// don't add the same concept or a hidden one
		if (IgnoreConceptH.IsKey(ConceptId) || IsHidden(ConceptId))
			continue;
		TStr Type = NewsBase->ConceptStore->GetFieldNmStr((uint64) ConceptId, "Type");
		if (!NewsConst.IsValidType(Type, ValidTypes))
			continue;
		ConceptArrV.Add(ConceptTrend(ConceptId, MonthlyTrendV));
	}
	return TJsonVal::NewArr(ConceptArrV);
}

// get stories in which the concept appears
PJsonVal TConceptDailyTrends::GetConceptEvents(const uint64& ConceptRecId, const int& Page, const int& Count, const TStr& EventLang, const int& MinConceptScore, const int& MaxDaysBack, const int& MaxDaysAhead) const
{
	// 1. find latest stories for the concept
	if (!NewsBase->ConceptStore->IsRecId(ConceptRecId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown concept id %I64u", ConceptRecId)));
	
	TEventQuery Query;
	Query.ConceptUriV.Add(NewsBase->ConceptStore->GetRecNm(ConceptRecId));
	if (MaxDaysBack != TInt::Mx) {
		TTm Tm = TTm::GetCurLocTm();
		Tm.AddDays(-MaxDaysBack);
		Query.SetDateStart(Tm);
	}
	if (MaxDaysAhead != TInt::Mx) {
		TTm Tm = TTm::GetCurLocTm();
		Tm.AddDays(MaxDaysAhead);
		Query.SetDateEnd(Tm);
	}
	
	PJsonVal ErrorJson;
	PRecSet EventSet; 
	bool Ok = NewsSearch->SearchEvents(Query, EventSet, ErrorJson);
	if (!Ok)
		return ErrorJson;

	// keep only events where the concept is at least MinConceptScore important 
	EventSet->FilterByFq(MinConceptScore, TInt::Mx);
	EventSet->SortByField(false, NewsBase->EventStore->GetFieldId("EventDate"));	// sort stories by decreasing date

	// 2. based on the popularity of the concept in the stories determine the treshold - for popular concepts just show the most popular stories

	// 3. for each story extract the summary of the story and return it
	PRecSet ShortEventSet = EventSet->GetLimit(Count, Count*Page);
	TJsonValV EventJsonV;
	for (int N = 0; N < ShortEventSet->GetRecs(); N++) {
		PJsonVal Event = GetEvent(ShortEventSet->GetRecId(N), 50, EventLang, NewsConst.LangCoreV, ctAll);
		Event->AddToObj("wgt", ShortEventSet->GetRecFq(N));
		EventJsonV.Add(Event);
	}
	return TJsonVal::NewArr(EventJsonV);
}


// get stories in which the concept appears
PJsonVal TConceptDailyTrends::GetConceptStories(const uint64& ConceptRecId, const int& Page, const int& Count, const int& MinConceptScore) const
{
	// 1. find latest stories for the concept
	if (!NewsBase->ConceptStore->IsRecId(ConceptRecId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown concept id %I64u", ConceptRecId)));
	PRecSet StorySet = NewsBase->ConceptStore->GetRec(ConceptRecId).DoJoin(NewsBase->Base, "hasStory");
	// keep only stories where the concept is at least MinConceptScore important 
	StorySet->FilterByFq(MinConceptScore, TInt::Mx);
	//StorySet->SortById(false);
	StorySet->SortByField(false, NewsBase->StoryStore->GetFieldId("AverageDate"));	// sort stories by decreasing date

	// 2. based on the popularity of the concept in the stories determine the treshold - for popular concepts just show the most popular stories

	// 3. for each story extract the summary of the story and return it
	PRecSet ShortStorySet = StorySet->GetLimit(Count, Count*Page);
	TJsonValV StoryJsonV;
	for (int N = 0; N < ShortStorySet->GetRecs(); N++) {
		PJsonVal Story = GetStory(ShortStorySet->GetRecId(N), 10);
		Story->AddToObj("wgt", ShortStorySet->GetRecFq(N));
		StoryJsonV.Add(Story);
	}
	return TJsonVal::NewArr(StoryJsonV);
}

PJsonVal TConceptDailyTrends::GetConceptArticles(const uint64& ConceptRecId, const int& Page, const int& Count, const int& MinConceptScore) const
{
	PJsonVal ResJson = TJsonVal::NewObj();
	// 1. find latest stories for the concept
	if (!NewsBase->ConceptStore->IsRecId(ConceptRecId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown concept id %I64u", ConceptRecId)));
	PRecSet ArticleSet = NewsBase->ConceptStore->GetRec(ConceptRecId).DoJoin(NewsBase->Base, "hasArticle");
	// keep only stories where the concept is at least MinConceptScore important 
	ArticleSet->FilterByFq(MinConceptScore, TInt::Mx);
	// if we get very few articles then don't filter by min score
	if (ArticleSet->GetRecs() < 5)
		ArticleSet = NewsBase->ConceptStore->GetRec(ConceptRecId).DoJoin(NewsBase->Base, "hasArticle");
	ArticleSet->SortById(false);
	PRecSet ShortArticleSet = ArticleSet->GetLimit(Count, Count*Page);
	return NewsSrvUtil->Articles(ShortArticleSet, -1, TUInt64Set(), true);
}

// return details about the story StoryRecId
// details include the title, summary, date, article count, (maybe) top concepts
PJsonVal TConceptDailyTrends::GetStory(const uint64& StoryRecId, const int& TopConcepts, const TStrV& LabelLanguageV, const EnumConceptType& ConceptTypes) const
{
	if (!NewsBase->StoryStore->IsRecId(StoryRecId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown story id %I64u", StoryRecId)));
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("uri", NewsBase->StoryStore->GetRecNm(StoryRecId));
	ResJson->AddToObj("title", NewsBase->StoryStore->GetFieldNmStr(StoryRecId, "Title"));
	ResJson->AddToObj("summary", NewsBase->StoryStore->GetFieldNmStr(StoryRecId, "Summary"));
	ResJson->AddToObj("date", NewsBase->StoryStore->GetFieldNmStr(StoryRecId, "AverageDate"));
	ResJson->AddToObj("articleCount", NewsBase->StoryStore->GetFieldNmInt(StoryRecId, "ArticleCount"));
	ResJson->AddToObj("lang", NewsBase->StoryStore->GetFieldNmStr(StoryRecId, "Language"));

	if (TopConcepts > 0) {
		PRecSet ConceptRecSet = NewsBase->StoryStore->GetRec(StoryRecId).DoJoin(NewsBase->Base, "hasConcept");
		ConceptRecSet->SortByFq(false);
		TUInt64V ConceptSet = GetValidConcepts(ConceptRecSet, TopConcepts, ConceptTypes);
		TJsonValV ConceptV;
		for (int N = 0; N < ConceptSet.Len(); N++)
			ConceptV.Add(GetConceptInfo(ConceptSet[N], LabelLanguageV, false, false));
		ResJson->AddToObj("topConcepts", ConceptV);
	}
	PRecSet ArticleRecSet = NewsBase->StoryStore->GetRec(StoryRecId).DoJoin(NewsBase->Base, "hasArticle");
	ArticleRecSet->SortByFq(false);
	for (int N = 0; N < ArticleRecSet->GetRecs(); N++) {
		uint64 ArticleRecID = ArticleRecSet->GetRecId(N);
		PRecSet ImageRecSet = ArticleRecSet->GetRec(N).DoJoin(NewsBase->Base, "hasImage");
		if (ImageRecSet->GetRecs() > 0) {
			ResJson->AddToObj("img", NewsBase->ImageStore->GetRecNm(ImageRecSet->GetRecId(0)));
			break;
		}
	}
	return ResJson;
}

// return details about the story StoryRecId
// details include the title, summary, date, article count, (maybe) top concepts
PJsonVal TConceptDailyTrends::GetEvent(const uint64& EventRecId, const int& TopConcepts, const TStr& EventLang, const TStrV& LabelLanguageV, const EnumConceptType& ConceptTypes) const
{
	if (!NewsBase->EventStore->IsRecId(EventRecId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown event id %I64u", EventRecId)));
	PJsonVal ResJson = TJsonVal::NewObj();
	ResJson->AddToObj("uri", NewsBase->EventStore->GetRecNm(EventRecId));
	PRecSet StorySet = NewsBase->EventStore->GetRec(EventRecId).DoJoin(NewsBase->Base, "hasStory");
	StorySet->SortByFq(false);
	for (int N = 0; N < StorySet->GetRecs(); N++) {
		const uint64 StoryId = StorySet->GetRecId(N);
		const TStr Lang = NewsBase->StoryStore->GetFieldNmStr(StoryId, "Language");
		if (Lang == EventLang) {
			ResJson->AddToObj("title", NewsBase->StoryStore->GetFieldNmStr(StoryId, "Title"));
			ResJson->AddToObj("summary", NewsBase->StoryStore->GetFieldNmStr(StoryId, "Summary"));
			break;
		}
	}
	ResJson->AddToObj("date", NewsBase->EventStore->GetFieldNmStr(EventRecId, "EventDate"));
	ResJson->AddToObj("articleCount", NewsSrvUtil->GetEventArticleCount(EventRecId));
	// add location
	TRec LocRec = NewsBase->EventStore->GetRec(EventRecId).DoSingleJoin(NewsBase->Base, "hasLocation");
	if (LocRec.IsDef())
		ResJson->AddToObj("location", NewsSrvUtil->Location(LocRec.GetRecId()));

	if (TopConcepts > 0) {
		PRecSet ConceptRecSet = NewsBase->EventStore->GetRec(EventRecId).DoJoin(NewsBase->Base, "hasConcept");
		ConceptRecSet->SortByFq(false);
		TUInt64V ConceptSet = GetValidConcepts(ConceptRecSet, TopConcepts, ConceptTypes);
		TJsonValV ConceptV;
		for (int N = 0; N < ConceptSet.Len(); N++)
			ConceptV.Add(GetConceptInfo(ConceptSet[N], LabelLanguageV, false, false));
		ResJson->AddToObj("topConcepts", ConceptV);
	}
	PRecSet ArticleRecSet = StorySet->DoJoin(NewsBase->Base, "hasArticle");
	ArticleRecSet->SortByFq(false);
	for (int N = 0; N < ArticleRecSet->GetRecs(); N++) {
		uint64 ArticleRecID = ArticleRecSet->GetRecId(N);
		PRecSet ImageRecSet = ArticleRecSet->GetRec(N).DoJoin(NewsBase->Base, "hasImage");
		if (ImageRecSet->GetRecs() > 0) {
			ResJson->AddToObj("img", NewsBase->ImageStore->GetRecNm(ImageRecSet->GetRecId(0)));
			break;
		}
	}
	return ResJson;
}

// return the articles assigned to the story
PJsonVal TConceptDailyTrends::GetStoryArticles(const uint64& StoryRecId, const int& Page, const int& Count) const
{
	if (!NewsBase->StoryStore->IsRecId(StoryRecId))
		return TJsonVal::NewObj("error", TJsonVal::NewStr(TStr::Fmt("Unknown story id %I64u", StoryRecId)));
	PRecSet ArticleRecSet = NewsBase->StoryStore->GetRec(StoryRecId).DoJoin(NewsBase->Base, "hasArticle");
	ArticleRecSet->SortById(false);
	PRecSet SubArticleRecSet = ArticleRecSet->GetLimit(Count, Count*Page);
	return NewsSrvUtil->Articles(SubArticleRecSet, DefaultArticleBodyLen);	
}

PJsonVal TConceptDailyTrends::GetStoryArticles(const PRecSet& ArticleRecSet, const int& Page, const int& Count) const
{
	PRecSet SubArticleRecSet = ArticleRecSet->GetLimit(Count, Count*Page);
	return NewsSrvUtil->Articles(SubArticleRecSet, DefaultArticleBodyLen);
}

// return information about daily concept frequencies for concepts in ConceptIdV
PJsonVal TConceptDailyTrends::GetConceptDailyFreq(const TUInt64V& ConceptIdV, const TStrV& LabelLanguageV)
{	
	TJsonValV DateArrJsonV;
	for (int DayOff = 0; DayOff < 30; DayOff++) {
		TStr PopKey = TrendsManager->GetDayPopKey(DayOff);
		PConceptPop Pop = TrendsManager->GetPopForKey(PopKey);
		if (Pop.Empty())
			continue;
		PJsonVal DateJson = TJsonVal::NewObj("date", PopKey);
		PJsonVal ConceptFqs = TJsonVal::NewObj();
		for (int N = 0; N < ConceptIdV.Len(); N++)
			ConceptFqs->AddToObj(ConceptIdV[N].GetStr(), TJsonVal::NewNum(Pop->GetConceptPop((TConceptId) ConceptIdV[N])));
		DateJson->AddToObj("conceptFreq", ConceptFqs);
		DateArrJsonV.Add(DateJson);
	}
	// go from the oldest to newest
	DateArrJsonV.Reverse();
	PJsonVal RetJson = TJsonVal::NewObj("trends", TJsonVal::NewArr(DateArrJsonV));
	
	PRecSet ConceptRecSet = TRecSet::New(NewsBase->ConceptStore, ConceptIdV);
	RetJson->AddToObj("conceptInfo", NewsSrvUtil->Concepts(ConceptRecSet, TUInt64Set(), TInt::Mx, ctAll, LabelLanguageV));
	
	return RetJson;
}

// return the subset of the concepts that are of the correct type and are not supposed to be hidden
TUInt64V TConceptDailyTrends::GetValidConcepts(const PRecSet& ConceptRecSet, const int& Count, const EnumConceptType& ConceptTypes)  const
{
	TUInt64V ValidConceptV;
	for (int N = 0; N < ConceptRecSet->GetRecs() && ValidConceptV.Len() < Count; N++) {
		const uint64 ConceptId = ConceptRecSet->GetRecId(N);
		// was the concept hidden by the admin?
		if (IsHidden((TConceptId) ConceptId))
			continue;
		// do we want to ignore concepts of this type?
		const TStr Type = NewsBase->ConceptStore->GetFieldNmStr(ConceptId, "Type");
		if (!NewsConst.IsValidType(Type, ConceptTypes))
			continue;
		ValidConceptV.Add(ConceptId);
	}
	return ValidConceptV;
}

// get non-duplicated articles for top trends
PRecSet TConceptDailyTrends::GetNonDuplArticles(const PRecSet& ArtSet, const int& Page, const int& Count) const 
{ 
	TUInt64V NonDupArtV;
	const int Arts = ArtSet->GetRecs(); 
	int ToIgnore = Count * max(0, Page);		// count non-dups found
	for (int N=0; N < Arts; N++) {
		if (NonDupArtV.Len() >= Count)
			break;
		uint64 ArtId = ArtSet->GetRecId(N);
		TRec OrigArt = NewsBase->ArticleStore->GetRec(ArtId).DoSingleJoin(NewsBase->Base, "isDuplicateOf");
		if (! OrigArt.IsDef()) {
			if (ToIgnore <= 0)
				NonDupArtV.Add(ArtId);
			else 
				ToIgnore--;
		}
	}
	PRecSet SubArticleRecSet = TRecSet::New(NewsBase->ArticleStore, NonDupArtV);
	return SubArticleRecSet;
}

//
// methods for settings the featured entities and events
// 
void TConceptDailyTrends::SetFeaturedEventsAndConcepts(const TVec<TPair<TUInt64, TUInt64> > & _EventIdConceptIdPrV)
{
	EventIdConceptIdPrV = TVec<TPair<TUInt64, TUInt64> >(_EventIdConceptIdPrV);
}

PJsonVal TConceptDailyTrends::GetFeaturedEventsAndConcepts()
{
	TStrV LabelLangV;
	LabelLangV.Add(LabelLang);
	EnumConceptInfoFlags ConceptFlags = (EnumConceptInfoFlags) (picIncludePrimImage);

	TJsonValV FeaturedJsonV;
	for (int N = 0; N < EventIdConceptIdPrV.Len(); N++) {
		uint64 EventId = EventIdConceptIdPrV[N].Val1;
		const uint64 ConceptId = EventIdConceptIdPrV[N].Val2;
		
		if (!NewsBase->EventStore->IsRecId(EventId))
			continue;
		
		PRecSet StorySet = NewsBase->EventStore->GetRec(EventId).DoJoin(NewsBase->Base, "hasStory");
		// if no story replace EventId with a valid event id
		if (StorySet->GetRecs() == 0) {
			TRec PrimStoryRec = NewsBase->EventStore->GetRec(EventId).DoSingleJoin(NewsBase->Base, "hasPrimaryStory");
			TRec CurrEventRec = PrimStoryRec.DoSingleJoin(NewsBase->Base, "hasEvent");
			EventId = CurrEventRec.GetRecId();
		}
		PJsonVal FeaturedItem = TJsonVal::NewObj();
		FeaturedItem->AddToObj("event", NewsSrvUtil->Event(EventId));
		
		if (NewsBase->ConceptStore->IsRecId(ConceptId))
			FeaturedItem->AddToObj("concept", NewsSrvUtil->Concept(ConceptId, -1, LabelLangV, ConceptFlags));
		
		FeaturedJsonV.Add(FeaturedItem);
	}
	return TJsonVal::NewArr(FeaturedJsonV);
}
