/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ConceptTrendsManager.h"
#include "../NewsMiner/NewsConst.h"

// ////
// TConceptPop
// ////
TConceptPop::TConceptPop()
{
	//Type = _Type;
	TotalCount = 0;
}

TConceptPop::TConceptPop(TSIn& SIn)
{
	//TCh TypeCh(SIn);
	//Type = (EnumConceptPopType) (char) TypeCh;
	TotalCount.Load(SIn);
	ConceptCountV.Load(SIn);
}

PConceptPop TConceptPop::Load(TSIn& SIn)
{
	return new TConceptPop(SIn);
}

PConceptPop TConceptPop::Copy(const PConceptPop& Pop)
{
	PConceptPop CopyPop = TConceptPop::New();
	CopyPop->TotalCount = Pop->TotalCount;
	CopyPop->ConceptCountV = TIntV(Pop->ConceptCountV);
	return CopyPop;
}

void TConceptPop::Save(TSOut& SOut) const
{
	//TCh TypeCh = (char) Type;
	//TypeCh.Save(SOut);
	TotalCount.Save(SOut);
	ConceptCountV.Save(SOut);
}

void TConceptPop::Resize(const int NewSize)
{
	if (NewSize < Len())
		return;
	int OldLen = ConceptCountV.Len();
	ConceptCountV.Reserve(NewSize, NewSize);
	int NewLen = ConceptCountV.Len();
	for (int N = OldLen; N < NewLen; N++)
		ConceptCountV[N] = 0;
}

// increase concept counts for a given set of concepts
// note: this assumes that these concepts come from the same article
void TConceptPop::AddConceptV(const TConceptIdV& ConceptIdV)
{
	// increase the article count
	TotalCount++; 

	// resize the vector if needed
	TConceptId Mx = 0;
	for (int N = 0; N < ConceptIdV.Len(); N++) 
		Mx = max(Mx, ConceptIdV[N]);
	if (ConceptCountV.Len() <= Mx) 
		Resize(Mx + 50);
	for (int N=0; N < ConceptIdV.Len(); N++)
		ConceptCountV[ConceptIdV[N]]++;
}

// decrease concept counts for a given set of concepts. called if an article is found to be a duplicate
// note: this assumes that these concepts come from the same article
void TConceptPop::RemoveConceptV(const TConceptIdV& ConceptIdV)
{
	// decrease the article count
	if (TotalCount > 0)
		TotalCount--; 

	for (int N=0; N < ConceptIdV.Len(); N++) {
		if (ConceptIdV[N] < ConceptCountV.Len())
			ConceptCountV[ConceptIdV[N]]--;
	}
}

// increase the values using the pop pop
void TConceptPop::AddConceptPop(const PConceptPop& Pop)
{
	// do we need to resize
	if (Len() < Pop->Len())
		Resize(Pop->Len());
	int Count = Pop->Len();
	TotalCount += Pop->TotalCount;
	for (int N=0; N < Count; N++) 
		ConceptCountV[N] += Pop->ConceptCountV[N];
}

// ////
// TConceptTrends
// ////
TConceptTrends::TConceptTrends()
{
	TestPop = TConceptPop::New();
	NullPop = TConceptPop::New();
}

TConceptTrends::TConceptTrends(const PConceptPop& _TestPop, const PConceptPop& _NullPop, const int IgnoreTestConceptsBelowFq, const int IgnoreNullConceptsBelowFq, const bool SortByFq, const EnumTrendMeasure Measure)
{
	TestPop = _TestPop;
	NullPop = _NullPop;
	
	if (TestPop.Empty() || NullPop.Empty())
		return;

	TUInt64 TestTot = TestPop->TotalCount;
	TUInt64 NullTot = NullPop->TotalCount;
	int MinInd = min(TestPop->ConceptCountV.Len(), NullPop->ConceptCountV.Len());
	for (int N=0; N < MinInd; N++) {
		TInt NullFq = NullPop->ConceptCountV[N], TestFq = TestPop->ConceptCountV[N];
		if (NullFq < IgnoreNullConceptsBelowFq || TestFq < IgnoreTestConceptsBelowFq)
			continue;
		if (NullPop->ConceptCountV[N] == 0)
			continue;
		double Observed = TestFq;
		double Expected = TestTot * NullFq / (double) NullTot;
		
		TFlt Val = 0;
		if (Measure == tmChiSquared)
			Val = (Observed - Expected) * (Observed - Expected) / (double) Expected;
		else if (Measure == tmPearsonResidual)
			Val = (Observed - Expected) / sqrt((float) Expected);
		TrendH.AddDat(N, Val);
	}
	if (SortByFq)
		TrendH.SortByDat(false);
}

PJsonVal TConceptTrends::GetConceptJson(const int& ConceptN) const
{
	if (ConceptN >= TrendH.Len())
		return TJsonVal::NewObj("error", "index out of range");
	TConceptId Id = TrendH.GetKey(ConceptN);
	PJsonVal Item = TJsonVal::NewObj("id", Id);
	Item->AddToObj("chi", TrendH[ConceptN]);
	Item->AddToObj("nullCount", NullPop->ConceptCountV[Id]);
	Item->AddToObj("testCount", TestPop->ConceptCountV[Id]);
	return Item;
}

PJsonVal TConceptTrends::GetJson(const int& Count) const
{
	PJsonVal Json = TJsonVal::NewObj();
	Json->AddToObj("testTotal", (int) TestPop->TotalCount);
	Json->AddToObj("nullTotal", (int) NullPop->TotalCount);
	PJsonVal Arr = TJsonVal::NewArr();
	Json->AddToObj("trends", Arr);
	for (int N=0; N < Count && N < TrendH.Len(); N++) {
		TConceptId Id = TrendH.GetKey(N);
		PJsonVal Item = TJsonVal::NewObj("id", Id);
		Item->AddToObj("chi", TrendH[N]);
		Item->AddToObj("nullCount", NullPop->ConceptCountV[Id]);
		Item->AddToObj("testCount", TestPop->ConceptCountV[Id]);
		Arr->AddToArr(Item);
	}
	return Json;
}

// ////
// TConceptTrendsManager
// ////
TStr TotalKey("Total");

TConceptTrendsManager::TConceptTrendsManager()
{
	Init();
}

TConceptTrendsManager::TConceptTrendsManager(const TStr& FileName, const TStr& _SettingsFileName)
{
	StoreFileName = FileName;
	SettingsFileName = _SettingsFileName;

	ChangeCount = 0;
	
	LoadSettings();
}

TConceptTrendsManager::~TConceptTrendsManager()
{
	Save();
}

bool TConceptTrendsManager::Load()
{
	// init first
	Init();
	
	if (TFile::Exists(StoreFileName)) {
		try {
			TNotify::StdNotify->OnStatus("Loading concept trends manager...");
			// load existing data	
			TFIn FIn(StoreFileName);
			MinConceptWgt.Load(FIn);
			KeepTotalPop.Load(FIn);
			KeepNrYears.Load(FIn);
			KeepNrMonths.Load(FIn);
			KeepNrDays.Load(FIn);

			TotalConceptPopH = TConceptPop::New(FIn);
			ConceptPopDH.Load(FIn);
			ConceptPopMH.Load(FIn);
			ConceptPopYH.Load(FIn);
			TNotify::StdNotify->OnStatus("Concept trends loaded.");
			return true;
		}
		catch (...) {
			Init();
			TNotify::StdNotify->OnStatus("Exception while loading concept trends file.");
			return false;
		}
	}
	else {
		TNotify::StdNotify->OnStatus("Concept trends file was missing...");
		return false;
	}
}

void TConceptTrendsManager::LoadSettings()
{
	const TStr SettingsStr = TFile::Exists(SettingsFileName) ? TStr::LoadTxt(SettingsFileName) : "{}";
	const PJsonVal AllSettings = TJsonVal::GetValFromStr(SettingsStr);
	const PJsonVal TrendsSettings = AllSettings->IsObjKey("trendsManager") ? AllSettings->GetObjKey("trendsManager") : TJsonVal::NewObj();

	MinConceptWgt = (int) TrendsSettings->GetObjNum("MinConceptWgt", 1);

	SetKeepNrDays((int) TrendsSettings->GetObjNum("KeepNrDays", 30));
	SetKeepNrMonths((int) TrendsSettings->GetObjNum("KeepNrMonths", 3));
	SetKeepNrYears((int) TrendsSettings->GetObjNum("KeepNrYears", 1));
	SetKeepTotal((int) TrendsSettings->GetObjBool("KeepTotal", true));
}

// clear all data from the trends manager
void TConceptTrendsManager::Init()
{
	MinConceptWgt = 0;
	KeepTotalPop = true;
	KeepNrYears = 1;
	KeepNrMonths = 3;
	KeepNrDays = 30;

	ConceptPopYH.Clr();
	ConceptPopMH.Clr();
	ConceptPopDH.Clr();
	
	TotalConceptPopH = TConceptPop::New();
}

void TConceptTrendsManager::Save()
{
	if (StoreFileName == "")
		return;
	
	TFOut FOut(StoreFileName);
	MinConceptWgt.Save(FOut);
	KeepTotalPop.Save(FOut);
	KeepNrYears.Save(FOut);
	KeepNrMonths.Save(FOut);
	KeepNrDays.Save(FOut);

	GarbageCollect();
	TotalConceptPopH->Save(FOut);
	
	ConceptPopDH.Save(FOut);
	ConceptPopMH.Save(FOut);
	ConceptPopYH.Save(FOut);
}

void TConceptTrendsManager::AddConceptV(const TTm& TimeStamp, const TConceptIdV& ConceptIdV)
{
	ChangeCount++;
	if (ChangeCount % 1000)
		GarbageCollect();
	
	if (KeepTotalPop)
		TotalConceptPopH->AddConceptV(ConceptIdV);
	if (KeepNrDays > 0) {
		TStr Key = TimeStamp.GetYMDDashStr();
		AddConceptsForKey(ConceptPopDH, Key, ConceptIdV);
	}
	if (KeepNrMonths > 0) {
		TStr Key = TStr::Fmt("%4d-%02d", TimeStamp.GetYear(), TimeStamp.GetMonth());
		AddConceptsForKey(ConceptPopMH, Key, ConceptIdV);
	}
	if (KeepNrYears > 0) {
		TStr Key = TInt(TimeStamp.GetYear()).GetStr();
		AddConceptsForKey(ConceptPopYH, Key, ConceptIdV);
	}
}

void TConceptTrendsManager::RemoveConceptV(const TTm& TimeStamp, const TConceptIdV& ConceptIdV)
{
	if (KeepTotalPop)
		TotalConceptPopH->RemoveConceptV(ConceptIdV);
	if (KeepNrDays > 0) {
		TStr Key = TimeStamp.GetYMDDashStr();
		PConceptPop Pop = GetPopForKey(Key);
		if (! Pop.Empty())
			Pop->RemoveConceptV(ConceptIdV);
	}
	if (KeepNrMonths > 0) {
		TStr Key = TStr::Fmt("%4d-%02d", TimeStamp.GetYear(), TimeStamp.GetMonth());
		PConceptPop Pop = GetPopForKey(Key);
		if (! Pop.Empty())
			Pop->RemoveConceptV(ConceptIdV);
	}
	if (KeepNrYears > 0) {
		TStr Key = TInt(TimeStamp.GetYear()).GetStr();
		PConceptPop Pop = GetPopForKey(Key);
		if (! Pop.Empty())
			Pop->RemoveConceptV(ConceptIdV);
	}
}

void TConceptTrendsManager::AddConceptsForKey(TConceptPopGroup& ConceptH, const TStr& Key, const TConceptIdV& ConceptIdV)
{
	if (! ConceptH.IsKey(Key))
		ConceptH.AddDat(Key, TConceptPop::New());
	ConceptH.GetDat(Key)->AddConceptV(ConceptIdV);
}

void TConceptTrendsManager::GarbageCollect()
{
	if (! KeepTotalPop)
		TotalConceptPopH->Clr();
	TStrV YearV = GetYearKeys();
	TStrV MonthV = GetMonthKeys();
	TStrV DayV = GetDayKeys();
	YearV.Sort(false);
	MonthV.Sort(false);
	DayV.Sort(false);
	for (int N = KeepNrYears; N < YearV.Len(); N++)
		RemoveKey(ConceptPopYH, YearV[N]);
	for (int N = KeepNrMonths; N < MonthV.Len(); N++)
		RemoveKey(ConceptPopMH, MonthV[N]);
	for (int N = KeepNrDays; N < DayV.Len(); N++)
		RemoveKey(ConceptPopDH, DayV[N]);
}

bool TConceptTrendsManager::IsPopKey(const TStr& PopKey)
{
	if (PopKey == TotalKey && KeepTotalPop) return true;
	else if (ConceptPopYH.IsKey(PopKey)) return true;
	else if (ConceptPopMH.IsKey(PopKey)) return true;
	else if (ConceptPopDH.IsKey(PopKey)) return true;
	else
		return false;
}

// return an existing population
PConceptPop TConceptTrendsManager::GetPopForKey(const TStr& PopKey)
{
	if (PopKey == TotalKey) return TotalConceptPopH;
	else if (ConceptPopYH.IsKey(PopKey)) return ConceptPopYH.GetDat(PopKey);
	else if (ConceptPopMH.IsKey(PopKey)) return ConceptPopMH.GetDat(PopKey);
	else if (ConceptPopDH.IsKey(PopKey)) return ConceptPopDH.GetDat(PopKey);
	else
		return NULL;
}

PConceptPop TConceptTrendsManager::GetPopForKeys(const TStrV& PopKeys)
{
	if (PopKeys.Len() == 0)
		return NULL;
	if (PopKeys.Len() == 1)
		return GetPopForKey(PopKeys[0]);
	
	PConceptPop MergedPop = TConceptPop::Copy(GetPopForKey(PopKeys[0]));
	for (int N=1; N < PopKeys.Len(); N++)
		MergedPop->AddConceptPop(GetPopForKey(PopKeys[N]));
	return MergedPop;
}

PConceptPop TConceptTrendsManager::GetPopForOff(const TStr& OffDesc)
{
	TStrV KeyV = ParsePopKeys(OffDesc);
	return GetPopForKeys(KeyV);
}

// for a given day offset (0, 1, ....) return key such as 2013-11-03
TStr TConceptTrendsManager::GetDayPopKey(const int DayOffset)
{
	TStrV DayV = GetDayKeys();
	DayV.Sort(false);
	return (DayOffset < DayV.Len()) ? DayV[DayOffset] : "";
}

TStr TConceptTrendsManager::GetMonthPopKey(const int MonthOffset)
{
	TStrV MonthV = GetMonthKeys();
	MonthV.Sort(false);
	return (MonthOffset < MonthV.Len()) ? MonthV[MonthOffset] : "";
}

TStr TConceptTrendsManager::GetYearPopKey(const int YearOffset)
{
	TStrV YearV = GetYearKeys();
	YearV.Sort(false);
	return (YearOffset < YearV.Len()) ? YearV[YearOffset] : "";
}

TStr TConceptTrendsManager::GetTotalPopKey()
{
	return KeepTotalPop ? TStr(TotalKey) : "";
}

PConceptTrends TConceptTrendsManager::ComparePops(const TStr& TestPopKey, const TStr& NullPopKey, const int IgnoreTestConceptsBelowFq, const int IgnoreNullConceptsBelowFq, bool SortByFq, const EnumTrendMeasure Measure)
{
	if (! IsPopKey(TestPopKey) || ! IsPopKey(NullPopKey))
		return NULL;
	PConceptPop TestPop = GetPopForKey(TestPopKey);
	PConceptPop NullPop = GetPopForKey(NullPopKey);
	return TConceptTrends::New(TestPop, NullPop, IgnoreTestConceptsBelowFq, IgnoreNullConceptsBelowFq, SortByFq, Measure);
}

PConceptTrends TConceptTrendsManager::ComparePops(const PConceptPop& TestPop, const PConceptPop& NullPop, const int IgnoreTestConceptsBelowFq, const int IgnoreNullConceptsBelowFq, bool SortByFq, const EnumTrendMeasure Measure)
{
	return TConceptTrends::New(TestPop, NullPop, IgnoreTestConceptsBelowFq, IgnoreNullConceptsBelowFq, SortByFq, Measure);
}

// get pop key from description like "Total,D0,M3,Y1,2014-03-01"
TStrV TConceptTrendsManager::ParsePopKeys(const TStr& OffDescOrKeyList)
{
	TStrV DescV; OffDescOrKeyList.SplitOnAllCh(',', DescV);
	TStrV KeyV;
	for (int N = 0; N < DescV.Len(); N++) {
		TStr Off = DescV[N].GetTrunc();
		if (Off.GetLc() == TotalKey.GetLc())
			KeyV.Add(GetTotalPopKey());
		else if (Off.GetSubStr(0, 4).IsInt())	// if the Off is actually a key, which should start with YYYY
			KeyV.Add(Off);
		else {
			// type should be Y, M, D
			char Type = Off.GetCap().GetCh(0);
			int Offset = Off.GetSubStr(1).GetInt(-1);
			if (Offset < 0)
				continue;
			if (Type == 'Y')
				KeyV.Add(GetYearPopKey(Offset));
			else if (Type == 'M')
				KeyV.Add(GetMonthPopKey(Offset));
			else if (Type == 'D')
				KeyV.Add(GetDayPopKey(Offset));
		}
	}
	// keep only the valid keys
	for (int N = KeyV.Len() - 1; N >= 0; N--) {
		if (!IsPopKey(KeyV[N]))
			KeyV.Del(N);
	}
	return KeyV;
}