#ifndef CONCEPTTRENDSMANAGER_H
#define CONCEPTTRENDSMANAGER_H

#include <base.h>

class TConceptPopComp;
class TConceptPop;
class TConceptTrends;
typedef TPt<TConceptPop> PConceptPop;
typedef TVec<PConceptPop> TConceptPopV;
typedef TPt<TConceptTrends> PConceptTrends;
typedef TVec<PConceptTrends> TConceptTrendsV;

//enum EnumConceptPopType { Yearly, Monthly, Daily };
typedef TInt TConceptId;
typedef TIntV TConceptIdV;
typedef THash<TStr, PConceptPop> TConceptPopGroup;

// /////
// CONCEPT POPULATION
// 
// class for keeping the number of concepts in the articles for a specific time period
class TConceptPop { 
public:
	friend class TPt<TConceptPop>;

private:
	TCRef CRef;

	//EnumConceptPopType Type;
	TUInt64 TotalCount;
	TIntV ConceptCountV;
	
	void Resize(const int NewSize);
	TConceptPop();
	static PConceptPop Load(TSIn& SIn);

public:
	/*TConceptPop(EnumConceptPopType _Type);*/
	static PConceptPop New() { return new TConceptPop(); }
	static PConceptPop New(TSIn& FIn) { return new TConceptPop(FIn);}

	TConceptPop(TSIn& SIn);
	void Save(TSOut& SOut) const;
	
	static PConceptPop Copy(const PConceptPop& Pop);	// create a copy

	// add values from another population (used when aggregating multiple populations into one)
	void AddConceptPop(const PConceptPop& Pop);
	// increase concept counts for concepts in ConceptIdV
	void AddConceptV(const TConceptIdV& ConceptIdV);
	// decrease concept counts for concepts in ConceptIdV
	void RemoveConceptV(const TConceptIdV& ConceptIdV);
	// get concept frequency for concept ConceptId
	int GetConceptPop(const TConceptId ConceptId) { return ConceptId < ConceptCountV.Len() ? ConceptCountV[ConceptId].Val : 0; }
	int GetTotalCount() { return (int) TotalCount; }

	void Clr() { ConceptCountV.Clr(); TotalCount = 0; }
	int Len() const { return ConceptCountV.Len(); }

friend class TConceptTrends;
};


enum EnumTrendMeasure { tmChiSquared, tmPearsonResidual };
// /////
// CONCEPT TRENDS
// 
// find concept trends
// compare two concept populations and return the ones with the highest chi squares
class TConceptTrends{ 
public:
	friend class TPt<TConceptTrends>;

private:
	TCRef CRef;

	PConceptPop TestPop;
	PConceptPop NullPop;
	
	THash<TConceptId, TFlt> TrendH;
	TConceptTrends(const PConceptPop& _TestPop, const PConceptPop& _NullPop, const int IgnoreTestConceptsBelowFq, const int IgnoreNullConceptsBelowFq, const bool SortByFq, const EnumTrendMeasure Measure);
	TConceptTrends();

public:
	static PConceptTrends New() {
		return new TConceptTrends();
	}
	static PConceptTrends New(const PConceptPop& _TestPop, const PConceptPop& _NullPop, const int IgnoreTestConceptsBelowFq = 10, const int IgnoreNullConceptsBelowFq = 10, const bool SortByFq = true, const EnumTrendMeasure Measure = tmPearsonResidual) {
		return new TConceptTrends(_TestPop, _NullPop, IgnoreTestConceptsBelowFq, IgnoreNullConceptsBelowFq, SortByFq, Measure);
	}

	int Len() const { return TrendH.Len(); }
	// get concept id and chi squared for concepts with highest chi value
	TConceptId GetTopConceptId(const int ConceptN) const { Assert(ConceptN < TrendH.Len()); return TrendH.GetKey(ConceptN); }
	double GetTopConceptChi(const int ConceptN) const { Assert(ConceptN < TrendH.Len()); return TrendH.GetDat(ConceptN); }

	bool IsConceptId(const TConceptId ConceptId) { return TrendH.IsKey(ConceptId); }
	float GetChiForConcept(const TConceptId ConceptId, const float& DefVal = 0) { return TrendH.IsKey(ConceptId) ? (float) TrendH.GetDat(ConceptId).Val: DefVal; }

	PJsonVal GetConceptJson(const int& ConceptN) const;
	PJsonVal GetJson(const int& Count) const;
	
	const PConceptPop GetNullPop() { return NullPop; }
	const PConceptPop GetTestPop() { return TestPop; }
};

// /////
// CONCEPT POPULATION MANAGER
// 
// manage concept populations on daily, monthly, yearly basis
class TConceptTrendsManager;
typedef TPt<TConceptTrendsManager> PConceptTrendsManager;
class TConceptTrendsManager{ 
public:
	friend class TPt<TConceptTrendsManager>;

private:
	TCRef CRef;

	TStr StoreFileName;
	TStr SettingsFileName;

	TBool KeepTotalPop;		// do we keep monitoring total number of concepts
	TInt KeepNrYears;		// number of last years to keep - 0 for none
	TInt KeepNrMonths;		// number of last months to keep - 0 for none
	TInt KeepNrDays;		// number of last days to keep - 0 for none

	int ChangeCount;		// the number of articles that were added - used to know when to do garbage collection
		
	TConceptPopGroup ConceptPopYH;
	TConceptPopGroup ConceptPopMH;
	TConceptPopGroup ConceptPopDH;
	PConceptPop TotalConceptPopH;
	
	void GarbageCollect();
	void AddConceptsForKey(TConceptPopGroup& ConceptPopH, const TStr& Key, const TConceptIdV& ConceptIdV);
	void RemoveKey(TConceptPopGroup& ConceptPopH, const TStr& Key) { if (ConceptPopH.IsKey(Key)) ConceptPopH.DelKey(Key); }

	TConceptTrendsManager();
	TConceptTrendsManager(const TStr& FileName, const TStr& SettingsFileName);

public:
	// concepts that have wgt below this val should be ignored - not included in ConceptIdV when called AddConceptV
	// this setting is not automatically taken into account - you should use it when preparing the ConceptIdV for AddConceptV
	TInt MinConceptWgt;

	static PConceptTrendsManager New(const TStr& FileName, const TStr& SettingsFileName) { return new TConceptTrendsManager(FileName, SettingsFileName); }
	~TConceptTrendsManager();
	void LoadSettings();
	void Init();
	bool Load();
	void Save();

	// set number of years we wish to monitor
	void SetKeepNrYears(const int NewVal) { KeepNrYears = NewVal; }
	// set number of months we wish to monitor
	void SetKeepNrMonths(const int NewVal) { KeepNrMonths = NewVal; }
	// set number of days we wish to monitor
	void SetKeepNrDays(const int NewVal) { KeepNrDays = NewVal; }
	// do we wish to monitor the total number of articles per concepts or not
	void SetKeepTotal(const bool NewVal) { KeepTotalPop = NewVal; if (NewVal == false) TotalConceptPopH->Clr(); }

	// add concepts ConceptIdV from article with TimeStamp
	void AddConceptV(const TTm& TimeStamp, const TConceptIdV& ConceptIdV);
	// remove concepts ConceptIdV from article with TimeStamp
	void RemoveConceptV(const TTm& TimeStamp, const TConceptIdV& ConceptIdV);
	
	bool IsManagingTotal() { return KeepTotalPop; }
	bool IsManagingYear(const int Year) { return ConceptPopYH.IsKey(TStr::Fmt("%04d", Year)); }
	bool IsManagingMonth(const int Year, const int Month) { return ConceptPopMH.IsKey(TStr::Fmt("%04d-%02d", Year, Month)); }
	bool IsManagingDay(const int Year, const int Month, const int Day) { return ConceptPopDH.IsKey(TStr::Fmt("%04d-%02d-%02d", Year, Month, Day)); }

	int GetTotalConceptFreq(const TConceptId ConceptId) { return TotalConceptPopH->GetConceptPop(ConceptId); }
	
	TStrV GetDayKeys() { TStrV KeyV; ConceptPopDH.GetKeyV(KeyV); return KeyV; }
	TStrV GetMonthKeys() { TStrV KeyV; ConceptPopMH.GetKeyV(KeyV); return KeyV; }
	TStrV GetYearKeys() { TStrV KeyV; ConceptPopYH.GetKeyV(KeyV); return KeyV; }

	// get key for managing concept population
	// if you wish to get concept population for 1 day ago call GetDayPopKey(1)
	TStr GetDayPopKey(const int DayOffset);
	TStr GetMonthPopKey(const int MonthOffset);
	TStr GetYearPopKey(const int YearOffset);
	TStr GetTotalPopKey();
	
	// OffsetDesc can be like T, D0 (0 day offset), M1, Y2 or a comma separated list of them (like D0,D1,D2)
	TStrV GetPopKeysForOffDesc(const TStrV& OffDesc);
	TStrV ParsePopKeys(const TStr& OffDesc);
	
	// key is like 2013 or 2013-01 or 2013-01-12
	bool IsPopKey(const TStr& PopKey);
	PConceptPop GetPopForKey(const TStr& PopKey);
	PConceptPop GetPopForKeys(const TStrV& PopKeys);

	PConceptPop GetPopForOff(const TStr& OffDesc);

	PConceptTrends ComparePops(const TStr& TestPopKey, const TStr& NullPopKey, const int IgnoreTestConceptsBelowFq = 10, const int IgnoreNullConceptsBelowFq = 10, bool SortByFq = true, const EnumTrendMeasure Measure = tmPearsonResidual);
	PConceptTrends ComparePops(const PConceptPop& TestPop, const PConceptPop& NullPop, const int IgnoreTestConceptsBelowFq = 10, const int IgnoreNullConceptsBelowFq = 10, bool SortByFq = true, const EnumTrendMeasure Measure = tmPearsonResidual);
};


#endif