/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
//#include "stdafx.h"
#include "DateExtractor.h"
#include <base.h>
#include <string>


class TCmpExtractedDate {
private:
  bool IsAsc;
public:
	TCmpExtractedDate(const bool& AscSort=true) : IsAsc(AscSort) { }
	bool operator () (const TExtractedDate& P1, const TExtractedDate& P2) const 
	{
		if (IsAsc) { return P1.GetFreq() < P2.GetFreq(); } else { return P2.GetFreq() < P1.GetFreq(); }
	}
};

// //////
// TExtractedDate
// //////
TExtractedDate::TExtractedDate(PJsonVal DateJson)
{
	Ambiguous = DateJson->GetObjBool("amb", true);
	ImputedYear = DateJson->GetObjBool("imp", false);
	Freq = (int) DateJson->GetObjNum("freq", 1);
	Date = DateJson->GetObjStr("date", "");
	BIsDateRange = DateJson->IsObjKey("dateEnd");
	PosInText = (int) DateJson->GetObjNum("posInText", -1);
	if (BIsDateRange)
		DateEnd = DateJson->GetObjStr("dateEnd");
	DetectedDate = DateJson->GetObjStr("detectedDate", "");
	TextSnippet = DateJson->GetObjStr("textSnippet", "");
}

PJsonVal TExtractedDate::GetJson(const bool& IncludeDetectedDateStr, const bool& IncludeTextSnippet) 
{
	PJsonVal DateJson = TJsonVal::NewObj("amb", Ambiguous);
	DateJson->AddToObj("imp", ImputedYear);
	DateJson->AddToObj("date", Date);
	DateJson->AddToObj("posInText", PosInText);
	if (Freq > 1)
		DateJson->AddToObj("freq", Freq);
	// date range props
	if (BIsDateRange)
		DateJson->AddToObj("dateEnd", DateEnd);
	if (IncludeDetectedDateStr) 
		DateJson->AddToObj("detectedDate", DetectedDate);
	if (IncludeTextSnippet) 
		DateJson->AddToObj("textSnippet", TextSnippet);
	return DateJson;
}


// //////
// TExtractedDateColl
// //////
TExtractedDateColl::TExtractedDateColl(const TStr& EncodedDateColl)
{
	PJsonVal ArrayJson = TJsonVal::GetValFromStr(EncodedDateColl);
	if (!ArrayJson->IsArr())
		return;
	int Len = ArrayJson->GetArrVals();
	for (int N=0; N < Len; N++) {
		PJsonVal DateJson = ArrayJson->GetArrVal(N);
		ExtractedDateV.Add(TExtractedDate(DateJson));
	}
}

// get extracted date object for given date in str (has to be exact match)
bool TExtractedDateColl::GetExtractedDate(const TStr& Date, TExtractedDate& ExtractedDate)
{
	for (int N = 0; N < ExtractedDateV.Len(); N++) {
		if (Date == ExtractedDateV[N].GetDate()) {
			ExtractedDate = ExtractedDateV[N];
			return true;
		}
	}
	return false;
}

// find the index in ExtractedDateV where we already have TExtracredDate with this date
int TExtractedDateColl::GetExtractedDatePos(const TExtractedDate& ExtractedDate)
{
	for (int N = 0; N < ExtractedDateV.Len(); N++) {
		if (ExtractedDate == ExtractedDateV[N])
			return N;
	}
	return -1;
}

void TExtractedDateColl::AddExtractedDate(const TExtractedDate& Date)
{
	TExtractedDate ExtractedDate;
	int Pos = GetExtractedDatePos(Date);
	if (Pos != -1) 
		ExtractedDateV[Pos].IncreaseFreq(Date.GetFreq());
	else 
		// first time we see the date. add it to the vector
		ExtractedDateV.Add(Date);
}

void TExtractedDateColl::AddColl(const TExtractedDateColl& Coll)
{
	int Dates = Coll.Len();
	for (int N=0; N < Dates; N++) {
		AddExtractedDate(Coll.GetExtractedDate(N));
	}
}

bool TExtractedDateColl::GetMostFreqDate(TExtractedDate& ExtractedDate, const int& MinFreq, const bool& IgnoreIncomplete)
{
	SortByFreq(false);
	for (int N = 0; N < Len(); N++) {
		const TExtractedDate& Date = ExtractedDateV[N];
		if (Date.GetFreq() < MinFreq)
			break;
		// if the date is not complete then ignore it as a valid event date
		if (IgnoreIncomplete && !Date.IsDateComplete())
			continue;
		ExtractedDate = Date;
		return true;
	}
	return false;
}

void TExtractedDateColl::FilterDatesByPosInText(const int& MinPos, const int& MaxPos)
{
	for (int N = Len()-1; N >= 0; N--) {
		const TExtractedDate& Date = ExtractedDateV[N];
		const int Pos = Date.GetPosInText();
		if (Pos == -1 || Pos < MinPos || Pos > MaxPos)
			ExtractedDateV.Del(N);
	}
}

// some TExtractedDate instances can have a missing year (e.g. -MM-DD).
// ImputeMissingYear checks if we can find other instances of TEXtractedDate that have YYYY-MM-DD set (with same MM-DD part)
// if found, then Freq is moved from -MM-DD to YYYY-MM-DD and the -MM-DD instance is removed
// if multiple YYYY-MM-DD are found with the same -MM-DD, the one with the largest Freq is used as the target for merging
// imputing works for TExtractedDate that represent single date as well as date range
void TExtractedDateColl::ImputeMissingYears(const int& MinFreq, PNotify Notify)
{
	// merge dates
	int i=0; 
	while (i < ExtractedDateV.Len()) {
		if (ExtractedDateV[i].IsDateRange()) { 
			i++; continue; 
		} 
		// we don't try to merge the dates with known year
		if (ExtractedDateV[i].DateHasValidYear()) {
			i++; continue; 
		}
		int BestMatchInd = -1;
		// find candidates with valid year
		for (int j=0; j < ExtractedDateV.Len(); j++) {
			if (ExtractedDateV[j].IsDateRange()) continue;
			if (i==j) continue;
			// if the MM-DD part is the same then remember it as BestMatchInd
			if (ExtractedDateV[j].GetFreq() >= MinFreq &&
				ExtractedDateV[j].DateHasValidYear() &&							 // has to have valid year
				ExtractedDateV[j].GetDate().SearchStr(ExtractedDateV[i].GetDate()) != -1) {    // has to contain the same MM-DD part
				if (BestMatchInd == -1 || ExtractedDateV[j].GetFreq() > ExtractedDateV[BestMatchInd].GetFreq())
					BestMatchInd = j;
			}
		}
		if (BestMatchInd != -1) {
			if (!Notify.Empty())
				Notify->OnNotifyFmt(ntInfo, "Imputing %s (%d articles) into %s (%d articles)", ExtractedDateV[i].GetDate().CStr(), ExtractedDateV[i].GetFreq(), ExtractedDateV[BestMatchInd].GetDate().CStr(), ExtractedDateV[BestMatchInd].GetFreq());
			ExtractedDateV[BestMatchInd].AddFreq(ExtractedDateV[i].GetFreq());
			ExtractedDateV.Del(i);
		}
		else
			i++;
	}
	// merge date ranges
	i=0; 
	while (i < ExtractedDateV.Len()) {
		if (! ExtractedDateV[i].IsDateRange()) { 
			i++; continue; 
		} 
		// we don't try to merge the dates with known year
		if (ExtractedDateV[i].DateHasValidYear() && ExtractedDateV[i].DateEndHasValidYear()) {
			i++; continue; 
		} 
		int BestMatchInd = -1;
		// find candidates with valid year
		for (int j=0; j < ExtractedDateV.Len(); j++) {
			if (! ExtractedDateV[j].IsDateRange()) continue;
			if (i==j) continue;
			// if the MM-DD part is the same then remember it as BestMatchInd
			if (ExtractedDateV[j].DateHasValidYear() && ExtractedDateV[j].DateEndHasValidYear() &&  // has to have valid year
				ExtractedDateV[j].GetDate().SearchStr(ExtractedDateV[i].GetDate()) != -1 && 
				ExtractedDateV[j].GetDateEnd().SearchStr(ExtractedDateV[i].GetDateEnd()) != -1) {    // has to contain the same MM-DD part
				if (BestMatchInd == -1 || ExtractedDateV[j].GetFreq() > ExtractedDateV[BestMatchInd].GetFreq())
					BestMatchInd = j;
			}
		}
		if (BestMatchInd != -1) {
			if (!Notify.Empty())
				Notify->OnStatusFmt("Imputing (%s, %s) (%d articles) into (%s, %s) (%d articles)", ExtractedDateV[i].GetDate().CStr(), ExtractedDateV[i].GetDateEnd().CStr(), ExtractedDateV[i].GetFreq(), ExtractedDateV[BestMatchInd].GetDate().CStr(), ExtractedDateV[BestMatchInd].GetDateEnd().CStr(), ExtractedDateV[BestMatchInd].GetFreq());
			ExtractedDateV[BestMatchInd].AddFreq(ExtractedDateV[i].GetFreq());
			ExtractedDateV.Del(i);
		}
		else
			i++;
	}
}

void TExtractedDateColl::SortByFreq(const bool& AscSort)
{
	ExtractedDateV.SortCmp(TCmpExtractedDate(AscSort));
}

TStr TExtractedDateColl::EncodeExtractedDates(const int& MaxCount) const
{
	if (ExtractedDateV.Len() == 0)
		return "";
	return TJsonVal::GetStrFromVal(EncodeExtractedDatesAsJson(MaxCount));
}

PJsonVal TExtractedDateColl::EncodeExtractedDatesAsJson(const int& MaxCount) const
{
	if (ExtractedDateV.Len() == 0)
		return TJsonVal::NewArr();
	// since we have const on the method we have to create a copy of the date coll
	// TODO: DEBUG!
	TExtractedDateColl Copy;
	Copy.AddColl(*this);
	Copy.SortByFreq(false);
	PJsonVal DateArr = TJsonVal::NewArr();
	int Count = (MaxCount == -1) ? Copy.ExtractedDateV.Len() : min(Copy.ExtractedDateV.Len(), MaxCount);
	for (int N=0; N < Count; N++) {
		TExtractedDate Date = Copy.ExtractedDateV[N];
		PJsonVal DateJson = Date.GetJson(true, true);
		DateArr->AddToArr(DateJson);
	}
	return DateArr;
}

TExtractedDateColl TExtractedDateColl::DecodeExtractedDates(const TStr& DatesStr)
{
	PJsonVal Json = TJsonVal::GetValFromStr(DatesStr);
	return TExtractedDateColl::DecodeExtractedDates(Json);
}

TExtractedDateColl TExtractedDateColl::DecodeExtractedDates(const PJsonVal& DatesJson)
{
	TExtractedDateColl DateColl;
	if (!DatesJson->IsArr())
		return DateColl;
	for (int N = 0; N < DatesJson->GetArrVals(); N++) {
		PJsonVal DateJson = DatesJson->GetArrVal(N);
		DateColl.AddExtractedDate(TExtractedDate(DateJson));
	}
	return DateColl;
}

// //////
// TDateExtractor
// //////
TDateExtractor::TDateExtractor(const PNotify& _NotifyInst, const PJsonVal& SettingsJson) : NotifyInst(_NotifyInst)
{
	// default is take the whole text
	TextLenToConsiderInExtraction = (int) SettingsJson->GetObjNum("TextLenToConsiderInExtraction", -1);
	MinValidYear = (int) SettingsJson->GetObjNum("MinValidYear", 1960);
	MaxValidYear = (int) SettingsJson->GetObjNum("MaxValidYear", 2030);
	ImputeYearIfMaxDaysApart = (int) SettingsJson->GetObjNum("ImputeYearIfMaxDaysApart", 30);
	SkipNumericDatesWithoutYear = SettingsJson->GetObjBool("SkipNumericDatesWithoutYear", true);	// skip dates such as 2.4. or 2/4 or 2-4 - too many false positives

	ConsoleNotify = TStdNotify::New();

	NotifyInfo("Created date extractor instance.");
}

void TDateExtractor::Clr()
{
	DateH.Clr();
	DateRangeH.Clr();
	MonthH.Clr();
}

// load regular expressions for a particular language
// the expressions should be in folder Path and should follow the naming scheme:
// single dates: Regex-date-LANG.txt
// date ranges: Regex-daterange-LANG.txt
// month name to number: Regex-months-LANG.txt

void TDateExtractor::LoadLangRegEx(const TStr& _Path, const TStr& _Lang, const TStr& CommentPrefix)
{
	TStr Lang = _Lang.GetLc();

	TStr Path = _Path.TrimRight("/").TrimRight("\\");
	TStr DateFName = TStr::Fmt("%s/Regex-date-%s.txt", Path.CStr(), Lang.CStr());
	if (TFile::Exists(DateFName)) {
		TStr LnStr; bool Foo;
		PSIn FIn = TFIn::New(DateFName, Foo, true);
		while (FIn->GetNextLn(LnStr)) {
			if (LnStr.IsWs() || LnStr.Len() == 0)
				continue;
			if (CommentPrefix != "" && LnStr.SearchStr(CommentPrefix, 0) == 0)
				continue;
			LnStr.ToTrunc();		// make sure we remove the possible ending newlines
			AddRegExDate(Lang, LnStr);
			NotifyInfo("Loaded %s date regular expression: %s", Lang.CStr(), LnStr.CStr());
		}
	}
	else
		ConsoleNotify->OnStatus("Failed to locate file " + DateFName);

	TStr RangeFName = TStr::Fmt("%s/Regex-daterange-%s.txt", Path.CStr(), Lang.CStr());
	if (TFile::Exists(RangeFName)) {
		TStr LnStr; bool Foo;
		PSIn FIn = TFIn::New(RangeFName, Foo, true);
		while (FIn->GetNextLn(LnStr)) {
			if (LnStr.IsWs() || LnStr.Len() == 0)
				continue;
			if (CommentPrefix != "" && LnStr.SearchStr(CommentPrefix, 0) == 0)
				continue;
			LnStr.ToTrunc();		// make sure we remove the possible ending newlines
			AddRegExDateRange(Lang, LnStr);
			NotifyInfo("Loaded %s date regular expression: %s", Lang.CStr(), LnStr.CStr());
		}
	}
	else
		ConsoleNotify->OnStatus("Failed to locate file " + RangeFName);

	TStr MonthFName = TStr::Fmt("%s/Regex-months-%s.txt", Path.CStr(), Lang.CStr());
	if (TFile::Exists(MonthFName)) {
		TStr LnStr; bool Foo;
		PSIn FIn = TFIn::New(MonthFName, Foo, true);
		THash<TStr, TInt> TextToMonthH;
		while (FIn->GetNextLn(LnStr)) {
			if (LnStr.IsWs() || LnStr.Len() == 0)
				continue;
			if (CommentPrefix != "" && LnStr.SearchStr(CommentPrefix, 0) == 0)
				continue;
			LnStr.ToTrunc();		// make sure we remove the possible ending newlines
			TStrV PartV; LnStr.SplitOnAllCh('|', PartV);
			
			
			if (PartV.Len() < 2 || !PartV[0].IsInt())
				continue;
			int Month = PartV[0].GetInt();
			for (int N = 1; N < PartV.Len(); N++)
				TextToMonthH.AddDat(PartV[N].GetLc(), Month);
		}
		MonthH.AddDat(Lang, TextToMonthH);
		NotifyInfo("Loaded %d month names for language %s", TextToMonthH.Len(), Lang.CStr());
	}
	else
		ConsoleNotify->OnStatus("Failed to locate file " + MonthFName);
}

void TDateExtractor::SetRegEx(const TStr& TextLang, const TStrV& DateRangeV, const TStrV DateV)
{
	Clr();
	for (int N=0; N < DateRangeV.Len(); N++) 
		AddRegExDateRange(TextLang, DateRangeV[N]);
	for (int N=0; N < DateV.Len(); N++) 
		AddRegExDate(TextLang, DateV[N]);
}

void TDateExtractor::AddRegExDate(const TStr& TextLang, const TStr& Date)
{
	if (!DateH.IsKey(TextLang))
		DateH.AddDat(TextLang, TStrV());
	DateH.GetDat(TextLang).Add(Date);
}

void TDateExtractor::AddRegExDateRange(const TStr& TextLang, const TStr& DateRange)
{
	if (!DateRangeH.IsKey(TextLang))
		DateRangeH.AddDat(TextLang, TStrV());
	DateRangeH.GetDat(TextLang).Add(DateRange);
}

// extract dates from Text
// text should be in a known language TextLang so that we know which regexes to use
// CurrentTime can be specified or not. if it is then we will for the dates with missing year try to impute the year to the closest one to CurrentTime
// year will be imputed only if the distance in days will be smaller than ImputeYearIfMaxDaysApart
void TDateExtractor::ExtractDates(const TStr& Text, const TStr& _TextLang, const TTm& CurrentTime, TExtractedDateColl& ExtractedDateColl)
{
	try {
		TStr SubText = (TextLenToConsiderInExtraction > 0) ? Text.GetSubStr(0, TextLenToConsiderInExtraction) : Text;
		TStr TextLang = _TextLang.GetLc();

		TVec<TIntPr> SeenMatchesV;
		const StringPiece SPText(SubText.CStr());
		RE2::Options ReOptions;
		ReOptions.set_case_sensitive(false);
		ReOptions.set_utf8(false);

		map<string, int> Groups;
		StringPiece Matches[10];

		// extract date ranges
		TStrV DateRangeV = DateRangeH.IsKey(TextLang) ? DateRangeH.GetDat(TextLang) : TStrV();
		for (int N = 0; N < DateRangeV.Len(); N++)
			AddDateRanges(SPText, CurrentTime, DateRangeV[N].CStr(), ReOptions, TextLang, ExtractedDateColl, SeenMatchesV);

		TStrV DateV = DateH.IsKey(TextLang) ? DateH.GetDat(TextLang) : TStrV();
		for (int N = 0; N < DateV.Len(); N++)
			AddIndividualDates(SPText, CurrentTime, DateV[N].CStr(), ReOptions, TextLang, ExtractedDateColl, SeenMatchesV);
	}
	catch (PExcept E) {
		NotifyErr("TDateExtractor::ExtractDates. Exception occured while extracting dates.\nException info: %s\nStack trace: %s", E->GetMsgStr().CStr(), E->GetLocStr().CStr());
	}
	catch (...) {
		NotifyErr("TDateExtractor::ExtractDates. General exception.");
	}
}


void TDateExtractor::AddIndividualDates(const StringPiece SPText, const TTm& CurrentTime, const char* RegEx, const RE2::Options& ReOptions, const TStr& TextLang, TExtractedDateColl& ExtractedDateColl, TVec<TIntPr>& SeenMatchesV)
{
	int PrintOffset = 20;
	
	StringPiece Matches[10];

	RE2 Re(RegEx, ReOptions);
	map<string, int> Groups = Re.NamedCapturingGroups();
	
	int off = 0;
	while(Re.Match(SPText, off, SPText.size(), RE2::UNANCHORED, Matches, 10)) {
		off = (int) (Matches[0].end() - SPText.begin());
		// make sure we haven't match this date already with some previous regex
		// regex expressions are ordered from most to least specific so we don't want to match multiple times cases such as 15 August 2013
		int MatchStart = (int) (Matches[0].begin() - SPText.begin());
		int MatchEnd = (int) (Matches[0].end() - SPText.begin());
		bool alreadySeen = false;
		for (int SeenN = 0; SeenN < SeenMatchesV.Len(); SeenN++) {
			if (SeenMatchesV[SeenN].Val1 <= MatchStart && MatchStart <= SeenMatchesV[SeenN].Val2) {
				NotifyInfo("skipping match '%s' since already seen", Matches[0].ToString().c_str());
				alreadySeen = true;
				break;
			}
		}
		if (alreadySeen)
			continue;

		int PrintOffset = 20;
		int StartPrint = max(0, MatchStart - PrintOffset);
		int EndPrint = min(SPText.length(), MatchEnd + PrintOffset);
		TStr PrintOrigContext = SPText.substr(StartPrint, EndPrint - StartPrint).ToString().c_str();
		PrintOrigContext.ChangeStrAll("\n", " ");

		TStr Day = Groups.count("day") > 0 ? Matches[Groups["day"]].ToString().c_str() : "";
		TStr Month = Groups.count("month") > 0 ? Matches[Groups["month"]].ToString().c_str() : "";
		TStr Year = Groups.count("year") > 0 ? Matches[Groups["year"]].ToString().c_str() : "";

		if (Day != "" && !Day.IsInt() && !Month.IsInt())
			continue;
		if (!Day.IsInt() && Month.IsInt()) { TStr Tmp = Day; Day = Month; Month = Tmp; }
		if (SkipNumericDatesWithoutYear && Day.IsInt() && Month.IsInt() && Year == "") {
			NotifyInfo("skipping date match '%s' due since with missing year and numeric day and month", PrintOrigContext.CStr());
			continue;
		}

		bool AmbiguousSep = true;
		if (Groups.count("separator1") > 0 && Groups.count("separator2") > 0) {
			TStr Sep1 = Matches[Groups["separator1"]].ToString().c_str();
			TStr Sep2 = Matches[Groups["separator2"]].ToString().c_str();
			if (Sep1 != Sep2) {
				NotifyInfo("skipping date match '%s' due to different separators", PrintOrigContext.CStr());
				continue;
			}
			// we assume that only dates with / as a separator produce ambiguous dates
			// if a different separator is used then we know what is day and month, even if both below 12
			if (Sep1 != "/")
				AmbiguousSep = false;
		}

		bool Ambiguous = false;
	
		// check if we have valid values for month and day and if we need to switch month and day due to their values
		if (Month.IsInt() && Day.IsInt() && AmbiguousSep) {
			int M = Month.GetInt(-1);
			int D = Day.GetInt(-1);
			Ambiguous = D <= 12 && M <= 12;
			if (M > 12 && D <= 12) { 
				int T = D; D = M; M = T; 
				TStr Tmp = Day; Day =Month; Month = Tmp;
			}
		}
		TStr OrigMonth = Month;	// remember the original versions of the month
		if (!Month.IsInt())
			Month = GetMonthAsInt(TextLang, Month);
	
		// we require at least two parts of the date to be valid
		if (Day == "" && Year == "")
			continue;
		// we treat as valid values only "" and range (MinValidYear, MaxValidYear)
		if (Year != "" && (!Year.IsInt() || Year.GetInt(0) < MinValidYear || Year.GetInt(0) > MaxValidYear)) {
			NotifyInfo("skipping invalid date match '%s'", Matches[0].ToString().c_str());
			continue;
		}
		if (Day != "" && (!Day.IsInt() || Day.GetInt(0) > 31 || Day.GetInt(0) < 1)) {
			NotifyInfo("skipping invalid date match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for day found - exit
		}
		if (!Month.IsInt() || Month.GetInt(0) < 1 || Month.GetInt(0) > 12) {
			NotifyInfo("skipping invalid date match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for month found - exit
		}

		TStr OriginalDate = Matches[0].ToString().c_str();
		if (Year.IsInt() && Year.Len() == 2 && (!OrigMonth.IsInt() || Year.GetInt() > 30))
			continue;
		
		if (Year.IsInt() && Year.Len() == 2)
			Year = "20" + Year;
		if (Day != "" && Day.Len() < 2)
			Day = "0" + Day;
		bool ImputedYear = (Year == "") ? TryImputingYear(CurrentTime, Day, Month, Year) : false;
		TStr NormalizedDate = TStr::Fmt("%s-%02d-%s", Year.CStr(), Month.GetInt(0), Day.CStr());
		NotifyInfo("found date: %s in text ...%s... (position: %d)", NormalizedDate.CStr(), PrintOrigContext.CStr(), MatchStart);
	
		ExtractedDateColl.AddExtractedDate(TExtractedDate(NormalizedDate, Ambiguous, ImputedYear, MatchStart, OriginalDate, PrintOrigContext));
		SeenMatchesV.Add(TIntPr(MatchStart, MatchEnd));
	}
}


void TDateExtractor::AddDateRanges(const StringPiece SPText, const TTm& CurrentTime, const char* RegEx, const RE2::Options& ReOptions, const TStr& TextLang, TExtractedDateColl& ExtractedDateColl, TVec<TIntPr>& SeenMatchesV)
{
	int PrintOffset = 20;
	
	StringPiece Matches[10];

	RE2 Re(RegEx, ReOptions);
	map<string, int> Groups = Re.NamedCapturingGroups();
	int off = 0;
	while(Re.Match(SPText, off, SPText.size(), RE2::UNANCHORED, Matches, 10)) {
		off = (int) (Matches[0].end() - SPText.begin());
		int MatchStart = (int) (Matches[0].begin() - SPText.begin());
		int MatchEnd = (int) (Matches[0].end() - SPText.begin());
		bool alreadySeen = false;
		for (int SeenN = 0; SeenN < SeenMatchesV.Len(); SeenN++) {
			if (SeenMatchesV[SeenN].Val1 <= MatchStart && MatchStart <= SeenMatchesV[SeenN].Val2) {
				//NotifyInfo("skipping match '%s' since already seen", Matches[0].ToString().c_str());
				alreadySeen = true;
				break;
			}
		}
		if (alreadySeen)
			continue;
		TStr Day1 = Groups.count("day1") > 0 ? Matches[Groups["day1"]].ToString().c_str() : "";
		TStr Day2 = Groups.count("day2") > 0 ? Matches[Groups["day2"]].ToString().c_str() : "";
		TStr Month1 = Groups.count("month1") > 0 ? Matches[Groups["month1"]].ToString().c_str() : "";
		TStr Month2 = Groups.count("month2") > 0 ? Matches[Groups["month2"]].ToString().c_str() : Month1;
		TStr Year1 = Groups.count("year1") > 0 ? Matches[Groups["year1"]].ToString().c_str() : "";
		TStr Year2 = Groups.count("year2") > 0 ? Matches[Groups["year2"]].ToString().c_str() : "";
		if (Year1 == "")
			Year1 = Year2;
		if (Year2 == "")
			Year2 = Year1;
	
		// we require at least two parts of the date to be valid
		if ( (Day1 == "" && Year1 == "") || (Day2 == "" && Year2 == "") )
			continue;
		if (Day1 != "" && !Day1.IsInt() && !Month1.IsInt()) 
			continue;
		if (Day2 != "" && !Day2.IsInt() && !Month2.IsInt()) 
			continue;
		TStr OrigMonth1 = Month1;	// remember the original versions of the month
		TStr OrigMonth2 = Month2;
		if (!Month1.IsInt())
			Month1 = GetMonthAsInt(TextLang, Month1);
		if (!Month2.IsInt())
			Month2 = GetMonthAsInt(TextLang, Month2);
	
		// we treat as valid values only "" and range (MinValidYear, MaxValidYear)
		if (Year1 != "" && (!Year1.IsInt() || Year1.GetInt(0) < MinValidYear || Year1.GetInt(0) > MaxValidYear)) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;
		}
		if (Year2 != "" && (!Year2.IsInt() || Year2.GetInt(0) < MinValidYear || Year2.GetInt(0) > MaxValidYear)) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;
		}
		if (Day1 != "" && (!Day1.IsInt() || Day1.GetInt(0) > 31 || Day1.GetInt(0) < 1) ) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for day found - exit
		}
		if (Day2 != "" && (!Day2.IsInt() || Day2.GetInt(0) > 31 || Day2.GetInt(0) < 1) ) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for day found - exit
		}
		if (!Month1.IsInt() || Month1.GetInt(0) < 1 || Month1.GetInt(0) > 12) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for month found - exit
		}
		if (!Month2.IsInt() || Month2.GetInt(0) < 1 || Month2.GetInt(0) > 12) {
			NotifyInfo("skipping invalid date range match '%s'", Matches[0].ToString().c_str());
			continue;		// invalid value for month found - exit
		}

		int StartPrint = max(0, MatchStart-PrintOffset);
		int EndPrint = min(SPText.length(), MatchEnd + PrintOffset);
		TStr PrintOrigContext = SPText.substr(StartPrint, EndPrint - StartPrint).ToString().c_str();
		PrintOrigContext.ChangeStrAll("\n", " ");
		TStr OriginalDate = Matches[0].ToString().c_str();
		
		// year specified with only two digits - then month has to be as a digit too (invalid 20 May 02)
		// also invalid are dates such as 20-03-40 (since a reference to 2040 in a short form is very unlikely)
		if (Year1.IsInt() && Year1.Len() == 2 && (!OrigMonth1.IsInt() || Year1.GetInt() > 30))
			continue;
		if (Year2.IsInt() && Year2.Len() == 2 && (!OrigMonth2.IsInt() || Year2.GetInt() > 30))
			continue;

		if (Year1.IsInt() && Year1.Len() == 2)
			Year1 = "20" + Year1;
		if (Year2.IsInt() && Year2.Len() == 2) 
			Year2 = "20" + Year2;
		if (Day1 != "" && Day1.Len() < 2)
			Day1 = "0" + Day1;
		if (Day2 != "" && Day2.Len() < 2)
			Day2 = "0" + Day2;
		
		bool Imp1 = (Year1 == "") ? TryImputingYear(CurrentTime, Day1, Month1, Year1) : false;
		bool Imp2 = (Year2 == "") ? TryImputingYear(CurrentTime, Day2, Month2, Year2) : false;

		TStr NormalizedDateStart = TStr::Fmt("%s-%02d-%0s", Year1.CStr(), Month1.GetInt(0), Day1.CStr());
		TStr NormalizedDateEnd = TStr::Fmt("%s-%02d-%0s", Year2.CStr(), Month2.GetInt(0), Day2.CStr());
		NotifyInfo("found date range: (%s, %s) in text ...%s... (position: %d)", NormalizedDateStart.CStr(), NormalizedDateEnd.CStr(), PrintOrigContext.CStr(), MatchStart);
		ExtractedDateColl.AddExtractedDate(TExtractedDate(NormalizedDateStart, NormalizedDateEnd, false, Imp1 || Imp2, MatchStart, OriginalDate, PrintOrigContext));
		SeenMatchesV.Add(TIntPr(MatchStart, MatchEnd));
	}
}

TStr TDateExtractor::GetMonthAsInt(const TStr& TextLang, const TStr& _Text)
{
	if (_Text == "")
		return "";
	const TStr Text = _Text.GetLc();
	if (!MonthH.IsKey(TextLang))
		return "";
	
	THash<TStr, TInt> TextToMonthH = MonthH.GetDat(TextLang);

	if (TextToMonthH.IsKey(Text))
		return TextToMonthH.GetDat(Text).GetStr();
	
	NotifyInfo("Unable to determine the month for '%s' in language %s", Text.CStr(), TextLang.CStr());
	return "";
}

bool TDateExtractor::TryImputingYear(const TTm& CurrentTime, const TStr& Day, const TStr& Month, TStr& Year)
{
	if (CurrentTime.IsDef() == false)
		return false;
	TTm PossibleTime1(CurrentTime.GetYear()-1, Month.GetInt(-1), Day != "" ? Day.GetInt(-1) : 1);
	TTm PossibleTime2(CurrentTime.GetYear(), Month.GetInt(-1), Day != "" ? Day.GetInt(-1) : 1);
	TTm PossibleTime3(CurrentTime.GetYear()+1, Month.GetInt(-1), Day != "" ? Day.GetInt(-1) : 1);
	if (!PossibleTime1.IsDef() || !PossibleTime2.IsDef() || !PossibleTime3.IsDef())
		return false;
	uint64 DiffDays1 = TSysTm::IsValidDate(PossibleTime1) ? TTm::GetDiffDays(CurrentTime, PossibleTime1) : (uint64) TUInt64::Mx;
	uint64 DiffDays2 = TSysTm::IsValidDate(PossibleTime2) ? TTm::GetDiffDays(CurrentTime, PossibleTime2) : (uint64) TUInt64::Mx;
	uint64 DiffDays3 = TSysTm::IsValidDate(PossibleTime3) ? TTm::GetDiffDays(CurrentTime, PossibleTime3) : (uint64) TUInt64::Mx;
	if (DiffDays1 < ImputeYearIfMaxDaysApart) {
		Year = TInt(PossibleTime1.GetYear()).GetStr();
		return true;
	}
	else if (DiffDays2 < ImputeYearIfMaxDaysApart) {
		Year = TInt(PossibleTime2.GetYear()).GetStr();
		return true;
	}
	if (DiffDays3 < ImputeYearIfMaxDaysApart) {
		Year = TInt(PossibleTime3.GetYear()).GetStr();
		return true;
	}
	return false;
}

void TDateExtractor::Test()
{
	TStr EngText =  "Jan. 11-15 | Feb 4. - 12. | "
					"Jan. 11- feb 15 | May 4 - Jun 12. | "
					"3/14 | 3/14/2011 | 3/14/11 | 03/14/11 | 14/03/2013 | 14/03/11 | 03/14/11 | "
					"14. 1. | 14.3. | 13.2. | "
					"2013-3-14 | 2011/3/14 | "
					"22nd September ,2006 | 22nd September 2006 | 22 September 2006 | 2nd February 2007 |  18 October | 18 October, 2009 | 14 March 2011 | 18th November, 2009 (OR 18 October, 2009) | 4th October, 2005 (OR 4 October, 2005) | "
					"March 14, 2011 |  March 14th | March 14 | September 22 2006 (without mma) | August 15, 1990. | September 22, 2006 | September 22nd, 2006 | September 22, 2006 | March 14th, 2011 | April-1 | "
					"Jan. 2012 | Feb 2013 | "
				"March the Fourteenth | A the Fourteenth of March, 2011	March the Fourteenth, 2011 | B	14th March 2011 | "
				"C	 | D	14/3/2011	3/14/2011 | E	14/3/11	3/14/11 | "
				" |  30.3.2005 | 19/2/2002 | 30-2-1999 | 26 May 1979  | 30 October 2003 | 18 October, 2003 | 29 January, 2008 | My son was born on 18 October, 2003. April-1";
	TExtractedDateColl Dates;
	ExtractDates(EngText, "eng", TTm(), Dates);
}

void TDateExtractor::RETest()
{
	int i;
	std::string s;
	bool ok = RE2::FullMatch("ruby:1234", "(\\w+):(\\d+)", &s, &i);
	ok = (s == "ruby");
	ok = (i == 1234);

	RE2 re("(?P<N1>is) (?P<N2>\\d+)|(?P<N3>car)|(object.*movie)");
    re2::StringPiece sp = "car is 12 the animation creation with the object as cars, its a very fine movie";
    int off = 0;
    re2::StringPiece matches[5];
    while(re.Match(sp, off, sp.size(), RE2::UNANCHORED, matches, 5)) {
		map<string, int> map = re.NamedCapturingGroups();
		int ind = map["N3"];
		re2::StringPiece carMatch = matches[ind];
		re2::StringPiece wholeMatch = matches[0];
        printf("match at %d,%d: %s\n", (int)(wholeMatch.begin() - sp.begin()), (int)(wholeMatch.end() - sp.begin()), wholeMatch.as_string().c_str());
        off = (int) (wholeMatch.end() - sp.begin());
    }

	//RE2::Arg Month, Day;
	std::string Month, Day, X;
	//RE2::Arg Month, Day, X;
	ok = RE2::PartialMatch("test March 14 20", "(Jan|(March|April)) (\\d+)(th)?", &Month, &Day, &X);

	// Fails: "ruby" cannot be parsed as an integer.
	ok = !RE2::FullMatch("ruby", "(.+)", &i);

	// Success; does not extract the number.
	ok = RE2::FullMatch("ruby:1234", "(\\w+):(\\d+)", &s);

	// Success; skips NULL argument.
	ok = RE2::FullMatch("ruby:1234", "(\\w+):(\\d+)", (void*)NULL, &i);

	// Fails: integer overflow keeps value from being stored in i.
	ok = !RE2::FullMatch("ruby:123456789123", "(\\w+):(\\d+)", &s, &i);

	//std::string s ("Bojan Jan., 10th has a submarine as a subsequence");
	//std::smatch m;
	//std::regex monthsLong("(\\b)(january|february|march|april|may|june|july|august|september|october|november|december)(\\b)",  std::regex_constants::ECMAScript | std::regex_constants::icase );
	//std::regex monthsShort("(\\b)(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[.]?(\\b)",  std::regex_constants::ECMAScript | std::regex_constants::icase );
	//std::regex months("(\\b)((january|february|march|april|may|june|july|august|september|october|november|december)|(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)[.]?)(\\b)",  std::regex_constants::ECMAScript | std::regex_constants::icase );

	//std::cout << "Target sequence: " << s << std::endl;
	//std::cout << "The following matches and submatches were found:" << std::endl;

	//while (std::regex_search (s, m, months)) {
	//	int len = m.length();
	//	std::string x = m[1];
	//	/*for (auto x:m) 
	//		std::cout << x << " ";*/
	//	std::cout << std::endl;
	//	s = m.suffix().str();
	//}
}

// notifications
void TDateExtractor::NotifyInfo(const char *FmtStr, ...) 
{
	va_list valist; va_start(valist, FmtStr);	
	Notify(ntInfo, FmtStr, valist); va_end(valist); 
}
void TDateExtractor::NotifyErr(const char *FmtStr, ...) 
{
	va_list valist; va_start(valist, FmtStr);
	Notify(ntErr, FmtStr, valist); va_end(valist); 
}

void TDateExtractor::Notify(const TNotifyType& Type, const char *FmtStr, va_list argptr) 
{
	const int RetVal=vsnprintf(NotifyBuf, 10*1024-2, FmtStr, argptr);
	if (RetVal < 0) return;
	if (! NotifyInst.Empty())
		NotifyInst->OnNotify(Type, NotifyBuf); 
}