#ifndef DATEEXTRACTOR_H
#define DATEEXTRACTOR_H

#include "base.h"
#include "re2/re2.h"

using std::string;
using std::map;
using re2::StringPiece;

//typedef TTriple<TStr, TStr, TBool> TExtractedDate;		// contains the detected string, the normalized string in the form (YYYY)?-MM-DD, ambiguity status (is it possible that DD-MM is switched with MM-DD)

class TExtractedDate 
{
private:
	bool BIsDateRange;
	TStr Date;
	TStr DateEnd;
	bool Ambiguous;
	bool ImputedYear;
	int Freq;
	int PosInText;
	TStr DetectedDate;	// date in the detected form
	TStr TextSnippet;	// date and the surrounding text

public:
	
	TExtractedDate() { 
		BIsDateRange = false;
		Date = "";
		DateEnd = "";
		Ambiguous = false;
		ImputedYear = false;
		Freq = 0;
		PosInText = -1;
		DetectedDate = "";
		TextSnippet = "";
	}

	TExtractedDate(const TStr& _Date, const bool& _Ambiguous, const bool& _ImputedYear, const int& _PosInText, const TStr& _DetectedDate, const TStr& _TextSnippet)
	{
		Date = _Date;
		Ambiguous = _Ambiguous;
		ImputedYear = _ImputedYear;
		BIsDateRange = false;
		Freq = 1;
		PosInText = _PosInText;
		DetectedDate = _DetectedDate;
		TextSnippet = _TextSnippet;
	}

	TExtractedDate(const TStr& _DateStart, const TStr& _DateEnd, const bool& _Ambiguous, const bool& _ImputedYear, const int& _PosInText, const TStr& _DetectedDate, const TStr& _TextSnippet)
	{
		Date = _DateStart;
		DateEnd = _DateEnd;
		Ambiguous = _Ambiguous;
		ImputedYear = _ImputedYear;
		BIsDateRange = true;
		Freq = 1;
		PosInText = _PosInText;
		DetectedDate = _DetectedDate;
		TextSnippet = _TextSnippet;
	}

	TExtractedDate(PJsonVal DateJson);

	TStr GetDate() const { return Date; }
	TStr GetDateEnd() const { return DateEnd; }
	TStr GetTextSnippet() const { return TextSnippet; }
	int GetPosInText() const { return PosInText; }

	bool operator==(const TExtractedDate& EDate) const { return (Date == EDate.Date) && (DateEnd == EDate.DateEnd); }

	// does the date have all the parts - year, month, day
	bool IsDateComplete() const { return Date.Len() > 0 && Date.GetCh(0) != '-' && Date.LastCh() != '-'; }
	bool DateHasValidYear() const { return Date.Len() > 0 && Date.GetCh(0) != '-'; }
	bool DateHasValidDay() const { return Date.Len() > 0 && Date.LastCh() != '-'; }
	
	bool IsDateEndComplete() const { return DateEnd.Len() > 0 && DateEnd.GetCh(0) != '-' && DateEnd.LastCh() != '-'; }
	bool DateEndHasValidYear() const { return DateEnd.Len() > 0 && DateEnd.GetCh(0) != '-'; }
	bool DateEndHasValidDay() const { return DateEnd.Len() > 0 && DateEnd.LastCh() != '-'; }

	// did we detect date  in form that can be interpreted in two ways (e.g. 2/1)
	bool IsDateAmbiguous() const { return Ambiguous; }
	bool IsImputedYear() const { return ImputedYear; }
	bool IsDateRange() const { return BIsDateRange; }
	int GetFreq() const { return Freq; }

	void IncreaseFreq(const int& _Freq = 1) { Freq += _Freq; }
	void SetFreq(const int _Freq) { Freq = _Freq; }
	void AddFreq(const int _Freq) { Freq += _Freq; }
	
	TStr GetStr() { return Date + (BIsDateRange ? " - " + DateEnd : ""); }
	
	PJsonVal GetJson(const bool& IncludeDetectedDate = false, const bool& IncludeTextSnippet = false);
};

class TExtractedDateColl
{
public:
	TVec<TExtractedDate> ExtractedDateV;

	TExtractedDateColl() { }
	TExtractedDateColl(const TStr& EncodedDateColl);

	void AddExtractedDate(const TExtractedDate& Date);
	void AddColl(const TExtractedDateColl& Coll);

	void SortByFreq(const bool& AscSort=true);
	void ImputeMissingYears(const int& MinFreq, PNotify Notify = NULL);

	//void DecodeExtractedDates(const TStr& EncodedDateColl);
	TStr EncodeExtractedDates(const int& MaxCount = -1) const;
	PJsonVal EncodeExtractedDatesAsJson(const int& MaxCount = -1) const;
	
	int Len() const { return ExtractedDateV.Len(); } ;
	TExtractedDate GetExtractedDate(const int& DateN) const { return ExtractedDateV[DateN]; }
	bool GetExtractedDate(const TStr& Date, TExtractedDate& ExtractedDate);
	int GetExtractedDatePos(const TExtractedDate& ExtractedDate);
	bool GetMostFreqDate(TExtractedDate& ExtractedDate, const int& MinFreq = 0, const bool& IgnoreIncomplete = true);
	
	void FilterDatesByPosInText(const int& MinPos, const int& MaxPos);

	static TExtractedDateColl DecodeExtractedDates(const TStr& DatesStr);
	static TExtractedDateColl DecodeExtractedDates(const PJsonVal& DatesJson);
};

//////////////////////////////////////
ClassTP(TDateExtractor, PDateExtractor)//{
typedef TPt<TDateExtractor> PDateExtractor;

private:

	THash<TStr, TStrV> DateRangeH;
	THash<TStr, TStrV> DateH;
	THash<TStr, THash<TStr, TInt> > MonthH;		// lang -> { jan: 1, feb: 2, ... }
	PNotify NotifyInst;
	char NotifyBuf[10*1024];

	PNotify ConsoleNotify;

	void NotifyInfo(const char *FmtStr, ...);
	void NotifyErr(const char *FmtStr, ...);
	void Notify(const TNotifyType& Type, const char *FmtStr, va_list argptr);

	int TextLenToConsiderInExtraction;
	int MinValidYear;
	int MaxValidYear;
	int ImputeYearIfMaxDaysApart;
	bool SkipNumericDatesWithoutYear;

	TDateExtractor(const PNotify& _NotifyInst, const PJsonVal& SettingsJson);

public:
	//enum ETextLang { LangEng, LangSpa, LangDeu, LangZho, LangSlv };

	static PDateExtractor New(const PNotify& NotifyInst, const PJsonVal& SettingsJson) {
		return new TDateExtractor(NotifyInst, SettingsJson);
	}

	void Clr();

	void LoadLangRegEx(const TStr& Path, const TStr& Lang, const TStr& CommentPrefix = "#");

	void SetRegEx(const TStr& TextLang, const TStrV& DateRangeV, const TStrV DateV);
	void AddRegExDate(const TStr& TextLang, const TStr& Date);
	void AddRegExDateRange(const TStr& TextLang, const TStr& DateRange);
	
	//void ExtractDates(const TStr& Text, const ETextLang& TextLang, TExtractedDateColl& ExtractedDateColl);
	void ExtractDates(const TStr& Text, const TStr& TextLang, const TTm& CurrentTime, TExtractedDateColl& ExtractedDateColl);
		
	void AddIndividualDates(const StringPiece SPText, const TTm& CurrentTime, const char* RegEx, const RE2::Options& ReOptions, const TStr& TextLang, TExtractedDateColl& ExtractedDateColl, TVec<TIntPr>& SeenMatchesV);
	void AddDateRanges(const StringPiece SPText, const TTm& CurrentTime, const char* RegEx, const RE2::Options& ReOptions, const TStr& TextLang, TExtractedDateColl& ExtractedDateColl, TVec<TIntPr>& SeenMatchesV);
	
	TStr GetMonthAsInt(const TStr& TextLang, const TStr& Text);
		
	bool TryImputingYear(const TTm& CurrentTime, const TStr& Day, const TStr& Month, TStr& Year);

	void Test();
	void RETest();

	static TStr EncodeExtractedDates(const TExtractedDateColl& ExtractedDates);
	static void DecodeExtractedDates(const TStr& EncodedDates, TExtractedDateColl& ExtractedDates);
};

#endif