/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "IterativeKMeans.h"

// SampleWhenDocsAboveCnt - if more than this number of docs are in Bow then take a sample of SampleWhenDocsAboveCnt docs and cluster only those. The remaining docs are simply assigned to the closest cluster centroid
// K - the number of clusters to create on each step
// ClustTrial - number of times we iterate in k-means
// MinClusterSizeToSplit - split a cluster only if more than this number of docs in current node
// MaxDepth - max times that we split clusters into subclusters
IterativeKMeans::IterativeKMeans(const PBowDocBs& _BowDocBs, const PBowDocWgtBs& _BowDocWgtBs, const int& SampleWhenDocsAboveCnt, const int& _K, const int& _ClustTrials, const int& _MinClusterSizeToSplit, const int& MaxDepth, const TBowSimType SimType)
	: BowDocBs(_BowDocBs), BowDocWgtBs(_BowDocWgtBs), K(_K), ClustTrials(_ClustTrials), MinClusterSizeToSplit(_MinClusterSizeToSplit)
{
	BowSim = TBowSim::New(SimType);
	Notify = TStdNotify::New();

	TClusterInfo Info;
	TIntV DIdV;
	BowDocWgtBs->GetDIdV(DIdV);

	TIntV RemainingDIdV;
	if (DIdV.Len() > SampleWhenDocsAboveCnt) {
		TRnd Rnd(1); DIdV.Shuffle(Rnd);
		DIdV.GetSubValV(SampleWhenDocsAboveCnt, DIdV.Len(), RemainingDIdV);
		DIdV.Trunc(SampleWhenDocsAboveCnt);
		AssertR(DIdV.Len() + RemainingDIdV.Len() == BowDocWgtBs->GetDocs(), "we skipped some docs");
	}

	Info.DIdV = DIdV;
	// put the sample docs into clusters and then cluster also the RemainingDIdV
	SplitDocsIntoClusters(MaxDepth, Info, RemainingDIdV);
	Info.DIdV.AddV(RemainingDIdV);		// assign all remaining documents back to the root so that we compute the correct count of events in the root
	printf("Finished KMeans.\n");
	
	// the similarities are not computed for the root note so we estimate it by adding the document similarities from the children clusters
	for (int k = 0; k < Info.SubClusterV.Len(); k++) {
		Info.DIdSimV.AddV(Info.SubClusterV[k].DIdSimV);
	}
	Info.DIdSimV.SortCmp(TCmpKeyDatByDat<TInt, TFlt>(false));

	// remember the cluster info
	ClusterInfo = Info;
}

// perform split based on the current set of documents in Info.DIdV
// RemainingDIdV documents will be assigned after k-means to the clostest clusters found
void IterativeKMeans::SplitDocsIntoClusters(const int& MaxDepth, TClusterInfo& Info, const TIntV& RemainingDIdV)
{
	TBowMatrix DocMtx(BowDocWgtBs, Info.DIdV);
	TBowKMeans KMeans(&DocMtx, K, ClustTrials, Notify, TRnd(1));
	KMeans.Init();
	KMeans.Apply();
	
	// compute cluster assignments and their similarities to the centroid
	Info.SubClusterV = TVec<TClusterInfo>(K);
	TVec<TIntV> DIdVV(K); TVec<TIntFltKdV> DIdSimKdVV(K);
	const TIntV& Assignment = KMeans.GetAssignment();
	IAssert(Assignment.Len() == Info.DIdV.Len());
	for (int N = 0; N < KMeans.GetDocs(); N++) {
		Info.SubClusterV[Assignment[N]].DIdV.Add(Info.DIdV[N]);
		/*double Sim = KMeans.GetDocClusterSim(Assignment[N], Info.DIdV[N]);*/
		double Sim = KMeans.GetDocClusterSim(Assignment[N], N);
		Info.SubClusterV[Assignment[N]].DIdSimV.Add(TIntFltKd(Info.DIdV[N], Sim));
	}
	
	// find to which clusters should the documents in RemainingDIdV go to
	// add these documents into the appropriate clusters
	TBowMatrix RemDocMtx(BowDocWgtBs, RemainingDIdV);
	TVec<TFltV> ClusterToSimVV(K);
	for (int k = 0; k < K; k++) {
		TFltV SimV(RemainingDIdV.Len()); RemDocMtx.MultiplyT(KMeans.GetCentroid(k), SimV);
		ClusterToSimVV[k] = SimV;
	}
	TVec<TIntV> RemDIdVV(K);			// to which of the clusters should each of the remaining documents be assigned
	TVec<TIntFltKdV> RemDIdSimKdV(K);	// for the document and the similarity to the cluster for each document
	for (int N = 0; N < RemainingDIdV.Len(); N++) {
		double MaxSim = ClusterToSimVV[0][N]; int MaxInd = 0;
		for (int k = 1; k < K; k++) {
			if (ClusterToSimVV[k][N] > MaxSim) { MaxSim = ClusterToSimVV[k][N]; MaxInd = k; }
		}
		RemDIdVV[MaxInd].Add(RemainingDIdV[N]);
		RemDIdSimKdV[MaxInd].Add(TIntFltKd(RemainingDIdV[N], MaxSim));
	}

	// find subclusters for each of the K clusters
	for (int k = 0; k < K; k++) {
		// take the created cluster and split it again if conditions are met
		if (Info.SubClusterV[k].DIdV.Len() > MinClusterSizeToSplit && MaxDepth > 0) {
			SplitDocsIntoClusters(MaxDepth - 1, Info.SubClusterV[k], RemDIdVV[k]);
		}
		// add to the subcluster k the remaining documents that we aproximated to be in the cluster
		Info.SubClusterV[k].DIdV.AddV(RemDIdVV[k]);
		Info.SubClusterV[k].DIdSimV.AddV(RemDIdSimKdV[k]);
		Info.SubClusterV[k].DIdSimV.SortCmp(TCmpKeyDatByDat<TInt, TFlt>(false));
	}
}

// compute keywords for each (sub)cluster
void IterativeKMeans::ComputeKeywords(const int& MxKeywords, const int& EventsUsedForKeywords)
{
	ComputeKeywords(MxKeywords, EventsUsedForKeywords, ClusterInfo);
}

// compute keywords for each of the clusters
void IterativeKMeans::ComputeKeywords(const int& MxKeywords, const int& EventsUsedForKeywords, TClusterInfo& Info)
{
	// take a sample if the DocV is too big
	TIntV DIdVSample; Info.DIdV.GetSubValV(0, EventsUsedForKeywords, DIdVSample);
	PBowSpV ClustSpV = TBowClust::GetConceptSpV(BowDocWgtBs, BowSim, DIdVSample);
	Info.KWordSet = ClustSpV->GetKWordSet(BowDocBs)->GetTopKWords(MxKeywords, 1.0);

	for (int N = 0; N < Info.SubClusterV.Len(); N++)
		ComputeKeywords(MxKeywords, EventsUsedForKeywords, Info.SubClusterV[N]);
}

// find all keywords in all subclusters in Info
TStrSet TClusterInfo::GetAllKeywords() const
{
	TStrSet KWords;
	if (!KWordSet.Empty()) {
		for (int N = 0; N < KWordSet->GetKWords(); N++)
			KWords.AddKey(KWordSet->GetKWordStr(N));
	}
	for (int N = 0; N < SubClusterV.Len(); N++) {
		TStrSet KW = SubClusterV[N].GetAllKeywords();
		TStrV KWV; KW.GetKeyV(KWV);
		KWords.AddKeyV(KWV);
	}
	return KWords;
}