/**
 * Author: Gregor Leban <gleban@gmail.com>
 * Copyright (c) 2015, Jozef Stefan Institute, Quintelligence d.o.o. and contributors
 * All rights reserved.
 *
 * This source code is licensed under the FreeBSD license found in the
 * LICENSE file in the root directory of this source tree.
 */
#include "ValidatedStoryLinks.h"


TValidatedStoryLinks::TValidatedStoryLinks(const TStr& _FName)
{
	FName = _FName;
}

TValidatedStoryLinks::~TValidatedStoryLinks()
{
	Save();
}


void TValidatedStoryLinks::Load()
{
	ValidatedDiffEventH.Clr();
	ValidatedSameEventH.Clr();
	try {
		if (TFile::Exists(FName)) {
			TFIn FIn(FName);
			ValidatedSameEventH.Load(FIn);
			ValidatedDiffEventH.Load(FIn);
		}
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("TValidatedStoryLinks::Load. Error while saving ValidatedStoryLinks.bin.");
	}
}

void TValidatedStoryLinks::Save()
{
	try {
		TFOut FOut(FName);
		ValidatedSameEventH.Save(FOut);
		ValidatedDiffEventH.Save(FOut);
	}
	catch (...) {
		TNotify::StdNotify->OnStatus("TValidatedStoryLinks::Save. Error while saving ValidatedStoryLinks.bin.");
	}
}

// save information about all same and different stories into a json file
void TValidatedStoryLinks::SaveToJson(const TStr& OutFNm)
{
	TFOut FOut(OutFNm);
	for (int KeyId = ValidatedSameEventH.FFirstKeyId(); ValidatedSameEventH.FNextKeyId(KeyId);) {
		const TUInt64 Id = ValidatedSameEventH.GetKey(KeyId);
		TUInt64V IdV = ValidatedSameEventH[KeyId];
		PJsonVal Json = TJsonVal::NewObj("id", Id.GetStr());
		TStrV IdStrV;
		for (int N = 0; N < IdV.Len(); N++)
			IdStrV.Add(IdV[N].GetStr());
		Json->AddToObj("sameIds", TJsonVal::NewArr(IdStrV));
		FOut.PutStrLn(TJsonVal::GetStrFromVal(Json));
	}
	for (int KeyId = ValidatedDiffEventH.FFirstKeyId(); ValidatedDiffEventH.FNextKeyId(KeyId);) {
		const TUInt64 Id = ValidatedDiffEventH.GetKey(KeyId);
		TUInt64V IdV = ValidatedDiffEventH[KeyId];
		PJsonVal Json = TJsonVal::NewObj("id", Id.GetStr());
		TStrV IdStrV;
		for (int N = 0; N < IdV.Len(); N++)
			IdStrV.Add(IdV[N].GetStr());
		Json->AddToObj("diffIds", TJsonVal::NewArr(IdStrV));
		FOut.PutStrLn(TJsonVal::GetStrFromVal(Json));
	}
}

// load the same and diff hash tables from json - used when we do data restoration
void TValidatedStoryLinks::LoadFromJson(const TStr& InFNm)
{
	if (TFile::Exists(InFNm)) {
		ValidatedDiffEventH.Clr();
		ValidatedSameEventH.Clr();
		TFIn FIn(InFNm);
		TChA LnA;
		while (FIn.GetNextLn(LnA)) {
			PJsonVal Json = TJsonVal::GetValFromStr(LnA);
			const uint64 Id = Json->GetObjStr("id").GetUInt64(TUInt64::Mx);
			TStrV IdStrV;
			if (Json->IsObjKey("sameIds")) {
				Json->GetObjStrV("sameIds", IdStrV);
				for (int N = 0; N < IdStrV.Len(); N++)
					AddCuratedStoryMergeInfo(Id, IdStrV[N].GetUInt64(TUInt64::Mx), true);
			}
			else if (Json->IsObjKey("diffIds")) {
				Json->GetObjStrV("diffIds", IdStrV);
				for (int N = 0; N < IdStrV.Len(); N++)
					AddCuratedStoryMergeInfo(Id, IdStrV[N].GetUInt64(TUInt64::Mx), false);
			}
		}
	}
	else 
		TNotify::StdNotify->OnStatusFmt("TValidatedStoryLinks: Unable to laod json file: %s. File does not exist.", InFNm.CStr());
}

// when the user provides feedback if two stories should be merged or not, we save it so that event updates don't break recreate mistakes again
void TValidatedStoryLinks::AddCuratedStoryMergeInfo(const uint64& StoryId1, const uint64& StoryId2, const bool& ShouldMerge)
{
	if (StoryId1 == TUInt64::Mx || StoryId2 == TUInt64::Mx) {
		TNotify::StdNotify->OnStatusFmt("Adding a curated story merge info with invalid story id (id1: %I64u, id2: %I64u)", StoryId1, StoryId2);
		return;
	}
	// first remove any existing information about the story pairs
	if (ValidatedSameEventH.IsKey(StoryId1))
		ValidatedSameEventH.GetDat(StoryId1).DelIfIn(StoryId2);
	if (ValidatedSameEventH.IsKey(StoryId2))
		ValidatedSameEventH.GetDat(StoryId2).DelIfIn(StoryId1);
	if (ValidatedDiffEventH.IsKey(StoryId1))
		ValidatedDiffEventH.GetDat(StoryId1).DelIfIn(StoryId2);
	if (ValidatedDiffEventH.IsKey(StoryId2))
		ValidatedDiffEventH.GetDat(StoryId2).DelIfIn(StoryId1);
	// set new info
	if (ShouldMerge) {
		ValidatedSameEventH.AddDat(StoryId1).AddUnique(StoryId2);
		ValidatedSameEventH.AddDat(StoryId2).AddUnique(StoryId1);
	}
	else {
		ValidatedDiffEventH.AddDat(StoryId1).AddUnique(StoryId2);
		ValidatedDiffEventH.AddDat(StoryId2).AddUnique(StoryId1);
	}
}

TUInt64V TValidatedStoryLinks::GetValidatedDiffEvents(const uint64& StoryId)
{
	if (!ValidatedDiffEventH.IsKey(StoryId))
		return TUInt64V();
	return ValidatedDiffEventH.GetDat(StoryId);
}

TUInt64V TValidatedStoryLinks::GetValidatedSameEvents(const uint64& StoryId)
{
	if (!ValidatedSameEventH.IsKey(StoryId))
		return TUInt64V();
	return ValidatedSameEventH.GetDat(StoryId);
}